package Liszt
import Liszt.Trees._
import Liszt.Types._

object ASTUtil {
	private object Log extends Liszt.Log("ASTUtil")
	def isMeshT(a : Type) = a match {
		case t : Topo => true
		case Set(_) => true
		case _ => false
	}
	def isMeshElement(a : Type) = a match {
		case Vertex => true
		case Edge => true
		case Face => true
		case Cell => true
		case _ => false
	}
	def isVec(a : Type) = a match {
		case VectorT(_,_) => true
		case _ => false
	}
        def isMat(a : Type) = a match {
                case MatrixT(_,_,_) => true
                case _ => false
        }
	def isFieldT(a : Type) = a match {
		case FieldT(_,_) => true
		case _ => false
	}
	def isSetT(a : Type) = a match {
		case _ : Set => true
		case _ => false
	}
	def isSparseMatrix(a : Type) = a match {
		case _ : SparseMatrixType => true
		case _ => false
	}
	def isSparseVector(a : Type) = a match {
		case _ : SparseVectorType => true
		case _ => false
	}
        def isSimdT(a : Type) = a match {
          case SimdT(_) => true
          case _ => false
        }
        def isSimdVec(a : Type) = a match {
          case SimdT(VectorT(_,_)) => true
          case _ => false
        }
        def isSimdMat(a : Type) = a match {
          case SimdT(MatrixT(_, _, _)) => true
          case _ => false
        }
        def isSimdScalar(a : Type) = a match {
          case SimdT(FloatT) => true;
          case SimdT(DoubleT) => true;
          case SimdT(IntT) => true;
          case SimdT(BoolT) => true;
          case _ => false;
        }
        def isSingleStatment(b : Block) = b.stmts.length == 1;

	object Names {
		var lastLabel = 0
		def create(str : String) : String = {
			lastLabel += 1
			str + "_" + lastLabel
		}
		//take variable 'str' and return a new unique variable tagged with 'tag'
		//assumes that variables are globally unique
		def tag(str : String, tag : String) : String = str + "_" + tag 
	}
	
	
	
	val moduleMatch = "(.*)___(.*)".r //angry face looks down upon you!	
	def moduleName(module : String, name : String) : String = "%s___%s".format(module,name)
	def stripModuleName(name : String) : String = name match {
		case a @ moduleMatch(m,v) => v
		case _ => name
	}
	//returns a block in place of a tree
	//it the tree was an expression, after running the block
	//blk.exp will have the value of that expression
	def blockWrap(tree : Tree) : Block = {
		tree match {
			case b : Block => b
			case x : Stmt => Block(List(x),EmptyExp(UnitT))
  			case x : SimpleExp => Block(List(),x)
  			case x : Exp => {
  				val v = Names.create("block_wrap")
  				Block(List(DefVar(v,x)),Var(x.t,v))
  			}
  			case _ => Log.ierror("unknown tree element: %s",tree)
		}
	}
	  	
  	//takes a tree, and removes any non-defining statements at the top level, creating a new
  	//function that includes the initialization statements
  	//it furthermore changes all variable/function definitions into declarations and 
  	
  	//returns (all declarations/function defs, variable initialization)
    class TopLevel(val var_decls : Block, val var_defs : Block, val func_decls : Block, val func_defs : Block)
  	def normalizeGlobals(tree : Tree) : TopLevel = {
  	
  		/*def Assign(defvar : DefVar) : Stmt = {
  			val tname = Names.create("assign")
	 		Block(
	 			List(DefVar(tname,defvar.init) copyAttr defvar),
	 			Binary(defvar.init.t,Var(defvar.init.t,defvar.name),Liszt.Ops.Assign,Var(defvar.init.t,tname))
	 		)
	 	}*/
  		tree match {
  			case Block(es,exp) => {
  				var func_decls : List[Stmt] = List()
  				var var_decls : List[Stmt] = List()
  				var var_inits : List[Stmt] = List()
  				var func_inits : List[Stmt] = List()
  				
  				for(e <- es) yield {
  					e match {
  						case dv @ DefVar(name,init) => {
  							if(dv.synthetic) {
  								var_inits = dv :: var_inits
  							} else {
  								val assign_stmts = ASTUtil.mkAssign(name,Ops.Assign,init)
  								var_inits = assign_stmts.reverse ::: var_inits
  								var_decls = (DefVar(name,EmptyExp(init.t)) copyAttr e) :: var_decls
  							}
  						}
  						case DefFunc(ftype,name,args,ExternStmt) => {}
  						case DefFunc(ftype,name,args,value) => {
  							func_decls = (DefFunc(ftype,name,args,ExternStmt)  copyAttr e) :: func_decls
  							func_inits = e :: func_inits
  						}
  						case _ => {
  							var_inits = e :: var_inits
  						}
  					}
  				}
  				var_inits = ExpStmt(exp) :: var_inits
  				new TopLevel ( Block(var_decls.reverse,EmptyExp(UnitT)) ,
  				              Block(var_inits.reverse, EmptyExp(UnitT)),
  				              Block(func_decls.reverse,EmptyExp(UnitT)),
  				              Block(func_inits.reverse, EmptyExp(UnitT)) )
  			}
  			case x => normalizeGlobals(blockWrap(tree))
  		}
  	}
  	def removeIfExp(stmt : Stmt) : List[Stmt] = stmt match {
		case DefVar(name,ife @ IfExp(tpt,p,t,e)) => {
			val decl = DefVar(name,EmptyExp(ife.t)) copyAttr stmt
			def addAssign(b : Block) = {
				val newS = ASTUtil.mkAssign(name,Ops.Assign,b.exp)
				Block(b.stmts ::: newS, EmptyExp(UnitT)) copyAttr b
			}
			val newIf = If(p,addAssign(t),addAssign(e)) copyAttr stmt
			List(decl,newIf)
		}
		case _ => Log.ierror(stmt.report("not an IfExp stmt"))
	}
	//creates a mapping from a variable/assign to the definition that created it, it assumes that
	//the tree is a top level block
	//TODO(zach): right now all function arguments and external module references will be mapped to None
	//which is fine for now, but if this gets reused it will need to be fixed
	def varToDefMap(stmts : List[Stmt]) : Map[Tree,Option[Tree]] = {
		object ctx {
			type Frame = scala.collection.mutable.HashMap[String,Option[Tree]]
			def push() = stk.push(new Frame)
			def pop() = stk.pop()
			def +=( pair : (String,Option[Tree]) ) { stk.top += pair }
			
			def find(a : String) : Option[Tree] = stk.toList.find(_.contains(a)) match { 
				case Some(x) => x(a)
				case None => None
			}
			
			private val stk = new scala.collection.mutable.Stack[Frame]()
		}
		ctx.push()
		//initialize top level functions
		for(stmt <- stmts) {
			stmt match {
				case DefVar(name,exp) => ctx += name -> Some(stmt)
				case DefFunc(t,name,args,exp) => ctx += name -> Some(stmt)
				case _ =>
			}
		}
		val result = new scala.collection.mutable.HashMap[Tree,Option[Tree]]
		def makeMap(tree : Tree) {
			tree match {
				case DefVar(name,init) => { makeMap(init); ctx += name -> Some(tree) }
				case DefFunc(t,name,args,exp) => { 
					ctx += name -> Some(tree)
					ctx.push()
					for(dv @ DefVar(name,_) <- args)
						ctx += name -> Some(dv)
					makeMap(exp)
					ctx.pop()
				}
				case Forall(dv,lblop,set,exp) => {
					makeMap(set)
					ctx.push()
					ctx += dv.name -> Some(dv)
					for(lbl <- lblop) {
						ctx += lbl.name -> Some(lbl)
					}
					makeMap(exp)
					ctx.pop()
				}
				case LinearSystemNonzeroes(linear_system,a, blk) => {
					ctx.push()
					ctx += a.name -> Some(a)
					makeMap(blk)
					ctx.pop()
				}
				case Var(t,x) => result += tree -> ctx.find(x)
				case Block(stmts,exp) => {
					ctx.push()
					stmts.foreach(makeMap)
					exp.foreach(makeMap)
					ctx.pop()
				}
				case ShapeAny(_,_,exps) => exps foreach makeMap
			}
		}
		stmts.foreach(makeMap)
		Map.empty ++ result
	}
	def mkAssign(name : String, op : Ops.AssignOp, rhs : Exp) : List[Stmt] = {
		val tmp = Names.create("%s_mkAssign".format(name))
		
		rhs match {
			case r : Var => List(new Assign(name,op,r) copyAttr rhs)
			case _ => {
				val decl = DefVar(tmp,rhs) copyAttr rhs
				val assign = new Assign(name,op,Var(rhs.t,tmp)) copyAttr rhs
				List(decl, assign)
			}
		}
	}
	def emptyDefVar(name : String,t : Type) : DefVar = DefVar(name,EmptyExp(t))
	def nameFromTempHack(name : String) : String = {
		val name_match = "(.*)_mkAssign_[0-9]+".r
		name match { 
			case name_match(a) => a 
			case _ => name
		}
	}
	def uniqueVariables[T <: Tree](tree : T) : T = {
		val ctx = new ContextStack[String,String]
		val used_names = new scala.collection.mutable.HashSet[String]
		val top_level_names = new scala.collection.mutable.HashSet[String]
		for(DefVar(tname,_) <- blockWrap(tree).stmts) {
			assert(!used_names.contains(tname))
			used_names += tname
			top_level_names += tname
			ctx.update(tname,tname)
		}
		def visit[T <: Tree](tree : T) : T = {
			val result = tree match {
				case DefVar(name,init) => {
					val n = if(used_names.contains(name) && !top_level_names.contains(name)) {
						name + Names.create("uniquevar")
					} else {
						name
					}
					used_names += n
					Log.log("mapping: %s -> %s",name,n)
					ctx.update(name,n)
					val ninit = visit(init)
					DefVar(n,ninit)
				}
				case Var(_ : Func,name) => tree
				case Var(t,name) => {
					Log.log("looking up %s = %s", name, ctx.get(name))
					Var(t,ctx(name))
				}
				case Forall(name,lblop,set,block) => {
					ctx.push()
					val nname = visit(name)
					val nlblop = for(lbl <- lblop) yield visit(lbl)
					val nset = visit(set)
					val nblock = visit(block)
					ctx.pop()
					Forall(nname,nlblop,nset,nblock)
				}
				case LinearSystemNonzeroes(linear_system,a, blk) => {
					ctx.push()
					val aa = visit(a)
					val nblock = visit(blk)
					ctx.pop()
					LinearSystemNonzeroes(linear_system,aa,nblock)
				}
				case DefFunc(t,name,args,exp) => {
					ctx.push()
					val nargs = for(a <- args) yield visit(a)
					val nexp = visit(exp)
					ctx.pop()
					DefFunc(t,name,nargs,nexp)
				}
				case Block(stmts,exp) => {
					ctx.push()
					val nstmts = for(s <- stmts) yield visit(s)
					val nexp = visit(exp)
					ctx.pop()
					Block(nstmts,nexp)
				}
				case ShapeAny(_,names,exps) => ShapeAny.trees(tree,exps map visit)
			}
			(result copyAttr tree).asInstanceOf[T]
		}
		visit(tree)
	}
	//returns a list of DefFuncs in topological order, such that no function appearing earlier
	//in the list calls a function later in the list.  Errors out if there is a cycle, since
	//Liszt does not support recursion.  Expects a top level block
	//will ignore external functions
	def toposortCallGraph(tree : Tree) : List[DefFunc] = {
		val blk = blockWrap(tree)
		class Node(val func : DefFunc) {
			var visited = false
			var in_progress = false
			var calls : List[Node] = List()
		}
		
		//map function names to their nodes
		val nameToNode = Map.empty ++ {
			for(dv @ DefFunc(ft,name,args,body) <- blk.stmts) 
				yield dv.name -> new Node(dv)
		}
		//build the calls lists
		for( (n,caller) <- nameToNode) {
			for(Call(_,Var(_,name),_) <- blockWrap(caller.func.body)) {
				nameToNode.get(name) match {
					case Some(called) => caller.calls ::= called
					case None => 
				}
			}
		}
		var visited : List[DefFunc] = List()
		def visit(node : Node) {
			if(node.in_progress)
				Log.error("cycle detected in call graph: I'm not implementing recusion on CUDA, so stop writing it.")
			if(!node.visited) {
				node.in_progress = true
				for(callee <- node.calls) {
					visit(callee)
				}
				node.in_progress = false
				node.visited = true
				visited ::= node.func
			}
		}
		for( (_,node) <- nameToNode )
			visit(node)
		visited.reverse
	}
	
	//makes a new deep-copy of tree, with new variable names maintaining the property that
	//variable names are globally unique.  Each node will have a new set of attributes
	//as well.
	
	def isExternalFunctionDefinition(tree : Tree) : Boolean = tree match {
		case DefFunc(_,_,_,ExternStmt) => true
		case _ => false
	}
	
	//prereq: code must already have unique variable names.
	def duplicate[T <: Tree](tree : T) : T = {
		//find all the defines, these will be renamed
		val renames = {
			def all_defs(tree : Tree) : List[String] = tree match {
				case dv : DefVar => dv.name :: all_defs(dv.init)
				case df : DefFunc if !isExternalFunctionDefinition(df) => 
					df.name ::
					(df.children flatMap all_defs)
				case _ => tree.children flatMap all_defs
			}
			val names = all_defs(tree)
			
			Map.empty ++ { 
				for(name <- names) 
					yield name -> {
					val split = name.split("_dup")(0)
					Names.create(split + "_dup")
				}
			} withDefault identity
		}
		//println("renames = %s".format(renames))
		def doRename(tree : Tree) : Tree = ShapeAny(tree,tree.t,tree.names map renames,tree.children map doRename)
		doRename(tree).asInstanceOf[T]
	}
	
	
	//gets a mapping of non-external functions defined in block
	def functionMap(block : Block) : Map[String,DefFunc] = {
		Map.empty ++ {
			for(df @ DefFunc(t,name,args,exp) <- block.stmts if !isExternalFunctionDefinition(df))
				yield name -> df 
		}
	}
	//takes a normal program block, and for every unique function call context, it clones
	//a copy of the call's target that is unique for that context.
	//This function basically inlines all function calls without actually inlining the 
	//calls, and simplifies analyses that follow it at the cost of potentially creating
	//an exponentially larger amount of code
	
	//main should the name of the entry point function
	def cloneContext(block : Block, main : String) : Block = {
		val function_map = functionMap(block) 
		var cloned_functions : List[DefFunc] = List.empty
		def clone(tree : Tree) : Tree = tree match {
			case Call(t,func,args) if function_map.contains(func.name) => {
				val definition = clone(duplicate(function_map(func.name))).asInstanceOf[DefFunc]
				cloned_functions ::= definition
				Call(t,Var(definition.t,definition.name),args)
			}
			case Var(t,n) => Var(t,n)
			case _ => tree match {
				case ShapeAny(_,_,ds) => ShapeAny.trees(tree, ds map clone)
			}
		}
		
		val top_level = for(stmt <- block.stmts if !stmt.isInstanceOf[DefFunc])
			yield stmt
		val main_fn = clone(function_map(main)).asInstanceOf[DefFunc]
		Block(main_fn :: cloned_functions ::: top_level,EmptyExp(UnitT))
	}
	
	def escapeString(s : String) : String = {
		s.flatMap(c => "\\%o".format(c.toInt))
	}
	
	//each statement in a block can produce multiple statements
	def flatMapBlocksAndExp(top : Tree)(fn : (Stmt) => List[Stmt], simple_fn : (SimpleExp) => (List[Stmt],SimpleExp) ) : Tree = {
		def flatMap(t : Tree) : Tree = t match {
			case Block(stmts,exp) => {
				val nstmts = for(s <- stmts; s2 <- fn(s))
					yield flatMap(s2).asInstanceOf[Stmt]
				val (simple_stmts, sexp) = simple_fn(exp)
				Block(nstmts ::: simple_stmts, sexp)
			}
			//special case the face that a def func might contain a single stmt that is not a block
			case df : DefFunc =>
				val new_body = df.body match {
					case blk : Block => flatMap(blk).asInstanceOf[Stmt]
					case stmt => 
						val stmts = fn(stmt)
						stmts match {
							case List() => Block(List.empty,EmptyExp(UnitT))
							case List(s) => s
							case ss => Block(ss,EmptyExp(UnitT))
						}
				}
				df.copy(body = new_body)
			case ShapeAny(_,names,exps) => ShapeAny.trees(t,exps map flatMap)
		}
		flatMap(top)
	}
	def flatMapBlocks(top : Tree)(fn : (Stmt) => List[Stmt]) : Tree = {
		def simple_fn(exp : SimpleExp) : (List[Stmt],SimpleExp) = (List.empty,exp)
		flatMapBlocksAndExp(top)(fn,simple_fn)
	}
	
	def asStmtList(stmt : Stmt) : List[Stmt] = stmt match {
		case Block(stmts,_) => stmts
		case stmt => List(stmt)
	}
	
	def addFlopCounter(t : Tree) : Tree = {
		def size(tpe : Type) : Int = tpe match {
			case FloatT => 1
			case IntT => 1
			case DoubleT => 1
			case BoolT => 1
			case VectorT(MetaIntT(i),v) => size(v) * i
			case MatrixT(MetaIntT(r),MetaIntT(c),v) => size(v) * r * c
			case _ => 0
		}		
		def flops(e : Exp) : Int = e match {
			case b : Binary => size(b.t) 
			case u : Unary => size(u.t) 
			case c : Call if c.isExternal => size(c.t)
			case _ => 0
		}
		def flopStmt(e : Exp) : List[Stmt] = {
			val fl = flops(e)
			if(fl > 0)
				List(PerfCounter("flops",fl))
			else
				List.empty
		}
		def annotate(stmt : Stmt) : List[Stmt] = {

			stmt match {
				//TODO(zach): total hack the way we count this now, we just say that the result type determines the number of operations performed,
				//which is correct for scala operators and scalar * vector, but not so correct for other operations (vec mat multiply), or
				//sqrt, etc.
				case DefVar(_,e) => stmt :: flopStmt(e)
				case ExpStmt(e) => stmt :: flopStmt(e)
				case FieldWrite(_,_,_ : Ops.Reduce,rhs) => {
					val fl = size(rhs.t)
					val counts = if(fl > 0)
						List(PerfCounter("flops",fl))
					else
						List()
					stmt :: counts
				}
				case _ => List(stmt)
			}
		}
		flatMapBlocks(t)(annotate)
	}
	def wrapForallInWorkgroup(t : Tree) : Tree = {
		t match {
			case fa : Forall => Workgroup(List(fa.copy(block=wrapForallInWorkgroup(fa.block).asInstanceOf[Block])))
			case ShapeAny(_,_,trees) => ShapeAny.trees(t,trees map wrapForallInWorkgroup)
		}
	}
}
