/*
 * Ops.scala
 * All built-in operators are declared here
 */

package Liszt.Ops


//built-in binary operations
@serializable
sealed abstract class BinOp {
  override def toString = this match {
    case Plus => "+"
    case Minus => "-"
    case Times => "*"
    case Div => "/"
    case Mod => "%"
    case Eq => "=="
    case NEq => "!="
    case Less => "<"
    case LessEq => "<="
    case Greater => ">"
    case GreaterEq => ">="
    case Min => "min"
    case Max => "max"
    case LShift => "<<"
    case RShift => ">>"
    case BOr => "|"
    case BAnd => "&"
    case BXor => "^"
    case And => "&&"
    case Or => "||"
  }
}
abstract class ArithOp extends BinOp
case object Plus extends ArithOp
case object Minus extends ArithOp
case object Times extends ArithOp
case object Div extends ArithOp
case object Mod extends ArithOp

abstract class CompOp extends BinOp
case object Eq extends CompOp
case object NEq extends CompOp
case object Less extends CompOp
case object LessEq extends CompOp
case object Greater extends CompOp
case object GreaterEq extends CompOp
abstract class PickOp extends CompOp //TODO(zach): what is a better superclass name for min/max
case object Min extends PickOp
case object Max extends PickOp


abstract class BitOp extends BinOp
case object LShift extends BitOp
case object RShift extends BitOp
case object BOr extends BitOp
case object BAnd extends BitOp
case object BXor extends BitOp

abstract class LogOp extends BinOp
case object And extends LogOp
case object Or extends LogOp

//assignment operators 
//either =, or op=, where op is one of the binary operators from above
@serializable
abstract class AssignOp {
	def asReduction : ReductionOp = asReduction(true)
    def asReduction(collapse : Boolean) : ReductionOp = this match {
      case Assign => RAssign
      case Reduce(op2) => op2 match {
        case Plus => RSum
        case Minus => if(collapse) RSum else RSub
        case Times => RProd
        case Div => if(collapse) RProd else RDiv
        case Min => RMin
        case Max => RMax
        case BOr => RBOr
        case BAnd => RBAnd
        case BXor => RBXor
        case And => RLAnd
        case Or => RLOr
        case _ => throw new Exception("unsupported reduction type: " + op2)
      }
   }
	override def toString = this match {
		case Assign => ":="
    	case Reduce(op) => op.toString + "="
    }
}
case object Assign extends AssignOp
case class Reduce(op : BinOp) extends AssignOp

//unary operators
@serializable
sealed abstract class UnOp {
  override def toString = this match {
    case Neg => "-"
    case BNot => "~"
    case Not => "!"
  }
}
case object Neg extends UnOp;
case object BNot extends UnOp;
case object Not extends UnOp;

//possible reduction operators that our field data structure supports
//this differs from the binary operators above in that
//+/-, and *// are considered the same reduction but different binary operators

sealed abstract class ReductionOp
case object RMax extends ReductionOp
case object RMin extends ReductionOp
case object RProd extends ReductionOp
case object RSum extends ReductionOp
case object RLAnd extends ReductionOp
case object RLOr extends ReductionOp
case object RLXor extends ReductionOp
case object RBAnd extends ReductionOp
case object RBOr extends ReductionOp
case object RBXor extends ReductionOp
case object RAssign extends ReductionOp
case object RSub extends ReductionOp
case object RDiv extends ReductionOp
case object RNone extends ReductionOp //ie a field is in a clean state
case object RModify extends ReductionOp

@serializable
abstract class TopoDir;
case object CW extends TopoDir;
case object CCW extends TopoDir;
case object NoDir extends TopoDir;

//class representing the build in topological functions
//such as vertices going counter-clockwise around a face
@serializable
abstract class TopoFn {
  def argts : List[Liszt.Types.Type]
  def rett : Liszt.Types.Type
  def asFunction : String
}

abstract class SingleTopoFn extends TopoFn //topo function returning a single element, not a set

//retrieve a set of topology elements:
//"get all $t's going $dir around $argt"

case class TopoSet(t : Liszt.Types.Topo, dir : TopoDir, argt : Liszt.Types.Topo) extends TopoFn {
  def argts = List(argt)
  def rett = Liszt.Types.Set(t)
  def asFunction = (t match {
    case Liszt.Types.Vertex => "vertices"
    case Liszt.Types.Edge => "edges"
    case Liszt.Types.Face => "faces"
    case Liszt.Types.Cell => "cells"
  }) + (dir match {
      case CW => "CW"
      case CCW => "CCW"
      case NoDir => ""
  }) + (argt match {
    case Liszt.Types.Vertex => "_v"
    case Liszt.Types.Edge => "_e"
    case Liszt.Types.Face => "_f"
    case Liszt.Types.Cell => "_c"
    case Liszt.Types.Mesh => "_m"
  })
}
//get the vertex on an edge, or the cell on a face
case class Head(argt : Liszt.Types.Topo) extends SingleTopoFn
{
  def argts = List(argt)
  def rett = argt match {
    case Liszt.Types.Edge => Liszt.Types.Vertex
    case Liszt.Types.Face => Liszt.Types.Cell
  }
  def asFunction = argt match {
    case Liszt.Types.Edge => "head"
    case Liszt.Types.Face => "outside"
    case _ => "ERROR"
  }
}
//get the other vertex on the edge, or cell on the face
case class Tail(argt : Liszt.Types.Topo) extends SingleTopoFn {
  def argts = List(argt)
  def rett = argt match {
    case Liszt.Types.Edge => Liszt.Types.Vertex
    case Liszt.Types.Face => Liszt.Types.Cell
  }
  def asFunction = argt match {
    case Liszt.Types.Edge => "tail"
    case Liszt.Types.Face => "inside"
    case _ => "ERROR"
  }
}
//flip the direction of an edge of face head(e) == tail(flip(e))
case class Flip(argt : Liszt.Types.Topo) extends SingleTopoFn {
  def argts = List(argt)
  def rett = argt
  def asFunction = argt match {
    case Liszt.Types.Edge => "flip_e"
    case Liszt.Types.Face => "flip_f"
    case _ => "ERROR"
  }
}
//orient the edge/face such that is pointing towards a vertex/cell
case class Towards(obj : Liszt.Types.Topo, refpoint : Liszt.Types.Topo) extends SingleTopoFn {
  def argts = List(obj,refpoint)
  def rett = obj
  def asFunction = obj match {
    case Liszt.Types.Edge => "towards_e"
    case Liszt.Types.Face => "towards_f"
    case _ => "ERROR"
  }
}
case class LabelExtract(rett : Liszt.Types.Topo, arg : Liszt.Types.Topo) extends SingleTopoFn {
	def argts = List(arg)
	def asTopoSet = TopoSet(rett,NoDir,arg)
	def asFunction = "%s_%s".format(
		rett match {
			case Liszt.Types.Vertex => "vertex"
			case Liszt.Types.Edge => "edge"
			case Liszt.Types.Face => "face"
			case Liszt.Types.Cell => "cell"
		},
		arg match {
			case Liszt.Types.Vertex => "v"
			case Liszt.Types.Edge => "e"
			case Liszt.Types.Face => "f"
			case Liszt.Types.Cell => "c"
		})
}
case class IDLookup(arg : Liszt.Types.Topo) extends SingleTopoFn {
	def argts = List(arg)
	def rett = Liszt.Types.IntT
	def asFunction = "ID_%s".format(arg match {
		case Liszt.Types.Vertex => "v"
		case Liszt.Types.Edge => "e"
		case Liszt.Types.Face => "f"
		case Liszt.Types.Cell => "c"
	})
}






