/*
 * Trees.scala
 * A definition of the Liszt AST, see Types.scala and Ops.scala for the types and operators
 * that the Liszt AST references
 */
package Liszt.Trees

import Liszt._
import Liszt.Types._

private object Log extends Liszt.Log("Trees")

//each Tree is assigned a unique ID
private object ID {
  private var id = 0
  def get() : Int = { id += 1; id}
}


@serializable
case class TraitApply(name : String, args : List[Type], isLinearSystem : Boolean)
@serializable
class Trait(val name : String, val type_args : List[String], val parents : List[TraitApply], val module_dependencies : List[String], val implementation : Block) {
	def instanciate(types : List[Type]) : Trait = {
		var arg_map : Map[String,Type] = Map.empty ++ (type_args zip types)
		
		Log.log("instanciating module %s with map %s".format(name,arg_map))
		def translateType(tpe : Type) : Type = tpe match {
			case TypeVar(v) => arg_map.getOrElse(v,tpe)
			case ShapeType(ts) => ShapeType(tpe,ts map translateType)
		}
		def translateTree(tree : Tree) : Tree = tree match {
			case DefFunc(ftype, name, args, body) => DefFunc(translateType(ftype),name,(args map translateTree).asInstanceOf[List[DefVar]], translateTree(body).asInstanceOf[Stmt])
			case Index(name,  arg_types, index_type) => Index(name, (arg_types map translateType).asInstanceOf[List[Topo]],translateType(index_type))
			case ShapeAny(t,names,exps) => ShapeAny(tree,translateType(t),names,exps map translateTree)
			//TODO: DefFuncs that define new args must mask out variables in arg_map that get shadowed
		}
		new Trait(name,List(),parents,module_dependencies,translateTree(implementation).asInstanceOf[Block])
	}
}

object Trait {
    private def moduleFile(s : String) : String = "%s.lobj".format(s)
	private def module_path = Config.getOrElse("module-path", List(".")).asInstanceOf[List[String]]
	
	private def find_module(name : String) : Option[java.io.File] = {
		val file_name = moduleFile(name)
		val file_path = module_path.find {
			path => (new java.io.File(path,file_name)).exists()
		} 
		for(path <- file_path)
			yield new java.io.File(path,file_name)
	}
	
	def load(name : String) : Option[Trait] = {
		for(file <- find_module(name)) yield {
			val fs = new java.io.FileInputStream(file)
			val trt = IO.decodeTrait(fs)
			fs.close()
			trt
		}
	}
	
	def save(path : String, trt : Trait) {
		val file_name = moduleFile(trt.name)
		val file = new java.io.File(path,file_name)
		val fs = new java.io.FileOutputStream(file)
		IO.encodeTrait(fs,trt)
	}
	
}

//Optional information about the line of scala code that resulted in this AST node
@serializable
case class Position(val filename : String, val lineno : Int, val column : Int, val line : String) {
  def formatString = "%s:%d: %s\n%s\n%s^".format(filename,lineno,"%s",line," " * (column-1))
  override def toString = {
    formatString.format("")
  }
}

@serializable
abstract sealed class Tree {
    def t : Type
    var id = ID.get()
    var context : Option[Tree] = None
    
    for(child <- this.children) {
    	child.context = Some(this)
    }
    
    //two tree nodes are only considered equal iff they share the same ID
    override def hashCode = id
    override def equals(o : Any) = o match {
      case o : Tree => id == o.id
      case _ => false
    }
    
    //copy all the non-AST node attributes of this Tree, including its position, and ID
    //this should only be used when you intend to _replace_ 'b' with 'this' in the 
    //program's AST, as it becomes equal to 'b'
    def copyAttr(b : Tree) : this.type = { id = b.id; position = b.position; this }
    def renumber() {
      // Taken from the IO code below...
      for(t <- this) {
    	print("")
      }
      this.id = ID.get();
      for (t <- this) {
        t.id = ID.get();
      }
    }

    //list of trees that this tree includes
    def children : List[Tree] = List.empty
    //list of names (variable defines, function names, etc.) that this tree includes
    def names : List[String] = List.empty
    //list of variable names that this statement writes to
    //in compiler parlance this is 'defines' but I want to distinguish it from
    //DefVar statments which declare and define variables
    def writes : List[String] = List.empty
    //flatten the tree and operator on every node
    def map[A](fn : Tree => A) : List[A] = this.flatMap( x => List(fn(x)) )
    def flatMap[A](fn : Tree => List[A]) : List[A] = fn(this) ::: this.children.flatMap(_.flatMap(fn))
    def foreach(fn : Tree => Unit) {
      map(fn)
    }
    def filter(fn : Tree => Boolean) : List[Tree] = this.flatMap {
      (tree) => if(fn(tree)) 
        List(tree) 
      else List()
    }
    //create an error message, using the Position information if avaliable
    def report(s : Any) : String = position match {
      case Some(x) => x.formatString.format(s)
      case None => s.toString
    }
    var position : Option[Position] = None
    def withPosition(p : Option[Position]) : this.type = { position = p; this }
    
    
    def nesting : (Boolean,Boolean) = {
    	Log.log("begin nesting %d %s %s".format(id,this.getClass, context))
    	val result = context match {
    		case Some(_ : Forall) => (false,true)
    		case Some(df : DefFunc) => df.info.nesting
    		case Some(x) => x.nesting
    		case None => (true,false)
    	}
    	Log.log("end nesting %d".format(id))
    	result
    }
    //def toString_ = {
    
    override def toString : String
    def toShortString : String = this.toString.lines.next.take(60)
    protected def indent(s : Any) = "    " + s.toString.replaceAll("\n","\n    ")
}
//tree node that only has side-effects (does not produce a value)_
abstract sealed class Stmt extends Tree {
  override def t  : Type = UnitT
  def isLeaf = true //does this statement recursively contain other statments? (e.g. is it a control flow node, or ifexp)
}
//tree node that produces a value
abstract sealed class Exp extends Tree 

//statments that can go on the LHS of a assign (e.g MapAccess, VecExtract, Select)
sealed trait LValue extends Exp {
  def sameRef(rhs : LValue) : Boolean 
  private var isWriteP = false
  def isWrite : Boolean = isWriteP //is this LValue actually used in a LHS position
  def isWrite_=(b : Boolean) {
  	isWriteP = b
  }
  override def copyAttr(b : Tree) : this.type = {
    b match {
      case lv : LValue => isWrite = lv.isWrite
      case _ =>
    }
    super.copyAttr(b)
  }
}

//serialization of Tree data structures, used to pass between compiler and runtime
object IO {

  private def encodeObj(output_stream : java.io.OutputStream, obj : Object) {
    val ostream = new java.io.ObjectOutputStream(output_stream)
    ostream.writeObject(obj)
    ostream.flush()
    
  }
  private def decodeObj[T](input_stream : java.io.InputStream) : T = {
      val istream = new java.io.ObjectInputStream(input_stream)
      val obj = istream.readObject().asInstanceOf[T]
      obj
  }
  
  def encode(tree : Tree) : String = {
    val bytes = new java.io.ByteArrayOutputStream()
    encodeObj(bytes,tree)
    bytes.toString("ISO8859_1")
  }
  
  private def treeAssignUniqueIDs(tree : Tree) {
    //this line is here to get around a scala pattern matching bug
    //that occurs the first time this tree is traversed but not subsequent times
    for(t <- tree) {
    	print("")
    }
    //assign new unique ids to the tree
    tree.id = ID.get()
    for(t <- tree) {
      t.id = ID.get()
    }
  }
  
  def decode(ts : String) : Tree = {
    val bytes = new java.io.ByteArrayInputStream(ts.getBytes("ISO8859_1"))
    val tree = decodeObj[Tree](bytes)
    treeAssignUniqueIDs(tree)
    tree
  }
  
  def encodeTrait(output_stream : java.io.OutputStream, t : Trait) { encodeObj(output_stream,t) }
  def decodeTrait(input_stream : java.io.InputStream) : Trait = {
  	val trt = decodeObj[Trait](input_stream)
  	treeAssignUniqueIDs(trt.implementation)
  	trt
  }
}

//STATMENTS

case class If(predE : Block, thenE : Block, elseE : Block ) extends Stmt { //if(predE) { thenE } else { elseE }
	override def children = List(predE,thenE,elseE)
	override def isLeaf = false
	override def toString = "if(\n%s\n) {\n%s\n} else {\n%s\n}".format(indent(predE),indent(thenE),indent(elseE))
}
case class FieldWrite(field : Var, arg : Var, op : Ops.AssignOp, rhs : Var) extends Stmt { // field(ar) op= rhs
	override def children = List(field,arg,rhs)
	override def toString = "%s(%s) %s %s".format(field,arg,op,rhs)
}
case class FieldSave(field : Var, location : Var) extends Stmt { // field(ar) op= rhs
	override def children = List(field,location)
	override def toString = "fieldSave(%s,%s)".format(field,location)
}
//inserted right before code generation to distinguish global scalar variables from locals
case class ScalarInit(scalar : Var) extends Stmt {
	override def children = List(scalar)
	override def toString = "ScalarInit(%s)".format(scalar)
}

case class ScalarWrite(scalar : Var, offset : VecExtractArgument, op : Ops.AssignOp, rhs : Var) extends Stmt { // field(ar) op= rhs
	override def children = List(scalar,offset,rhs)
	override def toString = "%s %s %s".format(scalar,op,rhs)
}
case class While(pred : Block, block : Block) extends Stmt {//while(pred) { block }
	override def children = List(pred,block)
	override def isLeaf = false
	override def toString = "while(\n%s\n) {\n%s\n}".format(indent(pred),indent(block))
}
case class Forall(name : DefVar, label : Option[DefVar], set : Var, block : Block) extends Stmt { //for(a [withLabel l] <- set) { block }
	override def children = name :: label.toList ::: List(set,block)
	override def isLeaf = false
	override def toString = label match {
		case Some(lbl) => "for( (%s,%s) in %s ) {\n%s\n}".format(name.name,lbl.name,set,indent(block))
		case None => "for( %s in %s ) {\n%s\n}".format(name.name,set,indent(block))
	}
	var mustBeTrivial : Boolean = false
	override def copyAttr(t : Tree) : this.type = {
		t match {
			case f : Forall => this.mustBeTrivial = f.mustBeTrivial
			case _ =>
		}
		super.copyAttr(t)
	}
}
case class Return(e : SimpleExp) extends Stmt { //return e
	override def children = List(e)
	override def toString = "return %s".format(e)
}
//a special print statment to make debugging easier, its debatable as to if
//this should be a special language statement or a function call in liszt, for now its easier
//to support a special statement
case class Print(es : List[Var]) extends Stmt {//Print(es(0),es(1),es(2), ... )
	override def children = es
	override def toString = "print(%s)".format(es.mkString("",",",""))
}

case object Exit extends Stmt;

case class DefVar(name : String, init : Exp) extends Stmt { //val name = init
  var synthetic = false
  var global = false
  init match {
    case f : FieldInit => f withName name
    case _ =>
  }
  override def isLeaf = !init.isInstanceOf[IfExp]
  def withSynthetic(b : Boolean) : this.type = { synthetic = b; this}
  def withGlobal(b : Boolean) : this.type = { global = b; this}
  override def copyAttr(t : Tree) : this.type = {
    t match {
      case d : DefVar => this withSynthetic d.synthetic withGlobal d.global
      case _ =>
    }
    super.copyAttr(t)
  }
  override def children = List(init)
  override def names = List(name)
  override def writes = List(name)
  override def toString = "%s : %s %s = %s".format(name,init.t,if(synthetic) "[synthetic]" else "",init)
}//
case class Assign(lv : LValue, op : Ops.AssignOp, r : Var) extends Stmt { //lv op= r
  lv.isWrite = true
  def this(name : String, op : Ops.AssignOp, r : Var) = this(Var(r.t,name),op,r)
  override def children = List(lv,r)
  override def writes = lv.writes
  override def toString = "%s %s %s".format(lv,op,r)
}
case class EnsureFieldState(field : Var, state : Ops.ReductionOp) extends Stmt { //compiler inserted to keep fields up-to-date
	override def children = List(field)
	override def toString = "%s.ensureState(%s)".format(field,state)
}
case class Workgroup(stmts : List[Stmt]) extends Stmt {//compiler inserted to denote blocks of parallel work
	override def children = stmts
	override def toString =
		if(stmts.size == 1) 
			"workgroup %d: ".format(this.id) + stmts.head.toString 
		else 
			"workgroup %d {\n%s\n}".format(this.id,indent(stmts.mkString("\n")) )
}
case class Block(stmts : List[Stmt],exp : SimpleExp) extends Stmt { //{ stmt(0); stmt(1); stmt(2); ... ; exp },  for statments exp = EmptyExp

	override def children = stmts ::: List(exp)
	override def isLeaf = false
	lazy val functions : List[DefFunc] = for(df @ DefFunc(_,_,_,_) <- stmts) yield df
	private lazy val function_calls : List[ (DefFunc,Call) ] = {
		def calls(t : Tree) : List[Call] = t match {
			case c : Call => List(c)
			case ShapeAny(_,_,es) => es flatMap calls
		}
		scala.Predef.Set.empty ++ {
			for(df <- functions; c <- calls(df) ) yield (df,c)
		}
	}.toList
	def definesFunc(name : String) : Boolean = functions.exists(df => df.name == name && df.body.getClass != ExternStmt.getClass) 
	abstract class FunctionInfo {
		def impl : DefFunc
		def calls: List[Call]
		def calledBy : List[Call]
		def nesting : (Boolean,Boolean)
	}
	def function_info(name : String) : FunctionInfo = new FunctionInfo {
		lazy val impl = { for(df @ DefFunc(_,nm,_,_) <- stmts if name == nm) yield df }.head
		
		lazy val calls : List[Call] = { scala.Predef.Set.empty ++  { for ( (er,ee) <- function_calls if er.name == name) yield ee } }.toList
		lazy val calledBy : List[Call] = { scala.Predef.Set.empty ++  { for( (_,ee) <- function_calls if name == ee.func.name) yield ee } }.toList
		lazy val nesting = if(name == "_liszt_start") (true,false) else {
			val nestings = for(c <- calledBy) yield c.nesting
			nestings.foldLeft( (false,false) ) {
				case ( (a,b) , (c,d) ) => (a || c, b || d) 
			}
		}
	}
	override def toString = (stmts ::: List(exp)).filter(x => !x.isInstanceOf[EmptyExp]).mkString("","\n","")
}

case class DefFunc(ftype : Type, name : String, args : List[DefVar], body : Stmt) extends Stmt {//def name(args(0),args(1),args(2), ...) { body }
	def typ = ftype.asInstanceOf[Func]
	def info = this.context.get.asInstanceOf[Block].function_info(name)
	override def children = args ::: List(body)
	override def names = List(name)
	override def isLeaf = false
	override def toString = body match {
		case ExternStmt => "extern def %s%s : %s".format(name,args.map(_.name).mkString("(",",",")"),t)
		case _ => "def %s%s : %s {\n%s\n}".format(name,args.map(_.name).mkString("(",",",")"),t,indent(body))
	}
}
case class ExpStmt(e : Exp) extends Stmt { // e;
	override def children = List(e)
	override def toString = e.toString
}

case class PerfCounter(name : String, count : Int) extends Stmt { //added to the code to count certain operations, currently only "flops" is supported
	override def toString = "perf_count(%s,%d)".format(name,count)
}
case object ExternStmt extends Stmt { // signifies that the value of the function/value is just a declaration and will be provided externally
	override def toString = ";"
}

//EXPRESSIONS

case class Binary( t : Type,l : Var, op : Ops.BinOp, r : Var) extends Exp {//l op r
	override def children = List(l,r)
	override def toString = "%s %s %s".format(l,op,r)
}
case class Unary(t : Type, op : Ops.UnOp, e : Var) extends Exp { //op e
	override def children = List(e)
	override def toString = "%s%s".format(op,e)
}
case class Call(t : Type, func : Var, args : List[Var]) extends Exp {//func(args)
	override def children = func :: args
	var type_params : List[Type] = List()
	
	def withTypeParameters(params : List[Type]) : this.type = {
		type_params = params
		this
	}
	override def copyAttr(t : Tree) = {
		t match {
			case c : Call => 
				type_params = c.type_params
			case _ =>
		}
		super.copyAttr(t)
	}
	//does the AST define this function or is it defined elsewhere
	def isExternal = {
		def definesFunc(t : Option[Tree]) : Boolean = t match {
			case Some(b : Block) => b.definesFunc(func.name) || definesFunc(b.context)
			case Some(x) => definesFunc(x.context)
			case None => false
		}
		!definesFunc(this.context)
	}
	override def toString ="%s%s".format(func,args.mkString("(",",",")"))
}
case class TopoCall( t : Type, func : Ops.TopoFn, args : List[Var]) extends Exp {//func(arg)
	override def children = args
	override def toString = "%s%s".format(func,args.mkString("(",",",")"))
}
case class FieldRead(t : Type, field : Var, arg : Var) extends Exp { //field(arg)
	override def children = List(field,arg)
	override def toString = "%s(%s)".format(field,arg)
}

//inserted right before code generation to distinguish global scalar variables from locals
case class  ScalarRead(scalar : Var) extends Exp { //field(arg)
	override def t = scalar.t
	override def children = List(scalar)
	override def toString = "%s".format(scalar)
}


trait VecExtractArgument extends Exp 
case class VecExtract(t : Type, vec : Var, n : VecExtractArgument) extends Exp with LValue { //vec(n)
	override def isWrite = vec.isWrite
	override def isWrite_=(b : Boolean) {
		vec.isWrite = b
	}
	def sameRef(rhs : LValue) = rhs match {
		case VecExtract(t, vv, nn) => (vec sameRef vv) && {
			(nn,n) match {
				case(IntL(x),IntL(y)) => x == y
				case (x : Var, y : Var) => x sameRef y
				case _ => false
			}
		}
		case _ => false
	}
	override def children = List(vec,n)
	override def writes = if(isWrite) vec.writes else List.empty
	override def toString = "%s[%s]".format(vec,n)
}
case class MatExtract(t : Type, mat : Var, n : Var, m : Var) extends Exp with LValue { //mat(n,m)
	override def isWrite = mat.isWrite
	override def isWrite_=(b : Boolean) {
		mat.isWrite = b
	}
	def sameRef(rhs : LValue) = rhs match {
		case MatExtract(t, mm, rr, cc) => (mat sameRef mm) && (n sameRef rr) && (m sameRef cc)
		case _ => false
	}
	override def children = List(mat,n,m)
	override def writes = if(isWrite) mat.writes else List.empty
	override def toString = "%s[%s,%s]".format(mat,n,m)
}
case class IfExp(t : Type, predE : Block, thenE : Block, elseE : Block) extends Exp {//if(predE) thenE else elseE 
	override def children = List(predE,thenE,elseE)
	override def toString = "ifexp(\n%s\n) {\n%s\n} else {\n%s\n}".format(indent(predE),indent(thenE),indent(elseE))
}
abstract sealed class SimpleExp extends Exp //an expression that requires no computation (e.g. constant, or variable)

case class Var(t : Type,name : String) extends SimpleExp with LValue with VecExtractArgument {//
  var global = false
  def sameRef(rhs : LValue) = rhs match {
    case Var(tt,nn) => name == nn
    case _ => false
  }
  def withGlobal(b : Boolean) : this.type = { global = b; this }
  override def copyAttr(b : Tree) : this.type = {
  	b match {
  		case v : Var => this withGlobal v.global
  		case _ => this
  	}
  	super.copyAttr(b)
  }
  override def names = List(name)
  override def writes = if(isWrite) List(name) else List.empty
  override def toString = name
}

case class EmptyExp(t : Type) extends SimpleExp {
	override def toString = "<empty:%s>".format(t)
}

//initialization sources for fields
//currently either the position of a vertex, or a constant initialization value
@serializable
abstract class FieldSource //
case class StringSource(e : String) extends FieldSource //
case class ConstSource(e : Var) extends FieldSource //

case class FieldInit(t : Type,src : FieldSource) extends Exp {
  var name : String = null
  override def toString = src match {
  	case StringSource(e) => "fieldFromString(%s)".format(e)
  	case ConstSource(e) => "fieldWithConst(%s)".format(e)
  }
  def withName(n : String) : this.type = { name = n; this }
  override def children = src match {
  	case ConstSource(e) => List(e)
  	case _ => List.empty
  }
}

//initialization sources for boundary sets
//currently extracted from the mesh file formats
case class SetInit(t : Type, name : String) extends Exp { //
	override def toString = "boundarySetWith(%s)".format(name)
}
//nodes representing constant expressions
abstract sealed class Literal extends Exp { //
	def c : Any //the literal expression
}
case class IntL(c : Int) extends Literal with VecExtractArgument { def t = IntT 
	override def toString = c.toString
} //
case class FloatL(c : Float) extends Literal { def t = FloatT 
	override def toString = c.toString
} //
case class DoubleL(c : Double) extends Literal { def t = DoubleT 
	override def toString = c.toString
} //
case class StringL(c : String) extends Literal { def t = StringT 
	override def toString = c.toString
} //
case class BoolL(c : Boolean) extends Literal { def t = BoolT 
	override def toString = c.toString
} //
case class SimdL(c : Literal) extends Literal {
  def t = SimdT(c.t);
  override def toString = "'%s'_simd".format(c.toString);
}

case class VectorL(t : Type, es : List[Var]) extends Exp {//
	override def children = es
	override def toString = es.mkString("[",",","]")
}
case class MatrixL(t : Type, es : List[Var]) extends Exp {//
	override def children = es
	override def toString = es.mkString("[",",","]")
}
case class MatrixFromVec(t : Type, es : List[Var]) extends Exp {
	override def children = es
	override def toString = es.mkString("matrixFromVec(",",",")")
}


case class ExecKernel(kernel_id : Int, set : Var) extends Stmt {
	override def children = List(set)
	override def toString = "ExecKernel(%d,%s)".format(id,set)
}
//EXPERIMENTAL: these nodes are not yet fully supported


//Indicies
case class Index(name : String,  arg_types : List[Type], index_type : Type) extends Stmt {
	override def toString = "index %s(%s) : %s".format(name,arg_types.mkString(","),index_type)
}
case class IndexCall(t : Type, linear_system : String, index : Index, args : List[Var]) extends Exp {
	override def children = index :: args
	override def toString = "%s.%s(%s)".format(linear_system,index.name,args.mkString(","))
}


//Linear System Operations
case class DefLinearSystem(name : String, solver : String, indices : List[Index]) extends Stmt {
	override def children = indices
	override def names = List(name)
	override def toString = "linear_system %s with solver %s {\n %s \n }\n".format(name, solver, indices.mkString("\n"))
}

abstract class LSElement
case object LS_A extends LSElement
case object LS_B extends LSElement
case object LS_X extends LSElement

case class LinearSystemAllocate(t : Type, linear_system : String, element : LSElement) extends Exp {
	 override def toString = {
		val name = element match {
			case LS_A => "a"
			case LS_B => "x"
			case LS_X => "b"
		}
		"%s.%s()".format(linear_system,element)
	}
}

case class LinearSystemCall(t : Type, linear_system : String, func : Var, args : List[Var]) extends Exp {
	override def children = args
	override def toString = "%s.%s(%s)".format(linear_system,func,args.mkString(","))
}
case class LinearSystemNonzeroes(linear_system : String,nonzeroes : DefVar, blk : Block) extends Stmt {
	override def children = List(nonzeroes,blk)
	override def isLeaf = false
	override def toString = "%s.nonzeroes(%s) { %s }".format(linear_system,nonzeroes,blk)
}
case class LinearSystemPlaceNonzero(obj : Var, element : LSElement, args : List[Var]) extends Stmt {
	override def children = obj :: args

}

//Matrix and Vector operations
case class SparseVectorAccess(t : Type, vec : Var, idx : Var) extends Exp with LValue {
  def sameRef(rhs : LValue) = rhs match {
    case SparseVectorAccess(t,vec2,idx2) => (vec sameRef vec2) && (idx sameRef idx2) 
    case _ => false
  }
  override def children = List(vec,idx)
  override def toString = "%s(%s)".format(vec,idx)
}
case class SparseMatrixAccess(t : Type, vec : Var, row : Var, col : Var) extends Exp with LValue {
  def sameRef(rhs : LValue) = rhs match {
    case SparseMatrixAccess(t,vec2,row2,col2) => (vec sameRef vec2) && (row sameRef row2) && (col sameRef col2)
    case _ => false
  }
  override def children = List(vec,row,col)
  override def toString = "%s(%s,%s)".format(vec,row,col)
}

//matching pattern modeled after Quotations.ExprShape in F#
object ShapeAny {
	def unapply(e : Tree) : Option[ (Type,List[String], List[Tree]) ] = Some( (e.t,e.names,e.children) )
	def trees(oldExp : Tree, es : List[Tree]) : Tree = apply(oldExp,oldExp.t,oldExp.names,es)
	def names(oldExp : Tree, names : List[String]) : Tree = apply(oldExp,oldExp.t,names,oldExp.children)
	def apply(oldExp : Tree, tt : Type, names : List[String], es : List[Tree]) : Tree = {
		//implicitly recast to original values when necessary
		implicit def expToVar(e : Tree) : Var = e.asInstanceOf[Var]
		implicit def expToBlock(e : Tree) : Block = e.asInstanceOf[Block]
		implicit def treeToExp(e : Tree) : Exp = e.asInstanceOf[Exp]
		implicit def treeToStmt(e : Tree) : Stmt = e.asInstanceOf[Stmt]
		implicit def esToA[A](es : List[Tree])(implicit etoa : (Tree) => A) : List[A] = es.map(etoa)
		val newExp = oldExp match {
			case Var(t,n) => Var(tt,names.head)
			case DefFunc(ftype,name,args,stmt) => DefFunc(ftype,names.head,for(dv <- es.init) yield dv.asInstanceOf[DefVar],es.last.asInstanceOf[Stmt])
			case DefVar(name,init) => DefVar(names.head,es.head.asInstanceOf[Exp])
			case Forall(name,None,set,block) => Forall(es(0).asInstanceOf[DefVar],None,es(1).asInstanceOf[Var],es(2).asInstanceOf[Block])
			case Forall(name,Some(lbl),set,block) => Forall(es(0).asInstanceOf[DefVar],Some(es(1).asInstanceOf[DefVar]),es(2).asInstanceOf[Var],es(3).asInstanceOf[Block])
			case DefLinearSystem(name , solver, indicies ) => DefLinearSystem(names.head,solver,es.asInstanceOf[List[Index]])
			case Binary(t,l,op,r) => Binary(tt,es(0),op,es(1))
			case If(predE,thenE,elseE) => If(es(0),es(1),es(2))
			case Unary(t,op,e) => Unary(tt,op,es(0))
			case Call(t,func,args) => Call(t,es.head,esToA(es.tail))
			case TopoCall(t,op,args) => TopoCall(tt,op,esToA(es))
			case FieldRead(t,field,arg) => FieldRead(tt,es(0),es(1))
			case FieldWrite(field,arg,op,exp) => FieldWrite(es(0),es(1),op,es(2))
			case FieldSave(field,location) => FieldSave(es(0),es(1))
			case SparseVectorAccess(t,vec,idx) => SparseVectorAccess(tt,es.head,es(1))
			case SparseMatrixAccess(t,vec,row,col) => SparseMatrixAccess(tt,es(0),es(1),es(2))
			case VecExtract(t,vec,arg) => VecExtract(tt,es(0),es(1).asInstanceOf[VecExtractArgument])
			case MatExtract(t,mat,n,m) => MatExtract(tt,es(0),es(1),es(2))
			case While(pred,block) => While(es(0),es(1))
			case Block(ess,ee) => Block(es.init.map(_.asInstanceOf[Stmt]),es.last.asInstanceOf[SimpleExp])
			case Return(e) => Return(es(0).asInstanceOf[SimpleExp])
			case Print(ess) => Print(esToA(es))
                        case Exit => Exit;
			case fi @ FieldInit(t,StringSource(e)) => FieldInit(tt,StringSource(e))
			case fi @ FieldInit(t,ConstSource(e)) => FieldInit(t,ConstSource(es(0)))
			case SetInit(t,s) => SetInit(tt,s)
			case IntL(c) => oldExp
			case FloatL(c) =>  oldExp
			case DoubleL(c) => oldExp
			case StringL(c) => oldExp
			case BoolL(c) => oldExp
                        case SimdL(c) => oldExp
			case VectorL(t,ess) => VectorL(tt,esToA(es))
			case MatrixL(t,ess) => MatrixL(tt,esToA(es))
			case MatrixFromVec(t,ess) => MatrixFromVec(tt,esToA(es))
			case ExpStmt(e) => ExpStmt(treeToExp(es(0))) 
			case IfExp(t,predE,thenE,elseE) => IfExp(t,es(0),es(1),es(2))
			case Assign(lval,op,exp) => Assign(es(0).asInstanceOf[LValue],op,es(1).asInstanceOf[Exp]) 
			case EnsureFieldState(f,reduction) => EnsureFieldState(es(0),reduction)
			case Workgroup(old) => Workgroup(es.map(_.asInstanceOf[Stmt]))
			case Index(name ,  arg_types , index_type ) => Index(name, arg_types, index_type)
			case IndexCall(t , linear_system , index , args ) => IndexCall(t, linear_system, es.head.asInstanceOf[Index], es.tail.asInstanceOf[List[Var]])
			case LinearSystemAllocate(t , linear_system , element ) => LinearSystemAllocate(t, linear_system, element)
			case LinearSystemCall(t, linear_system, func, args) => LinearSystemCall(t,linear_system,func,args)
			case LinearSystemNonzeroes(linear_system ,a , blk ) => LinearSystemNonzeroes(linear_system, es(0).asInstanceOf[DefVar], es(1).asInstanceOf[Block]) 
			case LinearSystemPlaceNonzero(obj,element, args) => LinearSystemPlaceNonzero(es.head,element,es.tail.asInstanceOf[List[Var]])
			case PerfCounter(n,c) => PerfCounter(n,c)
			case EmptyExp(t) => EmptyExp(tt)
			case ScalarInit(scalar) => ScalarInit(es(0))
			case ScalarWrite(scalar, offset, op , rhs) => ScalarWrite(es(0), es(1).asInstanceOf[VecExtractArgument],op,es(2))
			case ScalarRead(scalar) => ScalarRead(es(0))
			case ExecKernel(id : Int, set : Var) => ExecKernel(id, es(0))
			case x => {
				//there is a bug that can cause two instances of a singleton object to exist if you create one via java's object serialization
				//this causes the case matches against the object to fail
				//we get around this by comparing the class of the object, and then replacing it with the correct object
				if(x.getClass == ExternStmt.getClass)
					ExternStmt
				else
					Log.ierror("shape any failed: did you add an AST node and forget to add reconstruction code to ShapeAny?")
			}
		}
		newExp copyAttr oldExp
	}
}
object WriteStmt {
	def unapply(e : Tree) : Option[(Var, Ops.ReductionOp, List[Tree])] = e match {
		case FieldWrite(field,arg,op,exp) => Some( (field,op.asReduction, List(arg,exp)) )
		case Assign(lv : LValue, op_, exp) => {
			val op = op_.asReduction
			lv match {
				case v : Var => Some( (v,op,List(exp)) )
				case SparseVectorAccess(t, vec, idx) => Some( (vec,op,List(idx,exp)) )
				case SparseMatrixAccess(t, mat, row, col) => Some( (mat,op,List(row,col,exp)))
				case _ => None
			}
		}
		case _ => None
	}
}
