/*
 * Types.scala
 * types that Liszt supports
 */
package Liszt.Types
import Liszt.ASTUtil

@serializable
abstract class Type

abstract class Topo extends Type
case object Vertex extends Topo
case object Edge extends Topo
case object Face extends Topo
case object Cell extends Topo
case object Mesh extends Topo


//a special Liszt type that stores values at a particular mesh element
//i.e. a map from domain -> range, where domain is one of
//Vertex, Edge, Face, or Cell
case class FieldT(domain : Type, range : Type) extends Type {
	override def toString = "%s -> %s".format(domain,range)
}

abstract class BasicType extends Type
case object IntT extends BasicType
case object FloatT extends BasicType
case object DoubleT extends BasicType
case object StringT extends BasicType
case object BoolT extends BasicType
case class MetaIntT(a : Int) extends BasicType { //for vector/matrix entries
	override def toString = "%d".format(a)
}
case class TypeVar(a : String) extends Type //an abstract type, removed by the linking phase in the runtime

//short vector of constant length
case class VectorT(n : Type, t : Type) extends Type {
	def baseType = Some(t)
}

// Simd[T]
case class SimdT(t : Type) extends Type {
  def baseType = Some(t)
}


//a set of mesh elements e.g. vertices(mesh)
case class Set(t : Type) extends Type

//a function type
case class Func(args : List[Type], range : Type) extends Type {
	override def toString = "(%s) => %s".format(args.mkString(","),range)
}

case object UnitT extends BasicType


//Experimental Types: Not fully implemented in the Liszt compiler
case class MatrixT(n : Type, m : Type, t : Type) extends Type {
	def baseType = Some(t)
	def columnType = VectorT(n,t)
	def rowType = VectorT(m,t)
}

case class SparseVectorType(linear_system : String) extends Type
case class SparseMatrixType(linear_system : String) extends Type
case class NonzeroesSpecifierType(linear_system : String) extends Type

case class IndexType(isRowIndex : Boolean) extends Type

object ShapeType {
	def apply(t : Type, ts : List[Type]) : Type = t match {
		case FieldT(d,r) => FieldT(ts(0).asInstanceOf[Topo],ts(1))
		case Func(args,range) => Func(ts.init,ts.last)
		case Set(t) => Set(ts.head.asInstanceOf[Topo])
		case MatrixT(n,m,t) => MatrixT(ts(0),ts(1),ts(2))
		case VectorT(n,t) => VectorT(ts(0),ts(1))
		case _ => t
	}
	def unapply(t : Type) : Option[ List[Type] ] = t match {
		case FieldT(d,r) => Some(List(d,r))
		case Func(args,range) => Some(args ::: List(range))
		case Set(t) => Some(List(t))
		case MatrixT(n,m,t) => Some(List(n,m,t))
		case VectorT(n,t) => Some(List(n,t))
		case _ => Some(List())
	}
}
