package scala.LisztPlugin

import scala.tools.nsc
import nsc.Global

import Liszt.Types.{Type => LType, _}
import Liszt.Log

trait TypeTranslation {
private object Log extends Liszt.Log("TypeTranslation")
val global : Global
import global._

object ToLType {
	object MInt {
		lazy val types = for(i <- 0 to 9) yield (i,definitions.getClass("Liszt.Language._%d".format(i)).tpe)
		def unapply(a : Type) : Option[Int] = {
			val regex = "type _([0-9])".r
			a match {
				case TypeRef(_,sym,List()) => { 
				sym.toString match {
					case regex(i) => Some(i.toInt)
					case _ => None
				}
				}
				case _ => None
			}
		}
	}
	def liftOption[A](as : List[Option[A]]) : Option[List[A]] = {
		if(as.contains(None))
			None
		else
			Some(for(Some(a) <- as) yield a)
	}
	def toLTypeList(as : List[Symbol]) : Option[List[LType]] = liftOption(as.map(x => ToLType.unapply(x.tpe)))

	private case class LisztClass(name : String) {
		val clazz = definitions.getClass(name)
		def unapplySeq(a : Type) : Option[Seq[Type]] = {
			a match {
				case TypeRef(p,sym,params) if sym.tpe <:< clazz.tpe => {
					Some(a.typeArgs)
				}
				case _ => None
			}
		}
	}
	private val fieldType = LisztClass("Liszt.Language.Field")
	private val vecType = LisztClass("Liszt.Language.Vec")
	private val matType = LisztClass("Liszt.Language.Mat")
	
	private val meshType = LisztClass("Liszt.Language.Mesh")
	private val cellType = LisztClass("Liszt.Language.Cell")
	private val faceType = LisztClass("Liszt.Language.Face")
	private val edgeType = LisztClass("Liszt.Language.Edge")
	private val vertType = LisztClass("Liszt.Language.Vertex")
	private val setType = LisztClass("Liszt.Language.Set")

	private val zeroType = LisztClass("Liszt.Language.Zero")
	private val succType = LisztClass("Liszt.Language.Succ")
	private val annoType = LisztClass("Liszt.Language.lisztcode")
	private val sparseVecType = LisztClass("Liszt.Language.SparseVector")
	private val sparseMatType = LisztClass("Liszt.Language.SparseMatrix")
	private val nonZeroesType = LisztClass("Liszt.Language.NonZeroes")
	private val rowIndexType = LisztClass("Liszt.Language.RowIndex")
	private val colIndexType = LisztClass("Liszt.Language.ColIndex")
	
	def isLisztObjectSym(sym : Symbol) : Boolean = {
		sym.annotations match {
			case List(AnnotationInfo(tpe @ annoType(),args,assocs)) => true
			case _ => false
		}
	}
	object BType {
		private val types = List(
			(definitions.IntClass.tpe,IntT),
			(definitions.FloatClass.tpe,FloatT),
			(definitions.DoubleClass.tpe,DoubleT),
			(definitions.StringClass.tpe,StringT),
			(definitions.BooleanClass.tpe,BoolT),
			(definitions.UnitClass.tpe,UnitT)
		)
		def unapply(t : Type) : Option[BasicType] = {
			val tt = t match {
				case ConstantType(v) => v.tpe
				case _ => t
			}
			for ((_,lt) <- types.find(_._1 =:= tt))
				yield lt
		}
	}
	def unapply(t : Type) : Option[LType] = {
		Log.log("TYPE matching: %s (class: %s)",t,t.getClass)	
		val result = t match {
			case MethodType(params,result) => (toLTypeList(params),unapply(result)) match {
				case (Some(args),Some(r)) => Some(Func(args,r))
				case _ => None
			}
			case BType(tt) => Some(tt)
			case vertType() => Some(Vertex)
			case edgeType() => Some(Edge)
			case faceType() => Some(Face)
			case cellType() => Some(Cell)
			case meshType() => Some(Mesh)
			case _ => t match {
			//see: https://lampsvn.epfl.ch/trac/scala/ticket/1133, for why these
			//cannot be nested
			case vecType(a,b) => (a,b) match {
				case (ToLType(mint),ToLType(v)) => Some(VectorT(mint,v))
				case _ => None
			}
			/**/
			case fieldType(a,b) => (a,b) match {
				case (ToLType(m),ToLType(vt)) => Some(FieldT(m.asInstanceOf[Topo],vt))
				case _ => None
			}
			case setType(b) => b match {
				case ToLType(a) => Some(Set(a.asInstanceOf[Topo]))
				case _ => None
			}
			case SingleType(pre,sym) => unapply(sym.tpe)
			case PolyType(List(),resultType) => unapply(resultType) //sometimes things get typed as => typ, instead of typ, e.g a.B.update(x) a.B has type '=> a.B'
			case zeroType() => Some(MetaIntT(0))
			case succType(n) => unapply(n) match {
				case Some(MetaIntT(x)) => Some(MetaIntT(x+1))
				case _ => None
			}
			case sparseVecType() => {
				val TypeRef(SingleType(_,sym),_,_) = t.normalize
				Some(SparseVectorType(sym.name.toString))
			}
			case sparseMatType() => {
				val TypeRef(SingleType(_,sym),_,_) = t.normalize
				Some(SparseMatrixType(sym.name.toString))
			}

			case TypeRef(pre,sym,args) if sym.isNonClassType => {
				if(sym.isAbstractType) {
					Some(Liszt.Types.TypeVar(sym.name.toString))
				} else if(sym.isAliasType) {
					unapply(t.normalize)
				} else {
					Log.log("symbol: %s symbol type: %s info: %s",sym,sym.tpe,sym.info)
					unapply(sym.info)
				}
			}
			case _ => t match {
				case _ => {
					t match {
						case nonZeroesType() => {
							val TypeRef(SingleType(_,sym),_,_) = t.normalize
							Some(NonzeroesSpecifierType(sym.name.toString))
						}
						case rowIndexType() => Some(IndexType(true))
						case colIndexType() => Some(IndexType(false))
						case matType(r,c,v) => (r,c,v) match {
							case (ToLType(ri),ToLType(ci),ToLType(vt)) => Some(MatrixT(ri,ci,vt))
							case _ => None
						}
						case _ => None
					}
				}
			}
		}}
		Log.log("matched: %s",result)
		result
	}
}
class RichType(t : Type) {
	def lisztType : LType = lisztTypeOption match {
		case Some(tt) => tt
		case None => Log.ierror("not a liszt type (%s) %s",t.getClass, debugString(t))
	}
	def lisztTypeOption : Option[LType] = ToLType.unapply(t)
}
implicit def raiseType(t : Type) : RichType = new RichType(t)
//create an integer from a tree with a meta integer type
def mIntValueToInt(t : Tree) : Int = ToLType.MInt.unapply(t.tpe).get
def isImplicitType(t : Type) : Boolean = t match {
		case mt : MethodType if mt.isImplicit => true
		case _ => false
	}
} //trait TypeTranslation
