package Liszt

import Liszt.Types._
import Liszt.Trees._

object AbstractInterpretation {
	import scala.collection.immutable.{Set => SSet}
	object Log extends Liszt.Log("AbstractInterpretation")
	
	type ObjectID = Int
	class Value {
		def meet(rhs : Value) : Value = if(this == rhs) this else Union(this,rhs) 
		def copy() : Value = this
	}
	case object TopValue extends Value {
		override def meet(rhs : Value) = rhs
	}
	case class Union(a : Value, b : Value) extends Value
	
	case class TopoSet(members : SSet[Topology.Node]) extends Value {
		override def meet(rhs : Value) : Value = rhs match {
			case TopValue => this
			case TopoSet(members2) => TopoSet( members union members2 )
			case _ => Log.ierror("undefined meet: %s meet %s",this,rhs)
		}
	}
	
	case class Reference(obj : ObjectID) extends Value
	
	
	abstract class LObject extends Value with java.lang.Cloneable {
		override def copy() : LObject = this.clone().asInstanceOf[LObject]
	}
	
	type V = List[Value]
	
	abstract class Sparse extends LObject {
		
		protected var readset : SSet[V] = SSet.empty
		protected var writeset : SSet[V] = SSet.empty
		protected var nonzeroset : SSet[V] = SSet.empty
		protected def create() : Sparse
		
		override def meet(rhs : Value) : LObject = rhs match {
			case rhs : Sparse =>
			val a = create()
			a.readset = this.readset ++ rhs.readset
			a.writeset = this.writeset ++ rhs.writeset
			a.nonzeroset = this.nonzeroset ++ rhs.nonzeroset
			a
		}
		def read(e : V) {
			readset += e 
		}
		def write(e : V) {
			readset += e
		}
		def nonzeroes() : SSet[V] = writeset ++ nonzeroset
	}
	class SparseVector extends Sparse {
		override protected def create() = new SparseVector
		def +(sv : SparseVector) : SparseVector = {
			val a = new SparseVector
			a.nonzeroset = this.nonzeroes() ++ sv.nonzeroes()
			a
		}
		def read(e : Value) { read(List(e)) }
		def write(e : Value) { write(List(e)) }
		//scalar_variable * this, -this, etc.
		def scalar() : SparseVector = {
			val a = new SparseVector
			a.nonzeroset = this.nonzeroes()
			a
		}
		def *(v : SparseMatrix) : SparseVector = {
			val a = new SparseVector
			val r1 = for( List(r,c) <- v.nonzeroes() ) yield List(r)
			a.nonzeroset = r1
			a
		}
	}
	class SparseMatrix extends Sparse {
		protected override def create() = new SparseMatrix
		def +(sm : SparseMatrix) : SparseMatrix = {
			val a = create()
			a.nonzeroset = this.nonzeroes() ++ sm.nonzeroes()
			a
		}
		def read(e : Value, e2 : Value) { read(List(e,e2)) }
		def write(e : Value, e2 : Value) { write(List(e,e2)) }
		def scalar() : SparseMatrix = {
			val a = create()
			a.nonzeroset = this.nonzeroes()
			a
		}

	}
	
	case class Field(fi : FieldInit) extends LObject {
		var readset : SSet[Topology.Node] = SSet.empty
		var writeset : SSet[Topology.Node] = SSet.empty
		def read(t : Topology.Node) {
			readset += t
		}
		def write(t : Topology.Node) {
			writeset += t
		}
		def draw(gv : GraphViz) {
			for(r <- readset)
				gv(this -> r, "lightgray")
			for(w <- writeset)
				gv(this -> w, "gray")
		}
		override def meet(rhs : Value) : Field = rhs match {
			case rhs : Field =>
			val r = Field(fi)
			r.readset = readset ++ rhs.readset
			r.writeset = writeset ++ rhs.writeset
			r
		}
		override def toString = "Field: %s reads = %s, writes = %s".format(fi,readset,writeset)
	}
	
	
	/*  the member of nodes distinguish unique forall sets in the description
	    showing the difference between
	  
	  code:
	  for(f <- faces(mesh)) {
	    for(c <- cells(f)) {
	    	val c2 = towards(c,f)
	    }
	  }
	  
	  set builder notation:
	  
	  {c2 | c2 in { towards(c,f) } and c in cell(f) and f in faces(mesh) }
	  
	  
	  data structure: 
	  
	  mesh
	    |
	  MemberOf(faces)---
	    |               |
	  MemberOf(cells)   |
	    |              /
	  MemberOf(towards)
	  
	  
	  and
	  
	  
	  code:
	  
	  for(f <- faces(mesh)) {
	  	for(f2 <- faces(mesh)) {
	  		 
	  		 for(c <- cells(f)) {
	  		 	val c2 = towards(c,f2)
	  		 }
	  	}
	  }
	  
	  set builder notation:
	  { c2 | c2 in { towards(c,f2) } and c in cells(f) and f in cells(mesh) and f2 in faces(mesh) }
	  
	  data structure:
	  mesh----------------
	    |                 |
	  MemberOf(faces)    MemberOf(faces)
	    |                 |
	  MemberOf(cells)     |
	    |                /
	  MemberOf(towards)
	  
	  This distinction doesn't matter too much for pure mesh operations,
	  but will matter for row/column pairing on sparse matrices where the difference
	  between the two codes above determines a very different sparsity pattern
	*/
	
	object ObjectIDs {
		var nextID = 0
		def get() = {
			val i = nextID
			nextID += 1
			i
		}
	}
	case class Record[T <: Value](val depth : Int, val value : T)
	
	class Frame(var parent : Option[Frame]) {
		val depth : Int = parent match {
			case Some(p) => p.depth + 1
			case None => 0
		}
		import scala.collection.mutable.{ HashMap => VMap }
		
		val values = new VMap[String,Record[Value]]
		val objects = new VMap[ObjectID,Record[Value]]
		def newChild() : Frame = new Frame(Some(this))
		
		def mergeWith(f : Frame)( fn : (Value,Value) => Value ) {
			assert(this.depth <= f.depth)
			def perform[T,V <: Value](m : VMap[T,Record[V]], rhs : VMap[T,Record[V]]) {
				for( (k,Record(d,v)) <- rhs if d <= this.depth) {
					m.get(k) match {
						case Some(Record(d2,v2)) => {
							assert(d == d2)
							m(k) = Record(d,fn(v2,v).asInstanceOf[V])
						}
						case None => {
							m(k) = Record(d,v)
						}
					}
				}
			}
			perform(values,f.values)
			perform(objects,f.objects)
		}
		private def get(k : String) : Record[Value] = {
			Log.log("(%d) finding value %s",depth,k)
			values.getOrElse(k,parent.get.get(k))
		}
		private def getObj(k : ObjectID) : Record[Value] = {
			Log.log("(%d) finding object %s",depth,k)
			objects.getOrElse(k,parent.get.getObj(k))
		}
		def apply(k : String) : Value = {
			get(k).value
		}
		def define(k : String, v : Value) = {
			Log.log("(%d) defining %s = %s",depth,k,v)
			values(k) = Record(depth,v)
		}
		def assign(k : String, v : Value) = {
			Log.log("(%d) assigning %s = %s",depth,k,v)
			values(k) = Record(get(k).depth,v)
		}
		def newObj(v : Value) : Reference = {
			val i = ObjectIDs.get()
			objects(i) = Record(depth,v)
			Reference(i)
		}
		def dereferenceValue[T <: Value](r : Value, const : Boolean = true) : List[T] = {
			def points_to(r : Value) : List[ObjectID] = {
				r match {
					case Reference(i) => List(i)
					case Union(a,b) => points_to(a) ::: points_to(b)
					case _ => Log.ierror("not a pointer: %s",r)
				}
			}
			for(p <- points_to(r)) yield {
				val Record(d,o) = getObj(p)
				if(const) o.asInstanceOf[T] else {
					val cp = o.copy().asInstanceOf[T]
					objects(p) = Record(d,cp)
					cp
				}
			}
		}
		def dereference[T <: Value](n : String, const : Boolean = true) : List[T] = dereferenceValue[T](this.apply(n),const)
		def clear() {
			values.clear()
			objects.clear()
		}
	}
	def create(top : CompilationUnit) : (Topology.Graph,Map[Workgroup,List[WorkgroupWrite]],Map[Topology.Node, List[(Boolean,FieldInit)]]) = {
		val tgraph = new Topology.Graph
		def hasAbstractValue(t : Type) = {
			t match {
				case _ : Liszt.Types.Set => true
				case _ : Topo => true
				case _ : SparseVectorType => true
				case _ : SparseMatrixType => true
				case VectorT(_,_: IndexType) => true
				case _ => false
			}
		}
		val code_block = top.code
		val function_table : Map[String,DefFunc] = Map.empty ++ (
			for( df @ DefFunc(t,name,args,body) <- code_block.stmts) 
				yield df.name -> df
		)
		var forall_context : SSet[Topology.Node] = SSet(Topology.TopLevel)
		
		var current_workgroup_writes : List[WorkgroupWrite] = List.empty 
		var workgroup_info : Map[Workgroup,List[WorkgroupWrite]] = Map.empty
		
		
		var field_list : List[Reference] = List.empty
		
		def interpStmt(call_frame : Frame)(ctx : Frame)(stmt : Stmt) {
			Log.log("INTERP: %s",stmt.toShortString)
			val iStmt = interpStmt(call_frame)(ctx) _
			val iExp = interpExp(call_frame)(ctx) _
			def iPredicate(blk : Block) { //predicates, unlike while or if bodies don't create new contexts
				for(stmt <- blk.stmts)
					iStmt(stmt)
			}
			stmt match {
				case Block(stmts,v) => {
					val ctxn = ctx.newChild()
					for(stmt <- stmts)
						interpStmt(call_frame)(ctxn)(stmt)
					ctx.mergeWith(ctxn) { (_,t) => t }
				}
				case If(p,t,e) => {
					val old_forall = forall_context
					iPredicate(p)
					val tctx = ctx.newChild()
					val ectx = ctx.newChild()
					interpStmt(call_frame)(tctx)(t)
					forall_context = old_forall
					interpStmt(call_frame)(ectx)(e)
					forall_context = old_forall
					ctx.mergeWith(ectx) { (_,t) => t }
					ctx.mergeWith(tctx) { _ meet _ }
				}
				case While(p,b) => {
					val old_forall = forall_context
					//TODO(zach): may need to interate to fixed point when we deal with re-binding matrix stuff, pointers changing
					iPredicate(p)
					iStmt(b)
					forall_context = old_forall
				}
				case Forall(df,lbl,st,Block(stmts,_)) => {
					val old_forall = forall_context
					
					val ctxn = ctx.newChild()
					val TopoSet(nodes) = ctx(st.name)
					
					forall_context = for(n <- nodes; ctx <- old_forall if tgraph.avaliableIn(n,ctx)) 
						yield tgraph.forall(ctx,n)
						
					ctxn.define(df.name, TopoSet(forall_context) )
					for( s <- stmts) { 
						interpStmt(call_frame)(ctxn)(s)
					}
					ctx.mergeWith(ctxn) { (_,t) => t }
					forall_context = old_forall
				}
				//TODO: LinearSystemNonzeroes
				case DefVar(name,SetInit(t,str)) => {
					val st = for(ctx <- forall_context) 
						yield tgraph.boundary(ctx,t,name,str)
					ctx.define(name,TopoSet(st))
				}
				case DefVar(name,rhs) => {
					iExp(rhs) match {
						case Some(v) => ctx.define(name,v)
						case _ =>
					}
				}
				case Return(v : Var) => iExp(v) match {
					case Some(v) => {
						ctx.assign("$return_value",v)
						call_frame.mergeWith(ctx) { _ meet _ }
						ctx.clear()
					}
					case None => {
						ctx.clear()
					}
				}
				case Assign(SparseVectorAccess(t,vec,idx),op,rhs) => {
					val vs = ctx.dereference[SparseVector](vec.name,false)
					val si = ctx(idx.name)
					for(v <- vs) {
						v.write(si)
					}
				}
				case Assign(SparseMatrixAccess(t,vec,row,col),op,rhs) => {
					val vs = ctx.dereference[SparseMatrix](vec.name,false)
					val rowi = ctx(row.name)
					val coli = ctx(col.name)
					for(v <- vs) {
						v.write(rowi,coli)
					}
				}
				case Assign(v : Var,Ops.Assign,rhs) if hasAbstractValue(rhs.t) => {
					ctx.assign(v.name,ctx(rhs.name))
				}
				case FieldWrite(f,e,op,rhs) => {
					val fs = ctx.dereference[Field](f.name)
					val TopoSet(es) = ctx(e.name)
					for(f <- fs) {
						for(e <- es) {
							f.write(e)
							for(ctx <- forall_context) {
								if(tgraph.avaliableIn(e,ctx)) {
									current_workgroup_writes ::= new WorkgroupWrite(stmt,ctx,FieldLocation(f.fi.t,f.fi,e),tgraph)
								}
							}
						}
						
					}
				}
				case wg @ Workgroup(stmts) => {
					val parent_workgroup_writes = current_workgroup_writes
					current_workgroup_writes = List.empty
					for(stmt <- stmts) {
						iStmt(stmt)
					}
					//TODO(zach): BUG BUG BUG, if workgroup is called from more than one dynamic context, only the last one is remembered
					workgroup_info += wg -> current_workgroup_writes
					current_workgroup_writes :::= parent_workgroup_writes
				}
				case ExpStmt(e) => iExp(e)
				case _ =>
			}
		}
		def interpExp(call_frame : Frame)(ctx : Frame)(exp : Exp) : Option[Value] = {
			val iStmt = interpStmt(call_frame)(ctx) _
			exp match {
				case Call(t,func,List()) if func.name == "mesh" => {
					Some(TopoSet(SSet(Topology.TopLevel))) 
				}
				case Call(t,func,args) => {
					Log.log("calling %s",func)
					def funct = func.t.asInstanceOf[Func]
					function_table.get(func.name) match {
						case Some(df) => {
							val call_framen = ctx.newChild()
							call_framen.define("$return_value",TopValue)
							val function_ctx = call_framen.newChild()
							val argvs = for(a <- args) yield interpExp(call_frame)(ctx)(a)
							val argfs = for(dv <- df.args) yield dv.name
							for( (k,Some(v)) <- argfs zip argvs)
								function_ctx.define(k,v)
								
							interpStmt(call_framen)(function_ctx)(df.body)
							call_framen.mergeWith(function_ctx) { (_,t) => t }
							if(hasAbstractValue(funct.range)) 
								Some(call_framen("$return_value")) 
							else
								None
						}
						case None => {
							def warn() = Log.log("calling an external function %s, assuming that value returned can point to anything and that arguments do not use topological information",func.name)
							exp.t match {
								case mo : Topo => Log.error("External functions cannot return mesh objects: %s",func.name)
								case Set(mo) => Log.error("External functions cannot return mesh objects: %s",func.name)
								case _ => {
									if(funct.args exists hasAbstractValue)
										warn()
									None
								}
							}
						}
					}
				}
				case Var(t,name) => if(hasAbstractValue(t)) Some(ctx(name)) else None
				case IfExp(_,p,t,e) => {
					for(stmt <- p.stmts)
						iStmt(stmt)
					def interpBranch( b : Block) : (Option[Value], Frame) = {
						val ctxn = ctx.newChild()
						for( s <- b.stmts) {
							interpStmt(call_frame)(ctxn)(s)
						}
						(interpExp(call_frame)(ctxn)(b.exp),ctxn)
					}
					val old_forall = forall_context
					val (tv,tctx) = interpBranch(t)
					val tforall = forall_context
					forall_context = old_forall
					val (ev,ectx) = interpBranch(e)
					val eforall = forall_context
					forall_context = eforall ++ tforall 
					ctx.mergeWith(tctx) { (_,t) => t }
					ctx.mergeWith(ectx) { _ meet _ }
					(tv,ev) match {
						case (Some(a),Some(b)) => Some(a meet b)
						case (None,None) => None
						case _ => Log.ierror("one branch of if producing a interpreter value but the other is that; that is weird... %s %s",tv,ev)
					}
				}
				case TopoCall(_,op,args) => {
					val argvs = op match {
						case _ : Ops.LabelExtract => List(ctx(args.head.name)) //second argument does not have value in abstract interp (it is any integer)
						case _ => for( a <- args) yield ctx(a.name)
					}
					val nodes : SSet[Topology.Node] = op match {
						case _ : Ops.Towards => {
							val TopoSet(tos) = argvs(0)
							val TopoSet(orienteds) = argvs(1)
							for(to <- tos; oriented <- orienteds; ctx <- forall_context if tgraph.avaliableIn(to,ctx) && tgraph.avaliableIn(oriented,ctx))
								yield tgraph.orientedTopo(ctx, op, to, oriented)
						}
						case op : Ops.LabelExtract => {
							val temp_name = ASTUtil.Names.create("ai")
							val TopoSet(base) = argvs(0)
							val nodes = for(node <- base; ctx <- forall_context if tgraph.avaliableIn(node,ctx))
								yield tgraph.simpleTopo(ctx, op.asTopoSet, node)
							val new_ctx = for(node <- nodes; ctx <- forall_context if tgraph.avaliableIn(node,ctx))
								yield tgraph.forall(ctx,node)
							//we assume that the label extract can refer to any of the elements in the set it is extracting from
							//so we must insert a forall statement here so that the value will refer to all possible extractions
							//this scope of this forall statement extends until we reach the end of the forall statement it was created in
							forall_context = new_ctx //changing the forall context here means that all statments following this become nested under the label extract 
							new_ctx
						}
						case _ => {
							val TopoSet(base) = argvs(0)
							for(node <- base; ctx <- forall_context if tgraph.avaliableIn(node,ctx)) 
								yield tgraph.simpleTopo(ctx, op, node)
						}
					}
					Some(TopoSet(nodes))
				}
				case EmptyExp(t) if t.isInstanceOf[Topo] || t.isInstanceOf[Liszt.Types.Set] => {
					Some(TopoSet(SSet.empty))
				}
				//TODO(zach): we need a new object that represents an index that points
				//back to the nodes in the topology graph that generated the index
				case IndexCall(_,ls,idx,args) => Some(TopoSet(SSet.empty))
				
				case fi @ FieldInit(t,s) => {
					val r = ctx.newObj(Field(fi))
					field_list ::= r
					Some(r)
				}
				
				case SparseVectorAccess(t,vec,idx) => {
					val vs = ctx.dereference[SparseVector](vec.name)
					val si = ctx(idx.name)
					for(v <- vs) {
						v.read(si)
					}
					None
				}
				case SparseMatrixAccess(t,vec,row,col) => {
					val vs = ctx.dereference[SparseMatrix](vec.name)
					val rowi = ctx(row.name)
					val coli = ctx(col.name)
					for(v <- vs) {
						v.read(rowi,coli)
					}
					None
				}
				case FieldRead(t,f,e) => {
					val vs = ctx.dereference[Field](f.name)
					val TopoSet(nodes) = ctx(e.name)
					for(n <- nodes; f <- vs) {
						f.read(n)
					}
					None
				}
				case LinearSystemAllocate(t,linear_system,elem) => Some(ctx.newObj(
					elem match {
						case LS_A => new SparseMatrix
						case LS_X => new SparseVector
						case LS_B => new SparseVector
					}))
				//TODO: indices can be put in vec literals, so we need to represent vec literals
				//as well
				case VectorL(VectorT(_,_ : IndexType),es) => Some(TopoSet(SSet.empty))
				
				/*case Binary(_, Var(SparseVectorType,lhs), op, Var(SparseVectorType,rhs) ) => {
					val lhsv = ctx.dereference[SparseVector](lhs)
					val rhsv = ctx.dereference[SparseVector](rhs)
					val objs : List[Value] = for(l <- lhsv; r <- rhsv) yield l + r
					Some(ctx.newObj(objs reduceRight { _ meet _ }))
				}
				case Binary(_, Var(SparseMatrixType, lhs), op, Var(SparseMatrixType, rhs) ) => {
					val lhsv = ctx.dereference[SparseMatrix](lhs)
					val rhsv = ctx.dereference[SparseMatrix](rhs)
					val objs : List[Value] = for(l <- lhsv; r <- rhsv) yield l + r
					Some(ctx.newObj(objs reduceRight { _ meet _ }))
				}
				case Binary(_, lhs, op, Var(SparseMatrixType, rhs)) => {
					val rhsv = ctx.dereference[SparseMatrix](rhs)
					val objs : List[Value] = for(r <- rhsv) yield r.scalar()
					Some(ctx.newObj(objs reduceRight { _ meet _ }))
				}
				case Binary(_, Var(SparseMatrixType, lhs), op, Var(SparseVectorType, rhs) ) => {
					val lhsv = ctx.dereference[SparseMatrix](lhs)
					val rhsv = ctx.dereference[SparseVector](rhs)
					val objs : List[Value] = for(l <- lhsv; r <- rhsv) yield r * l
					Some(ctx.newObj(objs reduceRight { _ meet _ }))
				}
				case Binary(_, lhs, op, Var(SparseVectorType, rhs)) => {
					val rhsv = ctx.dereference[SparseVector](rhs)
					val objs : List[Value] = for(r <- rhsv) yield r.scalar()
					Some(ctx.newObj(objs reduceRight { _ meet _ }))
				}*/
				case _ => None
			}
		}
		Log.log("Program is\n%s",code_block)
		
		val top_level = new Frame(None)
		for(stmt <- code_block.stmts) {
			interpStmt(top_level)(top_level)(stmt)
		}
		interpStmt(top_level)(top_level)(function_table("_liszt_start").body)
		val gv = new GraphViz
		
		val field_info : Map[Topology.Node, List[(Boolean,FieldInit)]]= {
			val m = new scala.collection.mutable.HashMap[Topology.Node,List[(Boolean,FieldInit)]]
			val fields = for(fr <- field_list; f <- top_level.dereferenceValue[Field](fr)) 
				yield f 
			for(f <- fields; n <- f.readset) {
				m(n) = (true,f.fi) :: m.getOrElse(n,List.empty)
			}
			for(f <- fields; n <- f.writeset) {
				m(n) = (false,f.fi) :: m.getOrElse(n,List.empty)
			}
			Map.empty ++ m
		}
		
		for( (wg,infos) <- workgroup_info ) {
			Log.log("*"*80)
			Log.log("%s",wg)
			if(!infos.isEmpty) {
				Log.log(">"*80)
				for(info <- infos)
					Log.log(info)
			}
		}
		//tgraph.draw(gv)
		//gv.draw(field_info.asInstanceOf[Map[AnyRef,List[Field]]])
		(tgraph,workgroup_info,field_info)
	}
}