/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Liszt.CFGs

import Liszt._
import Liszt.Types._
import Liszt.Trees._

import scala.collection.immutable.{Set => SET}
import scala.collection.mutable.ListBuffer
/*
 CFG provides an interface to convert the Liszt AST into a CFG, insert statements into the CFG, and
 then restort the AST with the changes intact.  If statements and While statements annotate BasicBlocks
 such that we can restore them to ASTs when finished.
 */
object Log extends Liszt.Log("CFG")
abstract sealed class BasicBlockType
case object BlockHeader extends BasicBlockType
case object IfHeader extends BasicBlockType
case class IfExpHeader(typ : Type) extends BasicBlockType
case object WhileHeader extends BasicBlockType
case class ReturnHeader(dead_code : BasicBlock) extends BasicBlockType
case object Exit extends BasicBlockType

class BasicBlock {
  var typ : BasicBlockType = BlockHeader
  def withType(t : BasicBlockType) : this.type = {
  	typ = t
  	this
  }
  def asFooter(exp : SimpleExp) : this.type = {
  	expOption = Some(exp)
  	this
  }
  val stmts : ListBuffer[Stmt] = new ListBuffer[Stmt]
  def exp : SimpleExp = expOption.get
  var expOption : Option[SimpleExp] = None
  def trees : List[Tree] = stmts.toList ::: expOption.toList
  def isFooter = expOption match { case None => false; case Some(_) => true }
  val predecessors = new ListBuffer[BasicBlock]
  val successors = new ListBuffer[BasicBlock]
  override def toString = {
  	"BasicBlock(%s) { pred = %s, succ = %s, trees = %s }".format(typ,predecessors.size,successors.size,trees)
  }
}

//TODO(zach): we need to keep track of the old ast nodes so that when we restore them we retain their attributes (e.g. file position)
class CFG(val deffunc : DefFunc) {
	val (entry : BasicBlock,exit : BasicBlock) = {
		val func_exit = new BasicBlock withType Exit asFooter EmptyExp(UnitT)
		def makeBlock(block : Block) : (BasicBlock,BasicBlock) = {
			val first = new BasicBlock
			var current = first
			
			def makeIf(head_type : BasicBlockType, predE : Block, thenE : Block, elseE : Block) {
				val footer = new BasicBlock withType BlockHeader
				
			
				val (header,header_end) = makeBlock(predE)
				
				val (tstart,tend) = makeBlock(thenE)
				val (estart,eend) = makeBlock(elseE)
				
				current.successors += header
				header_end.successors += tstart
				header_end.successors += estart
				
				tend.successors += footer
				eend.successors += footer
				header withType head_type
				current = footer
			}
			def makeIfExp(typ : Type, name : String, vf : Stmt, p : Block, t : Block, e : Block) {
				val at = ASTUtil.mkAssign(name,Liszt.Ops.Assign,t.exp)
				val ae = ASTUtil.mkAssign(name,Liszt.Ops.Assign,e.exp)
						
				val np = Block(vf :: p.stmts,p.exp) copyAttr p
				val tp = Block(t.stmts ::: at,t.exp) copyAttr t
				val ep = Block(e.stmts ::: ae,e.exp) copyAttr e
						
				makeIf(IfExpHeader(typ),np,tp,ep)
			}
			for(stmt <- block.stmts) {
				stmt match {
					case Return(ee) => {
					   val dead_code = new BasicBlock withType BlockHeader
					   val return_block = new BasicBlock withType ReturnHeader(dead_code)
					   
					   return_block.stmts += stmt
					   current.successors += return_block
					   return_block.successors += func_exit
					   
					   current = dead_code //the rest is dead code
					}
					//for if/while constructs, a new basic block will be created with
					//the predicate in it. The type of this block will reflect the statement
					//it represents, so that we can reconstruct the AST after we are finished doing
					//control flow
					case While(pred,block) => {
						
						val footer = new BasicBlock withType BlockHeader
						
						val (header,header_end) = makeBlock(pred)
						current.successors += header
						
						header withType WhileHeader
						
						header_end.successors += footer
						val (start,end) = makeBlock(block)
						header_end.successors += start
						end.successors += header
					
						current = footer
					}
					case If(predE,thenE,elseE) => makeIf(IfHeader,predE,thenE,elseE)
					case DefVar(name,IfExp(typ,p,t,e)) => makeIfExp(typ,name,DefVar(name,EmptyExp(typ)) copyAttr stmt,p,t,e)
					case _ => current.stmts += stmt
				}
			}
			(first,current asFooter block.exp)
		}
		deffunc.body match {
			case ExternStmt => {
				val entry = new BasicBlock withType BlockHeader
				entry.successors += func_exit
				(entry,func_exit)
			}
			case blk @ Block(stmts,exp) => {
				val (entry,last) = makeBlock(blk)
				last.successors += func_exit
				(entry,func_exit)
			}
			case stmt => {
				val (entry,last) = makeBlock( Block(List(stmt),EmptyExp(UnitT)) )
				last.successors += func_exit
				(entry,func_exit)
			}
		}
	}
  val blocks : List[BasicBlock] = {
     val visited = scala.collection.mutable.Set[BasicBlock]()
     def visit(bb : BasicBlock) {
       if(!visited.contains(bb)) {
         visited += bb
         for(b <- bb.successors) {
           visit(b)
         }
       }
     }
     visit(entry)
     visited.toList
  }
  for(b <- blocks; s <- b.successors) {
    s.predecessors += b
  }
  
  def toptrees : List[Tree] = for(blk <- blocks; t <- blk.trees) yield t
  def alltrees : List[Tree] = toptrees.flatMap{ stmt => 
  	stmt :: stmt.map(x => x)
  }
  
  def modify( fn : (Stmt) => List[Stmt] ) {
    for(b <- blocks) {
       val nstmts = for(e <- b.stmts; ne <- fn(e)) yield ne
       //val nexp = for(exp <- b.expOption) 
       //		yield fn(exp).asInstanceOf[SimpleExp]
       b.stmts.clear
       b.stmts ++= nstmts
       //b.expOption = nexp
    }
  }
  
  def toAST : DefFunc = {
    def tBlock(bb : BasicBlock) : (Block, List[BasicBlock]) = {
    	var cur_bb = bb
    	val stmts = new ListBuffer[Stmt]
    	while(!cur_bb.isFooter) {
    		stmts ++= cur_bb.stmts
    		val List(succ) = cur_bb.successors.toList
    		val (nstmts,next) = tStmt(succ)
    		stmts ++= nstmts
    		assert(!(cur_bb eq next))
    		cur_bb = next
    	}
    	(Block(stmts.toList ::: cur_bb.stmts.toList, cur_bb.exp), cur_bb.successors.toList)
    }
    def tStmt(bb : BasicBlock) : (List[Stmt],BasicBlock) = {
      bb.typ match {
        case IfHeader => {
            val (predE,List(thenbb,elsebb)) = tBlock(bb)
			val (thenE,List(s0)) = tBlock(thenbb)
			val (elseE,List(s1)) = tBlock(elsebb)
			
			assert(s0 eq s1)
			val newIf = If(predE,thenE,elseE)
			(List(newIf),s0)
        }
        case IfExpHeader(typ) => {
        	val (predE,List(thenbb,elsebb)) = tBlock(bb)
			val (thenE,List(s0)) = tBlock(thenbb)
			val (elseE,List(s1)) = tBlock(elsebb)
			assert(s0 eq s1)
			val DefVar(name,_ : EmptyExp) = predE.stmts.head
			val np = Block(predE.stmts.tail,predE.exp) copyAttr predE
			val nt = Block(thenE.stmts.init,thenE.exp) copyAttr thenE
			val ne = Block(elseE.stmts.init,elseE.exp) copyAttr elseE
			val newStmt = DefVar(name,IfExp(typ,np,nt,ne))
			(List(newStmt),s0)
        }
        case WhileHeader => {
        	val(predE,List(footer,body)) = tBlock(bb)
			val (bodyE,List(h)) = tBlock(body) //this header BB is the body's exit point
			assert(h eq bb)
			val newWhile = While(predE,bodyE)
			(List(newWhile),footer)
        }
        case BlockHeader => {
        	(List(),bb)
        }
        case ReturnHeader(dead_code) => {
        	(bb.stmts.toList,dead_code)
        }
        case _ => Log.ierror("Unexpected Header Type %s", bb.typ)
      }
    }
    val (body,List(exitb)) = tBlock(entry)
    assert(exitb eq exit)
    DefFunc(deffunc.ftype,deffunc.name,deffunc.args,body) copyAttr deffunc
  }
  def draw() {
    draw( (_) => "", (_) => "")
  }
  def draw(begin : BasicBlock => Any, end : BasicBlock => Any) {
    val header =
"""
digraph g {
graph [
rankdir = "TB"
];
node [
fontsize = "16"
shape = "ellipse"
];
edge [
];
"""
    val footer = "}"
    val blktoid = Map.empty ++ blocks.zipWithIndex
    val nodefmt = """ "%d" [ label = "{ <f0> %s: %s|%s | %s | <f1> }" shape = "record" ]; """
    val edgefmt = """ "%d":f1 -> "%d":f0; """
    def quote(s : Any) : String = {
      val q = List("<",">","{","}")
      val r = q.foldLeft(s.toString) { (str:String,q:String) => str.replace(q,"\\"+q) }
      r.replace("\n","\\l")
    }
    def fmtes(es : List[Tree]) = (for(i <- es) yield quote(i.toString)).mkString("","\\l| ","\\l")
    def fmtype(bt : BasicBlockType) = bt match {
      case BlockHeader => "Block"
      case Exit => "Exit Node "
      case WhileHeader => "While Header"
      case IfHeader => "If Header"
      case IfExpHeader(t) => "IfExp Header %s".format(t)
      case ReturnHeader(_) => "Return Header"
    }
    val nodes = for((b,i) <- blocks.zipWithIndex)
      yield nodefmt.format(i,fmtype(b.typ),quote(begin(b)),fmtes(b.trees),quote(end(b)))
    val edges = for(b <- blocks; s <- b.successors) yield edgefmt.format(blktoid(b),blktoid(s))
    val dotstring = header + nodes.mkString("","\n","\n") + edges.mkString("","\n","\n") + footer
    val proc = Runtime.getRuntime().exec("dot -Tx11")
    val out = new java.io.PrintWriter(proc.getOutputStream())
    out.println(dotstring)
    out.close()
  }
}


trait OptimizationPass {
  type V
  def top : V
  def bottom : V
  def meet(a : V, B : V) : V
  def isForward : Boolean
  def inputValue : V
  def cfg : CFG
  def transferFn(e : Tree)(v : V):V
  lazy val (inmap,outmap) = run()
  private def run() : (Map[BasicBlock,V], Map[BasicBlock,V]) = {
    val inmap = scala.collection.mutable.Map[BasicBlock,V]()
    val outmap = scala.collection.mutable.Map[BasicBlock,V]()
    def tfn(bb : BasicBlock)(v : V) =
      if(isForward)
        bb.trees.map(transferFn).foldLeft(v)((v,f) => f(v))
      else
        bb.trees.map(transferFn).foldRight(v)((f,v) => f(v))
    for(b <- cfg.blocks) {
      inmap(b) = top
      outmap(b) = top
    }
    if(isForward) {
      inmap(cfg.entry) = inputValue
    } else {
      outmap(cfg.exit) = inputValue
    }
    var changed = true
    while(changed) {
      changed = false
      if(isForward) {
        for(b <- cfg.blocks if b != cfg.entry) {
           val input = b.predecessors.map(outmap).reduceLeft(meet)
           if(input != inmap(b))
             changed = true
           inmap(b) = input
        }
        for(b <- cfg.blocks) {
          val output = tfn(b)(inmap(b))
          if(output != outmap(b))
            changed = true
          outmap(b) = output
        }
      } else {
        for(b <- cfg.blocks if b != cfg.exit) {
           val input = b.successors.map(inmap).reduceLeft(meet)
           if(input != outmap(b))
             changed = true
           outmap(b) = input
        }
        for(b <- cfg.blocks) {
          val output = tfn(b)(outmap(b))
          if(output != inmap(b))
            changed = true
          inmap(b) = output
        }
      }
    }
    (Map.empty ++ inmap,Map.empty ++ outmap)
  }
  def scanLeft[A,B](es : List[A], init : B, fn : (B,A) => B) : List[B] = {
    val rlist = es.foldLeft(List(init)) { (i,e) => fn(i.head,e) :: i }
    rlist.reverse
  }
  //expMod will be given an expression e and its input and output values for this
  //analysis, the output is a list of expressions to replace e
  def modifyBlocks(expMod : (Option[Stmt],V,V) => List[Stmt]) {
    def modifyBlock(bb : BasicBlock) {
      val successors = bb.successors
      val predecessors = bb.predecessors
     
      val newes : List[Stmt] =
      if(isForward) {
        val v = inmap(bb)
        val values = scanLeft(bb.stmts.toList,v,(a:V,b:Stmt) => transferFn(b)(a))
        for( ((e,i),o) <- bb.stmts.toList.zip(values.init).zip(values.tail);
              ne <- expMod(Some(e),i,o)) yield ne

      } else {
        val v = outmap(bb)
        val res = bb.stmts.toList.reverse
        val values = scanLeft(bb.stmts.toList,v,(a:V,b:Stmt) => transferFn(b)(a))
        for( ((e,i),o) <- res.zip(values.tail).zip(values.init);
             ne <- expMod(Some(e),i,o)) yield ne
      }

      val beginstmts = if(bb.predecessors.size == 1) {
        expMod(None,outmap(bb.predecessors.head),inmap(bb))
      } else List()
      val endstmts = if(bb.successors.size == 1) {
        expMod(None,outmap(bb),inmap(bb.successors.head))
      } else List()
      bb.stmts.clear
      bb.stmts ++= (beginstmts ::: newes ::: endstmts)
    }
    for(b <- cfg.blocks) {
      modifyBlock(b)
    }
  }
}





class CFG2(val df : DefFunc) {
	
	type Node = Int
	case class Edge(from : Node,to : Node)
	
	
	//values stored on CFG, each node has a value defined at entry and at exit
	class Attribute[V <: AnyRef] {
		class Entry {
			var in : V = null.asInstanceOf[V]
			var out : V = null.asInstanceOf[V]
		}
		private val entries : Array[Entry] = new Array[Entry](nodes.length)
		for(i <- 0 until nodes.length)
			entries(i) = new Entry
			
		def apply(n : Node) : Entry = entries(n)
		def map[V2 <: AnyRef](fn : V => V2) : Attribute[V2] = {
			val a2 = new Attribute[V2]
			for(i <- 0 until entries.length) {
				a2(i).in = fn(this(i).in)
				a2(i).out = fn(this(i).out)
			}
			a2
		}
		def zipFunction[V2 <: AnyRef,V3 <: AnyRef](a2 : Attribute[V2])(fn : (V,V2) => V3) : Attribute[V3] = {
			val a3 = new Attribute[V3]
			for( i <- 0 until entries.length) {
				a3(i).in = fn(this(i).in,a2(i).in)
				a3(i).out = fn(this(i).out,a2(i).out)
			}
			a3
		}
		def zip[V2 <: AnyRef](a2 : Attribute[V2]) : Attribute[(V,V2)] = zipFunction(a2) { (a,b) => (a,b) }
	}
	def newAttribute[V <: AnyRef](default : V) : Attribute[V] = {
		val a = new Attribute[V]
		for(n <- 0 until stmts.length) {
			a(n).in = default
			a(n).out = default
		}
		a
	}
	
	def entry : Node = 0
	def exit : Node = 1
	def first_node : Node = 2
	
	class Replacement
	case class ControlFlowReplacement(stmt : Stmt) extends Replacement
	case class NodeReplacement(node : Node) extends Replacement
	
	val (replacement_map,edges,stmts,artificial_nodes) = init()
	
	def init() = {
		val blk_map = new scala.collection.mutable.HashMap[Int,List[Replacement]]()
		var node_list : List[Option[Tree]] = List(None,None)
		var artificial : List[Node] = List.empty
		var edges : List[Edge] = List.empty
		
		var next_node = node_list.length
		
		def emptyNode() : Node = {
			node_list ::= None
			next_node += 1
			next_node - 1
			
		}
		def newNode(tree : Tree) : Node = {
			node_list ::= Some(tree)
			next_node += 1
			next_node - 1
		}
		
		def newEdge(from : Node, to : Node) {
			edges ::= Edge(from,to)
		}
		
		//adds block to CFG, returns (entry,exit) nodes for the block
		def addBlockStmts(blkid : Int, stmts : List[Stmt]) : (Node,Node) = {
		
			val start_end = stmts map addStmt
			val (initial,start_footer,end_footer) = if(stmts.isEmpty) {
				val n = emptyNode()
				(List.empty,n,n)
			} else {
				(start_end.init,start_end.last._1,start_end.last._2)
			}
			val header = initial.foldRight(start_footer) {
				case ( (start,end,repl) , next_node) =>
					newEdge(end,next_node)
					start
			}
			val replacements = start_end.flatMap {
				case (_,_,repls) => repls
			}
			
			blk_map(blkid) = replacements
			(header,end_footer)
		}
		def addBlock(blk : Block) : (Node,Node) = addBlockStmts(blk.id,blk.stmts)
		
		def addPredicate(blk : Block) : (Node,Node) = {
			val (start,end) = addBlockStmts(blk.id,blk.stmts ::: List(ExpStmt(blk.exp)))
			artificial ::= end
			(start,end)
		}
		//create the AST structure for a statment, 
		def addStmt(stmt : Stmt) : (Node,Node,List[Replacement]) = stmt match {
			case blk : Block => {
				val (s,e) = addBlock(blk)
				(s,e,List(ControlFlowReplacement(stmt)))
			}
			case While(pred,blk) => {
				//  [ loop header ] <--
				//     |   |           |
				//     |  [loop body] -
				//   [ footer ]
				// the footer ensures we dont create a node with multiple successors that connects
				//to a node with multiple predecessors
				
				val (pstart,pend) = addPredicate(pred)
				val (bstart,bend) = addBlock(blk)
				newEdge(pend,bstart)
				newEdge(bend,pstart)
				val footer = emptyNode()
				newEdge(pend,footer)
				val repl = List( ControlFlowReplacement(stmt), NodeReplacement(footer) )
				(pstart,footer,repl)
			}
			case If(p,t,e) => {
			//       [predicate]
			//        /        \
			//[true-branch] [false-branch]
			//        \       /
			//        [footer]
			//footer ensures we have a single node to return to the enclosing block
				val (ps,pe) = addPredicate(p)
				val (ts,te) = addBlock(t)
				val (es,ee) = addBlock(e)
				newEdge(pe,ts)
				newEdge(pe,es)
				val footer = emptyNode()
				newEdge(te,footer)
				newEdge(ee,footer)
				val repl = List( ControlFlowReplacement(stmt), NodeReplacement(footer) )
				(ps,footer,repl)
			}
			case DefVar(v,IfExp(tpe,p,t,e)) => {
				val (start,end,init) = addStmt(DefVar(v,EmptyExp(tpe)))
				artificial ::= start
				def asAssign(blk : Block) : List[Stmt] = blk.stmts ::: List(Assign(Var(tpe,v),Ops.Assign,blk.exp.asInstanceOf[Var]))
				
				val (ps,pe) = addPredicate(p)
				val (ts,te) = addBlockStmts(t.id,asAssign(t))
				val (es,ee) = addBlockStmts(e.id,asAssign(e))
				
				newEdge(end,ps)
				newEdge(pe,ts)
				newEdge(pe,es)
				val footer = emptyNode()
				newEdge(te,footer)
				newEdge(ee,footer)
				
				val repl = init ::: List( ControlFlowReplacement(stmt), NodeReplacement(footer))
				
				(start,footer,repl)
			}
			case Forall(name,label,set,block) => {
				val header = newNode(set)
				artificial ::= header
				val footer = emptyNode()
				val defs = name :: { for(lbl <- label.toList) yield lbl }
				val (start,end) = addBlockStmts(block.id,defs ::: block.stmts)
				artificial ::= start
				if(defs.length == 2)
					artificial ::= (start+1)
				// [ set ] 
				//    |
				//   [block]<-
				//     |    \|
				// [footer]
				//note that control flow says we always enter the block
				//(i.e. the topology is empty)
				//this only occurs when the for comprehension is unnested
				//(nested topology operators never return empty sets right now)
				//in this case, partial redundancy elimination will cause
				//instructions in the for comprehension to be speculated
				//this is the behavior we want.
				newEdge(header,start)
				newEdge(end,start)
				newEdge(end,footer)
				
				val repl = List( NodeReplacement(header),
				                 ControlFlowReplacement(stmt),
				                 NodeReplacement(footer) )
				(header,footer, repl)
			}
			case Return(e) => {
				val r = newNode(stmt)
				val footer = emptyNode()
				newEdge(r,exit)
				newEdge(r,footer) //TODO(zach): footer is dead code, but we add an edge to it here
				                  //because we don't want to drop code from the AST without thinking
				                  //carefully about the implications. Right now post-order traversal
				                  //in OptimizationPass expects no dead code in the CFG
				val repl = List( NodeReplacement(r),NodeReplacement(footer))
				(r,footer,repl)
			}
			case _ => {
				val n = newNode(stmt)
				(n,n,List(NodeReplacement(n)))
			}
		}
		val stmts = ASTUtil.asStmtList(df.body)
		val (start,end) = addBlockStmts(df.body.id, df.args ::: stmts)
		for(i <- 0 until df.args.size) {
			artificial ::= (start + i)
		}
		newEdge(entry,start)
		newEdge(end,exit)
		(Map.empty ++ blk_map,  edges.toArray, node_list.reverse.toArray, SET.empty ++ artificial)
	}
	
	def treeForNode(a : Node) : Option[Tree] = stmts(a)
	lazy val successors : Map[Node,List[Node]] = {
		val succ = new scala.collection.mutable.HashMap[Node,List[Node]]
		for(Edge(from,to) <- edges)
			succ(from) = to :: succ.getOrElseUpdate(from,List.empty) 
		Map.empty.withDefaultValue(List()) ++ succ
	}
	lazy val predecessors : Map[Node,List[Node]] = {
		val succ = new scala.collection.mutable.HashMap[Node,List[Node]]
		for(Edge(from,to) <- edges)
			succ(to) = from :: succ.getOrElseUpdate(to,List.empty) 
		Map.empty.withDefaultValue(List()) ++ succ
	}
	def isArtificial(n : Node) = artificial_nodes.contains(n)
	def nodes = 0 until stmts.size
	
	
	//if we don't want to change this node, what statements should it return?
	def nodeIdentity(n : Node) : List[Stmt] = treeForNode(n) match {
		case Some(x : Stmt) if !isArtificial(n) => List(x)
		case _ => List.empty
	}
	//recreate the AST, at 
	def toTree(fn : (Node) => List[Stmt]) : DefFunc = {
		def replaceBlock(id : Int, exp : SimpleExp) : Block = {
			val repls = replacement_map(id)
			
			def replaceNestedBlocks(stmt : Stmt) : Stmt = {
				def find_blocks(tree : Tree) : Tree = tree match {
					case blk : Block => replaceBlock(blk.id,blk.exp)
					case ShapeAny(t,names,trees) => ShapeAny.trees(tree, trees map find_blocks)
				}
				find_blocks(stmt).asInstanceOf[Stmt]
			}
			def restoreIfExp(stmt : Stmt) : Stmt = stmt match {
				case dv @ DefVar(x,IfExp(tpe,p,t,e)) => {
					//TODO(zach): if statements get inserted after the assign that make this
					//no longer a simple if expression we need to do more fixup, or
					//change it into an If statement
					def extractAssign(blk : Block) : Block = {
						val (List(Assign(_,_,ifexp)), stmts) = blk.stmts.partition {
							case Assign(v : Var,Ops.Assign,_) if v.name == x => true
							case _ => false
						}
						Block(stmts,ifexp)
					}
					DefVar(x,IfExp(tpe,p,extractAssign(t),extractAssign(e))) copyAttr dv
				}
				case x => x
			}
			def replace(r : Replacement) : List[Stmt] = r match {
				case NodeReplacement(node) => fn(node)
				case ControlFlowReplacement(stmt) => {
					val replaced = replaceNestedBlocks(stmt)
					List(restoreIfExp(replaced))
				}
			}
			val stmts = for(r <- repls; stmt <- replace(r))
				yield stmt
			Block(stmts,exp)
		}
		df.copy(body = replaceBlock(df.body.id,EmptyExp(UnitT)))
	}
	//same as above, but also allow the caller to change the aritificial nodes, based on some data flow value
	//artificial nodes include: expressions at the ends of predicates, and defines of forall variables,
	//and the set expressions in a forall statment the function 'artificial' will be given the tree and the node
	//in the CFG that represents where in the control flow that node would normall execute.
	//it should return the new node
	def toTreeWithArtificial(fn : (Node) => List[Stmt], artificial : (Tree,Node) => Tree) : DefFunc = {
		val artificial_tree_to_node : Map[Tree,Node] = { Map.empty ++ {
			for(n <- artificial_nodes; tree <- treeForNode(n))
				yield tree -> n
		} }
		
		//println("artificial map is")
		//for( (k,v) <- artificial_tree_to_node)
		//	println("%s (%d) -> %s".format(k,k.id,v)) 
		val newDef = toTree(fn)
		def replace_artificial(t : Tree) : Tree = {
			//println("replacing: %s (id: %d)".format(t.toShortString,t.id)) 
			val node = artificial_tree_to_node.get(t)
			val new_t = node match {
				case Some(n) => artificial(t,n)
				case None => t
			}
			
			ShapeAny.trees(new_t, new_t.children map replace_artificial)
		}
		replace_artificial(newDef).asInstanceOf[DefFunc]
	}
	def draw( fn : Node => String, name_prefix : String = "")  = {
		val gv = new GraphViz
		class GVNode(n : Node) {
			override def toString =  { treeForNode(n) match {
				case None =>  if(n == entry) "entry" else if(n == exit) "exit" else ""
				case Some(x) if isArtificial(n) => "(*)" + x.toString
				case Some(x) => x.toString
			} } + fn(n)
		}
		val gv_nodes = { for(n <- nodes) yield new GVNode(n) }.toArray
		for(Edge(from,to) <- edges) {
			gv(gv_nodes(from) -> gv_nodes(to))
		}
		gv.draw(Map.empty, "%s_%s".format(name_prefix,df.name) )
	}
	//map from node to immediate domminator and the depth of the node
	abstract class DomNode(val dominator : Node) {
		def depth : Int
	}
	
	lazy val dominatorTree : Map[Node, DomNode] = {
		import scala.collection.immutable.BitSet
		val dominators = new OptimizationPass2 {
			type V = BitSet
			override def cfg = CFG2.this
			override def isForward = true
			
			//consider n -> y, where n has a single successor and y, a single predecessor
			//y's immediate dominator must be n, and n is not an immediate dominator of a node x, such that x != y
			//so we can forgo propagating dominator information for n, and fill in y's dominator easily
			val trivial_dom : Map[Node,Boolean] = Map.empty ++ {
				for(node <- cfg.nodes) yield {
					node -> { cfg.successors(node) match {
						case List(p) => cfg.predecessors(p).length == 1
						case _ => false
					} }
				}
			}
			override def nodeTransferFn(n : Node)( v : V) = if(trivial_dom(n)) v else v + n 
			override def meet(a : V, b : V) = a & b
			override def inputValue = BitSet.empty
			override def top = BitSet.empty ++ cfg.nodes
			override def name = "dominators"
			override def equal(a : V, b : V) = a == b
		}.result
		object ClassScope {
			val node_to_dom_map : Map[Node,DomNode] = Map.empty ++ { 
				for(node <- nodes) yield {			
					lazy val pred = CFG2.this.predecessors(node)
					val dn : DomNode = if(node == CFG2.this.entry) {
						new DomNode(node) {
							override def depth = 0
						}
					} else if(pred.size == 1) {
						val x = pred.head
						new DomNode(x) {
							override lazy val depth = node_to_dom_map(x.asInstanceOf[Node]).depth + 1
						}
					} else {
						val all_doms = dominators(node).in
						val doms_doms = for(dom <- all_doms) yield dominators(dom).in
						val doms_doms_union = doms_doms.foldRight(BitSet.empty)(_ | _)
						val immediate_doms = all_doms &~ doms_doms_union
						immediate_doms.toList match {
							case List(x) => new DomNode(x.asInstanceOf[Node]) {
								override lazy val depth = node_to_dom_map(x.asInstanceOf[Node]).depth + 1
							}
							case x => Log.ierror("immediate dominator?: %d %s doms: %s preds: %s x: %s".format(node,treeForNode(node),all_doms,predecessors(node),x))
						}
					}
					node -> dn
				}
			}
		}
		ClassScope.node_to_dom_map
	}
	def lowestCommonDominator( nodes : List[Node] ) : Node = {
		def lcd(a : Node, b : Node) : Node = {
			val (ad,bd) = (dominatorTree(a),dominatorTree(b))
			if(ad.depth < bd.depth)
				lcd(a,bd.dominator)
			else if(ad.depth > bd.depth)
				lcd(ad.dominator,b)
			else if(a == b)
				a
			else
				lcd(ad.dominator,bd.dominator)
		}
		if(nodes.isEmpty)
			this.exit
		else
			nodes.reduceRight(lcd)
	}
	def nodeID(node : Node) : Int = node
}


trait OptimizationPass2 {
  type V <: AnyRef
  def top : V
  def meet(a : V, B : V) : V
  def equal(a : V, b : V) : Boolean //used to determine if a value has changed from one iteration to the next
  
  def isForward : Boolean
  def inputValue : V
  def cfg : CFG2
  type Node = CFG2#Node
  type Attribute[V <: AnyRef] = CFG2#Attribute[V]
  
  //subclass should override either transferFn or nodeTransferFn
  def transferFn(e : Tree)(v : V) : V = v
  def nodeTransferFn(n : Node)(v : V) : V = {
    cfg.treeForNode(n) match {
    	case Some(tree) => transferFn(tree)(v)
    	case None => v
    }
  }
  
  def valueToString( v : V) : String = v.toString
  def name = "unknown_pass"
  
  def meetList(vs : List[V]) : V = vs match {
    case List() => top
    case List(a) => a
    case h :: t => t.foldRight(h)(meet)
  }
  def dependants(n : Node) : List[Node] = if(isForward) cfg.successors(n) else cfg.predecessors(n)
  def root = if(isForward) cfg.entry else cfg.exit
  
  val reverse_post_order : List[Node] = {
     import scala.collection.mutable.BitSet
     
     val is_visited = new BitSet
     is_visited += cfg.nodeID(root)
     def unvisited(h : Node) = for(s <- dependants(h) if !is_visited.contains(cfg.nodeID(s))) yield {
     	is_visited += cfg.nodeID(s)
     	s
     }
     
     var visited : List[Node] = List.empty
     var context : List[(Node,List[Node])] = List((root,unvisited(root)))
     
     while(!context.isEmpty) {
     	val (n,deps) :: tail = context
     	context = tail
     	deps match {
     		case h :: t =>
     			context = (h,unvisited(h)) :: (n,t) :: context 
     		case List() =>
     			visited ::= n
     	}
     }
     //println("reverse post order is: %s".format(visited))
     if(BitSet.empty ++ visited != BitSet.empty ++ cfg.nodes)
     	Log.ierror("dead nodes!: %s".format(BitSet.empty ++ cfg.nodes -- visited))
     if(Config.getOrElse("reverse_post_order",true).asInstanceOf[Boolean])
     	visited
     else
     	cfg.nodes.toList
  }
  
  
  lazy val result : Attribute[V] = {
    val state = cfg.newAttribute(top)
    if(isForward)
    	state(cfg.entry).in = inputValue
    else
    	state(cfg.exit).out = inputValue
    
    var changed = true
    var iter = 0
    while(changed) {
      changed = false
      if(isForward) {
        for(node <- reverse_post_order) {
           val inputs = for(n <- cfg.predecessors(node)) yield state(n).out
           val input = if(node == cfg.entry) state(node).in else meetList(inputs)
           if(!changed && !equal(input,state(node).in))
              changed = true
           state(node).in = input
           val output = nodeTransferFn(node)(input)
           if(!changed && !equal(output,state(node).out))
              changed = true
           state(node).out = output
        }
      } else {
        for(node <- reverse_post_order) {
           val inputs = for(n <- cfg.successors(node)) yield state(n).in
           val input = if(node == cfg.exit) state(node).out else meetList(inputs)
           if(!changed && !equal(input,state(node).out))
             changed = true
           state(node).out = input
           val output = nodeTransferFn(node)(input)
           if(!changed && !equal(output,state(node).in))
             changed = true
           state(node).in = output
        }
      }
      iter += 1
    }
    Log.log("it took %d iterations to complete %s".format(iter,this.name))
    if(LogSettings.enabled("CFGDraw"))
    	cfg.draw(node => valueToString(state(node).in), this.name )
    state
  }
}

