package Liszt
import scala.collection.immutable.{Set => SSet}
import Liszt.Types._
import Liszt.Trees._

trait CLikeCodeGen extends CodeGenBase with TypeInferenceClient {
	import ASTUtil._
	type StmtResult = List[CodeLine]
	type ExpResult = String
	type OpResult = String
	type TreeResult = String
	type ProgramResult = String
	type TypeResult = String
	class CodeLine(var indent : Int, line : String) {
		override def toString = ("    " * indent) + line
	}
	def line(str : String) = List(new CodeLine(0,str))
	def indent(code : StmtResult) = { for(i <- code) { i.indent += 1 }; code } //for(s <- code) yield ("    " + code)

	
	class Erasure0
	class Erasure1
	class StmtBuilder {
		var cur_indent = 0
		var rev_lines : List[CodeLine] = List.empty
		def add(line : String) : this.type = {
			rev_lines ::= new CodeLine(cur_indent,line)
			this
		}
		//implicit arguments are to get around type erasure for two sequences
		def add(line : Seq[String])(implicit e0 : Erasure0) : this.type = {
			for(l <- line)
				add(l)
			this
		}
		def add(c : CodeLine) : this.type = {
			c.indent += cur_indent
			rev_lines ::= c
			this
		}
		def add(cs : Seq[CodeLine])(implicit e1 : Erasure1) : this.type = {
			for(c <- cs)
				add(c)
			this
		}
		def begin : this.type = {
			cur_indent += 1
			this
		} 
		def end : this.type = {
			cur_indent -= 1
			this
		}
		def toCodeLines = rev_lines.reverse
	}
	def NewStmt = new StmtBuilder
	
	object CodeLineImplicits {
		implicit def promoteToCodeLinesStr(stmt : String) : List[CodeLine] = List(new CodeLine(0,stmt))
		implicit def promoteToCodeLinesList(stmts : List[String]) : List[CodeLine] = stmts.map(x => new CodeLine(0,x))
		implicit def builderToCodeLines(stmtb : StmtBuilder) : List[CodeLine] = stmtb.toCodeLines
		implicit val iErasure0 = new Erasure0
		implicit val iErasure1 = new Erasure1
	}
	import CodeLineImplicits._
	
	val debuggingEnabled = Config.getOrElse("debug",false).asInstanceOf[Boolean]
	//a set of all the vector and matrix types for which we found 
	//VectorL and MatrixL initializer for which we need to generate helper make_vec and make_mat functions
	private var vec_mat_types : SSet[Type] = SSet.empty
	
	def tBinOp(op : Ops.BinOp) : OpResult = {
		import Liszt.Ops._
		op match {
			case Plus => "+"
			case Minus => "-"
			case Times => "*"
			case Div => "/"
			case Mod => "%"
			case Eq => "=="
			case NEq => "!="
			case Less => "<"
			case LessEq => "<="
			case Greater => ">"
			case GreaterEq => ">="
			case Min => Log.ierror("not a binOp")
			case Max => Log.ierror("not a binOp")
			case LShift => "<<"
			case RShift => ">>"
			case BOr => "|"
			case BAnd => "&"
			case BXor => "^"
			case And => "&&"
			case Or => "||"
		}
	}
	def tUnOp(op : Ops.UnOp) : OpResult = op match {
		case Ops.Neg => "-"
		case Ops.BNot => "~"
		case Ops.Not => "!" 
	}
	def tPickOp(op : Ops.PickOp) : OpResult = op match {
		case Ops.Min => "min"
		case Ops.Max => "max"
	}
	def tAssignOp(op : Ops.AssignOp) : OpResult = op match {
		case Ops.Assign => "="
		case Ops.Reduce(op : Ops.PickOp) => tPickOp(op) + "With"
		case Ops.Reduce(Ops.Or) => "|=" //HACK: c++ doesn't have ||=, or &&= for no good reason... because you cannot implement short-circuit semantics in C++ cause it's not lazy like Haskell
		case Ops.Reduce(Ops.And) => "&="
		case Ops.Reduce(op : Ops.BinOp)  => tBinOp(op) + "="
	}
	def tWorkgroup(e : Workgroup) = NewStmt.add("/*Workgroup %d*/".format(e.id)).begin.add(e.stmts flatMap tStmt)
	lazy val integer_type = Config.getOrElse("integer-type","int").toString
	lazy val floating_point_type = Config.getOrElse("floating-point-type","double").toString
	lazy val double_type = Config.getOrElse("double-type","double").toString
	def tType(t : Type) : TypeResult = t match {
			case Vertex => "Vertex"
			case Edge => "Edge"
			case Face => "Face"
			case Cell => "Cell"
			case Mesh => "const Mesh*"
			case FieldT(mo,t) => "Field< %sType , %s >".format(tType(mo),tType(t))
			case Set(t) => "Set< %s >".format(tType(t))
			case IntT =>  integer_type
			case FloatT => floating_point_type
			case DoubleT => double_type
			case StringT => "std::string"
			case BoolT => "bool"
			case VectorT(n,t) => "vec< %s , %s >".format(tType(n),tType(t))
			case MatrixT(n,m,t) => "matrix< %s , %s , %s >".format(tType(n),tType(m),tType(t))
			case UnitT => "void"
			case SparseVectorType(ls) => "%s_Vector".format(ls)
			case SparseMatrixType(ls) => "%s_Matrix".format(ls)
			case _ : IndexType => "LinearSystemIndex::Index"
			case _ : NonzeroesSpecifierType => "int" //this is ignored at runtime, but needs to be a valid type
			case SpecialType(_,str) => str
			case MetaIntT(i) => "%d".format(i)
			case _ => Log.ierror("unexpected type: %s",t)
	    }
	override def producesSpecialType(t : Type) = t match {
		case _ : Set => true
		case _ => false
	}
	override def runtimeTypeFor(ctx : CtxT, exptoctype : ExpT, tree : Tree)  = tree match {
		case SetInit(t,_) => SpecialType(t,t match {
			case Set(Vertex) => "BoundarySet<vertex_set>::type"
			case Set(Edge) => "BoundarySet<edge_set>::type"
			case Set(Face) => "BoundarySet<face_set>::type"
			case Set(Cell) => "BoundarySet<cell_set>::type"
			case _ => Log.ierror("not a set  type %s",t)
		})
		case _ => super.runtimeTypeFor(ctx,exptoctype,tree)
	}
	def meet(a : Type, b : Type) = (a,b) match {
		case (SpecialType(t,str),SpecialType(t2,str2)) => {
			if(str == str2)
				a
			else (t,t2) match {
				case (_ : Set, _ : Set) => SpecialType(t,"EitherSet< %s , %s >".format(str,str2))
				case _ => Log.ierror("unexpected meet: %s %s",a,b)
			}
		}
		case _ => Log.ierror("unexpected meet: %s %s",a,b)
	}
	def treeForCTypes = top.code
	lazy val getCType : Map[Tree,Type] = CTypes.create(treeForCTypes,this)
	def tTypeFor(t : Tree) : TypeResult = tType(getCType(t))
	
	//in the C implementation, is the type a pointer?
	//i.e. should we use -> when calling a method and *v for binary operations performed on the type?
	def isPointerType(t : Type) : Boolean
	
	private def tVarUse(e : Var) : String = (if(isPointerType(e.t)) "(*%s)" else "%s" ).format(tVar(e)) 
	def tBinary(e : Binary) = e match {
		case Binary(t,l,op : Ops.PickOp,r) => "ReductionImpl< %s >::%s(%s,%s)".format(tType(t),tPickOp(op),tVarUse(l), tVarUse(r) )
		case Binary(t,l,op,r) => "%s %s %s".format(tVarUse(l),tBinOp(op),tVarUse(r))
	}

	private lazy val fnlist = SSet.empty ++ {
		for(DefFunc(_,name,_,stmt) <- top.code.stmts if stmt != ExternStmt)
			yield name
	}
	def isExternal(fn : String) : Boolean = !(fnlist contains fn)
	
	def tUnary(e : Unary) = "%s%s".format(tUnOp(e.op),tVarUse(e.e))
	def tVecExtract(e : VecExtract) = "%s[%s]".format(tVar(e.vec),tExp(e.n))
	def tMatExtract(e : MatExtract) = "%s(%s,%s)".format(tVar(e.mat),tVar(e.n),tVar(e.m))
	def tCall(e : Call) = {
		val fn = tVar(e.func)
		val arg_list = e.args.map(tVar).mkString(",")
		val param_list = e.type_params.filter(x => !ASTUtil.isMeshT(x))
		if(isExternal(fn) && param_list.isEmpty == false) {
			"%s < %s >(%s)".format(fn,param_list.map(tType).mkString(","),arg_list)
		} else {
			"%s(%s)".format(fn, arg_list)
		}
	}
	def tTopoCall(e : TopoCall) =  "%s(%s)".format(e.func.asFunction,e.args.map(tVar).mkString(","))
	def tFieldRead(e : FieldRead) = "%s(%s)".format(tVar(e.field),tVar(e.arg))
	def tIntL(e : IntL) = e.c.toString
	def tFloatL(e : FloatL) = e.c.toString + "f"
	def tDoubleL(e : DoubleL) = e.c.toString
	def tBoolL(e : BoolL) = e.c.toString
	
	def tVectorL(e : VectorL) = {
		vec_mat_types += e.t
		"Liszt_make_%s(%s)".format(typeID(e.t),e.es.map(tVar).mkString(",")) 
	}
	def tMatrixL(e : MatrixL) = {
		vec_mat_types += e.t
		"Liszt_make_%s(%s)".format(typeID(e.t),e.es.map(tVar).mkString(","))
	}
	def tMatrixFromVec(e : MatrixFromVec) = {
		vec_mat_types += e.t
		"Liszt_make_%s_fromvec(%s)".format(typeID(e.t),e.es.map(tVar).mkString(","))
	}
	
	def tStringL(e : StringL) = {
		"\"%s\"".format(escapeString(e.c))
	}
	
	def tVar(e : Var) = e.name
	


	
	def tFieldWrite(e : FieldWrite) = e match {
		case FieldWrite(field,arg,Ops.Reduce(op : Ops.PickOp),rhs) => "%s[%s].%sWith(%s);".format(tVar(field),tVar(arg),tPickOp(op),tVar(rhs))
		case FieldWrite(field,arg,op,rhs) => "%s[%s] %s %s;".format(tVar(field),tVar(arg),tAssignOp(op),tVar(rhs))
	}
	
	def tIf(e : If) = {
		val If(blk,thenE,elseE) = e
		val inits = blk.stmts.flatMap(tStmt)
		inits ::: 
		"if(%s) {".format(tExp(blk.exp)) :::
			indent(tStmt(thenE)) :::
		"} else {" :::
			indent(tStmt(elseE)) :::
		"}"
	}
	def tWhile(e : While) = {
		val While(pred,block) = e
		val begin = Names.create("while_begin")
		val end = Names.create("while_end")
		val bexp = pred.exp
		val exps = pred.stmts
		
		"%s: {".format(begin) :::
		(indent {
				exps.flatMap(tStmt) :::
				"if(!(%s)) goto %s;".format(tExp(bexp),end) :::
				tStmt(block)
			}):::
		"}" :::
		"goto %s;".format(begin) :::
		"%s: ;".format(end)
	}
	def tBlock(e : Block) =  { e.stmts.flatMap(tStmt) ::: tStmt(ExpStmt(e.exp)) }
	
	def tDefFunc(e : DefFunc) = e match {
		case DefFunc(ftype,name,args,ExternStmt) => "%s;".format(tFunctionHeader(e,true))
		case DefFunc(ftype,name,args,body) => {
			"%s {".format(tFunctionHeader(e,false)) :::
				indent(tStmt(body)) ::: 
			"}"
		}
	}
	
	def tDefVar(e : DefVar) = {
		e.init match {
			case l : Literal if e.name.startsWith("literal") => "#define %s %s".format(e.name,tExp(l))
			case _ : EmptyExp => "%s %s;".format(tTypeFor(e),e.name)
			case exp : IfExp => removeIfExp(e) flatMap tStmt
			case exp => "%s %s = %s;".format(tTypeFor(e),e.name,tExp(exp))
		}
	}
	
	def tReturn(e : Return) = e.e match {
		case _ : EmptyExp => "return;"
		case exp => "return %s;".format(tExp(exp))
	}
	def tPrintStream() = "std::cout"
	def tPrintContext() = "System::print_context"
	def tPrint(e : Print) = {
		val dbg = if(debuggingEnabled) "%s <<".format(tPrintContext()) else ""
		"%s << %s\"liszt: \" << %s std::endl;".format(tPrintStream(),dbg,e.es.map(tVar).mkString(""," << "," << ")) 
	}
        def tExit = "exit(-1);"
		
	def tExpStmt(e : ExpStmt) = e.e match {
		case _ : EmptyExp => List()
		case exp => "%s;".format(tExp(exp))
	}
	
	def tExternStmt = {
		Log.ierror("unexpected extern statement")
	}
	
	def tReductionOp(e : Ops.ReductionOp) = e match {
		case Ops.RMax => "LISZT_MAX"
		case Ops.RMin => "LISZT_MIN"
		case Ops.RProd => "LISZT_PROD"
		case Ops.RSum => "LISZT_SUM"
		case Ops.RLAnd => "LISZT_LAND"
		case Ops.RLOr => "LISZT_LOR"
		case Ops.RLXor => "LISZT_LXOR"
		case Ops.RBAnd => "LISZT_BAND"
		case Ops.RBOr => "LISZT_BOR"
		case Ops.RBXor => "LISZT_BXOR"
		case Ops.RAssign => "LISZT_ASSIGN"
		case Ops.RNone => "LISZT_READ"
		case Ops.RDiv => "LISZT_DIV"
		case Ops.RSub => "LISZT_SUB"
	}

	def tEnsureFieldState(e : EnsureFieldState) = {
		val dot = if(isPointerType(e.field.t)) "->" else "."
		"%s%stemplate ensureState<__liszt_nested,%s>();".format(e.field.name,dot,tReductionOp(e.state))
	}
	
	def tAssign(e : Assign) = {
		
		def makeAssign(lhs : String) = e.op match {
			case Ops.Reduce(po : Ops.PickOp) => "%s = ReductionImpl< %s >::%s(%s,%s);".format(lhs,tType(e.r.t),tPickOp(po),lhs,tVar(e.r))
			case op => "%s %s %s;".format(lhs,tAssignOp(op),tVar(e.r))
		}
		e.lv match {
			case v : Var => makeAssign(tVar(v))
			case ve : VecExtract => makeAssign( tVecExtract(ve) )
			case me : MatExtract => makeAssign( tMatExtract(me) )
			case sv @ SparseVectorAccess(t,obj,arg) => "SPARSE_VECTOR_WRITE_%s(%s,%s,%s,%s);".format(sparseReturn(sv),
		                                                                                         tVar(obj),
		                                                                                         tVar(arg),
		                                                                                         tReductionOp(e.op.asReduction(false)),
		                                                                                         tVar(e.r))
			case sm @ SparseMatrixAccess(_,obj,row,col) => "SPARSE_MATRIX_WRITE_%s(%s,%s,%s,%s,%s);".format(sparseReturn(sm),
		                                                                                              tVar(obj),
		                                                                                              tVar(row),
		                                                                                              tVar(col),
		                                                                                              tReductionOp(e.op.asReduction(false)),
		                                                                                              tVar(e.r))
			case _ => Log.ierror("unsupported LValue: %s",e)
		}
	}

	def sparseReturn(e : Exp) = e.t match {
		case _ : VectorT => "N"
		case _ : MatrixT => "N"
		case _ => "1"
	}
	def tSparseVectorAccess(e : SparseVectorAccess) = "SPARSE_VECTOR_READ_%s(%s,%s)".format(sparseReturn(e),tVar(e.vec),tVar(e.idx))
	def tSparseMatrixAccess(e : SparseMatrixAccess) = "SPARSE_MATRIX_READ_%s(%s,%s,%s)".format(sparseReturn(e),tVar(e.vec),tVar(e.row),tVar(e.col))
	
	def tIndexCall(e : IndexCall) = "%s_index.%s(%s)".format(e.linear_system, e.index.name, (e.args map tVar).mkString(","))
	def tLinearSystemAllocate(e : LinearSystemAllocate) = "%s.%s()".format(e.linear_system, e.element match {
		case LS_A => "A"
		case LS_X => "x"
		case LS_B => "b"
	})
	def tLinearSystemCall(e : LinearSystemCall) = "%s.%s(%s)".format(e.linear_system, tVar(e.func), e.args.map(tVar).mkString(","))
	
	def tDefLinearSystem( e : DefLinearSystem) = {
		def numIndicesPerCall(e : Index) : Int = {
			e.index_type match {
				case _ : IndexType => 1
				case VectorT(MetaIntT(n),_ : IndexType) => n
				case _ => Log.ierror("invalid index type")
			}
		}
		def isRowIndex(e : Index) = {
			e.index_type match {
				case IndexType(true) => true
				case VectorT(_,IndexType(true)) => true
				case _ => false
			}
		}
		def meshElementSize(topo : Topo) = topo match {
			case Vertex => "nV"
			case Edge => "nE"
			case Face => "nF"
			case Cell => "nC"
			case Mesh => Log.ierror("unexpected mesh type")
		}
		def indexSizeExp(e : Index) = {
			val sizes = for(t <- e.arg_types) yield meshElementSize(t.asInstanceOf[Topo])
			(numIndicesPerCall(e).toString :: sizes).mkString(" * ")
		}
		def indexInitExp(e : Index) = {
			List("%s_offset = current_offset;".format(e.name),
			     "current_offset += %s;".format(indexSizeExp(e)))
		}
		def indexFunction(e : Index) = {
			def offset(args : List[(Type,Int)]) : String = args match {
				case (head_type : Topo,head_index) :: t => "getID(a%d) + %s * ( %s )".format(head_index,meshElementSize(head_type), offset(t)) 
				case Nil => "0"
			}
			val off = "this->local%sOffset() + %s_offset + %d * (%s)".format(if(isRowIndex(e)) "Row" else "Col",e.name,numIndicesPerCall(e),offset(e.arg_types zipWithIndex))
			
			val arg_list = for((a,i) <- e.arg_types zipWithIndex)
				yield "%s a%d".format(tType(a),i)
			val arg_string = arg_list.mkString(",")
			val ret_type = tType(e.index_type)
			
			val decl = "%s %s(%s) {".format(ret_type,e.name,arg_string)
			
			val impl = e.index_type match {
				case _ : IndexType => NewStmt.add("%s ret = %s;".format(ret_type,off))
				case VectorT(MetaIntT(n), idx : IndexType) => NewStmt.
				                            add("%s offset = %s;".format(tType(idx),off)).
				                            add("%s ret;".format(ret_type)).
				                            add {
				                             for(i <- 0 until n)
				                             	yield "ret[%d] = offset + %d;".format(i,i) 
				                            }
				
			}
			NewStmt.
			add(decl).
			begin.
				add(impl).
				add("return ret;").
			end.
			add("}")
		}
		val solver_name = e.solver
		NewStmt.
		add("class %s_t : public LinearSystemIndex {".format(e.name)).
		add("public:").
		begin.
			add("public:").
			add("void generated_init() {").
				begin.
				add("size_t current_offset = 0;").
				add {
					for(idx <- e.indices if isRowIndex(idx); l <- indexInitExp(idx)) 
						yield l
				}.
				add("this->setRowSize(current_offset);").
				add("current_offset = 0;").
				add{
					for(idx <- e.indices if !isRowIndex(idx); l <- indexInitExp(idx))
						yield l
				}.
				add("this->setColSize(current_offset);").
				end.
			add("}").
			add{
				for(idx <- e.indices)
					yield "%s %s_offset;".format(tType(IndexType(true)), idx.name)
			}.
			add{ 
				for(idx <- e.indices; line <- indexFunction(idx))
					yield line
			}.
		end.
		add("};").
		add("%s_t %s_index;".format(e.name,e.name)).
		add("typedef %s::Vector %s_Vector;".format(solver_name,e.name)).
		add("typedef %s::Matrix %s_Matrix;".format(solver_name,e.name)).
		add("%s %s;".format(solver_name,e.name))
	}
	def tLinearSystemNonzeroes( e : LinearSystemNonzeroes) = {
		NewStmt.
		add("%s.beginNonZeroes();".format(e.linear_system)).
		begin.
			add("int %s = 0;".format(e.nonzeroes.name)).
			add(tStmt(e.blk)).
		end.
		add("%s.endNonZeroes();".format(e.linear_system))
	}
	def tLinearSystemPlaceNonzero( e : LinearSystemPlaceNonzero) = {
		val NonzeroesSpecifierType(ls) = e.obj.t
		val element = e.element match {
			case LS_A => "A"
			case LS_X => "X"
			case LS_B => "B"
		}
		"%s.entry%s(%s);".format(ls,element,(e.args map tVar).mkString(","))
	}
	
	def tForallPrefix(e : Forall) : StmtResult = if(debuggingEnabled) "%s.push();".format(tPrintContext()) else List()
	def tForallSuffix(e : Forall) : StmtResult = if(debuggingEnabled) "%s.pop();".format(tPrintContext()) else List()
	def tForallBodyPrefix(e : Forall) : StmtResult = if(debuggingEnabled) "%s.mark(%s);".format(tPrintContext(),e.name.name) else List()
	
	def tPerfCounter(e : PerfCounter) : StmtResult = "/* %s:%d */".format(e.name,e.count)
	def tHelperFunctionModifiers() = ""
	
	def tHelperFunctions() = {
		def fnTemplate(t : Type, name : String, args : List[String], assigns : StmtResult) : StmtResult = {
			val typ = tType(t)
			"%s %s Liszt_make_%s(%s) {".format(tHelperFunctionModifiers(), typ,name,args.mkString(",")) :::
				indent {
					"%s v;".format(typ) :::
					assigns :::
					"return v;"
				} :::
			"}"
		}
		def genHelper(t : Type) = t match {
			case VectorT(MetaIntT(n),v) => {
				val args = for(i <- 0 until n) 
					yield "const %s & a%d".format(tType(v),i)
				val assigns = for(i <- 0 until n)
					yield "v.v[%d] = a%d;".format(i,i)
				fnTemplate(t,typeID(t),args.toList,assigns.toList)
			}
			case m @ MatrixT(MetaIntT(r),MetaIntT(c),v) => {
				val args = for(i <- 0 until r; j <- 0 until c)
					yield "const %s & a%d_%d".format(tType(v),i,j)
				val assigns = for(i <- 0 until r; j <- 0 until c)
					yield "v.v[%d][%d] = a%d_%d;".format(i,j,i,j)
				val argvs = for(i <- 0 until r)
					yield "const %s & a%d".format(tType(m.rowType),i)
				val assignvs = for(i <- 0 until r; j <- 0 until c)
					yield "v.v[%d][%d] = a%d[%d];".format(i,j,i,j)
				fnTemplate(t,typeID(t),args.toList,assigns.toList) :::
				fnTemplate(t,"%s_fromvec".format(typeID(t)),argvs.toList,assignvs.toList)
			}
                        case SimdT(VectorT(MetaIntT(n), base)) => {
				val args = for(i <- 0 until n)
					yield "const %s & a%d".format(tType(SimdT(base)),i)
				val assigns = for(i <- 0 until n)
					yield "v[%d] = a%d;".format(i,i)
				fnTemplate(t,typeID(t),args.toList,assigns.toList)
                        }
                        case SimdT(MatrixT(MetaIntT(n), MetaIntT(m), base)) => {
				val args = for(i <- 0 until n; j <- 0 until m)
					yield "const %s & a%d_%d".format(tType(SimdT(base)),i,j)
				val assigns = for(i <- 0 until n; j <- 0 until m)
					yield "v(%d,%d) = a%d_%d;".format(i,j,i,j)
				val from_floats = fnTemplate(t,typeID(t),args.toList,assigns.toList)

                                val vecs = for(i <- 0 until n) yield "const %s & a%d".format(tType(SimdT(VectorT(MetaIntT(m), base))), i);
                                val vec_assigns = for(i <- 0 until n; j <- 0 until m)
                                  yield "v(%d,%d) = a%d[%d];".format(i, j, i, j);

                                val from_vectors = fnTemplate(t, "%s_fromvec".format(typeID(t)), vecs.toList, vec_assigns.toList);
                                from_floats ::: from_vectors;
                        }
			case _ => Log.ierror("unexpected type (%s)".format(tType(t)))
		}
		Log.log("vec_mat_types are: %s",vec_mat_types)
		vec_mat_types.toList flatMap genHelper
	}
}
