/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Liszt
import Liszt.Trees._
import Liszt.Types._
import Liszt.ASTUtil._


//certain runtimes may have more than one type to represent a Liszt type.

//For instance the C++ runtime uses different sets: Mesh::vertex_set, or Cell::vertex_set, etc.
//depending on how the set is created.  SpecialType tracks this runtime-specific set
//original is the type of the expression in Liszt (e.g. Set[Vertex]) while custom_type
//is the type in the runtime (e.g. Cell::vertex_set)
case class SpecialType(original : Type, custom_type : String) extends Type

//a lazy type so that contexts can be resolved later and we don't have to worry about function declaration
//orders
class LazyType(typfn : => Type) extends Type {
	lazy val touch : Type = LazyType.touchAll(typfn)
}
object LazyType {
	def touchAll(v : Type) : Type = v match {
		case lt : LazyType => lt.touch
		case Func(args,range) => Func(args map touchAll,touchAll(range))
		case v => v
	}
}
trait TypeInferenceClient {
	//returns the type (possibly SpecialType) for the AST node tree
	//by default we assume nothing needs special types, this should be overloaded to track special
	//types, and the call super class implementation for the remaining tree nodes
	
	//The default implement does not explicitly make anything a special type.  However, it tracks
	//special types through functions, and across if expressions if the client creates them.  
	//We assume for now that special types cannot appear on the lhs of an Assign node (i.e. they re not reassigned)
	type CtxT = String => Type
	type ExpT = scala.collection.Map[Tree,Type]
	def runtimeTypeFor(ctx : CtxT, exptoctype : ExpT, tree : Tree) : Type = {
		tree match {
			case Block(es,e) => exptoctype(e) 
			case df @ DefFunc(t,name,args,stmt) => {
				val argts = (args map {x => x.name}) map ctx
				val typ = Func(argts, stmt match {
					case ExternStmt => typeForReturn(df)
					case _ if !producesSpecialType(df.typ.range) => df.typ.range
					case body => {
						val typs = for(Return(exp) <- body) yield exptoctype(exp)
						typs reduceLeft lmeet
					}
				})
				typ
			}
			case DefVar(name,init) => exptoctype(init)
			case Var(t,name) => ctx(name)
			case IfExp(t,p,tn,e) => if(producesSpecialType(t)) lmeet(exptoctype(tn),exptoctype(e)) else t
			case _ => {
				if(producesSpecialType(tree.t)) {
					tree match {
						case Call(t,func,args) => {
							val Func(args,result) = ctx(func.name)
							result
						}
						case TopoCall(t,op,args) => {
							SpecialType(tree.t,specialReturnType(op.asFunction))
						}
						case _ => tree.t
					}
				} else {
					tree.t
				}
			}
		}
	}
	
	//take two types (possibly SpecialType) and return a (runtime) type that can 
	//work if the actual value of the expression is either a or b
	//if you run into a case where there is not type fitting this property, you potentially have a 
	//problem where you cannot codegen a piece of code because you cannot determine the correct
	//runtime type
	def meet(a : Type, b : Type) : Type
	def lmeet(a : Type, b : Type) : LazyType = new LazyType ( meet(LazyType.touchAll(a),LazyType.touchAll(b)) )
	//is it possible for an expression of type t to produce a SpecialType with type t?
	//this should mirror the implementation of runtimeTypeFor
	def producesSpecialType(t : Type) = false
	
	def typeForArgs(df : DefFunc) : List[Type] = {
		for(a <- df.typ.args) yield 
			if(producesSpecialType(a))
				SpecialType(a,ASTUtil.Names.create("CType"))
			else 
				a
	}
	def typeForReturn(df : DefFunc) : Type = typeForReturn(df.typ,df.name)
	def typeForReturn(ft : Func, name : String) : Type = {
		if(producesSpecialType(ft.range))
			SpecialType(ft.range,specialReturnType(name))
		else
			ft.range
	}
	
	//type given to return values that may produces special types
	def specialReturnType(fn_name : String) = fn_name + "_return_type"
	def isExternal(fn : String) : Boolean
}
object CTypes {
  object Log extends Liszt.Log("CTypes")
  
  def create(code : Tree, client : TypeInferenceClient) : Map[Tree,Type] = {
     import scala.collection.mutable.HashMap

     val exptoctype = new HashMap[Tree,Type]
     
     //mapping from variable to calculated Type, note that CompilationUnit ensures that
     //variable names are unique across all scopes, so there is not need for more compilicated lookup
     val ctx = new HashMap[String,Type]
     def lctx(a : String) : LazyType = new LazyType ( ctx(a) )
     
     def convert(exp : Tree) {
     
       def getT(exp : Tree) : Type = client.runtimeTypeFor(lctx,exptoctype,exp)
       //convert all trees in exp, for nodes that can define variables,
       //we must update the context
       exp match {
           case df @ DefFunc(ftype, name , args , stmt) => {
               val argts = client.typeForArgs(df)
               for((t,a) <- argts zip args)
                   ctx(a.name) = t
               convert(stmt)
               val typ = getT(exp)
               ctx(name) = typ
               exptoctype(exp) = typ
           }
           case Forall(iter,lblop,set,block) => {
               convert(set)
               for(lbl <- lblop) {
                 ctx(lbl.name) = IntT
               }

               //ctx(iter.name) = set.t.asInstanceOf[Set].t
               if (isSimdT(set.t)) {
                 // If the set is actually a vector of sets (e.g. a
                 // nested topocall) then the result type is a vector of
                 // the "scalar" read)
                 val simd_type = set.t.asInstanceOf[SimdT];
                 val set_type = simd_type.t.asInstanceOf[Set];
                 val elem_type = set_type.t;
                 ctx(iter.name) = SimdT(elem_type);
               } else {
                 ctx(iter.name) = set.t.asInstanceOf[Set].t
               }

               convert(block)
               exptoctype(exp) = getT(exp)
           }
           case LinearSystemNonzeroes(linear_system ,a , blk ) => {
               ctx(a.name) = SparseMatrixType(linear_system)
               convert(blk)
               exptoctype(exp) = getT(exp)
           }
           case DefVar(name,init) => {
               convert(init)
               exptoctype(exp) = getT(exp)
               ctx(name) = exptoctype(init)
           }
           case Var(ft : Func,name) => {
               val typ = if(client.isExternal(name))
                   client.typeForReturn(ft,name)
               else
                   lctx(name)
               exptoctype(exp) = typ
           }
           case Var(t,name) => {
               exptoctype(exp) = lctx(name)
           }
           case x @ SimdL(l) => exptoctype(exp) = x.t;
           case ShapeAny(_,_,trees) => {
               trees foreach convert
               exptoctype(exp) = getT(exp)
           }
       }
     }
    convert(code)

    //remove the Lazy types now that the ctxs are fully created
    Map.empty ++ { for( (k,v) <- exptoctype) yield { k -> LazyType.touchAll(v) } }
  }
}
