/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Liszt
import scala.collection.immutable.{Set => SSet}
import Liszt.Types._
import Liszt.Trees._


trait CodeGenBase {
	def top : CompilationUnit
	val Log : Liszt.Log
	
	//Expected output types for stmts and expressions
	//in our current implementation expressions become strings and statements become lists of CodeLines
	type StmtResult
	type ExpResult
	type OpResult
	type TreeResult
	type TypeResult
	type ProgramResult
	
	def indent(e : StmtResult) : StmtResult //indent the representation
	
	def tBinOp(op : Ops.BinOp) : OpResult
	def tUnOp(op : Ops.UnOp) : OpResult
	def tPickOp(op : Ops.PickOp) : OpResult
	def tAssignOp(op : Ops.AssignOp) : OpResult
	def tReductionOp(op : Ops.ReductionOp) : OpResult
	
	//translate type directly (some types might not have direct translations)
	def tType(t : Type) : TypeResult
	
	//translate the type that t has, may have specialized type (e.g. in the case of set representations)
	def tTypeFor(t : Tree) : TypeResult
	
	
	//Expression translators
	def tBinary(e : Binary) : ExpResult
	def tUnary(e : Unary) : ExpResult
	def tCall(e : Call) : ExpResult
	def tTopoCall(e : TopoCall) : ExpResult
	def tFieldRead( e : FieldRead ) : ExpResult
	def tVecExtract( e : VecExtract ) : ExpResult
	def tMatExtract( e : MatExtract)  : ExpResult
	def tIfExp(e : IfExp) : ExpResult
	def tFieldInit(e : FieldInit) : ExpResult
	def tSetInit(e : SetInit) : ExpResult
	

	def tIntL(e : IntL)  : ExpResult
	def tFloatL(e : FloatL) : ExpResult
	def tDoubleL(e : DoubleL) : ExpResult
	def tBoolL(e : BoolL) : ExpResult
	def tVectorL(e : VectorL) : ExpResult
        //def tSimdL(e : SimdL) : ExpResult
	def tMatrixL(e : MatrixL) : ExpResult
	def tStringL(e : StringL) : ExpResult
	
	def tMatrixFromVec(e : MatrixFromVec) : ExpResult
	def tSparseVectorAccess( e : SparseVectorAccess ) : ExpResult
	def tSparseMatrixAccess( e : SparseMatrixAccess ) : ExpResult
	
	def tVar( e : Var ) : ExpResult
	
	def tIndexCall(e : IndexCall) : ExpResult
	def tLinearSystemAllocate(e : LinearSystemAllocate) : ExpResult
	def tLinearSystemCall(e : LinearSystemCall) : ExpResult
		
	def tLiteral(e : Literal) : ExpResult = e match {
		case e : IntL => tIntL(e)
		case e : FloatL => tFloatL(e)
		case e : DoubleL => tDoubleL(e)
		case e : BoolL => tBoolL(e)
		case e : StringL => tStringL(e)
		//case e : SimdL => tSimdL(e)
	}
	
	def tExp(e : Exp) : ExpResult = e match {
		case e : Binary => tBinary(e)
		case e : Unary => tUnary(e)
		case e : Call => tCall(e)
		case e : TopoCall => tTopoCall(e)
		case e : FieldRead => tFieldRead(e)
		case e : VecExtract => tVecExtract(e)
		case e : MatExtract => tMatExtract(e)
		case e : IfExp => tIfExp(e)
		case e : FieldInit => tFieldInit(e)
		case e : SetInit => tSetInit(e)
		case e : Literal => tLiteral(e)
		case e : MatrixFromVec => tMatrixFromVec(e)
		case e : SparseMatrixAccess => tSparseMatrixAccess(e)
		case e : SparseVectorAccess => tSparseVectorAccess(e)
		case e : VectorL => tVectorL(e)
		case e : MatrixL => tMatrixL(e)
		case e : Var => tVar(e)
		
		case e : IndexCall => tIndexCall(e)
		case e : LinearSystemAllocate => tLinearSystemAllocate(e)
		case e : LinearSystemCall => tLinearSystemCall(e)
		case _ : EmptyExp => Log.ierror("unexpected EmptyExp")
	}
	
	//Statement translators
	
	def tIf(e : If) : StmtResult
	def tFieldWrite(e : FieldWrite) : StmtResult
	def tWhile(e : While) : StmtResult
	def tForall(e : Forall) : StmtResult
	def tReturn(e : Return) : StmtResult
	def tPrint(e : Print) : StmtResult
        def tExit : StmtResult
	def tDefVar(e : DefVar) : StmtResult
	def tAssign(e : Assign) : StmtResult
	def tEnsureFieldState(e : EnsureFieldState) : StmtResult
	def tBlock(e : Block) : StmtResult
	
	def tFunctionHeader(e : DefFunc, isDeclaration : Boolean) : ExpResult
	def tDefFunc(e : DefFunc) : StmtResult
	
	def tExpStmt(e : ExpStmt) : StmtResult
	def tExternStmt : StmtResult
	
	def tDefLinearSystem( e : DefLinearSystem) : StmtResult
	def tLinearSystemNonzeroes( e : LinearSystemNonzeroes) : StmtResult
	def tLinearSystemPlaceNonzero(e : LinearSystemPlaceNonzero) : StmtResult
	
	def tWorkgroup(e : Workgroup) : StmtResult
	def tPerfCounter(e : PerfCounter) : StmtResult
	
	def tStmt(e : Stmt) : StmtResult = e match {
		case e : If => tIf(e)
		case e : FieldWrite => tFieldWrite(e)
		case e : While => tWhile(e)
		case e : Forall => tForall(e)
		case e : Return => tReturn(e)
		case e : Print => tPrint(e)
		case e : DefVar => tDefVar(e)
		case e : Assign => tAssign(e)
		case e : EnsureFieldState => tEnsureFieldState(e)
		case e : Block => tBlock(e)
		case e : DefFunc => tDefFunc(e)
		case e : ExpStmt => tExpStmt(e)
                case Exit => tExit;
		case e : DefLinearSystem => tDefLinearSystem( e )
		case e : LinearSystemNonzeroes => tLinearSystemNonzeroes( e )
		case e : LinearSystemPlaceNonzero => tLinearSystemPlaceNonzero( e )
		case e : Index => Log.ierror("unexpected Index node")
		case e : Workgroup => tWorkgroup(e)
		case e : PerfCounter => tPerfCounter(e)
		case ExternStmt => tExternStmt
		
	}
	
	
	//a unique identifier for each type matching [A-Za-z][A-Za-z0-9_]+
	//can be used to differentiate function overloads
	def typeID(t : Type) : String = t match {
		case Vertex => "v"
		case Edge => "e"
		case Face => "f"
		case Cell => "c"
		case Mesh => "m"
		case FieldT(mo,t) => "f_%s_%s".format(typeID(mo),typeID(t))
		case Set(t) => "set_%s".format(typeID(t))
		case IntT =>  "i"
		case FloatT => "f"
		case DoubleT => "s"
		case StringT => "str"
		case BoolT => "b"
		case VectorT(n,t) => "vec_%s_%s".format(typeID(n),typeID(t))
                case SimdT(t) => "simd_%s".format(typeID(t));
		case MatrixT(n,m,t) => "mat_%s_%s_%s".format(typeID(n),typeID(m),typeID(t))
		case UnitT => "u"
		case _ : SparseVectorType => "sv"
		case _ : SparseMatrixType => "sm"
		case IndexType(true) => "idx_r"
		case IndexType(false) => "idx_c"
		case MetaIntT(i) => "%d".format(i)
		case _ => Log.ierror("unexpected type: %s",t)
	}
	//during codegen, code may generate helper functions which are needed to get the code to compile
	//this function should return those helper functions as generated code.
	def tHelperFunctions() : StmtResult
	
	def create(build_dir : String) : Unit
}
