package Liszt
import scala.collection.immutable.{Set => SSet}
import Liszt.Types._
import Liszt.Trees._
import scala.collection.mutable.Map
class CommonRuntimeCodeGen(val top : CompilationUnit, build_dir : String) {
	val Log = new Log("CommonRuntimeCodeGen")
	class FunctionEntry(var tree : DefFunc) {
		def name = tree.name
		var isCalledNested = false
		var isCalledUnnested = false
	}
	
	class ScalarEntry(val name : String, val typ : Type) {
		var isUsedNested = false
	}
	class FieldEntry(val name : String, val id : Int, val typ : Type)
	class SetEntry(val name : String, val typ : Type)
	class KernelEntry(val name : Int, val loop_var : DefVar, val body : Block, val mustBeTrivial : Boolean)
	
	val function_table = new scala.collection.mutable.HashMap[String,FunctionEntry]
	val scalar_table = new scala.collection.mutable.HashMap[String,ScalarEntry]
	val field_table = new scala.collection.mutable.HashMap[String,FieldEntry]
	val set_table = new scala.collection.mutable.HashMap[String,SetEntry]
	val kernel_table = new scala.collection.mutable.HashMap[Int,KernelEntry]
	
	//1. build symbol tables
	buildSymbolTable()
	//2. calculate function usage
	calculateFunctionUsage()
	//3. calculate variable usage
	calculateVariableUsage()
	//4. find local variables used in a nested context and promote to globals
	promoteLocalsToGlobals()
	//5. promote all globally-scope scalars to lScalar objects
	//   and remove reductions from assignment blocks
	insertScalarAccessors()
	//6. promote unnested for-comprehensions to kernel launches and fill in kernel table
	promoteKernels()
	
	//7. Remove references to mesh objects (they are not reified in the runtime and are safe to remove mow)
	removeMeshReferences()
	
	
	//open the output file
	val filename = "%s/generated.cpp".format(build_dir)
	val file_output = new java.io.FileOutputStream(filename)
	val out = new java.io.PrintWriter(file_output)
	var indent = 0
	
	def emit(line : String) {
		out.println( ("    " * indent) + line )
	}
	def begin(line : String) {
		emit(line)
		indent += 1
	}
	def end(line : String) {
		indent -= 1
		emit(line)
	}
		
	//9. emit code
	emitCode()
	
	
	//8. Generate stencil by abstractly interpreting the code
	emitStencil()

	out.close()
	
	//1. build symbol tables
	
	def buildSymbolTable() {
		val init_stmts = new scala.collection.mutable.ListBuffer[Stmt]
		def asAssign(dv : DefVar) : Assign = {
			val name = ASTUtil.Names.create("assign_to_global")
			val v = Var(dv.init.t,name)
			init_stmts += DefVar(name,dv.init)
			Assign(Var(dv.init.t,dv.name),Ops.Assign,v)
		}
		var next_field_id = 0
		def fresh_field_id() = {
			next_field_id += 1
			next_field_id - 1
		}
		def processStmt(stmt : Stmt) {
	
			stmt match {
				case df : DefFunc => 
					if(df.body != ExternStmt)
						function_table += (df.name -> new FunctionEntry(df))
				case dv @ DefVar(name,init) =>
					init match {
						case SetInit(t,_) =>
							set_table += (name -> new SetEntry(name,t))
							init_stmts += dv
						case FieldInit(t,_) =>
							field_table += (name -> new FieldEntry(name,fresh_field_id(),t))
							init_stmts += dv
						case _  => 
							if(!dv.synthetic) {
								scalar_table += (name -> new ScalarEntry(name,init.t))
								init_stmts += asAssign(dv)
							} else {
								init_stmts += dv
							}
					}
				case _ =>
					init_stmts += stmt
			}
		}

		
		for(stmt <- top.code.stmts) {
			processStmt(stmt)
		}
		val unit_fn = Func(List(UnitT),UnitT)
		val call_start = ExpStmt(Call(UnitT,Var(unit_fn,"_liszt_start"),List.empty))
		init_stmts += call_start
		
		val block =  Block(init_stmts.toList, EmptyExp(UnitT))
		val init_func = DefFunc(unit_fn,"_liszt_init_globals",List.empty,block)
		function_table += (init_func.name -> new FunctionEntry(init_func))
	}
	
	def calculateFunctionUsage() {	
		def markFunction(function : String, nested : Boolean, unnested : Boolean) {
			def markCallsNested(tree : Tree) { 
				tree match {
					case c : Call => markFunction(c.func.name,true,false)
					case tree => tree.children foreach markCallsNested
				}
			}
			def markCallsUnnested(tree : Tree) {
				tree match {
					case c : Call => markFunction(c.func.name,false,true)
					case f : Forall => markCallsNested(f.block)
					case tree => tree.children foreach markCallsUnnested
				}
			}
			for(fe <- function_table.get(function)) {
				if(nested && !fe.isCalledNested) {
					fe.isCalledNested = true
					markCallsNested(fe.tree)
				}
				if(unnested && !fe.isCalledUnnested) {
					fe.isCalledUnnested = true
					markCallsUnnested(fe.tree)
				}
			}
		}
		val entry_points = List("_liszt_init_globals")
		for(e <- entry_points) { 
			markFunction(e,false,true)
		}
	}

	def calculateVariableUsage() {
		def findUses(tree : Tree, nested : Boolean) {
			tree match {
				case fa : Forall if !nested =>
					findUses(tree,true)
				case v : Var if nested => 
					for(se <- scalar_table.get(v.name))
						se.isUsedNested = true 
				case e =>
					for(c <- e.children)
						findUses(c,nested)
				
			}
		}
		for(fe <- function_table.values)
			findUses(fe.tree,fe.isCalledNested)
	}
	
	def promoteLocalsToGlobals() {
		def isScalar(t : Type) = t match {
			case _ : Topo => false
			case _ : Set => false
			case _ : FieldT => false
			case _ => true
		}
		def unnestedScalarDefines(tree : Tree) : List[DefVar] = tree match {
			case f : Forall => List.empty
			case dv : DefVar if isScalar(dv.init.t) => List(dv)
			case _ => tree.children flatMap unnestedScalarDefines
		}
		def nestedUses(nested : Boolean, tree : Tree) : List[String] = tree match {
			case f : Forall if !nested => 
				nestedUses(true,f)
			case v : Var if nested => List(v.name)
			case _ => 
				for(c <- tree.children; df <- nestedUses(nested,c)) 
					yield df
		}	
		for(fe <- function_table.values if fe.isCalledUnnested) {
			val nus = SSet.empty ++ nestedUses(false,fe.tree)	
			val nds = unnestedScalarDefines(fe.tree)
			for(nd <- nds if nus.contains(nd.name)) {
				scalar_table += (nd.name -> new ScalarEntry(nd.name,nd.init.t))
			}
		}
	}
	
	def insertScalarAccessors() {
		def insertScalar(stmt_ : Stmt) : List[Stmt] = {
			
			val reads = new scala.collection.mutable.ListBuffer[Stmt]
			def insertReads(t : Tree) : Tree = t match {
				case v : Var if !v.isWrite && scalar_table.contains(v.name) =>
					val tmp = ASTUtil.Names.create(v.name)
					reads += DefVar(tmp, ScalarRead(v))
					Var(v.t,tmp)
				case _ => ShapeAny.trees(t,t.children map insertReads)
			}
			val stmt : Stmt = stmt_ match {
				case _ : EnsureFieldState => stmt_ //these statements need to operate on the lScalar object, so no read is performed
				case s : Stmt if s.isLeaf => insertReads(s).asInstanceOf[Stmt]
				case _ => stmt_
			}
			
			val writes : List[Stmt] = stmt match {
				case dv @ DefVar(name,init) if scalar_table.contains(name) =>
					List( stmt,
				    	  ScalarWrite(Var(init.t,name), IntL(0), Ops.Assign, Var(init.t,name)) )
				case Assign(v : Var, op, rhs) if scalar_table.contains(v.name) =>
					(v.t,rhs.t) match {
						case (lt,rt) if lt == rt =>List( ScalarWrite(v,IntL(0),op, rhs) )
						case (VectorT(MetaIntT(n),_),rt) =>
							val inits = for(i <- 0 until n)
								yield rhs
							val tmp = ASTUtil.Names.create("tmp")
							List( 
							  DefVar(tmp,VectorL(VectorT(MetaIntT(n),rhs.t),inits.toList)),
							  ScalarWrite(v,IntL(0),op,Var(v.t,tmp))
							)
						case (MatrixT(MetaIntT(r), MetaIntT(c),_),rt) =>
							val inits = for(i <- 0 until r;  j <- 0 until c)
								yield rhs
							val tmp = ASTUtil.Names.create("tmp")
							List( 
							  DefVar(tmp,MatrixL(MatrixT(MetaIntT(r),MetaIntT(c),rhs.t),inits.toList)),
							  ScalarWrite(v,IntL(0),op,Var(v.t,tmp))
							)
					}
				case Assign(ve : VecExtract, op, rhs) if scalar_table.contains(ve.vec.name) =>
					List( ScalarWrite(ve.vec, ve.n, op, rhs) )
				case Assign(me : MatExtract, op, rhs) if scalar_table.contains(me.mat.name) =>
					val MatrixT(MetaIntT(r), MetaIntT(c),_) = me.mat.t
					//offset = c * me.n + me.m
					val tmp0 = ASTUtil.Names.create("tmp")
					val tmp1 = ASTUtil.Names.create("tmp")
					val tmp2 = ASTUtil.Names.create("tmp")
					List(
						DefVar(tmp0,IntL(c)),
						DefVar(tmp1,Binary(IntT, Var(IntT,tmp0), Ops.Times, me.n)),
						DefVar(tmp2,Binary(IntT, Var(IntT,tmp1), Ops.Plus, me.m)),
						ScalarWrite(me.mat, Var(IntT,tmp2), op, rhs)
						)
				//remove a += b and replace with c = a + b;, a = c;
				//this is only safe to do after we have determined this isn't a scalar reduction variable (which we just did above)
				case Assign(lhs, op : Ops.Reduce, rhs) =>
					val tmp = ASTUtil.Names.create("assign")
					val prefix = lhs match {
						case v : Var => List( DefVar(tmp,Binary(lhs.t,v,op.op,rhs)) )
						case exp =>
							val tmp2 = ASTUtil.Names.create("assign")
							List( DefVar(tmp2,lhs),
							      DefVar(tmp,Binary(lhs.t,Var(lhs.t,tmp2),op.op,rhs)) )
					}
					val as = Assign(lhs,Ops.Assign,Var(lhs.t,tmp))
					prefix ::: List(as)
				case _ => List(stmt)
				
			}
			reads.toList ::: writes 
		}
		def insertReadForExp(exp : SimpleExp) : (List[Stmt],SimpleExp) = {
			exp match {
				case v @ Var(t,name) if scalar_table.contains(name) => 
					val tmp = ASTUtil.Names.create(name)
					(List( DefVar(tmp,ScalarRead(v)) ),
					 Var(v.t,tmp))
				case _ => (List.empty,exp)
					 
			}
		}
		
		for(fe <- function_table.values) {
			//paramaters to the function may have been promoted to globals. In this case, the function call acts as an assignment to the global,
			//so we need to insert explicit assignments at the beginning of the function
			val param_writes = for(a <- fe.tree.args if scalar_table.contains(a.name)) 
				yield ScalarWrite(Var(a.init.t,a.name), IntL(0), Ops.Assign, Var(a.init.t,a.name))
				
			val new_body = ASTUtil.flatMapBlocksAndExp(fe.tree.body)(insertScalar,insertReadForExp).asInstanceOf[Block]
			val new_block = Block(param_writes ::: new_body.stmts, new_body.exp)
			fe.tree = fe.tree.copy(body = new_block)
		}
		
		//add scalar init statments to the entry point
		val init = function_table("_liszt_init_globals")
		val scalar_inits = for(se <- scalar_table.values)
			yield ScalarInit(Var(se.typ,se.name))
		init.tree = init.tree.copy(body = Block(scalar_inits.toList ::: init.tree.body.asInstanceOf[Block].stmts, init.tree.body.asInstanceOf[Block].exp))
	}
	def promoteKernels() {
		var next_id = 0
		def getID() = { next_id += 1; next_id - 1}
		
		for(fe <- function_table.values if fe.isCalledUnnested) {
			def insertExec(tree : Tree) : Tree = tree match {
				case fa : Forall => {
					if(fe.isCalledNested) 
						Log.error("forall stmt in nested+unnested code, this shouldn't be possible...")
					val id = getID()
					kernel_table += (id -> new KernelEntry(id, fa.name, fa.block, fa.mustBeTrivial))
					ExecKernel(id,fa.set)
				}
				case _ => ShapeAny.trees(tree,tree.children map insertExec)
			}
			fe.tree = insertExec(fe.tree).asInstanceOf[DefFunc]
		}
	}
	
	def removeMeshReferences() {
		def removeUses(tree : Tree) : Tree = tree match {
			case c : Call => c.copy(args = c.args.filter(x => x.t != Mesh))
			case Return(v : Var)  if v.t == Mesh => Return(EmptyExp(UnitT))
			case _ => tree
		}
		def removeMesh(stmt : Stmt) : List[Stmt] = stmt match {
			case DefVar(_,init) if init.t == Mesh => List.empty
			case _ => List(removeUses(stmt).asInstanceOf[Stmt])
		}
		for(fe <- function_table.values) {
			val df = fe.tree
			val ntargs = for(a <- df.typ.args if a != Mesh)
				yield a
			val nrange = if (df.typ.range == Mesh) UnitT else df.typ.range
			val ntyp = Func(ntargs,nrange)
			
			val nargs = for(a <- df.args if a.init.t != Mesh) 
				yield a
			
			val nbody = ASTUtil.flatMapBlocks(df.body)(removeMesh).asInstanceOf[Stmt]
			fe.tree = DefFunc(ntyp, df.name, nargs ,nbody)
		}
	}
	
	
	
	//functions shared between emitCode and emitStencil
	
	def tKernelStencilHeader(ke : KernelEntry) : String = {
		"L_STENCIL lStencilData kernel_%d_stencil_data()".format(ke.name)
	}
	def tFunctionStencilHeader(fe : FunctionEntry) : String = {
		"L_STENCIL void %s_stencil(lsFunctionTable * __vtbl, lsContext * __ctx)".format(fe.tree.name)
	}
	
	def typeTag(t : Type) : String = t match {
		case IntT => "L_INT"
		case FloatT => "L_FLOAT"
		case DoubleT => "L_DOUBLE"
		case StringT => "L_STRING"
		case BoolT => "L_BOOL"
		case Vertex => "L_VERTEX"
		case Edge => "L_EDGE"
		case Face => "L_FACE"
		case Cell => "L_CELL"
		case MatrixT(_,_,t) => typeTag(t)
		case VectorT(_,t) => typeTag(t)
		case _ => Log.ierror("unexpected tagged type")
	}
	def tReduction(op : Ops.ReductionOp) = op match {
		case Ops.RSum => "PLUS"
		case Ops.RSub => "MINUS"
		case Ops.RProd => "MULTIPLY"
		case Ops.RDiv => "DIVIDE"
		case Ops.RMin => "MIN"
		case Ops.RMax => "MAX"
		case Ops.RBOr => "BOR"
		case Ops.RBAnd => "BAND"
		case Ops.RLAnd => "AND"
		case Ops.RBXor => "XOR"
		case Ops.RLOr => "OR"
		case Ops.RAssign => "ASSIGN"
		case _ => Log.ierror("not a reduction")
	}
	def tPhase(op : Ops.ReductionOp) = op match {
		case Ops.RAssign => "L_WRITE_ONLY"
		case Ops.RNone => "L_READ_ONLY"
		case Ops.RModify => "L_MODIFY"
		case _ => "L_REDUCE_" + tReduction(op)
	}
	def opTag(op : Ops.AssignOp) = "L_" + tReduction(op.asReduction(false))
			
	def tTopoFn(op : Ops.TopoFn) : String = {
		import Liszt.Ops._
		def tString(t : Topo) = t match {
			case Mesh => "Mesh"
			case Vertex => "Vertex"
			case Edge => "Edge"
			case Face => "Face"
			case Cell => "Cell"
		}
		def tsString(t : Topo) = t match {
			case Vertex => "Vertices"
			case Mesh => "Meshes"
			case _ => tString(t) + "s"
		}
		def dString(dir : TopoDir) = dir match {
			case CW => "CW"
			case CCW => "CCW"
			case NoDir => ""
		}
		op match {
			case TopoSet(result, _, Mesh) =>
				"%sOfMesh".format(tsString(result))
			case TopoSet(result, dir, arg) =>
				"%sOf%s%s".format(tsString(result),tString(arg),dString(dir))
			case Head(argt) => 
				"HeadOf%s".format(tString(argt))
			case Tail(argt) =>
				"TailOf%s".format(tString(argt))
			case Flip(argt) => 
				"Flip%s".format(tString(argt))
			case Towards(argt, ref) =>
				"Towards%s".format(tString(argt))
			case LabelExtract(rett, argt) =>
				"%sOf%sWithLabel".format(tString(rett),tString(argt))
			case IDLookup(argt) =>
				"IDOf%s".format(tString(argt))
		}
	}
	
	def emitCode() { 
		def tType(nested : Boolean)(t : Type) : String = t match {
			case Vertex => "lkElement"
			case Edge => "lkElement"
			case Face => "lkElement"
			case Cell => "lkElement"
			case IntT =>  "int"
			case FloatT => "float"
			case DoubleT => "double"
			case StringT => "const char *"
			case BoolT => "bool"
			case VectorT(n,t) => "vec< %s , %s >".format(tType(nested)(n),tType(nested)(t))
			case MatrixT(n,m,t) => "matrix< %s , %s , %s >".format(tType(nested)(n),tType(nested)(m),tType(nested)(t))
			case UnitT => "void"
			case MetaIntT(n) => n.toString
			case _ : FieldT if nested =>  "lkField"
			case _ : FieldT if !nested => "lField"
			case _ : Set if nested => "lkSet"
			case _ : Set if !nested => "lSet"
			case _ => Log.ierror("unexpected type: %s",t)
	    }
		
		
		def tFunctionHeader(df : DefFunc, nested : Boolean) = {
			val nesting = if(!nested) "UNNESTED" else "NESTED"
			val Func(args,r) = df.ftype
			val return_type = tType(nested)(r)
			val ctx = if(!nested) "lContext" else "lkContext"
			val arguments = for( (t,dv) <- args zip df.args)
				yield ", %s %s".format(tType(nested)(t), dv.name)
			"L_%s %s %s_%s(%s * __ctx %s)".format(nesting, return_type, df.name, nesting.toLowerCase,ctx, arguments.mkString)
		}
		def tKernelHeader(ke : KernelEntry) = {
			"L_KERNEL void kernel_%d(lkContext __ctx_struct)".format(ke.name)
		}
		def emitStruct(name : String, member_type : String, members : Iterable[String]) {
			begin("struct %s {".format(name))
			for(m <- members)
				emit("%s %s;".format(member_type,m))
			end("};")
		}
		def tBinOp(op : Ops.BinOp) : String = {
			import Liszt.Ops._
			op match {
				case Plus => "+"
				case Minus => "-"
				case Times => "*"
				case Div => "/"
				case Mod => "%"
				case Eq => "=="
				case NEq => "!="
				case Less => "<"
				case LessEq => "<="
				case Greater => ">"
				case GreaterEq => ">="
				case Min => "min"
				case Max => "max"
				case LShift => "<<"
				case RShift => ">>"
				case BOr => "|"
				case BAnd => "&"
				case BXor => "^"
				case And => "&&"
				case Or => "||"
			}
		}
		def tUnOp(op : Ops.UnOp) = op match {
			case Ops.Neg => "-"
			case Ops.BNot => "~"
			case Ops.Not => "!" 
		}
	
		//sets are sometimes in the set_table (when global) and sometimes not (when produced by functions like faces(mesh))
		def tVar(set : Var) = if (set_table.contains(set.name)) "lGetSets(__ctx)->" + set.name else set.name
		
		def emitBasicExp(nested : Boolean)(name : String, typ : Type, exp : String) {
			emit("%s %s = %s;".format(tType(nested)(typ),name,exp))
		}
		
		
		def typeLength(t : Type) : Int = t match {
			case _ : BasicType => 1
			case VectorT(MetaIntT(i),t) => i * typeLength(t)
			case MatrixT(MetaIntT(i),MetaIntT(j),t) => i * j * typeLength(t)
			case _ => Log.ierror("unexpected length type")
		}
		def typeRows(t : Type) : Int = t match {
			case _ : BasicType => 0
			case VectorT(MetaIntT(i), _) => i
			case MatrixT(MetaIntT(i), _, _) => i
			case _ => Log.ierror("unexpected type rows")
		}
		def typeCols(t : Type) : Int = t match {
			case _ : BasicType => 0
			case _ : VectorT => 0
			case MatrixT(_, MetaIntT(i), _) => i
			case _ => Log.ierror("unexpected type rows")
		}
		
		def emitTopoCall(name : String, op : Ops.TopoFn, args_ : List[Var]) {
			op match {
				case Ops.TopoSet(_ , _ , from) =>
					val prefix = if(from == Mesh) "l" else "lk"
					val args = for( (a,t) <- args_ zip op.argts if t != Mesh)
						yield "%s,".format(a)
					val arg_string = args.mkString("")
					emit("%sSet %s;".format(prefix,name))
					emit("%s%s(__ctx,%s&%s);".format(prefix,tTopoFn(op),arg_string,name))
				case Ops.IDLookup(argt) =>
					emit("int %s = lk%s(__ctx,%s);".format(name,tTopoFn(op),args_.mkString(",")))
				case _ => 
					emit("lkElement %s = lk%s(__ctx,%s);".format(name,tTopoFn(op),args_.mkString(",")))
			}
		}
		def emitExp(nested : Boolean)(nameOp : Option[String], exp : Exp) {
			
			def functionCall(c : Call) : String = {
				def nesting = if(nested) "nested" else "unnested"
				def arg_list = ( for(a <- c.args) yield ", " + tVar(a) ).mkString //tVar is called to change the variable into a set reference if it is a global set
				if(function_table.contains(c.func.name)) {
					"%s_%s(__ctx %s)".format(c.func,nesting,arg_list)
				}  else {
					c.func.name match {
					//TODO: there are some 'special' functions in the AST like size, and toFloat which should really be custom AST nodes
						case "size" =>
							val fn = if(nested) "lkSetSize" else "lSetSize"
							"%s(__ctx,&%s)".format(fn,tVar(c.args.head))
						case "wall_time" =>  "lWallTime(__ctx)"
						case _ => "%s(%s)".format(c.func.name,c.args.mkString(","))
					}
					
				}
			}
			nameOp match {
				case None =>
					exp match {
						case c @ Call(t,func,args) => emit("%s;".format(functionCall(c)))
						case _ => /* other ops have no side effects and can be dropped. 
						             (IfExp may have side effects but cannot be generated without a label) */
					
					}
				case Some(name) =>
					val typ = tType(nested)(exp.t)
					def d(s : String) {
						emit("%s %s = %s;".format(typ,name,s))
					}
					exp match {
						case v @ Var(t,name) => d(tVar(v))
						case c : Call => d(functionCall(c))
						case IntL(c) => d(c.toString)
						case FloatL(c) =>  d(c.toString)
						case DoubleL(c) => d(c.toString)
						case StringL(c) => d("\"%s\"".format(ASTUtil.escapeString(c)))
						case BoolL(c) => if(c) d("true") else d("false")
						case Unary(t,op,e) => d("%s%s".format(tUnOp(op),e))
						case VecExtract(t,vec,arg) => d("%s[%s]".format(vec,arg))
						case MatExtract(t,vec,r,c) => d("%s(%s,%s)".format(vec,r,c))
						case MatrixFromVec(t, es) =>
							val MatrixT(MetaIntT(r),MetaIntT(c),_) = t
							val elements = for(e <- es; i <- 0 until c)
								yield "%s[%d]".format(e,i)
							d("{ %s }".format(elements.mkString(",")))
						case VectorL(t,es) => d("{%s}".format(es.mkString(",")))
						case MatrixL(t,es) => d("{%s}".format(es.mkString(",")))
						case Binary(t,l,op : Ops.PickOp,r) => 
							val fn = tBinOp(op)
							d("%s(%s,%s)".format(fn,l,r))
						case Binary(t,l,op,r) => d("%s %s %s".format(l,tBinOp(op),r))
						case IfExp(t,predE,thenE,elseE) =>
							//thenE and elseE will define __if_exp_var when the block is generated
							//since the block will generate the assignment when it sees a non-empty block terminator
							
							emit("%s %s;".format(typ,name))
							begin("{")
							emit("%s __if_exp_var;".format(tType(nested)(t)))
							emitStmt(nested)(If(predE,thenE,elseE))
							emit("%s = __if_exp_var;".format(name))
							end("}")
						case SetInit(t,s) => 
							val Set(tt) = t
							emit("lSetInitBoundary(__ctx,&lGetSets(__ctx)->%s,%s,\"%s\");".format(name,typeTag(tt),s))
						case FieldInit(t,src) =>
							val FieldT(d,r) = t
							val dt = typeTag(d)
							val rt = typeTag(r)
							val length = typeLength(r)
							val id = field_table(name).id
							emit("lFieldInit(__ctx,&lGetFields(__ctx)->%s,%d,%s,%s,%d);".format(name,id,dt,rt,length))
							src match {
								case ConstSource(e) =>
									emit("lFieldBroadcast(__ctx,&lGetFields(__ctx)->%s,%s,%s,%d,&%s);".format(name,dt,rt,length,e))
								case StringSource(e) =>
									emit("lFieldLoadData(__ctx,&lGetFields(__ctx)->%s,%s,%s,%d,\"%s\");".format(name,dt,rt,length,ASTUtil.escapeString(e)))
							}
						case FieldRead(t,field,arg) =>
							val FieldT(d,r) = field.t
							val rt = typeTag(r)
							val length = typeLength(r)							
							emit("%s %s;".format(tType(nested)(t),name))
							emit("lkFieldRead(&lkGetFields(__ctx)->%s,%s,%s,%d,0,%d,&%s);".format(field,arg,rt,length,length,name))
						case ScalarRead(field) =>
							val p = if(nested) "lk" else "l"
							val rt = typeTag(field.t)
							val length = typeLength(field.t)							
							emit("%s %s;".format(tType(nested)(field.t),name))
							emit("%sScalarRead(__ctx,&%sGetScalars(__ctx)->%s,%s,%d,0,%d,&%s);".format(p,p,field,rt,length,length,name))
						case TopoCall(t, fn, args) => emitTopoCall(name,fn,args)
					}
			}
		}
		def emitStmt(nested : Boolean)(stmt : Stmt) { stmt match {
			case DefVar(name,exp) => emitExp(nested)(Some(name),exp)
			case ExpStmt(e) => emitExp(nested)(None,e)
			case If(predE,thenE,elseE) =>
				for(stmt <- predE.stmts)
					emitStmt(nested)(stmt)
				emit("if(%s)".format(predE.exp))
				emitStmt(nested)(thenE)
				emit("else")
				emitStmt(nested)(elseE)
			case Forall(name,lbl,set,block) =>
				val iter = ASTUtil.Names.create("iter")
				val Set(styp) = set.t
				val tTag = typeTag(styp)
				emit("lkIterator %s;".format(iter))
				emit("lkSetGetIterator(__ctx,&%s,%s,&%s);".format(set,tTag,iter))
				emit("lkElement %s;".format(name.name))
				
				val lbl_var = lbl match {
					case None => "NULL"
					case Some(x) => 
						emit("int %s;".format(x.name))
						"&%s".format(x.name)
				}
				
				emit("while(lkIteratorNext(__ctx,&%s,%s,&%s,%s))".format(iter,tTag,name.name,lbl_var))
				emitStmt(nested)(block)
			case Assign(VecExtract(_,vec,arg),Ops.Assign,rhs) =>
				emit("%s[%s] = %s;".format(vec,arg,rhs))
			case Assign(MatExtract(_,vec,a,b),Ops.Assign,rhs) =>
				emit("%s(%s,%s) = %s;".format(vec,a,b,rhs))
			case Assign(lhs : Var,Ops.Assign,rhs) =>
				emit("%s = %s;".format(lhs,rhs))
			case Return(_ : EmptyExp) =>
				emit("return;")
			case Return(v : Var) =>
				emit("return %s;".format(v))
			case While(pred,block) =>
				val tmp = ASTUtil.Names.create("pred")
				emit("bool %s;".format(tmp))
				begin("{")
				for(stmt <- pred.stmts)
					emitStmt(nested)(stmt)
				emit("%s = %s;".format(tmp,pred.exp))
				end("}")
				begin("if(%s) {".format(tmp))
				begin("do {")
				for(stmt <- block.stmts)
					emitStmt(nested)(stmt)
				for(stmt <- pred.stmts)
					emitStmt(nested)(stmt)
				emit("%s = %s;".format(tmp,pred.exp))
				end("} while(%s);".format(tmp))
				end("}")
			case Block(ess,ee) => 
				begin("{")
				for(stmt <- ess)
					emitStmt(nested)(stmt)
				ee match {
					case v : Var =>
						emit("__if_exp_var = %s;".format(tVar(v)))
					case _ =>
				}
				end("}")
			case FieldWrite(field,arg,op,exp) =>
				val FieldT(domain,range) = field.t
				val length = typeLength(range)
				val rT = typeTag(range)
				val opT = opTag(op)
				//we need to be careful about promoting the rhs if it is a scalar while the lhs is a vector
				//the compiler should really insert these conversions very early on, but lifting the AST from scala makes this difficult
				val rhs = range match {
					case r if r == exp.t => exp.name
					case VectorT(MetaIntT(n),_) =>
						val tmp = ASTUtil.Names.create("tmp")
						val names = for(i <- 0 until n)
							yield exp.name
						emit("%s  %s = { %s };".format(tType(nested)(range),tmp, names.mkString(",")))
						tmp
					case MatrixT(MetaIntT(r),MetaIntT(c),_) =>
						val tmp = ASTUtil.Names.create("tmp")
						val names = for(i <- 0 until r; j <- 0 until c)
							yield exp.name
						emit("%s %s = { %s };".format(tType(nested)(range),tmp, names.mkString(",")))
						tmp
				}
				emit("lkFieldWrite(&lkGetFields(__ctx)->%s,%s,%s,%s,%d,0,%d,&%s);".format(field,arg,opT,rT,length,length,rhs))
			case ScalarInit(scalar) =>
				val length = typeLength(scalar.t)
				val rTag = typeTag(scalar.t)
				emit("lScalarInit(__ctx,&lGetScalars(__ctx)->%s,%s,%d);".format(scalar,rTag,length))
			case ScalarWrite(scalar,offset,op,exp) =>
				val p = if(nested) "lk" else "l"
				val element_length = typeLength(scalar.t)
				val length = typeLength(exp.t)
				val rTag = typeTag(scalar.t)
				val opT = opTag(op)
				emit("%sScalarWrite(__ctx,&%sGetScalars(__ctx)->%s,%s,%s,%s,%s,%d,&%s);".format(p,p,scalar,opT,rTag,element_length,offset,length,exp))
			case Print(ess) =>
				val p = if(nested) "lk" else "l"
				emit("%sPrintBegin(__ctx);".format(p))
				for(e <- ess) {
					val tt = typeTag(e.t)
					
					if (e.t.isInstanceOf[Topo]) {
						emit("%sPrintElement(__ctx,%s,%s);".format(p,tt,e))
					} else {
						val et = if(e.t == StringT) "" else "&";
						emit("%sPrintValue(__ctx,%s,%s,%s,%s%s);".format(p,tt,typeRows(e.t),typeCols(e.t),et,e))
					}
				}
				emit("%sPrintEnd(__ctx);".format(p))
            case EnsureFieldState(f,reduction) =>
            	if(!nested) {
					f.t match {
						case FieldT(d,r) =>
							val tt = typeTag(r)
							val length = typeLength(r)
							val phase = tPhase(reduction)
							emit("lFieldEnterPhase(&lGetFields(__ctx)->%s,%s,%s,%s);".format(f,tt,length,phase))
						case _ =>
							val tt = typeTag(f.t)
							val length = typeLength(f.t)
							val phase = tPhase(reduction)
							emit("lScalarEnterPhase(&lGetScalars(__ctx)->%s,%s,%s,%s);".format(f,tt,length,phase))
					}
            	}
            case FieldSave(f,name) =>
            	if(!nested) {
            		val FieldT(d,r) = f.t
					val dt = typeTag(d)
					val rt = typeTag(r)
					val length = typeLength(r)
					emit("lFieldSaveData(__ctx,&lGetFields(__ctx)->%s,%s,%s,%d,%s);".format(f,dt,rt,length,name))
            	} else Log.error("fields cannot be saved from a nested function")
            case ExecKernel(id,set) =>
            	val set_var = tVar(set)
            	val Set(styp) = set.t
            	val tTag = typeTag(styp)
            	emit("lKernelRun(__ctx, &%s, %s, %d, kernel_%d, kernel_%d_stencil_data());".format(set_var,tTag,id,id,id))
            case _ : PerfCounter => 
		} /*end match*/ } /* end function */
		
		def emitKernel(ke : KernelEntry) {
			emit(tKernelHeader(ke))
			begin("{")
			emit("lkContext * __ctx = &__ctx_struct;")
			emit("lkElement %s;".format(ke.loop_var.name))
			emit("if(lkGetActiveElement(__ctx,&%s))".format(ke.loop_var.name))
			emitStmt(true)(ke.body)
			end("}")
		}
		
		//declare types to hold the contexts for fields, scalars and sets
		emitStruct("lFields","lField", field_table.keys)
		emitStruct("lkFields","lkField", field_table.keys)
		emitStruct("lScalars","lScalar", scalar_table.keys)
		emitStruct("lkScalars","lkScalar", scalar_table.keys)
		emitStruct("lSets","lSet", set_table.keys)
		//declare functions
		for(fe <- function_table.values) {
			if(fe.isCalledNested)
				emit(tFunctionHeader(fe.tree,true) + ";")
			if(fe.isCalledUnnested)
				emit(tFunctionHeader(fe.tree,false) + ";")
		}
		//declare entry point function stencil
		
		for(fe <- function_table.get("_liszt_init_globals")) {
			emit(tFunctionStencilHeader(fe) + ";")
		}
		
		//declare kernels and kernel stencil
		for(ke <- kernel_table.values) {
			emit(tKernelHeader(ke) + ";")
			emit(tKernelStencilHeader(ke) + ";")
		}
		//define functions
		for(fe <- function_table.values) {
			if(fe.isCalledNested) {
				emit(tFunctionHeader(fe.tree,true))
				emitStmt(true)(fe.tree.body)
			}
			if(fe.isCalledUnnested) {
				emit(tFunctionHeader(fe.tree,false))
				emitStmt(false)(fe.tree.body)
			}
		}
		//define kernels
		for(ke <- kernel_table.values) {
			emitKernel(ke)
		}
		
		//emit main
		begin("int main(int argc, char ** argv) {")
		 emit(    "lProgramArguments arguments;")
		 emit(    "lUtilParseProgramArguments(argc,argv,&arguments);")
	 	 emit(    "lExec(_liszt_init_globals_unnested,_liszt_init_globals_stencil,&arguments,%d,%d,%d);".format(field_table.size, set_table.size, scalar_table.size))
		  end("}")
	}
	def emitStencil() {
		type Label = Int
		var next_label = 0
		def fresh() : Int = {
			next_label += 1
			next_label - 1
		}
		
		class Value
		case class Function(op : Ops.TopoFn, args : List[Label]) extends Value
		case class Boundary(typ : Topo, name : String) extends Value
		case class Extract(set : Label, idx : Int) extends Value
		
		class Def(val name : Label, val value : Value)
		case class Access(val field : String, val element_type : Topo, val elem : Label, val op : Ops.ReductionOp)
		
		class Frame {
			val definitions = new scala.collection.mutable.ListBuffer[Def]
			val for_statements = new scala.collection.mutable.ListBuffer[ForStatement]
			val field_access = new scala.collection.mutable.HashSet[Access]
			
			private val value_to_lbl = new scala.collection.mutable.HashMap[Value,Label]
			private val lbl_to_for = new scala.collection.mutable.HashMap[Label,ForStatement]
			
			def defineValue(value : Value) : Label = {
				value_to_lbl.getOrElseUpdate(value, {
					val lbl = fresh()
					definitions += new Def(lbl,value)
					lbl
				})
			}
			def defineForall(lbl : Label) : ForStatement = {
				lbl_to_for.getOrElseUpdate(lbl, {
					val for_stmt = new ForStatement(fresh(), lbl)
					for_statements += for_stmt
					for_stmt
				})
			}
			def read(f : String, element_type : Topo, e : Label) {
				field_access += new Access(f, element_type,e, Ops.RNone)
			}
			def write(f : String, element_type : Topo,  e : Label, op : Ops.AssignOp) {
				field_access += new Access(f,element_type, e, op.asReduction(true))
			}
		}
		
		class Top extends Frame
		class Kernel(val elem : Label, val entry : KernelEntry) extends Frame
		class ForStatement(val elem : Label, val set : Label) extends Frame
		
		type Context = scala.collection.mutable.HashMap[String,Label]
		type ReturnValues = scala.collection.mutable.HashSet[Label]

		def interpStmt(ctx : Context, frame : Frame, return_values : ReturnValues, stmt : Stmt, continuation : List[Stmt]) : (Frame, List[Stmt]) = {
			def interpCall(c : Call) : Option[List[Label]] = {
				
				def processFn(fn_impl : FunctionEntry) : Option[List[Label]] = {
					for( (actual,formal) <- c.args zip fn_impl.tree.args if ctx.contains(actual.name) )
						ctx += (formal.name -> ctx(actual.name))
					val new_return_values = new ReturnValues
					interpStmt(ctx,frame, new_return_values, fn_impl.tree.body, List.empty)
					if(new_return_values.isEmpty)
						None
					else
						Some(new_return_values.toList)
				}
				
				for(fn_impl <- function_table.get(c.func.name);
					result <- processFn(fn_impl))
						yield result
			}
			def trampoline(frm : Frame, stmts : List[Stmt]) {
				stmts match {
					case h :: t => 
						val (frm2, rest) = interpStmt(ctx, frm, return_values, h,t)
						trampoline(frm2,rest)
					case _ =>
				}
			}
			stmt match {
				case DefVar(name,TopoCall(t,fn,args)) => 
					fn match {
						case Ops.TopoSet(Mesh, _, _) => 
							ctx += (name -> frame.defineValue(Function(fn,List.empty)))
							(frame,continuation)
						case _ : Ops.IDLookup =>
							(frame,continuation)
						case le : Ops.LabelExtract => 
							val set_lbl = frame.defineValue(Function(le.asTopoSet,List(ctx(args.head.name))))
							ctx.get(args(1).name) match {
								//label is constant, just insert the call to extract from the set
								case Some(lbl) =>
									val v = frame.defineValue(Extract(set_lbl,lbl))
									ctx += (name -> v)
									(frame,continuation)
								//label is not constant, start a new forall statement to loop over possibilities
								case None =>
									val frame2 = frame.defineForall(set_lbl)
									ctx += (name -> frame2.elem)
									(frame2,continuation)
							}
						case _ => 
							val lbls = for( (a,t) <- args zip fn.argts if t != Mesh) yield ctx(a.name)
							ctx += (name -> frame.defineValue(Function(fn,lbls)))
							(frame,continuation)
					}
				case DefVar(name,SetInit(Set(t : Topo),bs)) =>
					ctx += (name -> frame.defineValue(Boundary(t,bs)))
					(frame,continuation)
				case DefVar(name,c : Call) =>
					interpCall(c) match {
						case Some(lbls) =>
							for(lbl <- lbls) {
								ctx += (name -> lbl)
								trampoline(frame,continuation)
							}
							(frame,List.empty)
						case None =>
							(frame,continuation)
					}
				case DefVar(name,Var(Set(_),set)) =>
					ctx += (name -> ctx(set))
					(frame,continuation)
				case DefVar(_,FieldRead(t,field,elem)) =>
					frame.read(field.name,elem.t.asInstanceOf[Topo],ctx(elem.name))
					(frame,continuation)
				case FieldWrite(field,elem,op,rhs) =>
					frame.write(field.name,elem.t.asInstanceOf[Topo],ctx(elem.name),op)
					(frame,continuation)
				case DefVar(name,IfExp(t,predB,thenB,elseB))
				  if t.isInstanceOf[Topo] || t.isInstanceOf[Set] =>
					interpStmt(ctx,frame,return_values,predB,List.empty)
					
					//first interpret the then block
					interpStmt(ctx,frame,return_values,thenB,List.empty)
					ctx += (name -> ctx(thenB.exp.asInstanceOf[Var].name))
					//now run the rest of the local context with name == thenExp
					//until this variable goes out of scope
					trampoline(frame,continuation)
					
					//now interpret the else block in the same way
					interpStmt(ctx,frame,return_values,elseB,List.empty)
					ctx += (name -> ctx(elseB.exp.asInstanceOf[Var].name))
					trampoline(frame,continuation)
					
					(frame,List.empty) //continuation is empty since it was already run above
				case DefVar(name,IntL(x)) => //we keep track of integer literals so we can optimize extract functions
					ctx += (name -> x)
					(frame,continuation)
				case Forall(name,lbl, set, blk) =>
					val frm = frame.defineForall(ctx(set.name))
					ctx += (name.name -> frm.elem)
					interpStmt(ctx,frm,return_values,blk,List.empty)
					(frame,continuation)
				case ExecKernel(id,set) =>
					val ke = kernel_table(id)
					val frm = frame.defineForall(ctx(set.name))
					ctx += (ke.loop_var.name -> frm.elem)
					interpStmt(ctx,frm,return_values,ke.body,List.empty)
					(frame,continuation)
				case ExpStmt(c : Call) =>
					interpCall(c)
					(frame,continuation)
				case Block(stmts,exp) =>
					trampoline(frame,stmts)
					(frame,continuation)
				case Return(v : Var) if ctx.contains(v.name) =>
					return_values += ctx(v.name)
					(frame,continuation) 
				case _ => 
					for( b @ Block(stmts,exp) <- stmt.children) {
						interpStmt(ctx,frame,return_values,b,List.empty)
					
					}
					(frame,continuation)
			}
		}
		
		
		def emitFrame(f : Frame) {
			f match {
				case _ : Top =>
					begin("{")
				case f : ForStatement =>
					val id = fresh()
					emit("lsIterator it%d;".format(id))
					emit("lsElement e%d;".format(f.elem))
					emit("__vtbl->lsSetGetIterator(__ctx, &e%d, &it%d);".format(f.set,id))
					begin("while(__vtbl->lsIteratorNext(__ctx,&it%d, &e%d)) {".format(id,f.elem))
				case k : Kernel =>
					begin("{")
					emit("lsElement e%d;\n".format(k.elem))
					emit("__vtbl->lsGetActiveElement(__ctx,&e%d);".format(k.elem))
			}
			
			for(d <- f.definitions) {
				d.value match {
					case Boundary(typ, name) => 
						emit("lsSet e%d;".format(d.name))
						emit("__vtbl->lsSetInitBoundary(__ctx,&e%d,%s,\"%s\");".format(d.name, typeTag(typ), name))
					case Function(op, args_) => 
						val typ = if(op.rett.isInstanceOf[Set]) "lsSet" else "lsElement"
						emit("%s e%d;".format(typ,d.name))
						val args = for(a <- args_)
							yield "&e%d, ".format(a)
							
						emit("__vtbl->ls%s(__ctx,%s&e%d);".format(tTopoFn(op),args.mkString(""),d.name)) 
					case Extract(set,lbl) =>
						emit("lsElement e%d;".format(d.name))
						emit("__vtbl->lsSetExtractElement(__ctx,&e%d,%d,&e%d);".format(set,lbl,d.name))
				}
			}
			for(fi <- f.field_access) {
				val field = field_table(fi.field)
				val et = typeTag(fi.element_type)
				emit("__vtbl->lsFieldAccess(__ctx,%d,%s,&e%d,%s);".format(field.id, et , fi.elem, tPhase(fi.op)))
			}
			for(fr <- f.for_statements)
				emitFrame(fr)
			
			end("}")
		}
		
		def isKernelTrivial(k : Kernel) : Boolean = {
			val trivial_write_elem = k.elem
			def isAccessTrivial(a : Access) = {
				(a.op == Ops.RNone) || (a.elem == trivial_write_elem)
			}
			def isFrameTrivial(f : Frame) : Boolean = {
				f.for_statements.forall(isFrameTrivial) &&
				f.field_access.forall(isAccessTrivial)
			}
			isFrameTrivial(k)
		}

		//TODO: generate frames for each kernel / the top-level kernel, stencil code gen code
		for(ke <- kernel_table.values) {
			val context = new scala.collection.mutable.HashMap[String,Label]
			val elem = fresh()
			val frame = new Kernel(elem, ke)
			val return_values = new ReturnValues
			context += (ke.loop_var.name -> elem)
			interpStmt(context, frame, return_values,ke.body, List.empty)
			
			emit("L_STENCIL void kernel_%d_stencil(lsFunctionTable * __vtbl, lsContext * __ctx)".format(ke.name))
			emitFrame(frame)
			
			
			val isTrivial = isKernelTrivial(frame)
			if (ke.mustBeTrivial && !isTrivial) {
				Log.error("reader/writer conflict found in block: %s", ke.body.toString)
			}
			
			emit(tKernelStencilHeader(ke))
			begin("{")
			emit("lStencilData data = { kernel_%d_stencil, %s };".format(ke.name,isTrivial))
			emit("return data;")
			end("}")
		}
		//generate the total program stencil
		
		val fe = function_table("_liszt_init_globals")
		val context = new scala.collection.mutable.HashMap[String,Label]
		val frame = new Top
		val return_values = new ReturnValues
		interpStmt(context,frame,return_values,fe.tree.body, List.empty)
		emit(tFunctionStencilHeader(fe))
		emitFrame(frame)
		
	}
}