package Liszt
import Liszt._
import Liszt.Types._
import Liszt.Trees._
import scala.collection.immutable.{Set => SET}
import scala.collection.immutable.BitSet


//this class runs one iteration of the lazy code motion algorithm and inserts
//new DefVars for temporaries
//this.modified = true if any temporaries were added
//this.new_temporaries lists the names of the new temporaries
//note that this may place DefVar where they are out of scope to some uses
//the "RunIteratedLazyCodeMotion" object below will iterate the LazyCodeMotion
//pass and then clean up these DefVar nodes

class LazyCodeMotion(df : DefFunc) {

	object Log extends Liszt.Log("LazyCodeMotion")
	
	abstract class CExp {
		def tree : Tree
		def uses : List[String]
		def hashCode : Int
		def equals(a : Any) : Boolean
		//if we need to define a temporary for this CExp what does it look like?
		def defineStmt(name : String) : List[Stmt]
	}
	abstract class CExpSimple extends CExp {
		def copyTree() : Exp
		override def defineStmt(name : String) : List[Stmt] = List(DefVar(name,copyTree))
	}
	//to add a new CExp you must subclass CExp and fill in the missing functions
	//then you must add a new pattern match to the CExp to match and construct the CExp
	
	case class CTopoCall(tree : TopoCall) extends CExpSimple {
		override def uses : List[String] = for(v <- tree.args) yield v.name
		override def equals(rhs : Any) = rhs match {
			case CTopoCall(rhs) =>
				(tree.t == rhs.t) && 
				tree.func == rhs.func && 
				{
					(tree.args zip rhs.args).forall { 
						case (x,y) => x.name == y.name 
					}
				}
			case _ => false
		}
		override def hashCode = 
			tree.t.hashCode + 
			tree.func.hashCode + 
			( tree.args.foldRight(0) { (x,y) => x.name.hashCode + y } )
		override def copyTree() = tree.copy()
	}
	case class CFieldRead(tree : FieldRead) extends CExpSimple {
		override def uses : List[String] = List(tree.field.name,tree.arg.name)
		override def equals(rhs : Any) = rhs match {
			case CFieldRead(rhs) =>
				tree.t == rhs.t &&
				tree.field.name == rhs.field.name &&
				tree.arg.name == rhs.arg.name
			case _ => false
		}
		override def hashCode = 
				tree.t.hashCode +
				tree.field.name.hashCode +
				tree.arg.name.hashCode
		override def copyTree() = tree.copy()
	}
	case class CBinary(tree : Binary) extends CExpSimple {
		override def uses : List[String] = List(tree.l.name,tree.r.name)
		override def equals(rhs : Any) = rhs match {
			case cb @ CBinary(rhs) =>
				this.uses == cb.uses &&
				tree.op == rhs.op &&
				tree.t == rhs.t
			case _ => false
		}
		override def hashCode = 
			uses.hashCode + 
			tree.t.hashCode +
			tree.op.hashCode
		override def copyTree() = tree.copy()
	}
	object CExp {
		def unapply(exp : Exp) : Option[CExp] = exp match {
			case tree : TopoCall => Some(CTopoCall(tree))
			case tree : FieldRead => Some(CFieldRead(tree))
			case tree : Binary => Some(CBinary(tree))
			case _ => None
		}
	}
	
	
	
	type Node = cfg.Node
	
	val cfg = new CFGs.CFG2(df)
	
	def computed_expressions(node : Node) : List[CExp] = {
		cfg.treeForNode(node) match {
			case Some(x) => x match {
				case DefVar(_,CExp(exp)) => List(exp)
				case x => List()
			}
			case None => List()
		}
	}
	
	def kills(node : Node, exp : CExp) : Boolean = {
		val uses = exp.uses
		cfg.treeForNode(node) match {
			case Some(x) => x.writes.exists { y => uses.contains(y) }
			case None => false
		}
	}
	
	val (cexp_to_id : Map[CExp,Int],id_to_cexp : Array[CExp]) = {
		val map = new scala.collection.mutable.HashMap[CExp,Int]()
		var all_cexps : List[CExp] = List.empty
		var next_id = 0
		for(node <- cfg.nodes; cexp <- computed_expressions(node)) {
			map.getOrElseUpdate(cexp, {
				all_cexps ::= cexp
				next_id += 1
				next_id - 1
			})
		}
		(Map.empty ++ map, all_cexps.reverse.toArray)
	}
	
	val e_use : Map[Node, BitSet] = Map.empty ++ {
		for(node <- cfg.nodes) yield {
			val bs = computed_expressions(node).foldRight(BitSet.empty) {
				(cexp,y) => y + cexp_to_id(cexp)
			}
			node -> bs
		}
	}
	val e_kill : Map[Node,BitSet] = Map.empty ++ {
		
		def kills_for_node(node : Node) : List[Int] = {
			for((k,v) <- cexp_to_id if kills(node,k)) 
				yield v
		}.toList
		Map.empty ++ {
			for(node <- cfg.nodes) yield 
				{ node -> (BitSet.empty ++ kills_for_node(node)) }
		}
	}
	Log.log("cexp_to_id %s".format(cexp_to_id))
	Log.log("e_use %s".format(e_use))
	Log.log("e_kill %s".format(e_kill))
	
	val empty_set = BitSet.empty
	val universal_set = BitSet.empty ++ (0 until id_to_cexp.size)
	
	abstract class BitSetPass extends CFGs.OptimizationPass2 {
		type V = BitSet
		override def cfg = LazyCodeMotion.this.cfg
		override def equal(a : V, b : V) = a == b
		override def valueToString( v : V) : String = "\n%s = (%s)".format(this.name,{ for(i <- v) yield id_to_cexp(i).tree.toString }.mkString(","))
	}
	//Dragon Book pg. 649:
	val anticipated = new BitSetPass {
		override def isForward = false
		override def nodeTransferFn(n : Node)(x : V) = e_use(n) | (x &~ e_kill(n))
		override def inputValue = empty_set
		override def meet(a : V, b : V) = a & b
		override def top = universal_set
		override def name = "anticipated"
	}.result
	
	val available = new BitSetPass {
		override def isForward = true
		override def nodeTransferFn(n : Node)(x : V) = (anticipated(n).in | x) &~ e_kill(n)
		override def inputValue = empty_set
		override def meet(a : V, b : V) = a & b
		override def top = universal_set
		override def name = "available"
	}.result
	
	val earliest : Map[Node,BitSet] = Map.empty ++ {
		for(node <- cfg.nodes) 
			yield node -> (anticipated(node).in &~ available(node).in)
	}
	val postponable = new BitSetPass {
		override def isForward = true
		override def nodeTransferFn(n : Node)(x : V) = (earliest(n) | x) &~ e_use(n)
		override def inputValue = empty_set
		override def meet(a : V, b : V) = a & b
		override def top = universal_set
		override def name = "postponable"
	}.result
	
	val latest : Map[Node,BitSet] = Map.empty ++ {
		for(node <- cfg.nodes)
			yield {
			val a = (earliest(node) | postponable(node).in)
			val b = cfg.successors(node).foldRight(universal_set) {
				(s,v) => v & (earliest(s) | postponable(s).in)
			}
			node -> (a & (e_use(node) | b))
		}
	}
	
	val used = new BitSetPass {
		override def isForward = false
		override def nodeTransferFn(n : Node)(x : V) = (e_use(n) | x) &~ latest(n)
		override def inputValue = empty_set
		override def meet(a : V, b : V) = a | b
		override def top = empty_set
		override def name = "used"
	}.result
	
	val cexp_to_temporary : Array[String] = for( id <- id_to_cexp) yield ASTUtil.Names.create("lcm")
	
	
	var new_temporaries : SET[String] = SET.empty
	def modified = !new_temporaries.isEmpty
	def modify_tree(node : Node) : List[Stmt] = {
		val defines = latest(node) & used(node).out
		val prefix = for(d <- defines.toList; stmt <- id_to_cexp(d).defineStmt(cexp_to_temporary(d)))
			yield stmt
		new_temporaries ++= defines.map(cexp_to_temporary)
		def replace_with_var(id : Int) : Boolean = {
			val r = e_use(node) & ( (universal_set &~ latest(node)) | used(node).out )
			r.contains(id)
		}
		def replace_def(dv : DefVar, cexp : CExp) : Stmt = {
			val cexp_id = cexp_to_id(cexp)
			if(replace_with_var(cexp_id)) {
				DefVar(dv.name,Var(dv.init.t,cexp_to_temporary(cexp_id)))
			 } else dv
		}
		val replaced : List[Stmt] = for(n <- cfg.nodeIdentity(node)) yield n match {
			case dv @ DefVar(name, CExp(exp)) => replace_def(dv,exp)
			case x => x
		}
		prefix ::: replaced
	}
	val result : DefFunc = cfg.toTree(modify_tree)
	
}

//apply the above until it doesn't change or we reach max_iterations
class IteratedLazyCodeMotion(df : DefFunc, max_iterations : Option[Int]) {
	var new_temporaries : SET[String] = SET.empty
	val result : DefFunc = {
		var changed = true
		var def_func = df
		var iter = 0
		while(changed && (max_iterations.isEmpty || iter < max_iterations.get)) {
			val lzm = new LazyCodeMotion(def_func)
			def_func = remove_redundant_moves(lzm.result)
			new_temporaries ++= lzm.new_temporaries
			changed = lzm.modified
			iter += 1
		}
		def_func
	}
	def remove_redundant_moves(df : DefFunc) : DefFunc = {
		//if you write val a = b
		//then use a without any paths modifying a or b, we replace it with b
		//if a becomes dead, we remove a
		val cfg_ = new CFGs.CFG2(df)
		val must_reach_definitions = new CFGs.OptimizationPass2 {
			import scala.collection.immutable.BitSet
			type V = BitSet
			override def cfg = cfg_
			
			val id_to_def = {
				def find_defines(tree : Tree) : List[DefVar] = tree match {
					case df @ DefVar(_,_ : Var) => df :: (df.children flatMap find_defines)
					case t => t.children flatMap find_defines 
				}
				find_defines(df).toArray
			}
			override def valueToString(v : V) : String = { for(i <- v) yield id_to_def(i).toShortString}.mkString("[",",","]")
			val def_to_id = Map.empty ++ id_to_def.zipWithIndex
			val name_to_id = Map.empty ++ { for( (dv,i) <- def_to_id) yield dv.name -> i }
			
			val all_define_ids = BitSet.empty ++ (0 until id_to_def.length)
			def node_kills_define(node : Node, define_id: Int) = {
				val DefVar(name, rhs : Var) = id_to_def(define_id)
				cfg.treeForNode(node) match {
					case Some(x) => x.writes.contains(name) || x.writes.contains(rhs.name)
					case None => false
				}
			}
			val kills : Map[Node,V] = Map.empty ++ {
				for(node <- cfg.nodes) yield {
					val ks = all_define_ids.filter { x => node_kills_define(node,x) }
					node -> ks
				}
			}
			val defines : Map[Node,V] = Map.empty ++ {
				for(node <- cfg.nodes) yield node -> {
					cfg.treeForNode(node) match {
						case Some(dv : DefVar) => BitSet.empty ++ def_to_id.get(dv).toList
						case _ => BitSet.empty
					}
				}
			}
			
			override def isForward = true
			override def nodeTransferFn(n : Node)(x : V) = defines(n) | (x &~ kills(n))
			override def inputValue = BitSet.empty
			override def meet(a : V, b : V) = a & b
			override def top = all_define_ids
			override def name = "must_reach_definitions"
			override def equal(a : V, b : V) = a == b
		}
		def rename(nmap : String => String)(tree : Tree) : Tree = tree match {
			case v : Var if !v.isWrite => ShapeAny.names(tree,v.names map nmap)
			case _ => ShapeAny.trees(tree, tree.children map rename(nmap))
		}
		
		//create a map of possible replacement DefVar stmts.  The reason we dont use
		//toTree here is because we need to modify the variables inside a Forall set
		//which is not represented in the cfg
		type Node = cfg_.Node
		def renameTree(tree : Tree, node : Node) = {
			val renamed = rename({
				name => must_reach_definitions.name_to_id.get(name) match {
					case Some(id) 
						if must_reach_definitions.result(node).in.contains(id) =>
							must_reach_definitions.id_to_def(id).init.asInstanceOf[Var].name
					case _ => name
				}
			})(tree)
			renamed
		}
		def node_to_stmts(node : Node) : List[Stmt] = {
			for(tree <- cfg_.nodeIdentity(node))
				yield  renameTree(tree,node).asInstanceOf[Stmt]
		}
		def artificial(t : Tree, n : Node) = {
			val new_t = renameTree(t,n)
			//println("artificial: %s %s -> %s".format(t,n,new_t))
			new_t
		}
		val uses_removed = cfg_.toTreeWithArtificial(node_to_stmts,artificial)
		
		var used_defs = BitSet.empty
		def find_uses(t : Tree) { t match {
			case v : Var => used_defs ++= must_reach_definitions.name_to_id.get(v.name)
			case _ => t.children map find_uses
		} }
		find_uses(uses_removed)
		
		val defines_removed = ASTUtil.flatMapBlocks(uses_removed) {
			case dv : DefVar => must_reach_definitions.name_to_id.get(dv.name) match {
				case Some(id) if !used_defs.contains(id) => List()
				case _ => List(dv)
			}
			case x => List(x)
		}
		defines_removed.asInstanceOf[DefFunc]
	}
}

object RunLazyCodeMotion {
	def apply(blk : Block, max_iterations : Option[Int]) : Block = {
		def fixup_defines(df : DefFunc, temporaries : SET[String]) : DefFunc = {
			val cfg = new CFGs.CFG2(df)
			type Node = cfg.Node
			def lowest_common_dominator(temp : String) : Node = {
				def node_defines_temp(n : Node) : Boolean = cfg.treeForNode(n) match {
					case Some(DefVar(name,_)) => temp == name
					case _ => false
				}
				val defining_nodes = cfg.nodes.toList.filter(node_defines_temp)
				cfg.lowestCommonDominator(defining_nodes)
			}
			val temporary_types = Map.empty++ {
				for(node <- cfg.nodes; DefVar(name,init) <- cfg.treeForNode(node) if temporaries.contains(name))
					yield name -> init.t
			}
			val temporary_to_lowest_common_dominator = Map.empty ++ {
				for(temp <- temporaries) yield {
					temp -> lowest_common_dominator(temp)
				}
			}
			cfg.toTree { 
				node => 
				val new_decls = { for(temp <- temporaries if temporary_to_lowest_common_dominator(temp) == node)
					yield DefVar(temp,EmptyExp(temporary_types(temp))) }.toList
				cfg.treeForNode(node) match {
					case Some(dv @ DefVar(name,init)) if temporaries.contains(name) =>
						if(temporary_to_lowest_common_dominator(name) == node) {
							new_decls.filter(_.name != name) :::
							List(dv)
						} else {
							val nm = ASTUtil.Names.create("fixup")
							new_decls ::: 
							List( DefVar(nm,init),
							      Assign(Var(init.t,name),Ops.Assign,Var(init.t,nm)) )
						}
					case _ => 
						new_decls ::: cfg.nodeIdentity(node)
				}
				
			}
		}
		
		//variables referring to literals are frequently duplicated in the AST:
		//val temp0 = 1
		//val temp1 = 1
		//val temp2 = a + temp0
		//val temp3 = a + temp1
		//though a + temp0 is the same as a + temp1, since 
		//we use variable name equality to determine if 
		//the statement is the same, we would miss this in our analysis
		//this pass finds the trivially constant variables, de-duplicates them, and pre-defines
		//them at the beginning of the function
		def unique_constants(df : DefFunc) : DefFunc = {
			def find_all_literals(t : Tree) : List[(String,Literal)] = t match {
				case DefVar(name,l : Literal) => List(name -> l)
				case x => x.children flatMap find_all_literals
			}
			def find_all_writes(t : Tree) : SET[String] = t match {
				case DefVar(name,l : Literal) => SET.empty //we don't count the original define as a write
				case _ => SET.empty ++ (t.children flatMap find_all_writes) ++ t.writes
			}
			val writes = find_all_writes(df)
			//literals are all writes that do not contain x
			val all_literals = Map.empty ++ { find_all_literals(df).filter { case (name,_) => !writes.contains(name) } }
			
			//println(writes)
			//println(all_literals)
			
			//group the literals by the literal's value
			val grouped_literals = all_literals.groupBy{ case (k,v) => v.c }
			val unique_literals = for( (l,_) <- grouped_literals)
				yield ASTUtil.Names.create("literal")
			val grouped_literals_unique_literals = grouped_literals zip unique_literals
			val replacements = for( ((l,uses),new_name) <- grouped_literals_unique_literals;
			                           (name,_) <- uses.toList ) yield {
				name -> new_name
			}
			val replacement_map : Map[String,String] = (Map.empty ++ replacements).withDefault(x => x)
			def rename(t : Tree) : Tree = ShapeAny(t,t.t,t.names map replacement_map, t.children map rename)
			
			val new_defines = for( ((_,v),name) <- grouped_literals_unique_literals) 
				yield DefVar(name,v.values.head)
			
			val defs_removed = ASTUtil.flatMapBlocks(df.body) {
				case DefVar(name,l : Literal) if all_literals.contains(name) => List()
				case x => List(x)
			}
			val new_stmts = ASTUtil.asStmtList(defs_removed.asInstanceOf[Stmt]).map { x => rename(x).asInstanceOf[Stmt] }
			val new_block = Block(new_defines.toList ::: new_stmts,EmptyExp(UnitT))
			df.copy(body = new_block)
		}
		val nstmts = for(stmt <- blk.stmts) yield stmt match {
			case df : DefFunc if df.body.getClass != ExternStmt.getClass => {
				val constants_fixed = if(Config.getOrElse("unique_constants",true).asInstanceOf[Boolean]) 
					unique_constants(df)
				else df
				val ilcm = new IteratedLazyCodeMotion(constants_fixed,max_iterations)
				val result = ilcm.result
				val defines_fixed = fixup_defines(ilcm.result,ilcm.new_temporaries)
				defines_fixed
			}
			case x => x
		}
		Block(nstmts,blk.exp)
	}
}