package Liszt
import Liszt._
import Liszt.Types._
import Liszt.Trees._
import scala.collection.immutable.{Set => SSet}



class MPIStencil(graph : Topology.Graph, field_info : Map[Topology.Node, List[(Boolean,FieldInit)]], file_name : String) extends CodeLines {
	abstract class Tree {
		def t : Type
		def ref : String
		def depth : Int
	}
	
	class Block extends Tree {
		def t = UnitT
		def ref = error("cannot refer to a block")
		var uses : List[Tree] = List()
		var fields : SSet[FieldAccess] = SSet()
		def depth = 0
	}
	
	case class Forall(depth : Int, t : Type, ref : String, set : Tree, block : Block) extends Tree
	
	case class Request(depth : Int, t : Topo, ref : String) extends Tree
	case class AllElements(depth : Int, t : Set, ref : String) extends Tree
	case class Boundary(depth : Int, t : Set, ref : String, boundary_name : String) extends Tree
	case class Func(depth : Int, ref : String, call : Ops.TopoFn, args : List[Tree]) extends Tree {
		def t = call.rett
	}
	
	case class FieldAccess(id : Int, isRead : Boolean, entry : Tree)
	
	var set_declarations : List[String] = List()
	var obj_declarations : List[String] = List()
	var top_declarations : List[String] = List()
	var boundary_declarations : List[String] = List()
	
	def set_name() = {
		val nm = ASTUtil.Names.create("s")
		set_declarations ::= nm
		nm
	}
	def obj_name() = {
		val nm = ASTUtil.Names.create("s")
		obj_declarations ::= nm
		nm
	}
	def top_name() = {
		val nm = ASTUtil.Names.create("t")
		top_declarations ::= nm
		nm
	}
	def boundary_name() = {
		val nm = ASTUtil.Names.create("b")
		boundary_declarations ::= nm
		nm
	}
	
	val treeForNode : Topology.Node => Tree = Util.Memoize[Topology.Node,Tree] {
		case Topology.TopLevel => new Block
		case node : Topology.Node => {
			val (ctx_depth,ctx) = blockFor(node)
			val tree = node match {
				case e : Topology.Each => {
					val set = treeForNode(graph.defFor(e,Topology.Base))
					val depth = (set.depth max ctx_depth)
					Forall(depth,set.t.asInstanceOf[Set].t,obj_name(),set,new Block)
				}
				case Topology.TopoUse(Ops.TopoSet(t,_,Types.Mesh)) => AllElements(ctx_depth + 1,Set(t),top_name())
				case Topology.TopoUse(op) => {
					val args = treeForNode(graph.defFor(node,Topology.Base)) :: { 
						op match {
							case _ : Ops.Towards => List(treeForNode(graph.defFor(node,Topology.ReferenceElem)))
							case _ => List()
						}
					}
					val depths = args map { _.depth }
					val depth = { ctx_depth max { depths reduceLeft (_ max _) }} + 1
					
					for(arg <- args) {
						ctx.uses ::= Request(depth, arg.t.asInstanceOf[Topo],arg.ref)
					}
					val n = op match {
						case _ : Ops.TopoSet => set_name()
						case _ => obj_name()
					}
					Func(depth,n,op,args)
				}
				case Topology.Boundary(t,_,setname) => Boundary(ctx_depth + 1,t.asInstanceOf[Set],boundary_name(),setname)
			}
			ctx.uses ::= tree
			tree
		}
	}.apply
	def blockFor(node : Topology.Node) = treeForNode(graph.defFor(node,Topology.Context)) match {
		case b : Block => (0,b)
		case Forall(d,_,_,_,b) => (d,b)
	}
	
	//touch all the nodes in the graph to instanciate the full method
	for(n <- graph.allNodes) {
		treeForNode(n)
	}
	
	for( (k,fields) <- field_info) {
		val tree = treeForNode(k)
		val blk = k match {
			case _ : Topology.Each => tree.asInstanceOf[Forall].block
			case _ => 
				val (_,blk) = blockFor(k)
				blk
		}
		for((isRead,fi) <- fields) {
			blk.fields += FieldAccess(fi.id,isRead,tree)
		}
	}
	
	val tree = treeForNode(Topology.TopLevel)
	
	var max_depth = 0
	
	def gen(t : Tree) : List[Line] = {
		max_depth = max_depth max t.depth
		def typeToValue(t : Type) = t match {
			case Vertex => "VERTEX"
			case Edge => "EDGE"
			case Face => "FACE"
			case Cell => "CELL"
		}
		def typeToSet(t : Type) = t match {
			case Vertex => "vertices_m"
			case Edge => "edges_m"
			case Face => "faces_m"
			case Cell => "cells_m"
		}
		
		t match {
			case blk : Block => {
				val (stmts,foralls) = blk.uses.reverse.partition(x => !x.isInstanceOf[Forall])
				indent {
					stmts.flatMap(gen) ::: 
					foralls.flatMap(gen) :::
					"STENCIL_BEGIN_FIELDS".l :::
					indent { line {
						for(FieldAccess(id,isRead,tree) <- blk.fields.toList)
							yield "requestGhostIfAbsent(%s,%d,%s,elementID(%s,%s));".format(if(isRead) "true" else "false",id,typeToValue(tree.t),typeToValue(tree.t),tree.ref)
					} } :::
					"STENCIL_END_FIELDS".l 
				}
			}
			case Request(d,t,s) => "STENCIL_REQUEST_DEPTH(%d) requestTopologyIfAbsent(%s,elementID(%s,%s));".format(d,typeToValue(t),typeToValue(t),s).l
			case Forall(d,t,ref,set,block) => {
				val set_name = set match {
					case _ : AllElements => "*" + set.ref
					case _ => set.ref
				}
				"STENCIL_DEPTH(%d) STENCIL_BEGIN_FORALL(%s,%s)".format(d,ref,set_name).l :::
				indent {
					gen(block)
				} :::
				"STENCIL_END_FORALL".l
			}
			case AllElements(d,st,ref) => "STENCIL_DEPTH(%d) %s = &%s();".format(d,ref,typeToSet(st.t)).l
			case Boundary(d,t,ref,boundary_name) => {
				"STENCIL_DEPTH(%d) boundary(%s,\"%s\",&%s);".format(d,typeToValue(t.t),boundary_name,ref).l
			}
			case Func(d,ref,call,args) => {
				val args_ = for(a <- args) yield "elementTopo(%s,%s)".format(typeToValue(a.t),a.ref)
				val arg_str = args_.mkString(",")
				call match {
					case _ : Ops.TopoSet => "STENCIL_DEPTH(%d) %s(%s,&%s);".format(d,call.asFunction,arg_str,ref).l
					case _ => "STENCIL_DEPTH(%d) %s = %s(%s);".format(d,ref,call.asFunction,arg_str).l
				}
				
			}
		}
	}
	val template = """
#include "ParallelLoad/Stencil.h"
size_t Stencil::generatedMaxDepth() const { return %d; }
void Stencil::generatedDiscoveryAtDepth(bool discover_topology,size_t depth) {
%s
%s
}
"""
	val gen_tree = gen(tree).mkString("\n")
	
	val decl_sets = for(d <- set_declarations) yield "list_t %s;".format(d)
	val decl_objs = for(d <- obj_declarations) yield "HF * %s;".format(d)
	val decl_top = for(d <- top_declarations) yield "Array<id_type> * %s;".format(d)
	val decl_b = for(d <- boundary_declarations) yield "boundary_list_t %s;".format(d)
	
	val result = template.format(max_depth, (decl_sets ::: decl_objs ::: decl_top ::: decl_b).mkString("\n"),gen_tree)
	Util.WriteFile(file_name) {
		out =>
		out println result
	}
}