package Liszt

object Main {
	object Log extends Liszt.Log("Main")
	object Progress extends Liszt.Log("Progress")
	private def apply_specializations(varMap : Map[String,Any], blk : Liszt.Trees.Block) : Liszt.Trees.Block = {
		//given a map of variable specializations listed in the config,
		//replace those variables with constants
		val specializedCode = if(varMap.isEmpty) blk else {
			import Liszt.{Trees => T}
			import Liszt.Types._
			def getValue(t : Type, name : String, a : Any) : List[T.DefVar] = {
				(t match {
					case IntT => T.DefVar(name,T.IntL(a.asInstanceOf[Double].toInt)) :: List()
					case FloatT => T.DefVar(name,T.FloatL(a.asInstanceOf[Double].toFloat)) :: List()
					case DoubleT => T.DefVar(name,T.DoubleL(a.asInstanceOf[Double])) :: List()
					case BoolT => T.DefVar(name,T.BoolL(a.asInstanceOf[Boolean])) :: List()
					case StringT => T.DefVar(name,T.StringL(a.asInstanceOf[String])) :: List()
					case VectorT(MetaIntT(i),typ) => {
						val lst = a.asInstanceOf[List[Any]]
						if(lst.size == i) {
							val defs = lst map {
								i => getValue(typ,ASTUtil.Names.create("specialized"),i)
							}
							val vars = for(d <- defs) 
								yield T.Var(typ,d.last.name)
							val flatDefs = for(d <- defs; i <- d) yield i
							flatDefs ::: List(T.DefVar(name,T.VectorL(t,vars)))
						} else Log.error("constant vector %s has incorrect size %d expected %d",name,lst.size,i)
					}
				}) map (_ withSynthetic true) //extra defvars are synthetic
			}
			val regex = ASTUtil.moduleMatch
			val newStmts = 
			blk.stmts.flatMap {
				stmt => stmt match {
					case T.DefVar(name @ regex(module,v),init) => 
						varMap.get("%s.%s".format(module,v)) match {
							case Some(c) => {
								val stmts = getValue(init.t,name,c)
								stmts.last withSynthetic false //last stmt is the orignal and is not a generated var
								stmts
							}
							case None => List(stmt)
						}
					case _ => List(stmt)
				}
			}
			T.Block(newStmts,blk.exp) copyAttr blk
		}
		specializedCode
	}
	
	//remove all module calls that are really linear system calls and replace them with correct nodes
	private def resolve_linear_system_nodes(stmts : List[Liszt.Trees.Stmt]) : (List[Liszt.Trees.Stmt],List[Liszt.Trees.DefLinearSystem]) = {
		import Liszt.Trees._
		def filter(linear_system : DefLinearSystem, call : Call) : Exp = {
			def name(str : String) : String = ASTUtil.moduleName(linear_system.name,str)
			val allocate_A = name("A")
			val allocate_x = name("x")
			val allocate_b = name("b")
			
			val index_fns = linear_system.indices
			val fn_name = call.func.name
			val fn_stripped_name = ASTUtil.stripModuleName(call.func.name)
			
			if(fn_name == allocate_A) {
				LinearSystemAllocate(call.t, linear_system.name, LS_A)
			} else if(fn_name == allocate_x) {
				LinearSystemAllocate(call.t, linear_system.name, LS_X)
			} else if(fn_name == allocate_b) {
				LinearSystemAllocate(call.t, linear_system.name, LS_B)
			} else index_fns.find { x => name(x.name) == fn_name } match {
				case Some(idx) => IndexCall(call.t, linear_system.name, idx, call.args)
				case None => 
					if (fn_name.startsWith(linear_system.name) && call.isExternal) {
						Main.Log.log("Found linear system external call " + fn_stripped_name)
						LinearSystemCall(call.t, linear_system.name, Var(call.t, fn_stripped_name), call.args)
					} else {
						call
					}
			}
			
		}
		def translateTree(ls : List[DefLinearSystem],t : Tree) : Tree = ls match {
			case head :: rest => t match {
				case c : Call => translateTree(rest,filter(head,c))
				case ShapeAny(_,_,trees) => ShapeAny.trees(t, trees map { x => translateTree(ls,x) })
			}
			case _ => t
		}
		val linear_systems = for( ls @ DefLinearSystem(_,_,_) <- stmts)
			yield ls
		val filtered_stmts = for(stmt <- stmts if !stmt.isInstanceOf[DefLinearSystem])
			yield translateTree(linear_systems,stmt).asInstanceOf[Stmt]
		(filtered_stmts,linear_systems)
	}
	private def link(module_name : String) : CompilationUnit = {
		
		//resolve module dependencies, post-order traversal ensures that in any DAG of dependencies,
		//the code that a module requires is initialized before the module itself
		var dependencies : List[Module] = List.empty 
		var loaded_deps : Set[String] = Set.empty
		def load(name : String) {
			if(!loaded_deps.contains(name)) {
				val module = new Module(name)
				loaded_deps += name
				for(dep <- module.dependencies) {
					load(dep)
				}
				dependencies ::= module
			}
		}
		load(module_name)
		Log.log("Module Initialization order is: %s",for(m <- dependencies.reverse) yield m.name)
		
		val stmts_pre_linear_system = for(dep <- dependencies.reverse; stmt <- dep.implementation) 
			yield stmt
		
		val (stmts,linear_systems) = resolve_linear_system_nodes(stmts_pre_linear_system)
		
		//generate the entry point for the liszt program which calls the main module's main method
		val def_main = {
			import Liszt.Trees._
			import Liszt.Types._
			val main_fn_name = ASTUtil.moduleName(module_name,"main")
			//def _liszt_start() { module_name___main(); }
			val unit_fn = Func(List(),UnitT) 
			DefFunc(unit_fn,"_liszt_start",List(),ExpStmt(Call(UnitT,Var(unit_fn,main_fn_name),List())))
		}
		
		new CompilationUnit(Liszt.Trees.Block(stmts ::: List(def_main),Liszt.Trees.EmptyExp(Liszt.Types.UnitT)), linear_systems)
	}
	
	private def findScalarReductionVariables(blk : Liszt.Trees.Block) : Set[String] = {
		import Liszt.Trees._
		def unnestedDefines(t : Tree) : Set[String] = {
			t match {
				case Forall(name,lbl,set,block) => Set.empty
				case DefVar(name,init) => unnestedDefines(init) + name
				case ShapeAny(_,_,ts) => Set.empty ++ {for(t <- ts; v <- unnestedDefines(t)) yield v }
			}
		}
		def nestedAssigns(t : Tree, nested : Boolean) : Set[String] = {
			t match {
				case Forall(name,lbl,set,block) => nestedAssigns(block,true)
				case v @ Var(_,name) if v.isWrite => if(nested) Set(name) else Set.empty
				case _ => Set.empty ++ { for(t <- t.children; v <- nestedAssigns(t,nested)) yield v }
			}
		}
		val defs = unnestedDefines(blk)
		val assigns = nestedAssigns(blk,false)
		defs intersect assigns
	}
	
	def log_temporary(file_name : String, tree : Trees.Tree) {
		if(LogSettings.enabled("temporaries")) {
			Util.WriteFile(file_name) {
				out => out.println(tree)
			}
		}
	}
	
	
	def removeFieldVariables(blk : Liszt.Trees.Block) : Liszt.Trees.Block = {
		val Lg= Log
		import Liszt.Trees._
		import Liszt.Types._
		
		def isField(t : Type) = t.isInstanceOf[FieldT]
		
		def hasFieldArguments(stmt : Stmt) = stmt match {
			case df : DefFunc => df.typ.args.exists(isField)
			case _ => false 
		}
		val (fieldfunctions,other) = blk.stmts.partition(hasFieldArguments)
		
		class FieldFunction(val original : DefFunc) {
			val specializations = new scala.collection.mutable.HashMap[List[String],Var]
		}
		
		val fieldFunctionToDefinition = new scala.collection.mutable.HashMap[String,FieldFunction]
		
		for (ff_ <- fieldfunctions) {
			val ff = ff_.asInstanceOf[DefFunc]
			fieldFunctionToDefinition(ff.name) = new FieldFunction(ff)
		}
		
		//map from field variable name to its global field variable
		val varToField = new scala.collection.mutable.HashMap[String,String]
		
		for(DefVar(x,fi : FieldInit) <- blk.stmts) {
			varToField(x) = x
		}
		
		//list of statements left to process, contains both the original functions (not including those with field arguments)
		//and any specializations of functions to remove the field arguments
		val toProcess = new scala.collection.mutable.Queue[Stmt]
		
		toProcess ++= other
		
		def getFieldArgs(vs : List[Var]) : List[String] = vs.filter( (x : Var) =>isField(x.t) ).map(_.name)
		def filterArgs(vs : List[Var]) = vs.filter( (x : Var) => !isField(x.t) )
		
		def removeFieldTypes(fntyp : Func) = { 
			if(isField(fntyp.range))
				Lg.error("functions cannot return fields")
			Func(fntyp.args.filter(!isField(_)),fntyp.range)
		}
		
		def removeFieldDefs(typs : List[Type], dvs : List[DefVar]) = for( (t,d) <- typs zip dvs if !isField(t)) 
			yield d
		
		def addFieldMappings(args : List[Var],params : List[DefVar]) = {
			for((a,p) <- args zip params) {
				if(isField(a.t))
					varToField(p.name) = varToField(a.name)
			}
		}
		
		def specialize(c : Call) : Call = {
			fieldFunctionToDefinition.get(c.func.name) match {
				case None => c
				case Some(ff) =>
					val strs = getFieldArgs(c.args)
					ff.specializations.get(strs) match {
						case Some(v) => Call(c.t,v,filterArgs(c.args))
						case None =>
							//make a new specialization
							val newcopy = ASTUtil.duplicate(ff.original)
							addFieldMappings(c.args,newcopy.args)
							ff.specializations(strs) = Var(newcopy.typ,newcopy.name)
							toProcess += newcopy
							Call(c.t,Var(newcopy.typ,newcopy.name),filterArgs(c.args))
					}
			}
		}
		def resolveFieldNames(t : Tree) : Tree = {
			t match {
				case df : DefFunc => DefFunc(removeFieldTypes(df.typ),df.name,removeFieldDefs(df.typ.args,df.args),resolveFieldNames(df.body).asInstanceOf[Stmt])
				case DefVar(a,init) if isField(init.t) =>
					init match {
						case _ : FieldInit => //preprocessed above
						case v : Var =>
							varToField(a) = varToField(v.name)
						case _ =>
							Lg.error("unsupported definition of field:\n%s",init.toString)
					}
					t
				case v : Var if isField(v.t) =>
					Var(v.t,varToField(v.name))
				case c : Call => specialize(c)
				case x => ShapeAny.trees(x,x.children map resolveFieldNames)
			}
		}
		
		var newstmts = new scala.collection.mutable.ListBuffer[Stmt]
		
		while(!toProcess.isEmpty) {
			val resolved = resolveFieldNames(toProcess.dequeue).asInstanceOf[Stmt]
			//remove any remaining definitions of fields
			val newstmt = ASTUtil.flatMapBlocks(resolved) {
				(stmt : Stmt) => stmt match {
					case DefVar(a,_ : FieldInit) => List(stmt)
					case DefVar(a,init) if isField(init.t) => List.empty
					case x => List(x)
				}
			}
			newstmts += newstmt.asInstanceOf[Stmt]
		}
		
		Block(newstmts.toList,EmptyExp(UnitT))
	}
	
	def main(args : Array[String]) {
		
		val config_file :: _ = args.toList	//assigns the first two elements of args to config_file and build_dir
		Config.loadFromFile(config_file)
		Log.log("Configuration: %s",Config)
		val build_dir = Config("build-dir").asInstanceOf[String]
		val module_main = Config("main-class").asInstanceOf[String]
		
		val specializations = Config.getOrElse("constants",Map.empty[String,Any]).asInstanceOf[Map[String,Any]]
		Progress.log("linking")
		val linked_code = link(module_main)
		Progress.log("...done")
		log_temporary("linked.code",linked_code.code)
		
		linked_code.code = ASTUtil.addFlopCounter(linked_code.code).asInstanceOf[Liszt.Trees.Block]
		
		
		linked_code.code = ASTUtil.uniqueVariables(apply_specializations(specializations,linked_code.code))
		
		val topologicalDefList = ASTUtil.toposortCallGraph(linked_code.code)
		Log.log("Topological Order for Defs is: %s", for(dv <- topologicalDefList) yield dv.name)
		
		
		if(Config.getOrElse("inline_functions",false).asInstanceOf[Boolean]) {
			Progress.log("inlining functions")
			linked_code.code = new InlineFunctions(linked_code.code).result
			log_temporary("inlined.code",linked_code.code)
			Progress.log("...done")
		}
		val n_times = Config.getOrElse("lazy_code_motion",0.0).asInstanceOf[Double].toInt
		if(n_times > 0) {
			
			Progress.log("lazy code motion")
			linked_code.code = RunLazyCodeMotion(linked_code.code,Some(n_times))
			Progress.log("...done")
			
			log_temporary("lazy_code_motion.code",linked_code.code)
		}
		
		linked_code.code = removeFieldVariables(linked_code.code)
		
		//variables which are written from all nodes and need to be handled like a single-element
		//field
		val srVariables : Set[String] = findScalarReductionVariables(linked_code.code)
		
		Log.log("finding scalar reduction variables")
		Log.log("Scalar Reduction Variables: %s",srVariables)
		//tag all global variables and def-vars as globals:
		for(tree <- linked_code.code) {
			import Liszt.Trees._
			tree match {
				case dv @ DefVar(name,init) if srVariables.contains(name) => dv withGlobal true
				case v @ Var(t,name) if srVariables.contains(name) => v withGlobal true
				case _ =>
			}
		}
		
		val update_placement = new UpdatePlacement(linked_code.code,srVariables)
		
		//val update_placement2 = new UpdatePlacement2(linked_code.code,srVariables)
		
		
		Progress.log("update placement")
		linked_code.code = update_placement.withUpdates
		Progress.log("...done")
		log_temporary("updates.code",linked_code.code)
		//Testing new CFG code, purposely keeping both around while we switch to newer
		//implementation. We write both post-analysis ASTs out to files so we can compare
		//them to make sure we didnt mess anything up
		
		//Progress.log("update placement (2)")
		//val up2 = update_placement2.withUpdates
		//Progress.log("...done")
		
		log_temporary("updates2.code",linked_code.code)
		
		Progress.log("code gen")
		log_temporary("pre_codegen.code",linked_code.code)
		
		for(runtime <- Config("runtimes").asInstanceOf[List[String]]) runtime match {
			case "deprecated_mpi" => {
				val cg = new Liszt.CodeGen(linked_code,srVariables,"ScalaCppRuntime.h", "mpi")
				cg.create(build_dir)
			}
			case "deprecated_vis" => {
				val cg = new Liszt.VecCodeGen(linked_code)
				cg.create(build_dir)
			}
			case "deprecated_single" => {
				val cg = new Liszt.CodeGen(linked_code,srVariables,"ScalaSingleCoreRuntime.h", "single")
				cg.create(build_dir)
			}
			case "deprecated_gpu" => {
				val cg = new Liszt.GPUCodeGen(linked_code, srVariables)
				cg.create(build_dir)
			} 
			case "deprecated_smp" => {
				val cg = new Liszt.SMPCodeGen(linked_code,srVariables)
				cg.create(build_dir)
			}
                        case "deprecated_syrah" => {
                          val vectorizer = new Vectorizer(linked_code.code);
                          val vectorized_code = vectorizer.withVectorized;
                          // "Redo" name uniqeuing to hopefully get
                          // rid of the damn ctype shit.
                          val unique_vec = ASTUtil.uniqueVariables(vectorized_code);
                          val cg = new Liszt.SyrahCodeGen(new CompilationUnit(unique_vec, linked_code.linear_systems), srVariables);
                          cg.create(build_dir);
                        }
			case "abstract" => {
				//test-bed for compiler changes, does not actually generate any code
				ASTUtil.wrapForallInWorkgroup(linked_code.code)
				linked_code.code = new WorkgroupCreation(linked_code.code).workgroupified
				AbstractInterpretation.create(linked_code)
			}
			case x => {
				new CommonRuntimeCodeGen(linked_code, build_dir)
			}
		
		}
		Progress.log("...done")
	}
}
