package Liszt.Topology

import Liszt.Trees._
import Liszt.Types._

case class Edge(typ : EdgeType, dfn : Node, use : Node)
	//dfn -> the definition of a node
	//use -> where this defined node is used in a new node

object Node {
	private var next_id = 0
	def ID() = { next_id += 1; next_id - 1 }
}
class Node {
	private val id = Node.ID()
	override def hashCode = id
	override def equals(a : Any) = a match {
		case n : Node => id == n.id
		case _ => false
	}
}

class Each extends Node
case class TopoUse(op : Liszt.Ops.TopoFn) extends Node
object TopLevel extends Node
case class Boundary(t : Type, set : String, set_name : String) extends Node

class EdgeType
case object Context extends EdgeType //forall context of Node, it will dfn is TopLevel when not inside a forall loop
class Argument extends EdgeType
case object Base extends EdgeType //for node: towards(f,c) <- Base's dfn points to f
case object ReferenceElem extends EdgeType //towards(f,c) <- Base's dfn points to c

class Graph {
	def defFor(n : Node, et : EdgeType) : Node = {
		{
			for(Edge(typ,use,dfn) <- edges if (n eq use) && typ == et) 
				yield dfn
		}.head
	}
	def defsFor(n : Node) : List[Node] = {
		for(Edge(_,use,dfn) <- edges if (n eq use))
			yield dfn
	}
	def forall(ctx : Node, x : Node) : Node = {
		val a = for(Edge(Base,n : Each,dfn) <- edges if (dfn eq x) && (defFor(n,Context) eq ctx))
			yield n
		a.headOption.getOrElse {
			val e = new Each
			nodes ::= e
			edges ::= Edge(Base,e,x)
			edges ::= Edge(Context,e,ctx)
			e
		}
	}
	
	def simpleTopo(ctx : Node, op : Liszt.Ops.TopoFn, base : Node) : Node = {
		val tul = for(Edge(Base,tu @ TopoUse(op2),dfn) <- edges 
			if op2 == op && 
			   (base eq dfn) && 
			   (defFor(tu,Context) eq ctx)
		   ) yield tu
		tul.headOption.getOrElse {
			val tu = TopoUse(op)
			nodes ::= tu
			edges ::= Edge(Base,tu,base)
			edges ::= Edge(Context,tu,ctx)
			tu
		}
	}
	
	def orientedTopo(ctx : Node, op : Liszt.Ops.TopoFn, base : Node, orientation : Node) : Node = {
		val tul = for(Edge(Base,tu @ TopoUse(op2),dfn) <- edges
			if op == op2 &&
			   (base eq dfn) &&
			   (defFor(tu,Context) eq ctx) &&
			   (defFor(tu,ReferenceElem) eq orientation)
		   ) yield tu
		tul.headOption.getOrElse {
			val tu = TopoUse(op)
			nodes ::= tu
			edges ::= Edge(Base,tu,base)
			edges ::= Edge(Context,tu,ctx)
			edges ::= Edge(ReferenceElem,tu,orientation)
			tu
		}
	}
	def boundary(ctx : Node, t : Type, bs : String, set_name : String) = {
		val b = for(b @ Boundary(t,s,sn) <- nodes if s == bs && sn == set_name && defFor(b, Context) == ctx)
			yield ctx
		b.headOption.getOrElse {
			val nb = Boundary(t,bs,set_name)
			nodes ::= nb
			edges ::= Edge(Context,nb,ctx)
			nb
		}
	}
	def draw(gv : Liszt.GraphViz) {
		for(Edge(obj,u,d) <- edges)
			gv(u -> d, obj match {
				case Base => "black"
				case Context => "blue"
				case ReferenceElem => "red"
			})
		gv.draw()
	}
	def typeFor(node : Node) : Type = node match {
		case e : Each => typeFor(defFor(e,Base)).asInstanceOf[Set].t
		case TopoUse(op) => op.rett
		case Boundary(t,_,_) => t
		case TopLevel => UnitT
	}
	
	def isChildContext(child : Node, parent : Node) : Boolean = if(child eq parent) true else {
		child match {
			case TopLevel => false
			case _ => isChildContext(defFor(child,Context),parent)
		}
	}
	//if we genenerated the nested forall statements for this graph, will
	//_node_'s definition be avaliable from _ctx_
	//this helps us filter out imposible node graphs during abstract interpretation
	def avaliableIn(node : Node, ctx : Node) =  {
		val nodectx = node match {
			case _ : Each => node
			case TopLevel => node
			case _ => defFor(node,Context)
		}
		isChildContext(ctx,nodectx)
	}
	
	def allNodes : List[Node] = nodes
	def allEdges : List[Edge] = edges
	
	private var nodes : List[Node] = List.empty
	private var edges : List[Edge] = List.empty
}
