package Liszt
import Liszt._
import Liszt.Types._
import Liszt.Trees._

import scala.collection.immutable.{Set => SSet}
import scala.collection.mutable.{Set => MSet}
import scala.collection.mutable.{Stack => SStack}
import scala.collection.mutable.{Map => SMap}


class FunctionVectorizer() {
  def LiftOnce(t : Type) : Type = {
    t match {
      case SimdT(_) => t;
      case FloatT => SimdT(FloatT);
      case DoubleT => SimdT(DoubleT);
      case IntT => SimdT(IntT);
      case BoolT => SimdT(BoolT);
      case x @ VectorT(_,_) => SimdT(x);
      case x @ MatrixT(_,_,_) => SimdT(x);
      // The result of reading a Set(Vertex) should be SimdT(Vertex),
      // NOTE(boulos): Mesh isn't included. It shouldn't ever be
      // lifted as there is only one global mesh.
      case Vertex => SimdT(Vertex);
      case Edge => SimdT(Edge);
      case Face => SimdT(Face);
      case Cell => SimdT(Cell);

      // The result of say vertices(SimdT(face)) should be SimdT(Set(Vertex))
      case x @ Set(_) => SimdT(x);
      case _ => t;
    }
  }


  var localVars : MSet[String] = MSet.empty;
  var vectorVariables : MSet[String] = MSet.empty;
  var outsideForalls : MSet[String] = MSet.empty;
  var nestedCalls : MSet[String] = MSet.empty;

  def FindAllLocalVars(body : Tree) : Unit = {
    for (t <- body) {
      t match {
        case DefVar(name, init) => localVars.add(name);
        case _ => ;
      }
    }
  }

  // As a simple vectorization pass, mark each top-level forall as
  // having the induction variable be vectorized.
  def MarkTopLevelForalls(body : Tree) : Unit = {
    body match {
      case Forall(name, _, set, block) => setVec(name.name);
      case While(pred, block) => MarkTopLevelForalls(block);
      case If(pred, then_block, else_block) => { MarkTopLevelForalls(then_block); MarkTopLevelForalls(else_block); }
      case IfExp(t, pred, then_block, else_block) => { MarkTopLevelForalls(then_block); MarkTopLevelForalls(else_block); }
      case Block(stmts, exp) => { for (s <- stmts) MarkTopLevelForalls(s);   MarkTopLevelForalls(exp); }
      case _ => ;
    }
  }

  // Mark all DefVars outside of top-level forall blocks (because
  // these are currently the only vectorization targets) as being
  // "loop carried". This is not a general "Loop Carried Variable"
  // detector as clearly that would at least need to know which loop
  // we're talking about ;).
  def MarkTopLevelLoopCarriedVars(body : Tree) : Unit = {
    body match {
      case DefVar(name, init) => setLoopCarried(name);
      case While(pred, block) => MarkTopLevelLoopCarriedVars(pred); MarkTopLevelLoopCarriedVars(block);
      case If(pred, then_block, else_block) => MarkTopLevelLoopCarriedVars(pred); MarkTopLevelLoopCarriedVars(then_block); MarkTopLevelLoopCarriedVars(else_block);
      case IfExp(t, pred, then_block, else_block) => MarkTopLevelLoopCarriedVars(pred); MarkTopLevelLoopCarriedVars(then_block); MarkTopLevelLoopCarriedVars(else_block);
      case Block(stmts, exp) => { for (s <- stmts) MarkTopLevelLoopCarriedVars(s); MarkTopLevelLoopCarriedVars(exp); }
      case _ => ;
    }
  }

  def FlowVectorDependencies(depender : String, body : Tree) : Unit = {
    body match {
      // If an expression reads our loop variable (it can't write to
      // it conveniently) or on a variable that also depends on
      // ours, mark it as depending on it.

      // TODO(boulos): Reduction variables (scalars declared outside
      // a forall need to have their += instructions reduced into
      // them appropriately).
      case DefVar(name, init) => FlowVectorDependencies(name, init);
      // Variable reference. If we reference a vector, mark depender
      // as being "vectorized" too.
      case Var(t, name) => if (isVec(name)) { setVec(depender); } else { ; }
      case Assign(lhs, op, rhs) => {
        lhs match {
          // An assignment to a loop-carried var shouldn't necessarily
          // make the argument a vector.
          case Var(t, name) => /*println("assign of var '%s', isLoopCarried = %s".format(name, isLoopCarried(name))); */if (!isLoopCarried(name)) FlowVectorDependencies(name, rhs);
          case VecExtract(_, Var(_, name), _) => if (!isLoopCarried(name)) FlowVectorDependencies(name, rhs);
          case MatExtract(_, Var(_, name), _,_) => if (!isLoopCarried(name)) FlowVectorDependencies(name, rhs);
          case _ => /*println("Skipping non-var assignment '%s'".format(body.toString()));*/ ;
        }
      }
      case Binary(t, lhs, op, rhs) => FlowVectorDependencies(depender, lhs); FlowVectorDependencies(depender, rhs);
      case Unary(t, op, rhs) => FlowVectorDependencies(depender, rhs);
      case Call(t, func, args) => { for (a <- args) FlowVectorDependencies(depender, a); }
      case TopoCall(t, func, args) => { for (a <- args) FlowVectorDependencies(depender, a); }
      case FieldRead(t, field, arg) => FlowVectorDependencies(depender, field); FlowVectorDependencies(depender, arg);
      case VecExtract(t, vec, arg) => FlowVectorDependencies(depender, vec);
      case MatExtract(t, mat, n, m) => FlowVectorDependencies(depender, mat);
      // TODO(boulos): What should fieldwrite do? If you write to a
      // field with a vectorized argument that doesn't mean the
      // input was vectorized.
      case While(pred, block) => FlowVectorDependencies(depender, pred); FlowVectorDependencies(depender, block);
      case Block(stmts, exp) => { for (s <- stmts) FlowVectorDependencies(depender, s); FlowVectorDependencies(depender, exp); }
      case Forall(name,_,set, block) => FlowVectorDependencies(name.name, set); FlowVectorDependencies(depender, block);
      case If(pred, thenCode, elseCode) => FlowVectorDependencies(depender, pred); FlowVectorDependencies(depender, thenCode); FlowVectorDependencies(depender, elseCode);
      case IfExp(t, pred, thenCode, elseCode) => FlowVectorDependencies(depender, pred); FlowVectorDependencies(depender, thenCode); FlowVectorDependencies(depender, elseCode);

      case VectorL(t, vars) => { for (v <- vars) FlowVectorDependencies(depender, v); }
      case MatrixL(t, vars) => { for (v <- vars) FlowVectorDependencies(depender, v); }
      case MatrixFromVec(t, vars) => { for (v <- vars) FlowVectorDependencies(depender, v); }

      // Anything else, ignore for now.
      case _ => ;
    }
  }

  def setVec(name : String) = vectorVariables.add(name);
  def isVec(name : String) : Boolean = vectorVariables.contains(name);
  def isVec(v : Var) : Boolean = isVec(v.name);

  def getNestedFunc(func : Var, args : List[Var]) : Var = {
    val builtin_funcs : SSet[String] = SSet("exp", "expf", "sin", "sinf", "cos", "cosf", "tan", "tanf", "atan", "atanf", "atan2", "atan2f",
                                            "pow", "powf", "log", "logf", "log2", "log2f", "acos", "acosf", "asin", "asinf",
                                            "length", "dot", "normalize", "cross", "abs", "fabs", "fabsf", "sqrt", "sqrtf", "size", "length", "toFloat");
    if (builtin_funcs.contains(func.name)) {
      return func;
    } else {
      nestedCalls.add(func.name);
      Var(func.t, func.name + "_nested") copyAttr func;
    }
  }

  def clearLoopCarriedInfo() = outsideForalls.clear();
  def setLoopCarried(name : String) = outsideForalls.add(name);
  // Either it's a decl outside the foralls, or it's a global (!localVars)
  def isLoopCarried(name : String) : Boolean = outsideForalls.contains(name) || !localVars.contains(name);

  def VectorizeExp(e : Exp) : Exp = {
    e match {
      case Var(t, name) => if (isVec(name)) (Var(LiftOnce(t), name) copyAttr e) else e;
      case Binary(t, lhs, op, rhs) => {
        if (isVec(lhs) || isVec(rhs)) {
          Binary(LiftOnce(t), VectorizeExp(lhs).asInstanceOf[Var], op, VectorizeExp(rhs).asInstanceOf[Var]) copyAttr e
        } else { e }
      }
      case x @ FloatL(_) => SimdL(x);
      case x @ DoubleL(_) => SimdL(x);
      case x @ IntL(_) => SimdL(x);
      case VectorL(t, elements) => {
        if (elements.foldLeft(false)((agg, el) => agg || isVec(el))) {
          val lifted_elements : List[Var] = List.empty ++ {for (el <- elements) yield VectorizeExp(el).asInstanceOf[Var]; }
          VectorL(LiftOnce(t), lifted_elements) copyAttr e;
        } else { e };
      }
      case MatrixL(t, elements) => {
        if (elements.foldLeft(false)((agg, el) => agg || isVec(el))) {
          val lifted_elements : List[Var] = List.empty ++ {for (el <- elements) yield VectorizeExp(el).asInstanceOf[Var]; }
          MatrixL(LiftOnce(t), lifted_elements) copyAttr e;
        } else { e };
      }
      case MatrixFromVec(t, elements) => {
        if (elements.foldLeft(false)((agg, el) => agg || isVec(el))) {
          val lifted_elements : List[Var] = List.empty ++ {for (el <- elements) yield VectorizeExp(el).asInstanceOf[Var]; }
          MatrixFromVec(LiftOnce(t), lifted_elements) copyAttr e;
        } else { e };
      }
      case Unary(t, op, v) => {
        if (isVec(v)) {
          Unary(LiftOnce(t), op, VectorizeExp(v).asInstanceOf[Var]) copyAttr e;
        } else { e }
      }
      case Call(t, func, args) => {
        if (args.foldLeft(false)((agg, el) => agg || isVec(el))) {
          val lifted_args : List[Var] = List.empty ++ {for (a <- args) yield VectorizeExp(a).asInstanceOf[Var]};
          val nested_func = getNestedFunc(func, lifted_args);

          Call(LiftOnce(t), nested_func, lifted_args) copyAttr e
        } else { e }
      }
      case TopoCall(t, func, args) => {
        if (args.foldLeft(false)((agg, el) => agg || isVec(el))) {
          val lifted_args : List[Var] = List.empty ++ {for (a <- args) yield VectorizeExp(a).asInstanceOf[Var];}
          TopoCall(LiftOnce(t), func, lifted_args) copyAttr e
        } else { e }
      }
      case VecExtract(t, vec, arg) => {
        if (isVec(vec)) {
          VecExtract(LiftOnce(t), vec, arg) copyAttr e;
        } else { e }
      }
      case MatExtract(t, mat, n, m) => {
        if (isVec(mat)) {
          MatExtract(LiftOnce(t), mat, n, m) copyAttr e;
        } else { e }
      }
      case FieldRead(t, field, arg) => {
        if (isVec(arg) || isVec(field)) {
          val lifted_field : Var = VectorizeExp(field).asInstanceOf[Var];
          val lifted_arg : Var   = VectorizeExp(arg).asInstanceOf[Var];
          FieldRead(LiftOnce(t), lifted_field, lifted_arg) copyAttr e;
        } else { e }
      }
      case IfExp(t, predBlock, thenBlock, elseBlock) => {
        val vec_pred : Block = VectorizeStmt(predBlock).asInstanceOf[Block];
        val vec_then : Block = VectorizeStmt(thenBlock).asInstanceOf[Block];
        val vec_else : Block = VectorizeStmt(elseBlock).asInstanceOf[Block];
        (IfExp(LiftOnce(t), vec_pred, vec_then, vec_else) copyAttr e)
      }
      case _ => e;
    }
  }

  // If our expression includes a variable that is vectorized,
  // replace the expression with a lifted version.
  def VectorizeStmt(s : Stmt) : Stmt = {
    s match {
      case If(predBlock, thenBlock, elseBlock) => {
        val vec_pred : Block = VectorizeStmt(predBlock).asInstanceOf[Block];
        val vec_then : Block = VectorizeStmt(thenBlock).asInstanceOf[Block];
        val vec_else : Block = VectorizeStmt(elseBlock).asInstanceOf[Block];
        (If(vec_pred, vec_then, vec_else) copyAttr s)
      }
      case FieldWrite(field, arg, op, rhs) => {
        if (isVec(field) || isVec(arg) || isVec(rhs)) {
          FieldWrite(VectorizeExp(field).asInstanceOf[Var], VectorizeExp(arg).asInstanceOf[Var], op, VectorizeExp(rhs).asInstanceOf[Var]) copyAttr s;
        } else { s }
      }
      case While(pred, body) => {
        val vec_pred : Block = VectorizeStmt(pred).asInstanceOf[Block];
        val vec_body : Block = VectorizeStmt(body).asInstanceOf[Block];
        While(vec_pred, vec_body) copyAttr s;
      }
      case Forall(name, label, set, body) => {
        val vec_def  : DefVar = VectorizeStmt(name).asInstanceOf[DefVar];
        val vec_set  : Var    = VectorizeExp(set).asInstanceOf[Var];
        val vec_body : Block = VectorizeStmt(body).asInstanceOf[Block];
        Forall(vec_def, label, vec_set, vec_body) copyAttr s;
      }
      case Return(e) => Return(VectorizeExp(e).asInstanceOf[SimpleExp]) copyAttr s;
      case Print(vals) => {
        val vec_vals : List[Var] = List.empty ++ {for (arg <- vals) yield VectorizeExp(arg).asInstanceOf[Var]}
        Print(vec_vals) copyAttr s;
      }
      case DefVar(name, init) => {
        if (isVec(name)) {
          DefVar(name, VectorizeExp(init)) copyAttr s;
        } else {
          s
        }
      }
      case Assign(lv, op, rhs) => Assign(VectorizeExp(lv).asInstanceOf[LValue], op, VectorizeExp(rhs).asInstanceOf[Var]) copyAttr s;
      case Block(stmts, exp) => {
        val vec_stmts : List[Stmt] = List.empty ++ {for (stmt <- stmts) yield VectorizeStmt(stmt)}
        val vec_exp : SimpleExp = VectorizeExp(exp).asInstanceOf[SimpleExp];
        Block(vec_stmts, vec_exp) copyAttr s;
      }
      case ExpStmt(exp) => ExpStmt(VectorizeExp(exp)) copyAttr s;
      case _ => s;
    }
  }

  // Given a function, vectorize the foralls by lifting variables to
  // SimdT versions. Record for each called function whether or not
  // it's argument will be called with a vectorized
  // argument. TODO(boulos): There's a phase issue here, wherein we
  // need to go around vectorizing things that may have calls and then
  // go back and cleanup after ourselves (which then might make more
  // things not-vectorized, etc).
  def Vectorized(func : DefFunc, nested : Boolean) : DefFunc = {
    val debug = false;
    if (debug) {
      println("Producing  vectorized version of function '%s' (nested = %s)".format(func.name, nested));
      val nested_string = if (nested) "nested" else "";
      println("The %sfunction we're working on is:\n%s\n%s\n%s\n\n".format(nested_string, "-"*80, func, "-"*80));
    }

    // Set the return type to the (possibly) lifted version.
    // Consider adding __inputMask here? (need to deal with the tFunctionHeader as well..)
    vectorVariables.clear();
    clearLoopCarriedInfo();
    localVars.clear();
    FindAllLocalVars(func.body);

    // Before messing with the scalar funciton at all, determine which
    // variables are loop-carried. This allows us to preserve the
    // semantics of the scalar function information.
    if (!nested) {
      // NOTE(boulos): Since we don't change the way loops work in
      // nested functions, we don't need or even want this
      // analysis. In a nested function, a SIMD assignment to a "loop
      // carried" var means that the variable is "varying" while it's
      // "uniform" if that never happens (allowing us to keep the
      // value as a scalar).

      //println("marking loop-carried vars");
      MarkTopLevelLoopCarriedVars(func.body);
      //println("%s".format("-"*20));
      //for (loop_carried <- outsideForalls) println("  '%s'".format(loop_carried));
      //println("%s\n".format("-"*20));
    }

    val vec_args = if (nested) {List.empty ++ {for (arg <- func.args) yield VectorizeStmt(arg).asInstanceOf[DefVar];} } else { func.args };
    val arg_types = if (nested) { List.empty ++ {for (arg <- func.typ.args) yield LiftOnce(arg); }} else { func.typ.args }
    // TODO(boulos): Change nested to "Any arg is SimdT" (once
    // callInfo is worthwhile)
    val return_type = if (nested) LiftOnce(func.typ.range) else func.typ.range;

    // Label all args as vectorized if we're "nested".
    if (nested) {
      for (arg <- vec_args) setVec(arg.name);
    } else {
      // We're not nested, label each 1st level forall variable as
      // vectorized and then flow it.
      //println("marking top level foralls as vectorize-able");
      MarkTopLevelForalls(func.body);
    }

    // Analyze the body


    if (debug) {
      println("flowing dependencies for vector variables:");
      for (vecVar <- vectorVariables)
        println("  '%s'".format(vecVar));
      println("--------------------------");
    }
    FlowVectorDependencies("ERROR", func.body);
    if (debug) {
      println(" After Flowing Vec Status ");
      println("--------------------------");
      for (vecVar <- vectorVariables)
        println("  '%s'".format(vecVar));

      println("--------------------------");
    }

    // lift the body (respecting vector variables)
    val new_name = if (nested) func.name + "_nested" else func.name;
    var result = DefFunc(Func(arg_types, return_type), new_name, vec_args, VectorizeStmt(func.body)) copyAttr func;
    // Give brand new id's to everything in the cloned tree.
    if (nested) { result.renumber(); }

    //println("Vectorized the function. Got this =\n-----------------\n%s----------------".format(result.toString()));
    return result;
  }
}

// Take an input code block and take each function and vectorize
// it. Then take the results of the calls to generate the nested
// versions.
class Vectorizer(topLevel : Block) {
  def withVectorized : Block = {
    var func_vec = new FunctionVectorizer();
    var nameToDFMap : SMap[String, DefFunc] = SMap.empty;
    val stmts = for(s <- topLevel.stmts) yield s match {
      case df @ DefFunc(_,name,_,x) if x != ExternStmt => {
        // Have an actual body.
        val vectorized_fun = func_vec.Vectorized(df, false);
        nameToDFMap.put(name, df);
        vectorized_fun;
      }
      case x => x;
    }
    // Okay, now we can generate all the nested versions we need.
    var alreadyEmitted : MSet[String] = MSet.empty;
    var combined_stmts = stmts;

    while ( func_vec.nestedCalls.foldLeft(false)( (cur, call) => cur || !(alreadyEmitted.contains(call))) ) {
      val cloned_calls = func_vec.nestedCalls.clone();
      val new_stmts =
        for (nested_func <- cloned_calls if !alreadyEmitted.contains(nested_func))
          yield nameToDFMap.get(nested_func) match {
            case Some(df @ DefFunc(_,name,_,x)) => alreadyEmitted.add(nested_func); func_vec.Vectorized(df, true);
            case None => println("ERROR: Need to produce a 'nested' version of %s, but didn't encounter it in our top-level block... impossible?".format(nested_func)); alreadyEmitted.add(nested_func); ExternStmt;
          }


      combined_stmts ++= new_stmts;
    }

    Block(combined_stmts, topLevel.exp) copyAttr topLevel;
  }
}
