package Liszt
import Liszt._
import Liszt.Types._
import Liszt.Trees._
import scala.collection.immutable.{Set => SSet}

//A workgroup in Liszt is a block of statements that can all be run simultaneously with
//no synchronization except to handle incremental reductionsn to field entries.

//A forall-block always fits into a single work by the definition of the language
//Here we try to group multiple statements into single workgroups when possible.
//Statements inside a workgroup may include compound statements (statements containing 
//other statements) such as if/while.  In these cases, we assume the statements within
//a compound statement are run serially with respect to eachother, but in parallel with other 
//statements in the workgroup.
class WorkgroupCreation(code : Block) {
	type UseMap = Map[String, SSet[Ops.ReductionOp]]
	val function_map = ASTUtil.functionMap(code)
	//given a part of the AST return all the reads/writes that the tree performs
	//on the _free variables_ in the statement.
	val uses : Tree => UseMap = Util.Memoize[Tree,UseMap] { 
		(tree : Tree) => 
		def mergePair(a : UseMap, b : UseMap) : UseMap = Util.merge(a,b) { (x,y) => x union y }
		def merge(as : List[UseMap]) : UseMap = as match {
			case h :: t => t.foldLeft(h)(mergePair)
			case Nil => Map.empty
		}
		def single(v : Var, op : Ops.ReductionOp) : UseMap = Map { v.name -> SSet(op) } 
		def read(v : Var) = single(v,Ops.RNone)
		
		tree match {
			case Call(t,fn,args) if function_map.contains(fn.name) => 
				mergePair(merge(args map uses), uses(function_map(fn.name)))
			
			case v : Var => read(v)
			case Block(stmts,exp) => {
				val defined = for(DefVar(n,_) <- stmts)
					yield n
				merge((exp :: stmts) map uses) -- defined //defined are not free inside block
			}
			case EnsureFieldState(f,reduction) => Map.empty //we need to prevent from counting as a 'read'
			case DefVar(name,init) => mergePair(Map { name -> SSet(Ops.RAssign) }, uses(init))
			case _ => tree match {
				case WriteStmt(v, op, ts) => mergePair(merge(ts map uses), single(v,op))
				case ShapeAny(_,ns,ts) => merge(ts map uses) -- ns //ns are not free variables, so we remove them 
			}
		}
	}.apply
	def workgroupified : Block = {
		
		
		//determines if we can add the new uses generated from a statement (proposedAdditions)
		//to the uses already scheduled in the current map (currentUses)
		
		//in general we can only add the additional statement if in the set of intersecting
		//field states a field is in the same state in both maps (e.g. both for-stmts 
		//add to the same field but do not read it).
		
		//a proposed addition to a workblock can contain both reads/writes to the same field
		//In this case we can only add it to the workgroup if no other member of the workgroup
		//reads or writes that field (i.e. it gets exclusive rights to work on that field)
		//No future items in the work group may be added if they intersect
		
		def tryAdd(currentUses : UseMap, proposedAddition : UseMap) : Option[UseMap] = {
			val merged = Util.merge(currentUses,proposedAddition) { (lhs,rhs) =>
				if(lhs.size > 1 || //already have a work item that needs exclusive access to this field 
				   rhs.size > 1 || //the new work item needs exclusive access but an item already in the group needs access as well
				   lhs != rhs)     //conflicting uses between work group and proposed additions 
				   return None
				lhs
			}
			Some(merged)
		}
		
		def workgroupify(tree : Tree) : Tree = tree match {
			case Block(stmts_,exp) => {
				val stmts = stmts_ map workgroupify
				var workgroups : List[Workgroup] = List.empty
				var current_group_map : UseMap = Map.empty
				var current_group : List[Stmt] = List.empty
				def finishGroup() {
					if(!current_group.isEmpty)
						workgroups ::= Workgroup(current_group.reverse)
					current_group = List.empty
				}
				
				for(stmt <- stmts) {
					val stmt_uses = uses(stmt)
					tryAdd(current_group_map,stmt_uses) match {
						case Some(nm) => {
							current_group_map = nm
						}
						case None => {
							finishGroup()
							current_group_map = stmt_uses
						}
					}
					current_group ::= stmt.asInstanceOf[Stmt]
				}
				
				finishGroup()
				Block(workgroups.reverse,exp)
			}
			case ShapeAny(_,_,ts) => ShapeAny.trees(tree,ts map workgroupify)
		}
		
		val init_block = for(stmt <- code.stmts if !stmt.isInstanceOf[DefFunc]) 
			yield workgroupify(stmt).asInstanceOf[Stmt]
		val defs = for( (_,df) <- function_map) 
			yield workgroupify(df).asInstanceOf[Stmt]
		Block(defs.toList ::: init_block, code.exp)
	}
}

class Location
case class ScalarLocation(name : String) extends Location
case class FieldLocation(t : Type, name : FieldInit, entry : Topology.Node) extends Location
	
//record containing information about a write done inside a workgroup
class WorkgroupWrite(
	val stmt : Stmt, //statement that performs the write
	val context : Topology.Node, //in what forall nesting contexts can this take place, node will point to a forall node
	val location : Location ,// the abstract 'location' of a write, for scalars this is simply the unique name of the scalar, for fields, the name of the field and abstract representation of the set used to access it
	val tgraph : Topology.Graph
) {
	override def toString = "WorkgroupWrite { stmt = %s, context = %s, location = %s }".format(stmt,context,location)
	def isTrivial = location match {
			case l: FieldLocation => {
				(tgraph.defFor(l.entry,Topology.Context) == Topology.TopLevel)
			}
			case _ => true
		}
}


class ColoringGenerator(graph : Topology.Graph, workgroup_writes : Map[Workgroup,List[WorkgroupWrite]], workgroup_triviality : Map[Int,Boolean], tType : Type => String) {
	class Tree
	case class Def(t : Type, name : String, call : Ops.TopoFn, args : List[String]) extends Tree
	
	case class Edge(workgroupID : Int, field_type : Type, fieldID : Int, entry : String) extends Tree
	
	case class Block(loop : Option[(String,String)]) extends Tree {
		var stmts : List[Tree] = List.empty
		override def toString = "%s {\n %s \n }".format(loop match {
			case None => "TopLevel "
			case Some( (a,b) ) => "for(%s <- %s) ".format(a,b)
		} ,  {for(s <- stmts.reverse) yield s.toString }.mkString(";\n"))
	}
	def name() = ASTUtil.Names.create("v")
	
	val blockForContext : Topology.Node => Block = Util.Memoize[Topology.Node,Block] {
		case Topology.TopLevel => Block(None)
		case node =>
		 val ctx = blockForContext(graph.defFor(node,Topology.Context))
		 val st = graph.defFor(node,Topology.Base)
		 val blk = Block(Some( (name(),ref(st) ) ) )
		 ctx.stmts ::= blk
		 blk
	}.apply
	
	val ref : Topology.Node => String = Util.Memoize[Topology.Node,String] {
		(node : Topology.Node) => 
		
		val r = node match {
			case e : Topology.Each => {
				val Block(Some((a,_))) = blockForContext(e)
				a
			}
			case Topology.TopoUse(op : Ops.Towards) => {
				val ctx = blockForContext(graph.defFor(node,Topology.Context))
				val arg = ref(graph.defFor(node,Topology.Base))
				val arg2 = ref(graph.defFor(node,Topology.ReferenceElem))
				val n = name()
				ctx.stmts ::= Def(op.rett,n,op,List(arg,arg2))
				n
			}
			case Topology.TopoUse(op) => {
				val ctx = blockForContext(graph.defFor(node,Topology.Context))
				val arg = ref(graph.defFor(node,Topology.Base))
				val n = name()
				ctx.stmts ::= Def(op.rett,n,op,List(arg))
				n
			}
			case Topology.Boundary(t,str,name) => "COLORING_GET_BOUNDARY(%s,\"%s\")".format(str,name) 
			case Topology.TopLevel => "&cpu_mesh"
		}
		//println("created reference: %s = %s".format(r,node))
		r
	}.apply
	
	def add(workgroupID : Int, write : WorkgroupWrite) {
		val ctx = blockForContext(write.context)
		val FieldLocation(t,name,entry) = write.location
		val entry_ref = ref(entry)
		ctx.stmts ::= Edge(workgroupID,t,name.id,entry_ref)
	}
	def isNested(t : Liszt.Trees.Tree) : Boolean = t.nesting._2
	
	//generates tree
	for((wg,writes) <- workgroup_writes; write <- writes) {
		if (!workgroup_triviality(wg.id) && !isNested(wg))
			add(wg.id,write)

	}
	
	def write() : String = {
		class Line(idt : Int, line : String) {
			def indent : Line = new Line(idt + 1, line)
			override def toString = "%s%s".format("    " * idt,line)
		}
		def indent(lines : List[Line]) : List[Line] = for(line <- lines) yield line.indent
		def line(str : String) = List(new Line(0,str))
		def writeStmts(trees : List[Tree]) : List[Line] = {
			val (stmts,blks) = trees.partition{ x => !x.isInstanceOf[Block] }
			val stmt_lines = for(s <- stmts; t <- writeTree(s)) yield t
			val block_lines = for(s <- blks; t <- writeTree(s)) yield t
			stmt_lines ::: block_lines
		}
		def writeTree(tree : Tree) : List[Line] = tree match {
			case Def(t,name,op,args) => line { 
				"%s %s = %s(%s);".format(t match {
					case Set(_) => "%s_return_type".format(op.asFunction)
					case _ => tType(t)
				},name,op.asFunction,args.mkString(","))
			}
			case blk @ Block(None) => 
			line("void createColoringMetaData(const ColoringMesh & cpu_mesh, ::Synchronization::InterferenceBuilder * igraph) {") :::
			indent {
				line("printf(\"Start createColoringMetaData to build for->fieldwrite graph\\n\");\nstd::vector<id_type> context;") :::
				{ writeStmts(blk.stmts.reverse) }
			} ::: line("}")
			case blk @ Block( Some( (a,b) ) ) =>
			line("context.push_back(0);") :::
			line("FORALL_SET(%s,%s)".format(a,b)) :::
			indent { 
				line("context.back() = %s.ID();".format(a)) :::
				{ writeStmts(blk.stmts.reverse) }
			} ::: 
			line("ENDSET") :::
			line("context.pop_back();")
			case Edge(workgroupID, FieldT(t,_),fieldID, entry) => {
				val t_enum = t match {
					case Vertex => "VERTEX"
					case Liszt.Types.Edge =>  "EDGE"
					case Face => "FACE"
					case Cell => "CELL"
				}
				line("igraph->addEdge(::LisztPrivate::ElemTypes::%s,%d,%s.ID(),%d,::Synchronization::SimpleAbstractWriteContext(context));".format(t_enum,fieldID,entry,workgroupID))
			}
		}
		val lines = writeTree(blockForContext(Topology.TopLevel))
		lines.mkString("\n")
	}
}

//create a map from workgroup -> set of globals that are written (with reductions) within the workgroup
class GlobalWritesForWorkgroup(code : Block) {
	val function_map = ASTUtil.functionMap(code)
	val function_writes : String => SSet[String] = Util.Memoize[String, SSet[String]] { 
		(fname) => 
		val df = function_map(fname)
		writesInTree(df) 
	}.apply
	def writesInTree(tree : Tree) : SSet[String] = {
		val uses = SSet.empty ++ tree.flatMap { 
			case Assign(v @ Var(_,name),_,_) if v.global => List(name)
			case Call(_,fn,_) if function_map.contains(fn.name) => function_writes(fn.name).toList
			case _ => List()
		}
		val defs = SSet.empty ++ tree.flatMap {
			case dv : DefVar if dv.global => List(dv.name)
			case _ => List()
		}
		uses -- defs
	}
}
