import Liszt.Language._

@lisztcode
object Constants {
	val iterations = 30 ;
	val zero = 1.0e-9;
}
@lisztcode
object FEM {
  val wall = BoundarySet[Face]("wall")

  val position = FieldWithLabel[Vertex,Vec[_3,Float]]("position")
  val sol = FieldWithConst[Vertex,Double](0.0)
  val b    = FieldWithConst[Vertex,Double](0.0)
  val scratch  = FieldWithConst[Vertex,Double](0.0)
  val r    = FieldWithConst[Vertex,Double](0.0)
  val p    = FieldWithConst[Vertex,Double](0.0)
  val ref  = FieldWithConst[Vertex,Double](0.0)
  val isBoundary  = FieldWithConst[Vertex,Double](0.0)

  //some geometry fields
  val cell_volume = FieldWithConst[Cell,Double](0.f)

  def calcCellGeom() : Unit = {
    var numCells = 0.0
    for (c <- cells(mesh)) {
      numCells += 1.0
    }
    var volume = (2.0*2.0*2.0)/numCells
    for (c <- cells(mesh)) {
      cell_volume(c) = volume
    }
  }


  def DirichletValue(v : Vertex) : Double = {
    val c = position(v)
    return 0.5*(c(_0)*c(_0) + c(_1)*c(_1)) - c(_2)*c(_2)
  }

  def ReferenceSolution(v : Vertex) : Double = {
    val c = position(v)
    return DirichletValue(v);
  }

  /*def TriLinearHex(i : Int, j : Int) : Double = {
    var ii = 0
    var jj = 0
    if (i > j) {
      ii = j
      jj = i
    } else {
      ii = i
      jj = j
    }
    if (ii == jj) {
      return 2.0/3.0
    } else if ((jj == ii+1 && ii != 3)
               || (jj == ii+4)
	       || (jj == 3 && ii == 0)
	       || (jj == 7 && ii == 4)) {
      return 0.0
    } else {
      return -1.0/6.0
    }
  }

  // Does the product A*p = scratch.
  def ElementMVProduct(c : Cell, include_diag : Boolean) : Unit = {
    var je = cell_volume(c) / 8.0
    var i = 0
    var j = 0
    while (i < 8) {
      j = 0
      while (j < 8) {
	if (i != j || include_diag) {
	  scratch(vertex(c,i)) += (1/je)*TriLinearHex(i,j)*p(vertex(c,j))
	}
	j = j + 1
      }
      i = i + 1
    }
  }*/

  def TriLinearHex(i : Vertex, j : Vertex) : Double = {

    var ipos = position(i)
    var jpos = position(j)
    var diff = ipos - jpos

    var d0 = fabs(diff(_0))
    var d1 = fabs(diff(_1))
    var d2 = fabs(diff(_2))

    if (d0 == 0.0 && d1 == 0.0 && d2 == 0.0) {
      return 2.0/3.0
    } else {
      var a = 0.0
      var eps0 = 1e-8
      if (d0 > eps0) {
	a = d0
      } else if (d1 > eps0) {
	a = d1
      } else {
	a = d2
      }
      var eps1 = 1e-8
      if ( (d0+d1+d2)/a >= 2.0 - eps1) {
	return -1.0/6.0
      } else {
	return 0.0
      }
    }
  }

  // Does the product A*p = scratch.
  def ElementMVProduct(c : Cell, include_diag : Boolean) : Unit = {
    var je = cell_volume(c) / 8.0
    for (i <- vertices(c)) {
      var sc = 0.0;
      for (j <- vertices(c)) {
	sc += (1/je)*TriLinearHex(i,j)*p(j)
      }
      scratch(i) += sc;
    }
  }

  def SetBoundary() : Unit = {
    for(v <- vertices(mesh))
    	if(isBoundary(v) > 0.0)
			sol(v) = DirichletValue(v)
  }

  def AssembleRHS() : Unit = {
    ClearScratch()
    for (v <- vertices(mesh)) {
      p(v) = 1.0
    }
    
    //Calculate the product.
    for (c <- cells(mesh)) {
      ElementMVProduct(c, true)
    }

    // Move the values to the rhs b.
    for (v <- vertices(mesh)) {
      b(v) = 1.0 //scratch(v)
    }
  }

  def ClearSol() : Unit = {
    for (v <- vertices(mesh)) {
      sol(v) = 0.0
    }
  }

  def ClearB() : Unit = {
    for (v <- vertices(mesh)) {
      b(v) = 0.0
    }
  }

  def ClearR() : Unit = {
    for (v <- vertices(mesh)) {
      r(v) = 0.0
    }
  }

  def ClearP() : Unit = {
    for (v <- vertices(mesh)) {
      p(v) = 0.0
    }
  }

  def ClearScratch() : Unit = {
    for (v <- vertices(mesh)) {
      scratch(v) = 0.0
    }
  }

  def ClearDirichletNodesB() : Unit = {
    for(v <- vertices(mesh))
    	if(isBoundary(v) > 0.0)
    		b(v) = 0.0
  }

  def ClearDirichletNodesScratch() : Unit = {
    for(v <- vertices(mesh))
    	if(isBoundary(v) > 0.0)
			scratch(v) = 0.0
  }

  def SolveCG() : Unit = {
    var done = 0
    var old_residual = 0.0
    var rsold = 0.0
    var rsnew = 0.0
    var pap = 0.0
    var alpha = 0.0
    var normb = 0.0

    // Clear the working vectors and the initial guess.
    ClearSol()
    ClearR()
    ClearP()

    // Apply boundary to b
    ClearDirichletNodesB()

    // Calculate the 2-norm of b.
    for (v <- vertices(mesh)) {
      normb += b(v)*b(v)
    }
    normb = sqrt(normb)

    Print("The 2-norm of b is ", normb)

    // r = -b
    for (v <- vertices(mesh)) {
      r(v) =  -b(v)
    }
    
    // p = -r
    for (v <- vertices(mesh)) {
      p(v) = -r(v)
    }

    // rsold = r'*r
    rsold = 0.0
    for (v <- vertices(mesh)) {
      rsold += r(v)*r(v)
    }

	var iter = 0
	var start_time = 0.0;
    while (done == 0) {
		if (iter == 1) {
			start_time = wall_time();
		}

      // Clear the scratch vector.
      ClearScratch()
      
      // scratch = A*p
      for (c <- cells(mesh)) {
	ElementMVProduct(c, true)      
      }
      ClearDirichletNodesScratch()

      // pap = p'*A*p
      pap = 0.f
      for (v <- vertices(mesh)) {
	pap += p(v)*scratch(v)
      }

      // alpha = r'*r / p'*A*p
      alpha = rsold / pap
      //Print("\trsold = ", rsold, " and pAp = ", pap)

      // x = x + alpha*p
      for (v <- vertices(mesh)) {
	sol(v) += alpha*p(v)
      }

      // r = r + alpha*A*p
      for (v <- vertices(mesh)) {
	r(v) += alpha*scratch(v)
      }

      // rsnes = r'*r
      rsnew = 0.f
      for (v <- vertices(mesh)) {
	rsnew += r(v)*r(v)
      }

      // p = -r + beta*p
      // This require two passes since we cannot read and write p in one loop.
      for (v <- vertices(mesh)) {
	p(v) *= rsnew/rsold
      }
      for (v <- vertices(mesh)) {
	p(v) -= r(v)
      }
      
      rsold = rsnew

      // Check for convergence of the maximum number or iterations.
      if (iter % 1 == 0 || iter >= Constants.iterations) {
	var residual = sqrt(rsnew)/normb

	Print("cg iter, residual: ", iter, " ", residual)
	if (iter >= Constants.iterations) {
	  done = 1
	}
	old_residual = residual
      }
      iter += 1
    }
	Print("TIME_FOR_LOOP: ", wall_time() - start_time);

    if (iter < Constants.iterations) {
      Print("Converged in = ", iter, " iterations.")
    } else {
      Print("Maximum number of iterations reached.")
    }

  }

  def MPISanityCheck() : Unit = {
    for (v <- vertices(mesh)) {
      if (ID(v) == 999)
	p(v) = 1.0
      else
	p(v) = 0.0
    }
    for (c <- cells(mesh)) {
      ElementMVProduct(c, true)      
    }
    for (v <- vertices(mesh)) {
      Print(ID(v), " : ", scratch(v))
    }
  }
  
	def main() {
		for(f <- wall)
			for(v <- vertices(f))
				isBoundary(v) += 1.0
		// Calculate the cell geometries.
		for(f <- faces(mesh))
			for(v <- vertices(f))
				isBoundary(v) += 0.0

                // Calculate the cell geometries.
		calcCellGeom()
	
		// Assemble the problem.
		SetBoundary()
		AssembleRHS()
	
		// Solve the system.
		SolveCG();
		//MPISanityCheck()
	}
}
