import Liszt.Language._
import Liszt.MetaInteger._

@lisztcode
object JoeLS extends LinearSystem with PetSc {
   def row(c : Cell) : Vec[_5,RowIndex] = AutoIndex
   def col(c : Cell) : Vec[_5,ColIndex] = AutoIndex
}

@lisztcode
object JoeLiszt {

	// Flow variable initialization
	val rho = FieldWithConst[Cell,Double](IC.rho_init) ;
	val rhou = FieldWithConst[Cell,Vec[_3,Double]](IC.rho_init * IC.u_init) ;
	val rhoE = FieldWithConst[Cell,Double](IC.p_init / (IC.GAMMA - 1.0)  +  0.5 * IC.rho_init * dot( IC.u_init, IC.u_init )) ;

	val zone_all = BoundarySet[Face]("zone_all") ;
	val zone_interior = BoundarySet[Face]("zone_interior") ;
	val zone_boundary = BoundarySet[Face]("zone_boundary") ;
	val symmetry = BoundarySet[Face]("symmetry") ;
	val wall = BoundarySet[Face]("wall") ;
	val other_boundaries = BoundarySet[Face]("other_boundaries") ;
	val cbc = BoundarySet[Face]("cbc") ;
	val cbc_subsonic_outlet = BoundarySet[Face]("cbc_subsonic_outlet") ;
	val cbc_subsonic_inlet = BoundarySet[Face]("cbc_subsonic_inlet") ;
	val symmetry_pos = BoundarySet[Face]("symmetry_pos") ;
	val symmetry_neg = BoundarySet[Face]("symmetry_neg") ;
	
	// Definition of the residual vector
	val rhs_rho = FieldWithConst[Cell,Double](0.0) ;
	val rhs_rhou = FieldWithConst[Cell,Vec[_3,Double]](Constants.float3_zero) ;
	val rhs_rhoE = FieldWithConst[Cell,Double](0.0) ;

	// Definition of the implicit system
	JoeLS.nonzeroes {
		nz =>
			for(c1 <- cells(mesh)) {
				val row = JoeLS.row(c1)
				val col = JoeLS.col(c1)
				
				nz.b(row); 
				nz.x(col); 
				nz.A(row,col)
				
				for(c2 <- cells(mesh)) {
					nz.A(JoeLS.row(c1), JoeLS.col(c2))
					nz.A(JoeLS.row(c2), JoeLS.col(c1))
				}
			}
	}
	val A = JoeLS.A() ;
	val x = JoeLS.x() ;
	val b = JoeLS.b() ;
	
	JoeLS.max_iters(200)
	JoeLS.target_resid(0.001)
	

	def getInsideBoundaryCell( f : Face ) : Cell = {
		val c = if ( ID(inside(f)) == 0 ) outside(f) else inside(f) ;
		return c ;
	}	


	def run() {
		
		JoeLisztSolver.calcStateMaterialProp() ;
		
		setBoundaryConditions() ;
		
		var start_time = 0.0;
		var step = 0 ;
		
		// External loop
		while (step < Constants.nsteps ) {
			
			if (step == 1) {
				start_time = wall_time() ;
			}
			
			step += 1 ;
			
			// Initialize the residuals
			var my_resid_rho = 0.0 ;
			var my_resid_rhou = Constants.float3_zero ;
			var my_resid_rhoE = 0.0 ;

			// Compute the time step
			var dtMin = JoeLisztSolver.calcDt(IC.cfl) ;

			// Spatial integration. Compute convective and viscous residuals and Jacobians
			SpatialIntegration() ;

			// Time integration. Update the solution
			TimeIntegration() ;

			// Residual update for visualization purpouses
			for( c <- cells(mesh) ) {
				my_resid_rho += fabs(rhs_rho(c)) ;
				val absVec = Vec( fabs(rhs_rhou(c).x), fabs(rhs_rhou(c).y), fabs(rhs_rhou(c).z) ) ;
				my_resid_rhou += absVec ;
				my_resid_rhoE += fabs(rhs_rhoE(c)) ;
			}

			val my_resid = Vec( my_resid_rho, my_resid_rhou.x, my_resid_rhou.y, my_resid_rhou.z, my_resid_rhoE ) ;

			if ( step % IC.check_interval == 0 ) {
				if ( step % ( IC.check_interval * 10 ) == 0 ) {
					Print( "" ) ; // print nothing to simulate a \n !!!!
					Print( "done step: " , step , ", cfl: " , IC.cfl , ", min. dt:   " , dtMin ) ; 
				}

				showResidue( my_resid, step ) ;
			} 

			JoeLisztSolver.calcStateMaterialProp() ;

			setBoundaryConditions() ;
		}

		 Print( "TIME_FOR_LOOP: ", wall_time() - start_time ) ;
		 save("output", 1) ;
		
	}



	def SpatialIntegration() {
		
		// Initialize the value of the residuals
		for( c <- cells(mesh) ) {
			rhs_rho(c) = 0.0 ;
			rhs_rhou(c) =  Constants.float3_zero ;
			rhs_rhoE(c) = 0.0 ;
		}
		
		// Initialize the value of the jacobians, rhs, and solution
		for( c <- cells(mesh) ) {
			val row = JoeLS.row(c)
			val col = JoeLS.col(c)
			A(row,col) = Constants.float5x5_zero ;
			b(row) = Constants.float5_zero ;
			x(col) = Constants.float5_zero ;
		}

		// Compute velocity, pressure, and enthalpy gradient for the second order reconstruction
		calcConservGradient() ;
		calcVelocityGradient() ;
		calcEntalphyGradient() ;
	 	calcPressureGradient() ;

  		// Space integration (inviscid contribution)
		for (f <- zone_interior) { 
			calcRhsInteriorInviscid(f) ;
			calcJacInteriorInviscid(f) ;
		}
		
		// Space integration (viscous contribution)
		if (IC.mu_ref > 0.0) { 
		   for (f <- zone_interior) { 
		      calcRhsInteriorViscous(f) ;
		      calcJacInteriorViscous(f) ;
		   }
		}
		
		// Set boundary conditions
		for (f <- symmetry) { calcRhsSymmetry(f) }
		for (f <- wall) { calcRhsWall(f) }
		for (f <- other_boundaries) { calcRhsOtherBoundaries(f) }
	}

	def TimeIntegration() {
		if (NavierStokesSolvers.IMPL_EULER == 1) {
			for(c <- cells(mesh)) {
				val row = JoeLS.row(c)
				val col = JoeLS.col(c)
				val tmp = MeshGeometryCalc.cv_volume(c) / JoeLisztSolver.local_dt(c) ;
				val I = Mat( 
					Vec(tmp,0.0,0.0,0.0,0.0), 
					Vec(0.0,tmp,0.0,0.0,0.0), 
					Vec(0.0,0.0,tmp,0.0,0.0), 
					Vec(0.0,0.0,0.0,tmp,0.0), 
					Vec(0.0,0.0,0.0,0.0,tmp));
				
				A(row,col) += I
				
				val rho = -rhs_rho(c) ;
				val rhou = -rhs_rhou(c) ;
				val rhoE = -rhs_rhoE(c) ;
				b(row) = Vec(rho, rhou(0), rhou(1), rhou(1), rhoE)
			}
			
			JoeLS.solve(A,x,b)
			Print("Solver residue: ", JoeLS.resid(), " after ", JoeLS.iters(), " iterations.")
			
			for(c <- cells(mesh)) {
				val col = JoeLS.col(c)
				rho(c) += x(col)(0) ;
				rhou(c)(0) += x(col)(1) ;
				rhou(c)(1) += x(col)(2) ;
				rhou(c)(2) += x(col)(3) ;
				rhoE(c) += x(col)(4) ;
			}
		}
		if (NavierStokesSolvers.EXPL_EULER == 1) {
			for( c <- cells(mesh) ) {
				val tmp = JoeLisztSolver.local_dt(c) / MeshGeometryCalc.cv_volume(c) ;
				rho(c) += tmp * rhs_rho(c) ;
				rhou(c) += tmp * rhs_rhou(c) ;
				rhoE(c) += tmp * rhs_rhoE(c) ;
			}
		}
	}

	def setBoundaryConditions() {

		var bc_err = false ;

		val vector_bc = Vec( IC.u_cbc.x, IC.u_cbc.y, IC.u_cbc.z, IC.T_init, IC.p_init )
		val T_bc = vector_bc(_3)
		val p_bc = vector_bc(_4)
		val u_bc = Vec( vector_bc.x, vector_bc.y, vector_bc.z )

		// Convective bondary conditions (farfield)
		for (f <- cbc) {
			CBCBoundary(f, T_bc, p_bc, u_bc)
		}
		for (f <- cbc) {
			JoeLisztSolver.ComputeBCProperties_T(f)
		}

		// Convective bondary conditions (subsonic outlet)
		for (f <- cbc_subsonic_outlet) {
			CBCSubsonicOutletBoundary(f, p_bc);
		}
		for (f <- cbc_subsonic_outlet) {
			JoeLisztSolver.ComputeBCProperties_T(f)
		}

		// Symmetry and Euler wall boundary conditions
		for (f <- symmetry) {
			SymmetryBoundary(f);
		}
		for (f <- symmetry) {
			JoeLisztSolver.ComputeBCProperties_T(f)
		}

		// Navier-Stokes wall boundary conditions
		for (f <- wall) {
			NonSlipBoundary(f);
		}
		for (f <- wall) {
			JoeLisztSolver.ComputeBCProperties_T(f)
		}

		// Set density for all the boundary conditions
		for (f <- zone_boundary) {
			setRhoFa(f)
		}

	}

	
	def showResidue( rhsResid: Vec[_5,Double], step: Int ) {
		var sCount = 0;
		var checkCount = 0;
		for (c <- cells(mesh)) {
			checkCount += size(faces(c))
		}
		if ( step % (IC.check_interval * 10 ) == 0 ) {
			Print( "		rho		rhou-X		rhou-Y		rhou-Z		rhoE		ScatterCounts") ;
		}
		Print( "RESID: ", step, "    ", rhsResid(_0), "    ", rhsResid(_1), "    ", rhsResid(_2), "    ", rhsResid(_3), "    ", rhsResid(_4)) ;
	}

	
	def CBCBoundary( f: Face, T_bc : Double, p_bc : Double, u_bc : Vec[_3,Double] ) {
		JoeLisztSolver.T_fa(f) = T_bc ;
		JoeLisztSolver.vel_fa(f) = u_bc ;
		JoeLisztSolver.p_fa(f) = p_bc ;
	}

	def CBCSubsonicOutletBoundary( f: Face, p_bc : Double ) {
		val icv0 = getInsideBoundaryCell(f) ;
		JoeLisztSolver.T_fa(f) = JoeLisztSolver.temp(icv0) ;
		JoeLisztSolver.vel_fa(f) = rhou(icv0) / rho(icv0) ;
		JoeLisztSolver.p_fa(f) = p_bc ;
	}

	def SymmetryBoundary( f: Face ) {
		val icv0 = getInsideBoundaryCell(f) ;
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		nVec = normalize(nVec) ;
		
		// flip u, APPROXIMATION ---> take velocity at the cell center
		val u0 = JoeLisztSolver.vel(icv0) ;
		val un = dot( nVec, u0 ) ;
		JoeLisztSolver.vel_fa(f) = u0 - 1.0 * un * nVec ;

		JoeLisztSolver.T_fa(f) = JoeLisztSolver.temp(icv0) ;
		JoeLisztSolver.p_fa(f) = JoeLisztSolver.press(icv0) ;
		JoeLisztSolver.RoM_fa(f) = JoeLisztSolver.RoM(icv0) ;
		JoeLisztSolver.gam_fa(f) = JoeLisztSolver.gamma(icv0) ;
		JoeLisztSolver.h_fa(f) = JoeLisztSolver.enthalpy(icv0) ;
	}

	def NonSlipBoundary( f: Face ) {
		val icv0 = getInsideBoundaryCell(f) ;
		
		// Null velocity at the face
		JoeLisztSolver.vel_fa(f) = Constants.float3_zero ;

		JoeLisztSolver.T_fa(f) = JoeLisztSolver.temp(icv0) ;
		JoeLisztSolver.p_fa(f) = JoeLisztSolver.press(icv0) ;
		JoeLisztSolver.RoM_fa(f) = JoeLisztSolver.RoM(icv0) ;
		JoeLisztSolver.gam_fa(f) = JoeLisztSolver.gamma(icv0) ;
		JoeLisztSolver.h_fa(f) = JoeLisztSolver.enthalpy(icv0) ;
	}

	def setRhoFa( f: Face ) {
		JoeLisztSolver.rho_fa(f) = JoeLisztSolver.p_fa(f) / ( JoeLisztSolver.RoM_fa(f) * JoeLisztSolver.T_fa(f) ) ;
	}


	def calcEntalphyGradient() {

		for( c <- cells(mesh) ) {
			JoeLisztSolver.enthalpy_grad(c) = Constants.float3_zero ;
		}

		for (f <- zone_interior) {
		    val icv0 = outside(f) ; val icv1 = inside(f) ;
		    val h0 = JoeLisztSolver.enthalpy(icv0) ; val h1 = JoeLisztSolver.enthalpy(icv1) ;
		    val h01 = 0.5*(h0 + h1) ;
		    var nVec = MeshGeometryCalc.fa_normal(f) ;

		    JoeLisztSolver.enthalpy_grad(icv0) += h01 * nVec
		    JoeLisztSolver.enthalpy_grad(icv1) -= h01 * nVec;
		}
		
		for (f <- zone_boundary) {
		    val icv0 = getInsideBoundaryCell(f) ;
		    val h0 = JoeLisztSolver.h_fa(f) ;
		    var nVec = MeshGeometryCalc.fa_normal(f) ;
		    JoeLisztSolver.enthalpy_grad(icv0) += h0 * nVec;

		}

		for( c <- cells(mesh) ) {
			JoeLisztSolver.enthalpy_grad(c) /= MeshGeometryCalc.cv_volume(c) ;
		}
	}

	def calcPressureGradient() {

		for( c <- cells(mesh) ) {
			JoeLisztSolver.press_grad(c) = Constants.float3_zero ;
		}

		for (f <- zone_interior) {
		    val icv0 = outside(f) ; val icv1 = inside(f) ;
		    val p0 = JoeLisztSolver.press(icv0) ; val p1 = JoeLisztSolver.press(icv1) ;
		    val p01 = 0.5*(p0 + p1) ;
		    var nVec = MeshGeometryCalc.fa_normal(f) ;

		    JoeLisztSolver.press_grad(icv0) += p01 * nVec
		    JoeLisztSolver.press_grad(icv1) -= p01 * nVec;
		}
		
		for (f <- zone_boundary) {
		    val icv0 = getInsideBoundaryCell(f) ;
		    val p0 = JoeLisztSolver.p_fa(f) ;
		    var nVec = MeshGeometryCalc.fa_normal(f) ;
		    JoeLisztSolver.press_grad(icv0) += p0 * nVec;

		}

		for( c <- cells(mesh) ) {
			JoeLisztSolver.press_grad(c) /= MeshGeometryCalc.cv_volume(c) ;
		}
	}

	def calcVelocityGradient() {

		for( c <- cells(mesh) ) {
		    JoeLisztSolver.vel_grad(c) =  Constants.float3x3_zero ;
		}

		for (f <- zone_interior) {
		    val icv0 = outside(f) ; val icv1 = inside(f) ;
		    val vel0 = JoeLisztSolver.vel(icv0) ; val vel1 = JoeLisztSolver.vel(icv1) ;
		    var vel01 = 0.5*(vel0 + vel1) ;

		    var nVec = MeshGeometryCalc.fa_normal(f) ;
		    
		    val update = Constants.float3x3_zero
		    var i = 0
		    while(i < 2) {
		    	var j = 0
		    	while(j < 2) {
		    		update(i,j) = vel01(i) * nVec(j)
		    		j += 1
		    	}
		    	i += 1
		    }
		    JoeLisztSolver.vel_grad(icv0) += update
		    JoeLisztSolver.vel_grad(icv1) -= update
		}

		for (f <- zone_boundary) {
		    val icv0 = getInsideBoundaryCell(f) ;
		    val vel0 = JoeLisztSolver.vel_fa(f) ;
		    var nVec = MeshGeometryCalc.fa_normal(f) ;
		    
		    val update = Constants.float3x3_zero
		    var i = 0
		    while(i < 2) {
		    	var j = 0
		    	while(j < 2) {
		    		update(i,j) = vel0(i) * nVec(j)
		    		j += 1
		    	}
		    	i += 1
		    }
		    JoeLisztSolver.vel_grad(icv0) += update
		}

		for( c <- cells(mesh) ) {
		    JoeLisztSolver.vel_grad(c) /=  MeshGeometryCalc.cv_volume(c)
		}
	}

	def calcConservGradient() {

		for( c <- cells(mesh) ) {
		     JoeLisztSolver.rho_grad(c) = Constants.float3_zero ;
		     JoeLisztSolver.rhou_grad(c) =  Constants.float3x3_zero ;
		     JoeLisztSolver.rhoE_grad(c) = Constants.float3_zero ;
		}
		
		for (f <- zone_interior) {
		    val icv0 = outside(f) ; val icv1 = inside(f) ;
		    val rho0 = JoeLiszt.rho(icv0) ; val rho1 = JoeLiszt.rho(icv1) ;
		    val rhou0 = JoeLiszt.rhou(icv0) ;  val rhou1 = JoeLiszt.rhou(icv1) ;
		    val rhoE0 = JoeLiszt.rhoE(icv0) ; val rhoE1 = JoeLiszt.rhoE(icv1) ;
		    val rho01 = 0.5*(rho0 + rho1) ;
		    var rhou01 = 0.5*(rhou0 + rhou1) ;
		    val rhoE01 = 0.5*(rhoE0 + rhoE1) ;
		    var nVec = MeshGeometryCalc.fa_normal(f) ;
		    val update = Constants.float3x3_zero
		    var i = 0
		    while(i < 2) {
		    	var j = 0
		    	while(j < 2) {
		    		update(i,j) = rhou01(i) * nVec(j)
		    		j += 1
		    	}
		    	i += 1
		    }
		    JoeLisztSolver.rho_grad(icv0) += rho01 * nVec
		    JoeLisztSolver.rho_grad(icv1) -= rho01 * nVec;
		    JoeLisztSolver.rhou_grad(icv0) += update
		    JoeLisztSolver.rhou_grad(icv1) -= update
		    JoeLisztSolver.rhoE_grad(icv0) += rhoE01 * nVec
		    JoeLisztSolver.rhoE_grad(icv1) -= rhoE01 * nVec;
		}

		for( c <- cells(mesh) ) {
		     JoeLisztSolver.rho_grad(c) /= MeshGeometryCalc.cv_volume(c) ;
		     JoeLisztSolver.rhou_grad(c) /=  MeshGeometryCalc.cv_volume(c) ;
		     JoeLisztSolver.rhoE_grad(c) /= MeshGeometryCalc.cv_volume(c) ;
		}

	}

	def calcRhsInteriorViscous( f: Face ) {

		val icv0 = outside(f) ;
		val icv1 = inside(f) ;

		val rho0 = rho(icv0) ;
		val u0 = JoeLisztSolver.vel(icv0) ;
		val grad_u0 = JoeLisztSolver.vel_grad(icv0)
		val h0 = JoeLisztSolver.enthalpy(icv0) ;
		val grad_h0 = JoeLisztSolver.enthalpy_grad(icv0) ;
		val t0 =  JoeLisztSolver.temp(icv0) ;
		val muTurb0 = 0.0 ;
		val R0 = JoeLisztSolver.RoM(icv0) ;
		val gam0 = JoeLisztSolver.gamma(icv0) ;
		val kine0 = 0.0 ;

		val rho1 = rho(icv1) ;
		val u1 = JoeLisztSolver.vel(icv1) ;
		val grad_u1 = JoeLisztSolver.vel_grad(icv1) ;
		val h1 = JoeLisztSolver.enthalpy(icv1) ;
		val grad_h1 = JoeLisztSolver.enthalpy_grad(icv1) ;
		val t1 = JoeLisztSolver.temp(icv1) ;
		val muTurb1 =  0.0 ;
		val R1 = JoeLisztSolver.RoM(icv1) ;
		val gam1 = JoeLisztSolver.gamma(icv1) ;
		val kine1 = 0.0 ;

		// face unit normal and area...
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = nVec/area ;
		var sVec = MeshGeometryCalc.x_cv(icv1) - MeshGeometryCalc.x_cv(icv0)
		val smag = sqrt(dot(sVec,sVec)) ;
		sVec = sVec/smag ;

		val Frho5 = JoeLisztSolver.calcNSFlux( rho0, u0, grad_u0, h0,  grad_h0, t0, R0, gam0, kine0,
		    	    				  rho1, u1, grad_u1, h1,  grad_h1, t1, R1, gam1, kine1,
							   JoeLisztSolver.mu_fa(f), 0.5*(muTurb0+muTurb1), JoeLisztSolver.lamOcp_fa(f), 
							   0.5*(kine0+kine1), 0.5*(u0+u1), area, nVec, smag, nVec) ;

		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;

		rhs_rhou(icv1) += Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv1) += Frho5(_4) ;

	}

	def calcJacInteriorViscous( f: Face ) {

		val icv0 = outside(f) ;
		val icv1 = inside(f) ;

		val rho0 = rho(icv0) ;
		val u0 = JoeLisztSolver.vel(icv0) ;
		val grad_u0 = JoeLisztSolver.vel_grad(icv0)
		val h0 = JoeLisztSolver.enthalpy(icv0) ;
		val grad_h0 = JoeLisztSolver.enthalpy_grad(icv0) ;
		val t0 =  JoeLisztSolver.temp(icv0) ;
		val muTurb0 = 0.0 ;
		val R0 = JoeLisztSolver.RoM(icv0) ;
		val gam0 = JoeLisztSolver.gamma(icv0) ;
		val kine0 = 0.0 ;

		val rho1 = rho(icv1) ;
		val u1 = JoeLisztSolver.vel(icv1) ;
		val grad_u1 = JoeLisztSolver.vel_grad(icv1) ;
		val h1 = JoeLisztSolver.enthalpy(icv1) ;
		val grad_h1 = JoeLisztSolver.enthalpy_grad(icv1) ;
		val t1 = JoeLisztSolver.temp(icv1) ;
		val muTurb1 =  0.0 ;
		val R1 = JoeLisztSolver.RoM(icv1) ;
		val gam1 = JoeLisztSolver.gamma(icv1) ;
		val kine1 = 0.0 ;

		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = nVec/area ;
		var sVec = MeshGeometryCalc.x_cv(icv1) - MeshGeometryCalc.x_cv(icv0)
		val smag = sqrt(dot(sVec,sVec)) ;
		sVec = sVec/smag ;

		// val A_L =  Constants.float5x5_zero ;
		// val A_R =  Constants.float5x5_zero ;

		JoeLisztSolver.calcNSJacobian( icv0, icv1, rho0, u0, grad_u0, h0,  grad_h0, t0, R0, gam0, kine0,
                                               rho1, u1, grad_u1, h1,  grad_h1, t1, R1, gam1, kine1,
                                               JoeLisztSolver.mu_fa(f), 0.5*(muTurb0+muTurb1), JoeLisztSolver.lamOcp_fa(f), 
                                               0.5*(kine0+kine1), 0.5*(u0+u1), area, nVec, smag, nVec ) ;



		// val row0 = JoeLS.row(icv0) ;
		// val col0 = JoeLS.col(icv0) ;
		// val row1 = JoeLS.row(icv1) ;
		// val col1 = JoeLS.col(icv1) ;
		// 
		// A(row0,col0) += A_L ; 
		// A(row0,col1) += A_R ;
		// A(row1,col0) -= A_R ; 
		// A(row1,col1) -= A_L ;

	}

	def calcRhsInteriorInviscid( f: Face ) {

		val icv0 = outside(f) ;
		val icv1 = inside(f) ;

		var rho0 = rho(icv0) ;
		var u0 = JoeLisztSolver.vel(icv0) ;
		var p0 = JoeLisztSolver.press(icv0) ;
		var h0 = JoeLisztSolver.enthalpy(icv0) ;
		val x0 = MeshGeometryCalc.x_cv(icv0)

		var rho1 = rho(icv1) ;
		var u1 = JoeLisztSolver.vel(icv1) ;
		var p1 = JoeLisztSolver.press(icv1) ;
		var h1 = JoeLisztSolver.enthalpy(icv1) ;
		val x1 = MeshGeometryCalc.x_cv(icv1) ;

		// face unit normal and area
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		var xfa = MeshGeometryCalc.x_fa(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = normalize(nVec) ;

		// Second order reconstruction
		var xzfa = xfa - x0 ;
		var xofa = xfa - x1 ;

		rho0 += dot(JoeLisztSolver.rho_grad(icv0), xzfa) ;
		rho1 += dot(JoeLisztSolver.rho_grad(icv1), xofa) ;

		p0 += dot(JoeLisztSolver.press_grad(icv0), xzfa) ;
		p1 += dot(JoeLisztSolver.press_grad(icv1), xofa) ;

		h0 += dot(JoeLisztSolver.enthalpy_grad(icv0), xzfa) ;
		h1 += dot(JoeLisztSolver.enthalpy_grad(icv1), xofa) ;

		var grad_rhou0 = JoeLisztSolver.rhou_grad(icv0)
		u0(0) += grad_rhou0(0,0) * xzfa(0) + grad_rhou0(0,1) * xzfa(1) + grad_rhou0(0,2) * xzfa(2) ;
		u0(1) += grad_rhou0(1,0) * xzfa(0) + grad_rhou0(1,1) * xzfa(1) + grad_rhou0(1,2) * xzfa(2) ;
		u0(2) += grad_rhou0(2,0) * xzfa(0) + grad_rhou0(2,1) * xzfa(1) + grad_rhou0(2,2) * xzfa(2) ;

		var grad_rhou1 = JoeLisztSolver.rhou_grad(icv1)
		u1(0) += grad_rhou1(0,0) * xofa(0) + grad_rhou1(0,1) * xofa(1) + grad_rhou1(0,2) * xofa(2) ;
		u1(1) += grad_rhou1(1,0) * xofa(0) + grad_rhou1(1,1) * xofa(1) + grad_rhou1(1,2) * xofa(2) ;
		u1(2) += grad_rhou1(2,0) * xofa(0) + grad_rhou1(2,1) * xofa(1) + grad_rhou1(2,2) * xofa(2) ;

		val kine0 = JoeLisztSolver.kine(icv0) ;
		val kine1 = JoeLisztSolver.kine(icv1) ;

		val gamma0 = JoeLisztSolver.gamma(icv0) ;
		val gamma1 = JoeLisztSolver.gamma(icv1) ;

		val Frho5 = JoeLisztSolver.calcEulerFlux_HLLC(rho0, u0, p0, h0, gamma0, rho1, u1, p1, h1, gamma1, area, nVec, 0.0, kine0, kine1) ;
		
		rhs_rho(icv0) -= Frho5(_0) ;
		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;

		rhs_rho(icv1) += Frho5(_0) ;
		rhs_rhou(icv1) += Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv1) += Frho5(_4) ;

	}

	def calcJacInteriorInviscid( f: Face ) {

		val icv0 = outside(f) ;
		val icv1 = inside(f) ;

		var rho0 = rho(icv0) ;
		var u0 = JoeLisztSolver.vel(icv0) ;
		var p0 = JoeLisztSolver.press(icv0) ;
		var h0 = JoeLisztSolver.enthalpy(icv0) ;

		var rho1 = rho(icv1) ;
		var u1 = JoeLisztSolver.vel(icv1) ;
		var p1 = JoeLisztSolver.press(icv1) ;
		var h1 = JoeLisztSolver.enthalpy(icv1) ;

		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = normalize(nVec) ;

		val gamma0 = JoeLisztSolver.gamma(icv0) ;
		val gamma1 = JoeLisztSolver.gamma(icv1) ;

		val T0 = JoeLisztSolver.temp(icv0) ;
		val T1 = JoeLisztSolver.temp(icv1) ;

		val R0 = JoeLisztSolver.RoM(icv0) ;
		val R1 = JoeLisztSolver.RoM(icv1) ;

		// val A_L =  Constants.float5x5_zero ;
		// val A_R =  Constants.float5x5_zero ;

		val surfVeloc = 0.0 ;

		JoeLisztSolver.calcEulerJacobian_HLLC( icv0, icv1, rho0, u0, p0, T0, h0, R0, gamma0, 
                                                                 rho1, u1, p1, T1, h1, R1, gamma1, 
                                                                 area, nVec, surfVeloc ) ;

		// val row0 = JoeLS.row(icv0) ;
		// val col0 = JoeLS.col(icv0) ;
		// val row1 = JoeLS.row(icv1) ;
		// val col1 = JoeLS.col(icv1) ;
		// 
		// A(row0,col0) += A_L ; 
		// A(row0,col1) += A_R ;
		// A(row1,col0) -= A_R ; 
		// A(row1,col1) -= A_L ;
	}

	def calcRhsSymmetry( f: Face) {

		val icv0 = getInsideBoundaryCell(f) ;

		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = nVec/area ;
	
		val kine0 = 0.0 ;
		val kine1 = 0.0 ;

		val Frho5 = JoeLisztSolver.calcEulerFlux_HLLC(JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), JoeLisztSolver.p_fa(f), 
		    	    					 JoeLisztSolver.h_fa(f), JoeLisztSolver.gam_fa(f), JoeLisztSolver.rho_fa(f), 
								 JoeLisztSolver.vel_fa(f), JoeLisztSolver.p_fa(f), JoeLisztSolver.h_fa(f), 
								 JoeLisztSolver.gam_fa(f), area, nVec, 0.0, kine0, kine1) ;

		rhs_rho(icv0) -= Frho5(_0) ;
		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;


		// val A_L =  Constants.float5x5_zero ;
		// val A_R =  Constants.float5x5_zero ;

		JoeLisztSolver.calcEulerJacobian_HLLC( icv0, icv0, JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), JoeLisztSolver.p_fa(f), 
						       JoeLisztSolver.T_fa(f), JoeLisztSolver.h_fa(f), JoeLisztSolver.RoM_fa(f), JoeLisztSolver.gam_fa(f), 
                                                       JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), JoeLisztSolver.p_fa(f), 
						       JoeLisztSolver.T_fa(f), JoeLisztSolver.h_fa(f), JoeLisztSolver.RoM_fa(f), JoeLisztSolver.gam_fa(f), 
                                                       area, nVec, 0.0 ) ;
		
		// val row0 = JoeLS.row(icv0) ;
		// val col0 = JoeLS.col(icv0) ;		
		// A(row0,col0) += A_L ; 
	}

	def calcRhsWall( f: Face) {

		val icv0 = getInsideBoundaryCell(f) ;
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = nVec/area ;

		val Frho5 = JoeLisztSolver.calcEulerFlux_HLLC(JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), JoeLisztSolver.p_fa(f), 
		    	    					 JoeLisztSolver.h_fa(f), JoeLisztSolver.gam_fa(f), JoeLisztSolver.rho_fa(f), 
								 JoeLisztSolver.vel_fa(f), JoeLisztSolver.p_fa(f), JoeLisztSolver.h_fa(f), 
								 JoeLisztSolver.gam_fa(f), area, nVec, 0.0, 0.0, 0.0) ;

		rhs_rho(icv0) -= Frho5(_0) ;
		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;

		// var A_L =  Constants.float5x5_zero ;
		// var A_R =  Constants.float5x5_zero ;

		JoeLisztSolver.calcEulerJacobian_HLLC( icv0, icv0, JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), JoeLisztSolver.p_fa(f), 
						       JoeLisztSolver.T_fa(f), JoeLisztSolver.h_fa(f), JoeLisztSolver.RoM_fa(f), JoeLisztSolver.gam_fa(f), 
                                                       JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), JoeLisztSolver.p_fa(f), 
						       JoeLisztSolver.T_fa(f), JoeLisztSolver.h_fa(f), JoeLisztSolver.RoM_fa(f), JoeLisztSolver.gam_fa(f), 
                                                       area, nVec, 0.0 ) ;


		// val row0 = JoeLS.row(icv0) ;
		// val col0 = JoeLS.col(icv0) ;
		// 
		// A(row0,col0) += A_L ; 

		val grad_u0 = JoeLisztSolver.vel_grad(icv0)
		val grad_h0 = JoeLisztSolver.enthalpy_grad(icv0) ;
		val muLam0 = JoeLisztSolver.muLam(icv0) ;
		val muTurb0 = 0.0 ;
		val kine0 = 0.0 ;
		val kine1 = 0.0 ;
		val kine_fa = 0.0;

		val sVec = MeshGeometryCalc.x_fa(f) - MeshGeometryCalc.x_cv(icv0)
		val smag = sqrt(dot(sVec,sVec)) ;

		val Frho5_ = JoeLisztSolver.calcNSFlux(  rho(icv0), JoeLisztSolver.vel(icv0), grad_u0, JoeLisztSolver.enthalpy(icv0),  
		    	    				  grad_h0,  JoeLisztSolver.temp(icv0), JoeLisztSolver.RoM(icv0), JoeLisztSolver.gamma(icv0), kine0,
		    	    				  JoeLisztSolver.rho_fa(f),  JoeLisztSolver.vel_fa(f), grad_u0, JoeLisztSolver.h_fa(f),  
							  grad_h0, JoeLisztSolver.T_fa(f), JoeLisztSolver.RoM_fa(f), JoeLisztSolver.gam_fa(f), kine1,
							  JoeLisztSolver.mu_fa(f), .0, JoeLisztSolver.lamOcp_fa(f), kine_fa,  JoeLisztSolver.vel_fa(f), area, nVec, smag, nVec) ;

		rhs_rhou(icv0) -= Vec(Frho5_(_1), Frho5_(_2), Frho5_(_3)) ;
		rhs_rhoE(icv0) -= Frho5_(_4) ;

		// A_L =  Constants.float5x5_zero ;
		// A_R =  Constants.float5x5_zero ;

		JoeLisztSolver.calcNSJacobian( icv0, icv0, rho(icv0), JoeLisztSolver.vel(icv0), grad_u0, JoeLisztSolver.enthalpy(icv0),  grad_h0, JoeLisztSolver.temp(icv0), 
					       JoeLisztSolver.RoM(icv0), JoeLisztSolver.gamma(icv0), kine0, JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), grad_u0, JoeLisztSolver.h_fa(f),  
					       grad_h0, JoeLisztSolver.T_fa(f), JoeLisztSolver.RoM_fa(f), JoeLisztSolver.gam_fa(f), kine1, JoeLisztSolver.mu_fa(f), 0.0, JoeLisztSolver.lamOcp_fa(f), 
                                               kine_fa, JoeLisztSolver.vel_fa(f), area, nVec, smag, nVec ) ;

		// A(row0,col0) += A_L ;

	}

	def calcRhsOtherBoundaries( f: Face ) {

		val icv0 = getInsideBoundaryCell(f) ;

		val rho0 = rho(icv0) ;
		val u0 = JoeLisztSolver.vel(icv0) ;
		val p0 = JoeLisztSolver.press(icv0) ;
		val h0 = JoeLisztSolver.enthalpy(icv0) ;
		val t0 = JoeLisztSolver.temp(icv0) ;
		val gam0 = JoeLisztSolver.gamma(icv0) ;
		val RoM0 = JoeLisztSolver.RoM(icv0) ;

		// face unit normal and area...
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = normalize(nVec) ;
	
		val kine0 = 0.0 ;
		val kine1 = 0.0 ;

		val Frho5 = JoeLisztSolver.calcEulerFlux_HLLC(rho0, u0, p0, h0, gam0, JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), 
		    	    			              JoeLisztSolver.p_fa(f), JoeLisztSolver.h_fa(f), JoeLisztSolver.gam_fa(f), area, nVec, 0.0, kine0, kine1) ;
		


		rhs_rho(icv0) -= Frho5(_0) ;
		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;


		// var A_L =  Constants.float5x5_zero ;
		// var A_R =  Constants.float5x5_zero ;

		JoeLisztSolver.calcEulerJacobian_HLLC( icv0, icv0, rho0,u0, p0, 
						       t0, h0, RoM0, gam0, 
                                                       JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), JoeLisztSolver.p_fa(f), 
						       JoeLisztSolver.T_fa(f), JoeLisztSolver.h_fa(f), JoeLisztSolver.RoM_fa(f), JoeLisztSolver.gam_fa(f), 
                                                       area, nVec, 0.0 ) ;

		// val row0 = JoeLS.row(icv0) ;
		// val col0 = JoeLS.col(icv0) ;
		// A(row0,col0) += A_L ; 


		if (IC.mu_ref > 0.0) {
		   val grad_u0 = JoeLisztSolver.vel_grad(icv0)
		   val grad_h0 = JoeLisztSolver.enthalpy_grad(icv0) ;
		   val muLam0 = JoeLisztSolver.muLam(icv0) ;
		   val muTurb0 = 0.0 ;
		   val kine_fa = 0.0;

		   var sVec = MeshGeometryCalc.x_fa(f) - MeshGeometryCalc.x_cv(icv0)
		   val smag = sqrt(dot(sVec,sVec)) ;
		   sVec /= smag

		   val Frho5_ = JoeLisztSolver.calcNSFlux(  rho(icv0), JoeLisztSolver.vel(icv0), grad_u0, JoeLisztSolver.enthalpy(icv0),  
		    	    				  grad_h0,  JoeLisztSolver.temp(icv0), JoeLisztSolver.RoM(icv0), JoeLisztSolver.gamma(icv0), kine0,
		    	    				  JoeLisztSolver.rho_fa(f),  JoeLisztSolver.vel_fa(f), grad_u0, JoeLisztSolver.h_fa(f),  
							  grad_h0, JoeLisztSolver.T_fa(f), JoeLisztSolver.RoM_fa(f), JoeLisztSolver.gam_fa(f), kine1,
							  JoeLisztSolver.mu_fa(f), 0.0, JoeLisztSolver.lamOcp_fa(f), kine_fa,  JoeLisztSolver.vel_fa(f), area, nVec, smag, sVec) ;

		   rhs_rhou(icv0) -= Vec(Frho5_(_1), Frho5_(_2), Frho5_(_3)) ;
		   rhs_rhoE(icv0) -= Frho5_(_4) ;

		   // A_L =  Constants.float5x5_zero ;
		   // A_R =  Constants.float5x5_zero ;

		   JoeLisztSolver.calcNSJacobian( icv0, icv0, rho(icv0), JoeLisztSolver.vel(icv0), grad_u0, JoeLisztSolver.enthalpy(icv0),  grad_h0, JoeLisztSolver.temp(icv0), JoeLisztSolver.RoM(icv0), 
		   				  JoeLisztSolver.gamma(icv0), kine0, JoeLisztSolver.rho_fa(f), JoeLisztSolver.vel_fa(f), grad_u0, JoeLisztSolver.h_fa(f),  grad_h0, JoeLisztSolver.T_fa(f), 
						  JoeLisztSolver.RoM_fa(f), JoeLisztSolver.gam_fa(f), kine_fa, JoeLisztSolver.mu_fa(f), .0, JoeLisztSolver.lamOcp_fa(f),  kine_fa, JoeLisztSolver.vel_fa(f), 
						  area, nVec, smag, nVec ) ;
						
		  // A(row0,col0) += A_L ; 
		
		}

	}
}
