import Liszt.Language._
import Liszt.MetaInteger._

@lisztcode
object JoeLisztSolver {
	
	// Primitive variables
	val vel = FieldWithConst[Cell,Vec[_3,Double]](Constants.float3_zero) ;
	val press = FieldWithConst[Cell,Double](0.0) ;
	val temp = FieldWithConst[Cell,Double](0.0) ;
	val enthalpy = FieldWithConst[Cell,Double](0.0) ;
	val muLam = FieldWithConst[Cell,Double](0.0) ;
	val muTurb = FieldWithConst[Cell,Double](0.0) ;
	val LambdaOverCp = FieldWithConst[Cell,Double](0.0) ;
	val sos = FieldWithConst[Cell,Double](0.0) ;
	val kine = FieldWithConst[Cell,Double](0.0) ;
	val local_dt = FieldWithConst[Cell,Double](0.0) ;
	val RoM = FieldWithConst[Cell,Double](IC.R_gas) ;
	val gamma = FieldWithConst[Cell,Double](IC.GAMMA) ;

	// Gradients for viscous computation, and second order interpolation
	val vel_grad = FieldWithConst[Cell,Mat[_3,_3,Double]](Constants.float3x3_zero) ;
	val enthalpy_grad = FieldWithConst[Cell,Vec[_3,Double]](Constants.float3_zero) ;
	val press_grad = FieldWithConst[Cell,Vec[_3,Double]](Constants.float3_zero) ;
	val rho_grad = FieldWithConst[Cell,Vec[_3,Double]](Constants.float3_zero) ;
	val rhou_grad = FieldWithConst[Cell,Mat[_3,_3,Double]](Constants.float3x3_zero) ;
	val rhoE_grad = FieldWithConst[Cell,Vec[_3,Double]](Constants.float3_zero) ;

	// this should be called in runExplicitBackwardEuler
	val gam_fa = FieldWithConst[Face,Double](0.0) ;
	val RoM_fa = FieldWithConst[Face,Double](0.0) ;
	val h_fa = FieldWithConst[Face,Double](0.0) ;
	val T_fa = FieldWithConst[Face,Double](0.0) ;
	val vel_fa = FieldWithConst[Face,Vec[_3,Double]](Constants.float3_zero) ;
	val p_fa = FieldWithConst[Face,Double](0.0) ;
	val mu_fa = FieldWithConst[Face,Double](0.0) ;
	val lamOcp_fa = FieldWithConst[Face,Double](0.0) ;
	val rho_fa = FieldWithConst[Face,Double](0.0) ;


	def calcStateMaterialProp() {
		for ( c <- cells(mesh) ) {
			if ( JoeLiszt.rho(c) <= 0.0 ) {
				Print( "negative density at xcv: " , MeshGeometryCalc.x_cv(c) , " " , ID(c) ) ;
			}

			vel(c) = JoeLiszt.rhou(c) / JoeLiszt.rho(c) ;

			var kinecv = 0.0 ;
			kinecv = kine(c) ;

			val pr = ( gamma(c) - 1.0 ) * ( JoeLiszt.rhoE(c) - 0.5 * dot(JoeLiszt.rhou(c),JoeLiszt.rhou(c)) / 
			  JoeLiszt.rho(c) - JoeLiszt.rho(c) * kinecv ) ;
			if ( pr <= 0.0 ) Print( "negative pressure at xcv: " , MeshGeometryCalc.x_cv(c), " ", ID(c) ) ;
			else press(c) = pr ;

			val tp = pr / ( JoeLiszt.rho(c) * RoM(c) ) ;
			temp(c) = tp ;
			enthalpy(c) = gamma(c) * RoM(c) / ( gamma(c) - 1.0 ) * tp ;
			sos(c) = sqrt( gamma(c) * pr / JoeLiszt.rho(c) ) ;
		}
		for ( f <- faces(mesh) ) {
			if (IC.mu_ref > 0.0) {
				val mu = IC.mu_ref ; // IC.mu_ref*powf(tp/IC.SL_Tref, 1.5)*(IC.SL_Tref + IC.SL_Sref)/(tp + IC.SL_Sref);
				mu_fa(f) = mu;
				lamOcp_fa(f) = mu / IC.Pr;
			}
		}
	} 

	def ComputeBCProperties_T( f: Face ) {
		gam_fa(f) = IC.GAMMA ;
		RoM_fa(f) = IC.R_gas ;
		h_fa(f) = IC.GAMMA * IC.R_gas / ( IC.GAMMA - 1.0 ) * T_fa(f) ;
		if (IC.mu_ref > 0.0) {
		   mu_fa(f) = IC.mu_ref ; // IC.mu_ref*powf(T_fa(f)/IC.SL_Tref, 1.5)*(IC.SL_Tref + IC.SL_Sref)/(T_fa(f) + IC.SL_Sref);
		   lamOcp_fa(f) = IC.mu_ref / IC.Pr;
		}
	}
	

	def ComputeGeometry(){

		for( f <- faces(mesh) ){ MeshGeometryCalc.calcFaceGeom(f, true) ; }

//		Only for Hyshot_fine grid
		for (f <- JoeLiszt.symmetry_pos) {  MeshGeometryCalc.calcFaceGeom(f, false) ; }

		for( c <- cells(mesh) ){
			MeshGeometryCalc.calcCellGeom(c) ;
		}  
	}

	def init() {
		Print("JoeLisztSolver()") ;

		// ----------------------------------------------------------------------------------------
		// write some parameters on the screen
		// ----------------------------------------------------------------------------------------

		Print("") ;
		Print("--------------------------------------------") ;
		Print( "Gas properties         " ) ;
		Print("    GAMMA            : ", IC.GAMMA );
		Print("    R_GAS            : ", IC.R_gas );
		Print("    P_REF            : ", IC.p_init );
		Print("    RHO_REF          : ", IC.rho_init );
		Print("    T_REF            : ", IC.T_init );
		Print("    MU_REF            : ", IC.mu_ref );
		Print("    SOS_REF          : ", sqrt(IC.GAMMA*IC.R_gas*IC.T_init) );
		Print("Solver settings        ");
		Print("    nsteps           : ", Constants.nsteps);
		Print("    timeStepMode     : ", IC.timeStepMode);
		Print("--------------------------------------------") ;
		Print("") ;
		ComputeGeometry() ;

		for ( c <- cells(mesh) ) {
			gamma(c) = IC.GAMMA ;
			RoM(c) = IC.R_gas ;

			JoeLiszt.rho(c) = IC.rho_init ;
			JoeLiszt.rhou(c) = IC.rho_init * IC.u_init ;
			JoeLiszt.rhoE(c) = IC.p_init / (IC.GAMMA - 1.0)  +  0.5 * IC.rho_init * dot( IC.u_init, IC.u_init ) ;
		}	
	}


	def calcEulerFlux_HLLC( rhoL : Double, uL : Vec[_3,Double], pL : Double, h0 : Double, gammaL : Double, rhoR : Double, uR : Vec[_3,Double], pR : Double, h1 : Double, gammaR : Double, area : Double, nVec : Vec[_3,Double], surfVeloc : Double, kL : Double, kR : Double ) : Vec[_5,Double] = {
		
		var Frho = 0.0 ;
		var Frhou = Constants.float3_zero ;
		var FrhoE = 0.0 ;				
		val unL = dot(uL,nVec) ;
		val uLuL = dot(uL,uL) ;
		val cL = sqrt( gammaL * pL / rhoL ) ;
		val hL = gammaL / ( gammaL - 1.0 ) * pL / rhoL + 0.5 * uLuL + kL ;
		val eL = hL * rhoL - pL ;

		val unR = dot(uR,nVec) ;
		val uRuR = dot(uR,uR) ;
		val cR = sqrt( gammaR * pR / rhoR ) ;
		val hR = gammaR / ( gammaR - 1.0 ) * pR / rhoR + 0.5 * uRuR + kR ;
		val eR = hR * rhoR - pR ;

		// Roe's averaging
		val Rrho = sqrt( rhoR / rhoL ) ;
		val tmp = 1.0 / ( 1.0 + Rrho ) ;
		val velRoe = tmp * ( uL + uR *Rrho ) ;
		val uRoe = dot( velRoe, nVec ) ;
		//val hRoe = tmp * ( hL + hR * Rrho ) ;
		
		//val cRoe = sqrt( (gammaL - 1.0) * ( hRoe - 0.5 * dot( velRoe, velRoe ) ) ) ;
		val gamPdivRho = tmp * ( (gammaL * pL / rhoL + 0.5 * (gammaL - 1.0) * uLuL ) + (gammaR * pR / rhoR + 0.5 * (gammaR - 1.0) * uRuR ) * Rrho ) ;
		val cRoe = sqrt( gamPdivRho - ((gammaL + gammaR) * 0.5 - 1.0 ) * 0.5 * dot( velRoe, velRoe ) );

		// speed of sound at L and R 
		val sL = (uRoe - cRoe).min(unL - cL) ;
		val sR = (uRoe + cRoe).max(unR + cR) ;

		// speed of contact surface
		val sM = (pL - pR - rhoL * unL * (sL - unL) + rhoR * unR * (sR - unR) ) / (rhoR * (sR - unR) - rhoL * (sL - unL)) ;

		// pressure at right and left (pR =pL ) side contact surfaca
		val pStar = rhoR * (unR - sR) * (unR - sM) + pR ;

		if (sM >= 0.0) {
			if (sL > 0.0) {
				Frho = rhoL * unL ;
				Frhou = rhoL * unL * uL + pL * nVec ;
				FrhoE = (eL + pL) * unL ;
			} else {
				val invSLmSs = 1.0 / (sL - sM ) ;
				val sLmuL = sL - unL ;
				val rhoSL = rhoL * sLmuL * invSLmSs ;
				val rhouSL = (( rhoL * sLmuL) * uL + (pStar - pL) * nVec) * invSLmSs ;
				val eSL = ( sLmuL * eL - pL * unL + pStar * sM ) * invSLmSs ;

				Frho = rhoSL * sM ;
				Frhou = rhouSL * sM + pStar * nVec ;
				FrhoE = ( eSL + pStar ) * sM ; 
			}
		} else {
			if ( sR >= 0.0 ) {
				val invSRmSs = 1.0 / (sR - sM ) ;
				val sRmuR = sR - unR ;
				val rhoSR = rhoR * sRmuR * invSRmSs ;
				val rhouSR = (( rhoR * sRmuR) * uR + (pStar - pR) * nVec) * invSRmSs ;
				val eSR = ( sRmuR * eR - pR * unR + pStar * sM ) * invSRmSs ;

				Frho = rhoSR * sM ;
				Frhou = rhouSR * sM + pStar * nVec ;
				FrhoE = ( eSR + pStar ) * sM ;
			} else {
				Frho = rhoR * unR ;
				Frhou = rhoR * unR * uR + pR * nVec ;
				FrhoE = (eR + pR) * unR ;
			}
		}

		Frho *= area ;
		Frhou *= area ;
		FrhoE *= area ;
		
		return Vec(Frho, Frhou.x, Frhou.y, Frhou.z, FrhoE) ;
		
	}

	def calcEulerJacobian_HLLC( icv0 : Cell, icv1 : Cell,  
	    rhoL : Double, uL : Vec[_3,Double], pL : Double, TL : Double, h0 : Double, RL : Double, gammaL : Double, 
	    rhoR : Double, uR : Vec[_3,Double], pR : Double, TR : Double, h1 : Double, RR : Double, gammaR : Double, 
	    area : Double, nVec : Vec[_3,Double], surfVeloc : Double ) {
		
		val unL = dot(uL,nVec) ;
		val uLuL = dot(uL,uL) ;
		val cL = sqrt( gammaL * pL / rhoL ) ;
		val hL = gammaL / ( gammaL - 1.0 ) * pL / rhoL + 0.5 * uLuL ;
		val eL = hL * rhoL - pL ;

		val unR = dot(uR,nVec) ;
		val uRuR = dot(uR,uR) ;
		val cR = sqrt( gammaR * pR / rhoR ) ;
		val hR = gammaR / ( gammaR - 1.0 ) * pR / rhoR + 0.5 * uRuR ;
		val eR = hR * rhoR - pR ;

		// Roe's averaging
		val Rrho = sqrt( rhoR / rhoL ) ;
		val tmp = 1.0 / ( 1.0 + Rrho ) ;
		val velRoe = tmp * ( uL + uR *Rrho ) ;
		val uRoe = dot( velRoe, nVec ) ;
		
		val gamPdivRho = tmp * ( (gammaL * pL / rhoL + 0.5 * (gammaL - 1.0) * uLuL ) + (gammaR * pR / rhoR + 0.5 * (gammaR - 1.0) * uRuR ) * Rrho ) ;
		val cRoe = sqrt( gamPdivRho - ((gammaL + gammaR) * 0.5 - 1.0 ) * 0.5 * dot( velRoe, velRoe ) );

		// speed of sound at L and R 
		val sL = (uRoe - cRoe).min(unL - cL) ;
		val sR = (uRoe + cRoe).max(unR + cR) ;

		// speed of contact surface
		val sM = (pL - pR - rhoL * unL * (sL - unL) + rhoR * unR * (sR - unR) ) / (rhoR * (sR - unR) - rhoL * (sL - unL)) ;

		// pressure at right and left (pR =pL ) side contact surfaca
		val pStar = rhoR * (unR - sR) * (unR - sM) + pR ;
		
		var A_L = Constants.float5x5_zero;
		var A_R = Constants.float5x5_zero;
		
		if (sM >= 0.0) {
			
    		if (sL > 0.0) { // Shock wave and contact durface in the same directions
				
				val nVecArea = nVec*area
				// Left Jacobian
				A_L = calcJacobianA(uL, pL, rhoL, nVecArea, 0.5*(gammaL+gammaR), Vec(0.0,0.0,0.0))
				// Right Jacobian
				A_R =  Constants.float5x5_zero;
			
			} else { // Shock wave and contact durface in opposite directions
				val invSLmSs = 1.0 / (sL-sM)
				val sLmuL = sL - unL
				val rhoSL = rhoL * sLmuL * invSLmSs
				val rhouSL = (rhoL*uL*sLmuL+(pStar-pL)*nVec)*invSLmSs;
				val eSL = (sLmuL*eL-pL*unL+pStar*sM)*invSLmSs;
				val gammaLM1 = gammaL-1.0;
				val gammaRM1 = gammaR-1.0;
				val invrhotld = 1.0 / (rhoR*(sR-unR)-rhoL*(sL-unL));

				var dSMdUL =  Constants.float5_zero;
				var dSMdUR =  Constants.float5_zero;
				var dpsdUL =  Constants.float5_zero;
				var dpsdUR =  Constants.float5_zero;
				      
				// Left transformation matrices
				dSMdUL(_0) = -unL*unL + uLuL*gammaLM1/2.0 + sM*sL;
				dSMdUL(_1) =  nVec(_0)*(2.0*unL-sL-sM) - gammaLM1*uL(_0);
				dSMdUL(_2) =  nVec(_1)*(2.0*unL-sL-sM) - gammaLM1*uL(_1);
				dSMdUL(_3) =  nVec(_2)*(2.0*unL-sL-sM) - gammaLM1*uL(_2);
				dSMdUL(_4) =  gammaLM1;

				dSMdUL *= invrhotld;
				dpsdUL = rhoR*(sR-unR)*dSMdUL;

				// Right transformation matrices
				dSMdUR(_0) =  unR*unR - uRuR*gammaRM1/2.0 - sM*sR;
				dSMdUR(_1) = -nVec(_0)*(2.0*unR-sR-sM) + gammaRM1*uR(_0);
				dSMdUR(_2) = -nVec(_1)*(2.0*unR-sR-sM) + gammaRM1*uR(_1);
				dSMdUR(_3) = -nVec(_2)*(2.0*unR-sR-sM) + gammaRM1*uR(_2);
				dSMdUR(_4) = -gammaRM1;
				
				dSMdUR *= invrhotld;
				dpsdUR = rhoL*(sL-unL)*dSMdUR;
				
				A_L = calcSubSonicJacobeanHLLC_L(rhoL, uL, pL, eL, unL, uLuL, sL, 
					rhoSL, rhouSL, eSL, dSMdUL, dSMdUR, dpsdUL, dpsdUR, sM, pStar, 0.5*(gammaL+gammaR), nVec);
				A_R = calcSubSonicJacobeanHLLC_R(rhoL, uL, pL, eL, unL, uLuL, sL, 
					rhoSL, rhouSL, eSL, dSMdUL, dSMdUR, dpsdUL, dpsdUR, sM, pStar, 0.5*(gammaL+gammaR), nVec);

				// Left jacobian matrices
				var i = 0;
				while(i < 5) {
					A_L(0,i) =  A_L(0,i)*sM + dSMdUL(i)*rhoSL;
					A_L(1,i) =  A_L(1,i)*sM + dSMdUL(i)*rhouSL(0) + dpsdUL(i)*nVec(0);
					A_L(2,i) =  A_L(2,i)*sM + dSMdUL(i)*rhouSL(1) + dpsdUL(i)*nVec(1);
					A_L(3,i) =  A_L(3,i)*sM + dSMdUL(i)*rhouSL(2) + dpsdUL(i)*nVec(2);
					A_L(4,i) = (A_L(4,i)+dpsdUL(i))*sM + (eSL+pStar)*dSMdUL(i);
					i += 1
				}
				A_L *= area;
				
				// Right jacobian matrices
				i = 0;
				while(i < 5) {
					A_R(0,i) =  A_R(0,i)*sM + dSMdUR(i)*rhoSL;
					A_R(1,i) =  A_R(1,i)*sM + dSMdUR(i)*rhouSL(0) + dpsdUR(i)*nVec(0);
					A_R(2,i) =  A_R(2,i)*sM + dSMdUR(i)*rhouSL(1) + dpsdUR(i)*nVec(1);
					A_R(3,i) =  A_R(3,i)*sM + dSMdUR(i)*rhouSL(2) + dpsdUR(i)*nVec(2);
					A_R(4,i) = (A_R(4,i)+dpsdUR(i))*sM + (eSL+pStar)*dSMdUR(i);
					i += 1
				}
				A_R *= area;
			}
			
		} else {
			
			if (sR >= 0.0) {
				val invSRmSs = 1.0/(sR-sM);
				val sRmuR = sR-unR;
				val rhoSR = rhoR*sRmuR*invSRmSs;
				val rhouSR = (rhoR*uR*sRmuR+(pStar-pR)*nVec)*invSRmSs;
				val eSR = (sRmuR*eR-pR*unR+pStar*sM)*invSRmSs;      
				val gammaLM1 = (gammaL-1.0);
				val gammaRM1 = (gammaR-1.0);
				val invrhotld = 1.0/(rhoR*(sR-unR)-rhoL*(sL-unL));

				var dSMdUL =  Constants.float5_zero;
				var dSMdUR =  Constants.float5_zero;
				var dpsdUL =  Constants.float5_zero;
				var dpsdUR =  Constants.float5_zero;

				dSMdUL(0) = -unL*unL + uLuL*gammaLM1/2.0 + sM*sL;
				dSMdUL(1) =  nVec(0)*(2.0*unL-sL-sM) - gammaLM1*uL(0);
				dSMdUL(2) =  nVec(1)*(2.0*unL-sL-sM) - gammaLM1*uL(1);
				dSMdUL(3) =  nVec(2)*(2.0*unL-sL-sM) - gammaLM1*uL(2);
				dSMdUL(4) =  gammaLM1;

				dSMdUL *= invrhotld;
				dpsdUL = rhoR*(sR-unR)*dSMdUL;

				dSMdUR(0) =  unR*unR - uRuR*gammaRM1/2.0 - sM*sR;
				dSMdUR(1) = -nVec(0)*(2.0*unR-sR-sM) + gammaRM1*uR(0);
				dSMdUR(2) = -nVec(1)*(2.0*unR-sR-sM) + gammaRM1*uR(1);
				dSMdUR(3) = -nVec(2)*(2.0*unR-sR-sM) + gammaRM1*uR(2);
				dSMdUR(4) = -gammaRM1;

				dSMdUR *= invrhotld;
				dpsdUR = rhoL*(sL-unL)*dSMdUR;

				A_L = calcSubSonicJacobeanHLLC_L(rhoR, uR, pR, eR, unR, uRuR, sR,
					rhoSR, rhouSR, eSR, dSMdUR, dSMdUL, dpsdUR, dpsdUL, sM, pStar, 0.5*(gammaL+gammaR), nVec);
				
				A_R = calcSubSonicJacobeanHLLC_R(rhoR, uR, pR, eR, unR, uRuR, sR,
					rhoSR, rhouSR, eSR, dSMdUR, dSMdUL, dpsdUR, dpsdUL, sM, pStar, 0.5*(gammaL+gammaR), nVec);

				var i = 0;
				while(i < 5) {
					A_L(0,i) =  A_L(0,i)*sM + dSMdUL(i)*rhoSR;
					A_L(1,i) =  A_L(1,i)*sM + dSMdUL(i)*rhouSR(0) + dpsdUL(i)*nVec(0);
					A_L(2,i) =  A_L(2,i)*sM + dSMdUL(i)*rhouSR(1) + dpsdUL(i)*nVec(1);
					A_L(3,i) =  A_L(3,i)*sM + dSMdUL(i)*rhouSR(2) + dpsdUL(i)*nVec(2);
					A_L(4,i) = (A_L(4,i)+dpsdUL(i))*sM + (eSR+pStar)*dSMdUL(i);
					i += 1
				}
				A_L *= area;

				i = 0;
				while(i < 5) {
					A_R(0,i) =  A_R(0,i)*sM + dSMdUR(i)*rhoSR;
					A_R(1,i) =  A_R(1,i)*sM + dSMdUR(i)*rhouSR(0) + dpsdUR(i)*nVec(0);
					A_R(2,i) =  A_R(2,i)*sM + dSMdUR(i)*rhouSR(1) + dpsdUR(i)*nVec(1);
					A_R(3,i) =  A_R(3,i)*sM + dSMdUR(i)*rhouSR(2) + dpsdUR(i)*nVec(2);
					A_R(4,i) = (A_R(4,i)+dpsdUR(i))*sM + (eSR+pStar)*dSMdUR(i);
				}
				A_R *= area;
			} else {
				
				val nVecArea = nVec*area
				// Right Jacobian
				A_R = calcJacobianA(uR, pR, rhoR, nVecArea, 0.5*(gammaL+gammaR), Vec(0.0,0.0,0.0));
				// Left Jacobian
				A_L =  Constants.float5x5_zero;
			
			}
			
		}
		
		val row0 = JoeLS.row(icv0) ;
		val col0 = JoeLS.col(icv0) ;
		JoeLiszt.A(row0,col0) += A_L ; 
		if (ID(icv0) != ID(icv1)) {
			val row1 = JoeLS.row(icv1) ;
			val col1 = JoeLS.col(icv1) ;
			JoeLiszt.A(row0,col1) += A_R ;
			JoeLiszt.A(row1,col0) -= A_R ; 
			JoeLiszt.A(row1,col1) -= A_L ;
		}
		
	}

	def calcSubSonicJacobeanHLLC_L(rhoL : Double, uL : Vec[_3,Double], pL : Double, eL : Double, qL : Double, 
		  psiL : Double, SL : Double, rhoSL : Double, rhouSL : Vec[_3,Double], eSL : Double, dSMdUL : Vec[_5,Double], dSMdUR : Vec[_5,Double], dpsdUL : Vec[_5,Double], dpsdUR : Vec[_5,Double],
		  SM : Double, pS : Double, gamma : Double, nV : Vec[_3,Double] ) : Mat[_5,_5,Double] = {

		var A_L = Constants.float5x5_zero;
		
		val gammaMinus1 = (gamma-1.0);
		val omL = 1.0/(SL-SM);

		A_L(0,0) =  SL    + dSMdUL(0)*rhoSL;
		A_L(0,1) = -nV(0) + dSMdUL(1)*rhoSL;
		A_L(0,2) = -nV(1) + dSMdUL(2)*rhoSL;
		A_L(0,3) = -nV(2) + dSMdUL(3)*rhoSL;
		A_L(0,4) =          dSMdUL(4)*rhoSL;

		A_L(1,0) =    qL*uL(0)       - nV(0)*psiL*gammaMinus1/2.0   + nV(0)*dpsdUL(0) + rhouSL(0)*dSMdUL(0);
		A_L(1,1) =  SL - qL          + nV(0)*(gamma-2.0)*uL(0)      + nV(0)*dpsdUL(1) + rhouSL(0)*dSMdUL(1);
		A_L(1,2) =     - uL(0)*nV(1) + nV(0)*gammaMinus1*uL(1)      + nV(0)*dpsdUL(2) + rhouSL(0)*dSMdUL(2);
		A_L(1,3) =     - uL(0)*nV(2) + nV(0)*gammaMinus1*uL(2)      + nV(0)*dpsdUL(3) + rhouSL(0)*dSMdUL(3);
		A_L(1,4) = -gammaMinus1*nV(0)                               + nV(0)*dpsdUL(4) + rhouSL(0)*dSMdUL(4);

		A_L(2,0) =    qL*uL(1)       - nV(1)*psiL*gammaMinus1/2.0   + nV(1)*dpsdUL(0) + rhouSL(1)*dSMdUL(0);
		A_L(2,1) =     - uL(1)*nV(0) + nV(1)*gammaMinus1*uL(0)      + nV(1)*dpsdUL(1) + rhouSL(1)*dSMdUL(1);
		A_L(2,2) =  SL - qL          + nV(1)*(gamma-2.0)*uL(1)      + nV(1)*dpsdUL(2) + rhouSL(1)*dSMdUL(2);
		A_L(2,3) =     - uL(1)*nV(2) + nV(1)*gammaMinus1*uL(2)      + nV(1)*dpsdUL(3) + rhouSL(1)*dSMdUL(3);
		A_L(2,4) = -gammaMinus1*nV(1)                               + nV(1)*dpsdUL(4) + rhouSL(1)*dSMdUL(4);

		A_L(3,0) =    qL*uL(2)       - nV(2)*psiL*gammaMinus1/2.0   + nV(2)*dpsdUL(0) + rhouSL(2)*dSMdUL(0);
		A_L(3,1) =     - uL(2)*nV(0) + nV(2)*gammaMinus1*uL(0)      + nV(2)*dpsdUL(1) + rhouSL(2)*dSMdUL(1);
		A_L(3,2) =     - uL(2)*nV(1) + nV(2)*gammaMinus1*uL(1)      + nV(2)*dpsdUL(2) + rhouSL(2)*dSMdUL(2);
		A_L(3,3) =  SL - qL          + nV(2)*(gamma-2.0)*uL(2)      + nV(2)*dpsdUL(3) + rhouSL(2)*dSMdUL(3);
		A_L(3,4) = -gammaMinus1*nV(2)                               + nV(2)*dpsdUL(4) + rhouSL(2)*dSMdUL(4);

		A_L(4,0) =      qL*(eL+pL)/rhoL - qL*psiL*(gamma-1.0)/2.0   + SM*dpsdUL(0) + (pS+eSL)*dSMdUL(0);
		A_L(4,1) = - nV(0)*(eL+pL)/rhoL + gammaMinus1*uL(0)*qL      + SM*dpsdUL(1) + (pS+eSL)*dSMdUL(1);
		A_L(4,2) = - nV(1)*(eL+pL)/rhoL + gammaMinus1*uL(1)*qL      + SM*dpsdUL(2) + (pS+eSL)*dSMdUL(2);
		A_L(4,3) = - nV(2)*(eL+pL)/rhoL + gammaMinus1*uL(2)*qL      + SM*dpsdUL(3) + (pS+eSL)*dSMdUL(3);
		A_L(4,4) =   SL-qL*gamma                                    + SM*dpsdUL(4) + (pS+eSL)*dSMdUL(4);

		A_L *= omL;

		return A_L;
	}
	
	def calcSubSonicJacobeanHLLC_R(rhoL : Double, uL : Vec[_3,Double], pL : Double, eL : Double, qL : Double, 
		  psiL : Double, SL : Double, rhoSL : Double, rhouSL : Vec[_3,Double], eSL : Double, dSMdUL : Vec[_5,Double], dSMdUR : Vec[_5,Double], dpsdUL : Vec[_5,Double], dpsdUR : Vec[_5,Double],
		  SM : Double, pS : Double, gamma : Double, nV : Vec[_3,Double] ) : Mat[_5,_5,Double] = {

		val gammaMinus1 = (gamma-1.0);
		val omL = 1.0/(SL-SM);
		
		var A_R = Constants.float5x5_zero;
		
		var i = 0;
		while(i < 5) {
			A_R(0,i) = omL*rhoSL*dSMdUR(i);
			A_R(1,i) = omL*(nV(0)*dpsdUR(i)+rhouSL(0)*dSMdUR(i));
			A_R(2,i) = omL*(nV(1)*dpsdUR(i)+rhouSL(1)*dSMdUR(i));
			A_R(3,i) = omL*(nV(2)*dpsdUR(i)+rhouSL(2)*dSMdUR(i));
			A_R(4,i) = omL*(dpsdUR(i)*SM+(pS+eSL)*dSMdUR(i));
			i += 1
		}
		
		return A_R;
	}

	def calcJacobianA(vel : Vec[_3,Double], pp : Double, rrho : Double, nV : Vec[_3,Double], gamma : Double, surfVeloc : Vec[_3,Double]) : Mat[_5,_5,Double] = {

		val kapm1 = (gamma - 1.0)
		val nVel =  Constants.float3_zero
		nVel(0) = vel(0)*nV(0)
		nVel(1) = vel(1)*nV(1)
		nVel(2) = vel(2)*nV(2)
		val U_k = nVel(0) + nVel(1) + nVel(2)
		val vSquHlf = 0.5*dot(vel, vel)
		val c = sqrt(gamma*pp/rrho)
		val inv_kap_m1 = 1.0/kapm1

		val tempVelocity = 0.0

		var A = Constants.float5x5_zero ;

		// Jacobian matrix
		A(0,0) = -tempVelocity;
		A(0,1) = nV(0);
		A(0,2) = nV(1);
		A(0,3) = nV(2);
		A(0,4) = 0.0;

		A(1,0) = -vel(0)*(nVel(1)+nVel(2))+nV(0)*(kapm1*vSquHlf-vel(0)*vel(0));
		A(1,1) = (2.-gamma)*nVel(0)+U_k-tempVelocity;
		A(1,2) = vel(0)*nV(1)-kapm1*vel(1)*nV(0);
		A(1,3) = vel(0)*nV(2)-kapm1*vel(2)*nV(0);
		A(1,4) = kapm1*nV(0);

		A(2,0) = -vel(1)*(nVel(0)+nVel(2))+nV(1)*(kapm1*vSquHlf-vel(1)*vel(1));
		A(2,1) = -kapm1*vel(0)*nV(1)+ vel(1)*nV(0);
		A(2,2) = (2.-gamma)*nVel(1)+U_k-tempVelocity;
		A(2,3) = vel(1)*nV(2)-kapm1*vel(2)*nV(1);
		A(2,4) = kapm1*nV(1);

		A(3,0) = -vel(2)*(nVel(0)+nVel(1))+nV(2)*(kapm1*vSquHlf-vel(2)*vel(2));
		A(3,1) = -kapm1*vel(0)*nV(2)+vel(2)*nV(0);
		A(3,2) = -kapm1*vel(1)*nV(2)+vel(2)*nV(1);
		A(3,3) = (2.-gamma)*nVel(2)+U_k-tempVelocity;
		A(3,4) = kapm1*nV(2);

		A(4,0) = U_k*((gamma-2.)*vSquHlf-c*c*inv_kap_m1);
		A(4,1) = c*c*inv_kap_m1*nV(0)-kapm1*vel(0)*(nVel(1)+nVel(2))-(kapm1*vel(0)*vel(0)-vSquHlf)*nV(0);
		A(4,2) = c*c*inv_kap_m1*nV(1)-kapm1*vel(1)*(nVel(0)+nVel(2))-(kapm1*vel(1)*vel(1)-vSquHlf)*nV(1);
		A(4,3) = c*c*inv_kap_m1*nV(2)-kapm1*vel(2)*(nVel(0)+nVel(1))-(kapm1*vel(2)*vel(2)-vSquHlf)*nV(2);
		A(4,4) = gamma*U_k-tempVelocity;
		
		return A
	}

	def calcNSFlux( rho0 : Double, u0 : Vec[_3, Double], grad_u0 : Mat[_3,_3,Double], h0 : Double,  grad_h0 :Vec[_3, Double], T0 : Double, R0 : Double, gam0 : Double, kine0 : Double, 
		rho1 : Double, u1 : Vec[_3, Double], grad_u1 : Mat[_3,_3,Double], h1 : Double,  grad_h1 :Vec[_3, Double], T1 : Double, R1 : Double, gam1 : Double, kine1 : Double,
		mul : Double, mut : Double, lambdaOverCp : Double, kine_fa : Double, u_fa : Vec[_3, Double], area : Double, nVec : Vec[_3, Double], smag : Double, sVec : Vec[_3, Double]) : Vec[_5,Double] = {

		var Frho = 0.0 ;
		var Frhou = Constants.float3_zero ;
		var FrhoE = 0.0 ;

		var alpha = dot(nVec, sVec);

		var grad_u_f = 0.5*(grad_u0 + grad_u1);

		var fCorr = nVec - alpha*sVec;

		val muTotalMomentum = mul + mut;
		var tauij_nj = Constants.float3_zero ;

		tauij_nj(0) = muTotalMomentum * ( alpha * (u1(0)-u0(0)) / smag
		  + grad_u_f(0,0) * fCorr(0) + grad_u_f(0,0) * nVec(0)
		  + grad_u_f(0,1) * fCorr(1) + grad_u_f(1,0) * nVec(1)
		  + grad_u_f(0,2) * fCorr(2) + grad_u_f(2,0) * nVec(2));

		tauij_nj(1) = muTotalMomentum * ( alpha * (u1(1)-u0(1)) / smag
		  + grad_u_f(1,0) * fCorr(0) + grad_u_f(0,1) * nVec(0)
		  + grad_u_f(1,1) * fCorr(1) + grad_u_f(1,1) * nVec(1)
		  + grad_u_f(1,2) * fCorr(2) + grad_u_f(2,1) * nVec(2));

		tauij_nj(2) = muTotalMomentum * ( alpha * (u1(2)-u0(2)) / smag
		  + grad_u_f(2,0) * nVec(0) + grad_u_f(0,2) * nVec(0)
		  + grad_u_f(2,1) * nVec(1) + grad_u_f(1,2) * nVec(1)
		  + grad_u_f(2,2) * nVec(2) + grad_u_f(2,2) * nVec(2));

		val tmp = 2.0 / 3.0 * muTotalMomentum * (grad_u_f(0,0) + grad_u_f(1,1) + grad_u_f(2,2)); 
		tauij_nj -= tmp * nVec;

		Frhou = -area * tauij_nj;

		val enth = lambdaOverCp * (alpha * (h1 - h0) / smag
		  + 0.5 * ((grad_h0.x + grad_h1.x) * fCorr.x
		  + (grad_h0.y + grad_h1.y) * fCorr.y 
		  + (grad_h0.z + grad_h1.z) * fCorr.z));

		FrhoE = -area * (enth + tauij_nj.x * u_fa.x + tauij_nj.y * u_fa.y + tauij_nj.z * u_fa.z);		
		return Vec(Frho, Frhou.x, Frhou.y, Frhou.z, FrhoE) ;

	}

	def calcNSJacobian( icv0 : Cell, icv1 : Cell,  
		  rho0 : Double, u0 : Vec[_3, Double], grad_u0 : Mat[_3,_3,Double], h0 : Double,  grad_h0 :Vec[_3, Double], T0 : Double, R0 : Double, 
		  gam0 : Double, kine0 : Double, rho1 : Double, u1 : Vec[_3, Double], grad_u1 : Mat[_3,_3,Double], h1 : Double,  grad_h1 :Vec[_3, Double], T1 : Double, R1 : Double, gam1 : Double, kine1 : Double,
		  mul : Double, mut : Double, lambdaOverCp : Double, kine_fa : Double, u_fa : Vec[_3, Double], area : Double, nVec : Vec[_3, Double], smag : Double, sVec : Vec[_3, Double]) {
		
		var alpha = dot(nVec, sVec);
		var grad_u_f = 0.5*(grad_u0 + grad_u1);
		var fCorr = nVec - alpha*sVec;
		val muTotalMomentum = mul + mut;
		var tauij_nj = Constants.float3_zero ;

		tauij_nj(0) = muTotalMomentum * ( alpha * (u1(0)-u0(0)) / smag
		  + grad_u_f(0,0) * fCorr(0) + grad_u_f(0,0) * nVec(0)
		  + grad_u_f(0,1) * fCorr(1) + grad_u_f(1,0) * nVec(1)
		  + grad_u_f(0,2) * fCorr(2) + grad_u_f(2,0) * nVec(2));

		tauij_nj(1) = muTotalMomentum * ( alpha * (u1(1)-u0(1)) / smag
		  + grad_u_f(1,0) * fCorr(0) + grad_u_f(0,1) * nVec(0)
		  + grad_u_f(1,1) * fCorr(1) + grad_u_f(1,1) * nVec(1)
		  + grad_u_f(1,2) * fCorr(2) + grad_u_f(2,1) * nVec(2));

		tauij_nj(2) = muTotalMomentum * ( alpha * (u1(2)-u0(2)) / smag
		  + grad_u_f(2,0) * nVec(0) + grad_u_f(0,2) * nVec(0)
		  + grad_u_f(2,1) * nVec(1) + grad_u_f(1,2) * nVec(1)
		  + grad_u_f(2,2) * nVec(2) + grad_u_f(2,2) * nVec(2));

		val tmp = 2.0 / 3.0 * muTotalMomentum * (grad_u_f(0,0) + grad_u_f(1,1) + grad_u_f(2,2)); 
		tauij_nj -= tmp * nVec;

		val enth = lambdaOverCp * (alpha * (h1 - h0) / smag
		  + 0.5 * ((grad_h0.x + grad_h1.x) * fCorr.x
		  + (grad_h0.y + grad_h1.y) * fCorr.y 
		  + (grad_h0.z + grad_h1.z) * fCorr.z));

		var dfdU = Constants.float5x5_zero;
		var dUdQ = Constants.float5x5_zero;
		val keff = lambdaOverCp
					
		// calculate A_L
		dfdU(1,1) = - muTotalMomentum * alpha / smag * area;
		dfdU(1,2) = 0.0;
		dfdU(1,3) = 0.0;
		dfdU(2,1) = 0.0;
		dfdU(2,2) = - muTotalMomentum * alpha / smag * area;
		dfdU(2,3) = 0.0;
		dfdU(3,1) = 0.0;
		dfdU(3,2) = 0.0;
		dfdU(3,3) = - muTotalMomentum * alpha / smag * area;
		dfdU(4,1) = area * tauij_nj(0) * 0.5 + u_fa(0) * dfdU(1, 1);
		dfdU(4,2) = area * tauij_nj(1) * 0.5 + u_fa(1) * dfdU(2, 2);
		dfdU(4,3) = area * tauij_nj(2) * 0.5 + u_fa(2) * dfdU(3, 3);
		dfdU(4,4) = -area * keff * alpha / smag;

		val invRhoL = 1.0 / rho0;
		dUdQ(1,0) = -u0(0) * invRhoL;
		dUdQ(2,0) = -u0(1) * invRhoL;
		dUdQ(3,0) = -u0(2) * invRhoL;
		dUdQ(4,0) =  gam0 * invRhoL * (- h0 - kine0 + R0*T0 + 0.5 * dot(u0,u0));
		dUdQ(1,1) =  invRhoL;
		dUdQ(4,1) = -gam0 * u0(0) * invRhoL;
		dUdQ(2,2) =  invRhoL;
		dUdQ(4,2) = -gam0 * u0(1) * invRhoL;
		dUdQ(3,3) =  invRhoL;
		dUdQ(4,3) = -gam0 * u0(2) * invRhoL;
		dUdQ(4,4) =  gam0 * invRhoL;

		val A_L = dFdU_times_dUdQ(dfdU, dUdQ);
		
		val row0 = JoeLS.row(icv0) ;
		val col0 = JoeLS.col(icv0) ;	
		val row1 = JoeLS.row(icv1) ;
		val col1 = JoeLS.col(icv1) ;
			
		//are we on a boundary, so we do only one?
		
		if (ID(icv0) == ID(icv1)) {
			
			JoeLiszt.A(row0,col0) += A_L ; 
			
		} else  {
		
			JoeLiszt.A(row0,col0) += A_L ; 
			JoeLiszt.A(row1,col1) -= A_L ;
			
			// calculate A_R
			dfdU(1,1) = muTotalMomentum * alpha / smag * area;
			dfdU(1,2) = 0.0;
			dfdU(1,3) = 0.0;
			dfdU(2,1) = 0.0;
			dfdU(2,2) = muTotalMomentum * alpha / smag * area;
			dfdU(2,3) = 0.0;
			dfdU(3,1) = 0.0;
			dfdU(3,2) = 0.0;
			dfdU(3,3) = muTotalMomentum * alpha / smag * area;
			dfdU(4,1) = area * tauij_nj(0) * 0.5 + u_fa(0) * dfdU(1, 1);
			dfdU(4,2) = area * tauij_nj(1) * 0.5 + u_fa(1) * dfdU(2, 2);
			dfdU(4,3) = area * tauij_nj(2) * 0.5 + u_fa(2) * dfdU(3, 3);
			dfdU(4,4) = area * keff * alpha / smag;
			
			val invRhoR = 1.0 / rho1;
			dUdQ(1,0) = -u1(0) * invRhoR;
			dUdQ(2,0) = -u1(1) * invRhoR;
			dUdQ(3,0) = -u1(2) * invRhoR;
			dUdQ(4,0) =  gam1 * invRhoR * (- h1 - kine1 + R1 * T1 + 0.5 * dot(u1,u1));
			dUdQ(1,1) =  invRhoR;
			dUdQ(4,1) = -gam1 * u1(0) * invRhoR;
			dUdQ(2,2) =  invRhoR;
			dUdQ(4,2) = -gam1 * u1(1) * invRhoR;
			dUdQ(3,3) =  invRhoR;
			dUdQ(4,3) = -gam1 * u1(2) * invRhoR;
			dUdQ(4,4) =  gam1 * invRhoR;

			val A_R = dFdU_times_dUdQ(dfdU, dUdQ);
			
			JoeLiszt.A(row0,col1) += A_R ;
			JoeLiszt.A(row1,col0) -= A_R ; 

		}

	}

	def dFdU_times_dUdQ( dfdU : Mat[_5,_5,Double], dUdQ : Mat[_5,_5,Double] ) : Mat[_5,_5,Double] = {
		var A = Constants.float5x5_zero ;
		
		A(1,0) = dfdU(1,1)*dUdQ(1,0) + dfdU(1,2)*dUdQ(2,0) + dfdU(1,3)*dUdQ(3,0);
		A(2,0) = dfdU(2,1)*dUdQ(1,0) + dfdU(2,2)*dUdQ(2,0) + dfdU(2,3)*dUdQ(3,0);
		A(3,0) = dfdU(3,1)*dUdQ(1,0) + dfdU(3,2)*dUdQ(2,0) + dfdU(3,3)*dUdQ(3,0);
		A(4,0) = dfdU(4,1)*dUdQ(1,0) + dfdU(4,2)*dUdQ(2,0) + dfdU(4,3)*dUdQ(3,0) + dfdU(4,4)*dUdQ(4,0);
	
		A(1,1) = dfdU(1,1)*dUdQ(1,1);
		A(2,1) = dfdU(2,1)*dUdQ(1,1);
		A(3,1) = dfdU(3,1)*dUdQ(1,1);
		A(4,1) = dfdU(4,1)*dUdQ(1,1) + dfdU(4,4)*dUdQ(4,1);
	
		A(1,2) = dfdU(1,2)*dUdQ(2,2);
		A(2,2) = dfdU(2,2)*dUdQ(2,2);
		A(3,2) = dfdU(3,2)*dUdQ(2,2);
		A(4,2) = dfdU(4,2)*dUdQ(2,2) + dfdU(4,4)*dUdQ(4,2);
	
		A(1,3) = dfdU(1,3)*dUdQ(3,3);
		A(2,3) = dfdU(2,3)*dUdQ(3,3);
		A(3,3) = dfdU(3,3)*dUdQ(3,3);
		A(4,3) = dfdU(4,3)*dUdQ(3,3)   + dfdU(4,4)*dUdQ(4,3);
	
		A(4,4) = dfdU(4,1)*dUdQ(1,4) + dfdU(4,2)*dUdQ(2,4) + dfdU(4,3)*dUdQ(3,4) + dfdU(4,4)*dUdQ(4,4);
		
		return A
	}

	def calcDt( cfl_target : Double ) : Double = {

		var dt = 0.0 ;

		if ( IC.timeStepMode == 0 ) {
			return IC.const_dt ;
		}
		if ( IC.timeStepMode == 1 ) {
			dt = IC.DT ;
			IC.const_dt = IC.DT ;
			for( c <- cells(mesh) ) {
				local_dt(c) = dt ;
			}
			IC.timeStepMode = 0 ;
		}
		if ( IC.timeStepMode == 2 ) {
			for( icv <- cells(mesh) ) {
				var dt_cv = MAX_DOUBLE ;
				var lambdaMax = 0.0 ;

				val c = sqrt( gamma(icv) * press(icv) / JoeLiszt.rho(icv) ) ;

				for ( f <- faces(icv) ) {
					var nVec = MeshGeometryCalc.fa_normal(f) ;
					val area = sqrt(dot(nVec, nVec)) ;
					nVec = nVec / area ;

					val Uk = dot( JoeLiszt.rhou(icv), nVec ) / JoeLiszt.rho(icv) ;
					var lambda = ( fabs(Uk) + c ) * area ;
					//        				val lamdiff = 4.0 * (mu_fa(f) + mu_fa(f)) / JoeLiszt.rho(icv) * area * area / MeshGeometryCalc.cv_volume(icv);
					//				        lambda = lambda.max(lamdiff);
					lambdaMax = lambdaMax.max(lambda) ;
				}

				dt_cv = cfl_target * MeshGeometryCalc.cv_volume(icv) / lambdaMax ;
				IC.dt_minCPU = IC.dt_minCPU.min(dt_cv) ;
				local_dt(icv) = dt_cv ;
			}
			dt = IC.dt_minCPU ;
		}
		return dt ;
	}

}
