import Liszt.Language._

@lisztcode
object JoeWithModels {

	val rhs_rho = FieldWithConst[Cell,Float](0.f) ;
	val rhs_rhou = FieldWithConst[Cell,Vec[_3,Float]](Constants.float3_zero) ;
	val rhs_rhoE = FieldWithConst[Cell,Float](0.f) ;

	// Flow variable initialization
	val rho = FieldWithConst[Cell,Float](IC.rho_init) ;
	val rhou = FieldWithConst[Cell,Vec[_3,Float]](IC.rho_init * IC.u_init) ;
	val rhoE = FieldWithConst[Cell,Float](IC.p_init / (IC.GAMMA - 1.f)  +  0.5f * IC.rho_init * dot( IC.u_init, IC.u_init )) ;

	val zone_interior = BoundarySet[Face]("zone_interior")
	val zone_boundary = BoundarySet[Face]("zone_boundary")
	val symmetry = BoundarySet[Face]("symmetry")
	val wall = BoundarySet[Face]("wall")
	val other_boundaries = BoundarySet[Face]("other_boundaries")
	val cbc = BoundarySet[Face]("cbc")
	val cbc_subsonic_outlet = BoundarySet[Face]("cbc_subsonic_outlet")
	val cbc_subsonic_inlet = BoundarySet[Face]("cbc_subsonic_inlet")
	val symmetry_pos = BoundarySet[Face]("symmetry_pos")
	val symmetry_neg = BoundarySet[Face]("symmetry_neg")
			
	def init() { Print("JoeWithModels()") ;}

	def getInsideBoundaryCell( f : Face ) : Cell = {
		// NOTE(boulos): Sorry the "old" code pisses off my
		// vectorized return statements (and I've currently
		// forgotten why). This equivalent code works though.
		val inside_cell = inside(f);
		val outside_cell = outside(f);
		val result = if (ID(inside_cell) == 0) outside_cell else inside_cell;
		return result;
		//val c = if ( ID(inside(f)) == 0 ) outside(f) else inside(f) ;
		//return c ;
	}

	def run() {
		if ( IC.navierStokesSolver == NavierStokesSolvers.EXPL_EULER ) runExplicitForwardEuler() ;
		else Print("ERROR: no or wrong time integration scheme specified !") ;
	}

	def runExplicitForwardEuler() {

		val zeroFiveVec = Vec(0.f, 0.f, 0.f, 0.f, 0.f) ;

		UgpWithCvCompFlow.calcStateMaterialProp() ;
				
		setBoundaryConditions() ;
		var start_time = 0.0;
		var step = 0 ;
		while (step < Constants.iterations ) {
			if (step == 1) {
				start_time = wall_time();
			}

			step += 1 ;
			
			// Initialize the residuals
			var my_resid_rho = 0.f ;
			var my_resid_rhou = Constants.float3_zero ;
			var my_resid_rhoE = 0.f ;

			// Compute the time step
			var dtMin = UgpWithCvCompFlow.calcDt(IC.cfl) ;

			// Spatial integration. Compute convective and viscous residuals
			SpatialIntegration() ;

			// Time integration. Update the solution
			TimeIntegration() ;

			// Residual update for visualization purposes
			if ( step % IC.check_interval == 0 ) {
				if ( step % ( IC.check_interval * 10 ) == 0 ) {
					Print( "" ) ; // print nothing to simulate a \n !!!!
					Print( "done step: " , step , ", cfl: " , IC.cfl , ", min. dt:   " , dtMin ) ; 
				}
				var my_resid = Vec(0.f,0.f,0.f,0.f,0.f)
				for( c <- cells(mesh) ) {
                                  val read_rhou = rhs_rhou(c);
				  my_resid += Vec(fabsf(rhs_rho(c)),
				                  fabsf(read_rhou.x),
				                  fabsf(read_rhou.y),
				                  fabsf(read_rhou.z),
				                  fabsf(rhs_rhoE(c)))
				}
				showResidue( my_resid, step ) ;
			}

			UgpWithCvCompFlow.calcStateMaterialProp() ;

			setBoundaryConditions() ;
		}

		 Print( "TIME_FOR_LOOP: ", wall_time() - start_time ) ;
		
	}



	def SpatialIntegration() {
		// Initialize the value of the residuals
		for( c <- cells(mesh) ) {
			rhs_rho(c) = 0.f ;
			rhs_rhou(c) =  Constants.float3_zero ;
			rhs_rhoE(c) = 0.f ;
		}
		
		// Compute velocity and enthalpy gradient for viscous terms
		if (IC.mu_ref > 0.f) {
		   calcVelocityGradient() ;
		   calcEntalphyGradient() ;
		}
  		// Space integration (inviscid and viscous)
		for (f <- zone_interior) { calcRhsInteriorInviscid(f) }
		if (IC.mu_ref > 0.f) 
		   for (f <- zone_interior) { calcRhsInteriorViscous(f) }
		
		// Set boundary conditions
		for (f <- symmetry) { calcRhsSymmetry(f) }
		for (f <- wall) { calcRhsWall(f) }
		for (f <- other_boundaries) { calcRhsOtherBoundaries(f) }
	}

	def TimeIntegration() {
		for( c <- cells(mesh) ) {
			val tmp = UgpWithCvCompFlow.local_dt(c) / MeshGeometryCalc.cv_volume(c) ;
			rho(c) += tmp * rhs_rho(c) ;
			rhou(c) += tmp * rhs_rhou(c) ;
			rhoE(c) += tmp * rhs_rhoE(c) ;
		}
	}


	def setBoundaryConditions() {

		var bc_err = false ;

		val vector_bc = Vec( IC.u_cbc.x, IC.u_cbc.y, IC.u_cbc.z, IC.T_init, IC.p_init )
		val T_bc = vector_bc(_3)
		val p_bc = vector_bc(_4)
		val u_bc = Vec( vector_bc.x, vector_bc.y, vector_bc.z )

		// Convective bondary conditions (farfield)
		for (f <- cbc) {
			CBCBoundary(f, T_bc, p_bc, u_bc)
		}
		for (f <- cbc) {
			UgpWithCvCompFlow.ComputeBCProperties_T(f)
		}

		// Convective bondary conditions (subsonic outlet)
		for (f <- cbc_subsonic_outlet) {
			CBCSubsonicOutletBoundary(f, p_bc);
		}
		for (f <- cbc_subsonic_outlet) {
			UgpWithCvCompFlow.ComputeBCProperties_T(f)
		}

		// Symmetry and Euler wall boundary conditions
		for (f <- symmetry) {
			SymmetryBoundary(f);
		}
		for (f <- symmetry) {
			UgpWithCvCompFlow.ComputeBCProperties_T(f)
		}

		// Navier-Stokes wall boundary conditions
		for (f <- wall) {
			NonSlipBoundary(f);
		}
		for (f <- wall) {
			UgpWithCvCompFlow.ComputeBCProperties_T(f)
		}

		// Set density for all the boundary conditions
		for (f <- zone_boundary) {
			setRhoFa(f)
		}

	}

	
	def showResidue( rhsResid: Vec[_5,Float], step: Int ) {
		var sCount = 0;
		var checkCount = 0;
		for (c <- cells(mesh)) {
			checkCount += size(faces(c))
		}
		if ( step % (IC.check_interval * 10 ) == 0 ) {
			Print( "		rho		rhou-X		rhou-Y		rhou-Z		rhoE") ;
		}
		Print( "RESID: ", step, "    ", rhsResid(_0), "    ", rhsResid(_1), "    ", rhsResid(_2), "    ", rhsResid(_3), "    ", rhsResid(_4)) ;
	}

	
	def CBCBoundary( f: Face, T_bc : Float, p_bc : Float, u_bc : Vec[_3,Float] ) {
		UgpWithCvCompFlow.T_fa(f) = T_bc ;
		UgpWithCvCompFlow.vel_fa(f) = u_bc ;
		UgpWithCvCompFlow.p_fa(f) = p_bc ;
	}

	def CBCSubsonicOutletBoundary( f: Face, p_bc : Float ) {
		val icv0 = getInsideBoundaryCell(f) ;
		UgpWithCvCompFlow.T_fa(f) = UgpWithCvCompFlow.temp(icv0) ;
		UgpWithCvCompFlow.vel_fa(f) = rhou(icv0) / rho(icv0) ;
		UgpWithCvCompFlow.p_fa(f) = p_bc ;
	}

	def SymmetryBoundary( f: Face ) {
		val icv0 = getInsideBoundaryCell(f) ;
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		nVec = normalize(nVec) ;
		
		// flip u, APPROXIMATION ---> take velocity at the cell center
		val u0 = UgpWithCvCompFlow.vel(icv0) ;
		val un = dot( nVec, u0 ) ;
		UgpWithCvCompFlow.vel_fa(f) = u0 - 1.f * un * nVec ;

		UgpWithCvCompFlow.T_fa(f) = UgpWithCvCompFlow.temp(icv0) ;
		UgpWithCvCompFlow.p_fa(f) = UgpWithCvCompFlow.press(icv0) ;
		UgpWithCvCompFlow.RoM_fa(f) = UgpWithCvCompFlow.RoM(icv0) ;
		UgpWithCvCompFlow.gam_fa(f) = UgpWithCvCompFlow.gamma(icv0) ;
		UgpWithCvCompFlow.h_fa(f) = UgpWithCvCompFlow.enthalpy(icv0) ;
	}

	def NonSlipBoundary( f: Face ) {
		val icv0 = getInsideBoundaryCell(f) ;
		
		// Null velocity at the face
		UgpWithCvCompFlow.vel_fa(f) = Constants.float3_zero ;

		UgpWithCvCompFlow.T_fa(f) = UgpWithCvCompFlow.temp(icv0) ;
		UgpWithCvCompFlow.p_fa(f) = UgpWithCvCompFlow.press(icv0) ;
		UgpWithCvCompFlow.RoM_fa(f) = UgpWithCvCompFlow.RoM(icv0) ;
		UgpWithCvCompFlow.gam_fa(f) = UgpWithCvCompFlow.gamma(icv0) ;
		UgpWithCvCompFlow.h_fa(f) = UgpWithCvCompFlow.enthalpy(icv0) ;
	}

	def setRhoFa( f: Face ) {
		UgpWithCvCompFlow.rho_fa(f) = UgpWithCvCompFlow.p_fa(f) / ( UgpWithCvCompFlow.RoM_fa(f) * UgpWithCvCompFlow.T_fa(f) ) ;
	}


	def calcEntalphyGradient() {

		for( c <- cells(mesh) ) {
			UgpWithCvCompFlow.enthalpy_grad(c) = Constants.float3_zero ;
		}

		for (f <- zone_interior) {
		    val icv0 = outside(f) ;
		    val icv1 = inside(f) ;
		    val h0 = UgpWithCvCompFlow.enthalpy(icv0) ;
		    val h1 = UgpWithCvCompFlow.enthalpy(icv1) ;
		    val h01 = 0.5f*(h0 + h1) ;
		    var nVec = MeshGeometryCalc.fa_normal(f) ;

		    UgpWithCvCompFlow.enthalpy_grad(icv0) += h01 * nVec / MeshGeometryCalc.cv_volume(icv0)

		    UgpWithCvCompFlow.enthalpy_grad(icv1) -= h01 * nVec / MeshGeometryCalc.cv_volume(icv1)
		
		}
		
		for (f <- zone_boundary) {
		    val icv0 = getInsideBoundaryCell(f) ;
		    val h0 = UgpWithCvCompFlow.h_fa(f) ;
		    var nVec = MeshGeometryCalc.fa_normal(f) ;
		    UgpWithCvCompFlow.enthalpy_grad(icv0) += h0 * nVec / MeshGeometryCalc.cv_volume(icv0);

		}
	}

	def calcVelocityGradient() {

		for( c <- cells(mesh) ) {
		    UgpWithCvCompFlow.vel_grad(c) =  Constants.float3x3_zero ;
		}

		for (f <- zone_interior) {
		    val icv0 = outside(f) ;
		    val icv1 = inside(f) ;
		    val vel0 = UgpWithCvCompFlow.vel(icv0) ;
		    val vel1 = UgpWithCvCompFlow.vel(icv1) ;
		    var vel01 = 0.5f*(vel0 + vel1) ;

		    var nVec = MeshGeometryCalc.fa_normal(f) ;

		    val update = Mat(vel01(_0) * nVec,
                                     vel01(_1) * nVec,
                                     vel01(_2) * nVec);
		    UgpWithCvCompFlow.vel_grad(icv0) += update / MeshGeometryCalc.cv_volume(icv0)
		    UgpWithCvCompFlow.vel_grad(icv1) -= update / MeshGeometryCalc.cv_volume(icv1)
		}

		for (f <- zone_boundary) {
		    val icv0 = getInsideBoundaryCell(f) ;
		    val vel0 = UgpWithCvCompFlow.vel_fa(f) ;
		    var nVec = MeshGeometryCalc.fa_normal(f) ;
		    val update = Mat(vel0(_0) * nVec, vel0(_1) * nVec, vel0(_2) * nVec);
		    UgpWithCvCompFlow.vel_grad(icv0) += update / MeshGeometryCalc.cv_volume(icv0)
		}
	}

	def calcRhsInteriorViscous( f: Face ) {

		val icv0 = outside(f) ;
		val icv1 = inside(f) ;

		val rho0 = rho(icv0) ;
		val u0 = UgpWithCvCompFlow.vel(icv0) ;
		val grad_u0 = UgpWithCvCompFlow.vel_grad(icv0)
		val h0 = UgpWithCvCompFlow.enthalpy(icv0) ;
		val grad_h0 = UgpWithCvCompFlow.enthalpy_grad(icv0) ;
		val t0 =  UgpWithCvCompFlow.temp(icv0) ;
		val muTurb0 = 0.0f ;
		val R0 = UgpWithCvCompFlow.RoM(icv0) ;
		val gam0 = UgpWithCvCompFlow.gamma(icv0) ;
		val kine0 = 0.0f ;

		val rho1 = rho(icv1) ;
		val u1 = UgpWithCvCompFlow.vel(icv1) ;
		val grad_u1 = UgpWithCvCompFlow.vel_grad(icv1) ;
		val h1 = UgpWithCvCompFlow.enthalpy(icv1) ;
		val grad_h1 = UgpWithCvCompFlow.enthalpy_grad(icv1) ;
		val t1 = UgpWithCvCompFlow.temp(icv1) ;
		val muTurb1 =  0.0f ;
		val R1 = UgpWithCvCompFlow.RoM(icv1) ;
		val gam1 = UgpWithCvCompFlow.gamma(icv1) ;
		val kine1 = 0.0f ;

		// face unit normal and area...
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = nVec/area ;
		var sVec = MeshGeometryCalc.x_cv(icv1) - MeshGeometryCalc.x_cv(icv0)
		val smag = sqrt(dot(sVec,sVec)) ;
		sVec = sVec/smag ;

		val Frho5 = UgpWithCvCompFlow.calcNSFlux( rho0, u0, grad_u0, h0,  grad_h0, t0, R0, gam0, kine0,
		    	    				  rho1, u1, grad_u1, h1,  grad_h1, t1, R1, gam1, kine1,
							   UgpWithCvCompFlow.mu_fa(f), 0.5f*(muTurb0+muTurb1), UgpWithCvCompFlow.lamOcp_fa(f), 
							   0.5f*(kine0+kine1), 0.5f*(u0+u1), area, nVec, smag, nVec) ;

		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;

		rhs_rhou(icv1) += Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv1) += Frho5(_4) ;

	}

	def calcRhsInteriorInviscid( f: Face ) {

		val icv0 = outside(f) ;
		val icv1 = inside(f) ;

		val rho0 = rho(icv0) ;
		val u0 = UgpWithCvCompFlow.vel(icv0) ;
		val p0 = UgpWithCvCompFlow.press(icv0) ;
		val h0 = UgpWithCvCompFlow.enthalpy(icv0) ;

		val rho1 = rho(icv1) ;
		val u1 = UgpWithCvCompFlow.vel(icv1) ;
		val p1 = UgpWithCvCompFlow.press(icv1) ;
		val h1 = UgpWithCvCompFlow.enthalpy(icv1) ;

		// face unit normal and area...
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = normalize(nVec) ;
	
		val kine0 = UgpWithCvCompFlow.kine(icv0) ;
		val kine1 = UgpWithCvCompFlow.kine(icv1) ;

		val Frho5 = UgpWithCvCompFlow.calcEulerFlux_HLLC(rho0, u0, p0, h0, UgpWithCvCompFlow.gamma(icv0), rho1, u1, p1, h1, UgpWithCvCompFlow.gamma(icv1), area, nVec, 0.f, kine0, kine1) ;
		
		rhs_rho(icv0) -= Frho5(_0) ;
		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;

		rhs_rho(icv1) += Frho5(_0) ;
		rhs_rhou(icv1) += Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv1) += Frho5(_4) ;

	}

	def calcRhsSymmetry( f: Face) {

		val icv0 = getInsideBoundaryCell(f) ;

		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = nVec/area ;
	
		val kine0 = 0.0f ;
		val kine1 = 0.0f ;

		val Frho5 = UgpWithCvCompFlow.calcEulerFlux_HLLC(UgpWithCvCompFlow.rho_fa(f), UgpWithCvCompFlow.vel_fa(f), UgpWithCvCompFlow.p_fa(f), 
		    	    					 UgpWithCvCompFlow.h_fa(f), UgpWithCvCompFlow.gam_fa(f), UgpWithCvCompFlow.rho_fa(f), 
								 UgpWithCvCompFlow.vel_fa(f), UgpWithCvCompFlow.p_fa(f), UgpWithCvCompFlow.h_fa(f), 
								 UgpWithCvCompFlow.gam_fa(f), area, nVec, 0.f, kine0, kine1) ;

		rhs_rho(icv0) -= Frho5(_0) ;
		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;

	}

	def calcRhsWall( f: Face) {

		val icv0 = getInsideBoundaryCell(f) ;
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = nVec/area ;

		val Frho5 = UgpWithCvCompFlow.calcEulerFlux_HLLC(UgpWithCvCompFlow.rho_fa(f), UgpWithCvCompFlow.vel_fa(f), UgpWithCvCompFlow.p_fa(f), 
		    	    					 UgpWithCvCompFlow.h_fa(f), UgpWithCvCompFlow.gam_fa(f), UgpWithCvCompFlow.rho_fa(f), 
								 UgpWithCvCompFlow.vel_fa(f), UgpWithCvCompFlow.p_fa(f), UgpWithCvCompFlow.h_fa(f), 
								 UgpWithCvCompFlow.gam_fa(f), area, nVec, 0.f, 0.f, 0.f) ;

		rhs_rho(icv0) -= Frho5(_0) ;
		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;

		val grad_u0 = UgpWithCvCompFlow.vel_grad(icv0)
		val grad_h0 = UgpWithCvCompFlow.enthalpy_grad(icv0) ;
		val muLam0 = UgpWithCvCompFlow.muLam(icv0) ;
		val muTurb0 = 0.0f ;
		val kine0 = 0.0f ;
		val kine1 = 0.0f ;
		val kine_fa = 0.0f;

		val sVec = MeshGeometryCalc.x_fa(f) - MeshGeometryCalc.x_cv(icv0)
		val smag = sqrt(dot(sVec,sVec)) ;

		val Frho5_ = UgpWithCvCompFlow.calcNSFlux(  rho(icv0), UgpWithCvCompFlow.vel(icv0), grad_u0, UgpWithCvCompFlow.enthalpy(icv0),  
		    	    				  grad_h0,  UgpWithCvCompFlow.temp(icv0), UgpWithCvCompFlow.RoM(icv0), UgpWithCvCompFlow.gamma(icv0), kine0,
		    	    				  UgpWithCvCompFlow.rho_fa(f),  UgpWithCvCompFlow.vel_fa(f), grad_u0, UgpWithCvCompFlow.h_fa(f),  
							  grad_h0, UgpWithCvCompFlow.T_fa(f), UgpWithCvCompFlow.RoM_fa(f), UgpWithCvCompFlow.gam_fa(f), kine1,
							  UgpWithCvCompFlow.mu_fa(f), .0f, UgpWithCvCompFlow.lamOcp_fa(f), kine_fa,  UgpWithCvCompFlow.vel_fa(f), area, nVec, smag, nVec) ;

		rhs_rhou(icv0) -= Vec(Frho5_(_1), Frho5_(_2), Frho5_(_3)) ;
		rhs_rhoE(icv0) -= Frho5_(_4) ;
	}

	def calcRhsOtherBoundaries( f: Face ) {

		val icv0 = getInsideBoundaryCell(f) ;

		val rho0 = rho(icv0) ;
		val u0 = UgpWithCvCompFlow.vel(icv0) ;
		val p0 = UgpWithCvCompFlow.press(icv0) ;
		val h0 = UgpWithCvCompFlow.enthalpy(icv0) ;
		val gam0 = UgpWithCvCompFlow.gamma(icv0) ;


		// face unit normal and area...
		var nVec = MeshGeometryCalc.fa_normal(f) ;
		val area = sqrt(dot(nVec,nVec)) ;
		nVec = normalize(nVec) ;
	
		val kine0 = 0.0f ;
		val kine1 = 0.0f ;

		val Frho5 = UgpWithCvCompFlow.calcEulerFlux_HLLC(rho0, u0, p0, h0, gam0, UgpWithCvCompFlow.rho_fa(f), UgpWithCvCompFlow.vel_fa(f), 
		    	    					 UgpWithCvCompFlow.p_fa(f), UgpWithCvCompFlow.h_fa(f), UgpWithCvCompFlow.gam_fa(f), area, nVec, 0.f, kine0, kine1) ;
		
		rhs_rho(icv0) -= Frho5(_0) ;
		rhs_rhou(icv0) -= Vec(Frho5(_1), Frho5(_2), Frho5(_3)) ;
		rhs_rhoE(icv0) -= Frho5(_4) ;

		if (IC.mu_ref > 0.f) {
		   val grad_u0 = UgpWithCvCompFlow.vel_grad(icv0)
		   val grad_h0 = UgpWithCvCompFlow.enthalpy_grad(icv0) ;
		   val muLam0 = UgpWithCvCompFlow.muLam(icv0) ;
		   val muTurb0 = 0.0f ;
		   val kine_fa = 0.0f;

		   var sVec = MeshGeometryCalc.x_fa(f) - MeshGeometryCalc.x_cv(icv0)
		   val smag = sqrt(dot(sVec,sVec)) ;
		   sVec /= smag

		   val Frho5_ = UgpWithCvCompFlow.calcNSFlux(  rho(icv0), UgpWithCvCompFlow.vel(icv0), grad_u0, UgpWithCvCompFlow.enthalpy(icv0),  
		    	    				  grad_h0,  UgpWithCvCompFlow.temp(icv0), UgpWithCvCompFlow.RoM(icv0), UgpWithCvCompFlow.gamma(icv0), kine0,
		    	    				  UgpWithCvCompFlow.rho_fa(f),  UgpWithCvCompFlow.vel_fa(f), grad_u0, UgpWithCvCompFlow.h_fa(f),  
							  grad_h0, UgpWithCvCompFlow.T_fa(f), UgpWithCvCompFlow.RoM_fa(f), UgpWithCvCompFlow.gam_fa(f), kine1,
							  UgpWithCvCompFlow.mu_fa(f), .0f, UgpWithCvCompFlow.lamOcp_fa(f), kine_fa,  UgpWithCvCompFlow.vel_fa(f), area, nVec, smag, sVec) ;

		   rhs_rhou(icv0) -= Vec(Frho5_(_1), Frho5_(_2), Frho5_(_3)) ;
		   rhs_rhoE(icv0) -= Frho5_(_4) ;
		}

	}

}
