#ifndef _CELL_WALL_PARSE_H
#define _CELL_WALL_PARSE_H

#include <vector>
#include "MeshIO/LisztFileWriter.h"
#include "MeshIO/LisztFormat.h"
#include "MeshIO/EdgeTable.h"

namespace MeshIO {
class CellWallParse {
private:
	//output stuff
    FileField position_field;
    FileField material_field;

    FieldTableIndex index;
    LisztHeader* header;
    EdgeTable edge_table;
    LisztFileWriter* writer;
	
public:
	CellWallParse() {}
	void init(LisztFileWriter* wr, const std::string &meshfile, const std::string &datafile, const std::string & outfile) {
	  writer = wr;
	  header = &writer->header;
	  file = FOPEN(meshfile.c_str(), "r");
	  data = FOPEN(datafile.c_str(), "r");

		if(!file) {
			perror(NULL);
			assert(!"could not read off file");
		}
		writer->init(outfile);
		header->nV  = 0;
		header->nE  = 0;
		header->nBoundaries = 0;
		header->nF  = 0;
		header->nC  = 2;
		header->nFE = 0;

        // initialize position field
        position_field.domain        = VERTEX_T;
        position_field.range.type    = LISZT_DOUBLE;
        position_field.range.flags   = LISZT_VEC_FLAG;
        position_field.range.data[0] = 3;

        // initialize material field
        material_field.domain      = EDGE_T;
        material_field.range.type  = LISZT_INT;
        material_field.range.flags = 0;

        index.num_fields = 2;
	}

    void openDataFile ( ) {
        fclose(file);
        file = data;
    }

	void parseAll() {
	  writer->setBeginData();
		
	  char * line = readline();
	  int    nV   = 0;

	  position_field.data = writer->currentPosition();
	  
	  // scan vertices
	  while (line[0] == 'v') {
            char v;
            double p[3];
            sscanf(line, "%c %lf %lf %lf", &v, p, p+1, p+2); 

            writer->writeValue(p[0]);
            writer->writeValue(p[1]);
            writer->writeValue(p[2]);
            line = readline();
            nV++;
        }

        header->nV            = nV;
        position_field.nElems = nV;

	header->facet_edge_table = writer->currentPosition();
        int nF = 0;
        
        // Scan faces and construct facetEdges from each listed edge
        while (line[0] == 'f') {
	  std::vector<id_t> verts;
            int pos;
            id_t v;

            line++; // advance line pointer past 'f' char
            while (sscanf(line, " %d%n", &v, &pos) > 0) {
                verts.push_back(v-1); // This minus 1 is really important! vertices are indexed from 1
                line += pos;          // in the OBJ, but indexed from 0 in the lmesh format.
            }

            int nedges = verts.size();
			for(int j = 0; j < nedges; j++) {
				id_t tail = verts[j];
				id_t head = verts[(j + 1) % nedges];
				id_t edge = edge_table.get(tail,head);
				addFacetEdge(nF, edge, head, 0, tail, 1);
			}
            
            nF++;
            line = readline();
            if (!line)
                break;
		}

		//fill in header
        header->nF = nF;
	header->nE = edge_table.size();

        openDataFile();

        material_field.data   = writer->currentPosition();
        material_field.nElems = header->nE;

        int *edge_data = (int *) malloc(sizeof(int) * header->nE);
        char m;
        id_t v0, v1;

        int f_size = 0;
        line = readline();


        EdgeTable temp_table;
        while (line && sscanf(line, "%c %d %d", &m, &v0, &v1) == 3) {
            v0--; // update vertex indices to be zero-indexed.
            v1--;
            temp_table.get(v0, v1, NULL);

            bool added = false;
            id_t edge = edge_table.get(v0, v1, &added);


	    // We shouldn't new edges as this point...
            if (added) {
                printf("Error: Edge connecting vertices %d and %d in "
                       ".dat does not exist in mesh!\n", v0, v1);
            }
            assert(added == false);

            *(edge_data + edge) = ((m == 'S') ? 1 : 0);
            f_size++;
            line = readline();
        }

        EdgeTable::table_t::iterator it = edge_table.table.begin();

        while (it != edge_table.table.end()) {
            bool added = false;
            id_t v0 = it->first.first;
            id_t v1 = it->first.second;
            id_t edge = temp_table.get(v0, v1, &added);

            if (added) {
                printf("Error: edge from %d to %d not defined in edge properties file\n", v0, v1);
                assert(false);
            }
            it++;
        }


        assert(f_size == header->nE);
        writer->write(header->nE * sizeof(int), edge_data); 

        // Finalize field table info!
        position_field.name = writer->currentPosition();
        writer->writeString("position");

        material_field.name = writer->currentPosition();
        writer->writeString("springValues");

        header->field_table_index = writer->currentPosition();
        writer->writeValue(index);
        writer->writeValue(position_field);        
        writer->writeValue(material_field);
	}
private:
	char * readline() {
		assert(file);
		char * line;
		do {
			line = fgets(buf,2048,file);
			if(!line) {
				// perror(NULL);
				// assert(!"read failed");
                return NULL;
			}
			while(isspace(*line)) { line++; }
		} while(line[0] == '#' || strncmp(line,"OFF",3) == 0);
		char * cur = line;
		while(*cur && *cur != '#')
			cur++;
		*cur = '\0';
		return line;
	}
	void addFacetEdge(id_t face, id_t edge, id_t vert0, id_t cell0, id_t vert1, id_t cell1) {
		FileFacetEdge fe;
		fe.face = face;
		fe.edge = edge;
		fe.hf[0].vert = vert0;
		fe.hf[0].cell = cell0;
		fe.hf[1].vert = vert1;
		fe.hf[1].cell = cell1;
		// printf("fe = %d, f = %d, e = %d,  (v = %d, c = %d) (v = %d, c = %d)\n",
		// header->nFE,fe.face,fe.edge, fe.hf[0].vert, fe.hf[0].cell, fe.hf[1].vert, fe.hf[1].cell);
	
		writer->writeObject(&fe);
		header->nFE++;
	}
	char buf[2048];
    FILE * data;
	FILE * file;
};
} // namespace MeshIO
#endif /* _CELL_WALL_PARSE_H */
