#include <ctype.h>
#include <stdlib.h>
#include <iostream>
#include <vector>
#include <map>
#include "interpret.h"
#include "SingleCore/IObject.h"

static const char * token_str[] = { "NONE","!","*","/","%","+","-","<",">","<=",">=","=","!=", "&&", "||",
"(", ")","{", "}", "$", "IDENT", "NUM", ",", "~" };
enum TokenID {
	NONE,
	NOT,
	
	TIMES,
	DIV,
	MOD,
	
	PLUS,
	MINUS,
	
	LESS,
	GREAT,
	LESSE,
	GREATE,
	
	EQUAL,
	NOTEQUAL,
	
	AND,
	
	OR,
	
	LPAREN,
	RPAREN,
	
	LBRACE,
	RBRACE,
	APPLY,
	IDENT,
	NUM,
	COMMA,
	UMINUS,
	
};
/* parsegroups
!exp
-exp
(exp)
ident
num
[vec stuf]

low prec to high:
|

&&

!=
=

<
<=
>
>=

+
-*/
static const int num_levels = 6;
TokenID precedence_levels[][5] = 
{
	//{COMMA,NONE},
	{OR,NONE},
	{AND,NONE},
	{NOTEQUAL,EQUAL,NONE},
	{LESS,GREAT,LESSE,GREATE,NONE},
	{PLUS,MINUS,NONE},
	{TIMES,DIV,MOD,NONE}
};

struct Token {
	TokenID tok;
	std::string text;
};


static Exp::ctx builtin_funcs;
static Exp::ctx empty_context;
static IObject lambdaForExp(const std::string & str) {
   return interpret_compile(str)->eval(empty_context);
}
static void initialize_builtins() {
    builtin_funcs["map"] = IObject(IObjectLib::map);
    builtin_funcs["if"] = IObject(IObjectLib::ifthenelse);
    builtin_funcs["foldl"] = IObject(IObjectLib::foldl);
    builtin_funcs["foldl1"] = IObject(IObjectLib::foldl1);
    builtin_funcs["size"] = IObject(IObjectLib::size);
    builtin_funcs["int"] = IObject(IObjectLib::toInt);
    builtin_funcs["double"] = IObject(IObjectLib::toDouble);
    builtin_funcs["dot"] = IObject(IObjectLib::dot);
    builtin_funcs["sqrt"] = IObject(IObjectLib::sqrt);
    builtin_funcs["cross"] = IObject(IObjectLib::cross);
    builtin_funcs["isNaN"] = IObject(IObjectLib::isNaN);
    
    builtin_funcs["sum"] = lambdaForExp("{ foldl1({ _ 0 + _ 1}, _ ) }");
    builtin_funcs["avg"] = lambdaForExp("{ sum(_)/double(size(_)) }");
    builtin_funcs["min"] = lambdaForExp("{ foldl1({if(_ 0 < _ 1, _ 0, _ 1)}, _) }");
    builtin_funcs["max"] = lambdaForExp("{ foldl1({if(_ 0 > _ 1, _ 0, _ 1)}, _) }");
    builtin_funcs["any"] = lambdaForExp("{ foldl({ _ 0 || _ 1},0, map _ ) }");
    builtin_funcs["all"] = lambdaForExp("{ foldl({ _ 0 && _ 1},1, map _ ) }");
    builtin_funcs["length"] = lambdaForExp("{sqrt(dot(_,_))}");
}
IObject Exp::eval() const { ctx empty; return eval(empty); }

struct Binary : public Exp {
	Binary(ExpRef l_, TokenID o, ExpRef r_)
	: l(l_), op(o), r(r_) {}
	virtual void print() {
		std::cout << "(";
		l->print();
		std::cout << token_str[op];
		r->print();
		std::cout << ")";
	}
	virtual IObject eval(const ctx & c) const {
		IObject le = l->eval(c);
		IObject re = r->eval(c);
		switch(op) {
			case TIMES:
				return le * re;
			case DIV:
				return le / re;
			case MOD:
				return le % re;
			case PLUS:
				return le + re;
			case MINUS:
				return le - re;
			case LESS:
				return le < re;
			case GREAT:
				return le > re;
			case GREATE:
				return le >= re;
			case LESSE:
				return le <= re;
			case EQUAL:
				return le == re;
			case NOTEQUAL:
				return le != re;
			case AND:
				return le && re;
			case OR:
				return le || re;
			case APPLY:
				return le(re);
			default:
				assert(false);
				return IObject();
		}
	}
	ExpRef l;
	TokenID op;
	ExpRef r;
};
struct Unary : public Exp {
	Unary(TokenID o, ExpRef l_) : op(o),l(l_) {}
	virtual void print() {
		std::cout << token_str[op];
		l->print();
	}
	virtual IObject eval(const ctx & c) const {
		switch(op) {
			case NOT:
				return !l->eval(c);
			case UMINUS:
				return -l->eval(c);
			default:
				assert(false);
				return IObject();
		}
	}
	TokenID op;
	ExpRef l;
};
struct NumLiteral : public Exp {
	NumLiteral(const std::string & n) : num(n) {
		const char * cstr = n.c_str();
		char * end;
		int i = strtol(cstr,&end,10);
		if(!*end) {
			new (&v) IObject(i);
		} else {
			double d = strtod(cstr,&end);
			assert(!*end);
			new (&v) IObject(d);
		}
	}
	virtual IObject eval(const ctx & c) const { return v; }
	virtual void print() {
		std::cout << num;
	}
	std::string num;
	IObject v;
};
struct Var : public Exp {
	Var(const std::string & n) : name(n) {}
	virtual IObject eval(const ctx & c) const {
        if(builtin_funcs.size() == 0)
            initialize_builtins();
                
        ctx::const_iterator it  = builtin_funcs.find(name);
        if(it != builtin_funcs.end())
            return it->second;
                
        it = c.find(name);
		if(it != c.end())
			return it->second;
                
        return IObject();
	}
	virtual void print() {
		std::cout << name;
	}
	std::string name;
};

struct Lambda : public Exp {
	Lambda(const ExpRef e) 
	: exp(e) {}
	virtual void print() {
		std::cout << "{";
		exp->print();
		std::cout << "}";
	}
	virtual IObject eval(const ctx & c_) const {
		return IObject(ClosureRef(new Closure(c_,exp)));
	}
	ExpRef exp;
};
struct VecLiteral : public Exp {
        VecLiteral() {
        }
	virtual void print() {
		std::cout << "[";
                for(size_t i = 0; i < vec.size(); i++) {
                    vec[i]->print();
                    std::cout << ", ";
                }
		std::cout << "]";
	}
	virtual IObject eval(const ctx & c) const {
		IObject::vec_type v(vec.size());
                for(size_t i = 0; i < vec.size(); i++) {
                    v[i] = vec[i]->eval(c);
                    if(v[i].isUnknown())
                        return v[i];
                }
		return IObject(v);
	}
        std::vector<ExpRef> vec;
};
struct Error : public Exp {
	virtual void print() {
		std::cout << "ERR";
	}
	virtual IObject eval(const ctx & c) const { return IObject(); }
	virtual bool isError() { return true; }
};

int interpreter_lex(const std::string & stream, std::vector<Token> * result) {
	unsigned int pos = 0;
	#define ENSURE_NEXT do {if(pos + 1 >= stream.size()) return -1;} while(0)
	#define ADD_TOK(t) do { pos++; Token tok = { t, "" }; result->push_back(tok); } while(0)
	while(pos < stream.size()) {
	
	switch(stream[pos]) {
		case '!':
			ENSURE_NEXT;
			if(stream[pos+1] == '=') {
				ADD_TOK(NOTEQUAL);
				pos++;
			} else {
				ADD_TOK(NOT);
			}
			break;
		case '*':
			ADD_TOK(TIMES);
			break;
		case '/':
			ADD_TOK(DIV);
			break;
		case '%':
			ADD_TOK(MOD);
			break;
		case '+':
			ADD_TOK(PLUS);
			break;
		case '-':
			ADD_TOK(MINUS);
			break;
		case '<':
			ENSURE_NEXT;
			if(stream[pos+1] == '=') {
				pos++;
				ADD_TOK(LESSE);
			} else ADD_TOK(LESS);
			break;
		case '>':
			ENSURE_NEXT;
			if(stream[pos+1] == '=') {
				pos++;
				ADD_TOK(GREATE);
			} else ADD_TOK(GREAT);
			break;
		case '=':
			ADD_TOK(EQUAL);
			break;
		case '&':
			ENSURE_NEXT;
			if(stream[pos+1] != '&')
				return -1;
			pos++;
			ADD_TOK(AND);
			break;
		case '|':
			if(stream[pos+1] != '|')
				return -1;
			pos++;
			ADD_TOK(OR);
			break;
		case '(':
			ADD_TOK(LPAREN);
			break;
		case ')':
			ADD_TOK(RPAREN);
			break;
		case '{':
			ADD_TOK(LBRACE);
			break;
		case '}':
			ADD_TOK(RBRACE);
			break;
		case '$':
			ADD_TOK(APPLY);
			break;
		case ',':
			ADD_TOK(COMMA);
			break;
		case '~':
			ADD_TOK(UMINUS);
			break;
		default: 
			char c = stream[pos];
			const char * cstr = stream.c_str() + pos;
			char * end;
			strtod(cstr,&end);
			if(end != cstr) {
				Token tok = { NUM, stream.substr(pos,end - cstr) };
				result->push_back(tok);
				pos += end - cstr;
			} else if(isalpha(c) || c =='_') {
				int start = pos++;
				while(pos < stream.size() && (isalnum(stream[pos]) || stream[pos] == '_'))
					pos++;
				Token tok = { IDENT, stream.substr(start,pos - start) };
				result->push_back(tok); 
			} else if(isspace(c))
				pos++;
			else
				return -1;
				
	}
	}
	return 0;
}

typedef std::vector<Token>::const_iterator token_ptr;

static ExpRef interpreter_parse_high(token_ptr & start, const token_ptr end);
static ExpRef interpreter_parse_apply(token_ptr & start, const token_ptr end);
static ExpRef interpreter_parse_exp(token_ptr & start, const token_ptr end);
static ExpRef interpreter_parse_binary(token_ptr & start, const token_ptr end, int prec);
static ExpRef interpreter_parse_vec_literal(token_ptr & start, const token_ptr end);

static bool is_first_of_high(TokenID i) {
	return (i == UMINUS || i == NOT || i == LPAREN || i == NUM || i == IDENT || i == LBRACE);
}
static ExpRef interpreter_parse_high(token_ptr & start, const token_ptr end) {
	if(start->tok == UMINUS || start->tok == NOT) {
		TokenID t = start->tok;
		ExpRef exp = interpreter_parse_high(++start,end);
		if(exp->isError()) return exp;
		return ExpRef(new Unary(t,exp));
	} else if(start->tok == LPAREN) {
		if(start + 1 != end && start[1].tok == RPAREN) {
			start += 2;
			return ExpRef(new VecLiteral());
		}
		ExpRef exp = interpreter_parse_exp(++start,end);
		if(exp->isError()) return exp;
		if(start == end || start->tok != RPAREN)
			return ExpRef(new Error());
		++start;
		return exp;
	} else if(start->tok == LBRACE) {
		ExpRef exp = interpreter_parse_exp(++start,end);
		if(start == end || start->tok != RBRACE)
			return ExpRef(new Error());
		++start;
		return ExpRef(new Lambda(exp));
	}else if(start->tok == IDENT) {
		return ExpRef(new Var(start++->text));
	} else if(start->tok == NUM) {
		return ExpRef(new NumLiteral(start++->text));
		++start;
	} else return ExpRef(new Error());
}
static ExpRef interpreter_parse_apply(token_ptr & start, const token_ptr end) {
	ExpRef exp = interpreter_parse_high(start,end);
	if(exp->isError()) return exp;
	while(start != end && is_first_of_high(start->tok)) {
		ExpRef exp2 = interpreter_parse_high(start,end);
		if(exp2->isError())
			return exp2;
		exp = ExpRef(new Binary(exp,APPLY,exp2));
	}
    return exp;
}
static ExpRef interpreter_parse_binary(token_ptr & start, const token_ptr end, int prec) {
	if(prec >= num_levels)
		return interpreter_parse_apply(start,end);
	ExpRef exp = interpreter_parse_binary(start,end,prec+1);
	
	if(exp->isError()) return exp;
	
	TokenID * ops = precedence_levels[prec];
	while(*ops != NONE && start != end) {
		if(*ops == start->tok) {
			ExpRef exp2 = interpreter_parse_binary(++start,end,prec+1);
			if(exp2->isError())
				return exp2;
			exp = ExpRef(new Binary(exp,*ops,exp2));
			ops = precedence_levels[prec];
		} else
			ops++;
	}
	return exp;
}
static ExpRef interpreter_parse_vec_literal(token_ptr & start, const token_ptr end) {
    ExpRef exp = interpreter_parse_binary(start,end,0);
    if(exp->isError()) return exp;
    if(start != end && start->tok == COMMA) {
        VecLiteral * v = new VecLiteral();
        ExpRef vexp(v);
        v->vec.push_back(exp);
        do {
            ExpRef exp2 = interpreter_parse_binary(++start,end,0);
            if(exp2->isError())
                return exp2;
            v->vec.push_back(exp2);
        } while(start != end && start->tok == COMMA);
        return vexp;
    } else return exp;
}
static ExpRef interpreter_parse_exp(token_ptr & start, const token_ptr end) {
	if(start == end)
		return ExpRef(new Error());
	ExpRef exp = interpreter_parse_vec_literal(start,end);
	return exp;
}

static ExpRef interpreter_parse(const std::vector<Token> & result) {
	token_ptr start = result.begin();
	ExpRef exp = interpreter_parse_exp(start,result.end());
	if(start != result.end())
		return ExpRef(new Error());
	else
		return exp;
}

ExpRef interpret_compile(const std::string & str) {
	std::vector<Token> toks;
	if(interpreter_lex(str,&toks))
		return ExpRef(new Error());
	else
		return interpreter_parse(toks);
}
