#ifndef _MESH_IO_FACETEDGEITERATORS_H_
#define _MESH_IO_FACETEDGEITERATORS_H_
#include "FacetEdgeBuilder.h"

namespace MeshIO {

class VisitorTags {  // Simplified, not caring about overflow
public:
    typedef id_t tag_type;
    static VisitorTags& instance() {
        static VisitorTags instance_;
        return instance_;
    }
    tag_type alloc() {
    	if(next_tag == MeshIO::MAX_VISIT_TAG) 
    		LOG_ERROR("tags have overflowed, we need a bigger tag type");
        return next_tag++;
    }
private:
    VisitorTags() : next_tag(1) { }
    tag_type next_tag;
};

struct IterTypes {
	typedef FacetEdge::HalfFacet HF;
	typedef std::vector<HF *> list_t;
};

template<typename Self>
class DFSIterBase : public IterTypes {
public:
	DFSIterBase() 
	: tag(VisitorTags::instance().alloc()) {}
	void operator()(int d, FacetEdgeBuilder * b, HF * hf, list_t * l) {
		dual = d;
		stk = l;
		stk->clear();
		mesh = b;
		init = hf;
		next_to_scan = 0;
		visit(init);
		me()->run();
	}
protected:
	void visit(HF * value) {
		if(!me()->key(value)->visit(tag))
			stk->push_back(value);
	}
	bool nextUnscanned(HF ** hf) {
		if(next_to_scan < stk->size()) {
			*hf = (*stk)[next_to_scan++];
			return true;
		} else return false;
	}
	Self * me() {
		return static_cast<Self*>(this);
	}
	size_t next_to_scan;
	FacetEdgeBuilder * mesh;
	list_t * stk;
	HF * init;
	VisitorTags::tag_type tag;
	int dual;
};

//to use DFSIterBase, you need to implement key() which takes a HF and returns 
//the canonical HF for the element being searched over


struct edges_v : public DFSIterBase<edges_v> {
	void run() {
		HF * fe;
		while(nextUnscanned(&fe)) {
			HF * it = fe;
			do {
				it = it->ccwAroundEF[dual];
				HF * fe2 = it->ccwAroundEF[!dual]->flip();
				assert(fe2->vertOrCell(dual) == init->vertOrCell(dual));;
				visit(fe2);
			} while(it != fe);
		}
	}
	HF * key(HF * e) {
		return mesh->canonicalFacet(dual ? FACE_T : EDGE_T, e->edgeOrFace(dual));
	}
};

struct vertices_v : public IterTypes {
	void operator()(int dual, FacetEdgeBuilder * b, HF * hf, list_t * l) {
		l->clear();
		edges_v()(dual,b,hf,l);
		for(size_t i = 0; i < l->size(); i++)
			(*l)[i] = (*l)[i]->flip();
	}
};


struct faces_v : public DFSIterBase<faces_v> {
	void run() {
		HF * fe;
		while(nextUnscanned(&fe)) {
			visit(fe->ccwAroundEF[dual]);
			visit(fe->flip()->ccwAroundEF[dual]->flip());
			
			HF * fe2 = fe->ccwAroundEF[!dual]->flip();
			assert(fe2->vertOrCell(dual) == init->vertOrCell(dual));
			
			visit(fe2->ccwAroundEF[dual]);
			visit(fe2->flip()->ccwAroundEF[dual]->flip());
		}
	}
	HF * key(HF * e) {
		return mesh->canonicalFacet(dual ? EDGE_T : FACE_T, e->edgeOrFace(!dual));
	}
};

struct cells_v : public DFSIterBase<cells_v> {
	void run() {
		HF * fe;
		while(nextUnscanned(&fe)) {
			HF * it = fe;
			do {
				it = it->ccwAroundEF[dual];
				visit(it);
				it = it->ccwAroundEF[!dual]->flip();
			} while(it != fe);
		}
	}
	HF * key(HF * e) {
		return mesh->canonicalFacet(dual ? VERTEX_T : CELL_T, e->vertOrCell(!dual));
	}
	
};

struct vertices_e : public IterTypes {
	void operator()(int dual, FacetEdgeBuilder * b, HF * hf, list_t * l) {
		l->clear();
		l->push_back(hf);
		l->push_back(hf->flip());
	}
};

struct facets_e : public IterTypes {
	void operator()(int dual, FacetEdgeBuilder * b, HF * hf, list_t * l) {
		l->clear();
		HF * it = hf;
		do {
			l->push_back(it);
			it = it->ccwAroundEF[dual];
		} while(hf != it);
	}
};

struct facets_v : public DFSIterBase<facets_v> {
	void run() {
		HF * hf;
		while(nextUnscanned(&hf)) {
			visit(hf->ccwAroundEF[dual]);
			visit(hf->flip()->ccwAroundEF[dual]->flip());
			visit(hf->ccwAroundEF[!dual]->flip());
		}
	}
	HF * key(HF * e) {
		return e->isOther() ? e->flip() : e;
	}
};

typedef facets_e cells_e; 
typedef facets_e faces_e;

//returns true if all of the elements facets are present (i.e. there are no holes in the fe-rings associated with the element
struct has_all_facets_e : public IterTypes {
	void operator()(int dual, FacetEdgeBuilder * b, HF * hf, list_t * l) {
		
		HF * it = hf;
		do {
			id_t head = it->vertOrCell(!dual);
			it = it->ccwAroundEF[dual];
			id_t tail = it->flip()->vertOrCell(!dual);
			if(head != tail) { //this is an invalid link, so we don't have all the facet edges for this face
				//LOG_DBG("%d:%d: head = %d tail = %d\n",dual,(int)hf->edgeOrFace(dual),(int)head,(int)tail);
				result = false;
				return;
			}
		} while(hf != it);
		result = true;
		return;
	}
	bool result;
};

struct has_all_facets_v : public DFSIterBase<facets_v> {
	bool valid_link(int dual, HF * hf) {
		HF * hf2 = hf->ccwAroundEF[dual]->flip();
		return hf->vertOrCell(!dual) == hf2->vertOrCell(!dual);
	}
	
	void run() {
		HF * hf;
		while(nextUnscanned(&hf)) {
			if(!valid_link(dual,hf)) {
				result = false;
				return;
			}
			visit(hf->ccwAroundEF[dual]);
			
			if(!valid_link(dual,hf->flip())) {
				result = false;
				return;
			}
			
			visit(hf->flip()->ccwAroundEF[dual]->flip());
			
			if(!valid_link(!dual,hf)) {
				result = false;
				return;
			}
			visit(hf->ccwAroundEF[!dual]->flip());
		}
		result = true;
	}
	HF * key(HF * e) {
		return e->isOther() ? e->flip() : e;
	}
	bool result;
};

} //namespace MeshIO

#endif