#ifndef _VTK_PARSE_H
#define _VTK_PARSE_H

#include "MeshIO/LisztFileWriter.h"
#include "MeshIO/LisztFormat.h"
#include "MeshIO/EdgeTable.h"
#include <libgen.h>
#include <vtkPointSet.h>
#include <vtkCell.h>
#include <vtkSmartPointer.h>
#include <vtkPolyData.h>
#include <vtkUnstructuredGrid.h>
#include <vtkTriangleStrip.h>
#include <vtkCellArray.h>
#include <vtkIdList.h>

#include<vtkFacetReader.h> 
//.facet
#include<vtkOBJReader.h>
//.obj
#include<vtkPLYReader.h>
//.ply
#include<vtkSTLReader.h>
//.stl
#include<vtkUGFacetReader.h>
//.fac
#include<vtkExodusReader.h>
//.ex2
#include<vtkXMLPolyDataReader.h>
//'.vtp'
#include<vtkXMLUnstructuredGridReader.h> 
//'.vtu'
#include<vtkAVSucdReader.h>
//'.inp'
#include<vtkChacoReader.h> 
//'.graph'
#include<vtkGAMBITReader.h>
//'.neu'
#include<vtkUnstructuredGridReader.h>
//'.vtk'
#include<vtkPointSet.h>
#include<vtkPolyData.h>
#include<vtkUnstructuredGrid.h>
#include<vtkSmartPointer.h>

#define VP vtkSmartPointer

bool endsWith(const std::string & filename, const std::string & ending);

namespace MeshIO {


VP<vtkPointSet> loadVTKFromFile(const std::string & filename) {
#define MAKE_AND_RETURN(clazz) do { \
		VP<clazz> obj = VP<clazz>::New(); \
		obj->SetFileName(filename.c_str()); \
		obj->Update(); \
		if(obj->GetErrorCode()) { \
			printf("error code = %d\n",(int) obj->GetErrorCode()); \
			return NULL; \
		} \
		VP<vtkPointSet> ret = obj->GetOutput(); \
		return ret; \
	} while(0);
	
	if(endsWith(filename,".facet")) {
		MAKE_AND_RETURN(vtkFacetReader);
	} else if(endsWith(filename,".obj")) {
		MAKE_AND_RETURN(vtkOBJReader);
	} else if(endsWith(filename,".ply")) {
		MAKE_AND_RETURN(vtkPLYReader);
	} else if(endsWith(filename,".stl")) {
		MAKE_AND_RETURN(vtkSTLReader);
	} else if(endsWith(filename,".fac")) {
		MAKE_AND_RETURN(vtkUGFacetReader);
	} else if(endsWith(filename,".ex2")) {
		MAKE_AND_RETURN(vtkExodusReader);
	} else if(endsWith(filename,".vtp")) {
		MAKE_AND_RETURN(vtkXMLPolyDataReader);
	} else if(endsWith(filename,".vtu")) {
		MAKE_AND_RETURN(vtkXMLUnstructuredGridReader);
	} else if(endsWith(filename,".inp")) {
		MAKE_AND_RETURN(vtkAVSucdReader);
	} else if(endsWith(filename,".graph")) {
		char * name = strdup(filename.c_str());
		vtkChacoReader * obj = VP<vtkChacoReader>::New();
		obj->SetBaseName(dirname(name));
		obj->Update();
		vtkPointSet * ret = obj->GetOutput();
		obj->Delete();
		free(name);
		return ret;
	} else if(endsWith(filename,".neu")) {
		MAKE_AND_RETURN(vtkGAMBITReader);
	} else if(endsWith(filename,".vtk")) {
	    MAKE_AND_RETURN(vtkUnstructuredGridReader);
	} else {
		printf("meshconv does not understand mesh: %s\n",filename.c_str());
		return NULL;
	}
#undef MAKE_AND_RETURN

}

class VTKParse {
private:
	//output stuff
	LisztHeader* header;
	EdgeTable edge_table;
	EdgeTable face_table;
	LisztFileWriter* writer;
    FileField position_field;
    FieldTableIndex index;

public:
	VTKParse() {}
	void init(LisztFileWriter* wr, const char* filename, const std::string & outfile) {
		writer = wr;
		header = &writer->header;
		data_ = loadVTKFromFile(filename);
		if(data_ == NULL) {
			printf("Failed to load mesh into vtk: %s\n",filename);
			assert(!"VTK Failed to load mesh, is it a supported format?");
		}
		writer->init(outfile);
		header->nV = 0;
		header->nE = 0;
		header->nF = 0;
		header->nC = 0;
		header->nFE = 0;

        // initialize position field
        position_field.domain        = VERTEX_T;
        position_field.range.type    = LISZT_DOUBLE;
        position_field.range.flags   = LISZT_VEC_FLAG;
        position_field.range.data[0] = 3;

        index.num_fields = 1;
	}
	void parseAll() {
	
		writer->setBeginData();
		header->facet_edge_table = writer->currentPosition();
		
		vtkPolyData * pd = dynamic_cast<vtkPolyData*>(data_.GetPointer());
		if(pd) {
			printf("found 2d\n");
			is2D = true;
			parse2D(pd);
		} else {
			printf("found 3d\n");
			is2D = false;
			vtkUnstructuredGrid * gd = dynamic_cast<vtkUnstructuredGrid*>(data_.GetPointer());
			assert(gd);
			parse3D(gd);
		}
		
		//position field
        position_field.data   = writer->currentPosition();
        position_field.nElems = data_->GetNumberOfPoints();
		int npoints = data_->GetNumberOfPoints();
		for(int i = 0; i < npoints; i++) {
			double p[3];
			data_->GetPoint(i,p);
			for(int j = 0; j < 3; j++) {
				double d = p[j];
				writer->writeValue(d);
			}
		}

        // Finalize field table info!
        position_field.name = writer->currentPosition();
        writer->writeString("position");

        header->field_table_index = writer->currentPosition();
        writer->writeValue(index);
        writer->writeValue(position_field);
	}
	
	void printC(VP<vtkCell> cell) {
		printf("type = %d\n",cell->GetCellType());
		int npoints = cell->GetNumberOfPoints();
		int nedges = cell->GetNumberOfEdges();
		int nfaces = cell->GetNumberOfFaces();
		
		printf("ne = %d, nf = %d points: ",nedges,nfaces);
		for(int j = 0; j < npoints; j++) {
			printf("%d ",(int)cell->GetPointId(j));
		}
		printf("\n");
	}
	
	void parse2D(vtkPolyData * data) {
		data->BuildLinks();
		vtkIdType npoints = data->GetNumberOfPoints();
		vtkIdType nfaces = data->GetNumberOfCells();
		header->nF = nfaces;
		for(int i = 0; i < nfaces; i++) {
			VP<vtkCell> face = data->GetCell(i);
			add2D(face,i,0,1);
		}
		header->nV = npoints;
		header->nE = edge_table.size();
		header->nC = 2;
	}
	
	void parse3D(vtkUnstructuredGrid * data) {
		writer->setBeginData();
		
		//do work
		
		data->BuildLinks();
		vtkIdType npoints = data->GetNumberOfPoints();
		vtkIdType ncells = data->GetNumberOfCells();
		
		printf("npoints = %d",(int)npoints);
		printf("ncells = %d",(int)ncells);
		//loop over cells, then over faces, if we have not seen the face (unique (cell,cell) pair)
		//then add it using add2D
		header->nV = npoints;
		header->nC = ncells + 1;
		int next_fake_cell = header->nC;
		for(int i = 0; i < ncells; i++) {
			VP<vtkCell> cell = data->GetCell(i);
			assert(cell->GetCellDimension() == 3);
			//printf("cell %d ",i);
			printC(cell);
			int nfaces = cell->GetNumberOfFaces();
			for(int f = 0; f < nfaces; f++) {
				//printf("face f: ");
				VP<vtkCell> face = cell->GetFace(f);
				printC(face);
				VP<vtkIdList> points = face->GetPointIds();
				VP<vtkIdList> cell_neighbors = VP<vtkIdList>::New();
				data->GetCellNeighbors(i,points,cell_neighbors);
				assert(cell_neighbors->GetNumberOfIds() < 2);
				for(int n = 0; n < cell_neighbors->GetNumberOfIds(); n++) {
					//printf("neighbor cell = %d\n",cell_neighbors->GetId(n));
				}
				
				
				id_t outside = cell_neighbors->GetNumberOfIds() > 0 ?
				               cell_neighbors->GetId(0) + 1 : 0;
				id_t inside = i + 1;
				
				if(outside == 0) {
					//faces on the outside have never been seen before, but multiple faces
					//may point outside, so they each must be considered unique
					//we generate fake cell names for the outside cell here to guarentee
					//that face_table generates a unique id for this face
					id_t face_id = face_table.get(inside,next_fake_cell++);
					add2D(face,face_id,outside,inside);
				} else {
					bool new_face; //have we seen this face before?
					id_t face_id = face_table.get(inside,outside,&new_face);
					if(new_face) {
						add2D(face,face_id,outside,inside);
					}
				}
			}
			//printf("\n");
		}	
		header->nE = edge_table.size();
		header->nF = face_table.size();
	}
private:
	void add2D(VP<vtkCell> face,id_t face_id, id_t outside, id_t inside) {
		assert(face->GetCellDimension() == 2);
		VP<vtkIdList> vert = face->GetPointIds();
		add2D(face->GetCellType(),face_id,outside,inside,vert->GetNumberOfIds(),vert->GetPointer(0));
	}
	void add2D(int type, id_t face_id, id_t outside, id_t inside,int nverts, vtkIdType * verts) {
		vtkIdType pixel_buf[4];
		switch(type) {
			case VTK_PIXEL: {
				//reorder to make it CCW
				assert(nverts == 4);
				int order[] = { 0, 2, 3, 1};
				for(int i = 0; i < 4; i++) {
					pixel_buf[i] = verts[order[i]];
				}
				verts = pixel_buf;
				//FALLTHROUGH!
			}
			case VTK_TRIANGLE:
			case VTK_QUAD:
			case VTK_POLYGON: {
				for(int i = 0; i < nverts; i++) {
					id_t tail = verts[i];
					id_t head = verts[(i+1) % nverts];
					id_t edge = edge_table.get(tail,head);
					addFacetEdge(face_id,edge,head,outside,tail,inside);
				}
			} break;
			case VTK_TRIANGLE_STRIP: {
				assert(is2D); //doesn't work for 3d
				VP<vtkCellArray> tris = VP<vtkCellArray>::New();
				vtkTriangleStrip::DecomposeStrip(nverts,verts,tris);
				tris->InitTraversal();
				vtkIdType npts;
				vtkIdType * pts;
				while(tris->GetNextCell(npts,pts)) {
					assert(npts == 3);
					add2D(VTK_TRIANGLE,face_id,outside,inside,npts,pts);
					face_id = header->nF++; //add more tris for each part of the strip, first id is the original id
				}
			} break;
			default:
				printf("unexpected type: %d",type);
				break;
		}
	}
	void addFacetEdge(id_t face, id_t edge, id_t vert0, id_t cell0, id_t vert1, id_t cell1) {
		FileFacetEdge fe;
		fe.face = face;
		fe.edge = edge;
		fe.hf[0].vert = vert0;
		fe.hf[0].cell = cell0;
		fe.hf[1].vert = vert1;
		fe.hf[1].cell = cell1;
		//printf("fe = %d, f = %d, e = %d,  (v = %d, c = %d) (v = %d, c = %d)\n",
	   	//header->nFE,fe.face,fe.edge, fe.hf[0].vert, fe.hf[0].cell, fe.hf[1].vert, fe.hf[1].cell);
	
		writer->writeObject(&fe);
		header->nFE++;
	}
	bool is2D;
	VP<vtkPointSet> data_;
};
} // namespace MeshIO
#endif /* _VTK_PARSE_H */
