#ifndef _PARALLEL_LOAD_FIELD_COMMUNICATION_PATTERN_H
#define _PARALLEL_LOAD_FIELD_COMMUNICATION_PATTERN_H

#include "Common.h"
#include <vector>
#include <iostream>

class CommInfo {
    public:
        struct Message {
            int rank;
            uint32_t * values;
            size_t size;
            size_t buf_offset;
        };
        CommInfo() : data(NULL) {}
        void init(uint32_t * data_, const int * data_size, const int * data_disp, size_t n) {
            
            size_t buf_offset = 0;
            for(size_t i = 0; i < n; i++) {
                if(data_size[i] > 0) {
                    Message m = { i, data_ + data_disp[i], data_size[i],  buf_offset};
                    messages.push_back(m);
                    buf_offset += data_size[i];
                }
            }
            data.reset(data_,buf_offset); //take ownership of data_ array
            print();
        }
        void print() const {
            using namespace std;
            for(size_t i = 0; i < messages.size(); i++) {
                std::cout << "node " << messages[i].rank << ": ";
                for(size_t j = 0; j < messages[i].size; j++) {
                    std::cout << messages[i].values[j] << " ";
                }
                std::cout << std::endl;
            }
        }
        template<typename GlobalToLocal>
        void translateID(MeshIO::IOElemType typ, GlobalToLocal & func) {
            for(size_t i = 0; i < data.size(); i++)
                data[i] = func(typ,data[i]);
        }
        Array<uint32_t> data;
        std::vector<Message> messages;
        typedef std::vector<Message>::iterator message_iterator;
};

struct CommPair {
    CommInfo send;
    CommInfo recv;
};


struct FieldCommunicationPattern {
	bool reads_ghosts,writes_ghosts;
	shared_ptr<CommPair> reader;
	shared_ptr<CommPair> writer;
	
	void init(shared_ptr<LocalTopology> local_topology, FieldAccess * accesses) {
		reader.reset(new CommPair);
		writer.reset(new CommPair);
		LOG_LOG("Ghost reads");
		registerComm(accesses->field_type,local_topology, accesses->ghost_reads,&reader->recv,&reader->send);
		LOG_LOG("ghost writes");
		registerComm(accesses->field_type,local_topology, accesses->ghost_writes,&writer->send,&writer->recv);
		LocalTopology::GlobalToLocalFilter filter = local_topology->globalToLocalFilter();
		
		//translate comm structures to use local IDs
		reader->send.translateID((MeshIO::IOElemType)accesses->field_type,filter);
		reader->recv.translateID((MeshIO::IOElemType)accesses->field_type,filter);
		
		writer->send.translateID((MeshIO::IOElemType)accesses->field_type,filter);
		writer->recv.translateID((MeshIO::IOElemType)accesses->field_type,filter);
		reads_ghosts = accesses->ghost_reads.size() > 0;
		writes_ghosts = accesses->ghost_writes.size() > 0;
	}
	
	void registerComm(lElementType typ,  shared_ptr<LocalTopology> local_topology, FieldAccess::set_t & commset, CommInfo * from, CommInfo * to) {
		//establish a communication pathway to keep fields up to date
		//commset is the set of ghost elements for which we are communicating values for this field, represented by global ids
		//we will look up the owners of each of these elements and then notify them that we are reading/writing
	
		//if this is a Writer->Owner update
		//from will be a communicator representing the messages sent from this node to the owners
		//to will be a communicator representing messages received by this node from the writers
	
		//if this is a Owner->Reader update
		//from will be a communicator representing the messages to be received from the owners
		//to will be a communicator representing the messages to be sent to the readers
		int size;
		MPI_Comm_size(MPI_COMM_WORLD,&size);
		int rank;
		MPI_Comm_rank(MPI_COMM_WORLD,&rank);
		size_t mpi_size = size;
	
	   //commset.size() is the total number of ghosts for this field
	   int * send_buf = new int[commset.size()];
	
	   //calcuate the size of each message
	   int * send_sizes = new int[mpi_size];
	   std::fill_n(send_sizes,mpi_size,0);
	
	   for(FieldAccess::set_t::iterator it = commset.begin(), end = commset.end();
			it != end;
			it++) {
		   ElementStatus status;
		   bool result = local_topology->lookup(typ,*it,&status);
		   assert(result && status.isOwnerKnown());
		   PartitionID r = status.owner();
		   assert(r < mpi_size && (int)r != rank);
		   send_sizes[static_cast<size_t>(r)]++;
		}
	   int * send_disp = new int[mpi_size];
	   if(MPIU_GetDispFromSize(send_sizes,mpi_size,send_disp) != (int) commset.size())
			assert(!"commset is inconsistent");
	
	   //communicate to get the recv size
	   int * recv_sizes = new int[mpi_size];
	   LOG_LOG("alltoall1");
	   MPI_Alltoall(send_sizes,1,MPI_INT,recv_sizes,1,MPI_INT,MPI_COMM_WORLD);
	
	   int * recv_disp = new int[mpi_size];
	   int recv_total_size = MPIU_GetDispFromSize(recv_sizes,mpi_size,recv_disp);
	
	   int * recv_buf = new int [recv_total_size];
	
	   //fill the send_buf with the messages
	   //we send the list of global ids that we read/write to their owner
	   int * send_pos = new int[mpi_size];
	   std::fill_n(send_pos,mpi_size,0);
	   for(FieldAccess::set_t::iterator it = commset.begin(), end = commset.end();
			it != end;
			it++) {
		   ElementStatus status;
		   bool result = local_topology->lookup(typ,*it,&status);
		   assert(result && status.isOwnerKnown());
		   PartitionID r = status.owner();
		   assert(r < mpi_size);
		   send_buf[send_disp[static_cast<size_t>(r)] + send_pos[static_cast<size_t>(r)]++] = *it;
		}
	   for(size_t i = 0; i < mpi_size; i++)
		   assert(send_pos[i] == send_sizes[i]);
	   delete [] send_pos;
	   LOG_LOG("alltoall2");
	   MPI_Alltoallv(send_buf,send_sizes,send_disp,MPI_INT,recv_buf,recv_sizes,recv_disp,MPI_INT,MPI_COMM_WORLD);
	   //recv_buf now holds the messages that will be received from the readers/writers
	
	   LOG_LOG("frominit");
	   //we use these MPI buffers to initialize the Comm objects, which take OWNERSHIP send_buf/recv_buf
	   //but only use send_sizes, and send_disp for initialization
	   from->init(reinterpret_cast<uint32_t*>(send_buf),send_sizes,send_disp,mpi_size);
	   LOG_LOG("toinit");
	   to->init(reinterpret_cast<uint32_t*>(recv_buf),recv_sizes,recv_disp,mpi_size);
	
	
	   delete [] send_sizes;
	   delete [] send_disp;
	
	   delete [] recv_disp;
	   delete [] recv_sizes;
	}
};

#endif