#ifndef _PARALLEL_LOAD_FIELD_COMMUNICATOR_H
#define _PARALLEL_LOAD_FIELD_COMMUNICATOR_H

#include "Common.h"
#include "FieldCommunicationPattern.h"
#include "runtime_util.h"

#include <iostream>


class FieldBuffer {
public:
    FieldBuffer(CommInfo * c, size_t element_size_)
    : comm(c), element_size(element_size_) {
    	buf.reset(element_size * comm->data.size());
    	status.reset(comm->messages.size());
    }
    
   void getBuffer(size_t message, char ** buffer, size_t * size) {
        *buffer = &buf[element_size * comm->messages[message].buf_offset];
        *size = element_size * comm->messages[message].size;
   }
   size_t numberOfMessages() const {
       return comm->messages.size();
   }
   void issend(size_t message) {
       PartitionID rank = comm->messages[message].rank;
       char * b;
       size_t buf_size;
       getBuffer(message,&b,&buf_size);
       MPI_Issend(b,buf_size,MPI_BYTE,rank,0,MPI_COMM_WORLD,&status[message]);
   }
   void irecvAll() {
       for(size_t i = 0; i < numberOfMessages(); i++)
           irecv(i);
   }
   void irecv(size_t message) {
       size_t size;
       char * b;
       getBuffer(message,&b,&size);
       int rank = comm->messages[message].rank;
       MPI_Irecv(b,size,MPI_BYTE,rank,0,MPI_COMM_WORLD,&status[message]);
   }
   void waitAny(size_t * idx, char ** b, size_t * nelems) {
       int index;
       MPI_Waitany(numberOfMessages(),status.begin(),&index,MPI_STATUSES_IGNORE);
       //TODO(zach): what if all messages have been received
       *idx = static_cast<size_t>(index);
       getBuffer(*idx,b,nelems);
   }
   void waitAll() {
       MPI_Waitall(numberOfMessages(),status,MPI_STATUSES_IGNORE);
   }

private:
    size_t element_size;
    CommInfo * comm;
    Array<char> buf;
    Array<MPI_Request> status;
};

class FieldCommunicator {
public:
    void init(FieldCommunicationPattern * pattern)
     {
        communication_pattern = pattern;
        send_buf = NULL;
        recv_buf = NULL;
    }

    void begin(byte * field_data,  size_t first_ghost, size_t write_ghosts_offset, lType val_type, size_t val_length, lPhase current_phase, lPhase next_phase, shared_ptr<CommPair> pair) {

        size_t type_size = lUtilTypeSize(val_type);
        size_t element_size = type_size * val_length;
        
        if(send_buf != NULL) {
            // cout << "waiting on previous send" << endl;
            send_buf->waitAll();
            delete send_buf;
            send_buf = NULL;
        }

        assert(send_buf == NULL && recv_buf == NULL);

        if(pair->recv.messages.size() > 0) {
            // cout << "Posting recv" << endl;
            //post all the receives
            recv_buf = new FieldBuffer(&pair->recv, element_size);
            recv_buf->irecvAll();
        }
        
        if(pair->send.messages.size() > 0) {
            // cout << "Posting send" << endl;
            //lUtilValue identity;
            //bool set_identity = next_phase != L_READ_ONLY && next_phase != L_WRITE_ONLY;
            //if(set_identity)
            //   lUtilValueIdentity(next_phase, val_type,&identity);
            
            send_buf = new FieldBuffer(&pair->send, element_size);
            //fill the send buffers and post the sends
            const std::vector<CommInfo::Message> & messages = pair->send.messages;
            for(size_t i = 0; i < messages.size(); i++) {
              char * buf_start;
              size_t buf_size;
              send_buf->getBuffer(i,&buf_start,&buf_size);
              size_t nids = messages[i].size;
              uint32_t * ghost_ids = messages[i].values;
              
              char * buf = buf_start;
              for(size_t j = 0; j < nids; j++) {
                  size_t offset = ghost_ids[j] + ((ghost_ids[j] >= first_ghost) ? write_ghosts_offset : 0);
                  byte * field_entry = field_data + element_size * offset;
                  memcpy(buf,field_entry, element_size);
				  //std::cout << "Send    LocalID: " << ghost_ids[j] << " has diff: " << *(int*)field_entry << std::endl;
                  //if(set_identity) {
                  //    for(size_t i = 0; i < val_length; i++)
                  //        memcpy(field_entry + type_size * i,&identity, type_size);
                  //}
                  buf += element_size;
              }
              //post the send
              send_buf->issend(i);
            }
        }

    }

    void end(byte * field_data, size_t first_ghost, size_t write_ghosts_offset, lType val_type, size_t val_length, lPhase current_phase, lPhase next_phase, shared_ptr<CommPair> pair) {
        
        size_t type_size = lUtilTypeSize(val_type);
        size_t element_size = type_size * val_length;
        
        if(pair->recv.messages.size() > 0) {
            //std::cout << "recieving from " << pair->recv.messages.size() << " nodes" << std::endl;
            assert(recv_buf != NULL);
            const std::vector<CommInfo::Message> & messages = pair->recv.messages;
            for(size_t i = 0; i < messages.size(); i++) {
                size_t idx;
                char * buf_start;
                size_t buf_size;
                recv_buf->waitAny(&idx,&buf_start,&buf_size);
                char * buf = buf_start;
                uint32_t * ghost_ids = messages[idx].values;
                size_t nids = messages[idx].size;
                for(size_t j = 0; j < nids; j++) {
                    byte * field_entry = field_data + element_size * ghost_ids[j];
                    lUtilValueReduce(field_entry, lUtilPhaseToReduction(current_phase), val_type, val_length, buf);
                    //std::cout << "Receive LocalID: " << ghost_ids[j] << " has diff: " << *(int*)field_entry << std::endl;
                    buf += element_size;
                }
            }
            delete recv_buf;
            recv_buf = NULL;
        }
        //send_buf may not be done, but we don't need to free it yet.
    }
    
    void update(byte * field_data,  size_t first_ghost, size_t write_ghosts_offset, lType val_type, size_t val_length, lPhase current_phase, lPhase next_phase) {
        //  cout << "***** " << field->name() << " *****" << endl;
        // cout << "Writers -> Owner" << endl;
        if(current_phase != L_WRITE_ONLY) { //direct writes on only allowed to local elements, so they are not sent here
            begin(field_data, first_ghost, write_ghosts_offset, val_type, val_length, current_phase, next_phase, communication_pattern->writer);
            end(field_data, first_ghost, write_ghosts_offset, val_type, val_length, current_phase, next_phase, communication_pattern->writer);
        }
        //cout << "Owners -> Readers" << endl;
        begin(field_data,first_ghost,  write_ghosts_offset, val_type, val_length, L_WRITE_ONLY, L_READ_ONLY, communication_pattern->reader);
        end(field_data, first_ghost, write_ghosts_offset, val_type, val_length,   L_WRITE_ONLY, L_READ_ONLY, communication_pattern->reader);
    }
	FieldCommunicationPattern * communication_pattern;
private:
    FieldBuffer * send_buf;
    FieldBuffer * recv_buf;
};

#endif