#ifndef _PARALLEL_LOAD_LOCAL_TOPOLOGY_H
#define _PARALLEL_LOAD_LOCAL_TOPOLOGY_H

#include "Common.h"
#include "OwnedElements.h"
#include "OwnedElement.h"
#include "FacetEdgeWithID.h"
#include "MeshIO/FacetEdgeBuilder.h"
#include "NameService.h"
#include "mesh.h"

enum ElementStatusFlags {
	ESF_HAS_TOPOLOGY = 1,
	ESF_IS_GHOST = 2,
	ESF_OWNER_KNOWN = 4,
	ESF_HAS_ID = 8
};
	
struct ElementStatus {
	ElementStatus() {
		flags_ = 0;
		local_id_ = 0xDEADBEEF;
		owner_ = 0xDEADBEEF;
	}
	bool hasTopology() const {
		return (flags_ & ESF_HAS_TOPOLOGY) != 0;
	}
	bool isGhost() const {
		return (flags_ & ESF_IS_GHOST) != 0;
	}
	bool isOwnerKnown() const {
		return (flags_ & ESF_OWNER_KNOWN) != 0;
	}
	PartitionID owner() const {
		assert(isOwnerKnown());
		return owner_;
	}
	bool hasLocalID() const {
		return (flags_ & ESF_HAS_ID) != 0;
	}
	uint32_t localID() const {
		assert(hasLocalID());
		return local_id_;
	}
	PartitionID owner_;
	unsigned int flags_;
	uint32_t local_id_;
};

class LocalTopology {
public:
	LocalTopology(PartitionID sp, shared_ptr<OwnedElements> oe)
	: self_part(sp), 
	  owned_elements(oe) {
		for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++) {
			next_local_id[i] = 0;
			num_ghost[i] = 0;
		}
		
		initOutsideCell(); //all local topologies reference the outside cell and make sure its local ID is 0 as well
		                   //its owner and owner flags will only be set if we need to know them
		for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++) {
			Array<uint32_t> & arr = owned_elements->data[i];
			for(size_t j = 0; j < arr.size(); j++) {
				OwnedElement oe;
				oe.id = arr[j];
				oe.typ = (lElementType)i;
				oe.owner = self_part;
				addOwner(&oe);
				assignID(oe.typ,oe.id);
			}
			num_owned[i] = elem_map[i].size();
		}
	}
	bool lookup(lElementType typ, uint32_t id, ElementStatus * status) {
		elem_map_t::iterator it = elem_map[typ].find(id);
		bool ret = it != elem_map[typ].end();
		if(ret) {
			*status = it->second;
		}
		return ret;
	}
	void addTopology(const FacetEdgeWithID * feid) {
		const MeshIO::FileFacetEdge & ffe = feid->fe;
		local_mesh.insert(feid->id,&feid->fe);
		
		createOrGetEntry(L_VERTEX,ffe.hf[0].vert,NULL);
		createOrGetEntry(L_CELL,ffe.hf[0].cell,NULL);
		
		createOrGetEntry(L_VERTEX,ffe.hf[1].vert,NULL);
		createOrGetEntry(L_CELL,ffe.hf[1].cell,NULL);
		
		createOrGetEntry(L_EDGE,ffe.edge,NULL);
		createOrGetEntry(L_FACE,ffe.face,NULL);
	
	}
	void addOwner(const OwnedElement * oe) {
		ElementStatus * status;
		createOrGetEntry(oe->typ,oe->id,&status);
		status->flags_ |= ESF_OWNER_KNOWN;
		status->owner_ = oe->owner;
	}
	void topologyRequested(lElementType typ, uint32_t id) {
		ElementStatus * status;
		createOrGetEntry(typ,id,&status);
		status->flags_ |= ESF_HAS_TOPOLOGY;
	}
	void ownerRequested(lElementType typ, uint32_t id) {
		ElementStatus * status;
		createOrGetEntry(typ,id,&status);	
		status->flags_ |= ESF_OWNER_KNOWN;
	}
	void addGhost(lElementType typ, uint32_t id) {
		ElementStatus * status;
		createOrGetEntry(typ,id,&status);
		if(status->isGhost() == false) {
			status->flags_ |= ESF_IS_GHOST;
			num_ghost[typ]++;
		}
		assignID(typ,id);
	}
	
	shared_ptr<OwnedElements> ownedElements() {
		return owned_elements;
	}
	MeshIO::FacetEdgeBuilder * localFacetEdgeBuilder() {
		return &local_mesh;
	}
	
	void buildLocalMesh(Mesh * mesh) {
		GlobalToLocalFilter filter(this);
		local_mesh.filter(filter);
		local_mesh.init(elem_map[L_VERTEX].size(),
		                elem_map[L_EDGE  ].size(),
		                elem_map[L_FACE  ].size(),
		                elem_map[L_CELL  ].size(),0);
		lMeshInitFromFacetEdgeBuilder(mesh,&local_mesh);
	}
	
	void buildLocalToGlobal(lElementType typ, Array<uint32_t> * output) {
		assert(next_local_id[typ] == numberLocal(typ) && 
		       "All IDs must be assigned before call to buildLocalToGlobal (call buildLocalMesh first)");
		output->reset(numberLocal(typ));
		for(elem_map_t::iterator it = elem_map[typ].begin(),
		   end = elem_map[typ].end();
		   it != end;
		   it++) {
			ElementStatus & status = it->second;
			(*output)[status.localID()] = it->first;
		}
	}
	
	void print_debug(shared_ptr<NameService> name_service) {
#if 0
		for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++) {
			MPIFieldWriter<int> field;
			char buf[32];
			sprintf(buf,"partition_%d",(int)i);
			field.init(buf,lElementTypeToIO((lElementType)i), name_service->nElems((lElementType) i));
			for(elem_map_t::iterator it = elem_map[i].begin(),
			end = elem_map[i].end();
			it != end;
			it++) {
				char typ = (i == L_VERTEX) ? 'v' :
				           (i == EDGE) ? 'e' :
				           (i == L_FACE) ? 'f' :
				           (i == L_CELL) ? 'c' : ' ';
				ElementStatus & status = it->second;
				LOG_DBG("%c%d -> local_id = %d, owner = %d, has_topology = %d",typ,(int) it->first,status.hasLocalID() ? (int) status.localID() : -1,status.isOwnerKnown() ? (int) status.owner() : -1,(int) status.hasTopology()); 
				if(status.isOwnerKnown() && status.owner() == self_part) {
					field.write(it->first,(int)self_part);
				}
			}
			field.close();
		}
#endif
	}
	size_t numberGhost(lElementType typ) {
		return num_ghost[typ];
	}
	size_t numberLocal(lElementType typ) {
		return elem_map[typ].size();
	}
	//location of the first ghost element
	size_t firstGhost(lElementType typ) {
		return num_owned[typ];
	}
	//suitable to pass to the FacetEdgeBuilder to create a local mesh
	class GlobalToLocalFilter {
	public:
		GlobalToLocalFilter(LocalTopology * lt) 
		: local_topology(lt) {}
		MeshIO::id_t operator()(MeshIO::IOElemType iotyp, MeshIO::id_t id) {
			ElementStatus * status;
			local_topology->createOrGetEntry((lElementType)iotyp,id,&status);
			if(status->hasLocalID()) {
				return status->localID();
			} else {
				local_topology->assignID((lElementType)iotyp,id);
				return status->localID();
			}
		}
	private:
		LocalTopology * local_topology;
	};
	GlobalToLocalFilter globalToLocalFilter() {
		return GlobalToLocalFilter(this);
	}
private:
	void initOutsideCell() {
		assignID(L_CELL,0);
	}
	friend class GlobalToLocalFilter;
	
	void createOrGetEntry(lElementType typ, uint32_t id, ElementStatus ** status) {
		elem_map_t::iterator it = elem_map[typ].find(id);
		if(it != elem_map[typ].end()) {
			if(status)
				*status = &it->second;
		} else {
			ElementStatus nstatus;
			nstatus.flags_ = 0;
			std::pair<uint32_t,ElementStatus> tuple(id,nstatus);
			std::pair<elem_map_t::iterator,bool> it = elem_map[typ].insert(tuple);
			if(status)
				*status = &it.first->second;
		}
	}
	void assignID(lElementType typ, uint32_t id) {
		ElementStatus * status;
		createOrGetEntry(typ,id,&status);
		if(status->hasLocalID() == false) {
			status->flags_ |= ESF_HAS_ID;
			status->local_id_ = next_local_id[typ]++;
		}
	}
	typedef __gnu_cxx::hash_map<uint32_t, ElementStatus> elem_map_t;
	elem_map_t elem_map[L_ELEMENT_TYPE_SIZE];
	
	uint32_t num_owned[L_ELEMENT_TYPE_SIZE];
	uint32_t num_ghost[L_ELEMENT_TYPE_SIZE];
	uint32_t next_local_id[L_ELEMENT_TYPE_SIZE];
	
	MeshIO::FacetEdgeBuilder local_mesh;
    PartitionID self_part;
	shared_ptr<OwnedElements> owned_elements;
};
#endif
