#ifndef _PARALLEL_LOAD_NAME_SERVICE_H
#define _PARALLEL_LOAD_NAME_SERVICE_H
#include <mpi.h>
#include <assert.h>

#include "Common.h"
#include "MeshIO/LisztFormat.h"

//name service is reponsible for determining what node has a particular piece of information
//e.g. who knows the owner of cell x?

//we simply divide the entire set of mesh elements evenly across all cores, making it possible
//for name service to operation without any cross-node communication

typedef int PartitionID;
class NameService {
public:
	NameService(const MeshIO::LisztHeader * header) {
		
		int irank,isize;
		MPI_Comm_rank(MPI_COMM_WORLD,&irank);
		MPI_Comm_size(MPI_COMM_WORLD,&isize);
		rank = irank;
		size = isize;
		
		for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++)
			distributions[i].reset(size + 1);
		fe_distribution.reset(size + 1);
		
		createEvenDistribution(header->nV,size,distributions[L_VERTEX]);
		createEvenDistribution(header->nE,size,distributions[L_EDGE]);
		createEvenDistribution(header->nF,size,distributions[L_FACE]);
		createEvenDistribution(header->nC,size,distributions[L_CELL]);
	
		createEvenDistribution(header->nFE,size,fe_distribution);
	}


	bool inLocalPartition(lElementType typ, uint32_t id) {
		uint32_t begin,end;
		localPartition(typ,&begin,&end);
		return begin <= id && id < end;
	}

	//what range of elements is partition "part" responsible for
	void partitionRange(lElementType typ,PartitionID part, uint32_t * begin, uint32_t * end) {
		assert(part < size);
		*begin = distributions[typ][part];
		*end = distributions[typ][part + 1];
	}
	void localPartition(lElementType typ, uint32_t * begin, uint32_t * end) {
		partitionRange(typ,rank,begin,end);
	}

	//what range of facet edges is this node responsible for
	void localFEPartition(uint32_t * begin, uint32_t * end) {
		*begin = fe_distribution[rank];
		*end = fe_distribution[rank+1];
	}

	//what partition is responsible for id
	PartitionID partitionFor(lElementType typ, uint32_t id) {
		size_t owner = binary_search(id, distributions[typ]);
		return owner;
	}
	PartitionID self() const {
		return rank;
	}
	size_t numPartitions() const {
		return size;
	}
	size_t nElems(lElementType typ) {
		return distributions[typ][size];
	}
private:
	//given "total" number of elements and "size" nodes, create a distribution
	//dist is a size+1 sized array that holds the start and end ranges of each individual node
	void createEvenDistribution(size_t total, size_t size, size_t * dist) {
		size_t base = total / size;
		size_t r = total % size;
		size_t cur = 0;
		for(size_t i = 0; i < size; i++) {
			dist[i] = cur;
			cur += base;
			if(i < r)
				cur++;
		}
		dist[size] = cur;
	}
	size_t binary_search(uint32_t id, const Array<size_t> & dist) {
		const size_t * data = dist.begin();
		size_t low = 0;
		size_t high = dist.size() - 1;
		size_t mid = (low + high) / 2;
		while(low < high) {
			if(id >= data[mid + 1])
				low = mid + 1;
			else if(id < data[mid])
				high = mid - 1;
			else
				break;
			mid = (low + high) / 2;
		}
		assert(id < data[mid + 1] && id >= data[mid]);
		return mid;
	}
	
	Array<size_t> distributions[L_ELEMENT_TYPE_SIZE];
	Array<size_t> fe_distribution;
	
	size_t rank;
	size_t size;
};

#endif
