#ifndef _PARALLEL_LOAD_SIZED_ALL_TO_ALL_H
#define _PARALLEL_LOAD_SIZED_ALL_TO_ALL_H
#include <mpi.h>
#include "Common.h"

template<typename T>
class SizedAllToAll {
public:
	SizedAllToAll() {
		int isize,irank;
		MPI_Comm_size(MPI_COMM_WORLD,&isize);
		mpi_size = isize;
		MPI_Comm_rank(MPI_COMM_WORLD,&irank);
		mpi_rank = irank;
		MPI_Type_contiguous(sizeof(T),MPI_CHAR, &mpi_type);
		MPI_Type_commit(&mpi_type);
	}
	void beginCount() {
		send_sizes.reset(mpi_size);
		std::fill(send_sizes.begin(),send_sizes.end(),0);
	}
	void count(int to, size_t nelems = 1) {
		assert(to < mpi_size);
		send_sizes[to] += nelems;
	}
	void beginData() {
		send_offsets.reset(mpi_size);
		int total_send = getDispFromSize(send_sizes,mpi_size,send_offsets);
		send_data.reset(total_send);
		std::fill(send_sizes.begin(),send_sizes.end(),0);
	}
	void data(int to, const T & data) {
		assert(to <= mpi_size);
		assert( send_offsets[to] + send_sizes[to] < (int) send_data.size());
		
		send_data[ send_offsets[to] + send_sizes[to]++ ] = data;
	}
	size_t endData(size_t * npartitions, const int ** recieved_offsets, const T ** recieved_data) {
		size_t count = 0;
		for(size_t i = 0; i < send_sizes.size(); i++) {
			if((int)count != send_offsets[i])
				LOG_ERROR("counted objects != added objects, %d: count = %d, offset = %d",(int)i,(int)count,send_offsets[i]);
			count += send_sizes[i];
		}
		if(count != send_data.size())
			LOG_ERROR("count != send_data.size(): %d %d\n", (int) count, (int) send_data.size()); 
		
		recv_sizes.reset(mpi_size);
		MPI_Alltoall(send_sizes,1,MPI_INT, recv_sizes, 1, MPI_INT, MPI_COMM_WORLD);
		recv_offsets.reset(mpi_size);
		int total_recv = getDispFromSize(recv_sizes,mpi_size,recv_offsets);
		recv_data.reset(total_recv);
		
		MPI_Alltoallv(send_data,
					  send_sizes,
					  send_offsets,
					  mpi_type,
					  recv_data,
					  recv_sizes,
					  recv_offsets,
					  mpi_type,
					  MPI_COMM_WORLD);
		if(npartitions)
			*npartitions = mpi_size;
		if(recieved_offsets)
			*recieved_offsets = recv_offsets;
		if(recieved_data)
			*recieved_data = recv_data;
		return total_recv;
	}
	void reset() {
		send_sizes.reset();
		send_offsets.reset();
		send_data.reset();
		
		recv_sizes.reset();
		recv_offsets.reset();
		recv_data.reset();
	}
private:
	int getDispFromSize(const int * send_sizes, size_t n, int * send_disp) {
		send_disp[0] = 0;
		for(size_t i = 1; i < n; i++)
			send_disp[i] = send_sizes[i-1] + send_disp[i-1];
		return send_disp[n-1] + send_sizes[n-1];
	}
	size_t mpi_rank;
	size_t mpi_size;
	MPI_Datatype mpi_type;
	
	Array<int> send_sizes;
	Array<int> send_offsets;
	
	Array<T> send_data;
	
	Array<int> recv_sizes;
	Array<int> recv_offsets;
	
	Array<T> recv_data;
	
};


#endif
