#ifndef _PARALLEL_LOAD_TOPOLOGY_SERVICE_H
#define _PARALLEL_LOAD_TOPOLOGY_SERVICE_H
#include<vector>
#include<ext/hash_set>

#include "NameService.h"
#include "Common.h"
#include "SizedAllToAll.h"

#include "MeshIO/LisztFormat.h"
#include "MeshIO/LisztFileReader.h"
#include "MeshIO/FacetEdgeBuilder.h"
#include "MeshIO/FacetEdgeIterators.h"

#include "FacetEdgeWithID.h"
#include "Element.h"

using ::MeshIO::FacetEdge;
using ::MeshIO::FileFacetEdge;
using ::MeshIO::LisztFileReader;
using ::MeshIO::FacetEdgeBuilder;
using ::MeshIO::LisztHeader;



struct eqptr {
	bool operator()(FacetEdge * s1, FacetEdge * s2) const {
		return s1 == s2;
	}
};

struct hashptr {
	size_t operator()(FacetEdge * e) const {
		return (size_t) e;
	}
};

//the topology service answers queries about topology
//e.g. get me all the facet edges for cell x? individual requests are aggregated for performance
class TopologyService {
public:
	TopologyService(shared_ptr<NameService> ns,
					shared_ptr<LisztFileReader> r)
	: name_service(ns), reader(r) {
		uint32_t begin;
		uint32_t end;
		const LisztHeader & header = reader->header();

		//initialize our internal partial mesh
		builder.init(header.nV,header.nE,header.nF,header.nC,header.nFE);
		
		
		LOG_LOG("TopologyService init");

		//if we are responsible for cell C, we need to have all of its Facet Edges local
		//each node reads a slice of FEs, determines who is reponsible for the elements they refer to and then
		//sends them to that node

		//read in a slice of the facet edges
		name_service->localFEPartition(&begin,&end);
		FileFacetEdge * ffes = reader->facetEdges(begin,end);
		std::vector<PartitionID> tos; 
		tos.reserve(6); //a FE can go to at most 6 unique partitions, one each for its 2 cells, 2 verts, face, and edge
		
		//we now scan through each FE and check each mesh element it refers to
		//the FE must be sent to the node that the name service says is responsible for that element.
		LOG_LOG("read fes: %d to %d",(int)begin,(int)end);
		SizedAllToAll<FacetEdgeWithID> comm;
		comm.beginCount();
		for(uint32_t id = begin; id < end; id++) {
			tos.clear();
			FileFacetEdge & fe = ffes[id - begin];
			partitionsFor(fe,&tos);
			for(size_t i = 0; i < tos.size(); i++)
				comm.count(tos[i]);
		}
		comm.beginData();
		for(uint32_t id = begin; id < end; id++) {
			tos.clear();
			FileFacetEdge & fe = ffes[id - begin];
			partitionsFor(fe,&tos);
			FacetEdgeWithID feid = { id, fe };
			for(size_t i = 0; i < tos.size(); i++)
				comm.data(tos[i],feid);
		}
		reader->free(ffes);
		const FacetEdgeWithID * recieved;
		size_t total_recv = comm.endData(NULL,NULL,&recieved);
		LOG_LOG("recieved %d fes\n",(int)total_recv);

		//insert into builder mesh
		for(size_t i = 0; i < total_recv; i++) {
			LOG_DBG("adding %d edge: %d, face: %d", (int) recieved[i].id, (int) recieved[i].fe.edge, (int) recieved[i].fe.face);
			builder.insert(recieved[i].id,&recieved[i].fe);
		}
		
	}
	void beginRequest() {
		//nothing for now...
	}
	
	typedef __gnu_cxx::hash_set<FacetEdge *,hashptr,eqptr> fe_set;
	
	//request all the facet edges for a particular element
	void request(lElementType typ, uint32_t id) {
		//buffer up unique requests
		Element e = { typ, id };
		requests.push_back(e); //stencil consults local_topology to see if it was already requested
									//stencil will also mark pending requests as present, so we do not need to work about
									//duplicate requests here that would be sent over the wire
	}
	shared_ptr< Array<FacetEdgeWithID> > endRequest() {
		
		//sort requested topology int to requests to remove TopologyService objects
		SizedAllToAll<Element> request_mpi;
		request_mpi.beginCount();
		for(elem_list::iterator it = requests.begin(),
			end = requests.end();
			it != end;
			it++) {
			request_mpi.count(name_service->partitionFor(it->typ,it->id));
		}
		request_mpi.beginData();
		for(elem_list::iterator it = requests.begin(),
			end = requests.end();
			it != end;
			it++) {
			request_mpi.data(name_service->partitionFor(it->typ,it->id),*it);
		}
		requests.clear();
		size_t n_partitions;
		const int * offsets;
		const Element * request_data;
		LOG_LOG("Beginning request send");
                int total_recv = request_mpi.endData(&n_partitions,&offsets,&request_data);
		LOG_LOG("Ending request send");
		//take all requests and assemble lists of ffe as responses.
		
		Array<fe_set> responses(n_partitions);
		for(size_t i = 0; i < n_partitions; i++) {
			const Element * begin = request_data + offsets[i];
			const Element * end = request_data + ((i + 1 == n_partitions) ? total_recv : offsets[i + 1]);
			for(const Element * it = begin; it != end; it++) {
				addToResponse(i,*it,&responses);
			}
		}
		request_mpi.reset();
		SizedAllToAll<FacetEdgeWithID> response_mpi;
		response_mpi.beginCount();
		for(size_t i = 0; i < responses.size(); i++) {
			response_mpi.count(i,responses[i].size());
		}
		response_mpi.beginData();
		for(size_t i = 0; i < responses.size(); i++) {
			for(fe_set::iterator it = responses[i].begin(),
				end = responses[i].end();
				it != end;
				it++) {
				FacetEdge * fe = *it;
				FacetEdgeWithID feid;
				getFacetEdgeWithID(fe,&feid);
				response_mpi.data(i,feid);
			}
		}
		const FacetEdgeWithID * response_data;
		LOG_LOG("Beginning request response");
		total_recv = response_mpi.endData(&n_partitions,&offsets,&response_data);
		LOG_LOG("Ending request response");
		shared_ptr< Array<FacetEdgeWithID> > ret( new Array<FacetEdgeWithID>(total_recv) );
		std::copy(response_data,response_data + total_recv, ret->begin());
		response_mpi.reset();
		return ret;
		
	}
	FacetEdgeBuilder * partitionedTopology() {
		return &builder;
	}
private:
	bool meshIs2D() {
		return builder.nC() < 3;
	}
	//determine the partitions who need 'fe'
	void partitionsFor(const FileFacetEdge & fe, std::vector<PartitionID> * tos) {
		addToSet(name_service->partitionFor(L_VERTEX,fe.hf[0].vert),tos);
		addToSet(name_service->partitionFor(L_VERTEX,fe.hf[1].vert),tos);
		
		if(!meshIs2D()) {
			addToSet(name_service->partitionFor(L_CELL,fe.hf[0].cell),tos);
			addToSet(name_service->partitionFor(L_CELL,fe.hf[1].cell),tos);
		}
		
		addToSet(name_service->partitionFor(L_EDGE,fe.edge),tos);
		addToSet(name_service->partitionFor(L_FACE,fe.face),tos);
	}
	
	void addToSet(PartitionID to, std::vector<PartitionID> * tos) {
		for(size_t i = 0; i < tos->size(); i++) {
			if((*tos)[i] == to)
				return;
		}
		tos->push_back(to);
	}

	//use our local mesh to find all facet edges for 'elem' and add them to the response sets
	void addToResponse(PartitionID from, const Element & elem, Array< fe_set > * response) {
		MeshIO::IterTypes::list_t facets;
		MeshIO::IterTypes::HF * obj = builder.elem_maps[elem.typ][elem.id];
		assert(obj != NULL);
		switch(elem.typ) {
			case L_VERTEX:
				MeshIO::facets_v()(0,&builder,obj,&facets);
				break;
			case L_EDGE:
				MeshIO::facets_e()(0,&builder,obj,&facets);
				break;
			case L_FACE:
				MeshIO::facets_e()(1,&builder,obj,&facets);
				break;
			case L_CELL:
				if(obj->cell() != 0)
					MeshIO::facets_v()(1,&builder,obj,&facets);
				else
					facets.clear();

				break;
			default:
				break;
		}
		size_t n_facets = facets.size();
		for(size_t i = 0; i < n_facets; i++) {
			(*response)[from].insert(facets[i]->me());
		}
	}

	//marshall data from facet edge builder into network object
	void getFacetEdgeWithID(const FacetEdge * fe, FacetEdgeWithID * f) const {
		f->id = fe->id;
		f->fe.hf[0].vert = fe->hf[0].vert();
		f->fe.hf[0].cell = fe->hf[0].cell();
		
		f->fe.hf[1].vert = fe->hf[1].vert();
		f->fe.hf[1].cell = fe->hf[1].cell();
		
		f->fe.edge = fe->hf[0].edge();
		f->fe.face = fe->hf[0].face();
	}
	shared_ptr<NameService> name_service;
	shared_ptr<LisztFileReader> reader;
	FacetEdgeBuilder builder;
	typedef std::vector<Element> elem_list;
	elem_list requests;
};

#endif
