#ifndef _COLORING_H
#define _COLORING_H

#include "nested_topology.h"
#include "small_set.h"

//to use this interface a runtime must ensure that lsSet lsIterator and lsElement are large enough to hold
//the following types. It should do this by including this as a member (e.g.):
// struct lsElement {
//    union {
//       lColoringElement c;
//       lPartitioningElement e; 
//    };
// }

struct lColoringElement {
	lkElement e;
};

struct lColoringSet {
	lNestedSet s;
};

struct lColoringIterator {
	lNestedIterator it;
};

struct lColoringResult {
	size_t n_colors;
	uint32_t * row_idx;
	uint32_t * values;	
};

void lColoringGenerate(size_t n_elements_max, Mesh * mesh, size_t set_size, BoundarySetReader::range_list_t * ranges, void (*stencil)(lsFunctionTable*,lsContext*), lColoringResult * result);

size_t allocateColor(uint64_t color_mask) {
	int c = ffsll(~color_mask);
	if(c == 0) {
		fprintf(stderr,"exceeded maximum of 64 colors.");
		abort();
	}
	return c - 1;
}


struct lColoringContext {
	Mesh * mesh;
	SmallSet * set_entries; //edges from element -> field_entry
	lkElement active_element;
	uint32_t active_element_entry;
	uint32_t last_id_added; //1-element cache of the last edge added (active_element,last_id). Prevents many searches through sets.
};

void createColoringVTable(lsFunctionTable * vtbl);

void lColoringGenerate(size_t n_elements_max, Mesh * mesh, size_t set_size, BoundarySetReader::range_list_t * ranges, void (*stencil)(lsFunctionTable*,lsContext*), lColoringResult * result) {
	lColoringContext sctx;
	sctx.mesh = mesh;
	sctx.set_entries = new SmallSet[set_size];
	
	double before_stencil = lWallTime(NULL);
	
	sctx.active_element_entry = 0;
	
	lsFunctionTable vtbl;
	createColoringVTable(&vtbl);
	
	for(size_t i = 0; i < ranges->size(); i++) {
		uint32_t start = (*ranges)[i].first;
		uint32_t end = (*ranges)[i].second;
		for(size_t j = start; j < end; j++) {
			sctx.active_element.data = j;
			sctx.last_id_added = n_elements_max; //invalidate the cache by setting the last_id to an id that doesn't exist
			//record all edges from this element to the field entries that it writes
			stencil(&vtbl,(lsContext*)&sctx);
			sctx.active_element_entry++;
		}
	}

	double before_coloring = lWallTime(NULL);
	
	
	size_t * colors = new size_t[set_size];
	std::vector<int> color_frequency;
	
	
	uint64_t * field_entries = new uint64_t[n_elements_max];
	bzero(field_entries, sizeof(uint64_t) * n_elements_max);
	
	for(size_t i = 0; i < set_size; i++) {
		uint64_t coloring_mask = 0;
		SmallSet::Iterator it = sctx.set_entries[i].iterator();
		int entry;
		while(it.next(&entry)) {
			coloring_mask |= field_entries[entry];
		}
		int color = allocateColor(coloring_mask);
		it = sctx.set_entries[i].iterator();
		while(it.next(&entry)) {
			field_entries[entry] |= 1 << color;
		}
		colors[i] = color;
		if(color_frequency.size() == color) {
			color_frequency.push_back(1);
		} else {
			assert(color < color_frequency.size());
			color_frequency[color]++;
		}
	}
	
	delete [] field_entries;
	delete [] sctx.set_entries;
	
	result->n_colors = color_frequency.size();
	result->row_idx = new uint32_t[color_frequency.size() + 1];
	result->row_idx[0] = 0;
	for(size_t i = 0; i < color_frequency.size(); i++) {
		result->row_idx[i+1] = result->row_idx[i] + color_frequency[i];
		color_frequency[i] = 0; //zerod here because we will use this array in the counting sort to hold offsets in to the CRS structure
	}
	result->values = new uint32_t[set_size];


	double before_shuffle = lWallTime(NULL);

	int idx = 0;
	for(size_t i = 0; i < ranges->size(); i++) {
		uint32_t start = (*ranges)[i].first;
		uint32_t end = (*ranges)[i].second;
		for(size_t j = start; j < end; j++) {
			size_t color = colors[idx];
			result->values[result->row_idx[color] + color_frequency[color]] = j;
			color_frequency[color]++;
			idx++;
		}
	}
	
	double end = lWallTime(NULL);
	
	printf("stats: stencil %f, coloring %f, shuffle %f, total %f\n", before_coloring - before_stencil, before_shuffle - before_coloring, end - before_shuffle, end - before_stencil); 

	delete [] colors;
}

//stencil generations functions



#define CTX(ctx) ((lColoringContext*)ctx)
#define IT(it) ((lColoringIterator*)it)
#define SET(s) ((lColoringSet*)s)
#define ELEM(e) ((lColoringElement*)e)

void lColoringSetGetIterator(lsContext * ctx, lsSet * set, lsIterator * iterator) {
	lNestedSetGetIterator(&SET(set)->s,&IT(iterator)->it);
}
bool lColoringIteratorNext(lsContext * ctx, lsIterator * iterator, lsElement * e) {
	return lNestedIteratorNext(&IT(iterator)->it,&ELEM(e)->e,NULL);
}
void lColoringFieldAccess(lsContext * ctx, int field_id, lElementType typ, lsElement * e, lPhase phase) {
	uint32_t field_entry = elementID(ELEM(e)->e);
	if(phase != L_READ_ONLY && field_entry != CTX(ctx)->last_id_added && !(typ == L_CELL && ELEM(e)->e.data == 0)) {
		CTX(ctx)->set_entries[CTX(ctx)->active_element_entry].insert(field_entry);
		CTX(ctx)->last_id_added = field_entry;
	}
}

void lColoringGetActiveElement(lsContext * ctx, lsElement * e) {
	ELEM(e)->e = CTX(ctx)->active_element;
}

//stencil topology functions

void lColoringVerticesOfVertex(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->vtov,ELEM(c)->e,&SET(set)->s);
}
void lColoringVerticesOfEdge(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRSConst(&CTX(ctx)->mesh->etov,ELEM(c)->e,&SET(set)->s);
}
void lColoringVerticesOfFace(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->ftov,ELEM(c)->e,&SET(set)->s);
}
void lColoringVerticesOfCell(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->ctov,ELEM(c)->e,&SET(set)->s);
}

void lColoringCellsOfVertex(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->vtoc,ELEM(c)->e,&SET(set)->s);
}
void lColoringCellsOfEdge(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->etoc,ELEM(c)->e,&SET(set)->s);
}
void lColoringCellsOfFace(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRSConst(&CTX(ctx)->mesh->ftoc,ELEM(c)->e,&SET(set)->s);
}
void lColoringCellsOfCell(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->ctoc,ELEM(c)->e,&SET(set)->s);
}

void lColoringEdgesOfVertex(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->vtoe,ELEM(c)->e,&SET(set)->s);
}
void lColoringEdgesOfCell(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->ctoe,ELEM(c)->e,&SET(set)->s);
}

void lColoringEdgesOfFaceCCW(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->ftoe,ELEM(c)->e,&SET(set)->s);
}

void lColoringEdgesOfFace(lsContext * ctx, lsElement * c, lsSet * set) {
	lColoringEdgesOfFaceCCW(ctx,c,set);
}

void lColoringEdgesOfFaceCW(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->ftoe,lNestedFlip(ELEM(c)->e),&SET(set)->s);
}

void lColoringFacesOfVertex(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->vtof,ELEM(c)->e,&SET(set)->s);
}

void lColoringFacesOfCell(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->ctof,ELEM(c)->e,&SET(set)->s);
}

void lColoringFacesOfEdgeCCW(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->etof,ELEM(c)->e,&SET(set)->s);
}

void lColoringFacesOfEdge(lsContext * ctx, lsElement * c, lsSet * set) {
	lColoringFacesOfEdgeCCW(ctx,c,set);
}

void lColoringFacesOfEdgeCW(lsContext * ctx, lsElement * c, lsSet * set) {
	lNestedElementsOfCRS(&CTX(ctx)->mesh->etof,lNestedFlip(ELEM(c)->e),&SET(set)->s);
}

void lColoringHeadOfEdge(lsContext * ctx, lsElement * c, lsElement * result) {
	ELEM(result)->e = lNestedElementOfCRSConst(&CTX(ctx)->mesh->etov, ELEM(c)->e,0); 
}
void lColoringTailOfEdge(lsContext * ctx, lsElement * c, lsElement * result) {
	ELEM(result)->e = lNestedElementOfCRSConst(&CTX(ctx)->mesh->etov, ELEM(c)->e,1); 
}

void lColoringHeadOfFace(lsContext * ctx, lsElement * c, lsElement * result) {
	ELEM(result)->e = lNestedElementOfCRSConst(&CTX(ctx)->mesh->ftoc, ELEM(c)->e,0); 
}
void lColoringTailOfFace(lsContext * ctx, lsElement * c, lsElement * result) {
	ELEM(result)->e = lNestedElementOfCRSConst(&CTX(ctx)->mesh->ftoc, ELEM(c)->e,1); 
}

void lColoringFlipEdge(lsContext * ctx, lsElement * c, lsElement * result) {
	ELEM(result)->e = lNestedFlip(ELEM(c)->e);
}
void lColoringFlipFace(lsContext * ctx, lsElement * c, lsElement * result) {
	ELEM(result)->e = lNestedFlip(ELEM(c)->e);
}

void lColoringTowardsEdge(lsContext * ctx, lsElement * e, lsElement * v, lsElement * result) {
	ELEM(result)->e = lNestedTowardsCRSConst(&CTX(ctx)->mesh->etov,ELEM(e)->e,ELEM(v)->e);
}
void lColoringTowardsFace(lsContext * ctx, lsElement * e, lsElement * v, lsElement * result) {
	ELEM(result)->e = lNestedTowardsCRSConst(&CTX(ctx)->mesh->ftoc,ELEM(e)->e,ELEM(v)->e);
}

void lColoringSetExtractElement(lsContext * ctx, lsSet * set, uint32_t idx, lsElement * result) {
	bool valid = lNestedSetGetElement(&SET(set)->s,idx,&ELEM(result)->e);
	if(!valid) {
		//no element exists in the set
		abort();
	}
}
#undef SET
#undef IT
#undef CTX
#undef ELEM

void createColoringVTable(lsFunctionTable * vtbl) {
	bzero(vtbl, sizeof(lsFunctionTable));
	
	vtbl->lsSetGetIterator = lColoringSetGetIterator;
	vtbl->lsIteratorNext = lColoringIteratorNext;
	vtbl->lsFieldAccess = lColoringFieldAccess;
	vtbl->lsGetActiveElement = lColoringGetActiveElement;

	vtbl->lsVerticesOfVertex = lColoringVerticesOfVertex;
	vtbl->lsVerticesOfEdge = lColoringVerticesOfEdge;
	vtbl->lsVerticesOfFace = lColoringVerticesOfFace;
	vtbl->lsVerticesOfCell = lColoringVerticesOfCell;
	vtbl->lsCellsOfVertex = lColoringCellsOfVertex;
	vtbl->lsCellsOfEdge = lColoringCellsOfEdge;
	vtbl->lsCellsOfFace = lColoringCellsOfFace;
	vtbl->lsCellsOfCell = lColoringCellsOfCell;
	vtbl->lsEdgesOfVertex = lColoringEdgesOfVertex;
	vtbl->lsEdgesOfFace = lColoringEdgesOfFace;
	vtbl->lsEdgesOfCell = lColoringEdgesOfCell;
	vtbl->lsEdgesOfFaceCCW = lColoringEdgesOfFaceCCW;
	vtbl->lsEdgesOfFaceCW = lColoringEdgesOfFaceCW;
	vtbl->lsFacesOfVertex = lColoringFacesOfVertex;
	vtbl->lsFacesOfEdge = lColoringFacesOfEdge;
	
	vtbl->lsFacesOfCell = lColoringFacesOfCell;
	vtbl->lsFacesOfEdgeCCW = lColoringFacesOfEdgeCCW;
	vtbl->lsFacesOfEdgeCW = lColoringFacesOfEdgeCW;
	vtbl->lsHeadOfEdge = lColoringHeadOfEdge;
	vtbl->lsTailOfEdge = lColoringTailOfEdge;
	vtbl->lsHeadOfFace = lColoringHeadOfFace;
	vtbl->lsTailOfFace = lColoringTailOfFace;
	vtbl->lsFlipEdge = lColoringFlipEdge;
	vtbl->lsFlipFace = lColoringFlipFace;
	
	vtbl->lsTowardsEdge = lColoringTowardsEdge;
	vtbl->lsTowardsFace = lColoringTowardsFace;
	vtbl->lsSetExtractElement = lColoringSetExtractElement;
}

#endif
