#ifndef _GHOST_DISCOVERY_H
#define _GHOST_DISCOVERY_H

#include "Array.h"
#include "ParallelLoad/ParallelLoad.h"
#include "boundary_set_reader.h"

//stenciling functions
struct lDiscoveryElement {
	bool valid;
	int depth; //negative depth indicates an invalid element
	uint32_t id;
	MeshIO::FacetEdge::HalfFacet * hf; //will be null when (depth == NULL)
};

struct lDiscoverySet {
	int depth; //depth == 0 is a top level set, otherwise it is nested
	lElementType type;
	bool valid;
	//for top-level sets
	BoundarySetReader::range_list_t ranges;
	uint32_t * local_to_global;

	//for nested sets
	MeshIO::IterTypes::list_t elements;
};

struct lDiscoveryIterator {
	lDiscoverySet * set;
	size_t range_index;
	size_t i;
};


//provides functionality to discovery topology and load boundary sets for locally discovered topology

void lDiscoverySetInitRanges(shared_ptr<BoundarySetReader> boundary_reader, lElementType type, const char * boundary_name, size_t n_owned, uint32_t * local_to_global, BoundarySetReader::range_list_t * ranges, size_t * global_size, size_t * local_size, int * id);
void lDiscoveryRun(void (*stencil)(lsFunctionTable*,lsContext *), lsContext * ctx, shared_ptr<NameService> name_service, shared_ptr<TopologyService> topology_service, shared_ptr<OwnerService> owner_service, shared_ptr<BoundarySetReader> boundary_reader, /*inputs*/
                 shared_ptr<LocalTopology> local_topology, shared_ptr<LocalMemory> local_memory /*outputs*/);
                 
                 
                 
//implementation follows

struct lDiscoveryContext {
	bool record_fields;
	bool requests_occured;
	int topology_depth;
	
	shared_ptr<OwnedElements> owned_elements;
	shared_ptr<LocalTopology> local_topology;
	shared_ptr<LocalMemory> local_memory;
	shared_ptr<OwnerService> owner_service;
	shared_ptr<TopologyService> topology_service;
	shared_ptr<NameService> name_service;
	shared_ptr<BoundarySetReader> boundary_reader;
	FacetEdgeBuilder * mesh;
	MeshIO::IterTypes::list_t buffer;
};

//these functions need to be implemented by the specific runtime based on the format of the ls* objects
lDiscoveryElement * lDiscoveryGetElement(lsElement * e);
lDiscoverySet * lDiscoveryGetSet(lsSet * e);
lDiscoveryIterator * lDiscoveryGetIterator(lsIterator * e);
lDiscoveryContext * lDiscoveryGetContext(lsContext * ctx);

//stencil generations functions

void lDiscoveryCreateVTable(lsFunctionTable * vtbl);

void lDiscoveryRun(void (*stencil)(lsFunctionTable*,lsContext *), lsContext * lsCtx, shared_ptr<NameService> name_service, shared_ptr<TopologyService> topology_service, shared_ptr<OwnerService> owner_service, shared_ptr<BoundarySetReader> boundary_reader, /*inputs*/
                 shared_ptr<LocalTopology> local_topology, shared_ptr<LocalMemory> local_memory /*outputs*/) {
	lDiscoveryContext & ctx = *lDiscoveryGetContext(lsCtx);
	
	ctx.record_fields = false;
	ctx.owned_elements = local_topology->ownedElements();
	ctx.local_topology = local_topology;
	ctx.local_memory = local_memory;
	ctx.name_service = name_service;
	ctx.mesh = local_topology->localFacetEdgeBuilder();
	ctx.owner_service = owner_service;
	ctx.topology_service = topology_service;
	ctx.boundary_reader = boundary_reader;
	ctx.requests_occured = true;
	lsFunctionTable vtbl;
	lDiscoveryCreateVTable(&vtbl);
	
	for(ctx.topology_depth = 0; ctx.requests_occured; ctx.topology_depth++) {
		ctx.requests_occured = false;
		topology_service->beginRequest();
		LOG_LOG("Getting topology for depth %d",ctx.topology_depth);
		stencil(&vtbl,lsCtx);
		LOG_LOG("Beginning Request");
        shared_ptr< Array<FacetEdgeWithID> > data = topology_service->endRequest();
		LOG_LOG("Ending Request");
		for(FacetEdgeWithID * cur = data->begin(); cur != data->end(); cur++) {
			local_topology->addTopology(cur);
		}
		int anyrequests;
		int myrequests = ctx.requests_occured;
		MPI_Allreduce(&myrequests,&anyrequests,1,MPI_INT,MPI_LOR, MPI_COMM_WORLD);
		ctx.requests_occured = anyrequests;
	}
	LOG_LOG("Finished Discovery Last Depth %d", ctx.topology_depth);
	
	ctx.record_fields = true;
	owner_service->beginRequest();
	stencil(&vtbl,lsCtx);
	//when we discover ghost, we also request the owner of the ghost to setup communication for the field. Add this information to the local topology table now
	shared_ptr< Array< OwnedElement > > data = owner_service->endRequest();
	for(OwnedElement * cur = data->begin(); cur != data->end(); cur++) {
		local_topology->addOwner(cur);
	}             	
}
void lDiscoverySetInitRanges(shared_ptr<BoundarySetReader> boundary_reader, lElementType type, const char * boundary_name, size_t n_owned, uint32_t * local_to_global, BoundarySetReader::range_list_t * ranges, size_t * global_size, size_t * local_size, int * id) {
	boundary_reader->load((MeshIO::IOElemType) type,boundary_name,id,ranges,global_size);
	//translate the global ranges, into local ranges. Since the global_to_local id assignment is a stable
	//ordering global ranges can be translated into equivalent local ranges
	uint32_t * lbegin = local_to_global;
	uint32_t * lend = lbegin + n_owned;
	if(local_size)
		*local_size = 0;
	LOG_LOG("initializing set %s",boundary_name);
	for(size_t i = 0; i < ranges->size(); i++) {
		LOG_LOG("range is (%d,%d)\n", (int)(*ranges)[i].first, (int)(*ranges)[i].second);
		uint32_t * begin_set = std::lower_bound(lbegin,lend, (*ranges)[i].first);
		uint32_t * end_set = std::lower_bound(lbegin,lend, (*ranges)[i].second);
		if(begin_set != end_set) {
			size_t start_pos = begin_set - lbegin;
			size_t end_pos = end_set - lbegin;
			(*ranges)[i].first = start_pos;
			(*ranges)[i].second = end_pos;
			if(local_size)
			    *local_size += end_pos - start_pos;
			LOG_LOG("translated is (%d,%d)\n", (int)start_pos, (int)end_pos);
			assert(start_pos <= end_pos);
			assert(end_pos <= n_owned);
		} else {
			(*ranges)[i].first = (*ranges)[i].second = 0;
		}
	}
	LOG_LOG("finished %s",boundary_name);
}

#define CTX(ctx) (lDiscoveryGetContext(ctx))
#define SET(set) (lDiscoveryGetSet(set))
#define IT(it) (lDiscoveryGetIterator(it))
#define ELEM(e) (lDiscoveryGetElement(e))
          
void lDiscoverySetGetIterator(lsContext * ctx, lsSet * set, lsIterator * iterator) {
	IT(iterator)->set = SET(set);
	IT(iterator)->range_index = 0;
	IT(iterator)->i = 0;
}

void lDiscoverySetGetElement(lDiscoverySet * set, uint32_t idx, lDiscoveryElement * e) {
	assert(set->depth > 0);
	if(set->valid) {
		e->valid = true;
		e->depth = set->depth;
		assert(idx < set->elements.size()); 
		e->hf = set->elements[idx];
		switch(set->type) {
			case L_VERTEX: e->id = e->hf->vert(); break;
			case L_EDGE: e->id = e->hf->edge(); break;
			case L_FACE: e->id = e->hf->face(); break;
			case L_CELL: e->id = e->hf->cell(); break;
			default: abort(); break;
		}
	} else {
		e->valid = false;
		e->depth = set->depth;
		e->id = 0;
		e->hf = NULL;
	}
}

bool lDiscoveryIteratorNext(lsContext * ctx_, lsIterator * iterator_, lsElement * e_) {
	lDiscoveryContext * ctx = CTX(ctx_);
	lDiscoveryIterator * iterator = IT(iterator_);
	lDiscoveryElement * e = ELEM(e_);
	if(iterator->set->depth == 0) {
		e->valid = true;
		e->depth = iterator->set->depth;
		while(iterator->range_index < iterator->set->ranges.size()) {
			size_t begin = iterator->set->ranges[iterator->range_index].first;
			size_t size = iterator->set->ranges[iterator->range_index].second - begin;
			if(iterator->i < size) {
				e->id = iterator->set->local_to_global[begin + iterator->i];
				e->hf = NULL;
				iterator->i++;
				return true;
			} else {
				iterator->range_index++;
				iterator->i = 0;
			}
		}
		return false;
	} else {
		if(iterator->i < iterator->set->elements.size()) {
			lDiscoverySetGetElement(iterator->set,iterator->i,e);
			iterator->i++;
			return true;
		} else return false;
	}
}
void lDiscoveryFieldAccess(lsContext * ctx_, int field_id, lElementType typ, lsElement * e_, lPhase phase) {
	lDiscoveryContext * ctx = CTX(ctx_);
	lDiscoveryElement * e = ELEM(e_);
	if(ctx->record_fields) {
		ElementStatus status;
		uint32_t id = e->id;
		ctx->local_memory->registerType(field_id,typ);
		if(ctx->local_topology->lookup(typ,id,&status)) {
			if(status.isOwnerKnown() == false) {
				ctx->owner_service->request(typ,id);
				ctx->local_topology->ownerRequested(typ,id); //mark the owner present in local_topology to prevent requesting again
			}
			//skip the outside cell, you are not allowed to store values at it
			//is this element not owned by this partition, if so we have found a ghost
			if( !(typ == L_CELL && id == 0) &&
			    (status.isOwnerKnown() == false || status.owner() != ctx->name_service->self()) ) {
				ctx->local_topology->addGhost(typ,id);
				if(phase == L_READ_ONLY)
					ctx->local_memory->addGhostRead(field_id,typ,id);
				else
					ctx->local_memory->addGhostWrite(field_id,typ,id);
			}
		} else {
			LOG_ERROR("Missing Ghost: %d %d",typ,id);
		}
	}
}
void lDiscoverySetInitBoundary(lsContext * ctx_,  lsSet * set_, lElementType type, const char * boundary_name) {
	lDiscoveryContext * ctx = CTX(ctx_);
	lDiscoverySet * set = SET(set_);
	set->depth = 0;
	set->local_to_global = ctx->owned_elements->data[(int)type].begin();
	size_t n_owned = ctx->owned_elements->data[(int)type].size();
	size_t global_size;
	int id;
	lDiscoverySetInitRanges(ctx->boundary_reader, type, boundary_name, n_owned, set->local_to_global, &set->ranges, &global_size,NULL,&id);
	set->type = type;
}


void initTopLevelElements(lDiscoveryContext * ctx, lElementType type, lDiscoverySet * set) {
	set->depth = 0;
	set->valid = true;
	set->local_to_global = ctx->owned_elements->data[(int)type].begin();
	size_t n_owned = ctx->owned_elements->data[(int)type].size();
	set->ranges.push_back(std::make_pair(0,n_owned));
	set->type = type;
}
//stencil topology functions
void lDiscoveryVerticesOfMesh(lsContext * ctx, lsSet * set) {
	initTopLevelElements(CTX(ctx),L_VERTEX,SET(set));
}
void lDiscoveryEdgesOfMesh(lsContext * ctx, lsSet * set) {
	initTopLevelElements(CTX(ctx),L_EDGE,SET(set));
}
void lDiscoveryFacesOfMesh(lsContext * ctx, lsSet * set) {
	initTopLevelElements(CTX(ctx),L_FACE,SET(set));
}
void lDiscoveryCellsOfMesh(lsContext * ctx, lsSet * set) {
	initTopLevelElements(CTX(ctx),L_CELL,SET(set));
	//if we have the 0 (outside) element, skip it in the list of stuff
	if(SET(set)->ranges.begin()->second > 0 && SET(set)->local_to_global[0] == 0)
		SET(set)->ranges.begin()->first++;
}

bool elementTopologyIsPresent(lDiscoveryContext * ctx, lElementType typ, uint32_t id) {
	FacetEdgeBuilder::map_t & emap = ctx->mesh->elem_maps[typ];
	FacetEdgeBuilder::map_t::iterator it = emap.find(id);
	if(it == emap.end()) {
		//LOG_DBG("Found no reference: %d:%d",(int)typ,(int)id);
		return false;
	} else {
		
		MeshIO::FacetEdge::HalfFacet * hf = it->second;
		assert(hf != NULL);
		
		switch(typ) {
			case L_VERTEX: {
				MeshIO::has_all_facets_v haf;
				haf(0,ctx->mesh,hf,&ctx->buffer);
				return haf.result;
			}
			case L_EDGE: {
				//mesh->printRing(0,hf);
				MeshIO::has_all_facets_e haf;
				haf(0,ctx->mesh,hf,&ctx->buffer);
				return haf.result;
			}
			case L_FACE: {
				//mesh->printRing(1,hf);
				MeshIO::has_all_facets_e haf;
				haf(1,ctx->mesh,hf,&ctx->buffer);
				return haf.result;
			}
			case L_CELL: {
				MeshIO::has_all_facets_v haf;
				haf(1,ctx->mesh,hf,&ctx->buffer);
				return haf.result;
			}
			default:
				return false;
		}
	}
}

void requestAbsentTopology(lDiscoveryContext * ctx, lElementType typ, uint32_t id) {
	ElementStatus status;
	if(!ctx->local_topology->lookup(typ,id,&status) || status.hasTopology() == false) {
		ctx->local_topology->topologyRequested(typ,id); //mark the topology present in local_topology to prevent requesting again
		assert(ctx->local_topology->lookup(typ,id,&status));
		LOG_DBG("requesting %d:%d",(int)typ,(int)id);
		ctx->topology_service->request(typ,id);
		ctx->requests_occured = true;
	}
}

bool elementPrepareForFunction(lDiscoveryContext * ctx, lElementType type, lDiscoveryElement * c) {
	if(!c->valid || (type == L_CELL && c->id == 0))
		return false;
	
	if(c->depth < ctx->topology_depth || //we are guarenteed that a previous iteration requested the topology, or
	    elementTopologyIsPresent(ctx,type,c->id)) { //we happen to already have the topology from other requests
		if(c->hf == NULL) { //top-level sets return elements will null HFs because the toplogy won't be present on the first iteration
			c->hf = ctx->mesh->elem_maps[type][c->id];
			assert(c->hf != NULL);
		}
		return true;
	} else {
		//we don't have the topology but this is a valid function call
		//request the topology for the next round
		requestAbsentTopology(ctx,type,c->id);
		return false;	
	}
}

#define PREP_SET(set_type, elem_type) do {\
	SET(set)->depth = ELEM(c)->depth + 1; \
	SET(set)->type = set_type; \
	SET(set)->elements.clear(); \
	SET(set)->valid = false; \
	if(!elementPrepareForFunction(CTX(ctx),elem_type,ELEM(c))) return; \
	SET(set)->valid = true; \
} while(0)



void lDiscoveryVerticesOfVertex(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_VERTEX,L_VERTEX);
	MeshIO::vertices_v()(0,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}
void lDiscoveryVerticesOfEdge(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_VERTEX,L_EDGE);
	MeshIO::vertices_e()(0,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);	
}
void lDiscoveryVerticesOfFace(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_VERTEX,L_FACE);
	MeshIO::cells_e()(1,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}
void lDiscoveryVerticesOfCell(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_VERTEX,L_CELL);
	MeshIO::cells_v()(1,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}

void lDiscoveryCellsOfVertex(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_CELL,L_VERTEX);
	MeshIO::cells_v()(0,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}
void lDiscoveryCellsOfEdge(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_CELL,L_EDGE);
	MeshIO::cells_e()(0,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}
void lDiscoveryCellsOfFace(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_CELL,L_FACE);
	MeshIO::vertices_e()(1,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);	
}
void lDiscoveryCellsOfCell(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_CELL,L_CELL);
	MeshIO::vertices_v()(1,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}

void lDiscoveryEdgesOfVertex(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_EDGE,L_VERTEX);
	MeshIO::edges_v()(0,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}

void lDiscoveryEdgesOfCell(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_EDGE,L_CELL);
	MeshIO::faces_v()(1,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}

void lDiscoveryEdgesOfFaceCCW(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_EDGE,L_FACE);
	MeshIO::faces_e()(1,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}

void lDiscoveryEdgesOfFace(lsContext * ctx, lsElement * c, lsSet * set) {
	lDiscoveryEdgesOfFaceCCW(ctx,c,set);
}

void lDiscoveryEdgesOfFaceCW(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_EDGE,L_FACE);
	MeshIO::faces_e()(1,CTX(ctx)->mesh,ELEM(c)->hf->flip(),&SET(set)->elements);
}

void lDiscoveryFacesOfVertex(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_FACE,L_VERTEX);
	MeshIO::faces_v()(0,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}

void lDiscoveryFacesOfCell(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_FACE,L_CELL);
	MeshIO::edges_v()(1,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}

void lDiscoveryFacesOfEdgeCCW(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_FACE,L_EDGE);
	MeshIO::faces_e()(0,CTX(ctx)->mesh,ELEM(c)->hf,&SET(set)->elements);
}
void lDiscoveryFacesOfEdge(lsContext * ctx, lsElement * c, lsSet * set) {
	lDiscoveryFacesOfEdgeCCW(ctx,c,set);
}
void lDiscoveryFacesOfEdgeCW(lsContext * ctx, lsElement * c, lsSet * set) {
	PREP_SET(L_FACE,L_EDGE);
	MeshIO::faces_e()(0,CTX(ctx)->mesh,ELEM(c)->hf->flip(),&SET(set)->elements);
}
#undef PREP_SET


#define PREP_ELEM(typ)  do {\
	result->depth = c->depth + 1; \
	result->valid = false; \
	result->hf = NULL; \
	result->id = 0; \
	if(!elementPrepareForFunction(CTX(ctx),typ,c)) return; \
} while(0)

void lDiscoveryHeadOfEdge(lsContext * ctx, lsElement * c_, lsElement * result_) {
	lDiscoveryElement * c = ELEM(c_);
	lDiscoveryElement * result = ELEM(result_);
	PREP_ELEM(L_EDGE);
	result->hf = c->hf;
	result->id = result->hf->vert();
	
}
void lDiscoveryTailOfEdge(lsContext * ctx, lsElement * c_, lsElement * result_) {
	lDiscoveryElement * c = ELEM(c_);
	lDiscoveryElement * result = ELEM(result_);
	
	PREP_ELEM(L_EDGE);
	result->hf = c->hf->flip();
	result->id = result->hf->vert();
}

void lDiscoveryHeadOfFace(lsContext * ctx, lsElement * c_, lsElement * result_) {
	lDiscoveryElement * c = ELEM(c_);
	lDiscoveryElement * result = ELEM(result_);
	PREP_ELEM(L_FACE);
	result->hf = c->hf;
	result->id = result->hf->cell();
}
void lDiscoveryTailOfFace(lsContext * ctx, lsElement * c_, lsElement * result_) {
	lDiscoveryElement * c = ELEM(c_);
	lDiscoveryElement * result = ELEM(result_);
	PREP_ELEM(L_FACE);
	result->hf = c->hf->flip();
	result->id = result->hf->cell();
}

void lDiscoveryFlipEdge(lsContext * ctx, lsElement * c_, lsElement * result_) {
	lDiscoveryElement * c = ELEM(c_);
	lDiscoveryElement * result = ELEM(result_);
	PREP_ELEM(L_EDGE);
	result->hf = c->hf->flip();
	result->id = c->id;
}
void lDiscoveryFlipFace(lsContext * ctx, lsElement * c_, lsElement * result_) {
	lDiscoveryElement * c = ELEM(c_);
	lDiscoveryElement * result = ELEM(result_);
	PREP_ELEM(L_FACE);
	result->hf = c->hf->flip();
	result->id = c->id;
}

void lDiscoveryTowardsEdge(lsContext * ctx, lsElement * c_, lsElement * v_, lsElement * result_) {
	lDiscoveryElement * c = ELEM(c_);
	lDiscoveryElement * v = ELEM(v_);
	lDiscoveryElement * result = ELEM(result_);
	PREP_ELEM(L_EDGE);
	if(c->hf->vert() == v->id) {
		result->id = c->id;
		result->hf = c->hf;	
	} else {
		result->id = c->id;
		result->hf = c->hf->flip();
	}
}
void lDiscoveryTowardsFace(lsContext * ctx, lsElement * c_, lsElement * v_, lsElement * result_) {
	lDiscoveryElement * c = ELEM(c_);
	lDiscoveryElement * v = ELEM(v_);
	lDiscoveryElement * result = ELEM(result_);
	PREP_ELEM(L_FACE);
	if(c->hf->cell() == v->id) {
		result->id = c->id;
		result->hf = c->hf;	
	} else {
		result->id = c->id;
		result->hf = c->hf->flip();
	}
}

void lDiscoverySetExtractElement(lsContext * ctx, lsSet * set_, uint32_t idx, lsElement * result_) {
	lDiscoveryElement * result = ELEM(result_);
	lDiscoverySet * set = SET(set_);
	lDiscoverySetGetElement(set,idx,result);
}

#undef PREP_ELEM

#undef CTX
#undef SET
#undef IT
#undef ELEM

void lDiscoveryCreateVTable(lsFunctionTable * vtbl) {
	bzero(vtbl, sizeof(lsFunctionTable));
	vtbl->lsSetGetIterator = lDiscoverySetGetIterator;
	vtbl->lsFieldAccess = lDiscoveryFieldAccess;
	vtbl->lsIteratorNext = lDiscoveryIteratorNext;
	vtbl->lsFieldAccess = lDiscoveryFieldAccess;
	vtbl->lsSetInitBoundary = lDiscoverySetInitBoundary;
	vtbl->lsVerticesOfMesh = lDiscoveryVerticesOfMesh;
	vtbl->lsEdgesOfMesh = lDiscoveryEdgesOfMesh;
	vtbl->lsFacesOfMesh = lDiscoveryFacesOfMesh;
	vtbl->lsCellsOfMesh = lDiscoveryCellsOfMesh;
	vtbl->lsVerticesOfVertex = lDiscoveryVerticesOfVertex;
	vtbl->lsVerticesOfEdge = lDiscoveryVerticesOfEdge;
	vtbl->lsVerticesOfFace = lDiscoveryVerticesOfFace;
	vtbl->lsVerticesOfCell = lDiscoveryVerticesOfCell;
	vtbl->lsCellsOfVertex = lDiscoveryCellsOfVertex;
	vtbl->lsCellsOfEdge = lDiscoveryCellsOfEdge;
	vtbl->lsCellsOfFace = lDiscoveryCellsOfFace;
	vtbl->lsCellsOfCell = lDiscoveryCellsOfCell;
	vtbl->lsEdgesOfVertex = lDiscoveryEdgesOfVertex;
	vtbl->lsEdgesOfCell = lDiscoveryEdgesOfCell;
	vtbl->lsEdgesOfFaceCCW = lDiscoveryEdgesOfFaceCCW;
	vtbl->lsEdgesOfFace = lDiscoveryEdgesOfFace;
	vtbl->lsEdgesOfFaceCW = lDiscoveryEdgesOfFaceCW;
	vtbl->lsFacesOfVertex = lDiscoveryFacesOfVertex;
	vtbl->lsFacesOfCell = lDiscoveryFacesOfCell;
	vtbl->lsFacesOfEdgeCCW = lDiscoveryFacesOfEdgeCCW;
	vtbl->lsFacesOfEdge = lDiscoveryFacesOfEdge;
	vtbl->lsFacesOfEdgeCW = lDiscoveryFacesOfEdgeCW;
	vtbl->lsHeadOfEdge = lDiscoveryHeadOfEdge;
	vtbl->lsTailOfEdge = lDiscoveryTailOfEdge;
	vtbl->lsHeadOfFace = lDiscoveryHeadOfFace;
	vtbl->lsTailOfFace = lDiscoveryTailOfFace;
	vtbl->lsFlipEdge = lDiscoveryFlipEdge;
	vtbl->lsFlipFace = lDiscoveryFlipFace;
	vtbl->lsTowardsEdge = lDiscoveryTowardsEdge;
	vtbl->lsTowardsFace = lDiscoveryTowardsFace;
	vtbl->lsSetExtractElement = lDiscoverySetExtractElement;
}
#endif