/*
 * FieldInitializer.h
 *
 *  Created on: Jul 28, 2009
 *      Author: zdevito
 */

#ifndef FIELDINITIALIZER_H_
#define FIELDINITIALIZER_H_

#include "Liszt/ElemTypes.h"
#include "Liszt/ElemRefSet.h"

namespace FieldCommon {

    //field initializers provide initial values for the data in a field
    //each must implement the request method which will be called
    //with a list of elements that this mesh needs values for
    //for each value that is requested, the FieldInitializer must yield that value
    //using one of the two yield functions described below.


    //field initializer are copied internally in the Liszt runtime so they
    //should not hold a large amount of data before/after the call to request

    using namespace ::LisztPrivate::ElemTypes;

    //a Field initializer is a concept, and thus does not
    //have any implementation in the parent class
    //however, initializers should derive
    //from FieldInitialzer<ElemType,ValueType> to ensure
    //type correctness when used

    template<typename E, typename V>
    class FieldInitializer {
    public:
        typedef E elem_type;
        typedef V value_type;
    };


    class FieldInitializerExample : public FieldInitializer<VertexType,double> {
    public:

        //this is called before any request calls are made
        void init() {}

        //this is called (possibly multiple times) to request values that are needed for the
        //field. set.begin() and set.end() will be const_iterators for the set
        //the values of the set will be ElemRef<elem_type> objects, where elem_type
        //is specified in the FieldInitializer class
        //the return object provides an interface to yield values back to the field
        //as they are loaded, see below for its specific interface
        template<typename Set, typename Return>
        void request(const Set & required_elements, Return * obj) {}

        //this is called after all request calls are made
        template<typename Return>
        void finalize(Return * obj) {}

        virtual ~FieldInitializerExample() {}
    };

    //this object specifies the interface for the Return objects
    //used to interface with the field

    //for each element in the required_values, a FieldInitializer must call a yield
    //function to set its value.  If no yield call is made for a particular element
    //its value will be undefined. You may call yield in finalize() as long as the
    //value being yielded was requested.
    //If more than one yield is called, then the field
    //will have the value of the last yield

    //There are two calls to yield, with slightly different semantics
    //that allow for optimized field value iterators

    class ReturnObjExample {
        typedef double value_type;
        typedef VertexType elem_type;
        //provides the value for the elements in the request in the order specified in the reference set
        //it must be called in the order specified by the required_values set
        //this version of yield is faster and more coherent than the more general version
        //e.g. if the references set contains 1 5 7
        //then the first yield will the value of element 1, the next will set the value of 5, and the next 7.
        void yield(const value_type & v);

        //provides the value v for element ref in the field, there are no ordering requirements
       //but this version is less coherent and requires a lookup to find the correct entry

       //you may NOT call yield for element references not in the required_values set
       //TODO(zach): we can easily relax this requirement, which should ease the creation
       //of field initializers
        void yield(const ElemRef<elem_type> & ref, const value_type & v);
    };



    //this class is instantiated by a field and passed into
    //a client field helper
    //it provides the yield(value_type) and yield(elemRef,value_type) functions
    //as specified above
    template<typename Field>
    class FieldInitializerHelper {
    public:
        typedef typename Field::elem_type elem_type;
        typedef typename Field::value_type value_type;
        typedef Field field_type;
        typedef typename Field::global_to_local_type global_to_local_type;

        void yield(const value_type & v) {
            value_type * entry = field->getReadEntry(next_entry++);
            *entry = v;
        }
        void yield(const ElemRef<elem_type> & ref, const value_type & v) {
            value_type * entry = field->getReadEntry((*gtl)(ref));
            *entry = v;
        }
        void setField(field_type * f, const global_to_local_type * g) {
            field = f;
            gtl = g;
            next_entry = 0;
        }
    private:
        field_type * field;
        const global_to_local_type * gtl;
        id_type next_entry;
    };

    template<typename T, typename V>
    class  ConstFieldInitializer
    : public FieldInitializer<T,V> {
      public:
        ConstFieldInitializer(const V & v)
        : value(v) {}
        void init() {}
        template<typename Set, typename Return>
        void request(const Set & req, Return * obj) {
            for(size_t i = 0; i < req.size(); i++) {
                obj->yield(value);
            }
        }
        template<typename Return>
        void finalize(Return * obj) {}
      private:
        V value;
    };

    // Used only for multifields
    template<typename Field>
    class MultiFieldInitializerHelper {
    public:
        typedef typename Field::elem_type elem_type;
        typedef typename Field::value_type value_type;
        typedef Field field_type;
        typedef typename Field::global_to_local_type global_to_local_type;

        void yield(const id_type id, const id_type set_id, const value_type & v) {
            value_type * entry = field->getReadEntry(id, set_id);
            *entry = v;
        }
        void setField(field_type * f, const global_to_local_type * g) {
            field = f;
            gtl = g;
        }
    private:
        field_type * field;
        const global_to_local_type * gtl;
    };

    // Used only for multifields
    template<typename T, typename V>
    class  ConstMultiFieldInitializer
    : public FieldInitializer<T,V> {
      public:
        ConstMultiFieldInitializer(const V & v)
        : value(v) {}
        void init() {}
        template<typename Return>
        void request(const id_type id, const size_t set_size, Return * obj) {
            for(size_t i = 0; i < set_size; i++) {
                obj->yield(id, i, value);
            }
        }
        template<typename Return>
        void finalize(Return * obj) {}
      private:
        V value;
    };


}  // namespace FieldCommon

#endif /* FIELDINITIALIZER_H_ */
