/*
 * Identities.h
 *
 *  Created on: Jul 31, 2009
 *      Author: zdevito
 */
#ifndef IDENTITIES_H_
#define IDENTITIES_H_

#include <climits>
#include <limits>
#include <float.h>
#include "Vector.h"
#include "Matrix.h"
#include "type_traits.h"

#ifdef GPURUNTIME
#define GPU_DECLS __device__ __host__
#else
#define GPU_DECLS
#endif

//namespace {
    //this class provides a way to get the numeric identities for a field type and is pre-defined for built-in types
    //and liszt special type like vec and matrix, other classes can specialize this with this form:
    //for each operation op(x,identity) == x
/*
     class Identities<T> {

        static const T additive()
        static const T multiplicative()
        static const T min_identity()
        static const T max_identity()
        static const T or_identity()
        static const T and_identity()
        static const T xor_identity()
     }
 */



template<typename T>
struct numeric_limits;


#define NUMERIC_LIMITS_SPEC(type,max_const,min_const) \
template<> \
struct numeric_limits<type> { \
	GPU_DECLS static type max() {return max_const;} \
	GPU_DECLS static type min() {return min_const;} \
};

NUMERIC_LIMITS_SPEC(int, INT_MAX, INT_MIN);
NUMERIC_LIMITS_SPEC(float, FLT_MAX, FLT_MIN);
NUMERIC_LIMITS_SPEC(double, DBL_MAX, DBL_MIN);

template<typename T>
struct Identities;

template<typename T>
struct IdentitiesBase;

//identites for built-in numeric types
template<typename T>
struct IdentitiesBase {
	GPU_DECLS static T additive() { return (T) 0; }
	GPU_DECLS static T multiplicative() { return (T) 1; }
	GPU_DECLS static T or_identity() { return (T) 0; }
	GPU_DECLS static T and_identity() { return ~(T) 0; }
	GPU_DECLS static T xor_identity() { return (T) 0; }
	GPU_DECLS static T min_identity() { return numeric_limits<T>::max(); }
	GPU_DECLS static T max_identity() { return numeric_limits<T>::min(); }
};

//identities for vector types
template<size_t N, typename T>
struct IdentitiesBase< vec<N,T> > {
	typedef vec<N,T> value_type;
	typedef Identities<T> base_ids;
	GPU_DECLS static value_type additive() { return value_type(base_ids::additive());  }
	GPU_DECLS static value_type multiplicative() { return value_type(base_ids::multiplicative()); }
	GPU_DECLS static value_type min_identity() { return value_type(base_ids::min_identity()); }
	GPU_DECLS static value_type max_identity() { return value_type(base_ids::max_identity()); }
	GPU_DECLS static value_type or_identity() { return value_type(base_ids::or_identity()); }
	GPU_DECLS static value_type and_identity() { return value_type(base_ids::and_identity()); }
	GPU_DECLS static value_type xor_identity() { return value_type(base_ids::xor_identity()); }
};

//identities for Matrix types
template<size_t R, size_t C, typename T>
struct IdentitiesBase< matrix<R,C,T> > {
	typedef matrix<R,C,T> value_type;
	typedef Identities<T> base_ids;
	GPU_DECLS static value_type additive() { return value_type(base_ids::additive());  }
	GPU_DECLS static value_type multiplicative() { return value_type().eye(); }
	GPU_DECLS static value_type min_identity() { return value_type(base_ids::min_identity()); }
	GPU_DECLS static value_type max_identity() { return value_type(base_ids:: max_identity()); }
	GPU_DECLS static value_type or_identity() { return value_type(base_ids::or_identity()); }
	GPU_DECLS static value_type and_identity() { return value_type(base_ids::and_identity()); }
	GPU_DECLS static value_type xor_identity() { return value_type(base_ids::xor_identity()); }
};

template<typename T>
struct Identities : public IdentitiesBase<T> {};

//}
#endif /* IDENTITIES_H_ */
