/*
 * Field.h
 *
 *  Created on: Feb 19, 2009
 *      Author: zdevito
 */

#ifndef FIELD_H_
#define FIELD_H_
#include <vector>
#include <iostream>
#include <string>
#include "liszt_memory.h"
#include "SetAdaptor.h"
#include "Fields/FieldInitializer.h"
#include "Fields/FieldValueReference.h"
#include "Liszt/ElemTypes.h"
#include "Liszt/Range.h"
#include "noncopyable.h"

using std::vector;
using namespace LisztPrivate::ElemTypes;
using namespace std;

namespace LocalField {

//for use in initializer which needs to translate global IDs to local id
//for whatever field implementation is in use
template<typename T>
class GlobalToLocalIdentityFunc {
public:
    typedef T elem_type;
    GlobalToLocalIdentityFunc() {}
    LisztPrivate::id_type operator()(const LisztPrivate::ElemRef<elem_type> & e) const {
        return e.ID();
    }
    LisztPrivate::id_type operator()(LisztPrivate::id_type id) const {
        return id;
    }
};

template<typename E, typename vt> class Field : public Liszt::noncopyable {
public:
    typedef Field<E,vt> this_type;
	typedef vt value_type;
	typedef E elem_type;
	typedef GlobalToLocalIdentityFunc<elem_type> global_to_local_type;
	typedef FieldCommon::FieldValueReference<this_type> reference;
	Field(const std::string & name__,vt init, size_t size)
	: name_(name__), data(size,init) {}
	Field(const Field<elem_type,vt> & rhs) {
	    //If your code is failing here, it is because you are
	    //trying to copy a field (which is not allowed, declare a new field and copy the values)
	    //or you are trying to assign a field to mesh.newXField after already being declared
	    //(you should use the LocalField::newField(mesh, field,...) method instead in this case.

	    //This would be private, but the compiler chokes on that even in cases where
	    //this constructor is never run.
	    //TODO: I don't really understand the distinction on assignments.
	    assert(false);
	}
public: Field() : name_("undefined"), data() {}

    template<typename Init>
	void initWithObj(const std::string & name__,Init * init, size_t size) {
		name_ = name__;
		data.clear();
		data.insert(data.begin(),size,value_type());

		global_to_local_type gtl;
		FieldCommon::FieldInitializerHelper<this_type> helper;
		helper.setField(this,&gtl);
		LisztPrivate::Range<elem_type> allElems(0,size);
		init->init();
		init->request(allElems,&helper);
		init->finalize(&helper);
	}
    void init(const std::string & name__,const vt & init, size_t size) {
        name_ = name__;
        data.clear();
        data.insert(data.begin(),size,init);
    }
    template<typename MO>
	const vt & operator()(const MO & mo) const {
	    return data[mo.ID()];
	}
    template<typename MO>
    const vt & operator()(const MO & mo) {
        return (*(const this_type*)this)(mo);
    }
    template<typename MO>
	reference operator[](const MO & mo) {
	    return reference(*this,mo.ID());
	}

	const std::string & name() const {
		return name_;
	}
	template<typename Op>
	void update() {}
	template<typename Op>
	void setReductionOp() {}

	typedef typename vector<value_type>::const_iterator internal_it;
	internal_it internal_begin() const {
	    return data.begin();
	}
	internal_it internal_end() const {
	    return data.end();
	}
	value_type * getReadEntry(id_type id) {
	    return &data[id];
	}
    const value_type * getReadEntry(id_type id) const {
        return &data[id];
    }
    value_type * getWriteEntry(id_type id) {
        return &data[id];
    }
    const value_type * getWriteEntry(id_type id) const {
        return &data[id];
    }
private:
    std::string name_;
	vector<value_type> data;
};


template<typename E,typename VT>
class FieldSeries {
public:
    typedef VT value_type;
    typedef E elem_type;
    typedef Field<E,VT> field_type;
    typedef typename field_type::reference reference;
    FieldSeries(const std::string & name__,
                value_type init,
                size_t size, size_t n_series)
    : name_(name__), t0(0), fields(n_series,Field<E,VT>(name__,init,size)) {}

    FieldSeries() : name_("undefined") {}

    void init(const std::string & name__,value_type init, size_t size, size_t n_series) {
        name_ = name__;
        t0 = 0;
        fields.clear();
        fields.insert(fields.back(),n_series,field_type(name__,init,size));
    }
    template<typename MO>
    const value_type & operator()(const MO & mo) const {
        assert(mo != NULL);
        return fields[t0](mo);
    }
    template<typename MO>
    reference operator()(const MO & mo) {
        return fields[t0](mo);
    }
    const field_type & operator[](size_t n) const {
        n += t0;
        if(n < 0) n += fields.size();
        if(n >= fields.size()) n-= fields.size();

        assert(n >= 0 && n < fields.size());
        //cout << "Accessing " << n << endl;
        return fields[n];
    }
    field_type & operator[](size_t n) {
        n += t0;
        if(n < 0) n += fields.size();
        if(n >= fields.size()) n -= fields.size();

        assert(n >= 0 && n < fields.size());
        //cout << "Accessing " << n << endl;
        return fields[n];
    }
    const std::string & name() const {
        return name_;
    }
    void step() {
        t0 = (t0 + 1) % fields.size();
    }
private:
    std::string name_;
    typedef vector< field_type > FieldList;
    typedef typename FieldList::iterator FieldIter;

    FieldList fields;
    size_t t0;
};

// These functions replace the direct "newXField" functions directly within mesh.
// This is currently intended only for test code; actual user code should use
// buildField/buildFieldWithInit, which calls the appropriate build methods for
// the backend.
template<typename Mesh, typename ElemType, typename T> inline void newField(const Mesh& mesh, Field<ElemType,T> & field,const T & init,const std::string & name="<default>") {
    field.init(name,init,mesh.template get<ElemType>().size());
}
// TODO(mbarrien): With the demise of domains for fields, newSubField may no longer be necessary.
template<typename Mesh, typename ElemType, typename T, typename set_t> inline void newSubField(const Mesh& mesh, Field<ElemType,T> & field,const T & init,const set_t & s,const std::string & name="<default>") {
    field.init(name,init,mesh.template get<ElemType>().size());
}
template<typename Mesh, typename ElemType, typename T> inline void newFieldSeries(const Mesh& mesh, FieldSeries<ElemType,T> & field,size_t n, const T & init,const std::string & name="<default>") {
    field.init(name,init,mesh.template get<ElemType>().size(),n);
}

} // namespace LocalField

#endif /* FIELD_H_ */
