#ifndef _REDUCTION_IMPL
#define _REDUCTION_IMPL

#include "type_traits.h"

/*
    LISZT_MAX_TYPE,
    LISZT_MIN_TYPE,
    LISZT_PROD_TYPE,
    LISZT_SUM_TYPE,
    LISZT_LAND_TYPE,
    LISZT_LOR_TYPE,
    LISZT_LXOR_TYPE,
    LISZT_BAND_TYPE,
    LISZT_BOR_TYPE,
    LISZT_BXOR_TYPE,
*/

#ifdef GPURUNTIME
#define GPU_DECLS __device__ __host__
#else
#define GPU_DECLS
#endif

#ifdef __CUDA_ARCH__
#define REDUCTION_ABORT return T(); //really not what you want but oh well
#else
#define REDUCTION_ABORT abort();
#endif
	
template<typename T>
struct ReductionImpl;

template<typename enabled,typename T>
struct comparison_ops;

template<typename enabled,typename T>
struct sum_ops;

template<typename enabled,typename T>
struct prod_ops;

template<typename enabled,typename T>
struct logical_ops;

template<typename enabled,typename T>
struct binary_ops;

/*
        static const T additive()
        static const T multiplicative()
        static const T min_identity()
        static const T max_identity()
        static const T or_identity()
        static const T and_identity()
        static const T xor_identity()
*/

//missing ops are allow but will abort if ever called

template<typename T>
struct comparison_ops< ::Liszt::false_type,T> {
	static const T min_identity() { REDUCTION_ABORT; }
	static const T max_identity() { REDUCTION_ABORT; }
	GPU_DECLS static T min(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
	GPU_DECLS static T max(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
};

template<typename T>
struct sum_ops< ::Liszt::false_type,T> {
	static const T additive() { REDUCTION_ABORT; }
	GPU_DECLS static T plus(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
};

template<typename T>
struct prod_ops< ::Liszt::false_type,T> {
	static const T multiplicative() { REDUCTION_ABORT; }
	GPU_DECLS static T times(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
};

template<typename T>
struct logical_ops< ::Liszt::false_type,T> {
	static const T lor_identity() { REDUCTION_ABORT; }
	static const T land_identity() { REDUCTION_ABORT; }
	static const T lxor_identity() { REDUCTION_ABORT; }
	GPU_DECLS static T land(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
	GPU_DECLS static T lor(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
	GPU_DECLS static T lxor(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
};

template<typename T>
struct binary_ops< ::Liszt::false_type,T> {
	static const T bor_identity() { REDUCTION_ABORT; }
	static const T band_identity() { REDUCTION_ABORT; }
	static const T bxor_identity() { REDUCTION_ABORT; }
	GPU_DECLS static T band(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
	GPU_DECLS static T bor(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
	GPU_DECLS static T bxor(const T & a, const T & b) {
		REDUCTION_ABORT;
	}
};


//enabled has default implementations
template<typename T>
struct comparison_ops< ::Liszt::true_type,T> {
	static const T min_identity() { return Identities<T>::min_identity(); }
	static const T max_identity() { return Identities<T>::max_identity(); }
	GPU_DECLS static T min(const T & a, const T & b) {
		return a < b ? a : b;
	}
	GPU_DECLS static T max(const T & a, const T & b) {
		return a > b ? a : b;
	}
};

template<size_t N, typename T>
struct comparison_ops< ::Liszt::true_type, vec<N,T> > {
	
	GPU_DECLS static const vec<N,T> min_identity() { return Identities< vec<N,T> >::min_identity(); }
	GPU_DECLS static const vec<N,T> max_identity() { return Identities< vec<N,T> >::max_identity(); }
	
    GPU_DECLS static vec<N,T>  min(const vec<N,T> & a, const vec<N,T> & b) {
        vec<N,T> ret(Identities<T>::min_identity());
        for(size_t i = 0; i < N; i++)
            ret[i] = ReductionImpl<T>::min(a[i],b[i]);
        return ret;
    }
    GPU_DECLS static vec<N,T>  max(const vec<N,T> & a, const vec<N,T> & b) {
        vec<N,T> ret(Identities<T>::max_identity());
        for(size_t i = 0; i < N; i++)
            ret[i] =  ReductionImpl<T>::max(a[i],b[i]);
        return ret;
    }
};

template<size_t R, size_t C, typename T>
struct comparison_ops< ::Liszt::true_type, matrix<R,C,T> > {

	GPU_DECLS static const matrix<R,C,T> min_identity() { return Identities< matrix<R,C,T> >::min_identity(); }
	GPU_DECLS static const matrix<R,C,T> max_identity() { return Identities< matrix<R,C,T> >::max_identity(); }
    
    GPU_DECLS static matrix<R,C,T>  min(const matrix<R,C,T> & a, const matrix<R,C,T> & b) {
        matrix<R,C,T> ret(Identities<T>::min_identity());
        for(size_t i = 0; i < R; i++)
            for(size_t j = 0; j < C; j++)
                ret.v[i][j] = ReductionImpl<T>::min(a.v[i][j],b.v[i][j]);
        return ret;
    }
    GPU_DECLS static matrix<R,C,T>  max(const matrix<R,C,T> & a, const matrix<R,C,T> & b) {
        matrix<R,C,T> ret(Identities<T>::max_identity());
        for(size_t i = 0; i < R; i++)
            for(size_t j = 0; j < C; j++)
                ret.v[i][j] = ReductionImpl<T>::max(a.v[i][j],b.v[i][j]);
        return ret;
    }
};

template<typename T>
struct sum_ops< ::Liszt::true_type,T> {
	GPU_DECLS static const T additive() { return Identities<T>::additive(); }
	GPU_DECLS static T plus(const T & a, const T & b) {
		return a + b;
	}
};

template<typename T>
struct prod_ops< ::Liszt::true_type,T> {
	GPU_DECLS static const T multiplicative() { return Identities<T>::multiplicative(); }
	GPU_DECLS static T times(const T & a, const T & b) {
		return a * b;
	}
};

template<typename T>
struct logical_ops< ::Liszt::true_type,T> {
	GPU_DECLS static const T lor_identity() { return Identities<T>::or_identity(); }
	GPU_DECLS static const T land_identity() { return Identities<T>::and_identity(); }
	GPU_DECLS static const T lxor_identity() { return Identities<T>::xor_identity(); }
	GPU_DECLS static T land(const T & a, const T & b) {
		return a && b;
	}
	GPU_DECLS static T lor(const T & a, const T & b) {
		return a || b;
	}
	GPU_DECLS static T lxor(const T & a, const T & b) {
		return (a && ~b) || (~a && b);
	}
};

// template<size_t N, typename T>
// struct logical_ops<true_type,vec<N,T> > {
// 	GPU_DECLS static const T lor_identity() { return Identities<T>::or_identity(); }
// 	GPU_DECLS static const T land_identity() { return Identities<T>::and_identity(); }
// 	GPU_DECLS static const T lxor_identity() { return Identities<T>::xor_identity(); }
// 	GPU_DECLS static bool land(const T & a, const T & b) {
// 		return a && b;
// 	}
// 	GPU_DECLS static bool lor(const T & a, const T & b) {
// 		return a || b;
// 	}
// 	GPU_DECLS static bool lxor(const T & a, const T & b) {
// 		return (a && ~b) || (~a && b);
// 	}
// };
// 
// template<size_t R, size_t C, typename T>
// struct logical_ops<true_type,matrix<R,C,T> > {
// 	GPU_DECLS static const T lor_identity() { return Identities<T>::or_identity(); }
// 	GPU_DECLS static const T land_identity() { return Identities<T>::and_identity(); }
// 	GPU_DECLS static const T lxor_identity() { return Identities<T>::xor_identity(); }
// 	GPU_DECLS static bool land(const T & a, const T & b) {
// 		return a && b;
// 	}
// 	GPU_DECLS static bool lor(const T & a, const T & b) {
// 		return a || b;
// 	}
// 	GPU_DECLS static bool lxor(const T & a, const T & b) {
// 		return (a && ~b) || (~a && b);
// 	}
// };


template<typename T>
struct binary_ops< ::Liszt::true_type,T> {
	GPU_DECLS static const T bor_identity() { return Identities<T>::or_identity(); }
	GPU_DECLS static const T band_identity() { return Identities<T>::and_identity(); }
	GPU_DECLS static const T bxor_identity() { return Identities<T>::xor_identity(); }
	GPU_DECLS static T band(const T & a, const T & b) {
		return a & b;
	}
	GPU_DECLS static T bor(const T & a, const T & b) {
		return a | b;
	}
	GPU_DECLS static T bxor(const T & a, const T & b) {
		return a ^ b;
	}
};

template<typename T>
struct ReductionImpl : public comparison_ops< typename ::Liszt::is_arithmetic<T>::type, T >, 
                       public sum_ops< typename ::Liszt::is_arithmetic<T>::type, T >, 
                       public prod_ops< typename ::Liszt::is_arithmetic<T>::type, T >,
                       public logical_ops< typename ::Liszt::is_boolean<T>::type, T>,
                       public binary_ops< typename ::Liszt::is_integral<T>::type, T> {};
                       

#undef REDUCTION_ABORT
#endif