/*
 * Serialize.h
 *
 *  Created on: Jul 31, 2009
 *      Author: zdevito
 */

#ifndef SERIALIZE_H_
#define SERIALIZE_H_

#include "Vector.h"
#include "Matrix.h"

#include "type_traits.h"


//information about per-type serialization is obtained through the Serialize meta-funciton
//this function provides pack/unpack and size methods with this interface
/*
    class Serialize<T> {
        static const bool VARIABLE_SIZE; //does this type vary in size
        static const size_t SIZE; //equal to the size of each element if the class does not vary in size
        static size_t pack(const T * value, void * dest); place value in the memory at dest, return the size of dest written
        static size_t unpack(const void * src, T * value); place the memory in src in value, return the size of src read.
    }
 */


//SerializeBase implements some copy serialize operations
//is_self_contained is equal to true_type when a simply memcpy can copy the value in the field
//a specialized implementation is provided to do this copy when needed
template<typename T, bool is_self_contained>
struct SerializeBase;

//a class that is self contained (i.e. has a trivial copy constructor that can be replaced with memcpy)
//should inherit from this tag so that it can be placed into fields without implementing
//special methods
class LisztSelfContainedType {};




//serializer for self-contained types, this includes all built-in non-class types
//and any class tagged with the type LisztSelfContainedType

template<typename T>
struct SerializeBase<T, true > {
    static const bool VARIABLE_SIZE = false;
    static const size_t SIZE = sizeof(T);
    static size_t pack(const T * value, void * dest) {
        *( (T*) dest ) = *value;
        return SIZE;
    }
    static size_t unpack(const void * src, T * value) {
        *value = *((T*) src);
        return SIZE;
    }
};

template<typename T>
struct is_liszt_special;


template<typename T>
struct is_liszt_special : public Liszt::false_type {};


template<typename T>
struct liszt_is_self_contained : public Liszt::true_type {};

template<size_t N, typename T>
struct is_liszt_special< vec<N, T> > : liszt_is_self_contained<T> {};

template<size_t R, size_t C, typename scalar>
struct is_liszt_special< matrix<R,C,scalar> > : liszt_is_self_contained<scalar> {};

template<typename T>
class Serialize : public SerializeBase<T, liszt_is_self_contained<T>::value > {};

//you can specialize Serialize for non-constant types
//TODO(zach): but MPI communication for these types is not yet supported.

#endif /* SERIALIZE_H_ */
