#ifndef _GPU_BOUNDARY_SET_H
#define _GPU_BOUNDARY_SET_H

#include "Gpu/GPUMeshDecl.h"
#include "Gpu/Util.h"
#include <stdint.h>

namespace LisztGPU {
#define GPU_BOTH __host__ __device__
#define GPU_ONLY __device__

class ExplicitBoundarySet {
public:
	void beginInit(size_t set_size) {
		owner = true;
		hElems = new id_type[set_size];
		nelems = 0; //will be set by init code
	}
	void addSet(const Mesh * mesh,size_t start, size_t end) {
		dMesh = mesh;
		for(size_t i = start; i < end; i++) {
			hElems[nelems++] = i;
		}
	}
	void endInit() {
		//TODO(zach): copy to gpu, dealloc cpu copy, swap pointer
		DEVICEALLOC(&dElems,sizeof(id_type) * nelems);
		cudaMemcpy(dElems,hElems,sizeof(id_type) * nelems,cudaMemcpyHostToDevice);
	}
	
	GPU_BOTH id_type size() const {
		return nelems;
	}
	GPU_BOTH id_type kernel_size() const {
		return nelems;
	}
	GPU_ONLY bool elementAt(id_type idx, MeshElement * elem) const {
		if(idx < size()) {
			elem->init(dMesh,elems()[idx]);
			return true;
		} else return false;
	}
	class iterator {
	public:
		GPU_BOTH iterator() {}
		GPU_BOTH iterator(const Mesh * m, id_type * start, id_type * end)
		: mesh(m), start(start), end(end) {}
		typedef MeshElement value_type;
		GPU_BOTH bool hasNext() const {
			return start != end;
		}
		GPU_BOTH MeshElement next() {
			return MeshElement(mesh,*start++);
		}
	private:
		const Mesh * mesh;
		id_type * start;
		id_type * end;
	};
	GPU_BOTH iterator iter() const {
		return iterator(dMesh,elems(),elems()+nelems);
	}
	void dealloc() {
		if(dElems) {
			cudaFree(dElems);
			dElems = NULL;
		}
		if(hElems) {
			delete [] hElems;
			hElems = NULL;
		}
	}
	ExplicitBoundarySet withMesh(const Mesh * m) {
		ExplicitBoundarySet retval = *this;
		retval.dMesh = m;
		return retval;
	}
private:
	GPU_BOTH id_type * elems() {
		#ifdef __CUDA_ARCH__
			return dElems;
		#else
			return hElems;
		#endif
	}
	GPU_BOTH id_type * elems() const {
		#ifdef __CUDA_ARCH__
			return dElems;
		#else
			return hElems;
		#endif
	}
	const Mesh * dMesh; //pointer to mesh on GPU
	bool owner; //was this the original boundary set?
	size_t nelems;
	id_type * hElems;
	id_type * dElems;
};

#define CBS_BLOCK_NBITS 5
#define CBS_BLOCK_SIZE (1 << CBS_BLOCK_NBITS)
#define CBS_MASK 0x1F
class CompressedBoundarySet {
public:
	void beginInit(size_t set_size) {
		owner = true;
		entries = new Entry[set_size]; //this is potentially over provisioned now, we will shrink it later
		nelems = 0; //will be set by init code
		nentries = 0;
	}
	void addSet(Mesh * mesh,size_t start, size_t end) {
		dMesh = mesh;
		for(size_t i = start; i < end; i++) {
			size_t entry_id = i & ~CBS_MASK;
			if(nentries == 0 || entries[nentries-1].id != entry_id) {
				Entry & entry = entries[nentries++];
				entry.id = entry_id;
				entry.mask = 0;
			}
			entries[nentries - 1].mask |= (1 << (i & CBS_MASK));
			nelems++;
		}
	}
	void endInit() {
		if(nentries != nelems) {
			//pack the entries if necesssary
			Entry * packed_entries = new Entry[nentries];
			memcpy(packed_entries,entries,sizeof(Entry) * nentries);
			Entry * old_entries = entries;
			entries = packed_entries;
			delete [] old_entries;
		}
		
		Entry * dEntries;
		DEVICEALLOC(&dEntries, sizeof(Entry) * nentries);
		cudaMemcpy(dEntries,entries,sizeof(Entry) * nentries, cudaMemcpyHostToDevice);
		delete [] entries;
		entries = dEntries;
	}
	
	GPU_BOTH id_type size() const {
		return nelems;
	}
	GPU_BOTH id_type kernel_size() const {
		return nentries * CBS_BLOCK_SIZE;
	}
	
	GPU_ONLY bool elementAt(id_type idx, MeshElement * elem) const {
		if(idx < kernel_size()) {
			id_type entry_idx = (idx >> CBS_BLOCK_NBITS);
			id_type entry_flag = idx & CBS_MASK;
			Entry & entry = entries[entry_idx];
			elem->init(dMesh,entry.id | entry_flag);
			return entry.mask & (1 << entry_flag);
		} else return false;
	}
	
	struct Entry {
		id_type id;
		uint32_t mask;
	};
	class iterator {
	public:
		GPU_ONLY iterator() {}
		GPU_ONLY iterator(const Mesh * mesh, Entry * start, Entry * end)
		: mesh(mesh), start(start), end(end) {
			count = 0;
		}
		typedef MeshElement value_type;
		GPU_ONLY bool hasNext() const {
			return start != end && start->mask != 0;
		}
		GPU_ONLY MeshElement next() {
			if(start->mask == 0) {
				count = 0;
				start++;
			}
			while( (start->mask & 1) == 0 ) {
				start->mask >>= 1;
				count++;
			}
			start->mask >>= 1;
			
			return MeshElement(mesh,start->id | count);
		}
	private:
		const Mesh * mesh;
		Entry * start;
		Entry * end;
		id_type count;
	};
	GPU_ONLY iterator iter() const {
		return iterator(dMesh,entries,entries+nentries);
	}
	CompressedBoundarySet withMesh(const Mesh * m) {
		CompressedBoundarySet retval = *this;
		retval.dMesh = m;
		return retval;
	}
	void dealloc() {
		if(entries) {
			cudaFree(entries);
			entries = NULL;
		}
	}
private:
	const Mesh * dMesh; //pointer to mesh on GPU
	bool owner; //was this the original boundary set?
	size_t nelems;
	size_t nentries;
	Entry * entries;
};

#undef CBS_BLOCK_SIZE
#undef CBS_MASK
#undef CBS_BLOCK_NBITS
#undef GPU_BOTH
#undef GPU_ONLY
typedef ExplicitBoundarySet BoundarySet;

}//namespace LisztGPU
#endif