/*
 * GPUField.h
 *
 *  Created on: Mar 1, 2010
 *      Author: njoubert
 */

#ifndef GPUFIELD_H_
#define GPUFIELD_H_

#include <map>
#include "global.h"
#define BLOCK_SIZE 512

/*
 * A Gather Field on the GPU is used to spread data over many threads,
 * where a write to the field is written to a index as a (global thread id, field id) pair
 * rather than just a (field id) index. Thus, each thread generates separate values for
 * all the fields that it touches, and a reduction is done later. Or, dual, a field value is
 * written to independently be all the threads that potentially touches it.
 *
 * To not store n^2 values for a field with n keys, we store this as a compressed row, which
 * needs its nonzero values to be built up before moving to the GPU. We should be able to do this with
 * program analysis.
 *
 * Here is a diagram of the values:
 *             0 1 2 3 4 5 6 7 8 9
 * type_indices:  [ 0 2 4 7 8 ..........]
 * thread_indices:[ 5 6|7 5 9|6 5|.......]
 * values:        [ a b|c a d|b a|.......]
 */
using namespace std;

typedef unsigned int idx_type;
typedef unsigned int id_type;

__global__ void GPUMRField_reduceKernelADD_float(
		idx_type tyS,
		idx_type thS,
		idx_type * dtype_indices,
		idx_type * dthread_indices,
		float * dthread_values,
		float * dresult);

__global__ void GPUMRField_reduceKernelADD_float3(
		idx_type tyS,
		idx_type thS,
		idx_type * dtype_indices,
		idx_type * dthread_indices,
		float3 * dthread_values,
		float3 * dresult);

/**
 * THE QUESTION: "How to do += on a GPU in a massively paralle fashion?"
 *
 *
 * THIS CLASS' ANSWER:
 *
 * 		Compute: "for each entry in the field, which threads touch it?"
 * 		Create temporary space for each thread to write to.
 * 		Then, do a reduction to save this.
 *
 *
 */
template<typename VAL>
class GPUMRField {
	//TODO(njoubert): Expose the cresult array so that you can easily populate it.
public:
	/**
	 * Sets up a field of size (0, ..., i, ..., n) where n is size-1.
	 * fillWithOnetoOne indicates whether, by default, you have an entry
	 * corresponding to the i of the index. This is useful if the global
	 * thread index corresponds to a field index it will save a value to.
	 */
	GPUMRField(int size, VAL defaultVal, bool fillOneToOne):
			type_indices_size(size), _onGPU(false), _typeToThreadsMap(), _defaultVal(defaultVal) {

		_cTypeIndices = NULL;
		_cThIndices = NULL;
		_cValues = NULL;

		cresult = new VAL[type_indices_size];
		for (int i = 0; i < type_indices_size; i++) {
			cresult[i] = defaultVal;
		}

		if (fillOneToOne) {
			for (idx_type i = 0; i < type_indices_size; i++) {
				addMapping(i, i);
			}
		}

	}
	//No virtual functions in CUDA, so no polymorphism for yuuuuuu!
	~GPUMRField() {

	}

	//Add a thread that will write to a specific type index
	void addMapping(idx_type type_idx, idx_type th_idx) {
		_typeToThreadsMap.insert(std::pair<idx_type, idx_type>(type_idx, th_idx));
	}

private:
	int generateArraysLocally(idx_type ** typeH, idx_type ** thH, VAL ** vH) {
		idx_type * cTypeIndices = 0;
		idx_type * cThIndices = 0;
		VAL * cValues = 0;

		int countValues = _typeToThreadsMap.size();

		cTypeIndices = new idx_type[type_indices_size];
		cThIndices = new idx_type[countValues];
		cValues = new VAL[countValues];

		//fill values arr with default value
		for (int i = 0; i < countValues; i++) {
			cValues[i] = _defaultVal;
		}

		pair<multimap<idx_type,idx_type>::iterator,multimap<idx_type,idx_type>::iterator> th_index_it;
		multimap<idx_type,idx_type>::iterator it;

		idx_type globalValueIndex = 0; //stores which th_indx you're currently at
		//iterate over all keys in sorted ascending order
		for (idx_type i = 0; i < type_indices_size; i++) {
			//insert global value index into type_indices
			cTypeIndices[i] = globalValueIndex;
			th_index_it = _typeToThreadsMap.equal_range(i);
			for (it=th_index_it.first; it!=th_index_it.second; ++it) {
				//printf("type_idx %d, globalIdx %d, thread_idx %d\n", i, globalValueIndex, (*it).second);
				cThIndices[globalValueIndex] = (*it).second;
				globalValueIndex++;
			}
		}
		*typeH = cTypeIndices;
		*thH = cThIndices;
		*vH = cValues;

		return countValues;
	}

public:
	//converts multimap to CRS arrays on GPU with empty spots for values
	//NOTE: Run this ONLY ONCE
	//todo(njoubert): Make it symmetric so you can move things to and from GPU multiple times?
	void copyToGPU(bool saveCPUArrays) {
		idx_type * cTypeIndices;
		idx_type * cThIndices;
		VAL * cValues;
		thread_indices_size = generateArraysLocally(&cTypeIndices, &cThIndices, &cValues);

		//Allocate the type_indices, and copy it over
		DEVICEALLOC(&dtype_indices, sizeof(idx_type)*type_indices_size);
		cudaMemcpy(dtype_indices, cTypeIndices, type_indices_size*sizeof(idx_type),cudaMemcpyHostToDevice);

		//Allocate the thread_indices, and copy it over
		DEVICEALLOC(&dthread_indices, sizeof(idx_type)*thread_indices_size);
		cudaMemcpy(dthread_indices, cThIndices, thread_indices_size*sizeof(idx_type),cudaMemcpyHostToDevice);

		//Allocate the values, and copy default value over.
		DEVICEALLOC(&dthread_values, sizeof(VAL)*thread_indices_size);
		cudaMemcpy(dthread_values, cValues, thread_indices_size*sizeof(VAL),cudaMemcpyHostToDevice);

		//Allocate space for the result, copying over the same default value
		DEVICEALLOC(&dresult, sizeof(VAL)*type_indices_size);
		cudaMemcpy(dresult, cresult, type_indices_size*sizeof(VAL),cudaMemcpyHostToDevice);

		if (saveCPUArrays) {
			if (_cTypeIndices != NULL) {
				delete [] _cTypeIndices;
				delete [] _cThIndices;
				delete [] _cValues;
			}
			_cTypeIndices = cTypeIndices;
			_cThIndices = cThIndices;
			_cValues = cValues;
		} else {
			delete [] cTypeIndices;
			delete [] cThIndices;
			delete [] cValues;
		}
	}

	/*
	 * Returns NULL if no such mapping can be found.
	 */
	__device__ VAL * dGetLocationForMapping(idx_type type_idx, idx_type th_idx) {
		idx_type th_idx_s;
		idx_type th_idx_e;

		//get the start and end of map keys in th_indx array
		th_idx_s = dtype_indices[type_idx];
		if (type_idx = type_indices_size-1)
			th_idx_e = thread_indices_size;
		else
			th_idx_e = dtype_indices[type_idx+1];

		//search for my th_idx:
		for (idx_type i = th_idx_s; i < th_idx_e; i++) {
			if (dthread_indices[i] == th_idx) {
				return &dthread_values[i];
			}
		}
		return NULL;
	}

	/*
	 * Returns NULL if no such mapping can be found.
	 */
	__host__ VAL * cGetLocationForMapping(idx_type type_idx, idx_type th_idx) {
		if (_cTypeIndices == NULL)
			return NULL;

		idx_type th_idx_s;
		idx_type th_idx_e;

		//get the start and end of map keys in th_indx array
		th_idx_s = _cTypeIndices[type_idx];
		if (type_idx = type_indices_size-1)
			th_idx_e = thread_indices_size;
		else
			th_idx_e = _cTypeIndices[type_idx+1];

		//search for my th_idx:
		for (idx_type i = th_idx_s; i < th_idx_e; i++) {
			if (_cThIndices[i] == th_idx) {
				return &_cValues[i];
			}
		}
		return NULL;
	}

	//launches kernel that reduces field to a single flat array.
	//It will add the result to this, so you can keep reducing multiple passes.
	void reduce();

	/*
	 * Must run the reduce() function first!
	 */
	VAL const * copyResultFromGPU() {
		cudaMemcpy(cresult, dresult, type_indices_size*sizeof(VAL),cudaMemcpyDeviceToHost);
		return cresult;
	}

public:
	idx_type type_indices_size;
	idx_type thread_indices_size;

	idx_type * dtype_indices;
	idx_type * dthread_indices;
	VAL * dthread_values;	//same size as thread_indices_size;
	VAL * dresult;	//same size as type_indices_size, since it stores reduced value.

	VAL * cresult; //result as on the CPU
	idx_type * _cTypeIndices;
	idx_type * _cThIndices;
	VAL * _cValues;

public:
	bool _onGPU;
	VAL _defaultVal;
	std::multimap<idx_type, idx_type> _typeToThreadsMap;
};

template <>
inline void GPUMRField<float>::reduce() {
	dim3 dimBlock(BLOCK_SIZE);
	dim3 dimGrid(ceil((float)type_indices_size / (float)BLOCK_SIZE));
	GPUMRField_reduceKernelADD_float<<<dimGrid,dimBlock>>>(
			type_indices_size,thread_indices_size,
			dtype_indices,dthread_indices,
			dthread_values,
			dresult);
}

template <>
inline void GPUMRField<float3>::reduce() {
	dim3 dimBlock(BLOCK_SIZE);
	dim3 dimGrid(ceil((float)type_indices_size / (float)BLOCK_SIZE));
	GPUMRField_reduceKernelADD_float3<<<dimGrid,dimBlock>>>(
			type_indices_size,thread_indices_size,
			dtype_indices,dthread_indices,
			dthread_values,
			dresult);
}

#undef BLOCK_SIZE

#endif /* GPUFIELD_H_ */
