#ifndef _GPU_MESH_FUNCTIONS
#define _GPU_MESH_FUCNTIONS

#include <assert.h>
#include "MeshIO/LisztFormat.h"

#define LISZT_EXTERN(ret,name) \
typedef ret name##_return_type; \
GPU_DECLS static inline ret name

//this is implemented in kernel.cu
__device__ static inline id_type GPU_getGlobalID(MeshIO::IOElemType, id_type id);

namespace LisztGPU {
#define GPU_DECLS __device__ __host__

GPU_DECLS static inline id_type INTERNAL_ID(id_type id) {
	return ~MESH_DIRECTION_FLAG & id;
}

typedef MeshElement Vertex;
typedef MeshElement Edge;
typedef MeshElement Face;
typedef MeshElement Cell;

class TopLevelSet {
public:
	GPU_DECLS TopLevelSet() {}
	GPU_DECLS TopLevelSet(const Mesh * m, id_type s, id_type e) 
	: mesh(m), start(s), end(e) {}
	class iterator {
	public:
		typedef MeshElement value_type;
		GPU_DECLS iterator() {}
		GPU_DECLS iterator(const Mesh * m, id_type s, id_type e) 
		: mesh(m), start(s), end(e) {}
		GPU_DECLS bool hasNext() const {
			return start != end;
		}
		GPU_DECLS MeshElement next() {
			return MeshElement(mesh,start++);
		}
	private:
		const Mesh * mesh;
		id_type start,end;
	};
	GPU_DECLS iterator iter() const {
		return iterator(mesh,start,end);
	}
	GPU_DECLS size_t size() const {
		return end - start;
	}
	GPU_DECLS size_t kernel_size() const {
		return size();
	}
	GPU_DECLS bool elementAt(id_type id,MeshElement * elem) const {
		id_type idx = start + id;
		if(idx < end) {
			elem->init(mesh,idx);
			return true;
		} else return false;
	}
	GPU_DECLS TopLevelSet withMesh(const Mesh * m) {
		return TopLevelSet(m, start, end);
	}
	private:
		const Mesh * mesh;
		id_type start,end;
};

class DirectedSet {
public:
	GPU_DECLS DirectedSet() {}
	
	GPU_DECLS DirectedSet(const Mesh * m, id_type * s, id_type * e) 
	: mesh(m), start(s), end(e) {
		incr = (end < start)  ? -1 : 1;
	}
	
	class iterator {
	public:
		typedef MeshElement value_type;
		GPU_DECLS iterator() {}
		GPU_DECLS iterator(const Mesh * m, id_type * s, id_type * e, id_type i) 
		: mesh(m), start(s), end(e), incr(i) {}
		GPU_DECLS bool hasNext() const {
			return start != end;
		}
		GPU_DECLS MeshElement next() {
			//if we are going backward (i.e. MESH_DIRECTION_FLAG & incr != 0),
			//we need to flip the directions of the objects we return
			id_type new_id = *start ^ (MESH_DIRECTION_FLAG & incr);
			MeshElement ret(mesh,new_id);
			start += incr;
			return ret;
		}
	private:
		const Mesh * mesh;
		id_type * start;
		id_type * end;
		int incr;
	};
	
	GPU_DECLS iterator iter() const {
		return iterator(mesh,start,end,incr);
	}
	GPU_DECLS size_t size() const {
		return abs(end - start);
	}
	GPU_DECLS size_t kernel_size() const {
		return size();
	}
	GPU_DECLS bool elementAt(id_type index, MeshElement * elem) const {
		if(index < size()) {
			id_type * value = (incr > 0) ? start + index : start - index;
			elem->init(mesh,*value);
			return true;
		} else return false;
	}
	GPU_DECLS DirectedSet withMesh(const Mesh * m) {
		return DirectedSet(m, start, end);
	}
private:
	const Mesh * mesh;
	id_type * start;
	id_type * end;
	id_type incr;
};

class ConstSet {
public:
	GPU_DECLS ConstSet() {}
	GPU_DECLS ConstSet(const Mesh * m, id_type * d) 
	: mesh(m), data(d) {}
	class iterator {
	public:
		typedef MeshElement value_type;
		GPU_DECLS iterator() {}
		GPU_DECLS iterator(const Mesh * m, id_type * d) : mesh(m), data(d), cur(0) {}
		GPU_DECLS bool hasNext() const {
			return  cur < 2;
		}
		GPU_DECLS MeshElement next() {
			return MeshElement(mesh,data[cur++]);
		}
	private:
		const Mesh * mesh;
		id_type * data;
		id_type cur;
	};
	GPU_DECLS iterator iter() const {
		return iterator(mesh,data);
	}
	GPU_DECLS size_t size() const {
		return 2;
	}
	GPU_DECLS size_t kernel_size() const {
		return size();
	}
	GPU_DECLS bool elementAt(id_type index, MeshElement * elem) const {
		if(index < size()) {
			elem->init(mesh,data[index]);
			return true;
		} else return false;
	}
	GPU_DECLS ConstSet withMesh(const Mesh * m) {
		return ConstSet(m, data);
	}
private:
	const Mesh * mesh;
	id_type * data;
};

template<typename Set1, typename Set2>
class EitherSet {
public:
    class iterator {
        typedef typename Set1::iterator it1_t;
        typedef typename Set2::iterator it2_t;
        
    public:
        typedef typename Set1::iterator::value_type value_type;
        GPU_DECLS iterator() {}
        GPU_DECLS iterator(const Set1 & s) {
            data.i1 = s.iter();
            isLeft = true;
        }
        GPU_DECLS iterator(const Set2 & s) {
            data.i2 = s.iter();
            isLeft = false;
        }
        GPU_DECLS bool hasNext() const {
            if(isLeft)
                return data.i1.hasNext();
            else
                return data.i2.hasNext();
        }
        GPU_DECLS value_type next() {
            if(isLeft)
                return data.i1.next();
            else
                return data.i2.next();
        }
    private:
        struct {
            it1_t i1;
            it2_t i2;
        } data;
        bool isLeft;
    };
    typedef typename iterator::value_type value_type;
    GPU_DECLS  EitherSet() {}
    GPU_DECLS EitherSet(const Set1 & s) {
        data.s1 = s;
        isLeft = true;
    }
    GPU_DECLS EitherSet(const Set2 & s) {
        data.s2 = s;
        isLeft = false;
    }
    GPU_DECLS  size_t size() const {
        if(isLeft)
            return data.s1.size();
        else
            return data.s2.size();
    }
    GPU_DECLS  size_t kernel_size() const {
        if(isLeft)
            return data.s1.kernel_size();
        else
            return data.s2.kernel_size();
    }
    GPU_DECLS iterator iter() const {
        if(isLeft)
            return iterator(data.s1);
        else
            return iterator(data.s2);
    }
    GPU_DECLS bool elementAt(id_type index, MeshElement * elem) const {
        if(isLeft)
            return data.s1.elementAt(index,elem);
        else
            return data.s2.elementAt(index,elem);
    }
    GPU_DECLS EitherSet<Set1,Set2> withMesh(const Mesh * m) {
        if(isLeft)
            return EitherSet<Set1,Set2>(data.s1.withMesh(m));
        else
            return EitherSet<Set1,Set2>(data.s2.withMesh(m));
    }
private:
    bool isLeft;
    struct {
        Set1 s1;
        Set2 s2;
    } data;
};

GPU_DECLS static inline bool operator==(const MeshElement & a, const MeshElement & b) {
	return INTERNAL_ID(a.id) == INTERNAL_ID(b.id);
}

GPU_DECLS static inline bool operator!=(const MeshElement & a, const MeshElement & b) {
	return  !(a == b);
}

__device__ static inline id_type ID_v(const MeshElement & v) {
	return GPU_getGlobalID(MeshIO::VERTEX_T,INTERNAL_ID(v.id));
}
__device__ static inline id_type ID_e(const MeshElement & v) {
	return GPU_getGlobalID(MeshIO::EDGE_T,INTERNAL_ID(v.id));
}
__device__ static inline id_type ID_f(const MeshElement & v) {
	return GPU_getGlobalID(MeshIO::FACE_T,INTERNAL_ID(v.id));
}
__device__ static inline id_type ID_c(const MeshElement & v) {
	return GPU_getGlobalID(MeshIO::CELL_T,INTERNAL_ID(v.id));
}
GPU_DECLS static inline DirectedSet INTERNAL_RANGE(const CRS & crs, MeshElement elem) {
	const Mesh * mesh = elem.mesh;
	id_type id = elem.id;
	id_type * start;
	id_type * end;
	id_type iid = INTERNAL_ID(id);
	id_type * first = crs.values + crs.row_idx[iid];
	id_type * last = crs.values + crs.row_idx[iid+1];
	if((id & MESH_DIRECTION_FLAG) == 0) {
		start = first;
		end = last;
	} else {
		start = last - 1;
		end = first - 1;
	}
	return DirectedSet(mesh,start,end);
}


GPU_DECLS static inline Face flip_f(const Face & f) {
	return MeshElement(f.mesh,f.id ^ MESH_DIRECTION_FLAG);
}

GPU_DECLS static inline Edge flip_e(const Edge & e) {
	return MeshElement(e.mesh,e.id ^ MESH_DIRECTION_FLAG);
}

LISZT_EXTERN(TopLevelSet,vertices_m)(const Mesh * m) {
	return TopLevelSet(m,0,m->nvertices);
}

LISZT_EXTERN(DirectedSet,vertices_v)(const Vertex & m) {
   return INTERNAL_RANGE(m.mesh->vtov,m);
}

LISZT_EXTERN(ConstSet,vertices_e)(const Edge & m) {
    return ConstSet(m.mesh,m.mesh->etov.values[INTERNAL_ID(m.id)].data);
}

LISZT_EXTERN(DirectedSet,verticesCCW_f)(const Face & m) {
    return INTERNAL_RANGE(m.mesh->ftov,m);
}


LISZT_EXTERN(DirectedSet,vertices_f)(const Face & m) {
   return verticesCCW_f(m);
}

LISZT_EXTERN(DirectedSet,verticesCW_f)(const Face & m) {
   return verticesCCW_f(flip_f(m));
}

LISZT_EXTERN(DirectedSet,vertices_c)(const Cell & m) {
    return INTERNAL_RANGE(m.mesh->ctov,m);
}

LISZT_EXTERN(TopLevelSet,cells_m)(const Mesh * m) {
   return TopLevelSet(m,1,m->ncells);
}

LISZT_EXTERN(DirectedSet,cells_v)(const Vertex & m) {
    return INTERNAL_RANGE(m.mesh->vtoc,m);
}

LISZT_EXTERN(DirectedSet,cellsCCW_e)(const Edge & m) {
    return INTERNAL_RANGE(m.mesh->etoc,m);
}

LISZT_EXTERN(DirectedSet,cells_e)(const Edge & m) {
   return cellsCCW_e(m);
}

LISZT_EXTERN(DirectedSet,cellsCW_e)(const Edge & m) {
	return cellsCCW_e(flip_e(m));
}

LISZT_EXTERN(ConstSet,cells_f)(const Face & m) {
	return ConstSet(m.mesh,m.mesh->ftoc.values[INTERNAL_ID(m.id)].data);
}

LISZT_EXTERN(DirectedSet,cells_c)(const Cell & m) {
	return INTERNAL_RANGE(m.mesh->ctoc,m);
}

LISZT_EXTERN(TopLevelSet,edges_m)(const Mesh * m) {
	return TopLevelSet(m,0,m->nedges);
}

LISZT_EXTERN(DirectedSet,edges_v)(const Vertex & m) {
	return INTERNAL_RANGE(m.mesh->vtoe,m);
}


LISZT_EXTERN(DirectedSet,edgesCCW_f)(const Face & m) {
	return INTERNAL_RANGE(m.mesh->ftoe,m);
}


LISZT_EXTERN(DirectedSet,edges_f)(const Face & m) {
	return edgesCCW_f(m);
}

LISZT_EXTERN(DirectedSet,edges_c)(const Cell & m) {
	return INTERNAL_RANGE(m.mesh->ctoe,m);
}

LISZT_EXTERN(DirectedSet,edgesCW_f)(const Face & m) {
    return edgesCCW_f(flip_f(m));
}

LISZT_EXTERN(TopLevelSet,faces_m)(const Mesh * m) {
    return TopLevelSet(m,0,m->nfaces);
}

LISZT_EXTERN(DirectedSet,faces_v)(const Vertex & m) {
    return INTERNAL_RANGE(m.mesh->vtof,m);
}

LISZT_EXTERN(DirectedSet,facesCCW_e)(const Edge & m) {
    return INTERNAL_RANGE(m.mesh->etof,m);
}

LISZT_EXTERN(DirectedSet,faces_e)(const Edge & m) {
    return facesCCW_e(m);
}

LISZT_EXTERN(DirectedSet,faces_c)(const Cell & m) {
	return INTERNAL_RANGE(m.mesh->ctof,m);
}

LISZT_EXTERN(DirectedSet,facesCW_e)(const Edge & m) {
    return facesCCW_e(flip_e(m));
}


GPU_DECLS static inline Vertex head(const Edge & e) {
	id_type index = static_cast<uint32_t>(e.id & MESH_DIRECTION_FLAG) >> MESH_DIRECTION_BIT;
	return MeshElement(e.mesh,e.mesh->etov.values[INTERNAL_ID(e.id)].data[index]);
}

GPU_DECLS static inline Vertex tail(const Edge & e) {
	id_type index = static_cast<uint32_t>(e.id & MESH_DIRECTION_FLAG) >> MESH_DIRECTION_BIT;
	return MeshElement(e.mesh,e.mesh->etov.values[INTERNAL_ID(e.id)].data[index ^ 1]);
}

GPU_DECLS static inline Cell outside(const Face & e) {
	id_type index = static_cast<uint32_t>(e.id & MESH_DIRECTION_FLAG) >> MESH_DIRECTION_BIT;
	return MeshElement(e.mesh,e.mesh->ftoc.values[INTERNAL_ID(e.id)].data[index]);
}

GPU_DECLS static inline Cell inside(const Face & e) {
	id_type index = static_cast<uint32_t>(e.id & MESH_DIRECTION_FLAG) >> MESH_DIRECTION_BIT;
	return MeshElement(e.mesh,e.mesh->ftoc.values[INTERNAL_ID(e.id)].data[index ^ 1]);
}

//Extract by label functions
#include "LabelExtractXMacro.h"

template<typename SetT>
GPU_DECLS static inline MeshElement getElementAt(const SetT & s, int lbl) {
	MeshElement elem;
	s.elementAt(lbl,&elem);
	return elem;
}

#define LABEL_EXTRACT_TEMPLATE(fn,tag,r_type,a_type,plural_fn) \
GPU_DECLS static inline r_type fn##_##tag(const a_type & e, int lbl) { \
	return getElementAt(plural_fn##_##tag(e),lbl); \
}
LABEL_EXTRACTORS(LABEL_EXTRACT_TEMPLATE)
#undef LABEL_EXTRACT_TEMPLATE


#define LABEL_EXTRACT_CONST_TEMPLATE(fn,tag,r_type,a_type,zero_fn,one_fn) \
GPU_DECLS static inline r_type fn##_##tag(const a_type & e, int lbl) { \
	if(lbl == 0) return zero_fn(e); \
	else return one_fn(e); \
}
LABEL_EXTRACTORS_CONST(LABEL_EXTRACT_CONST_TEMPLATE)
#undef LABEL_EXTRACT_CONST_TEMPLATE

template<typename Set>
GPU_DECLS static inline size_t size_v(const Set & s) {
    return s.size();
}

template<typename Set>
GPU_DECLS static inline size_t size(const Set & s) {
    return s.size();
}


template<typename Set>
GPU_DECLS static inline size_t size_e(const Set & s) {
    return s.size();
}

template<typename Set>
GPU_DECLS static inline size_t size_f(const Set & s) {
    return s.size();
}

template<typename Set>
GPU_DECLS static inline size_t size_c(const Set & s) {
    return s.size();
}


GPU_DECLS static inline Face towards_f(const Face & f, const Cell & c) {
	if( INTERNAL_ID(outside(f).id) == INTERNAL_ID(c.id) )
		return f;
	else {
		return flip_f(f);
	}
}

GPU_DECLS static inline Edge towards_e(const Edge & e, const Vertex & v) {
	if( INTERNAL_ID(head(e).id) == INTERNAL_ID(v.id) )
		return e;
	else {
		return flip_e(e);
	}
}
#undef GPU_DECLS
} //namespace LisztGPU
#endif