/*
 * GPURuntime.h
 *
 * This file is only included for the nvcc compiler, and thus 
 * makes some things available to the CUDA side that is NOT
 * available on the CPU side in GPURuntime.h
 * 
 *  Created on: Apr 18, 2010
 *      Author: njoubert
 */

#ifndef GPUCUDARUNTIME_H_
#define GPUCUDARUNTIME_H_

#define GPURUNTIME

#define BLOCKSIZE 256

typedef unsigned long long int gpu_uint64_t;

#if defined(PROFILE_COUNTING)
__device__ gpu_uint64_t dev_bytes_accessed;
__device__ gpu_uint64_t dev_flops_performed;
__shared__ gpu_uint64_t dev_flops_performed_shared[BLOCKSIZE];
#endif

#include <cstdlib>
#include <cstdio>
#include <ctime>
#include <iostream>

#include <math_functions.h>
#include <cutil_math.h>

//Here we define things to manage printf-ing on different computa capabilities:
#if CUDA_COMPUTE_MAJOR >= 2
#define LisztGPU__printf printf
#define LisztGPU_DISPLAY_PRINTF cudaThreadSynchronize()
#define LisztGPU_INIT_PRINTF 0
#define LisztGPU_TERMINATE_PRINTF 0
#else
#include "Gpu/cuPrintf.cu"
#define LisztGPU__printf cuPrintf
#define LisztGPU_DISPLAY_PRINTF cudaPrintfDisplay(stdout, true)
#define LisztGPU_INIT_PRINTF cudaPrintfInit(104857600)
#define LisztGPU_TERMINATE_PRINTF cudaPrintfEnd()
#endif

#include "ForAll.h"
#include "Gpu/GPURuntimeJoint.h"
#include "Gpu/GPUMeshFunctions.h"
//#include "Gpu/Field.h"
#include "Liszt/ElemTypes.h"
#include "Gpu/GPUPrint.h"

#include "Gpu/Global.h"
#include "Gpu/Profiling.h"



/*
 * KERNEL CALL MACROS
 * This abstracts away how we launch kernels. Uses the x-macros pattern.
 *
 * The KERNEL_CALL macro expects the following to be defined when called:

#define ENSURE_SIZE(_) \
	_(sCount_1077.ensureSize(batch.kernel_size())) \
	_(checkCount_1078.ensureSize(batch.kernel_size()))

#define KERNEL_ARGS(_) \
	_(checkCount_1078) \
	_(sCount_1077)

 *
 */

//functions for implementing x-macros for KERNEL_CALL macro.
#define GRDM_ARGS(_) ,_
#define GRDM_CALL(_) _;

#define KERNEL_CALL(workgroup_id, mesh_set, workgroup_istrivial, kernelfunc, ensure_size_M, kernel_args_M) do {\
	WorkgroupLauncher launcher; \
	WorkgroupLauncher_forWorkgroup(workgroup_id,mesh_set,workgroup_istrivial,&launcher); \
	GRHM_RESET_BYTES_ACCESSED \
	GRDM_PROFILE_TIMING_START(workgroup_id) \
	ColorBatch batch; \
	while(launcher.nextBatch(&batch)) { \
		ensure_size_M(GRDM_CALL) \
		GlobalContext_copyToGPU(&cpuGC); \
		kernelfunc<<<(batch.kernel_size() + BLOCKSIZE - 1)/BLOCKSIZE,BLOCKSIZE>>>(batch kernel_args_M(GRDM_ARGS)); \
	} \
	GRDM_PROFILE_TIMING_STOP(workgroup_id) \
	GRHM_OUTPUT_COUNTING_DETAILS(workgroup_id, launcher.numColors()) \
	check_launch(#kernelfunc); \
	} while(0)
	
/*
 * END KERNEL CALL MACROS
 */
 
/* KERNEL DEFINITION MACROS */

#ifdef PROJECT_DEBUG

#define PRINT_CONTEXT_INIT(kernel_id,thread_var,loop_type_char) \
PrintContext_init(); \
PrintContext print_context_##kernel_id; \
PrintContext_push(&print_context_##kernel_id); \
PrintContext_mark_##loop_type_char(thread_var);

#else

#define PRINT_CONTEXT_INIT(kernel_id,thread_var,loop_type_char)

#endif


#define KERNEL_DEFINE_BEGIN(kernel_id,thread_var,loop_set,loop_type,loop_type_char) \
int thread_var##_##kernel_id = blockIdx.x * blockDim.x + threadIdx.x; \
loop_type thread_var; \
if(loop_set.elementAt(thread_var##_##kernel_id,&thread_var)) { \
PRINT_CONTEXT_INIT(kernel_id,thread_var,loop_type_char) \
GRDM_KERNEL_BEGIN

#define KERNEL_DEFINE_END(kernel_id,thread_var,loop_set,loop_type,loop_type_char) \
GRDM_KERNEL_END \
}

using namespace LisztPrivate;

namespace LisztGPU {

struct event_pair
{
  cudaEvent_t start;
  cudaEvent_t end;
};

void printMeshNumbers(Mesh & m) {
	printf("nr of output values: %ld\n", m.ncells);
}

inline void check_launch(char * kernel_name) {
	cudaThreadSynchronize();
	cudaError_t err;
	if((err = cudaGetLastError()) != cudaSuccess) {
		printf("error \"%s\" on %s kernel\n",cudaGetErrorString(err), kernel_name);
		exit(1);
	}
}

inline void start_timer(event_pair * p) {
	cudaEventCreate(&p->start);
	cudaEventCreate(&p->end);
	cudaEventRecord(p->start, 0);
}

inline void stop_timer(event_pair * p, char * kernel_name) {
	cudaEventRecord(p->end, 0);
	cudaEventSynchronize(p->end);
	
	float elapsed_time;
	cudaEventElapsedTime(&elapsed_time, p->start, p->end);
	printf("%s took %.4f ms\n",kernel_name, elapsed_time);
	cudaEventDestroy(p->start);
	cudaEventDestroy(p->end);
}

inline float stop_get_timer(event_pair * p) {
	cudaEventRecord(p->end, 0);
	cudaEventSynchronize(p->end);
	
	float elapsed_time;
	cudaEventElapsedTime(&elapsed_time, p->start, p->end);
	cudaEventDestroy(p->start);
	cudaEventDestroy(p->end);
	return elapsed_time;
}

template<typename MO,typename VT>
void Field_initWithConst(Field<MO,VT> * field, const VT & v) {
	size_t nelems;
	switch(MO::value) {
		case VERTEX:
			nelems = hMesh->nvertices;
			break;
		case EDGE:
			nelems = hMesh->nedges;
			break;
		case FACE:
			nelems = hMesh->nfaces;
			break;
		case CELL:
			nelems = hMesh->ncells;
			break;
	}
	field->init(nelems,v);
}

#undef GPURUNTIME

}

#endif /* GPUCUDARUNTIME_H_ */
