/*
 * GPURuntime.h
 *
 * THE AWESOMENESS WILL NOT BE OVERCOME!
 * This file is only included on the G++ side of the CUDA runtime
 * so it does not contain and GPU types.
 * 
 *  Created on: Apr 18, 2010
 *      Author: njoubert
 */

#ifndef GPURUNTIME_H_
#define GPURUNTIME_H_

#define GPURUNTIME
//// INCLUDE THIS IF YOU WANT TO BE ABLE TO MAKE CUDA CALLS FROM C++ CODE
/////////////////////////////////////////////////////////////////////////
#include <cuda_runtime.h>
#include <cuda.h>

#include "CRSMesh/CRSMesh.h"

#include "Vector.h"
#include "Matrix.h"
#include "ForAll.h"

#include <cstdlib>
#include <cstdio>
#include <ctime>

#include "Gpu/Util.h"


#include "Gpu/GPURuntimeJoint.h"
#include "Gpu/BoundarySet.h"

#include "MeshIO/LisztFileReader.h"
#include "MeshIO/FacetEdgeBuilder.h"

#include "Layout/BoundarySetBuilder.h"
#include "Layout/Partitioner.h"


using Synchronization::ColoringController;
using Synchronization::InterferenceBuilder;
using Synchronization::SimpleAbstractWriteContext;

#define COLORING_BLOCK_SIZE 1

#if defined(PROFILE_COUNTING) || defined(PROFILE_TIMING)
FILE* profiling_file;
#endif

namespace LisztGPU {

MeshIO::LisztFileReader mesh_file;
BoundarySetBuilder boundary_builder;
MeshIO::ResultFilter partitionRenameFilter;
bool partition_and_block;

CRS dAllocCRS(int size, int nvalues) {
	CRS dCRS;
	DEVICEALLOC((void**)&dCRS.row_idx, sizeof(idx_type) * (size+1));
	DEVICEALLOC((void**)&dCRS.values, sizeof(id_type) * nvalues);
	return dCRS;
}
void dFreeCRS(CRS dCRS) {
	cudaFree((void**)&dCRS.row_idx);
	cudaFree((void**)&dCRS.values);
}

CRSConst dAllocCRSConst(int size) {
	CRSConst dCRS;
	DEVICEALLOC((void**)&dCRS.values, sizeof(IDPair) * size);
	return dCRS;
}
void dFreeCRSConst(CRSConst dCRS) {
	cudaFree((void**)&dCRS.values);
}

//This will allocate, and then copy, the CPU representation to the GPU.
CRS dAdaptCRS(CRSMeshPrivate::CRS& a, int size) {
	int nvalues = a.row_idx[size];
	CRS dCRS = dAllocCRS(size, nvalues);
	cudaMemcpy(dCRS.row_idx, a.row_idx,(size+1)*sizeof(idx_type),cudaMemcpyHostToDevice);
	cudaMemcpy(dCRS.values, a.values,(nvalues)*sizeof(id_type),cudaMemcpyHostToDevice);
	return dCRS;
}

CRSConst dAdaptCRSConst(CRSMeshPrivate::CRSConst& a, int size) {
	CRSConst dCRSc = dAllocCRSConst(size);
	cudaMemcpy(dCRSc.values, a.values,size*sizeof(IDPair),cudaMemcpyHostToDevice);
	return dCRSc;
}

void Mesh_init(CRSMesh::Mesh * myMesh, Mesh** ptrhMesh, Mesh** dMesh) {
	(*ptrhMesh) = (Mesh*) malloc(sizeof(Mesh));
	Mesh* hMesh = (*ptrhMesh);
	
	hMesh->nvertices = myMesh->data.nvertices;
	hMesh->nedges =    myMesh->data.nedges;
	hMesh->nfaces =    myMesh->data.nfaces;
	hMesh->ncells =    myMesh->data.ncells;

	hMesh->vtov = dAdaptCRS(myMesh->data.vtov, myMesh->data.nvertices);
	hMesh->vtoe = dAdaptCRS(myMesh->data.vtoe, myMesh->data.nvertices);
	hMesh->vtof = dAdaptCRS(myMesh->data.vtof, myMesh->data.nvertices);
	hMesh->vtoc = dAdaptCRS(myMesh->data.vtoc, myMesh->data.nvertices);

	hMesh->etov = dAdaptCRSConst(myMesh->data.etov, myMesh->data.nedges);
	hMesh->etof = dAdaptCRS(myMesh->data.etof,      myMesh->data.nedges);
	hMesh->etoc = dAdaptCRS(myMesh->data.etoc,      myMesh->data.nedges);

	hMesh->ftov = dAdaptCRS(myMesh->data.ftov,      myMesh->data.nfaces);
	hMesh->ftoe = dAdaptCRS(myMesh->data.ftoe,      myMesh->data.nfaces);
	hMesh->ftoc = dAdaptCRSConst(myMesh->data.ftoc, myMesh->data.nfaces);

	hMesh->ctov = dAdaptCRS(myMesh->data.ctov, myMesh->data.ncells);
	hMesh->ctoe = dAdaptCRS(myMesh->data.ctoe, myMesh->data.ncells);
	hMesh->ctof = dAdaptCRS(myMesh->data.ctof, myMesh->data.ncells);
	hMesh->ctoc = dAdaptCRS(myMesh->data.ctoc, myMesh->data.ncells);

	DEVICEALLOC((void**)dMesh, sizeof(Mesh));
	cudaMemcpy(*dMesh, hMesh, sizeof(Mesh), cudaMemcpyHostToDevice);
	printf("Mesh allocated on CPU and GPU\n");
}

//create a version of the Mesh data structure capable of running on the cpu
void Mesh_buildCPUMesh(CRSMesh::Mesh * myMesh, Mesh * cpuMesh) {
	#define COPY_CRS(name) do { \
	cpuMesh->name.row_idx = myMesh->data.name.row_idx; \
	cpuMesh->name.values = myMesh->data.name.values; \
	} while(0)
	#define COPY_CRSCONST(name) do { \
	cpuMesh->name.values = (::LisztGPU::IDPair*) myMesh->data.name.values; \
	} while(0)
	
	cpuMesh->nvertices = myMesh->data.nvertices;
	cpuMesh->nedges = myMesh->data.nedges;
	cpuMesh->nfaces = myMesh->data.nfaces;
	cpuMesh->ncells = myMesh->data.ncells;
	
	COPY_CRS(vtov);
	COPY_CRS(vtoe);
	COPY_CRS(vtof);
	COPY_CRS(vtoc);
	
	COPY_CRSCONST(etov);
	COPY_CRS(etof);
	COPY_CRS(etoc);
	
	COPY_CRS(ftov);
	COPY_CRS(ftoe);
	COPY_CRSCONST(ftoc);
	
	COPY_CRS(ctov);
	COPY_CRS(ctoe);
	COPY_CRS(ctof);
	COPY_CRS(ctoc);
	
	#undef COPY_CRS
	#undef COPY_CRSCONST
}

template<typename MO>
void BoundarySet_init(const char * name, BoundarySet * boundary) {
	if(partition_and_block)
		boundary_builder.loadWithFilter<Mesh,MO>(dummydMesh,name,partitionRenameFilter,boundary);
	else
		boundary_builder.load<Mesh,MO>(dummydMesh,name,boundary);
}

template void BoundarySet_init<VertexType>(const char * name, BoundarySet * boundary);
template void BoundarySet_init<EdgeType>(const char * name, BoundarySet * boundary);
template void BoundarySet_init<FaceType>(const char * name, BoundarySet * boundary);
template void BoundarySet_init<CellType>(const char * name, BoundarySet * boundary);

template<typename VT>
class FieldInitWithPositions {
public:
	static void init(Field<VertexType,VT > * field);
};

template<typename VT>
void FieldInitWithPositions<VT>::init(Field<VertexType,VT > * field) {
	 MeshIO::PositionTable * table = mesh_file.positions();
	 size_t nV = mesh_file.header().nV;
	 VT * data = new VT[nV];
	 for(size_t i = 0; i < nV; i++) {
	 	size_t entry = (partition_and_block) ? partitionRenameFilter(MeshIO::VERTEX_T,i) : i;
	 	data[entry][0] = table->data[i][0];
	 	data[entry][1] = table->data[i][1];
	 	data[entry][2] = table->data[i][2];
	 }
	 mesh_file.free(table);
	 field->init(nV,data);
	 delete [] data;
}
template<>
void FieldInitWithPositions<float3>::init(Field<VertexType,float3 > * field) {
	 MeshIO::PositionTable * table = mesh_file.positions();
	 size_t nV = mesh_file.header().nV;
	 float3 * data = new float3[nV];
	 for(size_t i = 0; i < nV; i++) {
	 	size_t entry = (partition_and_block) ? partitionRenameFilter(MeshIO::VERTEX_T,i) : i;
	 	data[entry].x = table->data[i][0];
	 	data[entry].y = table->data[i][1];
	 	data[entry].z = table->data[i][2];
	 }
	 mesh_file.free(table);
	 field->init(nV,data);
	 delete [] data;
}


template<typename VT>
void Field_initWithPositions(Field<VertexType, VT > * field) {
	FieldInitWithPositions<VT>::init(field);
}

template void Field_initWithPositions< vec<3,double> >(Field<VertexType, vec<3,double> > * field);

template void Field_initWithPositions< vec<3,float> >(Field<VertexType, vec<3,float> > * field);

template void Field_initWithPositions< float3 >(Field<VertexType, float3 > * field);


} //namespace LisztGPU

static CUdevice cudaDevice;
static CUcontext cudaContext;
ColoringController coloring_controller;

void initialize_cuda() {
	cuInit(0);
	cuDeviceGet(&cudaDevice,0);
	cuCtxCreate(&cudaContext,CU_CTX_SCHED_SPIN | CU_CTX_MAP_HOST, cudaDevice);
	cudaThreadSetCacheConfig(cudaFuncCachePreferL1);
}

void teardown_cuda() {
	cuCtxDestroy(cudaContext);
}
inline void check_launch(const char * kernel_name) {
	cudaThreadSynchronize();
	cudaError_t err;
	if((err = cudaGetLastError()) != cudaSuccess) {
		printf("error \"%s\" on %s kernel\n",cudaGetErrorString(err), kernel_name);
		exit(1);
	}
}

int main(int argc, char** argv) {
	
	if (argc != 3) {
		printf("USAGE: main <mesh.lmsh> <partition_and_block>\n");
		exit(1);
	}
	initialize_cuda();
	
#ifdef PROJECT_DEBUG
	printf("adjusting printf and stack size for debug.\n");
	cuCtxSetLimit(CU_LIMIT_PRINTF_FIFO_SIZE,0x1000000);
	cudaThreadSetLimit(cudaLimitStackSize,0x2000);
#endif
#ifdef PROFILE_COUNTING
	profiling_file = fopen("profiling_counting_output.log", "w");
#endif
#ifdef PROFILE_TIMING
	profiling_file = fopen("profiling_timing_output.log", "w");
#endif

	printf("Initializing Mesh...\n");
	mesh_file.init(argv[1]);
	partition_and_block = atoi(argv[2]) ? true : false;
	MeshIO::FileFacetEdge * ffe = mesh_file.facetEdges();
	MeshIO::FacetEdgeBuilder builder;
	const MeshIO::LisztHeader & h = mesh_file.header();
	builder.init(h.nV,h.nE,h.nF,h.nC,h.nFE);
	builder.insert(0,h.nFE,ffe);
	mesh_file.free(ffe);
	
	CRSMesh::Mesh myMesh;
	boundary_builder.init(h.nBoundaries,mesh_file.boundaries());
	
	if(partition_and_block) {
		MeshIO::Partitioner partitioner;
		partitioner.init(&builder);
		partitioner.partition(256,&partitionRenameFilter);
		builder.filter(partitionRenameFilter);
		
		//allocate  gpu memory to hold local->global mapping for ID lookup
		for(unsigned int i = 0; i < MeshIO::TYPE_SIZE; i++) {
			DEVICEALLOC(&dummyRenamed_to_global_ids[i], sizeof(MeshIO::id_t) * partitionRenameFilter.nElems[i]);
			cudaMemcpy(dummyRenamed_to_global_ids[i],partitionRenameFilter.new_to_old[i],sizeof(MeshIO::id_t) * partitionRenameFilter.nElems[i],cudaMemcpyHostToDevice);
			check_launch("memcpy");
		}
	} else {
		for(unsigned int i = 0; i < MeshIO::TYPE_SIZE; i++) {
			dummyRenamed_to_global_ids[i] = NULL;
		}
	}
	
	builder.validateFullMesh();
	
	myMesh.initFromFacetEdgeBuilder(&builder);
	Mesh_init(&myMesh, &hMesh, &dummydMesh);
	Mesh cpu_mesh;
	Mesh_buildCPUMesh(&myMesh,&cpu_mesh);
	liszt_initialize_globals_gpu();	
	coloring_controller.init(cpu_mesh, COLORING_BLOCK_SIZE);
	liszt_run();
	teardown_cuda();
#if defined(PROFILE_COUNTING) || defined(PROFILE_TIMING)
	fclose(profiling_file);
#endif
	return 0;
}
#endif /* GPURUNTIME_H_ */
