#ifndef _GPU_GLOBAL_H
#define _GPU_GLOBAL_H

#include "Gpu/Profiling.h"

#include <iostream>
#include <thrust/device_ptr.h>
#include <thrust/fill.h>
#include <thrust/reduce.h>
#include "Gpu/ThrustReductions.h"

#define GPU_BOTH __device__ __host__
// enum LisztOpType {
//     LISZT_MAX_TYPE,
//     LISZT_MIN_TYPE,
//     LISZT_PROD_TYPE,
//     LISZT_SUM_TYPE,
//     LISZT_LAND_TYPE,
//     LISZT_LOR_TYPE,
//     LISZT_LXOR_TYPE,
//     LISZT_BAND_TYPE,
//     LISZT_BOR_TYPE,
//     LISZT_BXOR_TYPE,
//     LISZT_ASSIGN_TYPE,
//     LISZT_READ_TYPE
// };

template<typename T, bool enabled>
class Global;


template<typename T>
class Global<T,true> {
public:
	bool owner;

	GPU_BOTH Global()
	: state(LISZT_READ_TYPE), alloc_size(0), dirty_size(0), data(NULL), owner(true) {}
	
	GPU_BOTH Global(const Global & other)
	: read_value(other.read_value), state(other.state), alloc_size(other.alloc_size), dirty_size(other.dirty_size), data(other.data), owner(false) {}
	
	GPU_BOTH Global(const T & t) 
	: read_value(t), state(LISZT_READ_TYPE), alloc_size(0), dirty_size(0), data(NULL), owner(true) {}
	
	void ensureState(LisztOpType typ) {
		//printf("=================================== ENSURE STATE: %d to %d. dirty: %d, allocd: %d\n", state, typ, dirty_size, alloc_size);
		stateTransition(state,typ);
	}
	
	void ensureSize(size_t n_work_units) {
		if (state == LISZT_ASSIGN_TYPE) {
			n_work_units = 1;
		}
		if(alloc_size < n_work_units) {
			realloc(n_work_units);
		}
		if(dirty_size < n_work_units) {
			setIdentity(dirty_size, n_work_units - dirty_size);
			dirty_size = n_work_units;
		}
	}
	
	~Global() {
		if(data != NULL && owner) {
			cudaFree(data);
			data = NULL;
		}
	}
	__host__ T & read_cpu_entry() {
		return read_value;		
	}
	__host__ T & write_cpu_entry() {
		return read_value;		
	}
	__device__ T & read_gpu_entry() {
		GRDM_INCREMENT_BYTES_ACCESSED(T)
		return read_value;
	}
	__device__ T & write_gpu_entry() {
#ifdef __CUDA_ARCH__
		//TODO(zach): we should remove this if statement and just ask for the read_value for assignments
		GRDM_INCREMENT_BYTES_ACCESSED(T)
		if(state == LISZT_ASSIGN_TYPE)
			return data[0];
		else {
			//printf("writing to %d of %llx\n",BLOCKSIZE * blockIdx.x + threadIdx.x, data);
			return data[BLOCKSIZE * blockIdx.x + threadIdx.x];
		}
#else
		return data[0];
#endif
	}
	
	
private:
	void realloc(size_t n_work_units) {
		T * new_data;
		DEVICEALLOC((void**)&new_data,n_work_units * sizeof(T));
		if(data != NULL) {
			cudaMemcpy(new_data,data, alloc_size * sizeof(T), cudaMemcpyDeviceToDevice);
			cudaFree(data);
		}
		data = new_data;
		alloc_size = n_work_units;
	}
	void setIdentity(size_t offset, size_t size) {
		T identity = getIdentity();
		thrust::device_ptr<T> ptr(data);
		thrust::fill(ptr + offset, ptr + offset + size, identity);
	}
	void stateTransition(LisztOpType old_state, LisztOpType new_state) {
		if (new_state != LISZT_READ_TYPE)
			state = new_state;
		if(new_state != old_state && dirty_size > 0 && old_state != LISZT_READ_TYPE) {
			reduce(old_state,dirty_size);
		}
		dirty_size = 0;
	}
	T getIdentity() const {
		#define GET_IDENTITY(op) case op##_TYPE: return op::identity<T>();
		switch(state) {
			LISZT_REDUCTION_OPS(GET_IDENTITY)
		}
		
		#undef GET_IDENTITY
	}
	void reduce(LisztOpType reduction, size_t nelems) {
		::thrust::device_ptr<T> ptr(data);
		switch(reduction) {
		case LISZT_ASSIGN_TYPE:
			cudaMemcpy(&read_value, data, 1*sizeof(T), cudaMemcpyDeviceToHost);
			break;
		case LISZT_SUM_TYPE:
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::sum<T>());
			break;
		case LISZT_PROD_TYPE:
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::product<T>());
			break;
		case LISZT_MAX_TYPE:
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::max<T>());
			break;
		case LISZT_MIN_TYPE: {
			/*printf("min reduction: %llx!!!!\n",data);
			T * hdata = new T[nelems];
			
			cudaMemcpy(hdata,data, sizeof(T) * nelems, cudaMemcpyDeviceToHost);
			check_cuda_error("copy of data");
			for(int i = 0; i < nelems; i++) {
				std::cout << i << " " << hdata[i] << std::endl;
			}
			delete [] hdata;*/
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::min<T>());		
			break; }
		case LISZT_LAND_TYPE:
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::logical_and<T>());
			break;
		case LISZT_LOR_TYPE:
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::logical_or<T>());
			break;
		case LISZT_LXOR_TYPE:
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::logical_xor<T>());
			break;
		case LISZT_BAND_TYPE:
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::binary_and<T>()); 
			break;
		case LISZT_BOR_TYPE:
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::binary_or<T>()); 
			break;
		case LISZT_BXOR_TYPE:
			read_value = ::thrust::reduce(ptr, ptr+nelems, read_value, reductions::binary_xor<T>()); 
			break;
		case LISZT_READ_TYPE:
			std::cerr << "Liszt read type not expected!" << std::endl;
			break;
		}
	}

private:
	T read_value;
	LisztOpType state;
	size_t alloc_size;
	size_t dirty_size; //(dirty_size <= alloc_size) how many entries in data are dirty (in non-read state = this->state). This will be 0 if the field is actually clean
	T * data; //gpu_memory_pointer
};

template<typename T>
class Global<T,false> {
public:
	GPU_BOTH Global() {}
	GPU_BOTH Global(const T & t) : value(t) {}
	GPU_BOTH void ensureState(LisztOpType typ) {}
	GPU_BOTH void ensureSize(size_t n_work_units) {}
	__device__ T & read_gpu_entry() { return value; }
	__host__ T & read_cpu_entry() { return value; }
	__device__ T & write_gpu_entry() { return value; }
	__host__ T & write_cpu_entry() { return value; }
private:
	T value;
};

#undef GPU_BOTH

#endif
