//
//  ptx_intrinsics
//
//  Created by Niels Joubert on 2011-02-09.
//
//  Built for PTX ISA 2.2 as described in:
//  http://developer.download.nvidia.com/compute/cuda/3_2_prod/toolkit/docs/ptx_isa_2.2.pdf


#ifndef PTX_INTRINSICS_H_
#define PTX_INTRINSICS_H_

/*
 *
 *  READ-ONLY REGISTERS
 *  from CUDA PTX ISA 2.2 document, ch.9 page 152
 */

#define PTX_GENERATE_REGISTER_READ_u32_FUNC(name) \
	__device__ inline unsigned int __##name() { \
		unsigned temp##name; \
		asm volatile("mov.u32 %[index], %" #name ";" : [index] "=r" (temp##name)); \
		return temp##name; \
	}


#define PTX_SPECIAL_REGISTERS_u32(_) \
	_(laneid) \
	_(warpid) \
	_(nwarpid) \
	_(smid) \
	_(nsmid) \
	_(gridid) \
	_(lanemask_eq) \
	_(lanemask_le) \
	_(lanemask_lt) \
	_(lanemask_ge) \
	_(lanemask_gt) \
	_(clock) \
	_(pm0) \
	_(pm1) \
	_(pm2) \
	_(pm3)
	//note that there is also a clock64() function predefined by CUDA

PTX_SPECIAL_REGISTERS_u32(PTX_GENERATE_REGISTER_READ_u32_FUNC)


/*
 *
 *  PARALLEL SYNCHRONIZATION & COMMUNICATION
 *  from CUDA PTX ISA 2.2 document, ch.8.7.8 page 136
 */

#define PTX_BAR_SYNC(bar,thct) asm volatile("bar.sync " #bar ", %[thrcnt];" : : [thrcnt] "r" (thct) : "memory"); break

//  warpCount      = number of warps to barrier for [1-48]
//  threadsPerWarp = number of threads per warp to barrier for.  [1-32]
//  barrierNumber  = which named barrier? [1-15]
__device__ inline void __bar_sync(unsigned int warpCount, unsigned int threadsPerWarp, unsigned int barrierNumber) {
	switch (barrierNumber) {
		case 1:  PTX_BAR_SYNC(1, warpCount*threadsPerWarp);
		case 2:  PTX_BAR_SYNC(2, warpCount*threadsPerWarp);
		case 3:  PTX_BAR_SYNC(3, warpCount*threadsPerWarp);
		case 4:  PTX_BAR_SYNC(4, warpCount*threadsPerWarp);
		case 5:  PTX_BAR_SYNC(5, warpCount*threadsPerWarp);
		case 6:  PTX_BAR_SYNC(6, warpCount*threadsPerWarp);
		case 7:  PTX_BAR_SYNC(7, warpCount*threadsPerWarp);
		case 8:  PTX_BAR_SYNC(8, warpCount*threadsPerWarp);
		case 9:  PTX_BAR_SYNC(9, warpCount*threadsPerWarp);
		case 10: PTX_BAR_SYNC(10,warpCount*threadsPerWarp);
		case 11: PTX_BAR_SYNC(11,warpCount*threadsPerWarp);
		case 12: PTX_BAR_SYNC(12,warpCount*threadsPerWarp);
		case 13: PTX_BAR_SYNC(13,warpCount*threadsPerWarp);
		case 14: PTX_BAR_SYNC(14,warpCount*threadsPerWarp);
		case 15: PTX_BAR_SYNC(15,warpCount*threadsPerWarp);
		default: break;
	}
}

/*
 *
 *  UNIT TESTS
 *
 */

//If you want to run a unit test for all these registers,
//define the PTX_INSTRINSICS_UNIT_TEST macro.
//this will define an int main() so compile this alone.

//#define PTX_INTRINSICS_UNIT_TEST
#ifdef PTX_INTRINSICS_UNIT_TEST
	#include <stdio.h>
	
	#define PTX_INTRINSICS_CALL_KERNEL(name, kcall) \
		do { \
			kcall; \
			cudaThreadSynchronize(); \
			cudaError_t err = cudaGetLastError(); \
			if(err != cudaSuccess) { \
				fprintf(stderr,"%s:%d: %s\n",__FILE__,__LINE__,cudaGetErrorString(err)); \
				exit(err); \
			} \
		} while(0);

	#define PTX_INTRINSICS_CALL_SPECIAL_REGISTER_TEST_KERNEL(name) \
		PTX_INTRINSICS_CALL_KERNEL(name, (__ptx_unittest_##name<<<1,1>>>()))

	#define PTX_INTRINSICS_GEN_SPECIAL_REGISTER_TEST_KERNEL(name) \
			__global__ void __ptx_unittest_##name() { \
			int temp_##name = __##name(); \
			printf("Testing __" #name "(): %d.\n", temp_##name); \
		}
	PTX_SPECIAL_REGISTERS_u32(PTX_INTRINSICS_GEN_SPECIAL_REGISTER_TEST_KERNEL)
	#undef PTX_INTRINSICS_GEN_SPECIAL_REGISTER_TEST_KERNEL

	int main() {	
		PTX_SPECIAL_REGISTERS_u32(PTX_INTRINSICS_CALL_SPECIAL_REGISTER_TEST_KERNEL)
		
	}
	
	#undef PTX_INTRINSICS_CALL_SPECIAL_REGISTER_TEST_KERNEL
#endif // PTX_INTRINSICS_

/*
 *
 *  CLEAN UP
 *
 */
#undef PTX_SPECIAL_REGISTERS
#undef PTX_GENERATE_REGISTER_READ_FUNC


#endif /* PTX_INTRINSICS_H_ */ 