
template<typename LM>
inline Mesh::Mesh(LM & m) {
    init(m);
}

//allocate memory aligned to a multiple of 'alignment'
//alignment must be a power of two
//this function will use malloc to allocate more data than it needs and return an aligned address
//to free this data you must call aligned_free

static void * aligned_malloc(size_t size, size_t alignment) {
    size_t header_s = sizeof(void*) - 1 + alignment;
    unsigned char * mem = (unsigned char*) malloc(header_s+size);
    unsigned char * start = (unsigned char*) mem + sizeof(void*) - 1;
    arithmetic_pointer_type mask = (alignment - 1);
    start = (unsigned char *) (~mask & (arithmetic_pointer_type)start) + alignment;

    void * header = start - sizeof(void*);
    assert(header >= mem);
    assert(start + size < mem + header_s + size);

    *(void**)header = mem;
    return start;
}
static void aligned_free(void * data) {
    void * mem = *(void**) ( (unsigned char*) data - sizeof(void*));
    free(mem);
}


template<typename LM>
inline void Mesh::init(LM & m) {
    typedef typename LM::vertex_t vertex_o;
    typedef typename LM::edge_t edge_o;
    typedef typename LM::face_t face_o;
    typedef typename LM::cell_t cell_o;
    typedef typename LM::face_it face_it;
    typedef typename LM::edge_it edge_it;

    unsigned int nV = m.vertices().size();
    unsigned int nE = m.edges().size();
    unsigned int nF = m.faces().size();
    unsigned int nC = m.cells().size();

    data.v_verts.insert(data.v_verts.begin(),nV,VertexData(0,NULL));
    data.v_edges.insert(data.v_edges.begin(),nE,EdgeData(0,NULL));
    data.v_faces.insert(data.v_faces.begin(),nF,FaceData(0,NULL));
    data.v_cells.insert(data.v_cells.begin(),nC,CellData(0,NULL));

    for(unsigned int i = 0; i < nV; i++) {
        data.v_verts[i].setID(i);
    }
    for(unsigned int i = 0; i < nE; i++) {
        data.v_edges[i].setID(i);
    }
    for(unsigned int i = 0; i < nF; i++) {
        data.v_faces[i].setID(i);
    }
    for(unsigned int i = 0; i < nC; i++) {
        data.v_cells[i].setID(i);
    }
    data.n_ifes = 0;
    FORALL_SET_SER_TM(f_f,m.faces())
        data.n_ifes += f_f.edges().size();
    ENDSET
    data.n_ifes <<= 1;
    data.ifes = (FacetEdgeData*) aligned_malloc(data.n_ifes * sizeof(FacetEdgeData),sizeof(FacetEdgeData) << 1);
    cout << "data.ifes = " << data.ifes << " " << &data.ifes[0] << " " << data.ifes[1].opposite() << endl;
    cout << "sizeof(FacetEdgeData) = " << sizeof(FacetEdgeData) << endl;
    unsigned int next_ife = 0;

    vector<FacetEdgeData *> ring;
    vector< vector<FacetEdgeData *> > around_edge(nE);

    FORALL_SET_SER_TM(f_f,m.faces())
        cell_o c0_f = f_f.outside();
        cell_o c1_f = f_f.inside();
        ring.clear();

        assert(f_f.ID() < data.v_faces.size());
        FaceData & f = data.v_faces[f_f.ID()];
        assert(c0_f.ID() < data.v_cells.size());
        CellData & outside = data.v_cells[c0_f.ID()];
        assert(c1_f.ID() < data.v_cells.size());
        CellData & inside = data.v_cells[c1_f.ID()];
        FORALL_SET_SER_TM(cur,f_f.edges())
            assert(cur.ID() < data.v_edges.size());
            EdgeData & e = data.v_edges[cur.ID()];
            assert(cur.head().ID() < data.v_verts.size());
            assert(cur.tail().ID() < data.v_verts.size());

            VertexData & head = data.v_verts[cur.head().ID()];
            VertexData & tail = data.v_verts[cur.tail().ID()];

            FacetEdgeData * up = new(&data.ifes[next_ife]) FacetEdgeData(next_ife,&e,&f,&head,&outside); // new FacetEdgeData(fe_count++,&head,&tail,&e,&f,&outside,&inside);
            next_ife++;
            FacetEdgeData * down = new(&data.ifes[next_ife]) FacetEdgeData(next_ife,&e,&f,&tail,&inside);
            next_ife++;
            assert(up->opposite() == down);

            //these will get changed more than once but will point to a valid FacetEdgeData
            //when its over
            head.setFacetEdgeData(up);
            tail.setFacetEdgeData(down);
            e.setFacetEdgeData(up);
            f.setFacetEdgeData(down);
            outside.setFacetEdgeData(up);
            inside.setFacetEdgeData(down);

            //add this facet-edge to the face-ring
            ring.push_back(up);
            //add this facet-edge to the edge-ring
            assert(up->edge->id < around_edge.size());
            around_edge[up->edge->id].push_back(up);
            around_edge[down->edge->id].push_back(down);
        ENDSET
        for(unsigned int i = 0; i < ring.size(); i++) {
            FacetEdgeData * a = ring[i];
            FacetEdgeData * b = ring[(i+1)%ring.size()];
            assert(a->vertex == b->opposite()->vertex);
            assert(a->cell == b->cell);
            assert(a->opposite()->cell == b->opposite()->cell);

            a->ccwAroundFace = b;
            b->opposite()->ccwAroundFace = a->opposite();


            //some sanity checks
            FacetEdge atest = FacetEdge(a);
            FacetEdge btest = FacetEdge(b);
            FacetEdge atest2 = atest.facetedgeCCWAroundFace();
            assert(atest2.head() == btest.head());
            FacetEdge btest2 = btest.flip().facetedgeCCWAroundFace();
            assert(btest2.tail() == atest.head());
        }
    ENDSET
    assert(next_ife == data.n_ifes);
    //now we need to resolve the onEdge pointers for every edge.
    vector< vector<FacetEdgeData *> >::iterator cur = around_edge.begin();
    vector< vector<FacetEdgeData *> >::iterator end = around_edge.end();
    for(; cur != end; cur++) {
        vector< FacetEdgeData * > & fes = *cur;
        if(fes.size() == 2) {
            //degenerate edge ring this means that this face
            //has the same cell on both sides
            assert(fes[0]->opposite() == fes[1]);
            fes[0]->ccwAroundEdge = fes[0];
            fes[0]->opposite()->ccwAroundEdge = fes[0]->opposite();
            assert(fes[0]->cell == fes[0]->opposite()->cell);
        } else {
            for(unsigned int i = 0; i < fes.size(); i++) {
                FacetEdgeData * a = fes[i];

                if(a->opposite()->ccwAroundEdge == NULL) {
                    VertexData * v = fes[i]->vertex;
                    CellData * c = fes[i]->opposite()->cell;
                    FacetEdgeData * b = NULL;
                    for(unsigned int j = 0; j < fes.size(); j++) {
                        if(j != i && fes[j]->vertex == v && fes[j]->cell == c) {
                            b = fes[j];
                            break;
                        }
                    }
                    assert(b != NULL);
                    b->ccwAroundEdge = a;
                    a->opposite()->ccwAroundEdge = b->opposite();
                }
            }
        }
    }
    for(unsigned int i = 0; i < data.n_ifes; i++) {
        if(!data.ifes[i].initialized()) {
            cout <<"IFE Not initialized: " << data.ifes[i] << endl;
            assert(false);
        }
    }
}
inline Mesh::~Mesh() {
    aligned_free(data.ifes);
}

template<> inline Vertex Mesh::get<Vertex>(id_type id) const {
    return Vertex(data.v_verts[id]);
}
template<> inline Edge Mesh::get<Edge>(id_type id) const {
    return Edge(data.v_edges[id]);
}
template<> inline Face Mesh::get<Face>(id_type id) const {
    return Face(data.v_faces[id]);
}
template<> inline Cell Mesh::get<Cell>(id_type id) const {
    return Cell(data.v_cells[id]);
}

template<typename MO,typename Data>
class Mesh_iterator {
public:
    typedef typename vector<Data>::const_iterator iIt;
    typedef MO value_type;
    Mesh_iterator(const iIt & start, const iIt & end) : s(start),e(end) {}
    bool hasNext() const { return s != e; }
    value_type next() { return MO(*s++); }
    value_type operator[](unsigned int n) const {
        return MO(s[n]);
    }
private:
    iIt s;
    iIt e;
};
typedef Mesh_iterator<Vertex,VertexData> Mesh_vertex_it;
typedef Mesh_iterator<Edge,EdgeData> Mesh_edge_it;
typedef Mesh_iterator<Face,FaceData> Mesh_face_it;
typedef Mesh_iterator<Cell,CellData> Mesh_cell_it;

SUBCLASS_SIZEDMESHSET(Mesh,vertex_set,Mesh_vertex_it)
SUBCLASS_SIZEDMESHSET(Mesh,edge_set,Mesh_edge_it)
SUBCLASS_SIZEDMESHSET(Mesh,face_set,Mesh_face_it)
SUBCLASS_SIZEDMESHSET(Mesh,cell_set,Mesh_cell_it)


#define MAKE_SET(etype,name,data,offset) \
    inline Mesh::etype##_set Mesh::name() const { \
        return etype##_set(Mesh_##etype##_it(data.begin()+offset,data.end()),data.size()); \
    } \
    inline Mesh::etype##_set Mesh::name(id_type i, id_type j) const { \
        return etype##_set(Mesh_##etype##_it(data.begin()+i,data.begin()+j),j-i); \
    } \
    inline Mesh::etype##_set Mesh::name(const LisztPrivate::Range<etype##_t::type> & range) const { \
        return name(range.internal_begin(),range.internal_end()); \
    }
    MAKE_SET(vertex,vertices,data.v_verts,0)
    MAKE_SET(edge,edges,data.v_edges,0)
    MAKE_SET(face,faces,data.v_faces,0)
    MAKE_SET(cell,cells,data.v_cells,1)
#undef MAKE_SET
