#ifndef INTERFERENCEBUILDER_H_
#define INTERFERENCEBUILDER_H_

#include <map>
#include <set>
#include <algorithm>
#include "InterferenceGraph/AbstractWriteContext.h"
#include "Layout/Coloring.h"
#include "Layout/CRS.h"
#include "Liszt/Time.h"
#include "Layout/Util.h"

namespace Synchronization {
	
//using namespace std;
using namespace ::Layout;

typedef int WorkGroupID;
typedef int WorkUnitID;
typedef int FieldID;
typedef int FieldEntryID;
typedef int WorkBlockID;

class InterferenceBuilder {
public:
	struct FieldData {
		std::hash_map<FieldEntryID,std::set<WorkUnitID> > fieldEntries;
	};
	struct WorkGroupData {
		//this gathers all the data as addEdge is called
		std::set<WorkUnitID> workunits;
		std::map<FieldID,FieldData> fields;
		
		//this contains the actual colored interference graph after buildInterferenceGraph
		unsigned int numNodes;
		int numColors;
		CRS coloring;
	};  

	InterferenceBuilder() : _workgroups() {}

	void init(WorkBlockID blocksize);

	void addEdge(ElemType typ, int fieldID, int fieldEntryID, int workGroupID, const AbstractWriteContext & writeContext);

	template<typename MeshSet>
	int getCRSColoring(WorkGroupID wU, const MeshSet & mesh_set, bool isTrivial,  CRS** crs);

	void print();

private:
	std::map<WorkGroupID,WorkGroupData> _workgroups;
	WorkBlockID _blocksize;

	//builds interference graph from field-write structure, builds color, caches into WorkGroupData
	template<typename MeshSet>
	void buildAndColor(WorkGroupID wG, const MeshSet & mesh_set);

	//creates CRS structure for given workgroup
	void buildColoringCRS(WorkGroupData& wdata);
	
	template<typename MeshSet>
	void getTrivialCRSColoring(const MeshSet & mesh_set, WorkGroupData& wdata, WorkUnitID wG);

};

inline void InterferenceBuilder::init(WorkBlockID blocksize) {
       _blocksize = blocksize;
}

inline void InterferenceBuilder::addEdge(
          ElemType typ,
          FieldID fieldID, 
          FieldEntryID fieldEntryID, 
          WorkGroupID workGroupID, 
          const AbstractWriteContext & writeContext) {
	
	WorkUnitID workUnitID = writeContext.getWorkUnitID();
	if (typ == ::LisztPrivate::ElemTypes::CELL && fieldEntryID == 0) {
		//ignore the outside cell for now.
	} else {
		WorkGroupData& wdata = _workgroups[workGroupID];
		wdata.numColors = -1;
		wdata.workunits.insert(workUnitID);
		FieldData& fdata = wdata.fields[fieldID];
		fdata.fieldEntries[fieldEntryID].insert(workUnitID);
	}
}

template<typename MeshSet>
inline void InterferenceBuilder::buildAndColor(WorkGroupID wG, const MeshSet & mesh_set) {
	if (_workgroups.count(wG) == 0) {
		std::cerr << "Attempted to build interference graph for nonexistant workgroup." << std::endl;
		exit(1); //failing fast == easy bug identification
	}
	printf("Coloring on workgroup %d...", wG);
	double beginningTime = processor_time();

//#define PRINT_COLORING_TIMING
#ifdef PRINT_COLORING_TIMING
	printf("\n");
	double starttime = processor_time();
	#define CTIMING_EVENT(words) printf("   " words ": %fs.\n", processor_time() - starttime); starttime = processor_time()
#else
	#define CTIMING_EVENT(words)
#endif
	WorkGroupData& wdata = _workgroups[wG];
	
	typedef typename MeshSet::iterator IterType;
	typedef typename IterType::value_type ValueType;
	for(IterType __msh_it = (mesh_set).iter(); __msh_it.hasNext();) {
		ValueType e = __msh_it.next();
		wdata.workunits.insert(e.ID());
	}
	// FORALL_SET_TM(e,mesh_set)
	// 	wdata.workunits.insert(e.ID());
	// ENDSET
	
	unsigned int nN = wdata.workunits.size(); 
	wdata.numNodes = nN;
	
	if (mesh_set.size() != wdata.numNodes) {
		std::cerr << "mesh_set not same size as numNodes" << std::endl;
		exit(1);
	}
	
	//TODO: The two graph-building stages can become one. (edges for indices vs edges for blocks)
	
	//------------------------------------------------------------------------------
	// Populates the "edges" structure: 
	//    for each element, what other elements is it connected to?
	// uses datastructure gathered from add edges
	std::hash_map<WorkUnitID, std::set<WorkUnitID> > edges;
	
	unsigned int totalEdges = 0;
	for (std::map<FieldID,FieldData>::iterator f_it = wdata.fields.begin() ; f_it != wdata.fields.end(); f_it++) {
		std::hash_map<FieldEntryID,std::set<WorkUnitID> >::iterator fe_it;
		for (fe_it = (*f_it).second.fieldEntries.begin() ; fe_it != (*f_it).second.fieldEntries.end(); fe_it++) {
			
			//connect an edge between all work units writing to the same entry
			std::set<WorkUnitID>& workUnitsAtThisFieldEntry = (*fe_it).second;
			std::set<WorkUnitID>::iterator we_it;
			for (we_it = workUnitsAtThisFieldEntry.begin() ; we_it != workUnitsAtThisFieldEntry.end(); we_it++) {
				const WorkUnitID& WU = *we_it;	
				std::set<WorkUnitID>::iterator we_it2;
				for (we_it2 = workUnitsAtThisFieldEntry.begin() ; we_it2 != workUnitsAtThisFieldEntry.end(); we_it2++) {
					const WorkUnitID& oWU = *we_it2;
					std::set<WorkUnitID> &myEdges = edges[WU];
					if (WU != oWU) {
						//we have an edge, now to attempt to add it
						std::pair<std::set<WorkUnitID>::iterator, bool> didInsert = myEdges.insert(oWU);
						if (didInsert.second)
							totalEdges++;
					}
				}
			}
			
		}
	}
	CTIMING_EVENT("finished building interference graph");

	// Build a dense list of indices for the sparse list of elements in the top-level set
	//------------------------------------------------------------------------------
	size_t *elements = new size_t[mesh_set.size()];
	std::hash_map<WorkUnitID, unsigned int> elementsToIndices;
	unsigned int countEls = 0;

	typedef typename MeshSet::iterator IterType;
	typedef typename IterType::value_type ValueType;
	for(IterType __msh_it = (mesh_set).iter(); __msh_it.hasNext();) {
		ValueType e = __msh_it.next();
		elements[countEls] = e.ID();
		elementsToIndices[e.ID()] = countEls;
		countEls++;
	}

	// FORALL_SET_TM(e,mesh_set)
	// 	elements[countEls] = e.ID();
	// 	elementsToIndices[e.ID()] = countEls;
	// 	countEls++;
	// ENDSET

//#define SAVE_COLORING
#ifdef SAVE_COLORING	
	FILE * file = FOPEN("nodes.txt","a");
	fprintf(file, "Workgroup start\n");
	for (int i = 0; i < mesh_set.size(); i++) {
		fprintf(file, "%ld\n", elements[i]);
	}
	fclose(file);
#endif
	
	CTIMING_EVENT("finished building sparse elements -> dense indices map");

	// Build a interference graph between densely-numbered blocks from sparsely-numbered elements
	//------------------------------------------------------------------------------
	int nB = ceil(mesh_set.size() / (double) _blocksize);
	std::set<WorkBlockID> *dense_block_interference_edges = new std::set<WorkBlockID>[nB];
	unsigned int index = 0;
	typedef typename MeshSet::iterator IterType;
	typedef typename IterType::value_type ValueType;
	for(IterType __msh_it = (mesh_set).iter(); __msh_it.hasNext();) {
		ValueType e = __msh_it.next();
		WorkBlockID iblock = index / _blocksize;
		const std::set<WorkUnitID> connected_to = edges[e.ID()];
		for (std::set<WorkUnitID>::iterator e_it = connected_to.begin(); e_it != connected_to.end(); e_it++) {
			unsigned int index_j = elementsToIndices[*e_it];
			
			WorkBlockID jblock = index_j / _blocksize;
			dense_block_interference_edges[iblock].insert(jblock);
		}
		index++; 
	}
	// FORALL_SET_TM(e,mesh_set)
	// 	WorkBlockID iblock = index / _blocksize;
	// 	const std::set<WorkUnitID> connected_to = edges[e.ID()];
	// 	for (std::set<WorkUnitID>::iterator e_it = connected_to.begin(); e_it != connected_to.end(); e_it++) {
	// 		unsigned int index_j = elementsToIndices[*e_it];
	// 		
	// 		WorkBlockID jblock = index_j / _blocksize;
	// 		dense_block_interference_edges[iblock].insert(jblock);
	// 	}
	// 	index++; 
	// ENDSET
	CTIMING_EVENT("finished building block->block map");

	// Build a block to block interference graph CRS structure for coloring.
	//------------------------------------------------------------------------------
	size_t *edge_idx = new size_t[nB+1];
	size_t *edge_vals = new size_t[totalEdges]; //This will OVERESTIMATE the number of edges.
	size_t currEdgeV = 0;
	for (WorkBlockID i = 0; i < nB; i++) { //dense numbering
		edge_idx[i] = currEdgeV;
		const std::set<WorkUnitID> connected_to = dense_block_interference_edges[i];
		for (std::set<WorkUnitID>::iterator e_it = connected_to.begin(); e_it != connected_to.end(); e_it++) {
			edge_vals[currEdgeV] = *e_it;
			currEdgeV++;
		}
	}
	edge_idx[nB] = currEdgeV;
	CTIMING_EVENT("finished building block->block graph edge structure");

	// Run coloring
	//------------------------------------------------------------------------------	
	size_t* colors = new size_t[nB];
	Coloring colorer;
	wdata.numColors = colorer.run(nB, edge_idx, edge_vals, colors);
	CTIMING_EVENT("finished coloring");

	delete [] dense_block_interference_edges;
	
	// Build a color -> list of elements in ordered blocks CRS structure for launching threads
	//------------------------------------------------------------------------------	
	idx_type* color_idx = new idx_type[wdata.numColors+1];
	id_type* color_values = new id_type[nN];
	
//#define OUTPUT_COLORING
#define els_in_mesh 875
#ifdef OUTPUT_COLORING
      id_type* color_for_unit = new id_type[els_in_mesh];
      for (int i = 0; i < els_in_mesh; i++) {
              color_for_unit[i] = 1337;
      }
      printf("Number of blocks: %d, nr of els: %d, nr of this type of els: %d\n", nB, nN, els_in_mesh);
#endif

	int currentIdx = 0;
	for (int color = 0; color < wdata.numColors; color++) {
		color_idx[color] = currentIdx;
		for (int blocki = 0; blocki < nB; blocki++) {
			if (colors[blocki] == color) {
				unsigned int start_eli = blocki * _blocksize;
				unsigned int thisblocksize = (start_eli + _blocksize >= mesh_set.size()) ? (mesh_set.size() - start_eli) : _blocksize;
				for (int i = 0; i < thisblocksize; i++) {
					color_values[currentIdx] = elements[start_eli + i];
					currentIdx++;
#ifdef OUTPUT_COLORING
					color_for_unit[start_eli + i] = color;
#endif
				}
			}
		}
	}
	color_idx[wdata.numColors] = currentIdx;
	wdata.coloring.row_idx = color_idx;
	CTIMING_EVENT("finished building CRS structure");

#ifdef OUTPUT_COLORING        
      writeField<int>(ElemTypeToIO(::LisztPrivate::ElemTypes::FACE),color_for_unit,sizeof(id_type),els_in_mesh,"colors");
      delete [] color_for_unit;
#endif

#ifdef SAVE_COLORING
	file = FOPEN("colors.txt","a");
	fprintf(file, "Workgroup start, numColors=%d\n", wdata.numColors);
	for (int i = 0; i <= wdata.numColors; i++) {
		fprintf(file, "color_idx: %d %d\n", i, color_idx[i]);
	}
	for (int i = 0; i < wdata.numColors; i++) {
		for (int j = color_idx[i]; j < color_idx[i+1]; j++) {
			fprintf(file, "color_values: %d %d\n", i, color_values[j]);
		}
	}
	fclose(file);
#endif

	// Put this CRS structure at the right place
	//------------------------------------------------------------------------------	
#ifdef GPURUNTIME	
	DEVICEALLOC(&wdata.coloring.values,sizeof(id_type) * wdata.numNodes);
	cudaMemcpy(wdata.coloring.values,color_values,sizeof(id_type) * wdata.numNodes, cudaMemcpyHostToDevice);
	delete [] color_values;
#else
	wdata.coloring.values = color_values;
#endif
	CTIMING_EVENT("finished copying CRS around");

	printf(" found %d colors in %fs.\n", wdata.numColors, processor_time() - beginningTime);

	// Clean up the mess
	//------------------------------------------------------------------------------	
	delete [] elements;
	delete [] edge_idx;
	delete [] edge_vals;	
	delete [] colors;
}

template<typename MeshSet>
inline void InterferenceBuilder::getTrivialCRSColoring(const MeshSet & mesh_set, WorkGroupData& wdata, WorkUnitID wg) {
	int numNodes = mesh_set.size();
	if (numNodes == 0) {
		wdata.numColors = 0;
		std::cout << "Coloring on workgroup " << wg << " trivially assigned 0 colors." << std::endl;	
		return;
	}
	
	idx_type* color_idx = new idx_type[2];
	id_type* color_values = new id_type[numNodes];

	int currentIdx = 0;
	color_idx[0] = currentIdx;

	typedef typename MeshSet::iterator IterType;
	typedef typename IterType::value_type ValueType;
	for(IterType __msh_it = (mesh_set).iter(); __msh_it.hasNext();) {
		ValueType e = __msh_it.next();
		color_values[currentIdx] = e.ID();
		currentIdx++;
	}
	color_idx[1] = currentIdx;
	
	wdata.coloring.row_idx = color_idx;
	
#ifdef GPURUNTIME
	DEVICEALLOC(&wdata.coloring.values,sizeof(id_type) * numNodes);
	cudaMemcpy(wdata.coloring.values,color_values,sizeof(id_type) * numNodes, cudaMemcpyHostToDevice);
	delete [] color_values;
#else
	wdata.coloring.values = color_values;
#endif
	std::cout << "Coloring on workgroup " << wg << " trivially assigned 1 color." << std::endl;	
	
	wdata.numColors = 1;
	return;
}

template<typename MeshSet>
inline int InterferenceBuilder::getCRSColoring(WorkGroupID wG, const MeshSet & mesh_set, bool isTrivial, CRS** crs) {
	if (_workgroups.count(wG) <= 0) {
		WorkGroupData& wdata = _workgroups[wG];
		wdata.numColors = -1;
	}	
	WorkGroupData& wdata = _workgroups[wG];
	if (wdata.numColors < 0) {	
		if (isTrivial) {
			getTrivialCRSColoring(mesh_set,wdata,wG);
		} else {
			buildAndColor(wG,mesh_set);
		}
	}
	(*crs) = &wdata.coloring;
	return wdata.numColors;
}

inline void InterferenceBuilder::print() {
	std::map<WorkUnitID,WorkGroupData>::iterator wg_it;
	for (wg_it = _workgroups.begin() ; wg_it != _workgroups.end(); wg_it++ ) {
		std::cout << "Workgroup " << (*wg_it).first << ":" << std::endl;
		WorkGroupData& wdata = _workgroups[(*wg_it).first];
		
		std::set<WorkUnitID>::iterator w_it;
		std::cout << "  Work Units:";
		for (w_it = wdata.workunits.begin() ; w_it != wdata.workunits.end(); w_it++)
			std::cout << " " << *w_it;
		std::cout << std::endl;
		
		std::cout << " Fields touched:" << std::endl;
		std::map<FieldID,FieldData>::iterator f_it;
		for (f_it = wdata.fields.begin() ; f_it != wdata.fields.end(); f_it++) {
			std::cout << "    Field " << (*f_it).first << std::endl;
			
			FieldData& fdata = (*f_it).second;
			std::hash_map<FieldEntryID,std::set<WorkUnitID> >::iterator fe_it;
			for (fe_it = fdata.fieldEntries.begin() ; fe_it != fdata.fieldEntries.end(); fe_it++) {
				std::cout << "      Entry " << (*fe_it).first << " written to by workunits";
				std::set<WorkUnitID>::iterator we_it;
				for (we_it = (*fe_it).second.begin() ; we_it != (*fe_it).second.end(); we_it++) {
					std::cout << " " << *we_it;
				}
				std::cout << std::endl;
				
			}
			
		}
		
	}

}

} /* namespace Synchronization */

#endif /* INTERFERENCEBUILDER_H_ */