#ifndef _LAYOUT_PARTITIONER_H
#define _LAYOUT_PARTITIONER_H
#include <sstream>
extern "C" {
        #include <metis.h>
}
#include <queue>
#include <algorithm>
#include "MeshIO/FacetEdgeBuilder.h"
#include "Layout/Util.h"

namespace MeshIO {

struct ResultFilter {
        ResultFilter() {
                for(int i = 0; i < TYPE_SIZE; i++) {
                        new_to_old[i] = NULL;
                        old_to_new[i] = NULL;
                }
        }
        void init(lsize_t * nE) {
                std::copy(nE,nE + TYPE_SIZE, nElems);
        }
        //takes ownership of o_to_n
        void allocForward() {
                for(size_t i = 0; i < TYPE_SIZE; i++) {
                        assert(!old_to_new[i]);
                        old_to_new[i] = new id_t[nElems[i]];
                }
        }
        void allocReverse() {
                for(size_t i = 0; i < TYPE_SIZE; i++) {
                        assert(!new_to_old[i]);
                        new_to_old[i] = new id_t[nElems[i]];
                }
        }

        void initReverseFromForward() {
                if(!new_to_old[0])
                                allocReverse();
                for(size_t i = 0; i < TYPE_SIZE; i++) {
                        for(id_t j = 0; j < nElems[i]; j++) {
                                new_to_old[i][old_to_new[i][j]] = j;
                        }
                }
        }
        void initForwardFromReverse() {
                if(!old_to_new[0])
                        allocForward();
                for(size_t i = 0; i < TYPE_SIZE; i++) {
                        for(id_t j = 0; j < nElems[i]; j++) {
                                old_to_new[i][new_to_old[i][j]] = j;
                        }
                }
        }

        id_t operator()(IOElemType typ, id_t id) {
                return old_to_new[typ][id];
        }

        id_t reverse(IOElemType typ,id_t id) {
                return new_to_old[typ][id];
        }

        void clearForward() { for(id_t i = 0; i < TYPE_SIZE; i++) delete [] old_to_new[i]; }
        void clearReverse() { for(id_t i = 0; i < TYPE_SIZE; i++) delete [] new_to_old[i]; }

        id_t * old_to_new[TYPE_SIZE];
        id_t * new_to_old[TYPE_SIZE];
        lsize_t nElems[TYPE_SIZE];
        //write partition out to disk as a field for debug info
        void debug_write(const char * base) {
                for(int i = 0; i < TYPE_SIZE; i++) {
                        std::ostringstream name;
                        name << base << i;
                        writeField<int>(static_cast<IOElemType>(i),old_to_new[i],sizeof(id_t),nElems[i],name.str().c_str());
                }
        }

        ~ResultFilter() {
                for(size_t i = 0; i < TYPE_SIZE; i++) {
                        if(old_to_new[i])
                                delete [] old_to_new[i];
                        if(new_to_old[i])
                                delete [] new_to_old[i];
                }
        }
};

static const id_t UNASSIGNED_FLAG = 1u << 31u;

struct BlockForPartition {
        typedef FacetEdge::HalfFacet HalfFacet;
        void init(FacetEdgeBuilder * m, size_t np, size_t * * ps, size_t vid, float mbw, ResultFilter * f) {
                mesh = m;
                num_parts = np;
                partition_sizes = ps;
                filter = f;
                max_block_weight = mbw;
                visit_id = vid;
        }
        void run() {
                for(int i = 0; i < TYPE_SIZE; i++) {
                        nWeight[i] = mesh->nC() / (4.f * mesh->nElems[i]);
                        next_unvisited[i] = 0;
                        next_id[i] = (i == CELL_T) ? 1 : 0;

                        //all elements are initialially marked unassigned
                        for(size_t j = 0; j < mesh->nElems[i]; j++) {
                                if(i == CELL_T && j == 0)
                                        continue; //do not reassign the 0 cell, we mark it as assigned to it is never given a new value
                                filter->old_to_new[i][j] |= UNASSIGNED_FLAG;
                        }
                }
                block_weight = 0.f;
                for(cur_part = 0; cur_part < num_parts; cur_part++) {
                        id_t elem_id;
                        IOElemType elem_type;
                        while(nextElemInPartition(cur_part,&elem_type,&elem_id)) {
                                visit(mesh->canonicalFacet(elem_type,elem_id));
                        }
                        for(size_t i = 0; i < TYPE_SIZE; i++)
                                assert(next_id[i] == partition_sizes[i][cur_part + 1]);
                }

                for(size_t i = 0; i < TYPE_SIZE; i++)
                        assert(next_id[i] == mesh->nElems[i]);
                //the old_to_new mapping is now correct, but the new_to_old mapping is now out of date
                //we fix that by updaing the latter from the former
                filter->initReverseFromForward();
        }

        void visit(HalfFacet * start) {
                visit_id++;
                enqueue(start);
                HalfFacet * facet;
                while(nextUnsearched(&facet)) {
                        if(block_weight > max_block_weight) {
                                block_weight = 0.f;
                                while(!unsearched.empty())
                                        unsearched.pop();
                        }
                        assignIfNew(CELL_T,facet->cell());
                        assignIfNew(EDGE_T,facet->edge());
                        assignIfNew(FACE_T,facet->face());
                        assignIfNew(VERTEX_T,facet->vert());

                        enqueue(facet->flip());
                        if(inCurrentPartition(CELL_T,facet->cell()))
                                enqueue(facet->ccwAroundEdge());
                        if(inCurrentPartition(VERTEX_T,facet->vert()))
                                enqueue(facet->ccwAroundFace());
                }
        }
        void enqueue(HalfFacet * facet) {
                if(!facet->visit(visit_id)) {
                        unsearched.push(facet);
                }
        }
        bool nextUnsearched(HalfFacet ** hf) {
                if(!unsearched.empty()) {
                        *hf = unsearched.front();
                        unsearched.pop();
                        return true;
                } else return false;
        }
        bool inCurrentPartition(IOElemType type, id_t id) {
                //old_to_new holds either the new id post partitioning, or the new id after partitioning and blocking
                //since either is still in the same partition this method always returns wheter the id is in the current
                //partition, even if it has been assigend
                id_t new_id = filter->old_to_new[type][id] & ~UNASSIGNED_FLAG;
                return (partition_sizes[type][cur_part] <= new_id &&
                        new_id < partition_sizes[type][cur_part + 1]);
        }
        bool assignIfNew(IOElemType type, id_t id) {
                if(!assigned(type,id) && inCurrentPartition(type,id)) {
                        assert(next_id[type] < partition_sizes[type][cur_part + 1]);
                        filter->old_to_new[type][id] = next_id[type]++;
                        block_weight += nWeight[type];
                        return true;
                }
                return false;
        }
        bool nextElemInPartition(size_t part, IOElemType * type, id_t * elem_id) {
                bool advanced = true;
                while(advanced) {
                        advanced = false; //becomes true as long as one of the 4 lists is still within the partition
                        for(size_t i = 0; i < TYPE_SIZE; i++) {
                                if(next_unvisited[i] < partition_sizes[i][part + 1]) {
                                        size_t old_id = filter->new_to_old[i][next_unvisited[i]];
                                        if(!assigned(static_cast<IOElemType>(i),old_id)){
                                                *elem_id = old_id;
                                                *type = static_cast<IOElemType>(i);
                                                next_unvisited[i]++;
                                                return true;
                                        } else {
                                                next_unvisited[i]++;
                                                advanced = true;
                                        }
                                }
                        }
                }
                return false; //no more unassigned elements in this partition
        }
        bool assigned(IOElemType type, id_t id) {
                return ((filter->old_to_new[type][id] & UNASSIGNED_FLAG) == 0);
        }
        //these HF are in the current grouping but have not been searched yet.
        std::queue<HalfFacet *> unsearched;
        FacetEdgeBuilder * mesh;
        size_t num_parts;
        //partition_sizes[VERTEX_T][0] - partition_sizes[VERTEX_T][1] is the range of elements in f->new_to_old[VERTEX_T] that are in the first partition
        size_t ** partition_sizes;

        size_t next_unvisited[TYPE_SIZE];
        size_t next_id[TYPE_SIZE];

        ResultFilter * filter;
        size_t cur_part;
        size_t visit_id;

        float block_weight;
        float max_block_weight;
        float nWeight[TYPE_SIZE];
};
struct Partitioner {

        typedef FacetEdge::HalfFacet HalfFacet;

        typedef MeshIO::id_t id_t;

        void init(FacetEdgeBuilder * m) {
                mesh = m;
                visit_id = 1;
        }

        struct vertices_of_cell {
                vertices_of_cell(FacetEdgeBuilder * m, id_t c, id_t vid) {
                        mesh = m;
                        visit_id = vid;
                        cell = c;
                        push(mesh->canonicalFacet(CELL_T,c));
                }
                bool next(id_t * v) {
                        HalfFacet * facet = pop();
                        if(!facet)
                                return false;
                        HalfFacet * cur = facet;
                        do {
                                cur = cur->ccwAroundFace();
                                push(cur);
                                cur = cur->ccwAroundEdge()->flip();
                        } while(facet != cur);
                        *v = facet->vert();
                        return true;
                }
        private:
                void push(HalfFacet * hf) {
                        if(!mesh->canonicalFacet(VERTEX_T,hf->vert())->visit(visit_id))
                                unscanned.push_back(hf);
                }
                HalfFacet * pop() {
                        if(unscanned.size() == 0)
                                return NULL;
                        else {
                                HalfFacet * r = unscanned.back();
                                unscanned.pop_back();
                                return r;
                        }
                }
                id_t visit_id;
                id_t cell;
                FacetEdgeBuilder * mesh;
                std::vector<HalfFacet *> unscanned;
        };

        void partition(int nparts, ResultFilter * result) {
                int wgtflag = 1;
                int numflag = 0;
                int ncon = 2;

#define GET_CELL(c) (mesh->nV() + (c))
#define GET_VERTEX(v) ((v))

                //vertices first then cells
                int n = mesh->nV() + mesh->nC();
                idxtype * xadj = new idxtype[n + 1];
                std::fill_n(xadj,n + 1,0);

                //count the adjancencies, then scan to get xadj, fill in vweight


                //cell-to-cell, then cell-vertex
                for(id_t f = 0; f < mesh->nF(); f++) {
                        HalfFacet * hf = mesh->canonicalFacet(FACE_T,f);
                        if(hf->cell() != 0 && hf->flip()->cell() != 0) {
                                xadj[GET_CELL(hf->cell()) + 1]++;
                                xadj[GET_CELL(hf->flip()->cell()) + 1]++;
                        }
                }
                //cell-to-vertex, and vertex-to-cell
                for(id_t c = 1; c < mesh->nC(); c++) { //skip zero cell
                        vertices_of_cell voc(mesh,c,visit_id++);
                        id_t v;
                        while(voc.next(&v)) {
                                xadj[GET_CELL(c) + 1]++;
                                xadj[GET_VERTEX(v) + 1]++;
                        }
                }
                idxtype * vwgt = new idxtype[ncon * n];

                for(int i = 0; i < n; i++) {
                        if(i < mesh->nV()) {
                                vwgt[2*i] = 0;
                                vwgt[2*i + 1] = 1;
                        } else {
                                vwgt[2*i] = xadj[i + 1];
                                vwgt[2*i+1] = 0;
                        }
                        xadj[i + 1] += xadj[i];
                }


                idxtype * adjncy = new idxtype[xadj[n]];
                idxtype * ewgt = new idxtype[xadj[n]];

                idxtype * start = new idxtype[n];
                std::copy(xadj,xadj+n,start);
                //fill in the edges

                for(id_t f = 0; f < mesh->nF(); f++) {
                        HalfFacet * hf = mesh->canonicalFacet(FACE_T,f);
                        if(hf->cell() != 0 && hf->flip()->cell() != 0) {
                                size_t idx = start[GET_CELL(hf->cell())]++;
                                adjncy[idx] = GET_CELL(hf->flip()->cell());
                                ewgt[idx] = 1;
                                idx = start[GET_CELL(hf->flip()->cell())]++;
                                adjncy[idx] = GET_CELL(hf->cell());
                                ewgt[idx] = 1;
                        }
                }

                for(id_t c = 1; c < mesh->nC(); c++) { //skip zero cell
                        vertices_of_cell voc(mesh,c,visit_id++);
                        id_t v;

                        while(voc.next(&v)) {
                                size_t idx = start[GET_CELL(c)]++;
                                adjncy[idx] = GET_VERTEX(v);
                                ewgt[idx] = 1;
                                idx = start[GET_VERTEX(v)]++;
                                adjncy[idx] = GET_CELL(c);
                                ewgt[idx] = 1;
                        }
                }
                delete [] start;

                float ubvec[] = { 1.05f, 1.05f };
                int options[] = {0,0,0,0,0};

                int edgecut;
                idxtype * part = new idxtype[n];
                METIS_mCPartGraphKway(&n,&ncon,xadj,adjncy,vwgt,ewgt,&wgtflag,&numflag,&nparts,ubvec,options,&edgecut,part);

                writeField<int>(VERTEX_T, part + GET_VERTEX(0),sizeof(id_t),mesh->nV(),"partition_vert");
                writeField<int>(CELL_T, part + GET_CELL(0),sizeof(id_t),mesh->nC(),"partition_cell");
                delete [] xadj;
                delete [] vwgt;
                delete [] adjncy;
                delete [] ewgt;

                //cell 0 must map to 0, as it has special meaning in the mesh:
                part[GET_CELL(0)] = 0;

                //now counting sort the elements to group by partition, must ensure that cell 0 still maps to cell 0
                //for now this is accomplished because it is in partition zero, and the sort is stable
                result->init(mesh->nElems);
                result->allocReverse();

                //first sort the elements by partition, this gives an initial ordering of elements (
                //since this is sorted by partiton this produces a new_to_old.  The old_to_new mapping is
                //goes from the old global ID to its position in the new order)

                size_t * partition_sizes[TYPE_SIZE];
                for(size_t i = 0; i < TYPE_SIZE; i++)
                        partition_sizes[i] = new size_t[nparts + 1];
                sort_elems<SortVC>(mesh->nV(),nparts,SortVC(part),partition_sizes[VERTEX_T],result->new_to_old[VERTEX_T]);
                sort_elems<SortVC>(mesh->nC(),nparts,SortVC(part + mesh->nV()),partition_sizes[CELL_T],result->new_to_old[CELL_T]);

                sort_elems<SortEF>(mesh->nE(),nparts,SortEF(part,mesh->edgeMap(),0),partition_sizes[EDGE_T],result->new_to_old[EDGE_T]);
                sort_elems<SortEF>(mesh->nF(),nparts,SortEF(part + mesh->nV(),mesh->faceMap(),1),partition_sizes[FACE_T],result->new_to_old[FACE_T]);


                delete [] part;

                result->initForwardFromReverse();

                result->debug_write("partitioned_id");

                BlockForPartition blocker;
                blocker.init(mesh, nparts, partition_sizes, visit_id, 32.f, result);
                blocker.run();

                result->debug_write("partitioned_and_blocked_id");

                for(size_t i = 0; i < TYPE_SIZE; i++)
                        delete [] partition_sizes[i];
        }
        struct SortVC {
                SortVC(idxtype * p) : part(p) {}
                id_t operator()(id_t i) const { return part[i]; }
                idxtype * part;
        };

        struct SortEF {
                SortEF(idxtype * p, FacetEdgeBuilder::map_t & m, int d) : part(p), map(m), dual(d) {}
                id_t operator()(id_t i) const {
                        HalfFacet * hf = map[i];
                        id_t head = hf->vertOrCell(dual);
                        id_t tail = hf->flip()->vertOrCell(dual);
                        return ( dual && head == 0 ) ? part[tail] :
                               ( dual && tail == 0 ) ? part[head] :
                               ( (head < tail) ^ (i % 2 == 0) ) ? part[head] : part[tail];
                }
                idxtype * part;
                FacetEdgeBuilder::map_t & map;
                int dual;
        };



        template<typename Key>
        void sort_elems(size_t nElems, size_t nkeys, const Key & key, size_t * counts, id_t * sorted_keys) {
                std::fill_n(counts,nkeys + 1,0);
                for(id_t i = 0; i < nElems; i++) {
                        id_t k = key(i) + 1;
                        counts[k]++;
                }
                assert(counts[0] == 0);
                for(id_t i = 1; i < nkeys + 1; i++) {
                        counts[i] += counts[i-1];
                }
                size_t * offsets = new size_t[nkeys];
                std::copy(counts, counts + nkeys, offsets);
                for(id_t i = 0; i < nElems; i++) {
                        id_t k = key(i);
                        sorted_keys[offsets[k]++] = i;
                }
                delete [] offsets;
        }
private:
        id_t visit_id;
        FacetEdgeBuilder * mesh;
};


struct Blocker {
        void block(FacetEdgeBuilder * m, float mbw, ResultFilter * f) {
                filter = f;
                mesh = m;
                max_block_weight = mbw;
                block_weight = 0.f;
                block_id = 0;
                visit_id = 1;
                facets = mesh->facets();
                filter->init(mesh->nElems);
                filter->allocForward();

                for(size_t i = 0; i < TYPE_SIZE; i++)
                        for(size_t j = 0; j < mesh->nElems[i]; j++)
                                filter->old_to_new[i][j] = UNASSIGNED_FLAG;
                for(int i = 0; i < TYPE_SIZE; i++) {
                        nWeight[i] = mesh->nC() /(4.f * mesh->nElems[i]);
                        next_id[i] = 0;
                        block_ids[i] = new id_t[mesh->nElems[i]];
                }

                next_id[CELL_T] = 1; //do not renumber the outside cell
                filter->old_to_new[CELL_T][0] = 0;

                HalfFacet * hf;
                while(nextUnvisited(&hf)) {
                        enqueue(hf);
                        block_weight = 0.f;
                        //block_id++;
                        while(!unsearched_hf.empty()) {
                                HalfFacet * hf = unsearched_hf.front();
                                unsearched_hf.pop();
                                if(block_weight > max_block_weight) {
                                        block_weight = 0.f;
                                        block_id++;
                                        freeUnsearched(); //reset the local BFS, start a new seed
                                }
                                visit(hf);
                        }
                }

                for(int i = 0; i < TYPE_SIZE; i++) {
                        for(id_t j = 0; j < mesh->nElems[i]; j++) {
                                assert(assigned(static_cast<IOElemType>(i),j));
                        }
                        std::ostringstream n;
                        n << "partition_" << i;
                        writeField<int>(static_cast<IOElemType>(i),block_ids[i],sizeof(int),mesh->nElems[i],n.str().c_str());
                        delete [] block_ids[i];
                }

                filter->initReverseFromForward();
                filter->debug_write("blocked_id");
        }
private:
        typedef FacetEdge::HalfFacet HalfFacet;
        bool assignIfNew(IOElemType typ, id_t id) {
                if(!assigned(typ,id)) {
                        filter->old_to_new[typ][id] = next_id[typ]++;
                        //std::cout << "assigned " << typ << " " << id << " " << next_id[typ] - 1 << std::endl;
                        //update weight of the block with the normalized contribution
                        //of this mesh element
                        block_weight += nWeight[typ];
                        block_ids[typ][id] = block_id;
                        return true;
                } else return false;
        }
        void visit(HalfFacet * hf) {
                assert(hf->wasVisited(visit_id));

                //enqueue neighboring hf only if we the stuff between them gets assigned to this block
                enqueue(hf->flip());

                assignIfNew(VERTEX_T,hf->vert());
                enqueue(hf->ccwAroundFace());
                assignIfNew(EDGE_T, hf->edge());
                assignIfNew(FACE_T, hf->face());
                assignIfNew(CELL_T, hf->cell());
                enqueue(hf->ccwAroundEdge());
        }

        void enqueue(HalfFacet * hf) {
                //std::cout << "enqueue " << hf->ID() << " " << hf->isOther() << " " << hf->me()->visited << " " << std::endl;
                if(!hf->visit(visit_id)) {
                        //std::cout << "enqued! " << hf->ID() << " " << hf->isOther() << " " << hf->me()->visited << " " << std::endl;
                        unsearched_hf.push(hf);
                }
        }


        void freeUnsearched() {
                //std::cout << "freeing unsearched" << std::endl;
                while(!unsearched_hf.empty()) {
                        HalfFacet * hf = unsearched_hf.front();
                        unsearched_hf.pop();
                        hf->unvisit();
                        //std::cout << "free: " << hf->ID() << " " << hf->isOther() << std::endl;
                        unvisited_hf.push(hf);
                }
        }

        bool nextUnvisited(HalfFacet ** hf) {
                //search the 'free list' of restored half facets for ones not yet visited
                while(!unvisited_hf.empty()) {
                        HalfFacet * h = unvisited_hf.front();
                        unvisited_hf.pop();
                        if(!h->wasVisited(visit_id)) {
                                *hf = h;
                                return true;
                        }
                }
                //if we didnt find any in the free list, look for an unvisited HF in the
                //rest of the mesh
                assert(unvisited_hf.empty());
                FacetEdge * facet;
                while(facets.next(&facet)) {
                        for(int i = 0; i < 2; i++)
                                if(!facet->hf[i].wasVisited(visit_id)) {
                                        *hf = &facet->hf[0];
                                        return true;
                                }
                }
                //nothing is on the free list and we have seen each HF once so we have
                //visited all half facets
                return false;
        }
        bool assigned(IOElemType typ, id_t i) {
                return filter->old_to_new[typ][i] != UNASSIGNED_FLAG;
        }
        std::queue<HalfFacet *> unsearched_hf;
        std::queue<HalfFacet *> unvisited_hf;
        ResultFilter * filter;
        float nWeight[TYPE_SIZE];
        size_t next_id[TYPE_SIZE];
        float block_weight;
        float max_block_weight;
        FacetPool::iterator_t facets;
        FacetEdgeBuilder * mesh;
        id_t visit_id;
        id_t block_id;

        id_t * block_ids[TYPE_SIZE];
};
}
#endif
