/*
 * BoundarySet.h
 *
 *  Created on: May 15, 2010
 *      Author: mbarrien
 */
#ifndef BOUNDARYSET_H_
#define BOUNDARYSET_H_
#include <vector>
#include "liszt_memory.h"
#include "Liszt/ElemTypes.h"
#include "Liszt/Util.h"
#include "ParallelLoad/Util.h"
#include "MeshSet.h"
#include <algorithm>

namespace LisztPrivate {

template <typename SetType>
class BoundarySetBase {
    enum State { SET_EMPTY, SET_SINGLE, SET_MANY };
    typedef SetType set_type;
    typedef std::vector<set_type> set_list;
    typedef std::vector<size_t> size_list;
    typedef typename set_type::iterator base_iter_type;
public:
    BoundarySetBase() : size_(0), state(SET_EMPTY)  {}

    void addSetImpl(set_type new_set) {
        if (new_set.size() == 0) {
            //nothing to add, so we do nothing
        } else if(state == SET_EMPTY) {
            set = new_set;
            size_ = set.size();
            state = SET_SINGLE;
        } else {
            if(state == SET_SINGLE) {
                assert(set.size() != 0);
                // Convert from single set to multiple sets by pushing
                // existing single set into new vector.
                set_list* ss = new set_list;
                size_list * sizes = new size_list;
                // Separating out like this and in this specific order
                // for potential use of Union class.
                ss->push_back(set);
                sizes->push_back(0);
                sizes->push_back(size_);

                sets.reset(ss);
                size_scan.reset(sizes);

                state = SET_MANY;
            }
            sets->push_back(new_set);
            size_t nsize = new_set.size();
            size_ += nsize;
            size_scan->push_back(size_);
        }
    }
    class Iter {
    public:
        typedef typename base_iter_type::value_type value_type;
        typedef typename set_type::iterator iterator;
        Iter() {}
      Iter(size_t nelems, iterator it) : cur_set(NULL), iter(it), i(nelems), size_integral(NULL) {}
      Iter(size_t nelems, const set_type * cs, iterator it, const size_t* size_integral) :
        cur_set(cs), iter(it), i(nelems), size_integral(size_integral) {}

        bool hasNext() const {
            return i != 0;
        }
        value_type next() {
            // The assertion in addSet prevents adding empty sets, so no
            // need for a while loop
            assert(iter.hasNext());
            value_type retval = iter.next();
            --i;
            if (!iter.hasNext() && hasNext()) {
                iter = (++cur_set)->iter();
            }
            return retval;
        }

      value_type operator[](int idx) {
        // If we're a single set, return the answer
        if (__builtin_expect(!cur_set, 1)) return iter[idx];

        // Otherwise, assume offset was 0 and that next() has never
        // been called.  Then we just need to find out which set to
        // jump to in order to process the request.
        size_t which = 0;
        while (idx >= size_integral[which + 1]) {
          which++;
        }
        iterator temp_iter = (cur_set + which)->iter();
        return temp_iter[idx - size_integral[which]];
      }
    private:
        const set_type* cur_set;
        iterator iter;
        size_t i;
        const size_t* size_integral;
    };
    Iter iter(size_t start, size_t size) const {
        switch(state) {
                case SET_EMPTY: return iter();
                case SET_SINGLE:
                        assert( start + size <= set.size());
                        return Iter(size, set.iter(start));
                case SET_MANY: {
                        size_t min_set = 0;
                        size_t max_set = sets->size() - 1;
                        size_t mid_set = (min_set + max_set) / 2;;
                                while(min_set < max_set) {
                                        if(start >= (*size_scan)[mid_set + 1])
                                                min_set = mid_set + 1;
                                        else if(start < (*size_scan)[mid_set])
                                                max_set = mid_set - 1;
                                        else
                                                break;
                                        mid_set = (min_set + max_set) / 2;
                                }
                                assert(start < (*size_scan)[mid_set + 1] && start >= (*size_scan)[mid_set]);
                                size_t offset = start - (*size_scan)[mid_set];
                                assert(offset < (*sets)[mid_set].size());
                                return Iter(size, &*(sets->begin() + mid_set), (*sets)[mid_set].iter(offset), &((*size_scan)[mid_set]));
                }
                default: exit(1);
        }
    }
    Iter iter(size_t s) const {
        return iter(s,size() - s);
    }
    Iter iter() const {
        switch(state) {
                case SET_EMPTY: return Iter(0,base_iter_type());
                case SET_SINGLE: return Iter(size(),set.iter());
        case SET_MANY: return Iter(size(), &(*sets->begin()) , sets->front().iter(), &((*size_scan)[0]));
                default: exit(1);
        }
    }

    size_t size() const {
        return size_;
    }
    //empties the set
    void clear() {
        size_ = 0;
        state = SET_EMPTY;
        sets.reset();
        size_scan.reset();
    }
    typedef Iter iterator;
    typedef typename iterator::value_type value_type;
private:
    size_t size_;

    State state;
    set_type set;
    ::std::tr1::shared_ptr<size_list> size_scan;
    ::std::tr1::shared_ptr<set_list> sets;
};

template<typename SetType>
class BoundarySet : public BoundarySetBase<SetType> {
public:
        typedef BoundarySet<SetType> type;
        //called before any calls to addSet
    void beginInit(size_t num_elems) {}
    template<typename MeshT>
    void addSet(const MeshT * mesh, size_t start, size_t end) {
        BoundarySetBase<SetType>::addSetImpl(MeshFuncForType<MeshT,typename SetType::value_type::type>::get(*mesh,start,end));
    }
    //called after all calls to addSet
    void endInit() {}
};

namespace MPI {
template<typename SetType>
class BoundarySet : public ::LisztPrivate::BoundarySetBase<SetType> {
public:
        typedef BoundarySet<SetType> type;
        //called before any calls to addSet
    void beginInit(size_t num_elems) {
        real_size = num_elems;
    }
    template<typename Info>
    void addSet(Info * info, size_t start, size_t end) {
        id_type * local_to_global = info->local_to_global;
        id_type * lbegin = local_to_global;
        size_t n_owned = info->n_owned;
        id_type * lend = local_to_global + n_owned;
        LOG_DBG("start = %lld, end = %lld",start,end);
        id_type * begin_set = std::lower_bound(lbegin,lend,(id_type)start);
        id_type * end_set = std::lower_bound(lbegin,lend,(id_type)end);
        LOG_DBG("lbegin = %d begin_set = %d end_set = %d lend = %d",(int)(lbegin-lbegin),(int)(begin_set-lbegin),(int)(end_set-lbegin),(int)(lend-lbegin));

        if(begin_set != end_set) {
                size_t start_pos = begin_set - lbegin;
                size_t end_pos = end_set - lbegin;
                LOG_DBG("start_pos = %d end_pos = %d, n_owned = %d",(int)start_pos,(int)end_pos,(int)n_owned);
                assert(start_pos < end_pos);
                assert(end_pos <= n_owned);
                addSetImpl(MeshFuncForType<typename Info::mesh_type,typename SetType::value_type::type>::get(*info->mesh,start_pos,end_pos));
        }
    }
    //called after all calls to addSet
    void endInit() {}
    size_t size() const {
        return real_size;
    }
private:
        size_t real_size;
};
}

}  // namespace LisztPrivate
#endif /* BOUNDARY_SET_H_*/
