/*
 * Field.h
 *
 *  Created on: Jul 9, 2009
 *      Author: zdevito
 */

#ifndef FIELDS_FIELD_H_
#define FIELDS_FIELD_H_

#include "MPI/FieldComm.h"
#include "Liszt/ElemTypes.h"
#include "Fields/FieldInfo.h"
#include "Fields/FieldValueReference.h"
#include "Array.h"
#include "noncopyable.h"
#include "ParallelLoad/Common.h"
#include "ParallelLoad/FieldComms.h"
#include <mpi.h>

namespace MPIField {

using namespace ::LisztPrivate::ElemTypes;
using FieldCommon::FieldInfo;

template<typename E,typename V>
class FieldInitializer;


template<typename FieldType, typename ReductionOp, bool nested>
struct EnsureStateHelper;

template<typename E, typename vt>
class Field : public Liszt::noncopyable {
public:
    typedef Field<E,vt> this_type;
    typedef vt value_type;
    typedef E elem_type;
    typedef FieldCommon::FieldValueReference<this_type> reference;
    Field() : write_ghosts_offset(0), reduce_type(LISZT_SUM_TYPE) {
        data.size = 0;
    }

	void init(const std::string & name__,
                 FieldComms * comms,
                 size_t first_ghost,
                 size_t num_ghosts);
                 
    size_t IDToEntryRead(id_type id) const {
        return id;
    }
    size_t IDToEntryWrite(id_type id) const {
        if(id >= start_ghosts)
            return id + write_ghosts_offset;
        return id;
    }

    template<typename MO>
    const value_type & operator()(const MO & mo) const {
        size_t entry = IDToEntryRead(elemToID(LisztPrivate::GetRef(mo)));
#ifndef NDEBUG
        if (!(entry >= 0 && entry <= static_cast<id_type>(data.size))) {
            std::cout << "operator() " << name() << " about to crash on entry "
                    << entry << ", data.size() "
                    << data.size << ", start_ghosts " << start_ghosts
                    << ", write_ghosts_offset " << write_ghosts_offset
                    << std::endl;
        }
#endif
		assert(entry >= 0 && entry <= data.size);
        return data.array[entry];
    }

    template<typename MO>
    reference operator[](const MO & mo) {
        id_type id = elemToID(LisztPrivate::GetRef(mo));
        return reference(*this,id);
    }

    value_type * getReadEntry(id_type id) {
        id_type read_id = IDToEntryRead(id);
#ifndef NDEBUG
        if (!(read_id >= 0 && read_id <= static_cast<id_type>(data.size))) {
            std::cout << "getReadEntry " << name() << " about to crash on id " << id
                    << " with read id " << read_id << ", data.size() "
                    << data.size << ", start_ghosts " << start_ghosts
                    << ", write_ghosts_offset " << write_ghosts_offset
                    << std::endl;
        }
#endif
        assert(read_id >= 0 && read_id <= static_cast<id_type>(data.size));
        return &data.array[read_id];
    }
    const value_type * getReadEntry(id_type id) const {
        id_type read_id = IDToEntryRead(id);
#ifndef NDEBUG
        if (!(read_id >= 0 && read_id <= static_cast<id_type>(data.size))) {
            std::cout << "getReadEntry const " << name() << " about to crash on id " << id
                    << " with read id " << read_id << ", data.size() "
                    << data.size << ", start_ghosts " << start_ghosts
                    << ", write_ghosts_offset " << write_ghosts_offset
                    << std::endl;
        }
#endif
        assert(read_id >= 0 && read_id <= static_cast<id_type>(data.size));
        return &data.array[read_id];
    }
    value_type * getWriteEntry(id_type id) {
        id_type write_id = IDToEntryWrite(id);
#ifndef NDEBUG
        if (!(write_id >= 0 && write_id <= static_cast<id_type>(data.size))) {
            std::cout << "getWriteEntry " << name() << " about to crash on id " << id
                    << " with write id " << write_id << ", data.size() "
                    << data.size << ", start_ghosts " << start_ghosts
                    << ", write_ghosts_offset " << write_ghosts_offset
                    << std::endl;
        }
#endif
        assert(write_id >= 0 && write_id <= static_cast<id_type>(data.size));
        return &data.array[write_id];
    }
    const value_type * getWriteEntry(id_type id) const {
        id_type write_id = IDToEntryWrite(id);
#ifndef NDEBUG
        if (!(write_id >= 0 && write_id <= static_cast<id_type>(data.size))) {
            std::cout << "getWriteEntry const " << name() << " about to crash on id " << id
                    << " with write id " << write_id << ", data.size() "
                    << data.size << ", start_ghosts " << start_ghosts
                    << ", write_ghosts_offset " << write_ghosts_offset
                    << std::endl;
        }
#endif
        assert(write_id >= 0 && write_id <= static_cast<id_type>(data.size));
        return &data.array[write_id];
    }
    const size_t nReadEntries() const {
    	return start_ghosts + write_ghosts_offset;
    }
    const std::string & name() const {
        return name_;
    }
    template<typename T>
    void update() {
        //cout << "updating " << name_ << endl;
        assert(reduce_type == T::type);
        comm.template update<T>();
    }
    template<typename T,typename T2>
    void update2() {
    	//std::cout << name() << ".update2<"<<T::type<<","<<T2::type<<">"<<std::endl;
    	assert(reduce_type == T::type);
    	if(T2::type == LISZT_READ_TYPE) {
    		update<T>();
    	} else {
    		comm.template update2<T,T2>();
    		reduce_type = T2::type;
    	}
    }
    template<typename T>
    void setReductionOp() {
        reduce_type = T::type;
        if(reduce_type != LISZT_ASSIGN::type) {
           size_t field_size = data.size;
           const value_type IDENTITY = T::template identity<value_type>();
           for(size_t i = start_ghosts + write_ghosts_offset; i < field_size; i++) {
               data.array[i] = IDENTITY;
           }
        }
    }
    template<bool nested, typename T>
    void ensureState() {
    	if(T::type == LISZT_ASSIGN_TYPE && only_writes_ghosts)
    		EnsureStateHelper<this_type,LISZT_READ,nested>::run(this);
    	else
    		EnsureStateHelper<this_type,T,nested>::run(this);
    }
    
private:
	template<typename T>
	id_type elemToID(const ElemRef<T> & e) const {
		return e.ID();
	}
    friend class FieldInitializer<value_type,elem_type>;
    template<typename FieldType, typename ReductionOp, bool nested>
	friend struct EnsureStateHelper;
	
    //fields should not be copied
    Field(const Field<E,vt> & rhs) {}
    std::string name_;
    struct {
    	size_t size;
    	Array<value_type> array;
    } data;
    FieldComm<this_type> comm;
    id_type write_ghosts_offset;
    id_type start_ghosts;
    LisztOpType reduce_type;
    bool dirty;
    bool only_writes_ghosts;
};

}

namespace MPIField {

template<typename FieldType, typename ReductionOp>
struct EnsureStateHelper<FieldType,ReductionOp,true> {
	static void run(FieldType * field) {} //nested inside a forall context, ensure state does nothing
};
template<typename FieldType, typename ReductionOp>
struct EnsureStateHelper<FieldType,ReductionOp,false> {
	static void run(FieldType * field) {
		//std::cout << field->name() << ".ensureState("<< ReductionOpToString(ReductionOp::type)<<") state is: "<< ReductionOpToString(field->reduce_type) << " dirty? " << field->dirty << std::endl;
		if(field->reduce_type == ReductionOp::type) {//we are in the correct state already, but we still need to set the dirty bit
			field->dirty = true;
		} else {
			if(field->dirty) {
				//need to call field->update2<OldReductionOp,NewReductionOp>();, must find OldReductionOp
				if(ReductionOp::type == LISZT_READ_TYPE) {
					field->dirty = false;
				}
				switch(field->reduce_type) {
					case LISZT_MAX_TYPE:
						field->template update2<LISZT_MAX,ReductionOp>();
						break;
					case LISZT_MIN_TYPE:
						field->template update2<LISZT_MIN,ReductionOp>();
						break;
					case LISZT_PROD_TYPE:
						field->template update2<LISZT_PROD,ReductionOp>();
						break;
					case LISZT_SUM_TYPE:
						field->template update2<LISZT_SUM,ReductionOp>();
						break;
					case LISZT_LAND_TYPE:
						field->template update2<LISZT_LAND,ReductionOp>();
						break;
					case LISZT_LOR_TYPE:
						field->template update2<LISZT_LOR,ReductionOp>();
						break;
					case LISZT_LXOR_TYPE:
						field->template update2<LISZT_LXOR,ReductionOp>();
						break;
					case LISZT_BAND_TYPE:
						field->template update2<LISZT_BAND,ReductionOp>();
						break;
					case LISZT_BOR_TYPE:
						field->template update2<LISZT_BOR,ReductionOp>();
						break;
					case LISZT_BXOR_TYPE:
						field->template update2<LISZT_BXOR,ReductionOp>();
						break;
					case LISZT_ASSIGN_TYPE:
						field->template update2<LISZT_ASSIGN,ReductionOp>();
						break;
					default:
						assert(!"Unknown Reduction.");
				}
			} else {
				if(ReductionOp::type != LISZT_READ_TYPE) {
					field->dirty = true;
					field->template setReductionOp<ReductionOp>();
				}
			}
		}
		//std::cout << field->name() << " done." << std::endl;
	}
};

template <typename E, typename V>
void Field<E,V>::init(const std::string & name__,
                 FieldComms * comms,
                 size_t first_ghost,
                 size_t num_ghosts) {
    dirty = false;
    size_t nowned = first_ghost;
    size_t nghosts = num_ghosts;
	only_writes_ghosts = false;
    size_t size = nowned;

    start_ghosts = nowned;
    if(comms->reads_ghosts) {
        size += nghosts;
    }
    id_type write_ghosts_start = static_cast<id_type>(size);
    if(comms->writes_ghosts) {
        write_ghosts_offset = size - start_ghosts;
        size += nghosts;
        //if(!comms->reads_ghosts)
        only_writes_ghosts = true;
    } else if(comms->reads_ghosts) {
    	write_ghosts_offset = nghosts;
    } else {
    	write_ghosts_offset = 0;
    }
    name_ = name__;
    std::cout << "DATASIZE: " << size << " Ghosts: " << nghosts << " Owned: " << nowned << " Write Ghosts offset: " << write_ghosts_offset << std::endl;
    data.size = size;
    data.array.reset(size);


    const value_type IDENTITY = Identities<value_type>::additive();
    if(comms->writes_ghosts) {
        for(size_t i = write_ghosts_start; i < size; i++)
            data.array[i] = IDENTITY;
    }
    assert(comms);
    assert(comms->reader != NULL);
    assert(comms->writer != NULL);
    comm.init(MPI_COMM_WORLD,this,comms->reader,comms->writer);
}

} // namespace MPIField
#endif /* FIELD_H_ */
