/*
 * FieldComm.h
 *
 *  Created on: Jul 8, 2009
 *      Author: zdevito
 */

#ifndef FIELDCOMM_H_
#define FIELDCOMM_H_

#include "Vector.h"
#include "Fields/CommInfo.h"
#include "mpi.h"
#include "Fields/Serialize.h"
#include "Fields/Identities.h"
#include "Fields/ReductionOps.h"
// TODO(mbarrien): Should depend on something else for rank_type
#include "Liszt/ElemTypes.h"
#include "liszt_memory.h"

#include <iostream>
namespace FieldCommPrivate {
using namespace std;
using LisztPrivate::ElemTypes::id_type;
using LisztPrivate::ElemTypes::rank_type;

template<typename T>
class FieldBuffer {
public:
    typedef T value_type;
    FieldBuffer(const CommInfo * c, MPI_Comm s)
    : comm(c),
      socket(s),
      buf(new auto_ptr_array_adapter<char>(new char[Serialize<value_type>::SIZE * comm->total_size])),
      status(new auto_ptr_array_adapter<MPI_Request>(new MPI_Request[comm->messages.size()])) {}
   void getBuffer(size_t message, char ** b, size_t * size) {
        *b = (*buf).get() + Serialize<value_type>::SIZE * comm->messages[message].buf_offset;
        *size = Serialize<value_type>::SIZE * comm->messages[message].size;
   }
   size_t numberOfMessages() const {
       return comm->messages.size();
   }
   void issend(size_t message) {
       rank_type rank = comm->messages[message].rank;
       char * b;
       size_t buf_size;
       getBuffer(message,&b,&buf_size);

       MPI_Issend(b,buf_size,MPI_BYTE,static_cast<int>(rank),0,socket,&(*status)[message]);
   }
   void irecvAll() {
       for(size_t i = 0; i < numberOfMessages(); i++)
           irecv(i);
   }
   void irecv(size_t message) {
       size_t size;
       char * b;
       getBuffer(message,&b,&size);
       rank_type rank = comm->messages[message].rank;
       MPI_Irecv(b,size,MPI_BYTE,static_cast<int>(rank),0,socket,&(*status)[message]);
   }
   void waitAny(size_t * idx, char ** b, size_t * nelems) {
       int index;
       MPI_Waitany(numberOfMessages(),(*status).get(),&index,MPI_STATUSES_IGNORE);
       //TODO(zach): what if all messages have been received
       *idx = static_cast<size_t>(index);
       getBuffer(*idx,b,nelems);
   }
   void waitAll() {
       MPI_Waitall(numberOfMessages(),(*status).get(),MPI_STATUSES_IGNORE);
   }

private:
    const CommInfo * comm;
    MPI_Comm socket;
    ::std::auto_ptr<auto_ptr_array_adapter<char> > buf;
    ::std::auto_ptr<auto_ptr_array_adapter<MPI_Request> > status;
};

template<typename T>
class FieldComm {
public:
    typedef T field_type;
    typedef typename field_type::value_type value_type;
    typedef FieldBuffer<value_type> buffer_type;
    static const size_t ELEM_SIZE = Serialize<value_type>::SIZE;

    FieldComm()
    : comm(MPI_COMM_WORLD), send_buf(NULL), recv_buf(NULL) {}
    void init(MPI_Comm c, field_type * f, ::std::tr1::shared_ptr<CommPair> readers,::std::tr1::shared_ptr<CommPair> writers)
     {
        comm = c;
        reader = readers;
        writer = writers;
        send_buf = NULL;
        recv_buf = NULL;
        field = f;
    }

    value_type * fieldWriteEntry(id_type id) {
        return field->getWriteEntry(id);
    }
    value_type * fieldReadEntry(id_type id) {
        return field->getReadEntry(id);
    }

    template<typename Reduce, typename NextReduce, bool set_identity>
    void beginComm(const CommPair & pair) {

        if(send_buf != NULL) {
            // cout << "waiting on previous send" << endl;
            send_buf->waitAll();
            delete send_buf;
            send_buf = NULL;
        }

        assert(send_buf == NULL && recv_buf == NULL);

        if(pair.recv.messages.size() > 0) {
            // cout << "Posting recv" << endl;
            //post all the receives
            recv_buf = new buffer_type(&pair.recv,comm);
            recv_buf->irecvAll();
        }
        
        if(pair.send.messages.size() > 0) {
            __total_messages_sent += pair.send.messages.size();
            const value_type IDENTITY = NextReduce::template identity<value_type>();
            // cout << "Posting send" << endl;
            send_buf = new buffer_type(&pair.send,comm);
            //fill the send buffers and post the sends
            const vector<CommInfo::Message> & messages = pair.send.messages;
            for(size_t i = 0; i < messages.size(); i++) {
              char * buf_start;
              size_t buf_size;
              send_buf->getBuffer(i,&buf_start,&buf_size);
              size_t nids = messages[i].size;
              id_type * ghost_ids = messages[i].values;
              
              __total_bytes_sent += ELEM_SIZE * nids;
              
              char * buf = buf_start;
              for(size_t j = 0; j < nids; j++) {
                  value_type * elem = fieldWriteEntry(ghost_ids[j]);
                  // cout << "Send    LocalID: " << ghost_ids[j] << " has diff: " << *elem << endl;
                  Serialize<value_type>::pack(elem,buf);
                  if(set_identity)
                      *elem = IDENTITY;
                  buf += ELEM_SIZE;
              }
              //post the send
              send_buf->issend(i);
            }
        }

    }

    template<typename Reduce, typename NextReduce, bool set_identity>
    void endComm(const CommPair & pair) {
        
        if(pair.recv.messages.size() > 0) {
            __total_messages_recv += pair.recv.messages.size();
            //std::cout << "recieving from " << pair.recv.messages.size() << " nodes" << std::endl;
            assert(recv_buf != NULL);
            const vector<CommInfo::Message> & messages = pair.recv.messages;
            for(size_t i = 0; i < messages.size(); i++) {
                size_t idx;
                char * buf_start;
                size_t buf_size;
                recv_buf->waitAny(&idx,&buf_start,&buf_size);
                char * buf = buf_start;
                id_type * ghost_ids = messages[idx].values;
                size_t nids = messages[idx].size;
                __total_bytes_recv += ELEM_SIZE * nids;
                for(size_t j = 0; j < nids; j++) {
                    value_type diff;
                    value_type * v = fieldReadEntry(ghost_ids[j]);
                    Serialize<value_type>::unpack(buf,&diff);
                    // cout << "Receive LocalID: " << ghost_ids[j] << " has diff: " << diff << endl;
                    Reduce::apply(diff,v);
                    buf += ELEM_SIZE;
                }
            }
            delete recv_buf;
            recv_buf = NULL;
        }
        //send_buf may not be done, but we don't need to free it yet.
    }
    template<typename Reduce, typename NextReduce >
    void update2() {
      //  cout << "***** " << field->name() << " *****" << endl;
       // cout << "Writers -> Owner" << endl;
        beginComm< Reduce, NextReduce, true >(*writer);
        endComm< Reduce, NextReduce, true >(*writer);
        //cout << "Owners -> Readers" << endl;
        beginComm< LISZT_ASSIGN, LISZT_ASSIGN, false >(*reader);
        endComm< LISZT_ASSIGN, LISZT_ASSIGN, false >(*reader);
    }
    template<typename Reduce>
    void update() {
    	update2<Reduce,Reduce>();
    }
private:
    MPI_Comm comm;
    ::std::tr1::shared_ptr<CommPair> reader;
    ::std::tr1::shared_ptr<CommPair> writer;
    buffer_type * send_buf;
    buffer_type * recv_buf;
    field_type * field;
};

} //namespace FieldCommPrivate

using FieldCommPrivate::FieldComm;
#endif /* FIELDCOMM_H_ */
