/*
 * FieldInit.h
 *
 *  Created on: Jul 26, 2009
 *      Author: zdevito
 */

#ifndef FIELDINIT_H_
#define FIELDINIT_H_

#include "Fields/CommInfo.h"
#include "liszt_memory.h"
#include "Liszt/ElemTypes.h"
#include <mpi.h>

#include "ParallelLoad/LocalTopology.h"
#include "ParallelLoad/FieldAccess.h"
#include "ParallelLoad/FieldComms.h"
#include "MPI/MPIUtil.h"

namespace LisztMPIPrivate {

class FieldInit {
public:
	FieldInit() {}
	void init(MPI_Comm c, shared_ptr<LocalTopology> lt) {
		comm = c;
		local_topology = lt;
	}
	void initComms(ElemType typ, FieldAccess * ghost_access, FieldComms * comms) {
		comms->reader.reset(new CommPair);
		comms->writer.reset(new CommPair);
		LOG_LOG("Ghost reads");
                registerComm(typ,ghost_access->ghost_reads,&comms->reader->recv,&comms->reader->send);
		LOG_LOG("ghost writes");
                registerComm(typ,ghost_access->ghost_writes,&comms->writer->send,&comms->writer->recv);
		LocalTopology::GlobalToLocalFilter filter = local_topology->globalToLocalFilter();
		MeshIO::IOElemType iotyp = ElemTypeToIO(typ);
		
		//translate comm structures to use local IDs
		comms->reader->send.translateID(iotyp,filter);
		comms->reader->recv.translateID(iotyp,filter);
		
		comms->writer->send.translateID(iotyp,filter);
		comms->writer->recv.translateID(iotyp,filter);
		comms->reads_ghosts = ghost_access->ghost_reads.size() > 0;
		comms->writes_ghosts = ghost_access->ghost_writes.size() > 0;
	}
private:
	void registerComm(ElemType typ, FieldAccess::set_t & commset, CommInfo * from, CommInfo * to);
	MPI_Comm comm;
	shared_ptr<LocalTopology> local_topology;
};



inline void FieldInit::registerComm(ElemType typ, FieldAccess::set_t & commset, CommInfo * from, CommInfo * to) {
    //establish a communication pathway to keep fields up to date
    //commset is the set of ghost elements for which we are communicating values for this field, represented by global ids
    //we will look up the owners of each of these elements and then notify them that we are reading/writing

    //if this is a Writer->Owner update
    //from will be a communicator representing the messages sent from this node to the owners
    //to will be a communicator representing messages received by this node from the writers

    //if this is a Owner->Reader update
    //from will be a communicator representing the messages to be received from the owners
    //to will be a communicator representing the messages to be sent to the readers
    int size;
    MPI_Comm_size(comm,&size);
    int rank;
    MPI_Comm_rank(comm,&rank);
    size_t mpi_size = size;

   //commset.size() is the total number of ghosts for this field
   int * send_buf = new int[commset.size()];

   //calcuate the size of each message
   int * send_sizes = new int[mpi_size];
   std::fill_n(send_sizes,mpi_size,0);

   for(FieldAccess::set_t::iterator it = commset.begin(), end = commset.end();
        it != end;
        it++) {
       ElementStatus status;
       bool result = local_topology->lookup(typ,*it,&status);
       assert(result && status.isOwnerKnown());
       PartitionID r = status.owner();
       assert(r < mpi_size && (int)r != rank);
       send_sizes[static_cast<size_t>(r)]++;
    }
   int * send_disp = new int[mpi_size];
   if(MPIU_GetDispFromSize(send_sizes,mpi_size,send_disp) != (int) commset.size())
   		assert(!"commset is inconsistent");

   //communicate to get the recv size
   int * recv_sizes = new int[mpi_size];
   LOG_LOG("alltoall1");
   MPI_Alltoall(send_sizes,1,MPI_INT,recv_sizes,1,MPI_INT,comm);

   int * recv_disp = new int[mpi_size];
   int recv_total_size = MPIU_GetDispFromSize(recv_sizes,mpi_size,recv_disp);

   int * recv_buf = new int [recv_total_size];

   //fill the send_buf with the messages
   //we send the list of global ids that we read/write to their owner
   int * send_pos = new int[mpi_size];
   std::fill_n(send_pos,mpi_size,0);
   for(FieldAccess::set_t::iterator it = commset.begin(), end = commset.end();
        it != end;
        it++) {
       ElementStatus status;
       bool result = local_topology->lookup(typ,*it,&status);
       assert(result && status.isOwnerKnown());
       PartitionID r = status.owner();
       assert(r < mpi_size);
       send_buf[send_disp[static_cast<size_t>(r)] + send_pos[static_cast<size_t>(r)]++] = *it;
    }
   for(size_t i = 0; i < mpi_size; i++)
       assert(send_pos[i] == send_sizes[i]);
   delete [] send_pos;
   LOG_LOG("alltoall2");
   MPI_Alltoallv(send_buf,send_sizes,send_disp,MPI_INT,recv_buf,recv_sizes,recv_disp,MPI_INT,comm);
   //recv_buf now holds the messages that will be received from the readers/writers

   LOG_LOG("frominit");
   //we use these MPI buffers to initialize the Comm objects, which take OWNERSHIP send_buf/recv_buf
   //but only use send_sizes, and send_disp for initialization
   from->init(reinterpret_cast<id_type*>(send_buf),send_sizes,send_disp,mpi_size);
   LOG_LOG("toinit");
   to->init(reinterpret_cast<id_type*>(recv_buf),recv_sizes,recv_disp,mpi_size);


   delete [] send_sizes;
   delete [] send_disp;

   delete [] recv_disp;
   delete [] recv_sizes;
}

}
#endif /* FIELDINIT_H_ */
