/*
 * Global.h
 *
 *  Created on: Aug 4, 2009
 *      Author: zdevito
 */

#ifndef GLOBAL_H_
#define GLOBAL_H_

#include "Fields/ReductionOps.h"
#include "Vector.h"
#include "Matrix.h"
#include "mpi.h"
#include "Fields/Serialize.h"
#include "MPI/Field.h"

namespace LisztMPIPrivate {

template<typename T, bool nested = false>
class Global;

template<typename T>
class Global<T,false> {
public:
    typedef T value_type;
    typedef Global<T> this_type;
	Global(const T & t) {
		rvalue = t;
		reduce_type = LISZT_ASSIGN_TYPE;
		dirty = false;
	}
	Global() {
		reduce_type = LISZT_ASSIGN_TYPE;
		dirty = false;
	}
	const std::string name() {
		return "global";
	}
#define FORWARD_OP(op) \
        this_type & operator op(const value_type & rhs) { \
            lvalue op rhs; \
            return *this;\
        } \
        this_type & operator op(const this_type & r) { \
            return *this op (const value_type &) r; \
        }
        FORWARD_OP(=)
        FORWARD_OP(+=)
        FORWARD_OP(-=)
        FORWARD_OP(*=)
        FORWARD_OP(/=)
        FORWARD_OP(&=)
        FORWARD_OP(|=)
        FORWARD_OP(^=)

#undef FORWARD_OP
        this_type & minWith(const value_type & rhs) {
            lvalue = ReductionImpl<value_type>::min(lvalue,rhs);
            return *this;
        }
        this_type & minWith(const this_type & r) {
            return minWith( (const value_type &) r);
        }

        this_type & maxWith(const value_type & rhs) {
            lvalue = ReductionImpl<value_type>::max(lvalue,rhs);
            return *this;
        }
        this_type & maxWith(const this_type & r) {
            return maxWith( (const value_type &) r);
        }

        this_type & operator++(int) {
            lvalue++;
            return *this;
        }
        this_type & operator--(int) {
            lvalue--;
            return *this;
        }
        this_type & operator++() {
            ++lvalue;
            return *this;
        }
        this_type & operator--() {
            --lvalue;
            return *this;
        }
        const value_type operator-() const {
            return - (const value_type &) (*this);
        }
        operator const value_type &() const {
            return rvalue;
        }
        template<typename Op>
        void setReductionOp() {
            lvalue = Op::template identity<value_type>();
            reduce_type = Op::type;
        }
        template<typename Op>
        void update() {
        	if(Op::type == LISZT_READ_TYPE)
        		ensureState<false,Op>();
        	else
        		update2<Op,Op>();
        }
        
        template<typename Op1,typename Op2>
        void update2();
        
        template<bool nested2, typename ReductionOp>
        void ensureState() {
    	    ::MPIField::EnsureStateHelper<this_type,ReductionOp,nested2>::run(this);
        }
        value_type & writeEntry() {
        	return lvalue;
        }
        value_type & readEntry() {
        	return rvalue;
        }
private:
    T rvalue;
    T lvalue;
    unsigned char send_buf[Serialize<value_type>::SIZE];
    unsigned char recv_buf[Serialize<value_type>::SIZE];
public:
    LisztOpType reduce_type;
    bool dirty;
};

template<typename T>
class Global<T,true> {
public:
	Global(const T & t) {
		value = t;
	}
	Global() {}
    typedef T value_type;
    typedef Global<T> this_type;
        operator const value_type &() const {
            return value;
        }
        value_type & writeEntry() {
        	return value;
        }
        value_type & readEntry() {
        	return value;
        }
        template<bool nested, typename ReductionOp>
        void ensureState() {}
private:
    T value;
};


template<typename T>
struct GetMPIDatatype;
template<>
struct GetMPIDatatype<char> {
    static const MPI_Datatype type() { return MPI_CHAR; }
};
template<>
struct GetMPIDatatype<unsigned char> {
    static const MPI_Datatype type() { return MPI_UNSIGNED_CHAR; }
};
template<>
struct GetMPIDatatype<short int> {
    static const MPI_Datatype type() { return MPI_SHORT; }
};
template<>
struct GetMPIDatatype<unsigned short int> {
    static const MPI_Datatype type() { return MPI_UNSIGNED_SHORT; }
};

template<>
struct GetMPIDatatype<int> {
    static const MPI_Datatype type() { return MPI_INT; }
};

template<>
struct GetMPIDatatype<unsigned int> {
    static const MPI_Datatype type() { return MPI_UNSIGNED; }
};

template<>
struct GetMPIDatatype<long int> {
    static const MPI_Datatype type() { return MPI_LONG; }
};

template<>
struct GetMPIDatatype<unsigned long int> {
    static const MPI_Datatype type() { return MPI_UNSIGNED_LONG; }
};

template<>
struct GetMPIDatatype<float> {
    static const MPI_Datatype type() { return MPI_FLOAT; }
};

template<>
struct GetMPIDatatype<double> {
    static const MPI_Datatype type() { return MPI_DOUBLE; }
};

template<>
struct GetMPIDatatype<long double> {
    static const MPI_Datatype type() { return MPI_LONG_DOUBLE; }
};

template<>
struct GetMPIDatatype<bool> {
	static const MPI_Datatype type() { return MPI_UNSIGNED_CHAR; }
};

template<size_t N, typename T>
struct GetMPIDatatype< vec<N,T> > {
    static const MPI_Datatype type() { return GetMPIDatatype<T>::type(); }
};

template<size_t R, size_t C, typename T>
struct GetMPIDatatype< matrix<R,C,T> > {
    static const MPI_Datatype type() { return GetMPIDatatype<T>::type(); }
};

template<typename T>
struct GetMPILength {
    static size_t length() { return 1; }
};

template<size_t N, typename T>
struct GetMPILength< vec<N,T> > {
    static size_t length() { return N * GetMPILength<T>::length(); }
};

template<size_t R, size_t C, typename T>
struct GetMPILength< matrix<R,C,T> > {
    static size_t length()  { return R * C * GetMPILength<T>::length(); }
};

template<typename LisztOpType>
struct GetMPIOp;

template<>
struct GetMPIOp<LISZT_MAX> {
 static const MPI_Op op() { return MPI_MAX; }
};
template<>
struct GetMPIOp<LISZT_MIN> {
 static const MPI_Op op() { return MPI_MIN; }
};
template<>
struct GetMPIOp<LISZT_PROD> {
 static const MPI_Op op() { return MPI_PROD; }
};
template<>
struct GetMPIOp<LISZT_SUM> {
 static const MPI_Op op() { return MPI_SUM; }
};
template<>
struct GetMPIOp<LISZT_LAND> {
 static const MPI_Op op() { return MPI_LAND; }
};
template<>
struct GetMPIOp<LISZT_LOR> {
 static const MPI_Op op() { return MPI_LOR; }
};
template<>
struct GetMPIOp<LISZT_LXOR> {
 static const MPI_Op op() { return MPI_LXOR; }
};
template<>
struct GetMPIOp<LISZT_BAND> {
 static const MPI_Op op() { return MPI_BAND; }
};
template<>
struct GetMPIOp<LISZT_BOR> {
 static const MPI_Op op() { return MPI_BOR; }
};
template<>
struct GetMPIOp<LISZT_BXOR> {
 static const MPI_Op op() { return MPI_BXOR; }
};
template<>
struct GetMPIOp<LISZT_ASSIGN> {
 static const MPI_Op op() { return MPI_SUM; }
};

template<typename T>
template<typename Op1,typename Op2>
void Global<T,false>::update2() {
    assert(Op1::type == reduce_type);
    __total_global_reductions++;
    __total_global_reduction_bytes += sizeof(T);
    MPI_Op op = GetMPIOp<Op1>::op();
    size_t count = GetMPILength<value_type>::length();
    MPI_Datatype datatype = GetMPIDatatype<value_type>::type();
    Serialize<value_type>::pack(&lvalue,send_buf);
    MPI_Allreduce(send_buf, recv_buf,static_cast<int>(count),datatype,op, MPI_COMM_WORLD);
    T temp;
    Serialize<value_type>::unpack(recv_buf,&temp);
    Op1::apply(temp,&rvalue);
    if(Op2::type == LISZT_READ_TYPE) {
		lvalue = Op1::template identity<value_type>(); 
    } else {
    	lvalue = Op2::template identity<value_type>();
    	reduce_type = Op2::type;
    }
}
} // namespace LisztMPIPrivate
#endif /* GLOBAL_H_ */
