/*
 * LisztMPI.h
 *
 *  Created on: Aug 18, 2009
 *      Author: mbarrien
 */
#ifndef LISZTMPI_H_
#define LISZTMPI_H_

#include "MPI/Field.h"
#include "MPI/FieldInit.h"
#include "MPI/Global.h"
#include "Liszt/BoundarySet.h"
#include "liszt_memory.h"

#include "MeshIO/LisztFileReader.h"
#include "ParallelLoad/Common.h"
#include "ParallelLoad/NameService.h"
#include "ParallelLoad/TopologyService.h"
#include "ParallelLoad/OwnerService.h"
#include "ParallelLoad/LocalTopology.h"
#include "ParallelLoad/LocalMemory.h"
#include "ParallelLoad/Stencil.h"

#include "MPI/MeshFacade.h"
#include "Layout/BoundarySetBuilder.h"

#include <sstream>


namespace LisztMPIPrivate {

#ifndef LisztScalar
#define LisztScalar double
#endif


#define LISZT_RUNTIME_NAMESPACE LisztMPIPrivate

using ::LisztPrivate::ElemTypes::VertexType;
using ::LisztPrivate::ElemTypes::EdgeType;
using ::LisztPrivate::ElemTypes::FaceType;
using ::LisztPrivate::ElemTypes::CellType;


//LisztMPI is the main object that orchestrates the MPI runtime.
//Similar objects exists for some other runtimes, but they do not share identical APIs

//Code in the MPI runtime starts by calling init, further interaction with the liszt object
//occurs indirectly through calls to load fields and boundary sets
class LisztMPI {
public:
	//initialize the MPI runtime: perform mesh loading, partitioning, ghost discovery, and field communication setup.
    void init(int argc, char ** argv) {
		MPI_Init(&argc,&argv);
		assert(argc == 3);
		int log = atoi(argv[2]);
		if(log)
			redirectToLog("liszt_runtime");

		//set up the file reader which reads a mesh files, note that this will only read objects from a file when requested
		reader.reset(new MeshIO::LisztFileReader());
		reader->init(argv[1]);

		//every node reads the entire header and boundary set list from liszt mesh file
		const LisztHeader * header = &reader->header();
		boundary_builder.reset(new BoundarySetBuilder);
		boundary_builder->init(header->nBoundaries, reader->boundaries());
		

		//see individual objects for what each does during initialization in detail
		//name_service answers queries about who has mesh information
		//e.g. who is responsible for knowing the owner of cell x?
		name_service.reset(new NameService(header));

		//the topology service answers queries about topology
		//e.g. get me all the facet edges for cell x? individual requests are aggregated for performance
		shared_ptr<TopologyService> topology_service(new TopologyService(name_service, reader));

		//owner service determines ownership through partition and then answers questions like
		//what node owns cell x? or what are all the nodes that my local node owns?
		shared_ptr<OwnerService> owner_service( new OwnerService(name_service,topology_service->partitionedTopology()));

		//local topology stores the topology necessary to run the program and is used to construct the mesh that will be used
		//at runtime
		shared_ptr<LocalTopology> local_topology(new LocalTopology(name_service->self(),owner_service->locallyOwnedElements()));

		//local memory stores the memory access patterns to field, used to determine field communication.
		shared_ptr<LocalMemory> local_memory(new LocalMemory());

		//The stencil is the abstract representation of the program for the purpose of topology discovery and field memory discovery
		//it will populate local_topology and local_memory based on the initial partition
		shared_ptr<Stencil> stencil(new Stencil(name_service,topology_service,owner_service,local_topology,local_memory,boundary_builder));

		stencil->discoverLocalTopology();
		stencil->discoverLocalMemory();

		//convert the facet edge mesh of local topology into a fast CRSMesh.
		local_topology->buildLocalMesh(&mesh_facade);
		
		
		size_t nElems[ELEM_TYPE_SIZE];
		
		nElems[VERTEX] = header->nV;
		nElems[EDGE] = header->nE;
		nElems[FACE] = header->nF;
		nElems[CELL] = header->nC;
		
		//initialize meta-data needed to later initialize fields
		//local_topology will no longer exist when fields are constructed, so we save the relevent information here
		for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
			first_ghost[i] = local_topology->firstGhost((ElemType)i);
			num_ghosts[i] = local_topology->numberGhost((ElemType)i);
			num_local_elems[i] = first_ghost[i] + num_ghosts[i];
		}

		//mesh_facade is a thin wrapper around a CRSMesh that changes the meanings of vertices,edges,faces,and cells to only point
		//to a range of the entire set.  We use this functionality here to mask out the ghost elements from top level sets like
		//vertices(mesh)
		mesh_facade.initFacade(nElems,first_ghost);
		
		//we need to maintain a local ID -> global ID mapping for correct implementation of liszt's ID function
		for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
			local_topology->buildLocalToGlobal((ElemType)i,&local_to_global[i]);
		}

		//construct a map from field id to the information about field reads/writes it performs
		//when a field is initialized it will use these values to set up its comm structures
		local_memory->buildFieldCommMap(local_topology,field_info);

		//partitionedGIDs a number of mesh elements such that each partition has a range (start,end] of elements
		//partition 0 has elements, 0-N_1, partition 1 elements N_1-N_2.  These IDs are used in sparse matrices to refer to non-local
		//columns
		buildPartitionedGIDs(local_topology);
		local_topology->print_debug(name_service);

		//note that most of the parallel load infrastructure goes out of scope when init exits, freeing all memory but the things
		//that are necessary to perform the run
    }
    
    const MeshFacade * mesh() const {
    	return &mesh_facade;
    }
    
    void finalize() {
    	MPI_Finalize();
    }
    
    void redirectToLog(const char * base) {
    	std::ostringstream ss;
    	int mpi_rank;
    	MPI_Comm_rank(MPI_COMM_WORLD,&mpi_rank);
    	ss << base << "_" << mpi_rank << ".log";
        int file = fileno(fopen( /*(mpi_rank == 0) ?*/ ss.str().c_str() /*: "/dev/null"*/,"a"));
    	dup2(file,1);
    	dup2(file,2);
    }
    
    struct SetBuildInfo {
    	typedef CRSMesh::Mesh mesh_type;
    	id_type * local_to_global;
    	size_t n_owned;
    	const CRSMesh::Mesh * mesh;
    };
    template <typename MO, typename BoundarySet>
    void buildSet(const std::string & name, BoundarySet * bs) {
    	//boundary_builder is used by more than 1 runtime, and is templated on the actual boundary set type,
    	//as well as the argument it passes that boundary set (SetBuildInfo)
    	SetBuildInfo info = { local_to_global[MO::value], first_ghost[MO::value], static_cast<const CRSMesh::Mesh *>(mesh()) };
    	boundary_builder->load<SetBuildInfo,MO,BoundarySet>(&info,name.c_str(),bs);
    }
    
    void emptyFieldComm(FieldComms * fc) {
    	fc->reader.reset(new CommPair);
    	fc->writer.reset(new CommPair);
    	fc->reader->recv.init(NULL,NULL,NULL,0);
    	fc->reader->send.init(NULL,NULL,NULL,0);
    	fc->writer->recv.init(NULL,NULL,NULL,0);
    	fc->writer->send.init(NULL,NULL,NULL,0);
    }
    template<typename T,typename V>
    void buildFieldWithConst(id_type field_id, const std::string & name, const V & init,MPIField::Field<T,V> * field) {
    	typedef LocalMemory::field_comm_map_t map_t;

    	//find the communication info (ghost read/write structures) for this field
    	map_t::iterator it = field_info[T::value].find(field_id);
    	
    	
    	FieldComms * comms;
    	FieldComms empty;

    	//if the field performes no ghost reads/writes it is possible that we will have never created meta-data for it,
    	//in which case we simply allocate empty structures
    	if(it == field_info[T::value].end()) {
    		emptyFieldComm(&empty);
    		comms = &empty;
    	} else {
    		comms = &it->second;
    	}
    	
    	//allocate field memory
    	field->init(name,comms,first_ghost[T::value],num_ghosts[T::value]);

    	//initialize value to constant
    	std::fill_n(field->getReadEntry(0),field->nReadEntries(),init);
    
    }
    template<typename V>
    void buildFieldWithPositions(id_type field_id, const std::string & name, MPIField::Field<VertexType,V> * field) {
        typedef LocalMemory::field_comm_map_t map_t;
    	map_t::iterator it = field_info[VERTEX].find(field_id);
    	FieldComms * comms;
    	FieldComms empty;
    	if(it == field_info[VERTEX].end()) {
    		emptyFieldComm(&empty);
    		comms = &empty;
    		LOG_DBG("emtpy field comms");
    	} else {
    		comms = &it->second;
    	}
    	field->init(name,comms,first_ghost[VERTEX],num_ghosts[VERTEX]);


    	//load the positions from our mesh file in parallel
    	id_type begin,end;

    	//query name service to see which range this node is responsible for
    	name_service->localPartition(VERTEX,&begin,&end);
    	MeshIO::PositionTable * table;
    	table = reader->positions(begin,end); //read position range from the file
		SizedAllToAll<id_type> request;
		request.beginCount();
		
		//positions nows need to be sent to whoever potentially reads them
		LOG_DBG("read entries: %d, first_ghost %d, num_ghost %d",(int)field->nReadEntries(),(int)first_ghost[VERTEX],(int)num_ghosts[VERTEX]);

		//for everything in this fields read entries, request positions
		for(size_t i = 0; i < field->nReadEntries(); i++) {
			id_type global_id = local_to_global[VERTEX][i];
			request.count(name_service->partitionFor(VERTEX,global_id));
		}
		request.beginData();
		for(size_t i = 0; i < field->nReadEntries(); i++) {
			id_type global_id = local_to_global[VERTEX][i];
			request.data(name_service->partitionFor(VERTEX,global_id),global_id);
		}
		size_t n_part;
		const int * offsets;
		const id_type * data;
		size_t total_data = request.endData(&n_part,&offsets,&data);
		
		SizedAllToAll<double3> response;
		response.beginCount();
		for(size_t i = 0; i < n_part; i++) {
			size_t end = (i + 1 == n_part) ? total_data : offsets[i + 1];
			size_t begin = offsets[i];
			response.count(i,end - begin);
		}
		response.beginData();
		//respond to the initial request with the positions read from the local range we are responsible for
		for(size_t i = 0; i < n_part; i++) {
			size_t pend = (i + 1 == n_part) ? total_data : offsets[i + 1];
			size_t pbegin = offsets[i];
			for(size_t j = pbegin; j < pend; j++) {
				size_t e = data[j] - begin;
				double3 entry(table->data[e][0],table->data[e][1],table->data[e][2]);
				response.data(i,entry);
			}
		}
		reader->free(table);
		
		const double3 * rdata;
		total_data = response.endData(&n_part,&offsets,&rdata);
		Array<int> roffsets(n_part);
		std::copy(offsets,offsets+n_part,roffsets.begin());
		//fill the field with the response data
		for(size_t i = 0; i < field->nReadEntries(); i++) {
			id_type global_id = local_to_global[VERTEX][i];
			PartitionID part = name_service->partitionFor(VERTEX,global_id);
			*field->getReadEntry(i) = rdata[roffsets[part]++];
		}
    }

    // This method is intended for debug use only. Do not use in regular code.
    template <typename MO>
    id_type getGlobalID(MO e) const {
    	return local_to_global[MO::type::value][e.ID()];
    }
    
    template<typename MO>
	id_type getPID(MO e) const {
		size_t first_ghost = mesh_facade.firstGhost[MO::type::value];
		if(e.ID() < first_ghost)
			return local_id_offset[MO::type::value] + e.ID();
		else {
			return local_ghost_to_partitioned_gid[MO::type::value][e.ID() - first_ghost];
		}
    }
     //local_id_offset[type] + local_id = partitioned_global_id (note that this is not the same as the global_id which is listed in the file, but a new globally unique id that is continuous per core)
private:

	//we need to keep the name_service, and reader around to load positions, neither takes much memory
	//so it isn't costly to keep them around
	shared_ptr<NameService> name_service;
	shared_ptr<MeshIO::LisztFileReader> reader;
	shared_ptr<BoundarySetBuilder> boundary_builder;
	
	MeshFacade mesh_facade;
	Array<id_type> local_to_global[ELEM_TYPE_SIZE];
	size_t first_ghost[ELEM_TYPE_SIZE];
	size_t num_ghosts[ELEM_TYPE_SIZE];
	size_t num_local_elems[ELEM_TYPE_SIZE];
	LocalMemory::field_comm_map_t field_info[ELEM_TYPE_SIZE];
	
	
	id_type local_id_offset[ELEM_TYPE_SIZE];
	id_type *local_ghost_to_partitioned_gid[ELEM_TYPE_SIZE];
	
	void buildPartitionedGIDs(shared_ptr<LocalTopology> local_topology) {
		int mpi_size,mpi_rank;
		MPI_Comm_size(MPI_COMM_WORLD,&mpi_size);
		MPI_Comm_rank(MPI_COMM_WORLD,&mpi_rank);
		size_t (*local_sizes)[ELEM_TYPE_SIZE] = new size_t[mpi_size][ELEM_TYPE_SIZE];
		//TODO(zach): would be faster to do a parallel scan ...
		MPI_Allgather(num_local_elems,sizeof(size_t)*ELEM_TYPE_SIZE,MPI_BYTE,local_sizes,sizeof(size_t) * ELEM_TYPE_SIZE,MPI_BYTE,MPI_COMM_WORLD);
		//calc local offset
		std::fill(local_id_offset,local_id_offset + ELEM_TYPE_SIZE,0);
		for(int i = 0; i < mpi_rank; i++) {
			for(size_t j = 0; j < ELEM_TYPE_SIZE; j++)
				local_id_offset[j] += local_sizes[i][j];
		}
		delete [] local_sizes;
		
		std::cout << mpi_rank << ": offset is " << local_id_offset[0] << " " << local_id_offset[1] << " " << local_id_offset[2] << " " << local_id_offset[3] << std::endl;
		
		for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
			buildPGIDForElem( (ElemType) i, local_topology);
		}
	}
	
	void buildPGIDForElem(ElemType typ, shared_ptr<LocalTopology> local_topology) {
		int mpi_size,mpi_rank;
		MPI_Comm_size(MPI_COMM_WORLD,&mpi_size);
		MPI_Comm_rank(MPI_COMM_WORLD,&mpi_rank);
		int * send_sizes = new int[mpi_size];
		std::fill(send_sizes, send_sizes + mpi_size, 0);
		size_t n_ghosts = local_topology->numberGhost(typ);
		size_t first_ghost = local_topology->firstGhost(typ);
		id_type * owners = new id_type[n_ghosts];
		{
			for(size_t i = 0; i < n_ghosts; i++) {
				id_type ghost_id = local_to_global[typ][first_ghost + i];
				ElementStatus status;
				bool result = local_topology->lookup(typ,ghost_id,&status);
				assert(result && status.isOwnerKnown());
				PartitionID r = status.owner();
				owners[i] = r;
				send_sizes[r]++;
			}
		}
		int * send_disp = new int[mpi_size];
		int * send_off = new int[mpi_size];
		std::fill(send_off,send_off + mpi_size,0);
		
		int send_total_size = MPIU_GetDispFromSize(send_sizes,mpi_size,send_disp);
		/*cout << "Send Sizes: " << endl;
		for(int i = 0; i < mpi_size; i++) {
			cout << i << " " << send_sizes[i] << endl;
		}*/
		int * recv_sizes = new int[mpi_size];
		MPI_Alltoall(send_sizes,1, MPI_INT, recv_sizes, 1, MPI_INT, MPI_COMM_WORLD);
		/*cout << "Recieve Sizes: " << endl;
		for(int i = 0; i < mpi_size; i++) {
			cout << i << " " << recv_sizes[i] << endl;
		}*/
	
		int * recv_disp = new int[mpi_size];
		int recv_total_size = MPIU_GetDispFromSize(recv_sizes,mpi_size,recv_disp);
		
		int * send_buf = new int[send_total_size];
		int * recv_buf = new int[recv_total_size];
		{
			for(size_t i = 0; i < n_ghosts; i++) {
				id_type ghost_id = local_to_global[typ][first_ghost + i];
				PartitionID r = owners[i];
				int id_location = send_disp[r] + send_off[r];
				send_buf[id_location] = ghost_id;
				owners[i] = id_location;
				send_off[r]++;
			}
		}
		for(int i = 0; i < mpi_size; i++) {
			assert(send_sizes[i] == send_off[i]);
		}
		MPI_Alltoallv(send_buf,send_sizes,send_disp,MPI_INT,recv_buf,recv_sizes,recv_disp,MPI_INT,MPI_COMM_WORLD);
		
		for(int i = 0; i < recv_total_size; i++) {
			ElementStatus status;
			bool result = local_topology->lookup(typ,recv_buf[i],&status);
			assert(result && status.hasLocalID());
			recv_buf[i] = local_id_offset[typ] + status.localID();
		}
		
		MPI_Alltoallv(recv_buf,recv_sizes,recv_disp,MPI_INT,send_buf,send_sizes,send_disp,MPI_INT,MPI_COMM_WORLD);
		{
			for(size_t i = 0; i < n_ghosts; i++) {
				owners[i] = send_buf[owners[i]];
			}
		}
		local_ghost_to_partitioned_gid[typ] = owners;
		delete [] send_sizes;
		delete [] send_disp;
		delete [] send_off;
		delete [] recv_sizes;
		delete [] recv_disp;
		delete [] send_buf;
		delete [] recv_buf;
	}
};

} // namespace LisztMPIPrivate

#endif /* LISZTMPI_H_ */
