#ifndef _SPARSE_MATRIX_H
#define _SPARSE_MATRIX_H
#include <petscvec.h>
#include <petscmat.h>
#include <petscksp.h>
#include "liszt_memory.h"

#include "MPI/SparseVector.h"
//#include "/home/montse/lisztdsl/trunk/externals/petsc-3.1-p0/src/mat/impls/aij/seq/aij.h"

namespace PETSc {

template<typename T>
class SparseMatrix;

template<>
class SparseMatrix<double> {
public:
	typedef PetscScalar value_type;
	SparseMatrix() {
		n_local_rows = 0;
		n_local_cols = 0;
	}
	SparseMatrix(size_t nLR, size_t nLC) {
		init(nLR,nLC,0.0);
	}
	void init(size_t nLR, size_t nLC,const value_type & inital_value) {
		//n_rows = nR;
		//n_cols = nC;
		n_local_rows = nLR - 5 ;
		n_local_cols = nLC - 5 ;
		
		reduce_type = LISZT_ASSIGN_TYPE;
		dirty = false;
		//TODO(zach): size the matrices correctly to prevent dynamic allocation

		PETSC_DO(MatCreateMPIAIJ(MPI_COMM_WORLD,n_local_rows,n_local_cols,PETSC_DETERMINE,PETSC_DETERMINE,20,PETSC_NULL,0,PETSC_NULL,&m));
		//TODO(zach): KSPSolve needs the matrices to have a non-zero diagonal, this is a bit inefficient allocating a vector just
		//to set the diagonal to zero, but it works
		Vec zero;
		VecCreateMPI(MPI_COMM_WORLD,n_local_rows,PETSC_DETERMINE,&zero);
		VecSet(zero,0.0);
		//MatDiagonalSet(m,zero,INSERT_VALUES);
		VecDestroy(zero);
		
		KSPCreate(MPI_COMM_WORLD,&ksp);
		KSPSetOperators(ksp,m,m,SAME_NONZERO_PATTERN);
		KSPSetInitialGuessKnoll(ksp,PETSC_TRUE);
		KSPSetType(ksp,KSPGMRES);
		KSPGMRESSetRestart(ksp,100);
		const double zeroAbs = 1.0e-8;
		const double zeroRel = 1.0e-2;
		const int maxIter = 30;
		KSPSetTolerances(ksp, zeroRel, zeroAbs, 1.e10, maxIter);
		KSPSetFromOptions(ksp);
	}
	~SparseMatrix() {
		if(n_local_rows > 0) {
			KSPDestroy(ksp);
			MatDestroy(m);
		}
	}
	template<bool nested, typename Op>
	void ensureState() {
		assert(Op::type == LISZT_SUM_TYPE || Op::type == LISZT_ASSIGN_TYPE || Op::type == LISZT_READ_TYPE); //PETSc restriction
		if(Op::type == LISZT_READ_TYPE) {
			MatAssemblyBegin(m,MAT_FINAL_ASSEMBLY); 
			MatAssemblyEnd(m,MAT_FINAL_ASSEMBLY); 
			dirty = false;
		} else {
			if(Op::type != reduce_type) {
				MatAssemblyBegin(m,MAT_FINAL_ASSEMBLY); 
				MatAssemblyEnd(m,MAT_FINAL_ASSEMBLY);
			}
			dirty = true;
			reduce_type = Op::type;
		}
	}
	const value_type readEntry(sparse_index_t r, sparse_index_t c) {
		assert(!"unimplemented");
		return 0.0;
	}
	void writeEntry(sparse_index_t r, sparse_index_t c, const value_type & v) {
		assert(dirty);
		// HACK!!! PETSc expects the cells IDs to start with 0. But ours start with 1. So we need adjust for this.
		const PetscInt rr = r - 1;
		const PetscInt cc = c - 1;
	    InsertMode mode = reduce_type == LISZT_SUM_TYPE ? ADD_VALUES : INSERT_VALUES;
		PETSC_DO(MatSetValues(m,1,&rr,1,&cc,&v,mode));
	}
	template<size_t R, size_t C>
	void writeEntry(sparse_index_t r, sparse_index_t c, const matrix<R,C,value_type> & v) {
		PetscInt rr[R];
		PetscInt cc[C];
		// HACK!!! PETSc expects the cells IDs to start with 0. But ours start with 1. So we need adjust for this.
		for(size_t i = 0; i < R; i++)
			rr[i] = r + i - R;
		for(size_t i = 0; i < C; i++)
			cc[i] = c + i - C;
			
		InsertMode mode = reduce_type == LISZT_SUM_TYPE ? ADD_VALUES : INSERT_VALUES;
		PETSC_DO(MatSetValues(m,R,rr,C,cc,(const PetscScalar*)v.v,mode));
	}

	Mat & internal() { return m; }
	KSP & solver() { return ksp; }
private:
	Mat m;
	KSP ksp;
	size_t n_local_rows, n_local_cols;
	LisztOpType reduce_type;
	bool dirty;
};
}
#endif
