#ifndef _PARALLEL_LOAD_LOCAL_TOPOLOGY_H
#define _PARALLEL_LOAD_LOCAL_TOPOLOGY_H

#include "Liszt/Util.h"
#include "ParallelLoad/Common.h"
#include "ParallelLoad/MPIFieldWriter.h"
#include "ParallelLoad/OwnedElements.h"
#include "ParallelLoad/OwnedElement.h"
#include "ParallelLoad/FacetEdgeWithID.h"
#include "MeshIO/FacetEdgeBuilder.h"
#include "ParallelLoad/NameService.h"

enum ElementStatusFlags {
	ESF_HAS_TOPOLOGY = 1,
	ESF_IS_GHOST = 2,
	ESF_OWNER_KNOWN = 4,
	ESF_HAS_ID = 8
};
	
struct ElementStatus {
	ElementStatus() {
		flags_ = 0;
		local_id_ = 0xDEADBEEF;
		owner_ = 0xDEADBEEF;
	}
	bool hasTopology() const {
		return (flags_ & ESF_HAS_TOPOLOGY) != 0;
	}
	bool isGhost() const {
		return (flags_ & ESF_IS_GHOST) != 0;
	}
	bool isOwnerKnown() const {
		return (flags_ & ESF_OWNER_KNOWN) != 0;
	}
	PartitionID owner() const {
		assert(isOwnerKnown());
		return owner_;
	}
	bool hasLocalID() const {
		return (flags_ & ESF_HAS_ID) != 0;
	}
	id_type localID() const {
		assert(hasLocalID());
		return local_id_;
	}
	PartitionID owner_;
	unsigned int flags_;
	id_type local_id_;
};

class LocalTopology {
public:
	LocalTopology(PartitionID sp, shared_ptr<OwnedElements> oe)
	: self_part(sp), 
	  owned_elements(oe) {
		for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
			next_local_id[i] = 0;
			num_ghost[i] = 0;
		}
		
		initOutsideCell(); //all local topologies reference the outside cell and make sure its local ID is 0 as well
		                   //its owner and owner flags will only be set if we need to know them
		for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
			Array<id_type> & arr = owned_elements->data[i];
			for(size_t j = 0; j < arr.size(); j++) {
				OwnedElement oe;
				oe.id = arr[j];
				oe.typ = (ElemType)i;
				oe.owner = self_part;
				addOwner(&oe);
				assignID(oe.typ,oe.id);
			}
			num_owned[i] = arr.size();
		}
	}
	bool lookup(ElemType typ, id_type id, ElementStatus * status) {
		elem_map_t::iterator it = elem_map[typ].find(id);
		bool ret = it != elem_map[typ].end();
		if(ret) {
			*status = it->second;
		}
		return ret;
	}
	void addTopology(const FacetEdgeWithID * feid) {
		const MeshIO::FileFacetEdge & ffe = feid->fe;
		local_mesh.insert(feid->id,&feid->fe);
		
		createOrGetEntry(VERTEX,ffe.hf[0].vert,NULL);
		createOrGetEntry(CELL,ffe.hf[0].cell,NULL);
		
		createOrGetEntry(VERTEX,ffe.hf[1].vert,NULL);
		createOrGetEntry(CELL,ffe.hf[1].cell,NULL);
		
		createOrGetEntry(EDGE,ffe.edge,NULL);
		createOrGetEntry(FACE,ffe.face,NULL);
	
	}
	void addOwner(const OwnedElement * oe) {
		ElementStatus * status;
		createOrGetEntry(oe->typ,oe->id,&status);
		status->flags_ |= ESF_OWNER_KNOWN;
		status->owner_ = oe->owner;
	}
	void topologyRequested(ElemType typ, id_type id) {
		ElementStatus * status;
		createOrGetEntry(typ,id,&status);
		status->flags_ |= ESF_HAS_TOPOLOGY;
	}
	void ownerRequested(ElemType typ, id_type id) {
		ElementStatus * status;
		createOrGetEntry(typ,id,&status);	
		status->flags_ |= ESF_OWNER_KNOWN;
	}
	void addGhost(ElemType typ, id_type id) {
		ElementStatus * status;
		createOrGetEntry(typ,id,&status);
		if(status->isGhost() == false) {
			status->flags_ |= ESF_IS_GHOST;
			num_ghost[typ]++;
		}
		assignID(typ,id);
	}
	
	shared_ptr<OwnedElements> ownedElements() {
		return owned_elements;
	}
	MeshIO::FacetEdgeBuilder * localFacetEdgeBuilder() {
		return &local_mesh;
	}
	
	template<typename MeshT>
	void buildLocalMesh(MeshT * mesh) {
		GlobalToLocalFilter filter(this);
		local_mesh.filter(filter);
		local_mesh.init(elem_map[VERTEX].size(),
		                elem_map[EDGE  ].size(),
		                elem_map[FACE  ].size(),
		                elem_map[CELL  ].size(),0);
		
		mesh->initFromFacetEdgeBuilder(&local_mesh);
	}
	
	void buildLocalToGlobal(ElemType typ, Array<id_type> * output) {
		assert(next_local_id[typ] == numberLocal(typ) && 
		       "All IDs must be assigned before call to buildLocalToGlobal (call buildLocalMesh first)");
		output->reset(numberLocal(typ));
		for(elem_map_t::iterator it = elem_map[typ].begin(),
		   end = elem_map[typ].end();
		   it != end;
		   it++) {
			ElementStatus & status = it->second;
			(*output)[status.localID()] = it->first;
		}
	}
	
	void print_debug(shared_ptr<NameService> name_service) {
	  return;
		for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
			MPIFieldWriter<int> field;
			char buf[32];
			sprintf(buf,"partition_%d",(int)i);
			field.init(buf,ElemTypeToIO((ElemType)i), name_service->nElems((ElemType) i));
			for(elem_map_t::iterator it = elem_map[i].begin(),
			end = elem_map[i].end();
			it != end;
			it++) {
				char typ = (i == VERTEX) ? 'v' :
				           (i == EDGE) ? 'e' :
				           (i == FACE) ? 'f' :
				           (i == CELL) ? 'c' : ' ';
				ElementStatus & status = it->second;
				LOG_DBG("%c%d -> local_id = %d, owner = %d, has_topology = %d",typ,(int) it->first,status.hasLocalID() ? (int) status.localID() : -1,status.isOwnerKnown() ? (int) status.owner() : -1,(int) status.hasTopology()); 
				if(status.isOwnerKnown() && status.owner() == self_part) {
					field.write(it->first,(int)self_part);
				}
			}
			field.close();
		}
	}
	
	size_t numberOwned(ElemType typ) {
		return num_owned[typ];
	}
	size_t numberGhost(ElemType typ) {
		return num_ghost[typ];
	}
	size_t numberLocal(ElemType typ) {
		return elem_map[typ].size();
	}
	//location of the first ghost element
	size_t firstGhost(ElemType typ) {
		if(typ == CELL) {
			ElementStatus status;
			lookup(CELL,0,&status);
			return numberOwned(typ) + ( (!status.isOwnerKnown() || status.owner() != self_part) ? 1 : 0);
		} else return numberOwned(typ);
	}
	//suitable to pass to the FacetEdgeBuilder to create a local mesh
	class GlobalToLocalFilter {
	public:
		GlobalToLocalFilter(LocalTopology * lt) 
		: local_topology(lt) {}
		MeshIO::id_t operator()(MeshIO::IOElemType iotyp, MeshIO::id_t id) {
			ElementStatus * status;
			ElemType typ = IOToElemType(iotyp);
			local_topology->createOrGetEntry(typ,id,&status);
			if(status->hasLocalID()) {
				return status->localID();
			} else {
				local_topology->assignID(typ,id);
				return status->localID();
			}
		}
	private:
		LocalTopology * local_topology;
	};
	GlobalToLocalFilter globalToLocalFilter() {
		return GlobalToLocalFilter(this);
	}
private:
	void initOutsideCell() {
		assignID(CELL,0);
	}
	friend class GlobalToLocalFilter;
	
	void createOrGetEntry(ElemType typ, id_type id, ElementStatus ** status) {
		elem_map_t::iterator it = elem_map[typ].find(id);
		if(it != elem_map[typ].end()) {
			if(status)
				*status = &it->second;
		} else {
			ElementStatus nstatus;
			nstatus.flags_ = 0;
			std::pair<id_type,ElementStatus> tuple(id,nstatus);
			std::pair<elem_map_t::iterator,bool> it = elem_map[typ].insert(tuple);
			if(status)
				*status = &it.first->second;
		}
	}
	void assignID(ElemType typ, id_type id) {
		ElementStatus * status;
		createOrGetEntry(typ,id,&status);
		if(status->hasLocalID() == false) {
			status->flags_ |= ESF_HAS_ID;
			status->local_id_ = next_local_id[typ]++;
		}
	}
	typedef __gnu_cxx::hash_map<id_type, ElementStatus> elem_map_t;
	elem_map_t elem_map[ELEM_TYPE_SIZE];
	
	id_type num_owned[ELEM_TYPE_SIZE];
	id_type num_ghost[ELEM_TYPE_SIZE];
	id_type next_local_id[ELEM_TYPE_SIZE];
	
	MeshIO::FacetEdgeBuilder local_mesh;
    PartitionID self_part;
	shared_ptr<OwnedElements> owned_elements;
};
#endif
