#ifndef _PARALLEL_LOAD_MESH_ADAPTOR_H
#define _PARALLEL_LOAD_MESH_ADAPTOR_H

#include "MeshIO/FacetEdgeBuilder.h"
#include "Layout/BoundarySetBuilder.h"
#include "ParallelLoad/Common.h"
#include "ParallelLoad/FacetEdgeIterators.h"
#include "ParallelLoad/OwnedElements.h"
#include "Liszt/Util.h"

using ::MeshIO::FacetEdgeBuilder;
using ::MeshIO::IOElemType;

//this class takes a FacetEdgeBuilder mesh, and uses it to provide
//the standard liszt mesh operators

//instead of producing sets/itereators, each one of these methods
//simply produces a vector of half facet structures, to obtain an id, 
//you need to call the appropriate method (e.g. to get a cell id, call hf->cell())

class MeshAdaptor : public FEI::IterTypes {
public:
	MeshAdaptor() {}
	MeshAdaptor(FacetEdgeBuilder * b) 
	: mesh(b) {}
	
	void setMesh(FacetEdgeBuilder * b) {
		mesh = b;
	}
	void setOwnedElements(shared_ptr<OwnedElements> oe) {
		owned_elements = oe;
		//we cannot include the 0 cell in cells_m() so if
		//we have the 0 cell in our partition of cells
		//take a slice of the owned cells that does not contain it
		//note that the zero cell should always appear first in the list
		//because the owned elements are sorted
		Array<id_type> & cells = owned_elements->data[CELL];
		if(cells.size() > 0 && cells[0] == 0) {
			cells.slice(1,cells.size(),&cell_slice);
		} else {
			cells.slice(0,cells.size(),&cell_slice);
		}
	}
	void setBoundaryBuilder(shared_ptr<BoundarySetBuilder> bb) {
		boundary_builder = bb;
	}
	
	HF * flip_f(HF * f) {
		return f->flip();
	}
	
	HF * flip_e(HF * e) {
		return e->flip();
	}
	
	size_t vertices_v(HF * m, list_t * ret) {
		FEI::vertices_v()(0,mesh,m,ret);
		return ret->size();
	}
	
	size_t vertices_e(HF * m, list_t * ret) {
		FEI::vertices_e()(0,mesh,m,ret);
		return ret->size();
	}
	
	size_t verticesCCW_f(HF * m, list_t * ret) {
		FEI::cells_e()(1,mesh,m,ret);
		return ret->size();
	}
	
	size_t vertices_f(HF * m, list_t * ret) {
		return verticesCCW_f(m,ret);
	}
	
	size_t verticesCW_f(HF * m, list_t * ret) {
		return verticesCCW_f(flip_f(m),ret);
	}
	
	size_t vertices_c(HF * m, list_t * ret) {
		if(m->cell() == 0)
			ret->clear();
		else
			FEI::cells_v()(1,mesh,m,ret);
		return ret->size();
	}
	
	size_t cells_v(HF * m, list_t * ret) {
		FEI::cells_v()(0,mesh,m,ret);
		return ret->size();
	}
	
	size_t cellsCCW_e(HF * m, list_t * ret) {
		FEI::cells_e()(0,mesh,m,ret);
		return ret->size();
	}
	
	size_t cells_e(HF * m, list_t * ret) {
	   return cellsCCW_e(m,ret);
	}
	
	size_t cellsCW_e(HF * m, list_t * ret) {
		return cellsCCW_e(flip_e(m),ret);
	}
	
	size_t cells_f(HF * m, list_t * ret) {
		FEI::vertices_e()(1,mesh,m,ret);
		return ret->size();
	}
	
	size_t cells_c(HF * m, list_t * ret) {
		if(m->cell() == 0)
			ret->clear();
		else
			FEI::vertices_v()(1,mesh,m,ret);
		return ret->size();
	}
	
	size_t edges_v(HF * m, list_t * ret) {
		FEI::edges_v()(0,mesh,m,ret);
		return ret->size();
	}
	
	
	size_t edgesCCW_f(HF * m, list_t * ret) {
		FEI::faces_e()(1,mesh,m,ret);
		return ret->size();
	}
	
	size_t edges_f(HF * m, list_t * ret) {
		return edgesCCW_f(m,ret);
	}
	
	size_t edges_c(HF * m, list_t * ret) {
		if(m->cell() == 0)
			ret->clear();
		else
			FEI::faces_v()(1,mesh,m,ret);
		return ret->size();
	}
	
	size_t edgesCW_f(HF * m, list_t * ret) {
		return edgesCCW_f(flip_f(m),ret);
	}
	
	size_t faces_v(HF * m, list_t * ret) {
		FEI::faces_v()(0,mesh,m,ret);
		return ret->size();
	}
	
	size_t facesCCW_e(HF * m, list_t * ret) {
		FEI::faces_e()(0,mesh,m,ret);
		return ret->size();
	}
	
	size_t faces_e(HF * m, list_t * ret) {
		return facesCCW_e(m,ret);
	}
	
	size_t faces_c(HF * m, list_t * ret) {
		if(m->cell() == 0)
			ret->clear();
		else
			FEI::edges_v()(1,mesh,m,ret);
		return ret->size();
	}
	
	size_t facesCW_e(HF * m, list_t * ret) {
		return facesCCW_e(flip_e(m),ret);
	}
	
	size_t facets_v(HF * m, list_t * ret) {
		FEI::facets_v()(0,mesh,m,ret);
		return ret->size();
	}
	
	size_t facets_c(HF * m, list_t * ret) {
		if(m->cell() == 0)
			ret->clear();
		else
			FEI::facets_v()(1,mesh,m,ret);
		return ret->size();
	}
	
	size_t facets_e(HF * m, list_t * ret) {
		FEI::facets_e()(0,mesh,m,ret);
		return ret->size();
	}
	
	size_t facets_f(HF * m, list_t * ret) {
		FEI::facets_e()(1,mesh,m,ret);
		return ret->size();
	}
	
	HF * head(HF * e) {
		return e;
	}
	
	HF * tail(HF * e) {
		return e->flip();
	}
	
	HF * outside(HF * e) {
		return e;
	}
	
	HF * inside(HF * e) {
		return e->flip();
	}
	
	//Extract by label functions
	#include "LabelExtractXMacro.h"
	HF * getElementAt(list_t * s, int lbl) {
		return (*s)[lbl];
	}
#define LABEL_EXTRACT_TEMPLATE(fn,tag,r_type,a_type,plural_fn) \
	HF * fn##_##tag(HF * e, int lbl) { \
		list_t l; \
		plural_fn##_##tag(e,&l); \
		return getElementAt(&l,lbl); \
	}
	
LABEL_EXTRACTORS(LABEL_EXTRACT_TEMPLATE)

#undef LABEL_EXTRACT_TEMPLATE


#define LABEL_EXTRACT_CONST_TEMPLATE(fn,tag,r_type,a_type,zero_fn,one_fn) \
	HF * fn##_##tag(HF * e, int lbl) { \
		if(lbl == 0) return zero_fn(e); \
		else return one_fn(e); \
	}
	
LABEL_EXTRACTORS_CONST(LABEL_EXTRACT_CONST_TEMPLATE)
#undef LABEL_EXTRACT_CONST_TEMPLATE
	
	
	size_t size_v(const list_t * s) {
		return s->size();
	}
	
	size_t size(const list_t * s) {
		return s->size();
	}
	
	
	size_t size_e(const list_t * s) {
		return s->size();
	}
	
	size_t size_f(const list_t * s) {
		return s->size();
	}
	
	size_t size_c(const list_t * s) {
		return s->size();
	}
	
	
	HF * towards_f(HF * f, HF * c) {
		if(f->cell() == c->cell())
			return f;
		else {
			return f->flip();
		}
	}
	
	HF * towards_e(HF * e, HF * v) {
		if( e->vert() == v->vert()) 
			return e;
		else {
			return e->flip();
		}
	}
	
	Array<id_type> & all_elems(ElemType typ) {
		return owned_elements->data[typ];
	}
	
	Array<id_type> & vertices_m() {
		return all_elems(VERTEX);
	}
	
		
	Array<id_type> & edges_m() {
		return all_elems(EDGE);
	}
	
	Array<id_type> & faces_m() {
		return all_elems(FACE);
	}
	
		
	Array<id_type> & cells_m() {
	   return cell_slice;
	}
	typedef std::vector< id_type > boundary_list_t;
	
	
	size_t boundary(ElemType typ, const char * name, boundary_list_t * ret) {
		ret->clear();
		BoundarySetBuilder::range_list_t ranges;
		boundary_builder->load_internal(typ,name,&ranges,NULL);
		Array<id_type> & elems = all_elems(typ);
		size_t nelems = 0;
		for(size_t i = 0; i < elems.size(); i++) {
			id_type elem = elems[i];
			for(size_t j = 0; j < ranges.size(); j++) {
				MeshIO::id_t s = ranges[j].first;
				MeshIO::id_t e = ranges[j].second;
				if(elem >= s && elem < e)
					nelems++;
			}
		}
		ret->reserve(nelems);
		for(size_t i = 0; i < elems.size(); i++) {
			id_type elem = elems[i];
			for(size_t j = 0; j < ranges.size(); j++) {
				MeshIO::id_t s = ranges[j].first;
				MeshIO::id_t e = ranges[j].second;
				if(elem >= s && elem < e)
					ret->push_back(elem);
			}
		}
	}
	
	HF * ID_v(HF * hf) { return hf; }
	HF * ID_e(HF * hf) { return hf; }
	HF * ID_f(HF * hf) { return hf; }
	HF * ID_c(HF * hf) { return hf; }
	
	HF * elementTopo(ElemType typ, HF * hf) {
		return hf;
	}
	HF * elementTopo(ElemType typ, id_type id) {
		return mesh->elem_maps[ElemTypeToIO(typ)][id];
	}
	
	id_type elementID(ElemType typ, id_type id) {
		return id;
	}
	
	id_type elementID(ElemType typ, HF * hf) {
		switch(typ) {
			case VERTEX:
				return hf->vert();
			case EDGE:
				return hf->edge();
			case FACE:
				return hf->face();
			case CELL:
				return hf->cell();
			default:
				return 0;
		}
	}
	
	bool isTopologyPresent(ElemType typ, id_type id) {
		MeshIO::IOElemType iotyp = ElemTypeToIO(typ);
		FacetEdgeBuilder::map_t & emap = mesh->elem_maps[iotyp];
		FacetEdgeBuilder::map_t::iterator it = emap.find(id);
		if(it == emap.end()) {
			//LOG_DBG("Found no reference: %d:%d",(int)typ,(int)id);
			return false;
		} else {
			
			HF * hf = it->second;
			assert(hf != NULL);
			
			switch(typ) {
				case VERTEX: {
					FEI::has_all_facets_v haf;
					haf(0,mesh,hf,&buffer);
					return haf.result;
				}
				case EDGE: {
					//mesh->printRing(0,hf);
					FEI::has_all_facets_e haf;
					haf(0,mesh,hf,&buffer);
					return haf.result;
				}
				case FACE: {
					//mesh->printRing(1,hf);
					FEI::has_all_facets_e haf;
					haf(1,mesh,hf,&buffer);
					return haf.result;
				}
				case CELL: {
					FEI::has_all_facets_v haf;
					haf(1,mesh,hf,&buffer);
					return haf.result;
				}
				default:
					return false;
			}
		}
	}
private:
	list_t buffer; //used internally to avoid reallocating a vector on every call to isTopologyPresent
	FacetEdgeBuilder * mesh;
	shared_ptr<OwnedElements> owned_elements;
	shared_ptr<BoundarySetBuilder> boundary_builder;
	Array<id_type> cell_slice;
	
};

#endif