#ifndef _PARALLEL_LOAD_OWNER_SERVICE_H
#define _PARALLEL_LOAD_OWNER_SERVICE_H
#include "ParallelLoad/Common.h"
#include "ParallelLoad/OwnedElements.h"
#include "ParallelLoad/OwnedElement.h"

#include <mpi.h>
#include <parmetis.h>

class OwnerService {
public:
	OwnerService(shared_ptr<NameService> ns,
				 FacetEdgeBuilder * partitioned_topology)
	: name_service(ns){
	
		LOG_LOG("setting up owners arrays");
		
		//this function needs to fill the owners array with the owners for the locally handled partition
		//the name service tells us the range of elements of a particular type for which we are responsible for computing the owners
		//name_service->localPartition(VERTEX,&begin,&end) tells us that we must calculate the owner for the [begin,end) range of cells
		//they should be stored in owners[VERTEX][ vertex_id - begin ]
		
		for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
			id_type begin,end;
			//calculate the ranges we are responsible for, and allocate memory in the owners arrays
			name_service->localPartition((ElemType)i,&begin,&end);
			owners[i].reset(end - begin);
		}
		
		
		//when we add more intelligent partitioners, we should introduce logic here to determine which
		//one to use and then call the correct one, which will have the responsibility for actually
		//filling the ownership arrays
		if(is2D(partitioned_topology)) {
			printf("using 2D partitioner\n");
			partition2D(partitioned_topology);
		} else {
			partition3D(partitioned_topology);
		}
	}
	
private:
	//resolve the global vertex/cell ids into node ids in the paritioning graph
	//nodes are layed out like:
	// Partition 0       | Partition 1
	// [cells 0][verts 0]| [cells 1][verts 1]
	id_type partitioningCellID(id_type global_cell_id) const {
		PartitionID pid = name_service->partitionFor(CELL,global_cell_id);
		id_type vbegin;
		id_type vend;
		name_service->partitionRange(VERTEX,pid,&vbegin,&vend);
		return vbegin + global_cell_id;
	}
	id_type partitioningVertexID(id_type global_vert_id) const {
		PartitionID pid = name_service->partitionFor(VERTEX,global_vert_id);
		id_type cbegin;
		id_type cend;
		name_service->partitionRange(CELL,pid,&cbegin,&cend);
		return cend + global_vert_id;
	}
	void assignFaceAndEdgeOwnershipUsingCellAndVertexOwnership(FacetEdgeBuilder * partitioned_topology) {
		//once we know the CELL and VERTEX information, we then determine the owners for FACE and EDGE
		//we do this by assigning the owner of a FACE (EDGE) to the owner of one of its two cells (vertices).
		
		//since CELL and VERTEX arrays are already set up, it is safe to request ownership information for them
		
		//request owners of verts/cells for the faces/edges
		LOG_LOG("requesting owner information for edges/faces");
		for(int dual = 0; dual <= 1; dual++) {
			beginRequest();
			id_type begin,end;
			name_service->localPartition(dual ? FACE : EDGE,&begin,&end);
			for(size_t i = begin; i < end; i++) {
				FacetEdge::HalfFacet * hf = partitioned_topology->edge_face_m()[dual][i];
				id_t head = hf->vertOrCell(dual);
				id_t tail = hf->flip()->vertOrCell(dual);
				
				//request the owner of the head/tail of this edge, they might not be local so we need to go through the standard request interface
				
				request(dual ? CELL : VERTEX, head);
				request(dual ? CELL : VERTEX, tail);
			}
			shared_ptr< Array<OwnedElement> > owned = endRequest();
			
			//build a map from id -> owner from the requests we made.
			
			__gnu_cxx::hash_map<id_t,PartitionID> omap;
			elementToOwnerMap(owned, &omap);
			
			//look up the owner of the head/tail, and assigned the owner of the edge to one of them
			for(size_t i = begin; i < end; i++) {
				FacetEdge::HalfFacet * hf = partitioned_topology->edge_face_m()[dual][i];
				id_t head = hf->vertOrCell(dual);
				id_t tail = hf->flip()->vertOrCell(dual);
				PartitionID ho = omap[head];
				PartitionID to = omap[tail];
				
				PartitionID owner;
				if(dual && head == 0)
					owner = to;
				else if(dual && tail == 0)
					owner = ho;
				else
					owner = (i % 2 == 0) ? std::min(ho,to) : std::max(ho,to);
				owners[dual ? FACE : EDGE][i - begin] = owner;
				
			}
		}
	}
#if 0
	//we've disabled the more complicated partitioner here because it often produces worse results than the simpler thing
	void partition(FacetEdgeBuilder * partitioned_topology) {
		//set up the arguments to ParMetis
		idxtype * xadj;
		idxtype * adjncy;
		idxtype * vwgt;
		idxtype * ewgt;
		idxtype * part;
		size_t nelems;
		//CreateCSR creates the graph to be partitioned
		CreateCSR(partitioned_topology,&nelems,&xadj,&vwgt,&ewgt,&adjncy);
		int wgtflag = 3;
		int numflag = 0;
		int nparts = name_service->numPartitions();
		int ncon = 2;
		float * tpwgts = new float[ncon*nparts];
		float invSize = 1.f/nparts;
		for(size_t i = 0; i < (size_t) ncon*nparts; i++)
			tpwgts[i] = invSize;
		float ubvec[1] = { 1.05f };
		int options[3] = {1,1,15};
		
		int edgecut;
		
		idxtype * vtxdist = new idxtype[nparts+1];
		for(size_t i = 0; i < (size_t) nparts; i++) {
			id_type vbegin,vend,cbegin,cend;
			name_service->partitionRange(VERTEX,i,&vbegin,&vend);
			name_service->partitionRange(CELL,i,&cbegin,&cend);
			vtxdist[i] = vbegin + cbegin;
			vtxdist[i+1] = vend + cend;
			LOG_LOG("vbegin = %d, vend = %d, cbegin = %d, cend = %d",(int)vbegin,(int)vend,(int)cbegin,(int)cend);
		}
		
		part = new idxtype[nelems];
		
		if(vtxdist[nparts] < 16) {
			//TODO(zach): ParMETIS is crashing when we have really small meshes, I can't figure out why exactly, 
			//it might be our bug or it might be a bug in ParMETIS
			//for now I am just assigning everything to partition 0 in these cases
			id_type vbegin,vend,cbegin,cend;
			name_service->localPartition(CELL,&cbegin,&cend);
			name_service->localPartition(VERTEX,&vbegin,&vend);
			std::fill_n(part,(vend - vbegin) + (cend - cbegin),0);
		} else {
			LOG_LOG("calling ParMETIS");
			MPI_Comm comm = MPI_COMM_WORLD;
			ParMETIS_V3_PartKway(
				 vtxdist,xadj,adjncy,vwgt,ewgt,
				 &wgtflag,&numflag,&ncon,&nparts,tpwgts,ubvec,options,&edgecut,part,&comm);
		}
		for(size_t i = 0; i < owners[CELL].size(); i++)
			owners[CELL][i] = part[i];
		for(size_t i = 0; i < owners[VERTEX].size(); i++) 
			owners[VERTEX][i] = part[owners[CELL].size() + i];
			
		delete [] vtxdist;
		delete [] xadj;
		delete [] adjncy;
		delete [] vwgt;
		delete [] ewgt;
		delete [] part;
	}
	void CreateCSR(FacetEdgeBuilder * topology, size_t * nelems_ptr, idxtype ** xadj, idxtype ** vwgt, idxtype ** ewgt, idxtype ** adjncy) {
		id_type vbegin,vend,cbegin,cend;
		name_service->localPartition(VERTEX,&vbegin,&vend);
		name_service->localPartition(CELL,&cbegin,&cend);
		
		size_t vtxstart = vbegin + cbegin;
		size_t vtxend = vend + cend;
		
		size_t nelems = vtxend - vtxstart;
		
		*xadj = new idxtype[nelems + 1];
		*vwgt = new idxtype[2*nelems];
		*nelems_ptr = nelems;
		
		size_t cur = 0;
		size_t i = 0;
		
		//cout << "csr for: " << vtxstart << " " << vtxend << endl;
		MeshAdaptor p_mesh(topology);
		typedef MeshAdaptor::HF HF;
		MeshAdaptor::list_t cells,faces,verts;
		LOG_DBG("partitioning graph has %d nodes",(int)nelems);
		
		for(size_t c_id = cbegin; c_id < cend; c_id++) {
			HF * c = topology->cellMap()[c_id];
			(*xadj)[i] = cur;
			if(c_id == 0) {
				(*vwgt)[2*i] = 0;
			} else {
				(*vwgt)[2*i] = p_mesh.faces_c(c,&faces);
			}
			(*vwgt)[2*i+1] = 0;
			i++;
			if(c_id == 0)
				continue;
			size_t n_vertices = p_mesh.vertices_c(c,&verts);
			size_t n_cells = p_mesh.cells_c(c,&cells);
			cur += n_vertices;
			LOG_DBG("c = %d, has %d vertices and %d cells",(int)c_id,(int)n_vertices,(int)n_cells);
			
			for(size_t j = 0; j < n_cells; j++) {
				HF * c2 = cells[j];
				LOG_DBG("cell2: %d",(int)c2->cell());
				if(c2->cell() != 0)
					cur++;
			}
		}
		
		for(size_t v_id = vbegin; v_id < vend; v_id++) {
			HF * v = topology->vertMap()[v_id];
			(*xadj)[i] = cur;
			(*vwgt)[2*i] = 0;
			(*vwgt)[2*i+1] = 1;
			i++;
			size_t n_cells = p_mesh.cells_v(v,&cells);
			LOG_DBG("v = %d has %d cells",(int)v_id,(int)n_cells);
			for(size_t j = 0; j < n_cells; j++) {
				HF * c = cells[j];
				LOG_DBG("cell: %d",(int)c->cell());
				if(c->cell() != 0)
					cur++;
			}
		}
		assert(i == vtxend - vtxstart);
		(*xadj)[i] = cur;
		
		*adjncy = new idxtype[cur];
		*ewgt = new idxtype[cur];
		i = 0;
		
		for(size_t c_id = cbegin; c_id < cend; c_id++) {
			HF * c = topology->cellMap()[c_id];
			size_t n_verts = p_mesh.vertices_c(c,&verts);
			for(size_t j = 0; j < n_verts; j++) {
				HF * v = verts[j];
				id_t v_id = v->vert();
				if(c_id == 0)
					continue;
				(*adjncy)[i] = partitioningVertexID(v_id);
				(*ewgt)[i] = 0;
				i++;
			}
			size_t n_cells = p_mesh.cells_c(c,&cells);
			for(size_t j = 0; j < n_cells; j++) {
				HF * c2 = cells[j];
				id_t c2_id = c2->cell();
				if(c_id == 0)
					continue;
				if(c2_id == 0)
					continue;
				(*adjncy)[i] = partitioningCellID(c2_id);
				(*ewgt)[i] = 1;
				i++;
			}
		}
		for(size_t v_id = vbegin; v_id < vend; v_id++) {
			HF * v = topology->vertMap()[v_id];
			size_t n_cells = p_mesh.cells_v(v,&cells);
			for(size_t j = 0; j < n_cells; j++) {
				HF * c = cells[j];
				id_t c_id = c->cell();
				if(c_id == 0)
					continue;
				(*adjncy)[i] = partitioningCellID(c_id);
				(*ewgt)[i] = 0;
				i++;
			}
		}
		LOG_DBG("i = %d, expected %d",(int)i,(int)cur);
		assert(i == cur);
		assignFaceAndEdgeOwnershipUsingCellAndVertexOwnership(partitioned_topology);
	}
#else
	//this is an alternative partitioning routine that just partitions cells and assigns
	//vertices to an arbitrary cell they are next to
	void partition3D(FacetEdgeBuilder * partitioned_topology) {
	
		//this function needs to fill in VERTEX and CELL
		
		//we will set up parmetis by connecting each cell to each neighboring cell 
		
		//set up the arguments to ParMetis, see parmetis docs for what these are
		idxtype * xadj;
		idxtype * adjncy;
		idxtype * vwgt;
		idxtype * ewgt = NULL;
		idxtype * part;
		
		
		id_type cbegin,cend,vbegin,vend;
		name_service->localPartition(CELL,&cbegin,&cend);
		name_service->localPartition(VERTEX,&vbegin,&vend);
		size_t nelems = cend - cbegin;
		
		xadj = new idxtype[nelems + 1];
		vwgt = new idxtype[nelems];
		
		//mesh adaptor provides helper functions using the FacetEdgeBuilder to get lists of mesh topology
		MeshAdaptor p_mesh(partitioned_topology);
		typedef MeshAdaptor::HF HF;
		MeshAdaptor::list_t cells;
		
		size_t adjncy_idx = 0;
		for(size_t c_id = cbegin; c_id < cend; c_id++) {
			int xadj_idx = c_id - cbegin;
			
			xadj[xadj_idx] = adjncy_idx;
			if(c_id == 0)
				continue;
			size_t n_cells = p_mesh.cells_c(partitioned_topology->cellMap()[c_id],&cells);
			vwgt[xadj_idx] = n_cells;
			for(size_t j = 0; j < n_cells; j++) {
				if(cells[j]->cell() != 0) //we ignore cell 0, as it throws off the partitioning to have a single cell with many connections
					adjncy_idx++;
			}
		}
		xadj[nelems] = adjncy_idx;
		
		adjncy = new idxtype[adjncy_idx];
		adjncy_idx = 0;
		
		for(size_t c_id = cbegin; c_id < cend; c_id++) {
			if(c_id == 0)
				continue;
			size_t n_cells = p_mesh.cells_c(partitioned_topology->cellMap()[c_id],&cells);
			for(size_t j = 0; j < n_cells; j++) {
				if(cells[j]->cell() != 0) {
					//LOG_LOG("%d -> %d\n",c_id,cells[j]->cell());
					adjncy[adjncy_idx++] = cells[j]->cell();
				}
			}
		}
		
		int wgtflag = 2; //vert weights
		int numflag = 0;
		int nparts = name_service->numPartitions();
		int ncon = 1;
		float * tpwgts = new float[ncon*nparts];
		float invSize = 1.f/nparts;
		for(size_t i = 0; i < (size_t) ncon*nparts; i++)
			tpwgts[i] = invSize;
		float ubvec[1] = { 1.05f };
		int options[3] = {1,0x7F,15};
		
		int edgecut;
		
		idxtype * vtxdist = new idxtype[nparts+1];
		for(size_t i = 0; i < (size_t) nparts; i++) {
			id_type cbegin,cend;
			name_service->partitionRange(CELL,i,&cbegin,&cend);
			vtxdist[i] = cbegin;
			vtxdist[i+1] = cend;
		}
		LOG_LOG("calling ParMETIS");
		part = new idxtype[nelems];
		MPI_Comm comm = MPI_COMM_WORLD;
		ParMETIS_V3_PartKway(
			vtxdist,xadj,adjncy,vwgt,ewgt,
			&wgtflag,&numflag,&ncon,&nparts,tpwgts,ubvec,options,&edgecut,part,&comm);
		
		delete [] adjncy;
		delete [] xadj;
		delete [] tpwgts;
		delete [] vwgt;
		
		for(size_t i = 0; i < owners[CELL].size(); i++)
			owners[CELL][i] = part[i];	
		delete [] part;
		
		
		//we know assign vertices to one of their cells
		
		beginRequest();
		for(size_t v_id = vbegin; v_id < vend; v_id++) {
			request(CELL,partitioned_topology->vertMap()[v_id]->cell());
		}
		shared_ptr< Array<OwnedElement> > cell_owners = endRequest();
		//create a map from id -> owner
		__gnu_cxx::hash_map<id_t,PartitionID> omap;
		elementToOwnerMap(cell_owners,&omap);
		cell_owners.reset();
		//fill in vertex information
		for(size_t v_id = vbegin; v_id < vend; v_id++) {
			owners[VERTEX][v_id - vbegin] = omap[partitioned_topology->vertMap()[v_id]->cell()];
		}
		assignFaceAndEdgeOwnershipUsingCellAndVertexOwnership(partitioned_topology);
	}
	
	bool is2D(FacetEdgeBuilder * partitioned_topology) {
		return partitioned_topology->nC() < 3;
	}
	struct METISEdge {
		idxtype from;
		idxtype to;
	};
	void partition2D(FacetEdgeBuilder * partitioned_topology) {
		//set up the arguments to ParMetis, see parmetis docs for what these are
		idxtype * xadj;
		idxtype * adjncy;
		idxtype * vwgt;
		idxtype * ewgt = NULL;
		idxtype * part;
		
		typedef MeshAdaptor::HF HF;
		//we will set up parmetis by connecting each cell to each neighboring cell 
		
		SizedAllToAll<METISEdge> edge_list;
		{
			edge_list.beginCount();
			id_type ebegin,eend;
			name_service->localPartition(EDGE,&ebegin,&eend);
			for(id_type e_id = ebegin; e_id < eend; e_id++) {
				HF * edge = partitioned_topology->edgeMap()[e_id];
				edge_list.count(name_service->partitionFor(FACE,edge->face()));
				edge_list.count(name_service->partitionFor(FACE,edge->ccwAroundEdge()->face()));
			}
			edge_list.beginData();
			for(id_type e_id = ebegin; e_id < eend; e_id++) {
				HF * edge = partitioned_topology->edgeMap()[e_id];
				idxtype a = edge->face();
				idxtype b = edge->ccwAroundEdge()->face();
				assert( a != b);
				METISEdge e0 = { a, b }; 
				METISEdge e1 = { b, a };
				edge_list.data(name_service->partitionFor(FACE,a),e0);
				edge_list.data(name_service->partitionFor(FACE,b),e1);
			}
		}
		const METISEdge * metis_edges;
		size_t n_metis_edges = edge_list.endData(NULL,NULL,&metis_edges);
		adjncy = new idxtype[n_metis_edges];
		size_t adjncy_size = n_metis_edges;
		
		id_type fbegin,fend;
		name_service->localPartition(FACE,&fbegin,&fend);
		
		size_t nelems = fend - fbegin;
		
		xadj = new idxtype[nelems + 1];
		vwgt = new idxtype[nelems];
		std::fill(vwgt,vwgt + nelems, 0);
		for(size_t i = 0; i < n_metis_edges; i++) {
			vwgt[metis_edges[i].from - fbegin]++;
		}
		xadj[0] = 0;
		size_t total_weight = 0;
		for(size_t i = 0; i < nelems; i++) {
			total_weight += vwgt[i];
			xadj[i + 1] = total_weight;
			vwgt[i] = 0;
		}
		for(size_t i = 0; i < n_metis_edges; i++) {
			idxtype from = metis_edges[i].from - fbegin;
			idxtype to = metis_edges[i].to;
			idxtype offset = xadj[from] + vwgt[from]++;
			adjncy[offset] = to;
		}
		
		size_t total_faces = name_service->nElems(FACE);
		for(size_t i = 0; i < adjncy_size; i++) {
			assert(adjncy[i] < total_faces);
		}
		for(size_t i = 0; i < nelems; i++) {
			for(size_t j = xadj[i]; j < xadj[i+1]; j++) {
				int a  = i + fbegin;
				int b = adjncy[j];
				if(a == 10 || b == 10 || a == 445 || b == 445)
					fprintf(stderr,"edge %d %d\n",a,b);
			}
		}

		int wgtflag = 2; //vert weights
		int numflag = 0;
		int nparts = name_service->numPartitions();
		int ncon = 1;
		float * tpwgts = new float[ncon*nparts];
		float invSize = 1.f/nparts;
		for(size_t i = 0; i < (size_t) ncon*nparts; i++)
			tpwgts[i] = invSize;
		float ubvec[1] = { 1.05f };
		int options[3] = {1,1,15};
		
		int edgecut;
		
		idxtype * vtxdist = new idxtype[nparts+1];
		for(size_t i = 0; i < (size_t) nparts; i++) {
			id_type fbegin,fend;
			name_service->partitionRange(FACE,i,&fbegin,&fend);
			vtxdist[i] = fbegin;
			vtxdist[i+1] = fend;
		}
		LOG_LOG("calling ParMETIS");
		part = new idxtype[nelems];
		MPI_Comm comm = MPI_COMM_WORLD;
		ParMETIS_V3_PartKway(
			vtxdist,xadj,adjncy,vwgt,ewgt,
			&wgtflag,&numflag,&ncon,&nparts,tpwgts,ubvec,options,&edgecut,part,&comm);
		
		delete [] adjncy;
		delete [] xadj;
		delete [] tpwgts;
		delete [] vwgt;
		
		
		//fill the ownership arrays, since this is 2D, we only have 2 cells, which we assign to node 0
		
		for(size_t i = 0; i < owners[CELL].size(); i++)
			owners[CELL][i] = 0;
			
		for(size_t i = 0; i < owners[FACE].size(); i++) {
			owners[FACE][i] = part[i];	
		}
		delete [] part;
		
		
		//edges get assigned to one of the two faces, vertices to an arbitrary face they are adjacent to
		
		//assigning vertices
		{
			id_type vbegin,vend;
			name_service->localPartition(VERTEX,&vbegin,&vend);
			beginRequest();
			for(size_t v_id = vbegin; v_id < vend; v_id++) {
				request(FACE,partitioned_topology->vertMap()[v_id]->face());
			}
			shared_ptr< Array<OwnedElement> > vert_owners = endRequest();
			//create a map from id -> owner
			__gnu_cxx::hash_map<id_t,PartitionID> omap;
			elementToOwnerMap(vert_owners,&omap);
			vert_owners.reset();
			//fill in vertex information
			for(size_t v_id = vbegin; v_id < vend; v_id++) {
				owners[VERTEX][v_id - vbegin] = omap[partitioned_topology->vertMap()[v_id]->face()];
			}
		}
		
		//assigning edges: edges get assigned to one of their two faces
		{
			id_type ebegin,eend;
			name_service->localPartition(EDGE,&ebegin,&eend);
			beginRequest();
			for(size_t e_id = ebegin; e_id < eend; e_id++) {
				HF * edge = partitioned_topology->edgeMap()[e_id];
				request(FACE,edge->face());
				request(FACE,edge->ccwAroundEdge()->face());
			}
			shared_ptr< Array<OwnedElement> > face_owners = endRequest();
			__gnu_cxx::hash_map<id_t,PartitionID> omap;
			elementToOwnerMap(face_owners,&omap);
			
			face_owners.reset();
			for(size_t e_id = ebegin; e_id < eend; e_id++) {
				HF * edge = partitioned_topology->edgeMap()[e_id];
				PartitionID f0 = omap[edge->face()];
				PartitionID f1 = omap[edge->ccwAroundEdge()->face()];
				PartitionID owner = (e_id % 2 == 0) ? std::min(f0,f1) : std::max(f0,f1);
				owners[EDGE][e_id - ebegin] = owner;
			}
		}
	}
	void elementToOwnerMap(shared_ptr< Array<OwnedElement> > data, __gnu_cxx::hash_map<id_t,PartitionID> * omap) {
		for(size_t e = 0; e < data->size(); e++) {
			(*omap)[(*data)[e].id] = (*data)[e].owner;
		}
	}
#endif
public:
	shared_ptr< OwnedElements > locallyOwnedElements() const {
		shared_ptr< OwnedElements > owned_elements(new OwnedElements);
		for(size_t typ = 0; typ < ELEM_TYPE_SIZE; typ++) {
			SizedAllToAll<id_type> comm;
			id_type begin,end;
			name_service->localPartition((ElemType)typ,&begin,&end);
			
			comm.beginCount();
			for(id_type cur = begin; cur != end; cur++) {
				comm.count(localOwnerFor((ElemType)typ,cur));
			}
			comm.beginData();
			for(id_type cur = begin; cur != end; cur++) {
				comm.data(localOwnerFor((ElemType)typ,cur),cur);
			}
			const id_type * result;
			int size = comm.endData(NULL,NULL,&result);
			owned_elements->data[typ].reset(size);
			std::copy(result, result + size, owned_elements->data[typ].begin());
		}
		return owned_elements;
	}
	void beginRequest() {
		//nothing for now
		requests.clear();
	}
	void request(ElemType typ, id_type id) {
		Element e = { typ, id};
		requests.push_back(e);
	}
	shared_ptr< Array<OwnedElement> > endRequest() {
		SizedAllToAll<Element> request;
		request.beginCount();
		for( std::vector<Element>::iterator it = requests.begin(),
			end = requests.end();
			it != end;
			it++) {
			request.count(name_service->partitionFor(it->typ,it->id));
		}
		request.beginData();
		for( std::vector<Element>::iterator it = requests.begin(),
			end = requests.end();
			it != end;
			it++) {
			request.data(name_service->partitionFor(it->typ,it->id),*it);
		}
		size_t n_partitions;
		const int * offsets;
		const Element * data;
		int total_recv = request.endData(&n_partitions,&offsets,&data);
		SizedAllToAll<OwnedElement> response;
		response.beginCount();
		for(size_t i = 0; i < n_partitions; i++) {
			size_t next = (i + 1 == n_partitions) ? total_recv : offsets[i + 1];
			response.count(i,next - offsets[i]);
		}
		response.beginData();
		for(size_t i = 0; i < n_partitions; i++) {
			size_t begin = offsets[i];
			size_t end = (i + 1 == n_partitions) ? total_recv : offsets[i + 1];
			for(size_t cur = begin; cur != end; cur++) {
				const Element & elem = data[cur];
				OwnedElement oe; 
				oe.typ = elem.typ;
				oe.id = elem.id;
				oe.owner = localOwnerFor(elem.typ,elem.id);
				response.data(i,oe);
			}
		}
		request.reset();
		const OwnedElement * oe_data;
		total_recv = response.endData(NULL,NULL,&oe_data);
		shared_ptr< Array<OwnedElement> > ret(new Array<OwnedElement>(total_recv));
		std::copy(oe_data,oe_data + total_recv,ret->begin());
		response.reset();
		return ret;
	}
private:
	
	PartitionID localOwnerFor(ElemType typ, id_type id) const {
		id_type begin,end;
		name_service->localPartition(typ, &begin, &end);
		
		assert(begin <= id && id < end);
		
		PartitionID result = owners[typ][id - begin];
		//LOG_DBG("localOwnerFor %d %d is in range %d to %d: %d",(int)typ,(int)id,(int)begin,(int)end,(int)result);
		assert((id - begin) < owners[typ].size());
		return result;
	}
	
	Array<PartitionID> owners[ELEM_TYPE_SIZE];
	shared_ptr<NameService> name_service;
	std::vector<Element> requests;
};
#endif
