#ifndef _PARALLEL_LOAD_STENCIL_H
#define _PARALLEL_LOAD_STENCIL_H

#include "ParallelLoad/Common.h"
#include "ParallelLoad/Util.h"
#include "ParallelLoad/NameService.h"
#include "ParallelLoad/TopologyService.h"
#include "ParallelLoad/OwnerService.h"
#include "ParallelLoad/LocalTopology.h"
#include "ParallelLoad/LocalMemory.h"

class Stencil : public MeshAdaptor {
public:
	Stencil(shared_ptr<NameService> ns,
			shared_ptr<TopologyService> ts, 
			shared_ptr<OwnerService> os, 
			shared_ptr<LocalTopology> lt,
			shared_ptr<LocalMemory> lm,
			shared_ptr<BoundarySetBuilder> bb) 
	: name_service(ns),
	topology_service(ts), 
	owner_service(os), 
	local_topology(lt),
	local_memory(lm),
	boundary_builder(bb) {
	
		//initialize the mesh adapator
		this->setMesh(local_topology->localFacetEdgeBuilder());
		this->setOwnedElements(local_topology->ownedElements());
		this->setBoundaryBuilder(boundary_builder);
	}
	
	void discoverLocalTopology() {
		for(size_t i = 0; i < generatedMaxDepth(); i++) {
			topology_service->beginRequest();
			LOG_LOG("Getting topology for depth %d",(int)i);
			generatedDiscoveryAtDepth(true,i);
			LOG_LOG("Beginning Request");
                        shared_ptr< Array<FacetEdgeWithID> > data = topology_service->endRequest();
			LOG_LOG("Ending Request");
			for(FacetEdgeWithID * cur = data->begin(); cur != data->end(); cur++) {
				local_topology->addTopology(cur);
			}
		}
	}
	void discoverLocalMemory() {
		owner_service->beginRequest();
		generatedDiscoveryAtDepth(false,0); //depth is ignored for field discovery
		shared_ptr< Array< OwnedElement > > data = owner_service->endRequest();
		for(OwnedElement * cur = data->begin(); cur != data->end(); cur++) {
			local_topology->addOwner(cur);
		}
	}
private:
	void requestTopologyIfAbsent(ElemType typ, id_type id) {
		ElementStatus status;
		if(!local_topology->lookup(typ,id,&status) || status.hasTopology() == false) {
			local_topology->topologyRequested(typ,id); //mark the topology present in local_topology to prevent requesting again
			assert(local_topology->lookup(typ,id,&status));
			//there is a chance we already have the topology for this element even if 
			//we haven't specifically requested it.
			//we call isTopologyPresent to check locally before adding the request
			if(!isTopologyPresent(typ,id)) {
				LOG_DBG("requesting %d:%d",(int)typ,(int)id);
				topology_service->request(typ,id);
			} else {
				LOG_DBG("requested %d:%d, but topology is already present",(int)typ,(int)id);
			}
			
		}
	}
	
	void requestGhostIfAbsent(bool is_read, id_type field_id, ElemType typ, id_type id) {
		ElementStatus status;
		if(local_topology->lookup(typ,id,&status)) {
			if(status.isOwnerKnown() == false) {
				owner_service->request(typ,id);
				local_topology->ownerRequested(typ,id); //mark the owner present in local_topology to prevent requesting again
			}
			//skip the outside cell, you are not allowed to store values at it
			//is this element not owned by this partition, if so we have found a ghost
			if( !(typ == CELL && id == 0) &&
			    (status.isOwnerKnown() == false || status.owner() != name_service->self()) ) {
				local_topology->addGhost(typ,id);
				if(is_read)
					local_memory->addGhostRead(field_id,typ,id);
				else
					local_memory->addGhostWrite(field_id,typ,id);
			}
		} else {
			LOG_ERROR("Missing Ghost: %d %d",typ,id);
		}
	}
	
	shared_ptr<NameService> name_service;
	shared_ptr<TopologyService> topology_service;
	shared_ptr<OwnerService> owner_service;
	shared_ptr<LocalTopology> local_topology;
	shared_ptr<LocalMemory> local_memory;
	shared_ptr<BoundarySetBuilder> boundary_builder;
	//implemented by liszt compiler-generated code
	size_t generatedMaxDepth() const;
	void generatedDiscoveryAtDepth(bool discover_topology, size_t depth);
};


#define STENCIL_BEGIN_FIELDS \
if(!discover_topology) {

#define STENCIL_END_FIELDS \
}

//starts a loop, should only occur in a USE section
#define STENCIL_BEGIN_FORALL(v,setname) \
{\
typedef typeof(setname) __set_type; \
typedef __set_type::value_type __value_type; \
for(size_t __i = 0; \
    __i < (setname).size(); \
    __i++) { \
    __value_type v = (setname)[__i];

#define STENCIL_END_FORALL \
} \
} 

#define STENCIL_DEPTH(stmt_depth) \
if(!discover_topology || stmt_depth <= depth)

#define STENCIL_REQUEST_DEPTH(stmt_depth) \
if(discover_topology && stmt_depth == depth + 1)

#if 0
//example stencil code, normally generated by the scala compiler
size_t Stencil::generatedMaxDepth() const { return 3; }
void Stencil::generatedDiscoveryAtDepth(bool discover_topology, size_t depth) {
	STENCIL_BEGIN_DECLS
		list_t faces;
	STENCIL_END_DECLS
	
	STENCIL_BEGIN_REQUESTS
	STENCIL_END_REQUESTS
	
	STENCIL_BEGIN_USES
		Array<id_type> & cells = cells_m();
		STENCIL_BEGIN_FORALL(c,cells)
			STENCIL_BEGIN_REQUESTS
				requestTopologyIfAbsent(CELL,elementID(CELL,c));
			STENCIL_END_REQUESTS
			STENCIL_BEGIN_USES
				faces_c(elementTopo(CELL,c),&faces);
				STENCIL_BEGIN_FORALL(f,faces)
					STENCIL_BEGIN_REQUESTS
						requestTopologyIfAbsent(FACE,elementID(FACE,f));
						requestTopologyIfAbsent(FACE,elementID(FACE,f));
					STENCIL_END_REQUESTS
					STENCIL_BEGIN_USES
						HF * c0 = outside(elementTopo(FACE,f));
						HF * c1 = inside(elementTopo(FACE,f));
					STENCIL_END_USES
				STENCIL_END_FORALL
			STENCIL_END_USES
		STENCIL_END_FORALL
	STENCIL_END_USES
}

#endif

#endif
