#ifndef _SMP_GLOBAL_H
#define _SMP_GLOBAL_H


#define GLOBAL_CACHE_SIZE 1024
static char globals_cache[GLOBAL_CACHE_SIZE];

template<typename T, bool nested>
class Global;

template<typename T>
class Global<T,true> {
public:
	Global() {}
	Global(const T & rhs) : value(rhs) {}
	T & read_value() { return value; }
	T & write_value() { return value; }

	template<bool nested, typename Op>
	void ensureState(size_t num_threads, size_t thread_id) {}
	
private:
	T value;
};


template<typename T>
class Global<T,false> {
public:
	Global() : state(LISZT_READ_TYPE) {}
	Global(const T & rhs) : read(rhs), state(LISZT_READ_TYPE) {}
	T & read_value() { return read; }
	T & write_value() { return write; }
	
	template<typename Op>
	void ensureState(size_t num_threads, size_t thread_id) {
		//transitions are:
		// read -> write(x) //set write to identity(x)
		// write(x) -> read //runOp<x>
		// write(x) -> write(y) //runOp<x> set write to identity(x)
		if(Op::type != state) { //early out for fast path (i.e. no state change)
			//printf("%s -> %s\n",ReductionOpToString(state),ReductionOpToString(Op::type));
			if(state == LISZT_READ_TYPE) { //read->write(x)
				setIdentity<Op>();
			} else if(Op::type == LISZT_READ_TYPE) { //write(x) -> read
				runOp(state,num_threads,thread_id);
			} else { //write(x) -> write(y)
				runOp(state,num_threads,thread_id);
				setIdentity<Op>();
			}
			//update the state
			state = Op::type;
		}
	}
	
	template<typename Op>
	void setIdentity() {
		//printf("setting identity %s\n",ReductionOpToString(Op::type));
		write = Op::template identity<T>();
	}
	
	void runOp(LisztOpType type,size_t num_threads,size_t thread_id) {
		//using LISZT_REDUCTION_OPS x-macro
		#define RUN_X(x) case x##_TYPE: runOp<x>(num_threads,thread_id); break;
		switch(type) {
			LISZT_REDUCTION_OPS(RUN_X)
		}
		#undef RUN_X
	}
	template<typename Op>
	void runOp(size_t num_threads, size_t thread_id) {
		//printf("performing reduction %s\n",ReductionOpToString(Op::type));
		T * data = (T*) globals_cache;
		data[thread_id] = write;
		barrier_wait(&System::barrier);
		for(size_t i = 0; i < num_threads; i++) {
			//printf("reducing from %llu: %d\n",i,data[i]); 
			Op::apply(data[i],&read); 
		}
		barrier_wait(&System::barrier);
	}
	
private:
	LisztOpType state;
	T read;
	T write;
};

//MACROS FOR SCALA REDUCTIONS

#define SCALAR_REDUCTION_READ(v) (v.read_value())
#define SCALAR_REDUCTION_NESTED_WRITE(v) (v.write_value())
#define SCALAR_REDUCTION_UNNESTED_WRITE(v) (v.read_value())
#define SCALAR_REDUCTION_ENSURE_STATE(v,state) v.ensureState<state>(this->num_threads,this->thread_id)

#endif