// Matrix impl
template<typename MO, size_t N, size_t M>
class Field<MO, matrix<N, M, float> > : public FieldBase {
public:
  enum { MatrixSize = N*M };

  bool init(size_t sz, const std::string& name) {
    if (initialized) return false;
    FieldBase::init(sz, name, ElemTypeToIO(MO::value));
    size = sz;
    // round up to big enough alloc for vectorization
    alloc_size = (size + 15) & (~15);
    data = new float[MatrixSize * alloc_size];
    initialized = true;
    return true;
  }

  void init(size_t sz, const std::string& n, const matrix<N, M, float>& value) {
    init(sz, n);
    for (size_t r = 0; r < N; r++) {
      for (size_t c = 0; c < M; c++) {
        FixedVector<float, 16> splatted(value(r,c));
        float* dst = data + alloc_size * (r * M + c);
        for (size_t i = 0; i < alloc_size/16; i++) {
          splatted.store_aligned(dst + 16 * i);
        }
      }
    }
  }

  struct MatWriteBack {
    SYRAH_FORCEINLINE MatWriteBack(size_t which, float* data, size_t alloc_size) : which(which), data(data), alloc_size(alloc_size) {
    }

    SYRAH_FORCEINLINE matrix<N, M, float> operator()() {
      matrix<N, M, float> result;
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t i = M * r + c;
          cacheSim(&data[alloc_size * i + which]);
          result(r, c) = data[alloc_size * i + which];
        }
      }
      return result;
    }

    // Cast to matrix<N, M, float>
    SYRAH_FORCEINLINE operator matrix<N, M, float>() const {
      matrix<N, M, float> result;
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t i = M * r + c;
          cacheSim(&data[alloc_size * i + which]);
          result(r, c) = data[alloc_size * i + which];
        }
      }
      return result;
    }

    // Dimension indexing
    SYRAH_FORCEINLINE float& operator()(size_t r, size_t c) {
      size_t i = M * r + c;
      cacheSim(&data[alloc_size * i + which]);
      return data[i * alloc_size + which];
    }

    SYRAH_FORCEINLINE const float& operator()(size_t r, size_t c) const {
      size_t i = M * r + c;
      cacheSim(&data[alloc_size * i + which]);
      return data[i * alloc_size + which];
    }

    // Store mat
    SYRAH_FORCEINLINE MatWriteBack& operator=(const matrix<N, M, float>& mat) {
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t i = M * r + c;
          cacheSim(&data[alloc_size * i + which]);
          data[alloc_size * i + which] = mat(r, c);
        }
      }
      return *this;
    }

    // +=
    SYRAH_FORCEINLINE MatWriteBack& operator+=(const matrix<N, M, float>& mat) {
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t i = M * r + c;
          cacheSim(&data[alloc_size * i + which]);
          data[alloc_size * i + which] += mat(r, c);
        }
      }
      return *this;
    }

    // -=
    SYRAH_FORCEINLINE MatWriteBack& operator-=(const matrix<N, M, float>& mat) {
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t i = M * r + c;
          cacheSim(&data[alloc_size * i + which]);
          data[alloc_size * i + which] -= mat(r, c);
        }
      }
      return *this;
    }

    // *= scalar
    SYRAH_FORCEINLINE MatWriteBack& operator*=(const float& scalar) {
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t i = M * r + c;
          cacheSim(&data[alloc_size * i + which]);
          data[alloc_size * i + which] *= scalar;
        }
      }
      return *this;
    }

    // /= scalar
    SYRAH_FORCEINLINE MatWriteBack& operator/=(const float& scalar) {
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t i = M * r + c;
          cacheSim(&data[alloc_size * i + which]);
          data[alloc_size * i + which] /= scalar;
        }
      }
      return *this;
    }

    // Assign from other
    SYRAH_FORCEINLINE MatWriteBack& operator=(const MatWriteBack& other) {
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t i = M * r + c;
          cacheSim(&data[alloc_size * i + which]);
          data[alloc_size * i + which] -= other(r, c);
        }
      }
      return *this;
    }

    size_t which;
    float* data;
    size_t alloc_size;
  };

  template<typename ME>
  SYRAH_FORCEINLINE MatWriteBack operator()(const ME& mo) const {
    return MatWriteBack(mo.ID(), data, alloc_size);
  }

  template<typename ME>
  SYRAH_FORCEINLINE MatWriteBack operator[](const ME& mo) {
    return MatWriteBack(mo.ID(), data, alloc_size);
  }

  SYRAH_FORCEINLINE MatWriteBack internal(size_t which) {
    return MatWriteBack(which, data, alloc_size);
  }

  template<typename ME>
  SYRAH_FORCEINLINE const matrix<N, M, float> operator[](const ME& mo) const {
    size_t which = mo.ID();
    matrix<N, M, float> result;
    for (size_t r = 0; r < N; r++) {
      for (size_t c = 0; c < M; c++) {
        size_t i = r * M + c;
        cacheSim(&data[alloc_size * i + which]);
        result(r, c) = data[alloc_size * i + which];
      }
    }
    return result;
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE syrah::SimdMatrix<N, M, float, VecSize> simdMatRead(const syrah::SIMDObject<VecSize, ME>& mo) const {
    SYRAH_ALIGN(32) float local_vals[N][M][VecSize];
    for (size_t i = 0; i < VecSize; i++) {
      size_t id = mo[i].ID();
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t src = alloc_size * (r * M + c) + id;
          cacheSim(&data[src]);
          local_vals[r][c][i] = data[src];
        }
      }
    }
    return syrah::SimdMatrix<N, M, float, VecSize>(&(local_vals[0][0][0]), true);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE syrah::SimdMatrix<N, M, float, VecSize> simdMatReadSequential(const syrah::SIMDObject<VecSize, ME>& mo) const {
    size_t offset = mo[0].ID();
    const float* striped[MatrixSize];
    for (size_t i = 0; i < MatrixSize; i++) { striped[i] = &data[i * alloc_size + offset]; }
    return syrah::SimdMatrix<N, M, float, VecSize>(striped);
  }



  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdMatrix<N, M, float, (int)VecSize>& m) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[N][M][VecSize];
    m.store_aligned(&local_vals[0][0][0]);
    for (size_t i = 0; i < VecSize; i++) {
      size_t id = mo[i].ID();
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t dst = alloc_size * (r * M + c) + id;
          cacheSim(&data[dst]);
          data[dst] = local_vals[r][c][i];
        }
      }
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdMatrix<N, M, float, (int)VecSize>& m, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[N][M][VecSize];
    m.store_aligned(&local_vals[0][0][0]);
    for (size_t i = 0; i < VecSize; i++) {
      if (!mask.get(i)) continue;
      size_t id = mo[i].ID();
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t dst = alloc_size * (r * M + c) + id;
          cacheSim(&data[dst]);
          data[dst] = local_vals[r][c][i];
        }
      }
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdMatrix<N, M, float, (int)VecSize>& m) {
    size_t offset = mo[0].ID();
    float* striped[MatrixSize];
    for (size_t i = 0; i < MatrixSize; i++) { striped[i] = &data[i * alloc_size + offset]; }
    m.store(striped);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdMatrix<N, M, float, (int)VecSize>& m, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    size_t offset = mo[0].ID();
    float* striped[MatrixSize];
    for (size_t i = 0; i < MatrixSize; i++) { striped[i] = &data[i * alloc_size + offset]; }
    m.store(striped, mask);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const matrix<N, M, float>& mat){
    size_t offset = mo[0].ID();
    float* striped[MatrixSize];
    for (size_t i = 0; i < MatrixSize; i++) { striped[i] = &data[i * alloc_size + offset]; }
    syrah::SimdMatrix<N, M, float, (int)VecSize> m(mat);
    m.store(striped);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const matrix<N, M, float>& mat, const syrah::FixedVectorMask<(int)VecSize>& mask){
    size_t offset = mo[0].ID();
    float* striped[MatrixSize];
    for (size_t i = 0; i < MatrixSize; i++) { striped[i] = &data[i * alloc_size + offset]; }
    syrah::SimdMatrix<N, M, float, (int)VecSize> m(mat);
    m.store(striped, mask);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWritePlusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdMatrix<N, M, float, (int)VecSize>& m) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[N][M][VecSize];
    m.store_aligned(&local_vals[0][0]);
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t j = r * M + c;
          cacheSim(&data[alloc_size * j + which]);
          data[alloc_size * j + which] += local_vals[r][c][i];
        }
      }
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWritePlusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdMatrix<N, M, float, (int)VecSize>& m, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[N][M][VecSize];
    // use merge to make the += 0
    syrah::SimdMatrix<N, M, float, (int)VecSize> merged = select(m, syrah::SimdMatrix<N, M, float, (int)VecSize>::Zero(), mask);
    merged.store_aligned(&local_vals[0][0][0]);
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t j = r * M + c;
          cacheSim(&data[alloc_size * j + which]);
          data[alloc_size * j + which] += local_vals[r][c][i];
        }
      }
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteMinusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdMatrix<N, M, float, (int)VecSize>& m) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[N][M][VecSize];
    m.store_aligned(&local_vals[0][0][0]);
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t j = r * M + c;
          cacheSim(&data[alloc_size * j + which]);
          data[alloc_size * j + which] -= local_vals[r][c][i];
        }
      }
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteMinusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdMatrix<N, M, float, (int)VecSize>& v, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[N][M][VecSize];
    // use merge to make the += 0
    syrah::SimdMatrix<N, M, float, (int)VecSize> merged = select(v, syrah::SimdMatrix<N, M, float, (int)VecSize>::Zero(), mask);
    merged.store_aligned(&local_vals[0][0][0]);
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t j = r * M + c;
          cacheSim(&data[alloc_size * j + which]);
          data[alloc_size * j + which] -= local_vals[r][c][i];
        }
      }
    }
  }

  // scalar writes
  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const matrix<N, M, float>& mat) {
    assert(MO::value == ME::type::value);
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          size_t j = r * M + c;
          cacheSim(&data[alloc_size * j + which]);
          data[alloc_size * j + which] = mat(r, c);
        }
      }
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const matrix<N, M, float>& mat, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    for (size_t i = 0; i < VecSize; i++) {
      if (mask.get(i)) {
        size_t which = mo[i].ID();
        for (size_t r = 0; r < N; r++) {
          for (size_t c = 0; c < M; c++) {
            size_t j = r * M + c;
            cacheSim(&data[alloc_size * j + which]);
            data[alloc_size * j + which] = mat(r, c);
          }
        }
      }
    }
  }

  template<typename ME>
  SYRAH_FORCEINLINE void lock(const ME & mo) {
  }

  template<typename ME>
  SYRAH_FORCEINLINE void unlock(const ME & mo) {
  }

  float* data;
  size_t alloc_size;
};
