// Float impl
template<typename MO>
class Field<MO, float> : public FieldBase {
public:
  bool init(size_t sz, const std::string& name) {
    if (initialized) return false;
    FieldBase::init(sz, name, ElemTypeToIO(MO::value));
    size = sz;
    // round up to big enough alloc for vectorization
    alloc_size = (size + 31) & (~31);
    data = new float[alloc_size];
    initialized = true;
    return true;
  }

  void init(size_t sz, const std::string& n, const float& value) {
    init(sz, n);
#if 0
    for (int i = 0; i < alloc_size; i++) { data[i] = value; }
#else
    FixedVector<float, 16> splat(value);
    for (size_t i = 0; i < alloc_size/16; i++) {
      splat.store_aligned(data + 16 * i);
    }
#endif
  }

  template<typename ME>
  SYRAH_FORCEINLINE const float& operator()(const ME& mo) const {
    cacheSim(&data[mo.ID()]);
    return data[mo.ID()];
  }

  template<typename ME>
  SYRAH_FORCEINLINE float& operator[](const ME& mo) {
    cacheSim(&data[mo.ID()]);
    return data[mo.ID()];
  }

  template<typename ME>
  SYRAH_FORCEINLINE const float& operator[](const ME& mo) const {
    cacheSim(&data[mo.ID()]);
    return data[mo.ID()];
  }

  SYRAH_FORCEINLINE float& internal(size_t which) { return data[which]; }

  template<typename ME, size_t VecSize>
  syrah::FixedVector<float, VecSize> simdRead(const syrah::SIMDObject<VecSize, ME>& mo) const {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[VecSize];
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[mo[i].ID()]);
      local_vals[i] = data[mo[i].ID()];
    }
    return syrah::FixedVector<float, VecSize>(&(local_vals[0]), true);
  }

  template<typename ME, size_t VecSize>
  syrah::FixedVector<float, VecSize> simdReadSequential(const syrah::SIMDObject<VecSize, ME>& mo) const {
    assert(MO::value == ME::type::value);
    // unaligned read directly from memory.
    for (size_t i = 0; i < 16; i++) {
      cacheSim(&data[mo[i].ID()]);
    }
    return syrah::FixedVector<float, VecSize>(&data[mo[0].ID()]);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::FixedVector<float, (int)VecSize>& v) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[VecSize];
    v.store_aligned(local_vals);
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[mo[i].ID()]);
      data[mo[i].ID()] = local_vals[i];
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::FixedVector<float, (int)VecSize>& v) {
    assert(MO::value == ME::type::value);
    v.store(&(data[mo[0].ID()]));
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const float& v) {
    assert(MO::value == ME::type::value);
    syrah::FixedVector<float, (int)VecSize> splat(v);
    splat.store(&(data[mo[0].ID()]));
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::FixedVector<float, (int)VecSize>& v, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[VecSize];
    v.store_aligned(local_vals);
    // TODO(boulos): Figure out best code here.
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[mo[i].ID()]);
      data[mo[i].ID()] = (mask.get(i)) ? local_vals[i] : data[mo[i].ID()];
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::FixedVector<float, (int)VecSize>& v, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    v.store(&(data[mo[0].ID()]), mask);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const float& scalar, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    syrah::FixedVector<float, (int)VecSize> v(scalar);
    v.store(&(data[mo[0].ID()]), mask);
  }


  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWritePlusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::FixedVector<float, (int)VecSize>& v) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[VecSize];
    v.store_aligned(local_vals);
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[mo[i].ID()]);
      data[mo[i].ID()] += local_vals[i];
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWritePlusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::FixedVector<float, (int)VecSize>& v, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[VecSize];
    // use merge to make the += 0
    syrah::FixedVector<float, (int)VecSize> merged = select(v, syrah::FixedVector<float, (int)VecSize>::Zero(), mask);
    merged.store_aligned(local_vals);
    // TODO(boulos): Figure out best code here.
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[mo[i].ID()]);
      data[mo[i].ID()] += local_vals[i];
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteMinusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::FixedVector<float, (int)VecSize>& v) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[VecSize];
    v.store_aligned(local_vals);
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[mo[i].ID()]);
      data[mo[i].ID()] -= local_vals[i];
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteMinusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::FixedVector<float, (int)VecSize>& v, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[VecSize];
    syrah::FixedVector<float, (int)VecSize> merged = select(v, syrah::FixedVector<float, (int)VecSize>::Zero(), mask);
    merged.store_aligned(local_vals);
    // TODO(boulos): Figure out best code here.
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[mo[i].ID()]);
      data[mo[i].ID()] -= local_vals[i];
    }
  }

  // Optimized write scalar to all vals.
  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const float& val) {
    assert(MO::value == ME::type::value);
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[mo[i].ID()]);
      data[mo[i].ID()] = val;
    }
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const float& val, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    // TODO(boulos): Figure out best code here.
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[mo[i].ID()]);
      data[mo[i].ID()] = (mask.get(i)) ? val : data[mo[i].ID()];
    }
  }


  template<typename ME>
  SYRAH_FORCEINLINE void lock(const ME & mo) {
  }

  template<typename ME>
  SYRAH_FORCEINLINE void unlock(const ME & mo) {
  }

  float* data;
  size_t alloc_size;
};
