// Vec3f impl
template<typename MO>
class Field<MO, vec<3, float> > : public FieldBase {
public:
  bool init(size_t sz, const std::string& name) {
    if (initialized) return false;
    FieldBase::init(sz, name, ElemTypeToIO(MO::value));
    size = sz;
    // round up to big enough alloc for vectorization
    alloc_size = (size + 15) & (~15);
    data = new float[3 * alloc_size];
    initialized = true;
    return true;
  }

  void init(size_t sz, const std::string& n, const vec<3, float>& value) {
    init(sz, n);
    float* data_x = data;
    float* data_y = data + alloc_size;
    float* data_z = data + 2 * alloc_size;
#if 0
    for (size_t i = 0; i < alloc_size; i++) { data_x[i] = value[0]; }
    for (size_t i = 0; i < alloc_size; i++) { data_y[i] = value[1]; }
    for (size_t i = 0; i < alloc_size; i++) { data_z[i] = value[2]; }
#else
    FixedVector<float, 16> splat_x(value[0]);
    FixedVector<float, 16> splat_y(value[1]);
    FixedVector<float, 16> splat_z(value[2]);

    for (size_t i = 0; i < alloc_size/16; i++) {
      splat_x.store_aligned(data_x + 16 * i);
    }
    for (size_t i = 0; i < alloc_size/16; i++) {
      splat_y.store_aligned(data_y + 16 * i);
    }
    for (size_t i = 0; i < alloc_size/16; i++) {
      splat_z.store_aligned(data_z + 16 * i);
    }
#endif
  }

  struct VecWriteBack {
    SYRAH_FORCEINLINE VecWriteBack(size_t which, float* data, size_t alloc_size) : which(which), data(data), alloc_size(alloc_size) {
    }

    SYRAH_FORCEINLINE vec<3, float> operator()() {
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);
      return vec<3, float>(data[which], data[alloc_size + which], data[2 * alloc_size + which]);
    }

    // Cast to vec<3, float>
    SYRAH_FORCEINLINE operator vec<3, float>() const {
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);
      return vec<3, float>(data[which], data[alloc_size + which], data[2 * alloc_size + which]);
    }

    // Dimension indexing
    SYRAH_FORCEINLINE float& operator[](int d) {
      cacheSim(&data[alloc_size * d + which]);
      return data[d * alloc_size + which];
    }

    SYRAH_FORCEINLINE const float& operator[](int d) const {
      cacheSim(&data[alloc_size * d + which]);
      return data[d * alloc_size + which];
    }

    // Store vec
    SYRAH_FORCEINLINE VecWriteBack& operator=(const vec<3, float>& vec) {
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[which] = vec[0];
      data[alloc_size + which] = vec[1];
      data[2 * alloc_size + which] = vec[2];
      return *this;
    }

    // +=
    SYRAH_FORCEINLINE VecWriteBack& operator+=(const vec<3, float>& vec) {
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[which] += vec[0];
      data[alloc_size + which] += vec[1];
      data[2 * alloc_size + which] += vec[2];
      return *this;
    }

    // -=
    SYRAH_FORCEINLINE VecWriteBack& operator-=(const vec<3, float>& vec) {
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[which] -= vec[0];
      data[alloc_size + which] -= vec[1];
      data[2 * alloc_size + which] -= vec[2];
      return *this;
    }

    // *= scalar
    SYRAH_FORCEINLINE VecWriteBack& operator*=(const float& scalar) {
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[which] *= scalar;
      data[alloc_size + which] *= scalar;
      data[2 * alloc_size + which] *= scalar;
      return *this;
    }

    // /= scalar
    SYRAH_FORCEINLINE VecWriteBack& operator/=(const float& scalar) {
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[which] /= scalar;
      data[alloc_size + which] /= scalar;
      data[2 * alloc_size + which] /= scalar;
      return *this;
    }


    // Assign from other
    SYRAH_FORCEINLINE VecWriteBack& operator=(const VecWriteBack& other) {
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[which] = other[0];
      data[alloc_size + which] = other[1];
      data[2 * alloc_size + which] = other[2];
      return *this;
    }

    size_t which;
    float* data;
    size_t alloc_size;
  };

  template<typename ME>
  SYRAH_FORCEINLINE VecWriteBack operator()(const ME& mo) const {
    return VecWriteBack(mo.ID(), data, alloc_size);
  }

  template<typename ME>
  SYRAH_FORCEINLINE VecWriteBack operator[](const ME& mo) {
    return VecWriteBack(mo.ID(), data, alloc_size);
  }

  SYRAH_FORCEINLINE VecWriteBack internal(size_t which) {
    return VecWriteBack(which, data, alloc_size);
  }

  template<typename ME>
  SYRAH_FORCEINLINE const vec<3, float> operator[](const ME& mo) const {
    size_t which = mo.ID();
    cacheSim(&data[alloc_size * 0 + which]);
    cacheSim(&data[alloc_size * 1 + which]);
    cacheSim(&data[alloc_size * 2 + which]);

    return vec<3, float>(data[which], data[alloc_size + which], data[2 * alloc_size + which]);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE syrah::SimdVec<3, float, VecSize> simdVecRead(const syrah::SIMDObject<VecSize, ME>& mo) const {
    SYRAH_ALIGN(32) float local_vals[3][VecSize];
#if SYRAH_SIMD_FIELD_STRIPED
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 0 + mo[i].ID()]);
      local_vals[0][i] = data[mo[i].ID()];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 1 + mo[i].ID()]);
      local_vals[1][i] = data[alloc_size + mo[i].ID()];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 2 + mo[i].ID()]);
      local_vals[2][i] = data[2 * alloc_size + mo[i].ID()];
    }
#else
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      local_vals[0][i] = data[0 * alloc_size + which];
      local_vals[1][i] = data[1 * alloc_size + which];
      local_vals[2][i] = data[2 * alloc_size + which];
    }
#endif
    return syrah::SimdVec<3, float, VecSize>(&(local_vals[0][0]), true);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE syrah::SimdVec<3, float, VecSize> simdVecReadSequential(const syrah::SIMDObject<VecSize, ME>& mo) const {
    size_t offset = mo[0].ID();
    const float* striped[3] = { &data[0 * alloc_size + offset],
                                &data[1 * alloc_size + offset],
                                &data[2 * alloc_size + offset] };
    return syrah::SimdVec<3, float, VecSize>(striped);
  }



  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdVec<3, float, (int)VecSize>& v) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[3][VecSize];
    v.store_aligned(&local_vals[0][0]);

#if SYRAH_SIMD_FIELD_STRIPED
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 0 + mo[i].ID()]);
      data[mo[i].ID()] = local_vals[0][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 1 + mo[i].ID()]);
      data[alloc_size + mo[i].ID()] = local_vals[1][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 2 + mo[i].ID()]);
      data[2 * alloc_size + mo[i].ID()] = local_vals[2][i];
    }
#else
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[0 * alloc_size + which] = local_vals[0][i];
      data[1 * alloc_size + which] = local_vals[1][i];
      data[2 * alloc_size + which] = local_vals[2][i];
    }
#endif
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdVec<3, float, (int)VecSize>& v, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[3][VecSize];
    v.store_aligned(&local_vals[0][0]);
#if SYRAH_SIMD_FIELD_STRIPED
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 0 + mo[i].ID()]);
      data[mo[i].ID()] = (mask.get(i)) ? local_vals[0][i] : data[mo[i].ID()];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 1 + mo[i].ID()]);
      data[alloc_size + mo[i].ID()] = (mask.get(i)) ? local_vals[1][i] : data[alloc_size + mo[i].ID()];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 2 + mo[i].ID()]);
      data[2 * alloc_size + mo[i].ID()] = (mask.get(i)) ? local_vals[2][i] : data[2 * alloc_size + mo[i].ID()];
    }
#else
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[0 * alloc_size + which] = (mask.get(i)) ? local_vals[0][i] : data[0 * alloc_size + which];
      data[1 * alloc_size + which] = (mask.get(i)) ? local_vals[1][i] : data[1 * alloc_size + which];
      data[2 * alloc_size + which] = (mask.get(i)) ? local_vals[2][i] : data[2 * alloc_size + which];
    }
#endif
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdVec<3, float, (int)VecSize>& v) {
    size_t offset = mo[0].ID();
    float* striped[3] = { &data[0 * alloc_size + offset],
                          &data[1 * alloc_size + offset],
                          &data[2 * alloc_size + offset] };
    v.store(striped);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdVec<3, float, (int)VecSize>& v, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    size_t offset = mo[0].ID();
    float* striped[3] = { &data[0 * alloc_size + offset],
                          &data[1 * alloc_size + offset],
                          &data[2 * alloc_size + offset] };
    v.store(striped, mask);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const vec<3, float>& vec){
    size_t offset = mo[0].ID();
    float* striped[3] = { &data[0 * alloc_size + offset],
                          &data[1 * alloc_size + offset],
                          &data[2 * alloc_size + offset] };
    syrah::SimdVec<3, float, (int)VecSize> v(vec);
    v.store(striped);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteSequential(const syrah::SIMDObject<VecSize, ME>& mo, const vec<3, float>& vec, const syrah::FixedVectorMask<(int)VecSize>& mask){
    size_t offset = mo[0].ID();
    float* striped[3] = { &data[0 * alloc_size + offset],
                          &data[1 * alloc_size + offset],
                          &data[2 * alloc_size + offset] };
    syrah::SimdVec<3, float, (int)VecSize> v(vec);
    v.store(striped, mask);
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWritePlusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdVec<3, float, (int)VecSize>& v) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[3][VecSize];
    v.store_aligned(&local_vals[0][0]);
#if SYRAH_SIMD_FIELD_STRIPED
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 0 + mo[i].ID()]);
      data[mo[i].ID()] += local_vals[0][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 1 + mo[i].ID()]);
      data[alloc_size + mo[i].ID()] += local_vals[1][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 2 + mo[i].ID()]);
      data[2 * alloc_size + mo[i].ID()] += local_vals[2][i];
    }
#else
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[0 * alloc_size + which] += local_vals[0][i];
      data[1 * alloc_size + which] += local_vals[1][i];
      data[2 * alloc_size + which] += local_vals[2][i];
    }
#endif
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWritePlusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdVec<3, float, (int)VecSize>& v, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[3][VecSize];
    // use merge to make the += 0
    syrah::SimdVec<3, float, (int)VecSize> merged = select(v, syrah::SimdVec<3, float, (int)VecSize>::Zero(), mask);
    merged.store_aligned(&local_vals[0][0]);

#if SYRAH_SIMD_FIELD_STRIPED
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 0 + mo[i].ID()]);
      data[mo[i].ID()] += local_vals[0][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 1 + mo[i].ID()]);
      data[alloc_size + mo[i].ID()] += local_vals[1][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 2 + mo[i].ID()]);
      data[2 * alloc_size + mo[i].ID()] += local_vals[2][i];
    }
#else
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[0 * alloc_size + which] += local_vals[0][i];
      data[1 * alloc_size + which] += local_vals[1][i];
      data[2 * alloc_size + which] += local_vals[2][i];
    }
#endif
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteMinusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdVec<3, float, (int)VecSize>& v) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[3][VecSize];
    v.store_aligned(&local_vals[0][0]);
#if SYRAH_SIMD_FIELD_STRIPED
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 0 + mo[i].ID()]);
      data[mo[i].ID()] -= local_vals[0][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 1 + mo[i].ID()]);
      data[alloc_size + mo[i].ID()] -= local_vals[1][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 2 + mo[i].ID()]);
      data[2 * alloc_size + mo[i].ID()] -= local_vals[2][i];
    }
#else
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[0 * alloc_size + which] -= local_vals[0][i];
      data[1 * alloc_size + which] -= local_vals[1][i];
      data[2 * alloc_size + which] -= local_vals[2][i];
    }
#endif
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWriteMinusEq(const syrah::SIMDObject<VecSize, ME>& mo, const syrah::SimdVec<3, float, (int)VecSize>& v, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
    SYRAH_ALIGN(32) float local_vals[3][VecSize];
    // use merge to make the += 0
    syrah::SimdVec<3, float, (int)VecSize> merged = select(v, syrah::SimdVec<3, float, (int)VecSize>::Zero(), mask);
    merged.store_aligned(&local_vals[0][0]);
#if SYRAH_SIMD_FIELD_STRIPED
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 0 + mo[i].ID()]);
      data[mo[i].ID()] -= local_vals[0][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 1 + mo[i].ID()]);
      data[alloc_size + mo[i].ID()] -= local_vals[1][i];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 2 + mo[i].ID()]);
      data[2 * alloc_size + mo[i].ID()] -= local_vals[2][i];
    }
#else
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[0 * alloc_size + which] -= local_vals[0][i];
      data[1 * alloc_size + which] -= local_vals[1][i];
      data[2 * alloc_size + which] -= local_vals[2][i];
    }
#endif
  }

  // scalar writes
  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const vec<3, float>& vec) {
    assert(MO::value == ME::type::value);
#if SYRAH_SIMD_FIELD_STRIPED
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 0 + mo[i].ID()]);
      data[mo[i].ID()] = vec[0];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 1 + mo[i].ID()]);
      data[alloc_size + mo[i].ID()] = vec[1];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 2 + mo[i].ID()]);
      data[2 * alloc_size + mo[i].ID()] = vec[2];
    }
#else
    for (size_t i = 0; i < VecSize; i++) {
      size_t which = mo[i].ID();
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[0 * alloc_size + which] = vec[0];
      data[1 * alloc_size + which] = vec[1];
      data[2 * alloc_size + which] = vec[2];
    }
#endif
  }

  template<typename ME, size_t VecSize>
  SYRAH_FORCEINLINE void simdWrite(const syrah::SIMDObject<VecSize, ME>& mo, const vec<3, float>& vec, const syrah::FixedVectorMask<(int)VecSize>& mask) {
    assert(MO::value == ME::type::value);
#if SYRAH_SIMD_FIELD_STRIPED
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 0 + mo[i].ID()]);
      data[mo[i].ID()] = (mask.get(i)) ? vec[0] : data[mo[i].ID()];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 1 + mo[i].ID()]);
      data[alloc_size + mo[i].ID()] = (mask.get(i)) ? vec[1] : data[alloc_size + mo[i].ID()];
    }
    for (size_t i = 0; i < VecSize; i++) {
      cacheSim(&data[alloc_size * 2 + mo[i].ID()]);
      data[2 * alloc_size + mo[i].ID()] = (mask.get(i)) ? vec[2] : data[2 * alloc_size + mo[i].ID()];
    }
#else
    for (size_t i = 0; i < VecSize; i++) {
      if (!mask.get(i)) continue;
      size_t which = mo[i].ID();
      cacheSim(&data[alloc_size * 0 + which]);
      cacheSim(&data[alloc_size * 1 + which]);
      cacheSim(&data[alloc_size * 2 + which]);

      data[0 * alloc_size + which] = vec[0];
      data[1 * alloc_size + which] = vec[1];
      data[2 * alloc_size + which] = vec[2];
    }
#endif
  }

  template<typename ME>
  SYRAH_FORCEINLINE void lock(const ME & mo) {
  }

  template<typename ME>
  SYRAH_FORCEINLINE void unlock(const ME & mo) {
  }

  float* data;
  size_t alloc_size;
};
