#include <pthread.h>
#include <fstream>
#include "CRSMesh/CRSMesh.h"

using CRSMesh::Vertex;
using CRSMesh::Edge;
using CRSMesh::Face;
using CRSMesh::Cell;

typedef CRSMesh::Mesh GlobalMesh;

#include "Vector.h"
#include "ForAll.h"
#include <stdio.h>
#include <sys/time.h>
#include "MeshIO/FacetEdgeBuilder.h"
#include "MeshIO/LisztFileReader.h"
#include "SMP/Field.h"
#include "Layout/BoundarySetBuilder.h"
#include "SMP/MeshWrapper.h"
#include "SMP/BoundarySetWrapper.h"
#include "SMP/Synchronization.h"
#include "Layout/Partitioner.h"
#include "Liszt/BoundarySet.h"
#include "PrintContext.h"
#include "Liszt/Time.h"
#include "cache_sim.h"

SimpleCache* FieldBase::cache_sim = NULL;

typedef CRSMesh::Mesh Mesh;
typedef Mesh::vertex_set vertex_set;
typedef Mesh::edge_set edge_set;
typedef Mesh::face_set face_set;
typedef Mesh::cell_set cell_set;

using LisztPrivate::BoundarySet;

#define COLORING_BLOCK_SIZE (2048)
#define WORKGROUP_WORK_SIZE (2048)
#define MAX_THREAD_SIZE (128)

typedef Mesh mesh_type;
#include "ScalaStdLib.h"
typedef GlobalMesh ColoringMesh; //needed for coloring controller
#include "InterferenceGraph/ColoringController.h"


//Codegen subclasses this as its entry point after runtime initialization.
class LisztThread {
public:
        void init(const GlobalMesh * mesh, size_t nt, size_t tid, std::ostream * log) {
                num_threads = nt;
                thread_id = tid;
                log_out = log;
        }
        virtual void toplevel() = 0;
        virtual void main() = 0;
        virtual ~LisztThread() {}
        size_t num_threads;
        size_t thread_id;
        PrintContext print_context;
        std::ostream * log_out;
};
LisztThread * LisztClient_create();


struct RotateFilter {
        void init(size_t nv, size_t ne, size_t nf, size_t nc) {
                sz[MeshIO::VERTEX_T] = nv;
                sz[MeshIO::EDGE_T] = ne;
                sz[MeshIO::FACE_T] = nf;
                sz[MeshIO::CELL_T] = nc;
        }
        MeshIO::id_t operator()(MeshIO::IOElemType typ, MeshIO::id_t id) {
                if(id != 0) {
                        return ( id % (sz[typ] - 1) ) + 1;
                } else return 0;
        }
        MeshIO::id_t reverse(MeshIO::IOElemType typ, MeshIO::id_t id) {
                if(id != 0) {
                        if(id == 1)
                                return sz[typ] - 1;
                        else
                                return id - 1;
                } else return 0;
        }
        size_t sz[4];
};
typedef MeshIO::ResultFilter filter_t;

namespace System {

MeshIO::LisztFileReader mesh_file;
BoundarySetBuilder boundary_builder;
GlobalMesh global_mesh;
size_t num_threads;
Mesh local_mesh;
pthread_mutex_t boundary_load_l;
bool redirect_to_log;

filter_t partitionRenameFilter;
bool partition_and_block;
liszt_barrier_t barrier;
Synchronization::ColoringController coloring_controller;

void loadMesh() {
        MeshIO::FileFacetEdge * ffe = mesh_file.facetEdges();
        MeshIO::FacetEdgeBuilder builder;
        const MeshIO::LisztHeader & h = mesh_file.header();
        builder.init(h.nV,h.nE,h.nF,h.nC,h.nFE);
        builder.insert(0,h.nFE,ffe);

        if(partition_and_block) {
                MeshIO::Partitioner partitioner;
                partitioner.init(&builder);
                partitioner.partition(256,&partitionRenameFilter);
                //MeshIO::Blocker blocker;
                //blocker.block(&builder,1024,&partitionRenameFilter);
        }

        //TODO(zach): calculating blocking and partitioning would go here before
        //builder.filter is called

        if(partition_and_block)
                builder.filter(partitionRenameFilter);

        builder.validateFullMesh();
        mesh_file.free(ffe);
        global_mesh.initFromFacetEdgeBuilder(&builder);
        boundary_builder.init(h.nBoundaries,mesh_file.boundaries());
}


template<typename MO,typename VT>
void loadFieldConst(Field<MO,VT> * field, const std::string & name, const VT & value) {
        field->lock_all();
        size_t sz;
        switch (MO::value) {
                case VERTEX: sz = mesh_file.header().nV; break;
                case EDGE: sz = mesh_file.header().nE; break;
                case FACE: sz = mesh_file.header().nF; break;
                case CELL: sz = mesh_file.header().nC; break;
        }
        field->init(sz,name,value);
        field->unlock_all();
}
template<typename VT>
void loadPositions(Field<VertexType,VT> * field, const std::string & name) {
        field->lock_all();
        if(field->init(mesh_file.header().nV,name)) {
                MeshIO::PositionTable * table = mesh_file.positions();
                for(size_t i = 0; i < mesh_file.header().nV; i++) {
                        size_t entry = (partition_and_block) ? partitionRenameFilter(MeshIO::VERTEX_T,i) : i;
                        field->internal(entry)[0] = table->data[i][0];
                        field->internal(entry)[1] = table->data[i][1];
                        field->internal(entry)[2] = table->data[i][2];
                }
                mesh_file.free(table);
        }
        field->unlock_all();
}
template<typename MO>
void loadBoundary(LisztThread * thread, const char * name,
                  BoundarySet< typename MeshObjectForType<GlobalMesh,MO>::set > * boundary,
                  BoundarySetWrapper< BoundarySet< typename MeshObjectForType<GlobalMesh,MO>::set > > * boundary_wrapper) {

        pthread_mutex_lock(&boundary_load_l);
        if(partition_and_block)
                System::boundary_builder.loadWithFilter<GlobalMesh,MO>(&global_mesh,name,partitionRenameFilter,boundary);
        else
                System::boundary_builder.load<GlobalMesh,MO>(&global_mesh,name,boundary);
        pthread_mutex_unlock(&boundary_load_l);
        if(boundary_wrapper && thread)
                boundary_wrapper->init(boundary, thread->num_threads, thread->thread_id);
}

template<typename MeshSet>
BoundarySet< MeshSet > & loadBoundaryOnce(const char * name, BoundarySet< MeshSet > * dest) {
        loadBoundary<typename MeshSet::value_type::type>(NULL,name,dest,NULL);
        return *dest;
}

void * thread_entry(void * data) {
        size_t id = reinterpret_cast<size_t>(data);
        std::ostream * out = &std::cout;
        if(redirect_to_log) {
                std::ostringstream ss;
                ss << "liszt_runtime_" << id << ".log";
                out = new std::ofstream(ss.str().c_str());
                if(!out->good()) {
                        std::cerr << "Error opening log file: " << ss.str() << std::endl;
                        exit(1);
                }
        }
        LisztThread * thread = LisztClient_create();
        thread->init(&global_mesh, num_threads, id, out);
        thread->toplevel();
        double start_time = wall_time();

        // Now that the mesh has been loaded, create the cache
        FieldBase::cache_sim = new SimpleCache();

        thread->main();
        double end_time = wall_time();
        std::cout << "TIME_FOR_MAIN: " << end_time - start_time << std::endl;

        delete thread;
        if(redirect_to_log) {
                out->flush();
                delete out;
        }
        return NULL;
}


int run(int argc, char ** argv) {
        if(argc != 5) {
                printf("usage: %s <mesh_file> <num_threads> <redirect_to_log> <partition_and_block>\n",argv[0]);
                return 1;
        }

        const char * mesh_file_name = argv[1];
        num_threads = atoi(argv[2]);
        redirect_to_log = atoi(argv[3]) ? true : false;
        partition_and_block = atoi(argv[4]) ? true : false;

        pthread_mutex_init(&boundary_load_l,NULL);
        barrier_init(&barrier,num_threads);
        mesh_file.init(mesh_file_name);
        loadMesh();
        coloring_controller.init(global_mesh, COLORING_BLOCK_SIZE);
        pthread_t * threads = new pthread_t[num_threads];
        for(unsigned int i = 0; i < num_threads; i++) {
                pthread_create(&threads[i],NULL,thread_entry,reinterpret_cast<void *>(i));
        }
        for(unsigned int i = 0; i < num_threads; i++) {
                pthread_join(threads[i],NULL);
        }
        delete [] threads;

        FieldBase::cache_sim->printStats();
        return 0;
}


} //namespace System

static inline id_type ID_v(const Vertex & v) { return (System::partition_and_block) ? System::partitionRenameFilter.reverse(MeshIO::VERTEX_T,v.ID()) : v.ID(); }
static inline id_type ID_e(const Edge & e) { return (System::partition_and_block) ?  System::partitionRenameFilter.reverse(MeshIO::EDGE_T,e.ID()) : e.ID(); }
static inline id_type ID_f(const Face & f) { return (System::partition_and_block) ?  System::partitionRenameFilter.reverse(MeshIO::FACE_T,f.ID()) : f.ID(); }
static inline id_type ID_c(const Cell & c) { return (System::partition_and_block) ? System::partitionRenameFilter.reverse(MeshIO::CELL_T,c.ID()) : c.ID(); }

static inline const Mesh * mesh() {
        return &System::global_mesh;
}

int main(int argc, char ** argv) {
        return System::run(argc,argv);
}

#include "SMP/Global.h"

//MACROS NEEDED FOR CODE GEN
//we need to load the boundary sets before coloring, loadBoundary ensures that a boundary set is only ever loaded once,
//but we need to combine the load and first access into a single call, which is what COLORING_GET_BOUNDARY does here
#define COLORING_GET_BOUNDARY(boundary,set_name) System::loadBoundaryOnce( set_name , &boundary)

static inline int fetch_and_add(int * data, int value) {
        return __sync_fetch_and_add(data,value);
}

#ifdef PROJECT_DEBUG
#define PRINT_CONTEXT_BEGIN print_context.push();
#define PRINT_CONTEXT_END print_context.pop();
#define PRINT_CONTEXT_MARK(v) print_context.mark(v);
#else
#define PRINT_CONTEXT_BEGIN
#define PRINT_CONTEXT_END
#define PRINT_CONTEXT_MARK(v)
#endif

void Workgroup_getSize(int total_size, int num_parts, int part, int * begin, int * size) {
        int part_size = total_size / COLORING_BLOCK_SIZE / num_parts;
        *begin = part_size * COLORING_BLOCK_SIZE * part;
        if(part + 1 == num_parts)
                *size = total_size - *begin;
        else
                *size = part_size * COLORING_BLOCK_SIZE;
}

#define WORKGROUP_FORALL_BEGIN(wid,loop_var,set_value_type,set_var,trivial) \
{ \
/*TOTAL HACK: getCRSColoring is not thread safe if it hasn't been initialized but is thread safe if it has been, \
  we just have thread 0 initialize it first by calling it before other things have */ \
::Layout::CRS * __crs; \
static bool initialized_##wid = false; \
static int next_start_##wid [MAX_THREAD_SIZE]; \
if(!initialized_##wid) { \
        if(this->thread_id == 0) { \
                System::coloring_controller.getCRSColoring(wid,set_var,trivial,&__crs); \
                printf("initializing %s\n", #wid); \
                std::fill(next_start_##wid,next_start_##wid +  this->num_threads,0); \
        } \
        barrier_wait(&System::barrier);\
        initialized_##wid = true; \
} \
PRINT_CONTEXT_BEGIN \
int ncolors = System::coloring_controller.getCRSColoring(wid,set_var,trivial,&__crs); \
\
for(int color = 0; color < ncolors; color++) { \
        size_t __color_start = __crs->row_idx[color]; \
        size_t __color_size = __crs->row_idx[color+1] - __color_start; \
        int __steal_from = this->thread_id; \
        for(int __num_finished = 0; __num_finished < this->num_threads; __num_finished++) {  \
                if(__steal_from >= this->num_threads) \
                        __steal_from = 0; \
                int __start,__size; \
                Workgroup_getSize(__color_size,this->num_threads,__steal_from,&__start,&__size); \
                int __blk_start = __sync_fetch_and_add(&next_start_##wid [__steal_from],WORKGROUP_WORK_SIZE); \
                while(__blk_start < __size) { \
                        for(int __i = 0; __i < WORKGROUP_WORK_SIZE && __blk_start + __i < __size; __i++) { \
                                set_value_type loop_var = System::global_mesh.get<set_value_type>(__crs->values[__color_start + __start + __blk_start + __i]); \
                                PRINT_CONTEXT_MARK(loop_var)

#define WORKGROUP_FORALL_END(wid,loop_var,set_value_type,set_var,trivial) \
                        } \
                        __blk_start = fetch_and_add(&next_start_##wid [__steal_from],WORKGROUP_WORK_SIZE); \
                } \
                __steal_from++; \
        } \
        if(barrier_begin(&System::barrier)) { \
                std::fill(next_start_##wid,next_start_##wid +  this->num_threads,0); \
                barrier_end(&System::barrier); \
        } \
} \
PRINT_CONTEXT_END \
\
}

