/*
 * FieldDisplay.h
 *
 *  Created on: Nov 19, 2009
 *      Author: zdevito
 */

#ifndef FIELDDISPLAY_H_
#define FIELDDISPLAY_H_
#include "SingleCore/IObject.h"
#include <set>
//a class to define how fields of arbitrary template types map to scalar, 3 vector, or categorical values.
using namespace LisztPrivate::ElemTypes;

#include "FieldDisplayDecl.h"

template<typename MO>
IObject FieldDisplay::operator()(const MO & mo) const {
    assert( MO::type::value == this->fieldDomain());
    return this->atID(mo.ID());
}

template<typename T>
struct FieldEntryInfo {
    static std::string typeName()  { return "unknown"; }
    static IObject createObj(const T & t) {
        return IObject();
    }
};

template<>
struct FieldEntryInfo<int> {
    static std::string typeName()  { return "int"; }
    static IObject createObj(int i) {
        return IObject(i);
    }
};


template<>
struct FieldEntryInfo<float> {
    static IObject createObj(float i) {
        return IObject((double)i);
    }
    static std::string typeName()  { return "float"; }
};

template<>
struct FieldEntryInfo<double> {
    static std::string typeName()  { return "double"; }
    static IObject createObj(double i) {
        return IObject(i);
    }
};

template<>
struct FieldEntryInfo<bool> {
    static std::string typeName()  { return "bool"; }
    static IObject createObj(bool b) {
        return IObject(b ? 1 : 0);
    }
};

template<size_t N,typename T>
struct FieldEntryInfo< vec<N,T> > {
    static IObject createObj(const vec<N,T> & r) {
        std::vector<IObject> v(N);
        for(size_t i = 0; i < N; i++) {
            v[i] = FieldEntryInfo<T>::createObj(r[i]);
        }
        return IObject(v);
    }
    static std::string typeName()  {
        std::string v = FieldEntryInfo<T>::typeName();
        std::stringstream ss;
        ss << "vec<" << N << "," << v << ">";
        return ss.str();
    }
};


template<typename MO, typename VT>
class FieldDisplayImpl : public FieldDisplay {
    typedef MO elem_type;
    typedef VT value_type;
public:
    FieldDisplayImpl(size_t n_snapshots, const SingleField::Field<elem_type,value_type> * field_)
    : real_field(field_) {
    	reset();
    	this->reserveSnapshots(n_snapshots);
    }
	virtual std::string fieldRangeString() const {
        return FieldEntryInfo<value_type>::typeName();
    }
    virtual ElemType fieldDomain() const {
        return elem_type::value;
    }
    virtual std::string name() const {
        return real_field->name();
    }
    //TODO: this needs to be changed to have a most recent retrival mode
    virtual IObject atID(id_type id) const {
    	if(display_mode == DISPLAY_CURRENT) {
    		return FieldEntryInfo<value_type>::createObj(*real_field->getReadEntry(id));
    	}
    	else {
    		assert(cur_field.get());
    		return FieldEntryInfo<value_type>::createObj(*cur_field->getReadEntry(id));
    	}
    }
    virtual size_t snapshot() {
		if(real_field->isDirty()) {
			field_type_ref field(new field_type());
			(*field) = *real_field;
			field_map.push_back(field);
			real_field->clean();
		} else {
			assert(field_map.size() > 0);
			field_map.push_back(field_map.back());
		}
		return field_map.size();
    }
    virtual void setSnapshot(size_t i) {
    	assert(i < field_map.size());
    	cur_field = field_map[i];
    }
    virtual void reset() {
    	display_mode = DISPLAY_CURRENT;
    	field_map.clear();
    	field_map.push_back(field_type_ref(new field_type()));
    	*field_map.back() = *real_field;
    	cur_field = field_map.back();
    }
    virtual void setDisplayMode(DisplayMode dm) {
    	display_mode = dm;
    }
protected:
	virtual void reserveSnapshots(size_t t) {
    	assert(field_map.size() > 0);
    	while(field_map.size() < t) {
    		field_map.push_back(field_map.back());
    	}
    }
    virtual size_t size() const {
        return real_field->getSize();
    }
private:
    typedef SingleField::Field<elem_type,value_type> field_type;
    typedef ::std::tr1::shared_ptr<field_type> field_type_ref;
    typedef std::vector< field_type_ref > field_map_type;
    const field_type * real_field;
    field_type_ref cur_field;
    field_map_type field_map;
    DisplayMode display_mode;
};


/*
double doubleMax() const {
    //assert(hasDoubleValue());
    size_t len = size();
    double max = std::numeric_limits<double>::min();
    for(id_type i = 0; i < len; i++) {
        double v = atID(i)->doubleValue();
        if(v > max)
            max = v;
    }
    return max;
}
double doubleMin() const {
    //assert(hasDoubleValue());
    size_t len = size();
    double min = std::numeric_limits<double>::max();
    for(id_type i = 0; i < len; i++) {
        double v = atID(i)->doubleValue();
        if(v < min)
            min = v;
    }
    return min;
}
double doubleMean() const {
    //assert(hasDoubleValue());
    size_t len = size();
    double mean = 0.;
    for(id_type i = 0; i < len; i++) {
        double v = atID(i)->doubleValue();
        mean += v;
    }
    mean /= len;
    return mean;
}
double doubleStdDev() const {
    double mean = doubleMean();
    size_t len = size();
    double std = 0.;
    for(id_type i = 0; i < len; i++) {
        double v = atID(i)->doubleValue() - mean;
        std += v * v;
    }
    return sqrt(std);
}
void intSet(std::set<int> * s) const {
    //assert(hasIntValue());
    size_t len = size();
    for(id_type i = 0; i < len; i++) {
        s->insert(atID(i)->intValue());
    }
}
void stringSet(std::set<std::string> * s) const {
    //assert(hasStringValue());
    size_t len = size();
    for(id_type i = 0; i < len; i++) {
        s->insert(atID(i)->stringValue());
    }
}*/

#endif /* FIELDDISPLAY_H_ */
