#ifndef IOBJECT_H
#define IOBJECT_H
#include <cassert>
#include <ctype.h>
#include <stdlib.h>
#include <limits>
#include <iostream>
#include <vector>
#include "math.h"
#include "liszt_memory.h"

#include "FieldDisplayDecl.h"

#include "interpret.h"

enum IType { TINT, TDOUBLE, TVEC,TLAMBDA, TFIELD_VEC, TEXTERN_LAMBDA, TUNKNOWN};

static inline std::ostream & operator<<(std::ostream & out, const IObject & o);

struct IObject {
private:
	IType type;
public:
	typedef std::vector<IObject> vec_type;
        typedef IObject (*extern_func_type)(const IObject &);
        
	struct {
		double d;
		int i;
		vec_type v;
		ClosureRef l;
		::std::tr1::shared_ptr<FieldDisplay> f;
		extern_func_type func;        
	} data;
	IObject() {
		type = TUNKNOWN;
	}
	IObject(double d) {
		type = TDOUBLE;
		data.d = d;
	}
	IObject(int i) {
		type = TINT;
		data.i = i;
	}
	IObject(const ClosureRef er) {
		type = TLAMBDA;
		data.l = er;
	}
	template<typename T>
	IObject(const std::vector<T> & v) {
		type = TVEC;
		new (&data.v) vec_type(v.size());
		for(unsigned int i = 0; i < v.size(); i++) {
			data.v[i] = IObject(v[i]);
		}
	}
	IObject(const vec_type & v_) {
		type = TVEC;
		data.v = v_;
	}
	IObject(::std::tr1::shared_ptr<FieldDisplay> fd) {
		type = TFIELD_VEC;
		data.f = fd;
	}
        IObject(extern_func_type f) {
            type = TEXTERN_LAMBDA;
            data.func = f;
        }
	IType getType() const {
		return type;
	}
	bool isNumber() const {
		return type == TINT || type == TDOUBLE;
	}
	bool isInt() const {
		return type == TINT || type == TDOUBLE;
	}
	bool isDouble() const {
		return type == TINT || type == TDOUBLE;
	}
	bool isVec(size_t min_length = 0) const {
		return (type == TVEC && data.v.size() >= min_length) || (type == TFIELD_VEC && data.f->size() >= min_length);
	}
	bool isUnknown() const  {
		return type == TUNKNOWN;
	}
	bool isNaN() const {
	    if(type == TDOUBLE)
	        return isnan(doubleValue());
	    else if(isVec()) {
	        for(size_t i = 0; i < size(); i++) {
	            if(getEntry(i).isNaN())
	                return true;
	        }
	        return false;
	    } else return false;
	}
	int intValue() const {
		if(type == TINT)
			return data.i;
		else if(type == TDOUBLE)
			return (int) data.d;
		else
			assert(false);
		return 0;
	}
	double doubleValue() const {
		if(type == TINT)
			return (double) data.i;
		else if(type == TDOUBLE)
			return data.d;
		else
			assert(false);
		return 0.;
	}
	vec_type vecValue() const {
		if(type == TVEC) {
			return data.v;
		} else if(type == TFIELD_VEC) {
			vec_type ret(size());
			for(size_t i = 0; i < size(); i++)
				ret[i] = getEntry(i);
			return ret;
		} else { 
			std::cout << type;
			assert(false);
		}
		return vec_type(1);
	}
	int intValue(int e) const {
		if(!isInt())
			return e;
		else
			return intValue();
	}
	double doubleValue(double e) const {
		if(!isDouble())
			return e;
		else
			return doubleValue();
	}
	
	vec_type vecValue(const vec_type & e) const {
		if(!isVec())
			return e;
		else
			return vecValue();
	}
	IObject getEntry(size_t i) const {
		if(isVec(i+1)) {
			if(type == TVEC)
				return data.v[i];
			else
				return data.f->atID(i);
		} else return IObject();
	}
	size_t size() const {
		if(type == TVEC) 
			return data.v.size();
		else if(type == TFIELD_VEC)
			return data.f->size();
		else return 1;
	}
	IObject operator -() const {
		if(type == TINT)
			return IObject(-data.i);
		else if(type == TDOUBLE)
			return IObject(-data.d);
		else if(isVec()) {
			vec_type r = vecValue();
			for(size_t i = 0; i < r.size(); i++)
				r[i] = -r[i];
			return IObject(r);
		} else return IObject();
	}
	IObject operator !() const {
		if(type == TINT)
			return IObject(!data.i);
		else if(type == TDOUBLE)
			return IObject(!data.d);
		else if(isVec()) {
			vec_type r = vecValue();
			for(size_t i = 0; i < r.size(); i++)
				r[i] = !r[i];
			return IObject(r);
		} else return IObject();
	}
    IObject operator() (const IObject & rhs) const {
        static int nesting_depth = 0;
        nesting_depth++;
        if(nesting_depth > 100) { //stack depth limit reached
            nesting_depth--;
            return IObject();
        }
        IObject ret = apply_internal(rhs);
        nesting_depth--;
        return ret;
    }
private:
	IObject apply_internal(const IObject & rhs) const {
		if(isVec()) {
			if(rhs.type == TINT) {
				int v = rhs.intValue();
				return getEntry(v);
			} else if(rhs.isVec()) {
				const vec_type & v = rhs.data.v;
				vec_type r(v.size());
				for(size_t i = 0; i < v.size(); i++) {
					r[i] = (*this)(v[i]);
					if(r[i].isUnknown())
						return r[i];
				}
				return IObject(r);
			} else return IObject();
		} else if(getType() == TLAMBDA) {
			data.l->context["_"] = rhs;
			return data.l->exp->eval(data.l->context);
        } else if(getType() == TEXTERN_LAMBDA) {
            return data.func(rhs);
		} else return IObject();
	}
};

template<typename T>
struct ConformingOp {
	static inline IObject apply(const IObject & a, const IObject & b) {
		if(a.isUnknown()) return a;
		if(b.isUnknown()) return b;
		//TODO then v.data patterns
		switch(a.isVec() << 1 | b.isVec()) {
				case 3: {
					if(a.size() != b.size()) return IObject();
					IObject::vec_type rv(a.size());
					for(size_t i = 0; i < a.size(); i++) {
						rv[i] = ConformingOp<T>::apply(a.getEntry(i),b.getEntry(i));
					}
					return IObject(rv);
				}
				case 2: {
					IObject::vec_type av = a.vecValue();
					IObject::vec_type rv(av.size());
					for(size_t i = 0; i < av.size(); i++)
						rv[i] = ConformingOp<T>::apply(av[i],b);
					return IObject(rv);
				}
				case 1: {
					IObject::vec_type bv = b.vecValue();
					IObject::vec_type rv(bv.size());
					for(size_t i = 0; i < bv.size(); i++)
						rv[i] = ConformingOp<T>::apply(a,bv[i]);
					return IObject(rv);
				}
				case 0: {
					return T::apply(a,b);
				}
			}
			assert(false);
			return IObject();
	}
};

#define MAKE_OP(name,op) \
struct IOp_##name { \
    static IObject apply(const IObject & a, const IObject & b) { \
        if(a.getType() == TDOUBLE || b.getType() == TDOUBLE) \
              return IObject(a.doubleValue() op b.doubleValue()); \
        else if(a.getType() == TINT || b.getType() == TINT) \
              return IObject(a.intValue() op b.intValue()); \
		else return IObject(); \
    } \
}; \
static inline IObject operator op(const IObject & a, const IObject & b) { \
    return ConformingOp<IOp_##name>::apply(a,b); \
}

#define MAKE_OP_NO_ZERO_IMPL(name,op,dbl) \
struct IOp_##name { \
    static IObject apply(const IObject & a, const IObject & b) { \
        if(a.getType() == TDOUBLE || b.getType() == TDOUBLE) {\
              double d = b.dbl(); \
              if(d == 0.) \
                  return IObject(); \
              else \
                  return IObject(a.dbl() op b.dbl()); \
         } else if(a.getType() == TINT || b.getType() == TINT) {\
              int d = b.intValue(); \
              if(d == 0) \
                  return IObject(); \
              else \
                  return IObject(a.intValue() op d); \
          } else return IObject(); \
    } \
}; \
static inline IObject operator op(const IObject & a, const IObject & b) { \
    return ConformingOp<IOp_##name>::apply(a,b); \
}

MAKE_OP(mult,*)
MAKE_OP_NO_ZERO_IMPL(div,/,doubleValue)
MAKE_OP_NO_ZERO_IMPL(mod,%,intValue)
MAKE_OP(plus,+)
MAKE_OP(minus,-)
MAKE_OP(band,&&)
MAKE_OP(bor,||)

static inline IObject operator,(IObject a, IObject b) {
    if(a.isUnknown()) return a;
    if(b.isUnknown()) return b;
    IObject::vec_type alist(1,a);
    a = IObject(alist);
    a.data.v.push_back(b);
    return a;
}
static inline std::ostream & operator<<(std::ostream & out, const IObject & o) {
	if(o.getType() == TINT)
		out << o.intValue();
	else if(o.getType() == TDOUBLE)
		out << o.doubleValue();
	else if(o.isVec()) {
		out << "[";
	 	IObject::vec_type v = o.vecValue();
		for(size_t i = 0; i < v.size(); i++) {
			out << v[i] << " ";
		}
		out << "]";
	} else if(o.getType() == TLAMBDA) {
                out << "lambda";
        } else if(o.getType() == TEXTERN_LAMBDA) {
                out << "built-in";
        } else
		out << "None";
	return out;
}
#undef MAKE_OP
#undef MAKE_VEC_OP

struct IOp_min {
    static IObject apply(const IObject & a, const IObject & b) {
		if(a.getType() == TDOUBLE || b.getType() == TDOUBLE) {
            return IObject(std::min(a.doubleValue(),b.doubleValue()));
        } else if(a.getType() == TINT || b.getType() == TINT) {
            return IObject(std::min(a.intValue(),b.intValue()));
		} else return IObject();
    }
};
struct IOp_max {
    static IObject apply(const IObject & a, const IObject & b) {
        if(a.getType() == TDOUBLE || b.getType() == TDOUBLE) {
            return IObject(std::max(a.doubleValue(),b.doubleValue()));
        } else if(a.getType() == TINT || b.getType() == TINT) {
            return IObject(std::max(a.intValue(),b.intValue()));
        } else return IObject();
    }
};
struct IObjectLib {
    static IObject sqrt(const IObject & o) {
        if(o.isDouble()) {
            return IObject(::sqrt(o.doubleValue()));
        } else if(o.isVec()) {
            IObject::vec_type r = o.vecValue();
            for(size_t i = 0; i < r.size(); i++) {
                r[i] = IObjectLib::sqrt(r[i]);
                if(r[i].isUnknown())
                    return r[i];
            }
            return IObject(r);
        } else return IObject();
    }
    static IObject cross(const IObject & arg) {
        if(!arg.isVec(2))
            return IObject();
        IObject a = arg.getEntry(0);
        IObject b = arg.getEntry(1);
        if(!a.isVec(3) || !b.isVec(3))
            return IObject();
        IObject::vec_type ret(3);
        ret[0] = a.getEntry(1) * b.getEntry(2) - a.getEntry(2) * b.getEntry(1);
        ret[1] = a.getEntry(2) * b.getEntry(0) - a.getEntry(0) * b.getEntry(2);
        ret[2] = a.getEntry(0) * b.getEntry(1) - a.getEntry(1) * b.getEntry(0);
        return IObject(ret);
    }
    static IObject min(const IObject & a, const IObject & b) {
        return ConformingOp<IOp_min>::apply(a,b);
    }
    static IObject max(const IObject & a, const IObject & b) {
        return ConformingOp<IOp_max>::apply(a,b);
    }
    static IObject map(const IObject & arg) {
        if(!arg.isVec(2))
            return IObject();
        IObject fn = arg.getEntry(0);
        IObject es = arg.getEntry(1);
        if(!es.isVec())
            return IObject();
        size_t sz = es.size();
        IObject::vec_type v(sz);
        for(size_t i = 0; i < sz; i++) {
            v[i] = fn(es.getEntry(i));
            if(v[i].isUnknown())
                return v[i];
        }
        return IObject(v);
    }
    static IObject ifthenelse(const IObject & arg) {
        if(!arg.isVec(3))
            return IObject();
        IObject cond = arg.getEntry(0);
        if(!cond.isInt())
            return IObject();
        IObject then = arg.getEntry(1);
        IObject elseE = arg.getEntry(2);
        return cond.intValue() ? then : elseE;
    }
    static IObject foldl(const IObject & arg) {
        if(!arg.isVec(3))
            return IObject();
        IObject fn = arg.getEntry(0);
        IObject init = arg.getEntry(1);
        IObject es = arg.getEntry(2);
        if(!es.isVec())
            return IObject();
        size_t sz = es.size();
        for(size_t i = 0; i < sz; i++)
            init = fn( (init, es.getEntry(i)  ) );
        return init;
    }
    static IObject foldl1(const IObject & arg) {
        if(!arg.isVec(2))
            return IObject();
        IObject fn = arg.getEntry(0);
        IObject es = arg.getEntry(1);
        if(!es.isVec(1))
            return IObject();
        size_t sz = es.size();
        IObject init =  es.getEntry(0);
        for(size_t i = 1; i < sz; i++)
            init = fn( (init, es.getEntry(i)  ) );
        return init;
    }
    static IObject size(const IObject & arg) {
        if(!arg.isVec())
            return IObject();
        else
            return IObject((int) arg.size());
    }
    static IObject toInt(const IObject & arg) {
        if(arg.isNumber())
            return IObject(arg.intValue());
        else
            return IObject();
    }
    static IObject toDouble(const IObject & arg) {
        if(arg.isNumber())
            return IObject(arg.doubleValue());
        else
            return IObject();
    }
    static IObject dot(const IObject & arg) {
        if(!arg.isVec(2))
            return IObject();
        IObject a = arg.getEntry(0);
        IObject b = arg.getEntry(1);
        if(!a.isVec(1) || !b.isVec(1) || a.size() != b.size())
            return IObject();
        IObject c = a * b;
        assert(c.isVec(a.size()));
        size_t sz = c.size();
        IObject d = c.getEntry(0);
        for(size_t i = 1; i < sz; i++)
            d = d + c.getEntry(i);
        return d;
    }
    static IObject isNaN(const IObject & arg) {
    	return IObject( arg.isNaN() ? 1 : 0 );
    }
	bool operator()(const IObject & a, const IObject & b) const {
		return IObjectLib::compare(a,b) < 0;
	}
	static int compare(const IObject & a, const IObject & b) {

	    switch(a.getType()) {
	    case TINT:
	        if(b.getType() == TINT)
	            return a.intValue() - b.intValue();
	        /*fall through */
	    case TDOUBLE: {
	        if(!b.isNumber())
	            return a.getType() - b.getType();
            double r = a.doubleValue() - b.doubleValue();
            if(r < 0)
                return -1;
            else if(r > 0)
                return 1;
            else
                return 0;
        }
	    case TLAMBDA:
	        return (b.getType() == TLAMBDA) ? (char*) a.data.l.get() - (char*) b.data.l.get() : a.getType() - b.getType();
	    case TVEC: /* fall through */
	    case TFIELD_VEC: {
	        if(!b.isVec())
	            return a.getType() - b.getType();
	        if(a.size() != b.size())
                return a.size() - b.size();
            for(size_t i = 0; i < a.size(); i++) {
                int cmp = IObjectLib::compare(a.getEntry(i),b.getEntry(i));
                if(cmp)
                    return cmp;
            }
            return 0;
	    }
	    case TEXTERN_LAMBDA:
	        return (b.getType() == TEXTERN_LAMBDA) ? (char*) a.data.func - (char*) b.data.func : a.getType() - b.getType();
	    case TUNKNOWN:
	        return (b.getType() == TUNKNOWN) ? 0 : a.getType() - b.getType();
	    }
	    assert(!"comparison fell through");
	    return 0;
	}
};

static inline IObject operator<(const IObject & a, const IObject & b) {
	if(a.isUnknown() || b.isUnknown()) return IObject();
	else return IObject(IObjectLib::compare(a,b) < 0);
}

static inline IObject operator>(const IObject & a, const IObject & b) {
	return b < a;
}
static inline IObject operator==(const IObject & a, const IObject & b) {
	return IObject(IObjectLib::compare(a,b) == 0);
}
static inline IObject operator<=(const IObject & a, const IObject & b) {
	return IObject(IObjectLib::compare(a,b) <= 0);
}
static inline IObject operator>=(const IObject & a, const IObject & b) {
	return IObject(IObjectLib::compare(a,b) >= 0);
}
static inline IObject operator!=(const IObject & a, const IObject & b) {
	return !(a == b);
}


#endif
