
#ifndef _LINEAR_SYSTEM_PETSC_H
#define _LINEAR_SYSTEM_PETSC_H
#ifdef SPARSE_MATRIX_ENABLED

#include "petscksp.h"

// This should be set dynamically to a good upper 
// limit for the number of connections to neighbors 
// for each topological element in the matrix
// multiplied by the number of items stored on each connection.
#define DEFAULT_PER_ROW_STORAGE_SPACE 140

class PetSc {
public:
	typedef uint64_t Index;
	
	//called once per run,
	//nrows = number of rows in the matrix A, height of B
	//ncols = number of cols in the matrix A, width of X
	//my_partition = if in distributed memory the current partition this part of the matrix lives on (for now zero)
	//npartitions = number of partitions of memory (for now 1)
	//partitions = an array of npartitions+1 size, where the range [ partitions[i], partitions[i+1] ) specifies the part of the rows owned by partition i
	void init(size_t nrows, size_t ncols, size_t my_partition, size_t npartitions, Index * partitions) {
		printf("Initialize PetSc. nrows: %d, ncols: %d \n", nrows, ncols);
		_nrows = nrows;
		_ncols = ncols;
		_nnz = NULL;
		
		_nonzeroes.set_solver(this);
		
		_iters = 0;
		_max_iters = 1;
		_resid = 0.0;
		_target_resid = 1.0;

		int argc = 0;
		PetscInitialize(&argc,PETSC_NULL,PETSC_NULL,PETSC_NULL);
		KSPCreate(0, &ksp);
	}
	
	void finalize() {
		PetscFinalize();
	}
	
	///////////////////////////////////////////////////////////////////////////
	// NONZEROS: Functions for application to configure the sparsity pattern
	///////////////////////////////////////////////////////////////////////////	
	
		//non-zero specifying functions: called before the non-zero pattern changes, all previous non-zero patterns, including those applied to 
		//previously allocated arrays should be discarded in favor of the new nonzero pattern
		void beginNonZeroes() {
			printf("Calculating nonzero entries...\n");
			_nztime = processor_time();
			_nonzeroes.init(DEFAULT_PER_ROW_STORAGE_SPACE);
		}
	
		//these entries are only valid between begin and end NonZeroes calls, and specify that a non-zero lives at a location
		
		void entryA(Index row, Index col) {
			_nonzeroes.insert_entry(row, col);
		}
		void entryX(Index index) { }
		void entryB(Index index) { }
	
		template<size_t R, size_t C>
		void entryA(const vec<R,Index> & rows, const vec<C,Index> & cols) {
			for(size_t r = 0; r < R; ++r)
				for(size_t c = 0; c < C; ++c)
					entryA(rows[r], cols[c]);
		}
	
		template<size_t N>
		void entryX(const vec<N,Index> & indices) { }

		template<size_t N>
		void entryB(const vec<N,Index> & indices) { }
	
		//called at the end of a non-zero block
		void endNonZeroes();

	///////////////////////////////////////////////////////////////////////////
	// END NONZEROS
	///////////////////////////////////////////////////////////////////////////

	
	class Vector_t {
	public:
		Vector_t(PetSc* linear_system, size_t sz) {
			init(sz);
		}
		~Vector_t() {
		}
		
		double read(Index i) {
			PetscInt indices[1] = {i};
			double values[1] = {0};
			VecGetValues(x,1,indices,values);
			return values[0];
		}
		
		template<size_t N>
		vec<N,double> read(const vec<N,Index> idx) {
			vec<N,double> ret;
			PetscInt indices[N];
			for(int i = 0; i < N; i++)
				indices[i] = idx[i];
			VecGetValues(x,N,indices,ret.v);
			return ret;
		}
			
		//TODO: Specialize for LISZT_ASSIGN So that we don't bother reading.
		template<typename Operator>
		void write(Index i, double v) {
			PetscInt indices[1] = {i};
			double values[1] = {0};
			VecGetValues(x,1,indices,values);
			Operator::apply(v,&values[0]);
			VecSetValues(x,1,indices,values,INSERT_VALUES);
			VecAssemblyBegin(x);
			VecAssemblyEnd(x);
		}
		template<typename Operator, size_t N>
		void write(const vec<N,Index> & idx, vec<N,double> & vls) {
			PetscInt indices[N];
			double values[N];
			for(int i = 0; i < N; i++)
				indices[i] = idx[i];
			VecGetValues(x,N,indices,values);
			for(int i = 0; i < N; i++)
				Operator::apply((vls[i]),&values[i]);
			VecSetValues(x,N,indices,values,INSERT_VALUES);
			VecAssemblyBegin(x);
			VecAssemblyEnd(x);
		}
		
		
		//ensure the state is correct for upcoming operations
		//the possible states are in Fields/ReductionOps.h
		//if nested is true, this means that this ensure state is being called inside a forall loop
		//and can be safely ignored
		template<bool nested, typename T>
		void ensureState() { }
		
		size_t _sz;
		Vec x;
		
		void init(size_t sz) {
			_sz = sz;
			VecCreateSeq(0, _sz, &x);
		}
	};
	
	class Matrix_t {
	public:
		Matrix_t(PetSc* linear_system) {
			_linear_system = linear_system;
			_linear_system->add(this);
			init();
		}
		~Matrix_t() {
			_linear_system->remove(this);
		}
				
		double read(Index r, Index c) {
			PetscInt idxm[1] = {r};
			PetscInt idxn[1] = {c};
			double values[1];
			MatGetValues(A, 1, idxm, 1, idxn, values);
			return values[0];
		}

		//TODO: Optimize for block read
		template<size_t R, size_t C>
		matrix<R,C,double> read(const matrix<R,C,size_t> idx) {
			matrix<R,C,double> ret;
			for(int r = 0; r < R; r++)
				for(int c = 0; c < C; c++)
					ret(r,c) = read(idx[r],idx[c]);
			return ret;
		}
		
		template<typename Operator>
		void write(Index r, Index c, double v) {
			PetscInt idxm[1] = {r};
			PetscInt idxn[1] = {c};
			double values[1];
			// MatGetValues(A, 1, idxm, 1, idxn, values);
			// Operator::apply(v,&values[0]);
			MatSetValues(A, 1, idxm, 1, idxn, values, INSERT_VALUES);
			//MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
			//MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);
		}
		
		//TODO: Optimize for block write
		//TODO: Optimize for LISZT_ASSIGN
		template<typename Operator, size_t R, size_t C>
		void write(const vec<R,Index> & rows, 
		           const vec<C,Index> & cols,
		           ::matrix<R,C,double> & values) {
			for(size_t r = 0; r < R; r++)
				for(size_t c = 0; c < C; c++) {
					write<Operator>(rows[r],cols[c],values(r,c));
				}
		}
		
		void clear() {
			//set to 0.0
		}
		           
		//ensure the state is correct for upcoming operations
		//the possible states are in Fields/ReductionOps.h
		//if nested is true, this means that this ensure state is being called inside a forall loop
		//and can be safely ignored
		template<bool nested, typename T>
		void ensureState() { }
		
		void init() {
			//can be called whenever _nnz is available, and will reallocate matrix if necessary.
			MatCreateSeqAIJ(0,
			  _linear_system->_nrows,
			  _linear_system->_ncols,
			  0,
			  _linear_system->_nnz,
			  &A);
			MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);
			MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);
		}

		PetSc* _linear_system;
		Mat A; 
	};
	
	//types actually used by the compiler
	typedef shared_ptr<Matrix_t> Matrix;
	typedef shared_ptr<Vector_t> Vector;
	
	//allocators, note that even after allocation, a call to change non-zero layouts should still
	//affect already allocated objects
	
	Matrix A();
	Vector x();
	Vector b();
	
	//invoke the solver for Ax=b
	void solve(Matrix A, Vector x, Vector b);
	
	int iters() { return _iters; }
	void max_iters(int i) { _max_iters = i; }
	double resid() { return _resid; }
	void target_resid(double d) { _target_resid = d; }
	
protected:
	
	void add(Matrix_t * sm) {
		matrices.insert(sm);
	}
	void remove(Matrix_t * sm) {
		matrices.erase(sm);
	}
	void init_all_matrices();
	
	std::set<Matrix_t *> matrices;

	size_t _nrows;
	size_t _ncols;
	int* _nnz; //number of nonzeros per row, computed during endNonZeros
	
	int _iters;
	int _max_iters;
	double _resid;
	double _target_resid;
	double _nztime;
	
	KSP ksp;        // linear solver context

	
private:
		
	// Here we build the nonzero entries of a CRS-stored matrix as sets of nonzeros for each row.
	// We build these sets using linear unsorted arrays, ideal for a small amount of nonzeros per row.
	
	class LinearArrayNonzeros {
	public:
		
		LinearArrayNonzeros() { _nnr = NULL; }
		
		void set_solver(PetSc* s ) { _nrows = s->_nrows; }
		
		void clear() {
			if (_nnr != NULL) {
				for (size_t i = 0; i < _nrows; i++) {
					free(_nnr[i].entries);
				}
				free(_nnr);
				_nnr = NULL;
			}
		}
		void init(size_t default_storage_size) {
			clear();
			_nnr = (nonzero_row*) calloc(_nrows,sizeof(nonzero_row));
			if (default_storage_size < 1)
				default_storage_size = 1;
			for (size_t i = 0; i < _nrows; i++) {
				_nnr[i].entries = (Index*) malloc(default_storage_size * sizeof(Index));
				_nnr[i].num_entries = 0;
				_nnr[i].storage_size = default_storage_size;
			}
			printf("Allocated %ld KB for nonzero pattern.\n", _nrows * ((default_storage_size + 2) * sizeof(Index) + sizeof(Index*) + sizeof(nonzero_row)) / 1024);
		}
	
		void insert_entry(Index row, Index col) {
			nonzero_row* nzrow = &_nnr[row];
			bool found = false;
			for (Index i = 0; i < nzrow->num_entries; i++) {
				if (nzrow->entries[i] == col)
					found = true;
			}
			if (!found) {
				if (nzrow->num_entries == nzrow->storage_size) {
					nzrow->entries = (Index*) realloc(nzrow->entries, nzrow->storage_size*2 * sizeof(Index));
					nzrow->storage_size *= 2;
					printf("Had to realloc nonzero entries for row %ld.\n", row);
				}
				nzrow->entries[nzrow->num_entries] = col;
				nzrow->num_entries++;
			}
		}
	
		int* finalize() {
			int* nnz = (int*) malloc(_nrows * sizeof(int));
			for (size_t row = 0; row < _nrows; row++) {
				nnz[row] = _nnr[row].num_entries;
			}
			return nnz;
		}
		
	private:
		struct nonzero_row {
			Index* entries;
			Index num_entries;
			Index storage_size;
		};
		nonzero_row* _nnr; //nonzero entries in each row (unsorted), computed during nonzero block.
		
		int _nrows;
	};
	
	class SingleLinearArrayNonzeros {
	public:
		
		SingleLinearArrayNonzeros() { };
		
		void set_solver(PetSc* s) { _nrows = s->_nrows; };
		
		void clear() {
			
		}
		
		void init(size_t default_storage_size) {
			num_nz = (int*) malloc(_nrows * sizeof(int));
			for (size_t row = 0; row < _nrows; row++) {
				num_nz[row] = 0;
			}
		}
		
		void insert_entry(Index row, Index col) {
			num_nz[row] += 1;
		}
		
		int* finalize() {
			return num_nz;
		}
		
	private:
	
		int _nrows;
		int* num_nz;
	};
	
	LinearArrayNonzeros _nonzeroes;
	//SingleLinearArrayNonzeros _nonzeroes;
};


PetSc::Matrix PetSc::A() {
	return Matrix(new Matrix_t(this));
}
PetSc::Vector PetSc::x() {
	return Vector(new Vector_t(this,_ncols));
}
PetSc::Vector PetSc::b() {
	return Vector(new Vector_t(this,_nrows));
}

void PetSc::endNonZeroes() {
	
	if (_nnz != NULL) {
		free(_nnz);
	}
	_nnz = _nonzeroes.finalize();
	
	int largest = 0;
	double avg = 0.0;
	int lrow = 0;
	for (size_t row = 0; row < _nrows; row++) {
		avg += _nnz[row];
		if (_nnz[row] > largest) {
			largest = _nnz[row];
			lrow = row;
		}
	}
	avg /= _nrows;
	printf("Done in %lf ms.\n", processor_time() - _nztime);
	printf("Sparse Matrix Stats: %d x %d matrix, most occupied row is %d and has %d entries. average connectivity: %lf\n", _nrows, _ncols, lrow, largest, avg);

	_nonzeroes.clear();
	init_all_matrices();
}


void PetSc::init_all_matrices() {
	//iterate over set of matrix pointers
	//call init on it.
}
void PetSc::solve(Matrix A, Vector x, Vector b) {
	
	MatAssemblyBegin(A->A,MAT_FINAL_ASSEMBLY);
	MatAssemblyEnd(A->A,MAT_FINAL_ASSEMBLY);
	
	PC pc;
	KSPGetPC(ksp,&pc);
	PCSetType(pc, PCASM);
	
	printf("Getting ready to solve!!\n");
	
	
	KSPSetOperators(ksp, A->A, A->A, SAME_NONZERO_PATTERN);
	KSPSetInitialGuessKnoll(ksp, PETSC_TRUE);
	KSPSetType(ksp, KSPGMRES);
	KSPGMRESSetRestart(ksp, 100);
	KSPSetFromOptions(ksp);

	KSPSolve(ksp, b->x, x->x);
	
	KSPGetIterationNumber(ksp,&_iters);
	KSPGetResidualNorm(ksp,&_resid);
	
	MatAssemblyBegin(A->A,MAT_FINAL_ASSEMBLY);
	MatAssemblyEnd(A->A,MAT_FINAL_ASSEMBLY);

	
	printf("Done with solve!\n");
	
	return;
}

#endif //_LINEAR_SYSTEM_PETSC_H
#endif //SPARSE_MATRIX_ENABLED
