
#ifndef _LINEAR_SYSTEM_SUPERLU_H
#define _LINEAR_SYSTEM_SUPERLU_H
#ifdef SPARSE_MATRIX_ENABLED

#include "Array.h"
#include<set>
#include<stdlib.h>
#include<vector>

extern "C" {
#include "slu_ddefs.h"
}

using std::tr1::shared_ptr;

class SuperLU {
public:
	typedef uint64_t Index;
	
	void init(size_t nrows, size_t ncols, size_t my_partition, size_t npartitions, Index * partitions) { //called once at start of run
		numRows = nrows;
		numCols = ncols;
		nnz_ = 0;
	}
	void finalize() {}
	
		//non-zero specifying functions:
	void beginNonZeroes() {
	  nnz_ = 0;
	  nonzeros_.clear();
	  nonzeros_.resize(numCols);
	}

	void entryA(Index row, Index col) {
	  if(nonzeros_[col].find(row) == nonzeros_[col].end()) {
	    ++nnz_;
	  }
	  nonzeros_[col].insert(row);
	}
	void entryX(Index index) {}
	void entryB(Index index) {}
	
	
	template<size_t R, size_t C>
	void entryA(const vec<R,Index> & rows, const vec<C,Index> & cols) {
	  for(size_t r = 0; r < R; ++r)
	    for(size_t c = 0; c < C; ++c)
	      entryA(rows[r], cols[c]);
	}
	
	template<size_t N>
	void entryX(const vec<N,Index> & indices) {}
	
	template<size_t N>
	void entryB(const vec<N,Index> & indices) {}
	
	void endNonZeroes();
	
	class Vector_t {
	public:
		Vector_t(SuperLU * ls, size_t sz) {
			linear_system = ls;
			linear_system->add(this);
			data.reset(sz);
		}
		~Vector_t() {
			linear_system->remove(this);
		}
		double read(Index i) { return data[i]; }
		template<size_t N>
		vec<N,double> read(const vec<N,Index> idx) {
			vec<N,double> ret;
			for(int i = 0; i < N; i++)
				ret[i] = read(idx[i]);
			return ret;
		}
			
		template<typename Operator>
		void write(Index a, double v) {
			Operator::apply(v,&data[a]);
		}
		
		template<typename Operator, size_t N>
		void write(const vec<N,Index> & indices, vec<N,double> & values) {
			for(size_t i = 0; i < N; i++)
				write<Operator>(indices[i],values[i]);
		}
		
		
		template<bool nested, typename T>
		void ensureState() {}
		
		
		private:
			Array<double> data;
			friend class SuperLU;
		SuperLU * linear_system;
	};
	
	class Matrix_t {
	public:
		Matrix_t(SuperLU * ls) {
			linear_system = ls;
			linear_system->add(this);
		}
		~Matrix_t() {
		  // Do not delete the A matrix because the memory is not owned
		  // by SuperLU. It will be freed by Array.
		  // Destroy_CompCol_Matrix(&A_matrix);
		  Destroy_SuperNode_Matrix(&L);
		  Destroy_CompCol_Matrix(&U);
		  linear_system->remove(this);
		}
		double read(Index row, Index col) { return data[Key(col,row)]; }
		template<size_t R, size_t C>
		matrix<R,C,double> read(const matrix<R,C,size_t> idx) {
			matrix<R,C,double> ret;
			for(int r = 0; r < R; r++)
				for(int c = 0; c < C; c++)
					ret(r,c) = read(r,c);
			return ret;
		}
		
		template<typename Operator>
		void write(Index r, Index c, double v) {
			Operator::apply(v,&data[Key(c,r)]);
		}
		
		template<typename Operator, size_t R, size_t C>
		void write(const vec<R,Index> & rows, 
		           const vec<C,Index> & cols,
		           ::matrix<R,C,double> & values) {
			for(size_t r = 0; r < R; r++)
				for(size_t c = 0; c < C; c++) {
					write<Operator>(rows[r],cols[c],values(r,c));
				}
		}
		template<bool nested, typename T>
		void ensureState() {}
	private:
		
		typedef std::pair<Index,Index> Key;
		typedef std::map<Key,double> data_t;
		data_t data;
		friend class SuperLU;
		SuperLU * linear_system;
		Array<double> nonZeroes;
		SuperMatrix A_matrix, L, U;
	};
	typedef shared_ptr<Matrix_t> Matrix;
	typedef shared_ptr<Vector_t> Vector;

	Matrix A();
	Vector x();
	Vector b();
	
	void solve(Matrix A, Vector x, Vector b);
	
	int iters() { return 1; }
	void max_iters(int i) {}
	double resid() { return 0.0; }
	void target_resid(double d) {}

	
protected:
	typedef std::set<Index> NonzeroSet;
	typedef std::vector< NonzeroSet > CompressedColumnMap;
	
	void add(Vector_t * sv) {
		vectors.insert(sv);
	}
	void remove(Vector_t * sv) {
		vectors.erase(sv);
	}
	void add(Matrix_t * sm) {
		matrices.insert(sm);
	}
	void remove(Matrix_t * sm) {
		matrices.erase(sm);
	}
	std::set<Matrix_t *> matrices;
	std::set<Vector_t *> vectors;

	CompressedColumnMap nonzeros_;
	Array<int> asub_;
	Array<int> xa_;
	
	size_t numRows, numCols, nnz_;
};

SuperLU::Matrix SuperLU::A() {
	return Matrix(new Matrix_t(this));
}
SuperLU::Vector SuperLU::x() {
	return Vector(new Vector_t(this,numCols));
}
SuperLU::Vector SuperLU::b() {
	return Vector(new Vector_t(this,numRows));
}

void SuperLU::endNonZeroes() {
  // Reset the size of the arrays.
  asub_.reset(nnz_);
  xa_.reset(numCols+1);

  size_t index = 0;
  NonzeroSet::iterator rowIter;
  for(size_t c = 0; c < numCols; ++c) {
    xa_[c] = index;
    rowIter = nonzeros_[c].begin();
    for(; rowIter != nonzeros_[c].end(); ++rowIter) {
      asub_[index] = *rowIter;
      ++index;
    }
  }
  xa_[numCols] = index;
  nonzeros_.clear();
}

void SuperLU::solve(Matrix A, Vector x, Vector b) {
	int nnz = A->data.size();
	A->nonZeroes.reset(nnz);
	typedef SuperLU::Matrix_t::data_t::iterator it_t;
	int count = 0;
	for(it_t it = A->data.begin(); it != A->data.end(); it++) {
		double data = it->second;
		A->nonZeroes[count] = data;
		count++;
	}


	assert(numRows == numCols);
	
	SuperMatrix B_matrix;
	
	dCreate_CompCol_Matrix(&A->A_matrix, numRows, numCols, nnz, 
			       A->nonZeroes, asub_, xa_, SLU_NC, SLU_D, SLU_GE);
	
	memcpy(x->data,b->data,sizeof(double) * numRows);
	dCreate_Dense_Matrix(&B_matrix, numCols, 1, x->data, numCols, SLU_DN,
			     SLU_D, SLU_GE);
	
	superlu_options_t options;
	set_default_options(&options);
	options.ColPerm = NATURAL;
	
	SuperLUStat_t stat;
	
	Array<int> perm_c(numCols);
	Array<int> perm_r(numCols);
	
	StatInit(&stat);
	
	int info;
	dgssv(&options, &A->A_matrix, perm_c, perm_r, &A->L, &A->U, &B_matrix,
	      &stat, &info);
	
	StatFree(&stat);
	
	//dPrint_CompCol_Matrix("A",&A->A_matrix);
	//dPrint_Dense_Matrix("X",&B_matrix);
	
	//for(int i = 0; i < numCols; i++) {
	//	std::cout << i << " " << x->data[i] << std::endl;
	//}
}


#endif //SPARSE_MATRIX_ENABLED

#endif
