/*
 * LisztSingle.cpp
 *
 *  Created on: Aug 17, 2009
 *      Author: zdevito
 */

#include "SingleCore/LisztSingle.h"
#include "MeshIO/FacetEdgeBuilder.h"
#include <fstream>

namespace LisztSinglePrivate {

inline LisztSingle::LisztSingle() : is_prepared(false), has_sizes(false) {}

inline void LisztSingle::init(int argc, char ** argv) {
	assert(argc == 3);
	
	reader.init(argv[1]);
	int log = atoi(argv[2]);
	if(log)
		redirectToLog("liszt_single");
	
	watch_point_ref wp(new WatchPoint("Program Initialization","0"));
	snapshot_ref snap(new Snapshot(wp,IObject(),IObject(),0,"<Initialization>",0));
	snapshots_.push_back(snap);
	is_prepared = true;
	
	
	MeshIO::LisztHeader h = reader.header();
	MeshIO::FacetEdgeBuilder builder;
	builder.init(h.nV,h.nE,h.nF,h.nC,h.nFE);
	MeshIO::FileFacetEdge * fes = reader.facetEdges();
	builder.insert(0,h.nFE,fes);
	reader.free(fes);
	mesh_.initFromFacetEdgeBuilder(&builder);
	boundary_builder.init(h.nBoundaries,reader.boundaries());
	
    has_sizes = true;
    elem_sizes[VERTEX] = mesh()->vertices().size();
    elem_sizes[EDGE] = mesh()->edges().size();
    elem_sizes[FACE] = mesh()->faces().size();
    //TODO(zach): we need a +1 to include the zero cell which cells doesn't count, i don't like this hack however, so
    //we should clean up the way field sizes are set at some point.
    elem_sizes[CELL] = mesh()->cells().size() + 1;
}

inline void LisztSingle::finalize() {}

inline void LisztSingle::redirectToLog(const char * base) {
    ostringstream ss;
    ss << base << ".log";
    int file = fileno(fopen(ss.str().c_str(),"a"));
    dup2(file,1);
    dup2(file,2);
}


template <typename MO, typename BoundarySet>
inline void LisztSingle::buildSet(const std::string & name, BoundarySet * bs) {
	boundary_builder.load<CRSMesh::Mesh,MO,BoundarySet>(mesh(),name.c_str(),bs);
}

template<typename T,typename V>
inline void LisztSingle::buildFieldWithConst(id_type field_id, const std::string & name, const V & init,SingleField::Field<T,V> * field) {
	initField(field_id,name,field);
	std::fill_n(field->getReadEntry(0),field->nReadEntries(),init);
}

template<typename V>
inline void LisztSingle::buildFieldWithPositions(id_type field_id, const std::string & name, SingleField::Field<VertexType,V> * field) {
	initField(field_id,name,field);
	MeshIO::PositionTable * table = reader.positions();
	for(size_t i = 0; i < field->nReadEntries(); i++) {
		*field->getReadEntry(i) = double3(table->data[i][0],table->data[i][1],table->data[i][2]);
	}
	reader.free(table);
}
    
template<typename FieldType>
struct addToVisInfo {
	static void apply(VisInfo * vis_info, FieldType * field) {}
};

template<>
struct addToVisInfo< ::SingleField::Field<VertexType,LisztScalar> > {
	static void apply(VisInfo * vis_info, ::SingleField::Field<VertexType,LisztScalar> * field) {
		vis_info->scalar_vertex_fields.push_back(field);
	}
};
template<>
struct addToVisInfo< ::SingleField::Field<VertexType,vec<3,LisztScalar> > > {
	static void apply(VisInfo * vis_info, ::SingleField::Field<VertexType,vec<3,LisztScalar> > * field) {
		vis_info->vector_vertex_fields.push_back(field);
	}
};
template<>
struct addToVisInfo< ::SingleField::Field<CellType,LisztScalar> > {
	static void apply(VisInfo * vis_info, ::SingleField::Field<CellType,LisztScalar> * field) {
		vis_info->scalar_cell_fields.push_back(field);
	}
};
template<>
struct addToVisInfo< ::SingleField::Field<CellType,vec<3,LisztScalar> > > {
	static void apply(VisInfo * vis_info, ::SingleField::Field<CellType,vec<3,LisztScalar> > * field) {
		vis_info->vector_cell_fields.push_back(field);
	}
};

template<typename T,typename V>
inline void LisztSingle::initField(id_type field_id, const std::string & name, SingleField::Field<T,V> * field) {
	field->init(name,elem_sizes[T::value]);
	field_display_ref fd(new FieldDisplayImpl<T,V>(snapshots_.size(),field));
	registerField(T::value,fd);
	addToVisInfo< ::SingleField::Field<T,V> >::apply(&vis_info,field);
}

// This method is intended for debug use only. Do not use in regular code.
template <typename ElemType>
id_type LisztSingle::getGlobalID(ElemType e) const {
    return e.ID();
}

enum VTKCellType { VTK_UNKNOWN = 0, VTK_TETRA = 10, VTK_HEXAHEDRON = 12, VTK_WEDGE = 13, VTK_PYRAMID = 14};
template<typename Cell>
inline static VTKCellType getVTKType(const Cell & c) {
	size_t nv = c.vertices().size();
	size_t ne = c.edges().size();
	size_t nf = c.faces().size();
	if(nv == 4 && ne == 6 && nf == 4)
		return VTK_TETRA;
	else if(nv == 8 && ne == 12 && nf == 6)
		return VTK_HEXAHEDRON;
	else if(nv == 6 && ne == 9 && nf == 5)
		return VTK_WEDGE;
	else if(nv == 5 && ne == 8 && nf == 5)
		return VTK_PYRAMID;
	else
		return VTK_UNKNOWN;
}

template<typename T>
inline static void printVec(std::vector<T> & vec) {
	for(unsigned int i = 0; i < vec.size(); i++) {
		cout << vec[i]->name() << " ";
	}
	cout << endl;
}

inline void LisztSingle::saveVTKFile(const std::string & filename) {
	cout << "saveVTKFile" << endl;
	printVec(vis_info.scalar_vertex_fields);
	printVec(vis_info.vector_vertex_fields);
	printVec(vis_info.scalar_cell_fields);
	printVec(vis_info.vector_cell_fields);
	std::ofstream file(filename.c_str());
	file.setf(std::ios::fixed, std::ios::floatfield);
	size_t nverts = mesh()->vertices().size();
	size_t ncells = mesh()->cells().size(); //does not include outside
	file << "<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" byte_order=\"BigEndian\">" << endl;
	file << "<UnstructuredGrid>" << endl;
	file << "<Piece NumberOfPoints=\"" << nverts << "\" NumberOfCells=\"" << ncells << "\">" << endl;
	file << "<PointData>" << endl;
	for(unsigned int i = 0; i < vis_info.scalar_vertex_fields.size(); i++) {
		const ::SingleField::Field<VertexType, LisztScalar> & field = *vis_info.scalar_vertex_fields[i];
		file << "  <DataArray type=\"Float32\" Name=\"" << field.name() << "\" Format=\"ascii\">" << endl;
		FORALL_SET_SER(v,mesh()->vertices())
			file << field(v) << " ";
		ENDSET
		file << endl;
		file << "  </DataArray>" << endl;
	}
	for(unsigned int i = 0; i < vis_info.vector_vertex_fields.size(); i++) {
		const ::SingleField::Field<VertexType, vec<3,LisztScalar> > & field = *vis_info.vector_vertex_fields[i];
		file << "  <DataArray type=\"Float32\" Name=\"" << field.name() << "\" NumberOfComponents=\"3\" Format=\"ascii\">" << endl;
		FORALL_SET_SER(v,mesh()->vertices())
			vec<3,LisztScalar> e = field(v);
			file << "    " << e.x() << " " << e.y() << " " << e.z() << endl;
		ENDSET
		file << "  </DataArray>" << endl;
	}
	
	file << "</PointData>" << endl;
	file << "<CellData>" << endl;
	for(unsigned int i = 0; i < vis_info.scalar_cell_fields.size(); i++) {
		const ::SingleField::Field<CellType, LisztScalar> & field = *vis_info.scalar_cell_fields[i];
		file << "  <DataArray type=\"Float32\" Name=\"" << field.name() << "\" Format=\"ascii\">" << endl;
		FORALL_SET_SER(c,mesh()->cells())
			file << field(c) << " ";
		ENDSET
		file << endl;
		file << "  </DataArray>" << endl;
	}
	for(unsigned int i = 0; i < vis_info.vector_cell_fields.size(); i++) {
		const ::SingleField::Field<CellType, vec<3,LisztScalar> > & field = *vis_info.vector_cell_fields[i];
		file << "  <DataArray type=\"Float32\" Name=\"" << field.name() << "\" NumberOfComponents=\"3\" Format=\"ascii\">" << endl;
		FORALL_SET_SER(c,mesh()->cells())
			vec<3,LisztScalar> e = field(c);
			file << "    " << e.x() << " " << e.y() << " " << e.z() << endl;
		ENDSET
		file << "  </DataArray>" << endl;
	}
	file << "</CellData>" << endl;
	file << "<Points>" << endl;
	file << "  <DataArray type=\"Float32\" NumberOfComponents=\"3\" Format=\"ascii\">" << endl;
	MeshIO::PositionTable * table = reader.positions();
	for(size_t i = 0; i < nverts; i++) {
		file << "    " << table->data[i][0] << " " << table->data[i][1] << " " << table->data[i][2] << endl;
	}
	file << "  </DataArray>" << endl;
	file << "</Points>" << endl;
	file << "<Cells>" << endl;
	file << "  <DataArray type=\"Int32\" Name=\"connectivity\" Format=\"ascii\">" << endl;
	FORALL_SET_SER(c,mesh()->cells())
		switch(getVTKType(c)) {
			case VTK_UNKNOWN:
				cout << "VTKFILE: Unknown cell type for id = " << c.ID() << "bailing out." << endl;
				return;
				break;
			default:
				file << "    ";
				FORALL_SET_SER(v,c.vertices())
					file << v.ID() << " ";
				ENDSET
				file << endl;
				break;
		}
	ENDSET
	file << "  </DataArray>" << endl;
	
	file << "  <DataArray type=\"Int32\" Name=\"offsets\" Format=\"ascii\">" << endl;
	
	size_t offset = 0;
	FORALL_SET_SER(c,mesh()->cells())
		switch(getVTKType(c)) {
			case VTK_TETRA:
				offset += 4;
				break;
			case VTK_HEXAHEDRON:
				offset += 8;
				break;
			case VTK_WEDGE:
				offset += 6;
				break;
			case VTK_PYRAMID:
				offset += 5;
				break;
			default:
				assert(!"unexpected control flow");
				break;
		}
		file << offset << " ";
	ENDSET
	file << endl;
	file << "  </DataArray>" << endl;

	file << "  <DataArray type=\"Int32\" Name=\"types\" Format=\"ascii\">" << endl;
	FORALL_SET_SER(c,mesh()->cells())
		file << getVTKType(c) << " ";
	ENDSET
	file << endl;
	file << "  </DataArray>" << endl;
	file << "</Cells>" << endl;
	file << "</Piece>" << endl;
	file << "</UnstructuredGrid>" << endl;
	file << "</VTKFile>" << endl;
	file.close();
	reader.free(table);
}

}  // namespace LisztPrivate

