/*
 * LisztOpenMP.h
 *
 *  Created on: Aug 17, 2009
 *      Author: zdevito
 */

#ifndef LISZTSINGLE_H_
#define LISZTSINGLE_H_

#include "liszt_memory.h"
#include "SingleCore/Global.h"
#include "SingleCore/Field.h"
#include "SingleCore/FieldDisplay.h"
#include "SingleCore/WatchPoint.h"
#include "SingleCore/Snapshot.h"
#include "Liszt/BoundarySet.h"
#include "Layout/BoundarySetBuilder.h"

#include <sys/time.h>

namespace LisztSinglePrivate {

#ifndef LisztScalar
#define LisztScalar double
#endif

#define LISZT_RUNTIME_NAMESPACE LisztSinglePrivate

struct VisInfo {
    //keep track of fields for visualization purposes
    std::vector< SingleField::Field<VertexType, LisztScalar > * > scalar_vertex_fields;
    std::vector< SingleField::Field<CellType, LisztScalar >  * > scalar_cell_fields;
    std::vector< SingleField::Field<VertexType, vec<3,LisztScalar> > * > vector_vertex_fields;
    std::vector< SingleField::Field<CellType, vec<3,LisztScalar>  > * > vector_cell_fields;
};

class LisztSingle {
public:
    LisztSingle();
    void init(int argc, char ** argv);
    void prepare();
    void finalize();
    void redirectToLog(const char * base);


    template <typename MO, typename BoundarySet>
    void buildSet(const std::string & name, BoundarySet * bs);
    
    template<typename T,typename V>
    void buildFieldWithConst(id_type field_id, const std::string & name, const V & init,SingleField::Field<T,V> * field);
    
    template<typename V>
    void buildFieldWithPositions(id_type field_id, const std::string & name, SingleField::Field<VertexType,V> * field);

    // This method is intended for debug use only. Do not use in regular code.
    template <typename ElemType>
    id_type getGlobalID(ElemType e) const;

    typedef ::std::tr1::shared_ptr<FieldDisplay> field_display_ref;
    std::vector<field_display_ref> registered_fields[ELEM_TYPE_SIZE];

	typedef ::std::tr1::shared_ptr<WatchPoint> watch_point_ref;
    typedef ::std::tr1::shared_ptr<Snapshot> snapshot_ref;
	
    void takeSnapshot(watch_point_ref p, const IObject & oldv, const IObject & newv, const std::string & filename, size_t lineno) {
        size_t shid = snapshots_.size();
        snapshot_ref sh(new Snapshot(p,oldv,newv,shid,filename,lineno));
        snapshots_.push_back(sh);
        snapshotFields();
        cout << "snapshotting ..." << endl;
    }
    void testWatchPoints(const std::string & filename, size_t lineno) {
    	for(size_t i = 0; i < watch_points.size(); i++) {
    		watch_point_ref p = watch_points[i];
    		IObject oldv;
    		IObject newv;
    		cout << "testing ..." << endl;
    		if(p->eval(field_ctx,&oldv,&newv)) {
                takeSnapshot(p, oldv,newv,filename,lineno);
    		}
    		//cout << "old: " << oldv << " new: " << newv << endl;
    	}
    }
    void resetSnapshots() {
    	for(size_t i = 0; i < ELEM_TYPE_SIZE; i++) {
			registered_fields[i].clear();
		}
    	for(size_t i = 0; i < watch_points.size(); i++) {
    		watch_points[i]->reset();
    	}
    	field_ctx.clear();
    	snapshots_.clear();
    	watch_point_ref wp(new WatchPoint("Program Initialization","0"));
    	snapshot_ref snap(new Snapshot(wp,IObject(),IObject(),0,"<Initialization>",0));
    	snapshots_.push_back(snap);
    }
    void setCurrent() {
    	for(size_t i = 0; i < ELEM_TYPE_SIZE; i++)
		for(size_t j = 0; j < registered_fields[i].size();j ++) {
			field_display_ref f = registered_fields[i][j];
			f->setDisplayMode(FieldDisplay::DISPLAY_CURRENT);
		}
    }
    void setSnapshot(size_t snap_id) {
    	assert(snap_id < snapshots_.size());
    	for(size_t i = 0; i < ELEM_TYPE_SIZE; i++)
    	for(size_t j = 0; j < registered_fields[i].size();j ++) {
    		LisztSingle::field_display_ref f = registered_fields[i][j];
    		f->setDisplayMode(FieldDisplay::DISPLAY_HISTORY);
    		f->setSnapshot(snap_id);
    	}
    }
    void addWatchPoint(const std::string & name, const std::string & exps) {
    	watch_point_ref wp(new WatchPoint(name,exps));
    	watch_points.push_back(wp);
    }
    void removeWatchPoint(size_t i) {
    	assert(i < watch_points.size());
    	for(;i + 1 < watch_points.size(); i++) {
    		watch_points[i] = watch_points[i+1];
    	}
    	watch_points.pop_back();
    }
    std::vector<watch_point_ref> & watchPoints() {
    	return watch_points;
    }
    const std::vector<snapshot_ref> & snapshots() const {
    	return snapshots_;
    }
    Exp::ctx & fieldContext() {
    	return field_ctx;
    }
    const CRSMesh::Mesh * mesh() const {
    	return &mesh_;
    }
    void saveVTKFile(const std::string & filename);
private:
    template<typename T,typename V>
    void initField(id_type field_id, const std::string & name, SingleField::Field<T,V> * field);
    
    void snapshotFields() { //optionally the address of the field that changed
    	for(size_t i = 0; i < ELEM_TYPE_SIZE; i++)
    	for(size_t j = 0; j < registered_fields[i].size();j ++) {
    		field_display_ref f = registered_fields[i][j];
    		size_t nsnaps = f->snapshot();
    		if(nsnaps != snapshots_.size())
    			assert(!"snapshots are inconsistent with a field");
    	}
    }
    void registerField(ElemType typ, field_display_ref fd) {
    	registered_fields[typ].push_back(fd);
    	field_ctx[fd->name()] = IObject(fd);
    }
    Exp::ctx field_ctx;
    std::vector<watch_point_ref> watch_points;
    std::vector<snapshot_ref> snapshots_;
    size_t n_snapshots; //possibly includes snapshots of fields not from watchpoints (e.g. init and final)
    
    bool is_prepared;
    //must prepare the mesh store to get these
    bool has_sizes;
    
    //HACK: I need the number of vertices, edges, faces, to initialize fields
    size_t elem_sizes[ELEM_TYPE_SIZE];
    VisInfo vis_info;
    
    MeshIO::LisztFileReader reader;
    BoundarySetBuilder boundary_builder;
    CRSMesh::Mesh mesh_;
};

} // namespace LisztOpenMPPrivate
#include "SingleCore/LisztSingle.cpp"

static inline double single_wtime() {
   double wtime;
   struct timeval tv;
   gettimeofday(&tv, NULL);
   wtime = tv.tv_sec;
   wtime += (double)tv.tv_usec / 1000000.0;
   return wtime;
}

#define WTIME single_wtime()

#endif /* LISZTOPENMP_H_ */
