
#ifndef __LISZT_VECTORIZER_H
#define __LISZT_VECTORIZER_H

#include "Liszt/ElemTypes.h"
#include <stdlib.h>
#include <vector>

template<typename T>
class Value;
class Mask;
class VectorState;

using LisztPrivate::id_type;
template<typename T>
class Value {
private:
	typedef std::vector<T> data_type;
public:
	typedef T value_type;
	typedef typename data_type::const_reference const_reference;
	typedef typename data_type::reference reference;
	Value(bool init = true);
	const_reference operator[](id_type i) const;
	reference operator[](id_type i);
	void setDepth(size_t d);
	size_t size() const;
private:
	friend class VectorState;
	id_type getIndex(id_type i) const;
	size_t nestingDepth;
	data_type data;
};

class Mask {
public:
	void pushNew(size_t s);
	void push();
	void pop();
	Mask & operator&=(const Mask & rhs);
	template<typename VT>
	Mask & operator&=(const Value<VT> & rhs);
	Mask & andEqNot(const Mask & rhs);
	bool operator[](id_type idx) const;
	void print() const {
		std::cout << "Mask: ";
		for(size_t i = 0; i < mask_stack.back().size(); i++) {
			std::cout << mask_stack.back()[i] << " ";
		}
		std::cout << std::endl;
	}
private:
	typedef std::vector<bool> mask_row;
	std::vector<mask_row> mask_stack;
};

class VectorState {
public:	
	VectorState();
	static VectorState & instance();
	
	template<typename VT>
	void beginIf(const Value<VT> & pred);
	template<typename VT>
	void beginElse(const Value<VT> & pred);
	void endIf();
	bool beginFunction();
	//some or all lanes have returned
	void functionReturn();
	void endFunction();
	template<typename set_type>
	void beginForall(const Value<set_type> & new_sets, Value<typename set_type::value_type> * itervar, Value<int> * label = NULL);
	void endForall();
	void beginWhile();
	bool enabled(id_type i) const;
	bool anyEnabled() const;
	
	template<typename VT>
	bool whilePred(const Value<VT> & pred);
	void endWhile();
	size_t currentDepth() const;
	size_t currentWidth() const;
private:
	template<typename T>
	friend class Value;
	size_t nestingDepth;
	Mask cond_mask; //mask conditional operations,
	Mask func_mask; //mask for function returns;
	typedef std::vector<id_type> lane_type;
	std::vector<lane_type> lane_stack;
};

#define VEXEC \
for(size_t __iter_var = 0; __iter_var < VectorState::instance().currentWidth();__iter_var++) if(VectorState::instance().enabled(__iter_var))

#define VGET(a) ((a)[__iter_var])

#define VASSIGN(var) VEXEC var[__iter_var]

#define VDEFINE(typ,var)  \
Value<typ > var; \
VASSIGN(var)

#define VFORALL(e,v) \
{ \
typedef typeof(v) SetType; \
Value<SetType::value_type::iterator::value_type > e; \
VectorState::instance().beginForall(v,&e);

#define VFORALL_TM(e,v) \
{ \
typedef typeof(v) SetType; \
typedef typename SetType::value_type::iterator::value_type ValueType; \
Value< ValueType > e; \
VectorState::instance().beginForall(v,&e);

#define VFORALL_LABELED(e,l,v) \
{ \
typedef typeof(v) SetType; \
Value<SetType::value_type::iterator::value_type > e; \
Value< int > l; \
VectorState::instance().beginForall(v,&e, &l);

#define VFORALL_LABELED_TM(e,l,v) \
{ \
typedef typeof(v) SetType; \
typedef typename SetType::value_type::iterator::value_type ValueType; \
Value< ValueType > e; \
Value< int > l; \
VectorState::instance().beginForall(v,&e, &l);

#define VEND_FORALL VectorState::instance().endForall(); }

#define VIF(pred)  VectorState::instance().beginIf(pred); {
#define VELSE(pred) } VectorState::instance().beginElse(pred); { 
#define VEND_IF	} VectorState::instance().endIf();

#define VWHILE(unique) \
VectorState::instance().beginWhile(); \
unique : {
#define VWHILE_EXP(unique,exp) \
if(!VectorState::instance().whilePred(exp)) goto unique##_end;
#define VEND_WHILE(unique) goto unique; } unique##_end: VectorState::instance().endWhile();

#define VFUNCTION(typ) \
Value<typ > __return_value; \
if(!VectorState::instance().beginFunction()) return __return_value;

#define VRETURN(exp) \
VASSIGN(__return_value) = exp; \
VectorState::instance().functionReturn()

#define VEND_FUNCTION VectorState::instance().endFunction(); return __return_value;

#include "vectorizer_impl.h"
#endif
