/*
 *  vectorizer.cpp
 *  vector_test
 *
 *  Created by Zach DeVito on 11/23/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#include "vectorizer.h"
#include "ForAll.h"

template<typename T>
Value<T>::Value(bool init) 
: nestingDepth(init ? VectorState::instance().currentDepth() : 0),
  data(init ? VectorState::instance().currentWidth() : 0,T())
 {}

template<typename T>
typename Value<T>::const_reference Value<T>::operator[](id_type i) const {
	const_reference d = data[getIndex(i)];
	return d;
}
template<typename T>
typename Value<T>::reference Value<T>::operator[](id_type i) {
	return data[getIndex(i)];
}

template<typename T>
void Value<T>::setDepth(size_t d) {
	nestingDepth = d;
	data.clear();
}
template<typename T>
size_t Value<T>::size() const {
	return data.size();
}
template<typename T>
id_type Value<T>::getIndex(id_type i) const {
	size_t current_depth = VectorState::instance().currentDepth();
	while(current_depth != nestingDepth) {
		assert(i <  VectorState::instance().lane_stack[current_depth].size());
		i = VectorState::instance().lane_stack[current_depth--][i];
	}
	assert(i < data.size());
	return i;
}

void Mask::pushNew(size_t s) {
	mask_stack.push_back(mask_row(s,true));
}
void Mask::push() {
	assert(mask_stack.size() > 0);
	mask_stack.push_back(mask_stack.back());
}
void Mask::pop() {
	assert(mask_stack.size() > 0);
	mask_stack.pop_back();
}
Mask & Mask::operator&=(const Mask & rhs) {
	const mask_row & r = rhs.mask_stack.back();
	mask_row & cur = mask_stack.back();
	assert(cur.size() == r.size());
	for(size_t i = 0; i < cur.size(); i++)
		cur[i] = cur[i] && r[i];
	return *this;
}

template<typename VT>
Mask & Mask::operator&=(const Value<VT> & rhs) {
	mask_row & cur = mask_stack.back();
	assert(cur.size() == VectorState::instance().currentWidth());
	for(size_t i = 0; i < cur.size(); i++)
		cur[i] = cur[i] && rhs[i];
	return *this;
}
Mask & Mask::andEqNot(const Mask & rhs) {
	const mask_row & r = rhs.mask_stack.back();
	mask_row & cur = mask_stack.back();
	assert(cur.size() == r.size());
	for(size_t i = 0; i < cur.size(); i++)
		cur[i] = cur[i] && !r[i];
	return *this;
}
bool Mask::operator[](id_type idx) const {
	assert(idx < mask_stack.back().size());
	return mask_stack.back()[idx];
}


VectorState::VectorState() {
		nestingDepth = 0;
		lane_stack.push_back(lane_type(1,0));
		func_mask.pushNew(1);
		cond_mask.pushNew(1);
	}
VectorState & VectorState::instance() {
	static VectorState vs;
	return vs;
}
template<typename VT>
void VectorState::beginIf(const Value<VT> & pred) {
	cond_mask.push();
	cond_mask &= pred;
}

template<typename VT>
static Value<VT> flip(const  Value<VT> & v) {
	Value<VT> v2;
	for(size_t i = 0; i < v2.size(); i++)
		v2[i] = !v[i];
	return v2;
}

template<typename VT>
void VectorState::beginElse(const Value<VT> & pred) {//same as before
	cond_mask.pop();
	cond_mask.push();
	cond_mask &= flip(pred);
}

void VectorState::endIf() {
	cond_mask.pop();
}
bool VectorState::beginFunction() {
	if(anyEnabled()) {
		func_mask.push();
		return true;
	} else {
		return false;
	}
}
//some or all lanes have returned
void VectorState::functionReturn() {
	func_mask.andEqNot(cond_mask); //is this right?
}
void VectorState::endFunction() {
	func_mask.pop();
}
template<typename set_type>
void VectorState::beginForall(const Value<set_type> & new_sets, Value<typename set_type::value_type> * itervar, Value<int> * labels) {
	itervar->setDepth(nestingDepth + 1);
	if(labels)
	    labels->setDepth(nestingDepth + 1);
	lane_stack.push_back(lane_type());
	lane_type & current = lane_stack[lane_stack.size() - 2];
	//std::cout << current.size() << std::endl;
	lane_type & next = lane_stack.back();
	size_t position = 0;
	for(size_t i = 0; i < current.size(); i++) {
		if(enabled(i)) {
			const set_type & s = new_sets[i];
			next.insert(next.end(),s.size(),i);
			int count = 0;
			FORALL_SET_SER_TM(e,s)
				itervar->data.push_back(e);
				if(labels)
				    labels->data.push_back(count++);
				position++;
			ENDSET
		}
	}
	assert(position == next.size());
	func_mask.pushNew(currentWidth());
	cond_mask.pushNew(currentWidth());
	nestingDepth++;
}
void VectorState::endForall() {
	lane_stack.pop_back();
	func_mask.pop();
	cond_mask.pop();
	nestingDepth--;
}
void VectorState::beginWhile() {
	cond_mask.push();
}
bool VectorState::enabled(id_type i) const {
	return func_mask[i] && cond_mask[i];
}
bool VectorState::anyEnabled() const {
	for(size_t i = 0; i < lane_stack.back().size(); i++)
		if(func_mask[i] && cond_mask[i])
			return true;
	return false;
}
template<typename VT>
bool VectorState::whilePred(const Value<VT> & pred) {
	cond_mask &= pred;
	if(anyEnabled())
		return true;
	else
		return false;
}
void VectorState::endWhile() {
	cond_mask.pop();
}
size_t VectorState::currentDepth() const {
	return nestingDepth;
}
size_t VectorState::currentWidth() const {
	return lane_stack.back().size();
}
