#ifndef _SYRAH_LISZT_MATRIX_H_
#define _SYRAH_LISZT_MATRIX_H_

namespace syrah {
  template<size_t N, size_t M, typename ElemType, int SIMD_WIDTH>
  class SimdMatrix {
  public:
    enum { MatrixSize = N * M };
    typedef FixedVector<ElemType, SIMD_WIDTH> SimdElem;
    typedef FixedVectorMask<SIMD_WIDTH> SimdMask;
    typedef SimdVec<M, ElemType, SIMD_WIDTH> SimdColumnVector;
    //typedef vec<M, ElemType, SIMD_WIDTH> ColumnVector;

    SYRAH_FORCEINLINE SimdMatrix() {}

    SYRAH_FORCEINLINE SimdMatrix(const ElemType* all_data) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].load(all_data + i * SIMD_WIDTH);
      }
    }

    SYRAH_FORCEINLINE SimdMatrix(const ElemType** all_data) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].load(all_data[i]);
      }
    }

    SYRAH_FORCEINLINE SimdMatrix(const ElemType* all_data, bool /*aligned*/) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].load_aligned(all_data + i * SIMD_WIDTH);
      }
    }

    SYRAH_FORCEINLINE SimdMatrix(const ElemType** all_data, bool /*aligned*/) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].load_aligned(all_data[i]);
      }
    }

    SYRAH_FORCEINLINE void store_aligned(ElemType* all_data) const {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].store_aligned(all_data + i * SIMD_WIDTH);
      }
    }

    SYRAH_FORCEINLINE void store_aligned(ElemType* all_data, const SimdMask& mask) const {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].store_aligned(all_data + i * SIMD_WIDTH, mask);
      }
    }

    SYRAH_FORCEINLINE void store(ElemType* all_data) const {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].store(all_data + i * SIMD_WIDTH);
      }
    }

    SYRAH_FORCEINLINE void store(ElemType* all_data, const SimdMask& mask) const {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].store(all_data + i * SIMD_WIDTH, mask);
      }
    }

    SYRAH_FORCEINLINE void store_aligned(ElemType** all_data) const {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].store_aligned(all_data[i]);
      }
    }

    SYRAH_FORCEINLINE void store_aligned(ElemType** all_data, const SimdMask& mask) const {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].store_aligned(all_data[i], mask);
      }
    }

    SYRAH_FORCEINLINE void store(ElemType** all_data) const {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].store(all_data[i]);
      }
    }

    SYRAH_FORCEINLINE void store(ElemType** all_data, const SimdMask& mask) const {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].store(all_data[i], mask);
      }
    }

    template<typename OtherElem>
    SYRAH_FORCEINLINE SimdMatrix(const SimdMatrix<N, M, OtherElem, SIMD_WIDTH>& other) {
      for (size_t i = 0; i < MatrixSize; i++) data[i] = other.data[i];
    }

    template<typename OtherElem>
    SYRAH_FORCEINLINE SimdMatrix& operator=(const SimdMatrix<N, M, OtherElem, SIMD_WIDTH>& other) {
      for (size_t i = 0; i < N; i++) {
        data[i] = other.data[i];
      }
      return *this;
    }

    SYRAH_FORCEINLINE SimdMatrix(const SimdElem& other) {
      for (size_t i = 0; i < N; i++) {
        data[i] = other;
      }
    }

    template<typename OtherElem>
    SYRAH_FORCEINLINE SimdMatrix& operator=(const SimdElem& other) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i] = other;
      }
      return *this;
    }

    SYRAH_FORCEINLINE SimdMatrix(const matrix<N, M, ElemType>& other) {
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          operator()(r, c) = other(r, c);
        }
      }
    }

    SYRAH_FORCEINLINE SimdMatrix& operator=(const matrix<N, M, ElemType>& other) {
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          operator()(r, c) = other(r, c);
        }
      }
      return *this;
    }

    SYRAH_FORCEINLINE SimdMatrix(const ElemType& other) {
      for (size_t i = 0; i < N; i++) {
        data[i] = other;
      }
    }

    SYRAH_FORCEINLINE SimdMatrix& operator=(const ElemType& other) {
      for (size_t i = 0; i < N; i++) {
        data[i] = other;
      }
      return *this;
    }

    static SimdMatrix Zero() {
      SimdMatrix result;
      for (size_t i = 0; i < MatrixSize; i++) result.data[i] = SimdElem::Zero();
      return result;
    }

    SYRAH_FORCEINLINE SimdElem& operator()(size_t row, size_t col) { return data[M * row + col]; }
    SYRAH_FORCEINLINE const SimdElem& operator()(size_t row, size_t col) const { return data[M * row + col]; }

    SYRAH_FORCEINLINE void setMat(size_t which_simd, const matrix<N,M,ElemType>& mat) {
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          operator()(r, c)[which_simd] = mat(r, c);
        }
      }
    }

    SYRAH_FORCEINLINE matrix<N, M, ElemType> getMat(size_t which_simd) const {
      matrix<N, M, ElemType> result;
      for (size_t r = 0; r < N; r++) {
        for (size_t c = 0; c < M; c++) {
          result(r, c) = operator()(r,c)[which_simd];
        }
      }
      return result;
    }

    SYRAH_FORCEINLINE SimdMatrix operator+(const SimdMatrix& other) const {
      SimdMatrix result;
      for (size_t i = 0; i < MatrixSize; i++) {
        result.data[i] = data[i] + other.data[i];
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdMatrix operator-(const SimdMatrix& other) const {
      SimdMatrix result;
      for (size_t i = 0; i < MatrixSize; i++) {
        result.data[i] = data[i] - other.data[i];
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdMatrix operator-() const {
      SimdMatrix result;
      for (size_t i = 0; i < MatrixSize; i++) {
        result.data[i] = -data[i];
      }
      return result;
    }

    template<size_t OtherM>
    SYRAH_FORCEINLINE SimdMatrix<N, OtherM, ElemType, SIMD_WIDTH> operator*(const SimdMatrix<M, OtherM, ElemType, SIMD_WIDTH>& other) const {
      SimdMatrix<N, OtherM, ElemType, SIMD_WIDTH> result;
      for (size_t i = 0; i < N; i++) {
        for (size_t j = 0; j < OtherM; j++) {
          SimdElem elem = SimdElem::Zero();
          for (size_t k = 0; k < M; k++) {
            elem += operator()(i, k) * other(k, j);
          }
          result(i, j) = elem;
        }
      }
      return result;
    }

    SYRAH_FORCEINLINE SimdMatrix operator*(const SimdElem& other) const {
      SimdMatrix result;
      for (size_t i = 0; i < MatrixSize; i++) {
        result.data[i] = data[i] * other;
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdMatrix operator*(const ElemType& other) const {
      SimdMatrix result;
      for (size_t i = 0; i < MatrixSize; i++) {
        result.data[i] = data[i] * other;
      }
      return result;
    }

    SYRAH_FORCEINLINE SimdColumnVector operator*(const SimdColumnVector& other) const {
      SimdColumnVector result;
      for (size_t r = 0; r < N; r++) {
        SimdElem elem = SimdElem::Zero();
        for (size_t c = 0; c < M; c++) {
          elem += operator()(r,c) * other[c];
        }
        result[r] = elem;
      }
      return result;
    }

    // NOTE(boulos): Disallow Matrix/Matrix

    SYRAH_FORCEINLINE SimdMatrix operator/(const SimdElem& other) const {
      SimdMatrix result;
      for (size_t i = 0; i < MatrixSize; i++) {
        result.data[i] = data[i] / other;
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdMatrix operator/(const ElemType& other) const {
      SimdMatrix result;
      for (size_t i = 0; i < MatrixSize; i++) {
        result.data[i] = data[i] / other;
      }
      return result;
    }

    SYRAH_FORCEINLINE SimdMatrix& operator+=(const SimdMatrix& other) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i] += other.data[i];
      }
      return *this;
    }
    SYRAH_FORCEINLINE SimdMatrix& operator-=(const SimdMatrix& other) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i] -= other.data[i];
      }
      return *this;
    }

#if 0
    SYRAH_FORCEINLINE SimdMatrix& operator*=(const SimdMatrix& other) {
      for (size_t i = 0; i < N; i++) {
        data[i] *= other.data[i];
      }
      return *this;
    }
#endif

    SYRAH_FORCEINLINE SimdMatrix& operator*=(const SimdElem& other) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i] *= other;
      }
      return *this;
    }
    SYRAH_FORCEINLINE SimdMatrix& operator*=(const ElemType& other) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i] *= other;
      }
      return *this;
    }

    SYRAH_FORCEINLINE SimdMatrix& operator/=(const SimdElem& other) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i] /= other;
      }
      return *this;
    }
    SYRAH_FORCEINLINE SimdMatrix& operator/=(const ElemType& other) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i] /= other;
      }
      return *this;
    }

    SYRAH_FORCEINLINE SimdMask operator==(const SimdMatrix& other) const {
      SimdMask result = SimdMask::True();
      for (size_t i = 0; i < MatrixSize; i++) {
        result &= (data[i] == other.data[i]);
      }
      return result;
    }

    SYRAH_FORCEINLINE SimdMask operator!=(const SimdMatrix& other) const {
      SimdMask result = SimdMask::True();
      for (size_t i = 0; i < MatrixSize; i++) {
        result &= (data[i] != other.data[i]);
      }
      return result;
    }

    // okay, so min and max are defined as not BinaryOps in Liszt.
    SYRAH_FORCEINLINE void min(const SimdMatrix& other) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i] = min(data[i], other.data[i]);
      }
    }

    SYRAH_FORCEINLINE void max(const SimdMatrix& other) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i] = max(data[i], other.data[i]);
      }
    }

    SYRAH_FORCEINLINE void merge(const SimdMatrix& other, const SimdMask& mask) {
      for (size_t i = 0; i < MatrixSize; i++) {
        data[i].merge(other.data[i], mask);
      }
    }

    SimdElem data[MatrixSize];
  };

  template<size_t N, size_t M, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE std::ostream& operator<<(std::ostream& os, const SimdMatrix<N, M, ElemType, SIMD_WIDTH>& mat) {
    os << "(";
    for (size_t i = 0; i < SIMD_WIDTH; i++) {
      const char* spacing = (i == 0) ? "" : ", ";
      os << spacing << mat.getMat(i);
    }
    os << ")";
    return os;
  }


  template<size_t N, size_t M, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdMatrix<N, M, ElemType, SIMD_WIDTH> operator*(const ElemType& scalar, const SimdMatrix<N, M, ElemType, SIMD_WIDTH>& mat) {
    SimdMatrix<N, M, ElemType, SIMD_WIDTH> result;
    for (size_t i = 0; i < N*M; i++) {
      result.data[i] = scalar * mat.data[i];
    }
    return result;
  }

  template<size_t N, size_t M, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdMatrix<N, M, ElemType, SIMD_WIDTH> operator*(const FixedVector<ElemType, SIMD_WIDTH>& scalar, const SimdMatrix<N, M, ElemType, SIMD_WIDTH>& mat) {
    SimdMatrix<N, M, ElemType, SIMD_WIDTH> result;
    for (size_t i = 0; i < N * M; i++) {
      result.data[i] = scalar * mat.data[i];
    }
    return result;
  }

#if 0
  template<size_t N, size_t M, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdMatrix<N, M, ElemType, SIMD_WIDTH> operator*(const matrix<N, ElemType>& other, const SimdMatrix<N, M, ElemType, SIMD_WIDTH>& mat) {
    SimdMatrix<N, M, ElemType, SIMD_WIDTH> result;
    for (size_t r = 0; r < N; r++) {
      for (size_t c = 0; c < M; c++) {
        // Matrix-Matrix Multiply
      }
    }
    return result;
  }
#endif

  template<size_t N, size_t M, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdMatrix<N, M, ElemType, SIMD_WIDTH> operator+(const matrix<N, M, ElemType>& other, const SimdMatrix<N, M, ElemType, SIMD_WIDTH>& mat) {
    SimdMatrix<N, M, ElemType, SIMD_WIDTH> result;
    for (size_t r = 0; r < N; r++) {
      for (size_t c = 0; c < M; c++) {
        result(r, c) = (other(r, c) + mat(r, c));
      }
    }
    return result;
  }

  template<size_t N, size_t M, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdMatrix<N, M, ElemType, SIMD_WIDTH> select(const SimdMatrix<N, M, ElemType, SIMD_WIDTH>& a,
                                                                  const SimdMatrix<N, M, ElemType, SIMD_WIDTH>& b,
                                                                  const syrah::FixedVectorMask<SIMD_WIDTH>& mask) {
    SimdMatrix<N, M, ElemType, SIMD_WIDTH> result;
    for (size_t i = 0; i < N*M; i++) {
      result.data[i] = select(a.data[i], b.data[i], mask);
    }
    return result;
  }
};


#endif // _SYRAH_LISZT_VECTOR_H_
