#ifndef _SYRAH_LISZT_VECTOR_H_
#define _SYRAH_LISZT_VECTOR_H_

namespace syrah {
  template<size_t N, typename ElemType, int SIMD_WIDTH>
  class SimdVec {
  public:
    typedef FixedVector<ElemType, SIMD_WIDTH> SimdElem;
    typedef FixedVectorMask<SIMD_WIDTH> SimdMask;

    SYRAH_FORCEINLINE SimdVec() {}

    SYRAH_FORCEINLINE SimdVec(const ElemType* all_data) {
      for (int i = 0; i < N; i++) {
        v[i].load(all_data + i * SIMD_WIDTH);
      }
    }

    SYRAH_FORCEINLINE SimdVec(const ElemType** all_data) {
      for (int i = 0; i < N; i++) {
        v[i].load(all_data[i]);
      }
    }

    SYRAH_FORCEINLINE SimdVec(const ElemType* all_data, bool /*aligned*/) {
      for (int i = 0; i < N; i++) {
        v[i].load_aligned(all_data + i * SIMD_WIDTH);
      }
    }

    SYRAH_FORCEINLINE SimdVec(const ElemType** all_data, bool /*aligned*/) {
      for (int i = 0; i < N; i++) {
        v[i].load_aligned(all_data[i]);
      }
    }

    SYRAH_FORCEINLINE void store_aligned(ElemType* all_data) const {
      for (int i = 0; i < N; i++) {
        v[i].store_aligned(all_data + i * SIMD_WIDTH);
      }
    }

    SYRAH_FORCEINLINE void store_aligned(ElemType* all_data, const SimdMask& mask) const {
      for (int i = 0; i < N; i++) {
        v[i].store_aligned(all_data + i * SIMD_WIDTH, mask);
      }
    }

    SYRAH_FORCEINLINE void store(ElemType* all_data) const {
      for (int i = 0; i < N; i++) {
        v[i].store(all_data + i * SIMD_WIDTH);
      }
    }

    SYRAH_FORCEINLINE void store(ElemType* all_data, const SimdMask& mask) const {
      for (int i = 0; i < N; i++) {
        v[i].store(all_data + i * SIMD_WIDTH, mask);
      }
    }

    SYRAH_FORCEINLINE void store_aligned(ElemType** all_data) const {
      for (int i = 0; i < N; i++) {
        v[i].store_aligned(all_data[i]);
      }
    }

    SYRAH_FORCEINLINE void store_aligned(ElemType** all_data, const SimdMask& mask) const {
      for (int i = 0; i < N; i++) {
        v[i].store_aligned(all_data[i], mask);
      }
    }

    SYRAH_FORCEINLINE void store(ElemType** all_data) const {
      for (int i = 0; i < N; i++) {
        v[i].store(all_data[i]);
      }
    }

    SYRAH_FORCEINLINE void store(ElemType** all_data, const SimdMask& mask) const {
      for (int i = 0; i < N; i++) {
        v[i].store(all_data[i], mask);
      }
    }

    template<typename OtherElem>
    SYRAH_FORCEINLINE SimdVec(const SimdVec<N, OtherElem, SIMD_WIDTH>& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other.v[i];
      }
    }

    template<typename OtherElem>
    SYRAH_FORCEINLINE SimdVec& operator=(const SimdVec<N, OtherElem, SIMD_WIDTH>& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other.v[i];
      }
      return *this;
    }

    SYRAH_FORCEINLINE SimdVec(const SimdElem& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other;
      }
    }

    template<typename OtherElem>
    SYRAH_FORCEINLINE SimdVec& operator=(const SimdElem& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other;
      }
      return *this;
    }

#if 0
    // And implicit constructors from the non-SIMD version.
    template<typename OtherElem>
    SYRAH_FORCEINLINE SimdVec(const vec<N, OtherElem>& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other.v[i];
      }
    }

    template<typename OtherElem>
    SYRAH_FORCEINLINE SimdVec& operator=(const vec<N, OtherElem>& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other.v[i];
      }
      return *this;
    }
#endif

    SYRAH_FORCEINLINE SimdVec(const vec<N, ElemType>& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other.v[i];
      }
    }

    SYRAH_FORCEINLINE SimdVec& operator=(const vec<N, ElemType>& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other.v[i];
      }
      return *this;
    }

    SYRAH_FORCEINLINE SimdVec(const ElemType& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other;
      }
    }

    SYRAH_FORCEINLINE SimdVec& operator=(const ElemType& other) {
      for (int i = 0; i < N; i++) {
        v[i] = other;
      }
      return *this;
    }

    static SimdVec Zero() {
      SimdVec result;
      for (int i = 0; i < N; i++) result[i] = SimdElem::Zero();
      return result;
    }

    SYRAH_FORCEINLINE SimdElem& operator[](int i) { return v[i]; }
    SYRAH_FORCEINLINE const SimdElem& operator[](int i) const { return v[i]; }

    SYRAH_FORCEINLINE void setVec(int which_simd, const vec<N,ElemType>& vec) {
      for (int i = 0; i < N; i++) {
        v[i][which_simd] = vec[i];
      }
    }

    SYRAH_FORCEINLINE vec<N, ElemType> getVec(int which_simd) const {
      vec<N, ElemType> result;
      for (int i = 0; i < N; i++) {
        result[i] = v[i][which_simd];
      }
      return result;
    }

    SYRAH_FORCEINLINE SimdElem dot(const SimdVec& other) const {
      SimdElem result = SimdElem::Zero();
      for (int i = 0; i < N; i++) {
        result += v[i] * other.v[i];
      }
      return result;
    }

    SYRAH_FORCEINLINE SimdElem length() const {
      SimdElem result = SimdElem::Zero();
      for (int i = 0; i < N; i++) {
        result += v[i] * v[i];
      }
      return sqrt(result);
    }

    SYRAH_FORCEINLINE void normalize() {
      SimdElem result = SimdElem::Zero();
      for (int i = 0; i < N; i++) {
        result += v[i] * v[i];
      }
      // NOTE(boulos): Use /= because *= inv isn't guaranteed to be
      // precise...
      SimdElem len = sqrt(result);
      for (int i = 0; i < N; i++) {
        v[i] /= len;
      }
    }

    SYRAH_FORCEINLINE SimdVec operator+(const SimdVec& other) const {
      SimdVec result;
      for (int i = 0; i < N; i++) {
        result.v[i] = v[i] + other.v[i];
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdVec operator-(const SimdVec& other) const {
      SimdVec result;
      for (int i = 0; i < N; i++) {
        result.v[i] = v[i] - other.v[i];
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdVec operator-() const {
      SimdVec result;
      for (int i = 0; i < N; i++) {
        result.v[i] = -v[i];
      }
      return result;
    }

    SYRAH_FORCEINLINE SimdVec operator*(const SimdVec& other) const {
      SimdVec result;
      for (int i = 0; i < N; i++) {
        result.v[i] = v[i] * other.v[i];
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdVec operator*(const SimdElem& other) const {
      SimdVec result;
      for (int i = 0; i < N; i++) {
        result.v[i] = v[i] * other;
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdVec operator*(const ElemType& other) const {
      SimdVec result;
      for (int i = 0; i < N; i++) {
        result.v[i] = v[i] * other;
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdVec operator/(const SimdVec& other) const {
      SimdVec result;
      for (int i = 0; i < N; i++) {
        result.v[i] = v[i] / other.v[i];
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdVec operator/(const SimdElem& other) const {
      SimdVec result;
      for (int i = 0; i < N; i++) {
        result.v[i] = v[i] / other;
      }
      return result;
    }
    SYRAH_FORCEINLINE SimdVec operator/(const ElemType& other) const {
      SimdVec result;
      for (int i = 0; i < N; i++) {
        result.v[i] = v[i] / other;
      }
      return result;
    }


    SYRAH_FORCEINLINE SimdVec& operator+=(const SimdVec& other) {
      for (int i = 0; i < N; i++) {
        v[i] += other.v[i];
      }
      return *this;
    }
    SYRAH_FORCEINLINE SimdVec& operator-=(const SimdVec& other) {
      for (int i = 0; i < N; i++) {
        v[i] -= other.v[i];
      }
      return *this;
    }
    SYRAH_FORCEINLINE SimdVec& operator*=(const SimdVec& other) {
      for (int i = 0; i < N; i++) {
        v[i] *= other.v[i];
      }
      return *this;
    }
    SYRAH_FORCEINLINE SimdVec& operator*=(const SimdElem& other) {
      for (int i = 0; i < N; i++) {
        v[i] *= other;
      }
      return *this;
    }
    SYRAH_FORCEINLINE SimdVec& operator*=(const ElemType& other) {
      for (int i = 0; i < N; i++) {
        v[i] *= other;
      }
      return *this;
    }

    SYRAH_FORCEINLINE SimdVec& operator/=(const SimdVec& other) {
      for (int i = 0; i < N; i++) {
        v[i] /= other.v[i];
      }
      return *this;
    }
    SYRAH_FORCEINLINE SimdVec& operator/=(const SimdElem& other) {
      for (int i = 0; i < N; i++) {
        v[i] /= other;
      }
      return *this;
    }
    SYRAH_FORCEINLINE SimdVec& operator/=(const ElemType& other) {
      for (int i = 0; i < N; i++) {
        v[i] /= other;
      }
      return *this;
    }

    SYRAH_FORCEINLINE SimdMask operator==(const SimdVec& other) const {
      SimdMask result = SimdMask::True();
      for (int i = 0; i < N; i++) {
        result &= (v[i] == other.v[i]);
      }
      return result;
    }

    SYRAH_FORCEINLINE SimdMask operator!=(const SimdVec& other) const {
      SimdMask result = SimdMask::True();
      for (int i = 0; i < N; i++) {
        result &= (v[i] != other.v[i]);
      }
      return result;
    }

    // okay, so min and max are defined as not BinaryOps in Liszt.
    SYRAH_FORCEINLINE void min(const SimdVec& other) {
      for (int i = 0; i < N; i++) {
        v[i] = min(v[i], other.v[i]);
      }
    }

    SYRAH_FORCEINLINE void max(const SimdVec& other) {
      for (int i = 0; i < N; i++) {
        v[i] = max(v[i], other.v[i]);
      }
    }

    SYRAH_FORCEINLINE void merge(const SimdVec& other, const SimdMask& mask) {
      for (int i = 0; i < N; i++) {
        v[i].merge(other.v[i], mask);
      }
    }

    SYRAH_FORCEINLINE vec<N, ElemType> foldMin() const {
       vec<N, ElemType> result;
       for (int i = 0; i < N; i++) result[i] = v[i].foldMin();
       return result;
    }

    SYRAH_FORCEINLINE vec<N, ElemType> foldMax() const {
       vec<N, ElemType> result;
       for (int i = 0; i < N; i++) result[i] = v[i].foldMax();
       return result;
    }

    SYRAH_FORCEINLINE vec<N, ElemType> foldSum() const {
       vec<N, ElemType> result;
       for (int i = 0; i < N; i++) result[i] = v[i].foldSum();
       return result;
    }

    SYRAH_FORCEINLINE vec<N, ElemType> foldProd() const {
       vec<N, ElemType> result;
       for (int i = 0; i < N; i++) result[i] = v[i].foldProd();
       return result;
    }

    SimdElem v[N];
  };

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE std::ostream& operator<<(std::ostream& os, const SimdVec<N, ElemType, SIMD_WIDTH>& vec) {
    os << "(";
#if 0
    for (int i = 0; i < N; i++) {
      const char* spacing = (i == 0) ? "" : ", ";
      os << spacing << vec.v[i];
    }
#else
    for (int i = 0; i < SIMD_WIDTH; i++) {
      const char* spacing = (i == 0) ? "" : ", ";
      os << spacing << vec.getVec(i);
    }
#endif
    os << ")";
    return os;
  }


  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdVec<N, ElemType, SIMD_WIDTH> operator*(const ElemType& scalar, const SimdVec<N, ElemType, SIMD_WIDTH>& vec) {
    SimdVec<N, ElemType, SIMD_WIDTH> result;
    for (int i = 0; i < N; i++) {
      result.v[i] = scalar * vec.v[i];
    }
    return result;
  }

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdVec<N, ElemType, SIMD_WIDTH> operator*(const FixedVector<ElemType, SIMD_WIDTH>& scalar, const SimdVec<N, ElemType, SIMD_WIDTH>& vec) {
    SimdVec<N, ElemType, SIMD_WIDTH> result;
    for (int i = 0; i < N; i++) {
      result.v[i] = scalar * vec.v[i];
    }
    return result;
  }

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdVec<N, ElemType, SIMD_WIDTH> operator*(const vec<N, ElemType>& other, const SimdVec<N, ElemType, SIMD_WIDTH>& vec) {
    SimdVec<N, ElemType, SIMD_WIDTH> result;
    for (int i = 0; i < N; i++) {
      result.v[i] = other.v[i] * vec.v[i];
    }
    return result;
  }

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdVec<N, ElemType, SIMD_WIDTH> operator+(const vec<N, ElemType>& other, const SimdVec<N, ElemType, SIMD_WIDTH>& vec) {
    SimdVec<N, ElemType, SIMD_WIDTH> result;
    for (int i = 0; i < N; i++) {
      result.v[i] = other.v[i] + vec.v[i];
    }
    return result;
  }

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE FixedVector<ElemType, SIMD_WIDTH> length(const SimdVec<N, ElemType, SIMD_WIDTH>& vec) {
    FixedVector<ElemType, SIMD_WIDTH> result = FixedVector<ElemType, SIMD_WIDTH>::Zero();
    for (int i = 0; i < N; i++) {
      result += vec.v[i] * vec.v[i];
    }
    return sqrt(result);
  }

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdVec<N, ElemType, SIMD_WIDTH> normalize(const SimdVec<N, ElemType, SIMD_WIDTH>& vec) {
    SimdVec<N, ElemType, SIMD_WIDTH> result;
    FixedVector<ElemType, SIMD_WIDTH> len = length(vec);
    for (int i = 0; i < N; i++) {
      result.v[i] = vec.v[i] / len;
    }
    return result;
  }

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE FixedVector<ElemType, SIMD_WIDTH> dot(const SimdVec<N, ElemType, SIMD_WIDTH>& a,
                                                          const SimdVec<N, ElemType, SIMD_WIDTH>& b) {
    FixedVector<ElemType, SIMD_WIDTH> result = FixedVector<ElemType, SIMD_WIDTH>::Zero();
    for (int i = 0; i < N; i++) {
      result += a.v[i] * b.v[i];
    }
    return result;
  }

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE FixedVector<ElemType, SIMD_WIDTH> dot(const vec<N, ElemType>& a,
                                                          const SimdVec<N, ElemType, SIMD_WIDTH>& b) {
    FixedVector<ElemType, SIMD_WIDTH> result = FixedVector<ElemType, SIMD_WIDTH>::Zero();
    for (int i = 0; i < N; i++) {
      result += a.v[i] * b.v[i];
    }
    return result;
  }

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE FixedVector<ElemType, SIMD_WIDTH> dot(const SimdVec<N, ElemType, SIMD_WIDTH>& a,
                                                         const vec<N, ElemType>& b) {
    FixedVector<ElemType, SIMD_WIDTH> result = FixedVector<ElemType, SIMD_WIDTH>::Zero();
    for (int i = 0; i < N; i++) {
      result += a.v[i] * b.v[i];
    }
    return result;
  }

  template<size_t N, typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdVec<N, ElemType, SIMD_WIDTH> select(const SimdVec<N, ElemType, SIMD_WIDTH>& a,
                                                           const SimdVec<N, ElemType, SIMD_WIDTH>& b,
                                                           const syrah::FixedVectorMask<SIMD_WIDTH>& mask) {
    SimdVec<N, ElemType, SIMD_WIDTH> result;
    for (int i = 0; i < N; i++) {
      result[i] = select(a.v[i], b.v[i], mask);
    }
    return result;
  }

  template<typename ElemType, int SIMD_WIDTH>
  SYRAH_FORCEINLINE SimdVec<3, ElemType, SIMD_WIDTH> cross(const SimdVec<3, ElemType, SIMD_WIDTH>& a,
                                                          const SimdVec<3, ElemType, SIMD_WIDTH>& b) {
    SimdVec<3, ElemType, SIMD_WIDTH> result;
    result[0] = a[1] * b[2] - a[2] * b[1];
    result[1] = a[2] * b[0] - a[0] * b[2];
    result[2] = a[0] * b[1] - a[1] * b[0];
    return result;
  }
};


#endif // _SYRAH_LISZT_VECTOR_H_
