#ifndef _LISZT_CACHE_SIM_H_
#define _LISZT_CACHE_SIM_H_

#include <map>

class SimpleCache {
public:
  class CacheLevel {
  public:
    typedef std::map<size_t, size_t> AddressMap;
    CacheLevel(int total_size, int line_size, CacheLevel* next_level) : total_size(total_size), line_size(line_size), next_level(next_level) {
      num_lines = total_size / line_size;
      tags = new size_t[num_lines];
      recently_used = new bool[num_lines];

      memset(tags, -1, sizeof(size_t) * num_lines);
      memset(recently_used, 0, sizeof(bool) * num_lines);

      clock_ptr = 0;
      num_hits = 0;
      num_misses = 0;
    }

    void fetch(const void* ptr) {
      // get the ptr as a size_t
      size_t address = reinterpret_cast<size_t>(ptr);
      // Round down the address to be line-aligned (0 out the bottom bits)
      size_t cache_aligned = address & ~(line_size-1);
      // Fully associative cache

      AddressMap::iterator it = address_map.find(cache_aligned);
      if (it != address_map.end()) {
        if (tags[it->second] != cache_aligned) {
          throw "ERROR: inconsistency!";
        }
        recently_used[it->second] = true;
        num_hits++;
        return;
      }

      num_misses++;
      // Pass it upstream too.
      if (next_level) { next_level->fetch(ptr); }
      // Technically you'd now wait for the answer, we assume it comes back instantly.

      // Nope, not in the cache. Evict someone that isn't recently_used
      while (true) {
        for (; clock_ptr < num_lines; clock_ptr++) {
          // Not recently used, replace them.
          if (!recently_used[clock_ptr]) {
            // Erase the current entry
            size_t current_entry = tags[clock_ptr];
            if (current_entry != static_cast<size_t>(-1)) {
              address_map.erase(current_entry);
            }
            // Insert the new one
            tags[clock_ptr] = cache_aligned;
            address_map[cache_aligned] = clock_ptr;
            recently_used[clock_ptr] = true;
            return;
          }
          // They were used before the last miss but now we've gotten to them.
          recently_used[clock_ptr] = false;
        }
        // Reset the clock
        clock_ptr = 0;
      }
    }

    void printStats() {
      size_t num_requests = num_hits + num_misses;
      double hit_rate = static_cast<double>(num_hits) / static_cast<double>(num_requests);
      printf("%9zu misses, %9zu hits, %9zu requests, %5.3f hit rate\n", num_misses, num_hits, num_requests, hit_rate);
    }

    int total_size;
    int line_size;
    int num_lines; // num_lines = total_size / line_size

    size_t* tags;
    bool* recently_used;
    int clock_ptr;
    AddressMap address_map;

    size_t num_misses;
    size_t num_hits;

    CacheLevel* next_level;
  };

  // My mbp is 6MB, my nehalem is 8MB a nehalem EX is 24MB...
  SimpleCache() {
    l3_cache = new CacheLevel(8 * 1024 * 1024, 128, NULL);
    l2_cache = new CacheLevel(256 * 1024, 128, l3_cache);
    l1_cache = new CacheLevel(32 * 1024, 32, l2_cache);
  }
  ~SimpleCache() {
    delete l1_cache;
    delete l2_cache;
    delete l3_cache;
  }

  void fetch(const void* ptr) {
    l1_cache->fetch(ptr);
  }

  void printStats() {
    if (l1_cache->num_misses) {
      printf("L1Cache: "); l1_cache->printStats();
      printf("L2Cache: "); l2_cache->printStats();
      printf("L3Cache: "); l3_cache->printStats();
    }
  }

  CacheLevel* l1_cache;
  CacheLevel* l2_cache;
  CacheLevel* l3_cache;
};

#endif // _LISZT_CACHE_SIM_H_
