#ifndef _LISZT_RUNTIME_H
#define _LISZT_RUNTIME_H

#include <stdint.h>
#include <stddef.h>
#include <string>

enum lElementType {
	L_VERTEX = 0,
	L_CELL = 1,
	L_EDGE = 2,
	L_FACE = 3,
};
enum {
	L_ELEMENT_TYPE_SIZE = 4
};
enum lType {
	L_INT,
	L_FLOAT,
	L_DOUBLE,
	L_BOOL,
	L_STRING
};

enum lReduction {
    L_ASSIGN,
	L_PLUS,
	L_MINUS,
	L_MULTIPLY,
	L_DIVIDE,
	L_MIN,
	L_MAX,
	L_BOR,
	L_BAND,
	L_XOR,
	L_AND,
	L_OR
};

enum lPhase {
	L_READ_ONLY,
	L_WRITE_ONLY,
	L_MODIFY,
	L_REDUCE_PLUS,
	L_REDUCE_MULTIPLY,
	L_REDUCE_MIN,
	L_REDUCE_MAX,
	L_REDUCE_BOR,
	L_REDUCE_BAND,
	L_REDUCE_XOR,
	L_REDUCE_AND,
	L_REDUCE_OR
};
struct lContext;
struct lField;
struct lScalar;
struct lSet;

struct lFields;
/* 
{
   lField flux;
   lField temp;
   .... //compiler will will this in
};
*/
struct lScalars;
/* 
{
   lScalar error;
   .... //compiler will will this in
};
*/
struct lSets;
/*
{
   lSet boundary;
   ..... //compiler will will this in
}*/

struct lkField;
struct lkFields;

struct lkScalar;
struct lkScalars;

struct lkContext;

//for efficiency we define lkElement explicitly for all runtimes as a 32-bit value
struct lkElement { uint32_t data; };

struct lkSet;
struct lkIterator;

struct lsContext;
struct lsElement;
struct lsSet;
struct lsIterator;

struct lsFunctionTable;

struct lStencilData {
	void (*stencil_fn)(lsFunctionTable *,lsContext*);
	bool is_trivial;
};

//Modifiers for runtime functions, GPU modifiers are set rather than in GPU specific code to simplify definition order for C code
#ifdef __CUDACC__
#define L_RUNTIME_NESTED __device__
#define L_RUNTIME_UNNESTED __host__
#define L_RUNTIME_STENCIL
#define L_RUNTIME_ALL __device__ __host__
#else
#define L_RUNTIME_NESTED
#define L_RUNTIME_UNNESTED
#define L_RUNTIME_STENCIL
#define L_RUNTIME_ALL
#endif

//Unnested Runtime Calls

struct lProgramArguments {
	const char * mesh_file;
	bool redirect_output_to_log;
};

void lExec(void (*entry_point)(lContext*),void (*entry_stencil)(lsFunctionTable*,lsContext*), lProgramArguments * arguments, size_t n_fields, size_t n_sets, size_t n_scalars);

L_RUNTIME_UNNESTED lFields * lGetFields(lContext * ctx);
L_RUNTIME_UNNESTED lSets * lGetSets(lContext * ctx);
L_RUNTIME_UNNESTED lScalars * lGetScalars(lContext * ctx);

L_RUNTIME_UNNESTED void lFieldInit(lContext * ctx, lField * field, int id, lElementType key_type, lType val_type, size_t val_length);
L_RUNTIME_UNNESTED void lFieldBroadcast(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, void * data);
L_RUNTIME_UNNESTED void lFieldLoadData(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, const char * key);
L_RUNTIME_UNNESTED void lFieldSaveData(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, const char * key);


L_RUNTIME_UNNESTED void lFieldEnterPhase(lField * field, lType val_type, size_t val_length, lPhase phase);

L_RUNTIME_UNNESTED void lScalarInit(lContext * ctx, lScalar * scalar, lType val_type, size_t val_length);

L_RUNTIME_UNNESTED void lScalarRead(lContext * ctx, lScalar * scalar, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * result);
L_RUNTIME_UNNESTED void lScalarWrite(lContext * ctx, lScalar * scalar, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value);

L_RUNTIME_UNNESTED void lScalarEnterPhase(lScalar * scalar, lType val_type, size_t val_length, lPhase phase);

L_RUNTIME_UNNESTED void lSetInitBoundary(lContext * ctx, lSet * set, lElementType type, const char * boundary_name);
L_RUNTIME_UNNESTED size_t lSetSize(lContext * ctx, lSet * set);

L_RUNTIME_UNNESTED void lKernelRun(lContext * ctx,  lSet * set, lElementType typ, int id, void (*kernel)(lkContext), lStencilData data);
/*kernels have the signature and structure: 
void kernel_name(lkContext ctx_) {
	lkContext * ctx = &ctx_;
	lkElement e;
	if(lkGetActiveElement(ctx,&e)) {
	  <...>
	}
}*/

L_RUNTIME_UNNESTED void lPrintBegin(lContext * ctx);
L_RUNTIME_UNNESTED void lPrintEnd(lContext * ctx);
L_RUNTIME_UNNESTED void lPrintValue(lContext * ctx, lType typ, size_t r, size_t c, const void * value);
L_RUNTIME_UNNESTED double lWallTime(lContext * ctx); //counter, time reported in seconds

//Nested Runtime Calls
L_RUNTIME_NESTED void lkPrintBegin(lkContext * ctx);
L_RUNTIME_NESTED void lkPrintEnd(lkContext * ctx);
L_RUNTIME_NESTED void lkPrintValue(lkContext * ctx, lType typ, size_t r, size_t c, const void * value);
L_RUNTIME_NESTED void lkPrintElement(lkContext * ctx, lElementType typ, lkElement);

L_RUNTIME_NESTED lkFields * lkGetFields(lkContext * ctx);
L_RUNTIME_NESTED lkScalars * lkGetScalars(lkContext * ctx);
L_RUNTIME_NESTED bool lkGetActiveElement(lkContext * ctx, lkElement * e);

L_RUNTIME_NESTED void lkFieldRead(lkField * scalar, lkElement e, lType element_type, size_t element_length, size_t val_offset, size_t val_length,  void * result);
L_RUNTIME_NESTED void lkFieldWrite(lkField * scalar, lkElement e, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value);

L_RUNTIME_NESTED void lkScalarRead(lkContext * ctx, lkScalar * scalar, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * result);
L_RUNTIME_NESTED void lkScalarWrite(lkContext * ctx, lkScalar * scalar, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value);

L_RUNTIME_NESTED size_t lkSetSize(lkContext * ctx, lkSet * set);

L_RUNTIME_NESTED void lkSetGetIterator(lkContext * ctx, lkSet * set, lElementType typ, lkIterator * iterator);
L_RUNTIME_NESTED bool lkIteratorNext(lkContext * ctx, lkIterator * iterator, lElementType typ, lkElement * e, int * lbl);

L_RUNTIME_NESTED bool operator==(lkElement a, lkElement b);
L_RUNTIME_NESTED static inline bool operator!=(lkElement a, lkElement b) { return !(a == b); }

//now all the topological functions ....

L_RUNTIME_UNNESTED void lVerticesOfMesh(lContext * ctx, lSet * set);
L_RUNTIME_UNNESTED void lEdgesOfMesh(lContext * ctx, lSet * set);
L_RUNTIME_UNNESTED void lFacesOfMesh(lContext * ctx, lSet * set);
L_RUNTIME_UNNESTED void lCellsOfMesh(lContext * ctx, lSet * set);

L_RUNTIME_NESTED void lkVerticesOfVertex(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkVerticesOfEdge(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkVerticesOfFace(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkVerticesOfCell(lkContext * ctx, lkElement c, lkSet * set);

L_RUNTIME_NESTED void lkCellsOfVertex(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkCellsOfEdge(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkCellsOfFace(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkCellsOfCell(lkContext * ctx, lkElement c, lkSet * set);

L_RUNTIME_NESTED void lkEdgesOfVertex(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkEdgesOfFace(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkEdgesOfCell(lkContext * ctx, lkElement c, lkSet * set);

L_RUNTIME_NESTED void lkEdgesOfFaceCCW(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkEdgesOfFaceCW(lkContext * ctx, lkElement c, lkSet * set);

L_RUNTIME_NESTED void lkFacesOfVertex(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkFacesOfEdge(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkFacesOfCell(lkContext * ctx, lkElement c, lkSet * set);

L_RUNTIME_NESTED void lkFacesOfEdgeCCW(lkContext * ctx, lkElement c, lkSet * set);
L_RUNTIME_NESTED void lkFacesOfEdgeCW(lkContext * ctx, lkElement c, lkSet * set);


L_RUNTIME_NESTED lkElement lkHeadOfEdge(lkContext * ctx, lkElement c);
lkElement lkTailOfEdge(lkContext * ctx, lkElement c);

L_RUNTIME_NESTED lkElement lkHeadOfFace(lkContext * ctx, lkElement c);
L_RUNTIME_NESTED lkElement lkTailOfFace(lkContext * ctx, lkElement c);

L_RUNTIME_NESTED lkElement lkFlipEdge(lkContext * ctx, lkElement c);
L_RUNTIME_NESTED lkElement lkFlipFace(lkContext * ctx, lkElement c);

L_RUNTIME_NESTED lkElement lkTowardsEdge(lkContext * ctx, lkElement e, lkElement v);
L_RUNTIME_NESTED lkElement lkTowardsFace(lkContext * ctx, lkElement e, lkElement v);

L_RUNTIME_NESTED int lkIDOfVertex(lkContext * ctx, lkElement e);
L_RUNTIME_NESTED int lkIDOfEdge(lkContext * ctx, lkElement e);
L_RUNTIME_NESTED int lkIDOfFace(lkContext * ctx, lkElement e);
L_RUNTIME_NESTED int lkIDOfCell(lkContext * ctx, lkElement e);


L_RUNTIME_NESTED lkElement lkVertexOfCellWithLabel(lkContext * ctx, lkElement e, int l);
L_RUNTIME_NESTED lkElement lkEdgeOfCellWithLabel(lkContext * ctx, lkElement e, int l);
L_RUNTIME_NESTED lkElement lkFaceOfCellWithLabel(lkContext * ctx, lkElement e, int l);
L_RUNTIME_NESTED lkElement lkCellOfCellWithLabel(lkContext * ctx, lkElement e, int l);


L_RUNTIME_NESTED lkElement lkVertexOfFaceWithLabel(lkContext * ctx, lkElement e, int l);
L_RUNTIME_NESTED lkElement lkEdgeOfFaceWithLabel(lkContext * ctx, lkElement e, int l);
L_RUNTIME_NESTED lkElement lkCellOfFaceWithLabel(lkContext * ctx, lkElement e, int l);

L_RUNTIME_NESTED lkElement lkCellOfEdgeWithLabel(lkContext * ctx, lkElement e, int l);
L_RUNTIME_NESTED lkElement lkFaceOfEdgeWithLabel(lkContext * ctx, lkElement e, int l);

//stencil generations functions
struct lsFunctionTable {
	L_RUNTIME_STENCIL void (*lsSetGetIterator)(lsContext * ctx, lsSet * set, lsIterator * iterator);
	L_RUNTIME_STENCIL bool (*lsIteratorNext)(lsContext * ctx, lsIterator * iterator, lsElement * e);
	L_RUNTIME_STENCIL void (*lsFieldAccess)(lsContext * ctx, int field_id, lElementType type, lsElement * e, lPhase phase);
	L_RUNTIME_STENCIL void (*lsSetInitBoundary)(lsContext * ctx,  lsSet * set, lElementType type, const char * boundary_name);
	L_RUNTIME_STENCIL void (*lsGetActiveElement)(lsContext * ctx, lsElement * e);
	
	//stencil topology functions
	L_RUNTIME_STENCIL void (*lsVerticesOfMesh)(lsContext * ctx, lsSet * set);
	L_RUNTIME_STENCIL void (*lsEdgesOfMesh)(lsContext * ctx, lsSet * set);
	L_RUNTIME_STENCIL void (*lsFacesOfMesh)(lsContext * ctx, lsSet * set);
	L_RUNTIME_STENCIL void (*lsCellsOfMesh)(lsContext * ctx, lsSet * set);
	
	L_RUNTIME_STENCIL void (*lsVerticesOfVertex)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsVerticesOfEdge)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsVerticesOfFace)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsVerticesOfCell)(lsContext * ctx, lsElement * c, lsSet * set);
	
	L_RUNTIME_STENCIL void (*lsCellsOfVertex)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsCellsOfEdge)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsCellsOfFace)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsCellsOfCell)(lsContext * ctx, lsElement * c, lsSet * set);
	
	L_RUNTIME_STENCIL void (*lsEdgesOfVertex)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsEdgesOfFace)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsEdgesOfCell)(lsContext * ctx, lsElement * c, lsSet * set);
	
	L_RUNTIME_STENCIL void (*lsEdgesOfFaceCCW)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsEdgesOfFaceCW)(lsContext * ctx, lsElement * c, lsSet * set);
	
	L_RUNTIME_STENCIL void (*lsFacesOfVertex)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsFacesOfEdge)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsFacesOfCell)(lsContext * ctx, lsElement * c, lsSet * set);
	
	L_RUNTIME_STENCIL void (*lsFacesOfEdgeCCW)(lsContext * ctx, lsElement * c, lsSet * set);
	L_RUNTIME_STENCIL void (*lsFacesOfEdgeCW)(lsContext * ctx, lsElement * c, lsSet * set);
	
	
	L_RUNTIME_STENCIL void (*lsHeadOfEdge)(lsContext * ctx, lsElement * c, lsElement * result);
	L_RUNTIME_STENCIL void (*lsTailOfEdge)(lsContext * ctx, lsElement * c, lsElement * result);
	
	L_RUNTIME_STENCIL void (*lsHeadOfFace)(lsContext * ctx, lsElement * c, lsElement * result);
	L_RUNTIME_STENCIL void (*lsTailOfFace)(lsContext * ctx, lsElement * c, lsElement * result);
	
	L_RUNTIME_STENCIL void (*lsFlipEdge)(lsContext * ctx, lsElement * c, lsElement * result);
	L_RUNTIME_STENCIL void (*lsFlipFace)(lsContext * ctx, lsElement * c, lsElement * result);
	
	L_RUNTIME_STENCIL void (*lsTowardsEdge)(lsContext * ctx, lsElement * e, lsElement * v, lsElement * result);
	L_RUNTIME_STENCIL void (*lsTowardsFace)(lsContext * ctx, lsElement * e, lsElement * v, lsElement * result);
	L_RUNTIME_STENCIL void (*lsSetExtractElement)(lsContext * ctx, lsSet * set, uint32_t elem, lsElement * result);
};

#endif