
#include <sys/time.h>
#include <stdio.h>
#include <iostream>
#include <assert.h>
#include <cuda.h>
#include <map>

#include <thrust/reduce.h>
#include <thrust/fill.h>
#include <thrust/functional.h>
#include <thrust/iterator/counting_iterator.h>

#include "liszt_runtime.h"
#include "boundary_set_reader.h"
#include "mesh.h"
#include "runtime_util.h"
#include "coloring.h"
#include "nested_topology.h"

#define L_KERNEL __global__
#define L_NESTED L_ALWAYS_INLINE __device__
#define L_UNNESTED __host__
#define L_STENCIL

#define KERNEL_BLOCKSIZE (256)

#define PRINT_BUFFER_ELEMENT_SIZE (200000 * 128)
#define PRINT_BUFFER_DATA_SIZE (1024 * 1024 * 32)

//this really should be passed into lExec ...
#define SCALAR_MAX_SIZE (sizeof(double) * 22)
#define SCALAR_BYTE_ALIGNMENT (8)
//define opaque data types and implementation for specific runtime

#define CUDA_DO(x) do { \
	cudaError_t err = (x); \
	if(cudaSuccess != err) { \
		fprintf(stderr, "%s:%d: %s\n",__FILE__,__LINE__, cudaGetErrorString(err)); \
		abort(); \
	} \
} while(0)


typedef unsigned char byte;


struct DPrintContext {
	lElementType type;
	int id;
	int count;	
	DPrintContext * child;
	DPrintContext * parent;
};
struct PrintBuffer;

struct lkContext {
	lkFields * fields;
	lkScalars * scalars;
	Mesh * mesh;
	size_t n_elements_max;
	
	size_t n_elements;
	
	size_t first_element;
	uint32_t * colored_elements; //if this is null, then use first_element + idx as the mesh element, otherwise using colored_element[id] as the index

#ifndef NDEBUG
	size_t print_count;
	PrintBuffer * print_buffer;
	DPrintContext print_context;
	DPrintContext * current_print_context;
	int print_context_top_level_count;
#endif
};

struct lContext {
	lFields * fields;
	lSets * sets;
	lScalars * scalars;
	
	MeshIO::LisztFileReader mesh_reader; //we hang on to this so we can load the positions on demand
	Mesh mesh;
	size_t n_elements_max; //max(nV,nE,nF,nC)
	BoundarySetReader boundary_reader;
	lkElement active_element;
#ifndef NDEBUG
	size_t print_context_top_level_count;
#endif
	bool scalar_dirty;
	
	lkContext nested_template;
	
	size_t n_scalars;
	size_t scalar_data_bump_pointer;
	byte * scalar_data;
	byte * device_scalar_data;

	//key is (set_id,kernel_id)
	std::map< std::pair<uint32_t,uint32_t>, lColoringResult> coloring_cache;

	cudaEvent_t start_event,end_event;
};

struct lField {
	byte * device_data;
	lkField * device_field;
};


struct lScalar {
	byte * data;
	size_t max_dirty_elements;
	lPhase phase;
	lType type;
	size_t length;
	byte * device_data;
	byte * device_write_table;
	lkScalar * device_scalar;
	lContext * ctx;
};

struct lSet {
	BoundarySetReader::range_list_t ranges;
	int id; //unique identifier for this set (complete mesh sets have a negative number (e.g. -L_VERTEX), boundaries are numbered as they found in the file)
	size_t size;
};

struct lkField {
	byte * data;
};

struct lkScalar {
	byte * data;
	byte * write_table;
};

struct PrintElement {
	lType type;
	uint64_t order;
	char rows;
	char cols;
	uint32_t offset; //offset into PrintBuffer data
};

struct PrintSizes {
	uint32_t n_bytes;
	uint32_t n_elements;
};

struct PrintBuffer {
	PrintSizes sizes;
	PrintElement elements[PRINT_BUFFER_ELEMENT_SIZE];
	byte data[PRINT_BUFFER_DATA_SIZE];
};


struct lkSet {
	lNestedSet set;
};

struct lkIterator {
	lNestedIterator it;
#ifndef NDEBUG
	DPrintContext print_context;
#endif
};



L_ALWAYS_INLINE 
size_t numberOfElementsOfType(lContext * ctx, lElementType typ) {
	switch(typ) {
		case L_VERTEX: return ctx->mesh.nvertices;
		case L_EDGE: return ctx->mesh.nedges;
		case L_FACE: return ctx->mesh.nfaces;
		case L_CELL: return ctx->mesh.ncells;
	}
	return 0;
}

void copyCRS(size_t n_elems, CRS * dst, CRS * src) {
	size_t n_values = src->row_idx[n_elems];

	//size of row_idx is (n_elems + 1) since it holds the total size in row_idx[n_elems]
	CUDA_DO(cudaMalloc(&dst->row_idx, (n_elems + 1) * sizeof(uint32_t)));
	CUDA_DO(cudaMalloc(&dst->values,n_values * sizeof(uint32_t)));
	CUDA_DO(cudaMemcpy(dst->row_idx,src->row_idx, (n_elems + 1) * sizeof(uint32_t),cudaMemcpyHostToDevice));
	CUDA_DO(cudaMemcpy(dst->values,src->values, n_values * sizeof(uint32_t),cudaMemcpyHostToDevice));
	
}
void copyCRSConst(size_t n_elems, CRSConst * dst, CRSConst * src) {
	CUDA_DO(cudaMalloc(&dst->values,2 * n_elems * sizeof(uint32_t)));
	CUDA_DO(cudaMemcpy(dst->values,src->values, 2 * n_elems * sizeof(uint32_t),cudaMemcpyHostToDevice));
}
//dst is a host pointer to a mesh structure that will be copied to the device
//all pointers inside 'dst' should be device pointers
void initDeviceMesh(Mesh * dst, Mesh * src) {
	dst->nvertices = src->nvertices;
	dst->nedges = src->nedges;
	dst->nfaces = src->nfaces;
	dst->ncells = src->ncells;
	
	copyCRS(src->nvertices, &dst->vtov, &src->vtov);
	copyCRS(src->nvertices, &dst->vtoe, &src->vtoe);
	copyCRS(src->nvertices, &dst->vtof, &src->vtof);
	copyCRS(src->nvertices, &dst->vtoc, &src->vtoc);
	
	copyCRSConst(src->nedges, &dst->etov, &src->etov);
	copyCRS(src->nedges, &dst->etof, &src->etof);
	copyCRS(src->nedges, &dst->etoc, &src->etoc);

	copyCRSConst(src->nfaces, &dst->ftoc, &src->ftoc);
	copyCRS(src->nfaces, &dst->ftoe, &src->ftoe);
	copyCRS(src->nfaces, &dst->ftov, &src->ftov);

	copyCRS(src->ncells, &dst->ctov, &src->ctov);
	copyCRS(src->ncells, &dst->ctoe, &src->ctoe);
	copyCRS(src->ncells, &dst->ctof, &src->ctof);
	copyCRS(src->ncells, &dst->ctoc, &src->ctoc);
}

void kernelSize(size_t n_elems, size_t * n_blocks, size_t * block_size) {
	*n_blocks = (n_elems + KERNEL_BLOCKSIZE - 1) / KERNEL_BLOCKSIZE;
	*block_size = KERNEL_BLOCKSIZE;
}
//Unnested Runtime Calls

void printBufferClearSizes(PrintBuffer * buffer) {
	PrintSizes * sizes = &buffer->sizes;
	PrintSizes local = { 0, 0 };
	CUDA_DO(cudaMemcpy(sizes,&local,sizeof(PrintSizes), cudaMemcpyHostToDevice));
}

L_RUNTIME_UNNESTED
void lExec(void (*entry_point)(lContext*),void (*entry_stencil)(lsFunctionTable*,lsContext*), lProgramArguments * arguments, size_t n_fields, size_t n_sets, size_t n_scalars) {
	lContext ctx;
	
	//allocate memory to hold global state
	ctx.fields = (lFields*) malloc(sizeof(lField) * n_fields);
	ctx.sets = (lSets*) malloc(sizeof(lSet) * n_sets);
	ctx.scalars = (lScalars*) malloc(sizeof(lScalar) * n_scalars);
	ctx.scalar_dirty = false;
	
	//load mesh and boundary sets
	ctx.mesh_reader.init(arguments->mesh_file);
	MeshIO::FileFacetEdge * ffe = ctx.mesh_reader.facetEdges();
	const MeshIO::LisztHeader & h = ctx.mesh_reader.header();
	
	MeshIO::FacetEdgeBuilder builder;
	builder.init(h.nV,h.nE,h.nF,h.nC,h.nFE);
	builder.insert(0,h.nFE,ffe);
	//builder.validateFullMesh();
	ctx.mesh_reader.free(ffe);
	lMeshInitFromFacetEdgeBuilder(&ctx.mesh,&builder);
	ctx.boundary_reader.init(h.nBoundaries,ctx.mesh_reader.boundaries());
	ctx.n_elements_max = std::max(std::max(h.nV,h.nE),std::max(h.nF,h.nC));
	//initialize device-side state
	ctx.nested_template.n_elements_max = ctx.n_elements_max;
	CUDA_DO(cudaMalloc(&ctx.nested_template.fields, sizeof(lkField) * n_fields));
	CUDA_DO(cudaMalloc(&ctx.nested_template.scalars, sizeof(lkScalar) * n_scalars));
	Mesh device_mesh;
	initDeviceMesh(&device_mesh, &ctx.mesh); 
	
	CUDA_DO(cudaMalloc(&ctx.nested_template.mesh, sizeof(Mesh)));
	CUDA_DO(cudaMemcpy(ctx.nested_template.mesh,&device_mesh, sizeof(Mesh), cudaMemcpyHostToDevice));
	
#ifndef NDEBUG
	ctx.print_context_top_level_count = 0;
	CUDA_DO(cudaMalloc(&ctx.nested_template.print_buffer, sizeof(PrintBuffer)));
	printBufferClearSizes(ctx.nested_template.print_buffer);
	ctx.nested_template.print_count = 0;
#endif

	ctx.n_scalars = n_scalars;
	ctx.scalar_data_bump_pointer = 0;
	ctx.scalar_data = (byte*) malloc(SCALAR_MAX_SIZE * n_scalars);
	CUDA_DO(cudaMalloc(&ctx.device_scalar_data,SCALAR_MAX_SIZE * n_scalars));
	//launch the program
	entry_point(&ctx);
	
	//do cleanup
}
L_ALWAYS_INLINE L_RUNTIME_UNNESTED
lFields * lGetFields(lContext * ctx) {
	return ctx->fields;
}
L_ALWAYS_INLINE L_RUNTIME_UNNESTED
lSets * lGetSets(lContext * ctx) {
	return ctx->sets;
}
L_ALWAYS_INLINE L_RUNTIME_UNNESTED
lScalars * lGetScalars(lContext * ctx) {
	return ctx->scalars;
}

L_RUNTIME_UNNESTED
void lFieldInit(lContext * ctx, lField * field, int id, lElementType key_type, lType val_type, size_t val_length) {
	size_t n_elems = numberOfElementsOfType(ctx,key_type);
	size_t field_idx = field - (lField *) ctx->fields;
	field->device_field =  (lkField*)ctx->nested_template.fields + field_idx;
	lkField device_field;
	CUDA_DO(cudaMalloc(&field->device_data,n_elems * val_length * lUtilTypeSize(val_type)));
	device_field.data = field->device_data;
	CUDA_DO(cudaMemcpy(field->device_field,&device_field, sizeof(lkField), cudaMemcpyHostToDevice));
}


__global__ void broadcast_kernel(lkField * field, size_t n_field_elements,lElementType key_type, lType val_type, size_t val_length, byte * data) {
	int elem_id = blockIdx.x * blockDim.x + threadIdx.x;
	if(elem_id < n_field_elements) {
		lkElement e = {elem_id};
		lkFieldWrite(field,e,L_ASSIGN, val_type, val_length, 0, val_length, data);
	}
}

//this is total unoptimized, currently only expect to use it on field initialization
L_ALWAYS_INLINE L_RUNTIME_UNNESTED
void lFieldBroadcast(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, void * data) {
	size_t size = lUtilTypeSize(val_type) * val_length;
	
	byte * device_data;
	CUDA_DO(cudaMalloc(&device_data,size));
	CUDA_DO(cudaMemcpy(device_data,data,size,cudaMemcpyHostToDevice));
	
	size_t n_elems = numberOfElementsOfType(ctx,key_type);
	size_t n_blocks,n_threads;
	kernelSize(n_elems,&n_blocks,&n_threads);
	broadcast_kernel<<<n_blocks,n_threads>>>(field->device_field,n_elems,key_type,val_type,val_length,device_data);
	CUDA_DO(cudaGetLastError());
	cudaThreadSynchronize(); //does cudaFree go onto the normal push buffer? or do we need to syncrhonize here to avoid freeing before the kernel completes?
	CUDA_DO(cudaFree(device_data));
}

L_RUNTIME_UNNESTED void lFieldLoadData(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, const char * key) {
    
	if (!strcmp(key,"position")) {
        assert(key_type == L_VERTEX);
        assert(val_type == L_DOUBLE || val_type == L_FLOAT);
    }
    size_t  nelems    = numberOfElementsOfType(ctx, key_type);
    size_t  type_size = lUtilTypeSize(val_type);
    byte   *data      = (byte *) ctx->mesh_reader.fieldData(key_type, val_type, val_length, key);
    CUDA_DO(cudaMemcpy(field->device_data, data, val_length * type_size * nelems, cudaMemcpyHostToDevice));
    free(data);
}
L_RUNTIME_UNNESTED void lFieldSaveData(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, const char * key) {
	lFieldEnterPhase(field,val_type, val_length, L_READ_ONLY);
	printf("TODO: save field %s\n",key);
}


L_RUNTIME_UNNESTED
void lFieldEnterPhase(lField * field, lType val_type, size_t val_length, lPhase phase) {
	//nop
}

L_RUNTIME_UNNESTED
void lScalarInit(lContext * ctx, lScalar * scalar, lType val_type, size_t val_length) {
	size_t size = lUtilTypeSize(val_type) * val_length;
	size_t aligned_size = ((size + SCALAR_BYTE_ALIGNMENT - 1) / SCALAR_BYTE_ALIGNMENT) * SCALAR_BYTE_ALIGNMENT;
	
	
	scalar->ctx = ctx;
	scalar->data = ctx->scalar_data + ctx->scalar_data_bump_pointer;
	scalar->max_dirty_elements = 0;
	scalar->phase = L_READ_ONLY;
	scalar->type = val_type;
	scalar->length = val_length;
	scalar->device_data = ctx->device_scalar_data + ctx->scalar_data_bump_pointer;
	
	CUDA_DO(cudaMalloc(&scalar->device_write_table, ctx->n_elements_max * size));

	size_t idx = (scalar - (lScalar*)ctx->scalars);
	scalar->device_scalar = (lkScalar*) ctx->nested_template.scalars + idx;
	
	lkScalar temp;
	temp.data = scalar->device_data;
	temp.write_table = scalar->device_write_table;
	
	CUDA_DO(cudaMemcpy(scalar->device_scalar, &temp, sizeof(lkScalar), cudaMemcpyHostToDevice));

	ctx->scalar_data_bump_pointer += aligned_size;
}

L_ALWAYS_INLINE L_RUNTIME_UNNESTED
void lScalarRead(lContext * ctx, lScalar * scalar, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * result) {
	assert(scalar->phase == L_READ_ONLY);
	size_t scalar_size = lUtilTypeSize(element_type);
	memcpy(result,scalar->data + val_offset * scalar_size, scalar_size * val_length);
}

L_ALWAYS_INLINE L_RUNTIME_UNNESTED
void lScalarWrite(lContext * ctx, lScalar * scalar, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value) {
	assert(scalar->phase == L_READ_ONLY);
	size_t scalar_size = lUtilTypeSize(element_type);
	lUtilValueReduce(scalar->data + val_offset * scalar_size, reduction, element_type, val_length, value);
	scalar->ctx->scalar_dirty = true;
}

template<typename Type, typename Reduction>
L_ALWAYS_INLINE
void reduceWithThrustTemplate(void * begin, size_t n_elements, size_t n_elements_stride, size_t length, Type identity, Reduction reduction, byte * result) {
	for(int i = 0; i < length; i++) {
		Type * start = (Type*)begin + n_elements_stride * i;
		
		/*
		printf("reducing %d,%d\n",(int)n_elements_stride * i, (int) n_elements);
		Type * foo = new Type[n_elements];
		cudaMemcpy(foo,start,sizeof(Type) * n_elements, cudaMemcpyDeviceToHost);
		for(int i = 0; i < n_elements; i++) {
			printf("%d: %d\n",i,(int)foo[i]);
		}
		*/

		::thrust::device_ptr< Type > t_begin(start);
		::thrust::device_ptr< Type > t_end = t_begin + n_elements;
		((Type*)result)[i] = reduction(((Type*)result)[i],::thrust::reduce(t_begin, t_end, identity, reduction));
	}
}

L_ALWAYS_INLINE
void reduceWithThrust(void * begin, size_t n_elements, size_t n_elements_stride, lType val_type, size_t length, lPhase phase, byte * result) {

	#define DO(r,t,i) reduceWithThrustTemplate(begin, n_elements, n_elements_stride, length, i, ::thrust :: r < t >(), result); break;
	switch(phase) {
		case L_WRITE_ONLY: /*fallthrough*/
		case L_REDUCE_PLUS:
		    switch(val_type) {
		    	case L_INT: DO(plus,int,0)
		    	case L_FLOAT: DO(plus,float,0.f)
		    	case L_DOUBLE: DO(plus,double,0.0)
		    } break;
		case L_REDUCE_MULTIPLY:
		    switch(val_type) {
		    	case L_INT: DO(multiplies,int,1)
		    	case L_FLOAT: DO(multiplies,float,1.f)
		    	case L_DOUBLE: DO(multiplies,double,1.0)
		    } break;
		case L_REDUCE_MIN:
			switch(val_type) {
		    	case L_INT: DO(minimum,int,INT_MAX)
		    	case L_FLOAT: DO(minimum,float,FLT_MAX)
		    	case L_DOUBLE: DO(minimum,double,DBL_MAX)
		    } break;
		case L_REDUCE_MAX:
			switch(val_type) {
		    	case L_INT: DO(maximum,int,INT_MIN)
		    	case L_FLOAT: DO(maximum,float,FLT_MIN)
		    	case L_DOUBLE: DO(maximum,double,DBL_MIN)
		    } break;
		case L_REDUCE_BOR:
			switch(val_type) {
		    	case L_INT: DO(bit_or,int,0)
		    } break;
		case L_REDUCE_BAND:
			switch(val_type) {
		    	case L_INT: DO(bit_and,int,~0)
		    } break;
		case L_REDUCE_XOR:
			switch(val_type) {
		    	case L_INT: DO(bit_xor,int,0)
		    } break;
		case L_REDUCE_AND:
			switch(val_type) {
		    	case L_INT: DO(bit_and,int,1)
		    	case L_BOOL: DO(bit_and,char,true)
		    } break;
		case L_REDUCE_OR:
			switch(val_type) {
		    	case L_INT: DO(bit_or,int,0)
		    	case L_BOOL: DO(bit_or,char,false)
		    } break;
	}
	#undef DO
}

void scalarSetMaxDirtyElements(lScalar * scalar, size_t n_elems) {
	if(scalar->phase != L_READ_ONLY && scalar->max_dirty_elements < n_elems) {
		lUtilValue v;
		lUtilValueIdentity(scalar->phase, scalar->type,&v);
		
		for(int i = 0; i < scalar->length; i++) {
			size_t begin = (scalar->ctx->n_elements_max * i) + scalar->max_dirty_elements;
			size_t size = n_elems - scalar->max_dirty_elements;
			//printf("filling: %d, %d\n",(int)begin,(int)n_elems);
			switch(scalar->type) {
				case L_INT: {
					::thrust::device_ptr<int> data( ((int*)scalar->device_write_table) + begin);
					:: thrust :: fill_n( data , size, v.i);	
				} break;
				case L_BOOL: {
					::thrust::device_ptr<bool> data( ((bool*)scalar->device_write_table) + begin);
					:: thrust :: fill_n( data , size, v.b);	
				} break;
				case L_FLOAT: {
					::thrust::device_ptr<float> data( ((float*)scalar->device_write_table) + begin);
					:: thrust :: fill_n( data , size, v.f);	
				} break;
				case L_DOUBLE: {
					::thrust::device_ptr<double> data( ((double*)scalar->device_write_table) + begin);
					:: thrust :: fill_n( data , size, v.d);	
				} break;
				case L_STRING: 
					/* nop */	
					break;
			}
		}
		scalar->max_dirty_elements = n_elems;
	}

}
void scalarsSetMaxDirtyElements(lContext * ctx, size_t n_elems) {
	for(size_t j = 0; j < ctx->n_scalars; j++) {
		lScalar * s = (lScalar*) ctx->scalars + j;
		scalarSetMaxDirtyElements(s,n_elems);
	}
}

L_RUNTIME_UNNESTED
void lScalarEnterPhase(lScalar * scalar, lType val_type, size_t val_length, lPhase phase) {
	if(scalar->phase != phase) {
		
		if(scalar->max_dirty_elements > 0) {
			//get into clean state
		
			reduceWithThrust(scalar->device_write_table,scalar->max_dirty_elements,scalar->ctx->n_elements_max,val_type, val_length, scalar->phase, scalar->data);
			
			scalar->max_dirty_elements = 0;
			//reduction writes to the host value, making the device value stale
			scalar->ctx->scalar_dirty = true;
		}	
		scalar->phase = phase;
	}
}

L_RUNTIME_UNNESTED
void lSetInitBoundary(lContext * ctx, lSet * set, lElementType type, const char * boundary_name) {
	MeshIO::IOElemType etyp;
	switch(type) {
		case L_VERTEX:
			etyp = MeshIO::VERTEX_T;
			break;
		case L_EDGE:
			etyp = MeshIO::EDGE_T;
			break;
		case L_FACE:
			etyp = MeshIO::FACE_T;
			break;
		case L_CELL:
			etyp = MeshIO::CELL_T;
			break;
	}
	ctx->boundary_reader.load(etyp,boundary_name,&set->id,&set->ranges,&set->size);
}

L_ALWAYS_INLINE L_RUNTIME_UNNESTED
size_t lSetSize(lContext * ctx, lSet * set) {
	return set->size;
}


bool elementLess(const PrintElement & lhs, const PrintElement & rhs) {
	return lhs.order < rhs.order;
}

void printBufferFlush(lContext * ctx) {
#ifndef NDEBUG
	PrintBuffer * buffer  = ctx->nested_template.print_buffer;
	
	PrintSizes sizes = {0,0};
	CUDA_DO(cudaMemcpy(&sizes,&buffer->sizes,sizeof(PrintSizes), cudaMemcpyDeviceToHost));
	

	size_t element_size = sizes.n_elements * sizeof(PrintElement);
	PrintElement * elements = (PrintElement*) malloc(element_size);
	byte * data = (byte*) malloc(sizes.n_bytes);

	CUDA_DO(cudaMemcpy(data,&buffer->data[0],sizes.n_bytes,cudaMemcpyDeviceToHost));
	CUDA_DO(cudaMemcpy(elements,&buffer->elements[0],element_size,cudaMemcpyDeviceToHost));
	
	std::sort(elements,elements + sizes.n_elements, elementLess);

	uint32_t smallest_valid_order = ctx->nested_template.print_count;
	for(size_t i = 0; i < sizes.n_elements; i++) {
		PrintElement * e =  &elements[i];
		uint32_t order =  (uint32_t) e->order;
		if(order >= smallest_valid_order) {
			if(e->type == L_STRING)
				std::cout << data + e->offset;
			else {
				//ensure correct alignment regardless of buffer alignment
				size_t n_elements = ((e->rows == 0) ? 1 : e->rows) * ((e->cols == 0) ? 1 : e->cols);
				lUtilValue v[n_elements];
				memcpy(&v, data + e->offset, n_elements * lUtilTypeSize(e->type)); 
				lPrintValue(ctx,e->type, e->rows, e->cols,&v);
			}
			ctx->nested_template.print_count = std::max(ctx->nested_template.print_count,(size_t)order);
		}
	}
	
	printBufferClearSizes(buffer);

	free(elements);
	free(data);
#endif

}



void kernelBeginTiming(lContext * ctx) {
	CUDA_DO(cudaEventCreate(&ctx->start_event));
	CUDA_DO(cudaEventCreate(&ctx->end_event));
	CUDA_DO(cudaEventRecord(ctx->start_event,0));
}
void kernelEndTiming(lContext * ctx,int id) {
	cudaEventRecord(ctx->end_event);
	cudaEventSynchronize(ctx->end_event);
	float elapsed_time;
	cudaEventElapsedTime(&elapsed_time, ctx->start_event, ctx->end_event);
	cudaEventDestroy(ctx->start_event);
	cudaEventDestroy(ctx->end_event);
	//fprintf(stderr,"kernel_%d: %fms\n", id, elapsed_time);
}
L_ALWAYS_INLINE L_RUNTIME_UNNESTED
void lKernelRun(lContext * ctx,  lSet * set, lElementType typ, int id, void (*kernel)(lkContext), lStencilData stencil) {
	
	if(ctx->scalar_dirty) {
		CUDA_DO(cudaMemcpy(ctx->device_scalar_data,ctx->scalar_data,ctx->scalar_data_bump_pointer,cudaMemcpyHostToDevice));
		ctx->scalar_dirty = false;
	}

#ifndef NDEBUG
	ctx->nested_template.print_context.type = typ;
	ctx->nested_template.print_context.parent = NULL;
	ctx->nested_template.print_context.child = NULL;
	ctx->nested_template.print_context_top_level_count = ctx->print_context_top_level_count++;
#endif

	if(stencil.is_trivial) { /* TODO: determine if the coloring for this kernel is trivially 1 color */
		//prepare scalars by inserting identity elements into their write table up to the length of the largest kernel launch
		size_t max_size = 0;
		for(size_t i = 0; i < set->ranges.size(); i++) {
			uint32_t start = set->ranges[i].first;
			uint32_t end = set->ranges[i].second;
			max_size = std::max(max_size,(size_t)end - start);
		}
		
		scalarsSetMaxDirtyElements(ctx,max_size);

		//launch kernels, 1 per range
		
		//kernelBeginTiming(ctx);
		
		for(size_t i = 0; i < set->ranges.size(); i++) {
			uint32_t start = set->ranges[i].first;
			uint32_t end = set->ranges[i].second;
			
			ctx->nested_template.first_element = start;
			ctx->nested_template.n_elements = end - start;
			ctx->nested_template.colored_elements = NULL;

			size_t n_blocks,n_elements;
			kernelSize(end - start, &n_blocks, &n_elements);
			kernel<<<n_blocks,n_elements>>>(ctx->nested_template);
			CUDA_DO(cudaGetLastError());

			printBufferFlush(ctx);
		}
		
		//kernelEndTiming(ctx,id);
		
	} else {
		//coloring, first check to see if we already have a cached data structure
		std::pair<uint32_t,uint32_t> key(set->id,id);
		std::map< std::pair<uint32_t,uint32_t>, lColoringResult >::iterator it = ctx->coloring_cache.find(key),
		                                                                  end = ctx->coloring_cache.end();
		lColoringResult * coloring;
		if(it != end) {
			coloring = &it->second;
		} else {
			coloring =  &ctx->coloring_cache[key]; //insert into cache
			printf("(%d,%d): generating coloring.\n",(int)set->id,(int)id);
			lColoringGenerate(ctx->n_elements_max,&ctx->mesh,set->size,&set->ranges,stencil.stencil_fn,coloring); //calculate coloring
			printf("(%d,%d): finished coloring, found %d colors.\n",(int)set->id,(int)id,(int)coloring->n_colors);
			//transfer value array to gpu
			uint32_t * device_values;
			CUDA_DO(cudaMalloc(&device_values,sizeof(uint32_t) * set->size));
			CUDA_DO(cudaMemcpy(device_values, coloring->values, sizeof(uint32_t) * set->size, cudaMemcpyHostToDevice));
			delete [] coloring->values;
			coloring->values = device_values;
		}

		//prepare scalars by inserting identity elements into their write table up to the length of the largest kernel launch
		size_t max_size = 0;
		for(size_t i = 0; i < coloring->n_colors; i++) {
			uint32_t start = coloring->row_idx[i];
			uint32_t end = coloring->row_idx[i+1];
			max_size = std::max(max_size,(size_t)end - start);
		}

		scalarsSetMaxDirtyElements(ctx,max_size);

		//kernelBeginTiming(ctx);
		
		for(size_t i = 0; i < coloring->n_colors; i++) {
			uint32_t start = coloring->row_idx[i];
			uint32_t end = coloring->row_idx[i+1];

			ctx->nested_template.n_elements = end - start;
			ctx->nested_template.colored_elements = coloring->values + start;
			size_t n_blocks, n_elements;
			kernelSize(end - start, &n_blocks,&n_elements);
			kernel<<<n_blocks,n_elements>>>(ctx->nested_template);
			CUDA_DO(cudaGetLastError());
			printBufferFlush(ctx);
		}
		
		//kernelEndTiming(ctx,id);
		
	}

}

L_RUNTIME_UNNESTED
void lPrintBegin(lContext * ctx) {
#ifndef NDEBUG
	std::cout << ctx->print_context_top_level_count++ << " liszt: ";
#else
	std::cout << "liszt: ";
#endif
}
L_RUNTIME_UNNESTED
void lPrintEnd(lContext * ctx) {
	std::cout << "\n";
}

void printValue(lType typ, const void * value) {
	switch(typ) {
		case L_FLOAT:
			std::cout << *(float*)value;
			break;
		case L_INT:
			std::cout << *(int*)value;
			break;
		case L_BOOL:
			std::cout << *(bool*)value;
			break;
		case L_DOUBLE:
			std::cout << *(double*)value;
			break;
		case L_STRING:
			std::cout << (const char *)value;
			break;
	}
}

L_RUNTIME_UNNESTED
void lPrintValue(lContext * ctx, lType typ, size_t r, size_t c,  const void * value) {
	byte * v = (byte *) value;
	size_t size = lUtilTypeSize(typ);
	if(c != 0) { 
		std::cout << "[";
		for(size_t i = 0; i < r; i++) {
			for (size_t j = 0; j < c; j++) {
				printValue(typ,v);
				v += size;
				if(j < c - 1) std::cout << ",";
			}
			if(i < r - 1) std::cout << "|";
			else std::cout << "]";  
		}
	} else if(r != 0) {
		std::cout << "(";
		for(int i = 0; i < r - 1; i++) {
			printValue(typ,v);
			std::cout << ",";
			v += size;
		}
		printValue(typ,v);
		std::cout << ")";
	} else {
		printValue(typ,v);
	}
}

L_RUNTIME_UNNESTED
double lWallTime(lContext * ctx) { //counter, time reported in seconds
	timeval currentTime;
	gettimeofday(&currentTime, 0);
	return currentTime.tv_sec + 1e-6 * currentTime.tv_usec;
}
//Nested Runtime Calls

__device__
size_t deviceStrlen(const void * c_) {
	size_t s = 0;
	byte * c = (byte *) c_;
	while(*c++) s++;
	return s;
}

__device__
void deviceMemcpy(void * dst_, const void * src_, size_t n) {
	byte * dst = (byte*) dst_;
	byte * src = (byte*) src_;
	for(size_t i = 0; i < n; i++)
		*dst++ = *src++;
}

L_RUNTIME_NESTED
void lkPrintBegin(lkContext * ctx) {
#ifndef NDEBUG
	lkPrintValue(ctx,L_INT,0,0,&ctx->print_context_top_level_count);
	DPrintContext * pc = &ctx->print_context;
	while(1) {
		switch(pc->type) {
			case L_VERTEX: lkPrintValue(ctx,L_STRING,0,0,(void*)":v"); break;
			case L_EDGE: lkPrintValue(ctx,L_STRING,0,0,(void*)":e"); break;
			case L_FACE: lkPrintValue(ctx,L_STRING,0,0,(void*)":f"); break;
			case L_CELL: lkPrintValue(ctx,L_STRING,0,0,(void*)":c"); break;
		}

		lkPrintValue(ctx,L_INT,0,0,&pc->id);
		lkPrintValue(ctx,L_STRING,0,0,(void*)"-");
		lkPrintValue(ctx,L_INT,0,0,&pc->count);

		if(pc->child == NULL) {
			pc->count++;
			break;
		} else {
			pc = pc->child;
		}
	}
	lkPrintValue(ctx,L_STRING,0,0,(void*)" liszt: ");
#endif
}

L_RUNTIME_NESTED
void lkPrintEnd(lkContext * ctx) {
#ifndef NDEBUG
	lkPrintValue(ctx,L_STRING,0,0,(void*)"\n");
#endif
}

L_RUNTIME_NESTED
void lkPrintValue(lkContext * ctx, lType typ, size_t r, size_t c, const void * value) {
#ifndef NDEBUG
	//reserve an element
	uint32_t offset = atomicAdd(&ctx->print_buffer->sizes.n_elements,1);
	PrintElement * elem = &ctx->print_buffer->elements[offset];
	elem->type = typ;
	elem->rows = r;
	elem->cols = c;
	uint64_t id = blockIdx.x * blockDim.x + threadIdx.x;
	
	size_t data_size;


	if(typ == L_STRING) {
		const char * str = (const char *) value;
		data_size = deviceStrlen(str) + 1;
	} else {
		size_t rows = (r == 0) ? 1 : r;
		size_t cols = (c == 0) ? 1 : c;
		data_size = lUtilTypeSize(typ) * rows * cols;
	}
	elem->offset = atomicAdd(&ctx->print_buffer->sizes.n_bytes,data_size);
	deviceMemcpy(ctx->print_buffer->data + elem->offset, (const char * ) value, data_size);
	
	//must go last to ensure the host will not see a valid print count before all the data is in place
	//do we need a fence before this?
	elem->order = (id << 32) + ctx->print_count;
	
	ctx->print_count++;
#endif
}

L_RUNTIME_NESTED
void lkPrintElement(lkContext * ctx, lElementType typ, lkElement e) {
	const char * name;
	switch(typ) {
		case L_VERTEX:
			name = "Vertex ";
			break;
		case L_EDGE:
			name = "Edge ";
			break;
		case L_FACE:
			name = "Face ";
			break;
		case L_CELL:
			name = "Cell ";
			break;
	}
	
	lkPrintValue(ctx,L_STRING,0,0,(void*)name);
	int id = elementID(e);
	lkPrintValue(ctx,L_INT,0,0,&id);
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
lkFields * lkGetFields(lkContext * ctx) {
	return ctx->fields;
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
lkScalars * lkGetScalars(lkContext * ctx) {
	return ctx->scalars;
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
bool lkGetActiveElement(lkContext * ctx, lkElement * e) {
	int id = blockIdx.x * blockDim.x + threadIdx.x;
	if(id < ctx->n_elements) {
		if(ctx->colored_elements != NULL)
			e->data = ctx->colored_elements[id];
		else
			e->data = id + ctx->first_element;
#ifndef NDEBUG
		ctx->current_print_context = &ctx->print_context;
		ctx->print_context.id = elementID(*e);
		ctx->print_context.count = 0;
#endif
		return true;
	} else 
		return false;
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkFieldRead(lkField * f, lkElement e, lType element_type, size_t element_length, size_t val_offset, size_t val_length,  void * result) {
	size_t offset,size;
	lUtilValueLocation(elementID(e),element_type,element_length,val_offset,val_length,&offset,&size);
	//after constant propagation this should turn into assignments
	lUtilValueReduce(result, L_ASSIGN, element_type, val_length,f->data + offset);
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkFieldWrite(lkField * f, lkElement e, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value) {
	size_t offset,size;
	lUtilValueLocation(elementID(e),element_type,element_length,val_offset,val_length,&offset,&size);
	lUtilValueReduce(f->data + offset, reduction, element_type, val_length, value);
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkScalarRead(lkContext * ctx, lkScalar * scalar, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * result) {
	lUtilValueReduce(result, L_ASSIGN, element_type, val_length, scalar->data + val_offset * lUtilTypeSize(element_type));
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkScalarWrite(lkContext * ctx, lkScalar * scalar, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value_) {
	int id = blockDim.x * blockIdx.x + threadIdx.x;
	size_t elem_size = lUtilTypeSize(element_type);
	byte * value =  (byte *) value_;
	for(int i = 0; i < val_length; i++) {
		size_t offset_elem = (ctx->n_elements_max * (val_offset + i) + id);
		//printf("(%d) reducing into element %d\n",(int)id,(int)offset_elem);
		size_t offset = elem_size * offset_elem;
		lUtilValueReduce(scalar->write_table + offset, reduction, element_type, 1, value + elem_size * i);
	}
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
size_t lkSetSize(lkContext * ctx, lkSet * set) {
	return set->set.size;
}
L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkSetGetIterator(lkContext * ctx, lkSet * set, lElementType typ, lkIterator * iterator) {

	lNestedSetGetIterator(&set->set,&iterator->it);
	
#ifndef NDEBUG

	iterator->print_context.type = typ;
	iterator->print_context.count = 0;
	iterator->print_context.parent = ctx->current_print_context;
	ctx->current_print_context->child = &iterator->print_context;
	iterator->print_context.child = NULL;
	ctx->current_print_context = ctx->current_print_context->child;

#endif
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
bool lkIteratorNext(lkContext * ctx, lkIterator * iterator, lElementType typ, lkElement * e, int * lbl) {
	bool result = lNestedIteratorNext(&iterator->it, e, lbl);
#ifndef NDEBUG
	if(result) {
		iterator->print_context.id = elementID(*e);
		iterator->print_context.count = 0;
	} else {
		ctx->current_print_context = ctx->current_print_context->parent;
		ctx->current_print_context->child = NULL;
		ctx->current_print_context->count++;
	}
#endif
	return result;
}

//now all the topological functions ....
L_RUNTIME_UNNESTED
void lVerticesOfMesh(lContext * ctx, lSet * set) {
	set->id = -L_VERTEX - 1;
	set->size = ctx->mesh.nvertices;
	set->ranges.push_back(std::make_pair(0, set->size));
}
L_RUNTIME_UNNESTED
void lEdgesOfMesh(lContext * ctx, lSet * set) {
	set->id = -L_EDGE - 1;
	set->size = ctx->mesh.nedges;
	set->ranges.push_back(std::make_pair(0, set->size));
}
L_RUNTIME_UNNESTED
void lFacesOfMesh(lContext * ctx, lSet * set) {
	set->id = -L_FACE - 1;
	set->size = ctx->mesh.nfaces;
	set->ranges.push_back(std::make_pair(0, set->size));
}
L_RUNTIME_UNNESTED
void lCellsOfMesh(lContext * ctx, lSet * set) {
	set->id = -L_CELL - 1;
	set->size = ctx->mesh.ncells - 1;
	set->ranges.push_back(std::make_pair(1, ctx->mesh.ncells));
}

L_RUNTIME_NESTED
int lkIDOfVertex(lkContext * ctx, lkElement e) {
	return elementID(e);
}
L_RUNTIME_NESTED
int lkIDOfEdge(lkContext * ctx, lkElement e) {
	return elementID(e);
}
L_RUNTIME_NESTED
int lkIDOfFace(lkContext * ctx, lkElement e) {
	return elementID(e);
}
L_RUNTIME_NESTED
int lkIDOfCell(lkContext * ctx, lkElement e) {
	return elementID(e);
}

//this template defines all the 
#define NESTED_TOPOLOGY_GET_MESH(ctx) (ctx->mesh)
#define NESTED_TOPOLOGY_GET_SET(set) (set->set)
#include "nested_topology_functions.inc"


struct lsElement { lColoringElement e; };
struct lsSet { lColoringSet s; };
struct lsIterator { lColoringIterator it; };

#include "vector.h"
#include "matrix.h"
#include "liszt_math.h"

//code generated from the liszt compiler is now included here after runtime-specific
//headers have been included

#include "generated.cpp"

