
#include <sys/time.h>
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <assert.h>
#include <cuda.h>
#include <map>

#include <thrust/reduce.h>
#include <thrust/fill.h>
#include <thrust/functional.h>
#include <thrust/iterator/counting_iterator.h>


#include "liszt_runtime.h"
#include "boundary_set_reader.h"
#include "mesh.h"
#include "print_context.h"
#include "runtime_util.h"
#include "Array.h"
#include "ParallelLoad/ParallelLoad.h"
#include "nested_topology.h"
#include "ghost_discovery.h"
#include "vector.h"
#include "coloring.h"

#define L_KERNEL __global__
#define L_NESTED L_ALWAYS_INLINE __device__
#define L_UNNESTED __host__
#define L_STENCIL

#define KERNEL_BLOCKSIZE (256)

#define PRINT_BUFFER_ELEMENT_SIZE (200000 * 128)
#define PRINT_BUFFER_DATA_SIZE (1024 * 1024 * 32)

//this really should be passed into lExec ...
#define SCALAR_MAX_SIZE (sizeof(double) * 22)
#define SCALAR_BYTE_ALIGNMENT (8)

//define opaque data types and implementation for specific runtime

#define CUDA_DO(x) do { \
	cudaError_t err = (x); \
	if(cudaSuccess != err) { \
		fprintf(stderr, "%s:%d: %s\n",__FILE__,__LINE__, cudaGetErrorString(err)); \
		exit(1);\
	} \
} while(0)

struct DPrintContext {
	lElementType type;
	int id;
	int count;	
	DPrintContext * child;
	DPrintContext * parent;
};
struct PrintBuffer;

struct lkContext {
	lkFields * fields;
	lkScalars * scalars;
	Mesh * mesh;
	size_t n_elements_max;
	
	size_t n_elements;
	
	size_t first_element;
	uint32_t * colored_elements; //if this is null, then use first_element + idx as the mesh element, otherwise using colored_element[id] as the index

	uint32_t * local_to_global[L_ELEMENT_TYPE_SIZE];
	
#ifndef NDEBUG
	size_t print_count;
	PrintBuffer * print_buffer;
	DPrintContext print_context;
	DPrintContext * current_print_context;
	int print_context_top_level_count;
#endif
};


struct lContext {
	lFields * fields;
	lSets * sets;
	lScalars * scalars;
	shared_ptr<MeshIO::LisztFileReader> mesh_reader; //we hang on to this so we can load the positions on demand
	shared_ptr<BoundarySetReader> boundary_reader;
	
	Mesh mesh;

	shared_ptr<NameService> name_service;
	Array<uint32_t> local_to_global[L_ELEMENT_TYPE_SIZE];
	size_t num_global_elements[L_ELEMENT_TYPE_SIZE];
	size_t first_ghost[L_ELEMENT_TYPE_SIZE];
	size_t num_ghosts[L_ELEMENT_TYPE_SIZE];
	Array<FieldCommunicationPattern> field_communication_patterns; //map from field_id -> communication pattern
	
	size_t n_elements_max;
	
#ifndef NDEBUG
	size_t print_context_top_level_count;
#endif
	
	bool scalar_dirty;
	lkContext nested_template;
	
	size_t n_scalars;
	size_t scalar_data_bump_pointer;
	byte * scalar_data;
	byte * device_scalar_data;
	
	
	//key is (set_id,kernel_id)
	std::map< std::pair<uint32_t,uint32_t>, lColoringResult> coloring_cache;
	cudaEvent_t start_event,end_event;
};


struct lkField {
	byte * data;
	size_t first_ghost;
	size_t write_ghosts_offset;
	size_t n_write_ghosts;
	size_t n_read_ghosts;
};

struct FieldMessage {
	int N; //number of ranks to which this message will be sent
	int * ranks; //list of ranks 0..N
	size_t * offsets; //offsets into device idxs that partition the data into ranks
	uint32_t * device_idxs; //ids of the entries on the device that should be read/written.
	byte * message_data; //host pointer to the page accessible on the device/host for this message
	byte * device_message_data; //device pointer for the same data above
	MPI_Request * status;
};

struct GPUFieldCommunicator {
	FieldMessage writers_to_owners_send;
	FieldMessage writers_to_owners_recv;
	
	FieldMessage owners_to_readers_send;
	FieldMessage owners_to_readers_recv;
};

struct lField {
	GPUFieldCommunicator communicator;
	lPhase phase;
	lkField field_template; //same content as below but on the host
	lkField * device_field; //pointer to field on device
};



struct lkScalar {
	byte * data;
	byte * write_table;
};
struct lScalar {
	byte * data;
	byte * write_entry;
	lPhase phase;
	
	size_t max_dirty_elements;
	
	lType type;
	size_t length;
	
	lkScalar scalar_template;
	lkScalar * device_scalar;
	lContext * ctx;
};


struct lSet {
	BoundarySetReader::range_list_t ranges;
	int id; //unique identifier for this set (complete mesh sets have a negative number (e.g. -L_VERTEX), boundaries are numbered as they found in the file)
	size_t size; //total size, not partitioned size
    size_t local_size; //size of the set on the local partition
};

struct lkSet {
	lNestedSet set;
};
struct lkIterator {
	lNestedIterator it;
#ifndef NDEBUG
	DPrintContext print_context;
#endif
};

struct PrintElement {
	lType type;
	uint64_t order;
	char rows;
	char cols;
	uint32_t offset; //offset into PrintBuffer data
};

struct PrintSizes {
	uint32_t n_bytes;
	uint32_t n_elements;
};

struct PrintBuffer {
	PrintSizes sizes;
	PrintElement elements[PRINT_BUFFER_ELEMENT_SIZE];
	byte data[PRINT_BUFFER_DATA_SIZE];
};


void FieldMessage_allocateMessageData(FieldMessage * msg, lType val_type, size_t val_length) {
	if(msg->N > 0) {
		size_t size = lUtilTypeSize(val_type) * val_length;
		CUDA_DO(cudaHostAlloc(&msg->message_data, size * msg->offsets[msg->N],cudaHostAllocMapped));
		//CUDA_DO(cudaHostGetDevicePointer(&msg->device_message_data, msg->message_data,0));
		CUDA_DO(cudaMalloc(&msg->device_message_data, size * msg->offsets[msg->N]));
	}
}

void FieldMessage_initFromCommInfo(FieldMessage * msg, CommInfo * info,lType val_type, size_t val_length) {
	msg->N = info->messages.size();
	msg->ranks = (int*) malloc(sizeof(int) * msg->N);
	msg->offsets = (size_t*) malloc(sizeof(size_t) * (msg->N + 1));
	//for(int i = 0; i < info->data.size(); i++) {
	//	LOG_LOG("index[%d] = %d",i,info->data[i]);
	//}
	CUDA_DO(cudaMalloc(&msg->device_idxs, sizeof(uint32_t) * info->data.size()));
	CUDA_DO(cudaMemcpy(msg->device_idxs,info->data.begin(), sizeof(uint32_t) * info->data.size(),cudaMemcpyHostToDevice));
	msg->offsets[msg->N] = 0; //in case there are 0 messages
	for(size_t i = 0; i < msg->N; i++) {
		msg->ranks[i] = info->messages[i].rank;
		msg->offsets[i] = info->messages[i].buf_offset;
		msg->offsets[i+1] = info->messages[i].buf_offset + info->messages[i].size;
	}
	
	msg->status = (MPI_Request*) malloc(sizeof(MPI_Request) * msg->N);
	for(size_t i = 0; i < msg->N; i++) {
		msg->status[i] = MPI_REQUEST_NULL;
	}
	
	FieldMessage_allocateMessageData(msg,val_type,val_length);
}



L_ALWAYS_INLINE 
size_t numberOfOwnedElementsOfType(lContext * ctx, lElementType typ) {
	return ctx->first_ghost[(int)typ];
}
L_ALWAYS_INLINE 
size_t numberOfGlobalElementsOfType(lContext * ctx, lElementType typ) {
	return ctx->num_global_elements[(int)typ];
}

L_ALWAYS_INLINE 
size_t numberOfGhostElementsOfType(lContext * ctx, lElementType typ) {
	return ctx->num_ghosts[(int)typ];
}

L_ALWAYS_INLINE 
size_t numberOfLocalElementsOfType(lContext * ctx, lElementType typ) {
	return numberOfOwnedElementsOfType(ctx,typ) 
	     + numberOfGhostElementsOfType(ctx,typ);
}

void copyCRS(size_t n_elems, CRS * dst, CRS * src) {
	size_t n_values = src->row_idx[n_elems];

	//size of row_idx is (n_elems + 1) since it holds the total size in row_idx[n_elems]
	CUDA_DO(cudaMalloc(&dst->row_idx, (n_elems + 1) * sizeof(uint32_t)));
	CUDA_DO(cudaMalloc(&dst->values,n_values * sizeof(uint32_t)));
	CUDA_DO(cudaMemcpy(dst->row_idx,src->row_idx, (n_elems + 1) * sizeof(uint32_t),cudaMemcpyHostToDevice));
	CUDA_DO(cudaMemcpy(dst->values,src->values, n_values * sizeof(uint32_t),cudaMemcpyHostToDevice));
	
}
void copyCRSConst(size_t n_elems, CRSConst * dst, CRSConst * src) {
	CUDA_DO(cudaMalloc(&dst->values,2 * n_elems * sizeof(uint32_t)));
	CUDA_DO(cudaMemcpy(dst->values,src->values, 2 * n_elems * sizeof(uint32_t),cudaMemcpyHostToDevice));
}
//dst is a host pointer to a mesh structure that will be copied to the device
//all pointers inside 'dst' should be device pointers
void initDeviceMesh(Mesh * dst, Mesh * src) {
	dst->nvertices = src->nvertices;
	dst->nedges = src->nedges;
	dst->nfaces = src->nfaces;
	dst->ncells = src->ncells;
	
	copyCRS(src->nvertices, &dst->vtov, &src->vtov);
	copyCRS(src->nvertices, &dst->vtoe, &src->vtoe);
	copyCRS(src->nvertices, &dst->vtof, &src->vtof);
	copyCRS(src->nvertices, &dst->vtoc, &src->vtoc);
	
	copyCRSConst(src->nedges, &dst->etov, &src->etov);
	copyCRS(src->nedges, &dst->etof, &src->etof);
	copyCRS(src->nedges, &dst->etoc, &src->etoc);

	copyCRSConst(src->nfaces, &dst->ftoc, &src->ftoc);
	copyCRS(src->nfaces, &dst->ftoe, &src->ftoe);
	copyCRS(src->nfaces, &dst->ftov, &src->ftov);

	copyCRS(src->ncells, &dst->ctov, &src->ctov);
	copyCRS(src->ncells, &dst->ctoe, &src->ctoe);
	copyCRS(src->ncells, &dst->ctof, &src->ctof);
	copyCRS(src->ncells, &dst->ctoc, &src->ctoc);
}

void kernelSize(size_t n_elems, size_t * n_blocks, size_t * block_size) {
	*n_blocks = (n_elems + KERNEL_BLOCKSIZE - 1) / KERNEL_BLOCKSIZE;
	*block_size = KERNEL_BLOCKSIZE;
}
//Unnested Runtime Calls

void printBufferClearSizes(PrintBuffer * buffer) {
	PrintSizes * sizes = &buffer->sizes;
	PrintSizes local = { 0, 0 };
	CUDA_DO(cudaMemcpy(sizes,&local,sizeof(PrintSizes), cudaMemcpyHostToDevice));
}

struct lsContext {  lColoringContext c; lDiscoveryContext d; };
//Unnested Runtime Calls           

void initializeMPIContext(lContext * ctx, void (*entry_stencil)(lsFunctionTable*,lsContext*), lProgramArguments * arguments, size_t n_fields, size_t n_sets, size_t n_scalars) {
	
	int zero = 0;
	MPI_Init(&zero,NULL);
	
	if(arguments->redirect_output_to_log) {
		int mpi_rank;
		std::ostringstream ss;
		MPI_Comm_rank(MPI_COMM_WORLD,&mpi_rank);
		ss << "liszt_runtime_" << mpi_rank << ".log";
		int file = fileno(fopen( /*(mpi_rank == 0) ?*/ ss.str().c_str() /*: "/dev/null"*/,"a"));
		dup2(file,1);
		dup2(file,2);
	}
	
	//allocate memory to hold global state
	ctx->fields = (lFields*) new lField[n_fields];
	ctx->sets = (lSets*) new lSet[n_sets];
	ctx->scalars = (lScalars*) new lScalar[n_scalars];
	
	
	//set up the file reader which reads a mesh files, note that this will only read objects from a file when requested
	ctx->mesh_reader.reset(new MeshIO::LisztFileReader()); 
	ctx->mesh_reader->init(arguments->mesh_file);
	//every node reads the entire header and boundary set list from liszt mesh file
	const MeshIO::LisztHeader & h = ctx->mesh_reader->header();
	ctx->num_global_elements[L_VERTEX] = h.nV;
	ctx->num_global_elements[L_EDGE] = h.nE;
	ctx->num_global_elements[L_FACE] = h.nF;
	ctx->num_global_elements[L_CELL] = h.nC;
	
	ctx->boundary_reader.reset(new BoundarySetReader);
	ctx->boundary_reader->init(h.nBoundaries,ctx->mesh_reader->boundaries());
	
	

	const LisztHeader * header = &ctx->mesh_reader->header();
	ctx->boundary_reader->init(header->nBoundaries, ctx->mesh_reader->boundaries());
	

	//see individual objects for what each does during initialization in detail
	//name_service answers queries about who has mesh information
	//e.g. who is responsible for knowing the owner of cell x?
	ctx->name_service.reset(new NameService(header));

	//the topology service answers queries about topology
	//e.g. get me all the facet edges for cell x? individual requests are aggregated for performance
	shared_ptr<TopologyService> topology_service(new TopologyService(ctx->name_service, ctx->mesh_reader));

	//owner service determines ownership through partition and then answers questions like
	//what node owns cell x? or what are all the nodes that my local node owns?
	shared_ptr<OwnerService> owner_service( new OwnerService(ctx->name_service,topology_service->partitionedTopology()));

	//local topology stores the topology necessary to run the program and is used to construct the mesh that will be used
	//at runtime
	shared_ptr<LocalTopology> local_topology(new LocalTopology(ctx->name_service->self(),owner_service->locallyOwnedElements()));

	//local memory stores the memory access patterns to field, used to determine field communication.
	shared_ptr<LocalMemory> local_memory(new LocalMemory(n_fields));

	//The stencil is the abstract representation of the program for the purpose of topology discovery and field memory discovery
	//it will populate local_topology and local_memory based on the initial partition
	lsContext lsCtx;
	lDiscoveryRun(entry_stencil,&lsCtx,ctx->name_service,topology_service,owner_service,ctx->boundary_reader,local_topology,local_memory);
	
	//convert the facet edge mesh of local topology into a fast Mesh object.
	local_topology->buildLocalMesh(&ctx->mesh);
	
	//initialize meta-data needed to later initialize fields
	//local_topology will no longer exist when fields are constructed, so we save the relevent information here
	for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++) {
		ctx->first_ghost[i] = local_topology->firstGhost((lElementType)i);
		ctx->num_ghosts[i] = local_topology->numberGhost((lElementType)i);
	}
	//we need to maintain a local ID -> global ID mapping for correct implementation of liszt's ID function
	for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++) {
		local_topology->buildLocalToGlobal((lElementType)i,&ctx->local_to_global[i]);
	}
	
	//construct a map from field id to the information about field reads/writes it performs
	//when a field is initialized it will use these values to set up its comm structures
	
	//Fill in field communication patterns
	ctx->field_communication_patterns.reset(n_fields);
	for(size_t i = 0; i < n_fields; i++) {
		ctx->field_communication_patterns[i].init(local_topology,&local_memory->fields[i]);
	}

	local_topology->print_debug(ctx->name_service);

	//note that most of the parallel load infrastructure goes out of scope when init exits, freeing all memory but the things
	//that are necessary to perform the run
}


void initializeGPUContext(lContext * ctx, void (*entry_stencil)(lsFunctionTable*,lsContext*), lProgramArguments * arguments, size_t n_fields, size_t n_sets, size_t n_scalars) {
    int rank,device_count;
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    CUDA_DO(cudaGetDeviceCount(&device_count));
    int device_to_use = rank % device_count;
    LOG_LOG("using gpu %d",device_to_use);
    CUDA_DO(cudaSetDevice(device_to_use));

	ctx->scalar_dirty = false;
	ctx->n_elements_max = 0;
	for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++) {
		ctx->n_elements_max = std::max(ctx->n_elements_max,numberOfOwnedElementsOfType(ctx,(lElementType)i));
		CUDA_DO(cudaMalloc(&ctx->nested_template.local_to_global[i], sizeof(uint32_t) * ctx->local_to_global[i].size()));
		CUDA_DO(cudaMemcpy(ctx->nested_template.local_to_global[i],ctx->local_to_global[i].begin(), sizeof(uint32_t) * ctx->local_to_global[i].size(), cudaMemcpyHostToDevice));
	}
	ctx->nested_template.n_elements_max = ctx->n_elements_max;
	CUDA_DO(cudaMalloc(&ctx->nested_template.fields, sizeof(lkField) * n_fields));
	CUDA_DO(cudaMalloc(&ctx->nested_template.scalars, sizeof(lkScalar) * n_scalars));
	Mesh device_mesh;
	initDeviceMesh(&device_mesh, &ctx->mesh); 
	
	CUDA_DO(cudaMalloc(&ctx->nested_template.mesh, sizeof(Mesh)));
	CUDA_DO(cudaMemcpy(ctx->nested_template.mesh,&device_mesh, sizeof(Mesh), cudaMemcpyHostToDevice));
	
#ifndef NDEBUG
	ctx->print_context_top_level_count = 0;
	CUDA_DO(cudaMalloc(&ctx->nested_template.print_buffer, sizeof(PrintBuffer)));
	printBufferClearSizes(ctx->nested_template.print_buffer);
	ctx->nested_template.print_count = 0;
#endif

	ctx->n_scalars = n_scalars;
	ctx->scalar_data_bump_pointer = 0;
	ctx->scalar_data = (byte*) malloc(SCALAR_MAX_SIZE * n_scalars);
	CUDA_DO(cudaMalloc(&ctx->device_scalar_data,SCALAR_MAX_SIZE * n_scalars));
	
}

void lExec(void (*entry_point)(lContext*),void (*entry_stencil)(lsFunctionTable*,lsContext*), lProgramArguments * arguments, size_t n_fields, size_t n_sets, size_t n_scalars) {
	
	lContext ctx;
	
	initializeMPIContext(&ctx,entry_stencil,arguments,n_fields,n_sets,n_scalars);
	initializeGPUContext(&ctx,entry_stencil,arguments,n_fields,n_sets,n_scalars);
	
	//launch the program
	
	entry_point(&ctx);
	
    MPI_Finalize();
	//do cleanup
}


L_ALWAYS_INLINE
lFields * lGetFields(lContext * ctx) {
	return ctx->fields;
}
L_ALWAYS_INLINE
lSets * lGetSets(lContext * ctx) {
	return ctx->sets;
}
L_ALWAYS_INLINE
lScalars * lGetScalars(lContext * ctx) {
	return ctx->scalars;
}
L_RUNTIME_UNNESTED
void lFieldInit(lContext * ctx, lField * field, int id, lElementType key_type, lType val_type, size_t val_length) {

	FieldCommunicationPattern & p = ctx->field_communication_patterns[id];
	
	FieldMessage_initFromCommInfo(&field->communicator.writers_to_owners_send, &p.writer->send,val_type,val_length);
	FieldMessage_initFromCommInfo(&field->communicator.writers_to_owners_recv, &p.writer->recv,val_type,val_length);
	FieldMessage_initFromCommInfo(&field->communicator.owners_to_readers_send, &p.reader->send,val_type,val_length);
	FieldMessage_initFromCommInfo(&field->communicator.owners_to_readers_recv, &p.reader->recv,val_type,val_length);
	
	size_t n_owned = numberOfOwnedElementsOfType(ctx, key_type);
	size_t n_ghosts = numberOfGhostElementsOfType(ctx,key_type);
	field->field_template.n_read_ghosts = (p.reads_ghosts) ? n_ghosts : 0;
	field->field_template.n_write_ghosts = (p.writes_ghosts) ? n_ghosts : 0;
	
	field->field_template.first_ghost = n_owned;
	field->field_template.write_ghosts_offset = field->field_template.n_read_ghosts;
	
	size_t n_elems = n_owned + field->field_template.n_read_ghosts + field->field_template.n_write_ghosts;
	
	
	field->phase = L_READ_ONLY;
	
	//initialize GPU-size state
	
	size_t field_idx = field - (lField *) ctx->fields;
	field->device_field = (lkField*)ctx->nested_template.fields + field_idx;
	CUDA_DO(cudaMalloc(&field->field_template.data,n_elems * val_length * lUtilTypeSize(val_type)));
	CUDA_DO(cudaMemcpy(field->device_field,&field->field_template, sizeof(lkField), cudaMemcpyHostToDevice));
	
}

size_t fieldNumberOfReadElements(lContext * ctx, lField * field, lElementType key_type) {
	size_t n_owned = numberOfOwnedElementsOfType(ctx, key_type);
	return n_owned + field->field_template.n_read_ghosts;
}

__global__ void broadcast_kernel(byte * device_data, size_t n_field_elements,lType val_type, size_t val_length, byte * data) {
	int elem_id = blockIdx.x * blockDim.x + threadIdx.x;
	if(elem_id < n_field_elements) {
		size_t offset,size;
		lUtilValueLocation(elem_id,val_type,val_length,0,val_length,&offset,&size);
		lUtilValueReduce(device_data + offset, L_ASSIGN, val_type, val_length, data);
	}
}

void doBroadcast(byte * device_data, size_t n_elems,lType val_type, size_t val_length, byte * data) {
	size_t size = lUtilTypeSize(val_type) * val_length;
	
	byte * device_value;
	CUDA_DO(cudaMalloc(&device_value,size));
	CUDA_DO(cudaMemcpy(device_value,data,size,cudaMemcpyHostToDevice));
	
	size_t n_blocks,n_threads;
	kernelSize(n_elems,&n_blocks,&n_threads);
	broadcast_kernel<<<n_blocks,n_threads>>>(device_data,n_elems,val_type,val_length,device_value);
	CUDA_DO(cudaGetLastError());
	cudaThreadSynchronize(); //does cudaFree go onto the normal push buffer? or do we need to syncrhonize here to avoid freeing before the kernel completes?
	CUDA_DO(cudaFree(device_value));
}

L_ALWAYS_INLINE
void lFieldBroadcast(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, void * data) {
	size_t n_elems = fieldNumberOfReadElements(ctx,field,key_type);
	doBroadcast(field->field_template.data,n_elems,val_type,val_length,(byte*)data);
}

L_RUNTIME_UNNESTED void lFieldLoadData(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, const char * key) {
	if (!strcmp(key,"position")) {
        assert(key_type == L_VERTEX);
        assert(val_type == L_DOUBLE || val_type == L_FLOAT);
    }

    size_t n_read_elems = fieldNumberOfReadElements(ctx,field,key_type);

    // load field data from our mesh file in parallel
    uint32_t begin, end;

    // query name service to see which range this node is responsible for
    ctx->name_service->localPartition(key_type, &begin, &end);
    // read field data range from file
    const byte *data = ctx->mesh_reader->fieldData(key_type, val_type, val_length, key, begin, end);

    SizedAllToAll<uint32_t> request;
    request.beginCount();

    // Field data now needs to be requested from node responsible for partition and sent 
    // to whoever potentially reads them

    // for everything in this field's read entries, request field data
    for (size_t i = 0; i < n_read_elems; i++) {
        uint32_t global_id = ctx->local_to_global[key_type][i];
        request.count(ctx->name_service->partitionFor(key_type,global_id));
    }

    request.beginData();
    for (size_t  i = 0; i < n_read_elems; i++) {
        uint32_t global_id = ctx->local_to_global[key_type][i];
        request.data(ctx->name_service->partitionFor(key_type,global_id),global_id);
    }
        
    size_t n_part;
    const int *offsets;
    const uint32_t *rq_data;
    size_t total_data = request.endData(&n_part, &offsets, &rq_data);
    
    size_t elem_size = val_length * lUtilTypeSize(val_type);
    SizedAllToAll<uint8_t> response;

    response.beginCount();
    for (size_t i = 0; i < n_part; i++) {
        size_t begin = offsets[i];
        size_t end   = (i + 1 == n_part) ? total_data : offsets[i + 1];
        response.count(i, (end - begin) * elem_size);
    }

    // offsets - array listing the offset in rq_data where each node's request begins/ends
    // rq_data - array of global ids of requested elements
    // begin   - id of global element that is the 0th element of this node's partition

    // respond to the initial request with the data read from the local range we are responsible for
    response.beginData();
    for (size_t i = 0; i < n_part; i++) {
        size_t pend   = (i + 1 == n_part) ? total_data : offsets[i + 1];
        size_t pbegin = offsets[i];
        for (size_t j = pbegin; j < pend; j++) {
            size_t e = rq_data[j] - begin;
            
            uint8_t elem_data[elem_size];
            memcpy(elem_data, data + e * elem_size, elem_size);
            
            for (int k = 0; k < elem_size; k++)
                response.data(i, elem_data[k]);
        }
    }
    std::free((void *) data);


    // Reading request response as a client
    const uint8_t *rdata;
    total_data = response.endData(&n_part,&offsets,&rdata);
  
    Array<int> roffsets(n_part);
    std::copy(offsets,offsets+n_part,roffsets.begin());


	byte * field_data = (byte*) malloc(n_read_elems * elem_size);
    // fill the field with the response data
    //   (Remember, elem_size elems are read for each field entry)
    for (size_t i = 0; i < n_read_elems; i++) {
        uint32_t global_id = ctx->local_to_global[key_type][i];
        PartitionID part   = ctx->name_service->partitionFor(key_type, global_id);

        memcpy(field_data + i * elem_size, rdata + roffsets[part], elem_size);
        roffsets[part] += elem_size;
    }
    
    //now copy the result to the GPU
    CUDA_DO(cudaMemcpy(field->field_template.data, field_data, n_read_elems * elem_size, cudaMemcpyHostToDevice));
    free(field_data); //we're done with the host-side data
}

L_RUNTIME_UNNESTED void lFieldSaveData(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, const char * key) {
	lFieldEnterPhase(field,val_type, val_length, L_READ_ONLY);
	printf("TODO: save field %s\n",key);
}



__device__
size_t deviceStrlen(const void * c_) {
	size_t s = 0;
	byte * c = (byte *) c_;
	while(*c++) s++;
	return s;
}

__device__
void deviceMemcpy(void * dst_, const void * src_, size_t n) {
	byte * dst = (byte*) dst_;
	byte * src = (byte*) src_;
	for(size_t i = 0; i < n; i++)
		*dst++ = *src++;
}



__global__ void doGather(size_t n_indices, uint32_t * indexes, size_t elem_size, lkField * field, byte * output) {
	int i = blockIdx.x * blockDim.x + threadIdx.x;
	if(i < n_indices) {
		size_t id = indexes[i];
		uint32_t adjusted_id = (id >= field->first_ghost) ? (id + field->write_ghosts_offset) : id; 
		deviceMemcpy(output + elem_size * i,field->data + elem_size * adjusted_id, elem_size);
	}
}

void FieldMessage_gather(FieldMessage * msg, lField * field, lType val_type, size_t val_length) {
	if(msg->N == 0)
		return;
		
	CUDA_DO(cudaThreadSynchronize());
	double begin = lWallTime(NULL);
	size_t n_elems = msg->offsets[msg->N];
	
	size_t n_blocks,n_elements;
	kernelSize(n_elems, &n_blocks, &n_elements);
	doGather<<<n_blocks,n_elements>>>(n_elems, msg->device_idxs, lUtilTypeSize(val_type) * val_length,  field->device_field, msg->device_message_data);
   	size_t elem_size = lUtilTypeSize(val_type) * val_length;
	CUDA_DO(cudaMemcpy(msg->message_data,msg->device_message_data,elem_size * msg->offsets[msg->N],cudaMemcpyDeviceToHost)); 
   double end = lWallTime(NULL);
    LOG_LOG("gathered %d bytes in %f seconds",(int)msg->offsets[msg->N],(end - begin)*1000.0);
}

__global__ void doScatter(size_t n_indices, uint32_t * indexes, lType val_type, size_t val_length, lReduction reduction, lkField * field, byte * input) {
	int i = blockIdx.x * blockDim.x + threadIdx.x;
	if(i < n_indices) {
		size_t id = indexes[i]; 
		size_t offset,size;
		lUtilValueLocation(id,val_type,val_length,0,val_length,&offset,&size);
		lUtilValueReduce(field->data + offset, reduction, val_type, val_length,input + i * size);
	}
}

void FieldMessage_scatter(FieldMessage * msg, lField * field, lType val_type, size_t val_length, lReduction reduction) {
	if(msg->N == 0)
		return;
	CUDA_DO(cudaThreadSynchronize());
	double begin = lWallTime(NULL);
	size_t elem_size = lUtilTypeSize(val_type) * val_length;
	CUDA_DO(cudaMemcpy(msg->device_message_data,msg->message_data,elem_size * msg->offsets[msg->N],cudaMemcpyHostToDevice));	
	if (reduction != L_ASSIGN) {
		//we might have multiple reductions to the same field coming from different MPI nodes
		//so we handle the scatter for each incoming node separately
		for(size_t i = 0; i < msg->N; i++) {
			byte * buf = msg->device_message_data + elem_size * msg->offsets[i];
			uint32_t * idx = msg->device_idxs + msg->offsets[i];
			size_t n_elems = msg->offsets[i+1] - msg->offsets[i];
			size_t n_blocks,n_elements;
			kernelSize(n_elems, &n_blocks, &n_elements);
			doScatter<<<n_blocks,n_elements>>>(n_elems, idx, val_type, val_length,  reduction, field->device_field, buf);
		}
	} else {
		//assignments can happen all at once since assignments will occur only once and only to ghost elements
		size_t n_elems = msg->offsets[msg->N];
		size_t n_blocks,n_elements;
		kernelSize(n_elems, &n_blocks, &n_elements);
		//for(int i = 0; i < 10; i++) {
		//	LOG_LOG("data is %d",((int*)msg->message_data)[i]);
		//}
		doScatter<<<n_blocks,n_elements>>>(n_elems, msg->device_idxs, val_type, val_length,  reduction, field->device_field, msg->device_message_data);
	}
	
	CUDA_DO(cudaThreadSynchronize());
	double end = lWallTime(NULL);
	LOG_LOG("scattered %d bytes in %f seconds",(int)msg->offsets[msg->N],(end - begin)*1000.0);
}

//post MPI message saying that you expect to receive this message
void FieldMessage_recvbegin(FieldMessage * msg, lType val_type, size_t val_length) {
	size_t elem_size = lUtilTypeSize(val_type) * val_length;
	for(int i = 0; i < msg->N; i++) {
		int rank = msg->ranks[i];
		byte * buf = msg->message_data + elem_size * msg->offsets[i];
		size_t n_elements = msg->offsets[i+1] - msg->offsets[i];
		size_t n_bytes = lUtilTypeSize(val_type) * val_length * n_elements;
		MPI_Irecv(buf,n_bytes,MPI_BYTE,rank,0,MPI_COMM_WORLD,&msg->status[i]);
	}
}
void FieldMessage_recvend(FieldMessage * msg, lType val_type, size_t val_length) {
	if(msg->N > 0)
		MPI_Waitall(msg->N,msg->status,MPI_STATUSES_IGNORE);
}


//issues sends, must come after issuing receives
void FieldMessage_send(FieldMessage * msg, lType val_type, size_t val_length) {
	//wait for potential previous sends to complete
	if(msg->N > 0 && msg->status[0] != MPI_REQUEST_NULL) {
		MPI_Waitall(msg->N, msg->status, MPI_STATUSES_IGNORE);
	}
	
	size_t elem_size = lUtilTypeSize(val_type) * val_length;
	
	for(int i = 0; i < msg->N; i++) {
		int rank = msg->ranks[i];
		byte * buf = msg->message_data + elem_size * msg->offsets[i];
		size_t n_elements = msg->offsets[i+1] - msg->offsets[i];
		size_t n_bytes = lUtilTypeSize(val_type) * val_length * n_elements;
		MPI_Issend(buf,n_bytes,MPI_BYTE,rank,0,MPI_COMM_WORLD,&msg->status[i]);
	}
}

void doCommunication(lField * field, lType val_type, size_t val_length, lPhase from_phase) {
	lReduction r = lUtilPhaseToReduction(from_phase);
	
	//LOG_LOG("wto recv %d send %d",field->communicator.writers_to_owners_recv.N,field->communicator.writers_to_owners_send.N);
	//LOG_LOG("otr recv %d send %d",field->communicator.owners_to_readers_recv.N,field->communicator.owners_to_readers_send.N);
	//writers->owners
	if(from_phase != L_WRITE_ONLY) { //direct writes on only allowed to local elements, so they are not sent here
        FieldMessage_gather(&field->communicator.writers_to_owners_send, field, val_type, val_length);
        
        FieldMessage_recvbegin(&field->communicator.writers_to_owners_recv, val_type, val_length);
        FieldMessage_send(&field->communicator.writers_to_owners_send, val_type, val_length);
        FieldMessage_recvend(&field->communicator.writers_to_owners_recv, val_type, val_length);
        
        FieldMessage_scatter(&field->communicator.writers_to_owners_recv, field,val_type, val_length, r);
    }        
	//owners->readers
	FieldMessage_gather(&field->communicator.owners_to_readers_send, field, val_type, val_length);
	FieldMessage_recvbegin(&field->communicator.owners_to_readers_recv, val_type, val_length);
	FieldMessage_send(&field->communicator.owners_to_readers_send, val_type, val_length);
	FieldMessage_recvend(&field->communicator.owners_to_readers_recv, val_type, val_length);
	
	FieldMessage_scatter(&field->communicator.owners_to_readers_recv, field, val_type, val_length, L_ASSIGN);
	
}


void lFieldEnterPhase(lField * field, lType val_type, size_t val_length, lPhase phase) {
	if(phase == L_MODIFY)
		phase = L_WRITE_ONLY; //L_MODIFY and L_WRITE_ONLY have the same behavior since read/writes to non-ghosts always go directly to the element
		
	if(field->phase != phase) {
		if(field->phase != L_READ_ONLY) {
			//performs the communication to complete the reductions
			//field->communicator.update(field->data, field->first_ghost, field->write_ghosts_offset, val_type, val_length, field->phase, phase);
			doCommunication(field,val_type,val_length,field->phase);
		}
		if(phase != L_READ_ONLY && phase != L_WRITE_ONLY && field->field_template.n_write_ghosts > 0) {
			lUtilValue v;
			lUtilValueIdentity(phase, val_type, &v);
			//set the identity
			size_t type_size = lUtilTypeSize(val_type);
			size_t element_size = type_size * val_length;
			byte * write_ghosts = field->field_template.data + element_size * (field->field_template.first_ghost + field->field_template.write_ghosts_offset);
			doBroadcast(write_ghosts, val_length * field->field_template.n_write_ghosts, val_type,1, (byte*)&v);
		}
		field->phase = phase;
	}
}

void lScalarInit(lContext * ctx, lScalar * scalar, lType val_type, size_t val_length) {
	size_t entry_size = lUtilTypeSize(val_type) * val_length;
	scalar->data = ctx->scalar_data + ctx->scalar_data_bump_pointer;
	scalar->write_entry = (byte*) malloc(entry_size); 
	scalar->phase = L_READ_ONLY;

	scalar->ctx = ctx;
	scalar->scalar_template.data = ctx->device_scalar_data + ctx->scalar_data_bump_pointer;
	scalar->max_dirty_elements = 0;
	scalar->phase = L_READ_ONLY;
	scalar->type = val_type;
	scalar->length = val_length;
	scalar->scalar_template.data = ctx->device_scalar_data + ctx->scalar_data_bump_pointer;
	
	CUDA_DO(cudaMalloc(&scalar->scalar_template.write_table, ctx->n_elements_max * entry_size));

	size_t idx = (scalar - (lScalar*)ctx->scalars);
	scalar->device_scalar = (lkScalar*) ctx->nested_template.scalars + idx;

	
	CUDA_DO(cudaMemcpy(scalar->device_scalar, &scalar->scalar_template, sizeof(lkScalar), cudaMemcpyHostToDevice));

	size_t aligned_size = ((entry_size + SCALAR_BYTE_ALIGNMENT - 1) / SCALAR_BYTE_ALIGNMENT) * SCALAR_BYTE_ALIGNMENT;
	ctx->scalar_data_bump_pointer += aligned_size;
}


L_ALWAYS_INLINE 
void lScalarRead(lContext * ctx, lScalar * scalar, lType element_type, size_t element_length, size_t offset, size_t val_length, void * result) {
	assert(scalar->phase == L_READ_ONLY);
	size_t scalar_size = lUtilTypeSize(element_type);
	memcpy(result,scalar->data + offset * scalar_size, scalar_size * val_length);
}

L_ALWAYS_INLINE 
void lScalarWrite(lContext * ctx, lScalar * scalar, lReduction reduction, lType element_type, size_t element_length, size_t offset, size_t val_length, void * value) {
	assert(scalar->phase == L_READ_ONLY);
	size_t scalar_size = lUtilTypeSize(element_type);
	lUtilValueReduce(scalar->data + offset * scalar_size, reduction, element_type, val_length, value);
	ctx->scalar_dirty = true;
}

MPI_Datatype lTypeToMPIType(lType type) {
	switch(type) {
		case L_INT: return MPI_INT;
		case L_FLOAT: return MPI_FLOAT;
		case L_DOUBLE: return MPI_DOUBLE;
		case L_BOOL: return MPI_UNSIGNED_CHAR;
		case L_STRING: abort();
	}
}

MPI_Op lPhaseToMPIOp(lPhase phase) {
	switch(phase) {
		case L_READ_ONLY: abort();
		case L_WRITE_ONLY: abort();
		case L_MODIFY: abort();
		case L_REDUCE_PLUS: return MPI_SUM;
		case L_REDUCE_MULTIPLY: return MPI_PROD;
		case L_REDUCE_MIN: return MPI_MIN;
		case L_REDUCE_MAX: return MPI_MAX;
		case L_REDUCE_BOR: return MPI_BOR;
		case L_REDUCE_BAND: return MPI_BAND;
		case L_REDUCE_XOR: return MPI_BXOR;
		case L_REDUCE_AND: return MPI_LAND;
		case L_REDUCE_OR: return MPI_LOR;
	}
}


template<typename Type, typename Reduction>
L_ALWAYS_INLINE
void reduceWithThrustTemplate(void * begin, size_t n_elements, size_t n_elements_stride, size_t length, Type identity, Reduction reduction, byte * result) {
	for(int i = 0; i < length; i++) {
		Type * start = (Type*)begin + n_elements_stride * i;
		
		/*
		printf("reducing %d,%d\n",(int)n_elements_stride * i, (int) n_elements);
		Type * foo = new Type[n_elements];
		cudaMemcpy(foo,start,sizeof(Type) * n_elements, cudaMemcpyDeviceToHost);
		for(int i = 0; i < n_elements; i++) {
			printf("%d: %d\n",i,(int)foo[i]);
		}*/
		

		::thrust::device_ptr< Type > t_begin(start);
		::thrust::device_ptr< Type > t_end = t_begin + n_elements;
		((Type*)result)[i] = ::thrust::reduce(t_begin, t_end, identity, reduction);
		CUDA_DO(cudaGetLastError());
	}
}

L_ALWAYS_INLINE
void reduceWithThrust(void * begin, size_t n_elements, size_t n_elements_stride, lType val_type, size_t length, lPhase phase, byte * result) {

	#define DO(r,t,i) reduceWithThrustTemplate(begin, n_elements, n_elements_stride, length, i, ::thrust :: r < t >(), result); break;
	switch(phase) {
		case L_WRITE_ONLY: /*fallthrough*/
		case L_REDUCE_PLUS:
		    switch(val_type) {
		    	case L_INT: DO(plus,int,0)
		    	case L_FLOAT: DO(plus,float,0.f)
		    	case L_DOUBLE: DO(plus,double,0.0)
		    } break;
		case L_REDUCE_MULTIPLY:
		    switch(val_type) {
		    	case L_INT: DO(multiplies,int,1)
		    	case L_FLOAT: DO(multiplies,float,1.f)
		    	case L_DOUBLE: DO(multiplies,double,1.0)
		    } break;
		case L_REDUCE_MIN:
			switch(val_type) {
		    	case L_INT: DO(minimum,int,INT_MAX)
		    	case L_FLOAT: DO(minimum,float,FLT_MAX)
		    	case L_DOUBLE: DO(minimum,double,DBL_MAX)
		    } break;
		case L_REDUCE_MAX:
			switch(val_type) {
		    	case L_INT: DO(maximum,int,INT_MIN)
		    	case L_FLOAT: DO(maximum,float,FLT_MIN)
		    	case L_DOUBLE: DO(maximum,double,DBL_MIN)
		    } break;
		case L_REDUCE_BOR:
			switch(val_type) {
		    	case L_INT: DO(bit_or,int,0)
		    } break;
		case L_REDUCE_BAND:
			switch(val_type) {
		    	case L_INT: DO(bit_and,int,~0)
		    } break;
		case L_REDUCE_XOR:
			switch(val_type) {
		    	case L_INT: DO(bit_xor,int,0)
		    } break;
		case L_REDUCE_AND:
			switch(val_type) {
		    	case L_INT: DO(bit_and,int,1)
		    	case L_BOOL: DO(bit_and,char,true)
		    } break;
		case L_REDUCE_OR:
			switch(val_type) {
		    	case L_INT: DO(bit_or,int,0)
		    	case L_BOOL: DO(bit_or,char,false)
		    } break;
	}
	#undef DO
}

void scalarSetMaxDirtyElements(lScalar * scalar, size_t n_elems) {
	if(scalar->phase != L_READ_ONLY && scalar->max_dirty_elements < n_elems) {
		lUtilValue v;
		lUtilValueIdentity(scalar->phase, scalar->type,&v);
		
		for(int i = 0; i < scalar->length; i++) {
			size_t begin = (scalar->ctx->n_elements_max * i) + scalar->max_dirty_elements;
			size_t size = n_elems - scalar->max_dirty_elements;
			//printf("filling: %d, %d\n",(int)begin,(int)n_elems);
			switch(scalar->type) {
				case L_INT: {
					::thrust::device_ptr<int> data( ((int*)scalar->scalar_template.write_table) + begin);
					:: thrust :: fill_n( data , size, v.i);	
				} break;
				case L_BOOL: {
					::thrust::device_ptr<bool> data( ((bool*)scalar->scalar_template.write_table) + begin);
					:: thrust :: fill_n( data , size, v.b);	
				} break;
				case L_FLOAT: {
					::thrust::device_ptr<float> data( ((float*)scalar->scalar_template.write_table) + begin);
					:: thrust :: fill_n( data , size, v.f);	
				} break;
				case L_DOUBLE: {
					::thrust::device_ptr<double> data( ((double*)scalar->scalar_template.write_table) + begin);
					:: thrust :: fill_n( data , size, v.d);	
				} break;
				case L_STRING: 
					/* nop */	
					break;
			}
		}
		scalar->max_dirty_elements = n_elems;
	}

}
void scalarsSetMaxDirtyElements(lContext * ctx, size_t n_elems) {
	for(size_t j = 0; j < ctx->n_scalars; j++) {
		lScalar * s = (lScalar*) ctx->scalars + j;
		scalarSetMaxDirtyElements(s,n_elems);
	}
}

L_ALWAYS_INLINE
void lScalarEnterPhase(lScalar * scalar, lType val_type, size_t val_length, lPhase phase) {
	if(scalar->phase != phase) {
		//get the data off the gpu and into the write entry
		if(scalar->max_dirty_elements > 0) {
			//get into clean state
			assert(scalar->phase != L_READ_ONLY);
			reduceWithThrust(scalar->scalar_template.write_table,scalar->max_dirty_elements,scalar->ctx->n_elements_max,val_type, val_length, scalar->phase, scalar->write_entry);
			
			scalar->max_dirty_elements = 0;
			//reduction writes to the host value, making the device value stale
			scalar->ctx->scalar_dirty = true;
		}	
		
		//now perform the mpi reduce across nodes given the GPU values
		//perform the reduction required by scalar->phase (or do nothing if we are currently in a read phase)
		if(scalar->phase != L_READ_ONLY) {
			if(scalar->phase == L_WRITE_ONLY) {
				memcpy(scalar->data, scalar->write_entry, lUtilTypeSize(val_type) * val_length);
			} else {
				MPI_Op op = lPhaseToMPIOp(scalar->phase);
				MPI_Datatype type = lTypeToMPIType(val_type);
                byte temp[lUtilTypeSize(val_type) * val_length];
				MPI_Allreduce(scalar->write_entry, temp, val_length, type, op, MPI_COMM_WORLD);
                memcpy(scalar->write_entry, temp, lUtilTypeSize(val_type) * val_length);
				lUtilValueReduce(scalar->data, lUtilPhaseToReduction(scalar->phase), val_type, val_length, scalar->write_entry);
			}
		}
		
		scalar->phase = phase;
	}
	
}

void lSetInitBoundary(lContext * ctx, lSet * set, lElementType type, const char * boundary_name) {
	lDiscoverySetInitRanges(ctx->boundary_reader, type, boundary_name, numberOfOwnedElementsOfType(ctx,type), ctx->local_to_global[(int)type].begin(), &set->ranges, &set->size, &set->local_size, &set->id); 
}

L_ALWAYS_INLINE 
size_t lSetSize(lContext * ctx, lSet * set) {
	return set->size;
}

bool elementLess(const PrintElement & lhs, const PrintElement & rhs) {
	return lhs.order < rhs.order;
}

void printBufferFlush(lContext * ctx) {
#ifndef NDEBUG
	CUDA_DO(cudaGetLastError());
	PrintBuffer * buffer  = ctx->nested_template.print_buffer;
	
	PrintSizes sizes = {0,0};
	CUDA_DO(cudaMemcpy(&sizes,&buffer->sizes,sizeof(PrintSizes), cudaMemcpyDeviceToHost));
	

	size_t element_size = sizes.n_elements * sizeof(PrintElement);
	PrintElement * elements = (PrintElement*) malloc(element_size);
	byte * data = (byte*) malloc(sizes.n_bytes);

	CUDA_DO(cudaMemcpy(data,&buffer->data[0],sizes.n_bytes,cudaMemcpyDeviceToHost));
	CUDA_DO(cudaMemcpy(elements,&buffer->elements[0],element_size,cudaMemcpyDeviceToHost));
	
	std::sort(elements,elements + sizes.n_elements, elementLess);

	uint32_t smallest_valid_order = ctx->nested_template.print_count;
	for(size_t i = 0; i < sizes.n_elements; i++) {
		PrintElement * e =  &elements[i];
		uint32_t order =  (uint32_t) e->order;
		if(order >= smallest_valid_order) {
			if(e->type == L_STRING)
				std::cout << data + e->offset;
			else {
				//ensure correct alignment regardless of buffer alignment
				size_t n_elements = ((e->rows == 0) ? 1 : e->rows) * ((e->cols == 0) ? 1 : e->cols);
				lUtilValue v[n_elements];
				memcpy(&v, data + e->offset, n_elements * lUtilTypeSize(e->type)); 
				lPrintValue(ctx,e->type, e->rows, e->cols,&v);
			}
			ctx->nested_template.print_count = std::max(ctx->nested_template.print_count,(size_t)order);
		}
	}
	
	printBufferClearSizes(buffer);

	free(elements);
	free(data);
#endif

}



void kernelBeginTiming(lContext * ctx) {
	CUDA_DO(cudaEventCreate(&ctx->start_event));
	CUDA_DO(cudaEventCreate(&ctx->end_event));
	CUDA_DO(cudaEventRecord(ctx->start_event,0));
}
void kernelEndTiming(lContext * ctx,int id) {
	cudaEventRecord(ctx->end_event);
	cudaEventSynchronize(ctx->end_event);
	float elapsed_time;
	cudaEventElapsedTime(&elapsed_time, ctx->start_event, ctx->end_event);
	cudaEventDestroy(ctx->start_event);
	cudaEventDestroy(ctx->end_event);
	//fprintf(stderr,"kernel_%d: %fms\n", id, elapsed_time);
}

L_ALWAYS_INLINE L_RUNTIME_UNNESTED
void lKernelRun(lContext * ctx,  lSet * set, lElementType typ, int id, void (*kernel)(lkContext), lStencilData stencil) {
	
	if(ctx->scalar_dirty) {
		CUDA_DO(cudaMemcpy(ctx->device_scalar_data,ctx->scalar_data,ctx->scalar_data_bump_pointer,cudaMemcpyHostToDevice));
		ctx->scalar_dirty = false;
	}

#ifndef NDEBUG
	ctx->nested_template.print_context.type = typ;
	ctx->nested_template.print_context.parent = NULL;
	ctx->nested_template.print_context.child = NULL;
	ctx->nested_template.print_context_top_level_count = ctx->print_context_top_level_count++;
#endif

	if(stencil.is_trivial) { /* TODO: determine if the coloring for this kernel is trivially 1 color */
		//prepare scalars by inserting identity elements into their write table up to the length of the largest kernel launch
		size_t max_size = 0;
		for(size_t i = 0; i < set->ranges.size(); i++) {
			uint32_t start = set->ranges[i].first;
			uint32_t end = set->ranges[i].second;
			max_size = std::max(max_size,(size_t)end - start);
		}
		
		scalarsSetMaxDirtyElements(ctx,max_size);

		//launch kernels, 1 per range
		
		//kernelBeginTiming(ctx);
		
		for(size_t i = 0; i < set->ranges.size(); i++) {
			uint32_t start = set->ranges[i].first;
			uint32_t end = set->ranges[i].second;
			
			ctx->nested_template.first_element = start;
			ctx->nested_template.n_elements = end - start;
			ctx->nested_template.colored_elements = NULL;

			size_t n_blocks,n_elements;
			kernelSize(end - start, &n_blocks, &n_elements);
			//LOG_LOG("launching %d elements in kernel %d\n",end - start, (int) id);
			if(end - start > 0)
				kernel<<<n_blocks,n_elements>>>(ctx->nested_template);
			CUDA_DO(cudaThreadSynchronize());
			CUDA_DO(cudaGetLastError());
			
			printBufferFlush(ctx);
		}
		
		//kernelEndTiming(ctx,id);
		
	} else {
		//coloring, first check to see if we already have a cached data structure
		std::pair<uint32_t,uint32_t> key(set->id,id);
		std::map< std::pair<uint32_t,uint32_t>, lColoringResult >::iterator it = ctx->coloring_cache.find(key),
		                                                                  end = ctx->coloring_cache.end();
		lColoringResult * coloring;
		if(it != end) {
			coloring = &it->second;
		} else {
			coloring =  &ctx->coloring_cache[key]; //insert into cache
			printf("(%d,%d): generating coloring.\n",(int)set->id,(int)id);
			
			size_t max_local_elements = 0;
			for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++) {
				max_local_elements = std::max(max_local_elements,numberOfLocalElementsOfType(ctx,(lElementType)i));
			}
			//coloring needs to know how many elements might exist in the mesh, including ghost elements, so that it allocates enough space for the interference graph
			lColoringGenerate(max_local_elements,&ctx->mesh,set->local_size,&set->ranges,stencil.stencil_fn,coloring); //calculate coloring
			printf("(%d,%d): finished coloring, found %d colors.\n",(int)set->id,(int)id,(int)coloring->n_colors);
			//transfer value array to gpu
			uint32_t * device_values;
			CUDA_DO(cudaMalloc(&device_values,sizeof(uint32_t) * set->local_size));
			CUDA_DO(cudaMemcpy(device_values, coloring->values, sizeof(uint32_t) * set->local_size, cudaMemcpyHostToDevice));
			delete [] coloring->values;
			coloring->values = device_values;
		}

		//prepare scalars by inserting identity elements into their write table up to the length of the largest kernel launch
		size_t max_size = 0;
		for(size_t i = 0; i < coloring->n_colors; i++) {
			uint32_t start = coloring->row_idx[i];
			uint32_t end = coloring->row_idx[i+1];
			max_size = std::max(max_size,(size_t)end - start);
		}

		scalarsSetMaxDirtyElements(ctx,max_size);

		//kernelBeginTiming(ctx);
		
		for(size_t i = 0; i < coloring->n_colors; i++) {
			uint32_t start = coloring->row_idx[i];
			uint32_t end = coloring->row_idx[i+1];

			ctx->nested_template.n_elements = end - start;
			ctx->nested_template.colored_elements = coloring->values + start;
			size_t n_blocks, n_elements;
			kernelSize(end - start, &n_blocks,&n_elements);
			if(end - start > 0)
				kernel<<<n_blocks,n_elements>>>(ctx->nested_template);
			CUDA_DO(cudaGetLastError());
			printBufferFlush(ctx);
		}
		
		//kernelEndTiming(ctx,id);
		
	}

}

/*kernels have the signature and structure: 
void kernel_name(lkContext * ctx) {
	lkElement e;
	if(lkGetActiveElement(ctx,&e)) {
	  <...>
	}
}*/

L_RUNTIME_UNNESTED
void lPrintBegin(lContext * ctx) {
#ifndef NDEBUG
	std::cout << ctx->print_context_top_level_count++ << " liszt: ";
#else
	std::cout << "liszt: ";
#endif
}
L_RUNTIME_UNNESTED
void lPrintEnd(lContext * ctx) {
	std::cout << "\n";
}

void printValue(lType typ, const void * value) {
	switch(typ) {
		case L_FLOAT:
			std::cout << *(float*)value;
			break;
		case L_INT:
			std::cout << *(int*)value;
			break;
		case L_BOOL:
			std::cout << *(bool*)value;
			break;
		case L_DOUBLE:
			std::cout << *(double*)value;
			break;
		case L_STRING:
			std::cout << (const char *)value;
			break;
	}
}

L_RUNTIME_UNNESTED
void lPrintValue(lContext * ctx, lType typ, size_t r, size_t c,  const void * value) {
	byte * v = (byte *) value;
	size_t size = lUtilTypeSize(typ);
	if(c != 0) { 
		std::cout << "[";
		for(size_t i = 0; i < r; i++) {
			for (size_t j = 0; j < c; j++) {
				printValue(typ,v);
				v += size;
				if(j < c - 1) std::cout << ",";
			}
			if(i < r - 1) std::cout << "|";
			else std::cout << "]";  
		}
	} else if(r != 0) {
		std::cout << "(";
		for(int i = 0; i < r - 1; i++) {
			printValue(typ,v);
			std::cout << ",";
			v += size;
		}
		printValue(typ,v);
		std::cout << ")";
	} else {
		printValue(typ,v);
	}
}
double lWallTime(lContext * ctx) { //counter, time reported in seconds
	timeval currentTime;
	gettimeofday(&currentTime, 0);
	return currentTime.tv_sec + 1e-6 * currentTime.tv_usec;
}

L_RUNTIME_NESTED
void lkPrintBegin(lkContext * ctx) {
#ifndef NDEBUG
	lkPrintValue(ctx,L_INT,0,0,&ctx->print_context_top_level_count);
	DPrintContext * pc = &ctx->print_context;
	while(1) {
		switch(pc->type) {
			case L_VERTEX: lkPrintValue(ctx,L_STRING,0,0,(void*)":v"); break;
			case L_EDGE: lkPrintValue(ctx,L_STRING,0,0,(void*)":e"); break;
			case L_FACE: lkPrintValue(ctx,L_STRING,0,0,(void*)":f"); break;
			case L_CELL: lkPrintValue(ctx,L_STRING,0,0,(void*)":c"); break;
		}
		int global_id = ctx->local_to_global[pc->type][pc->id];
		
		lkPrintValue(ctx,L_INT,0,0,&global_id);
		lkPrintValue(ctx,L_STRING,0,0,(void*)"-");
		lkPrintValue(ctx,L_INT,0,0,&pc->count);

		if(pc->child == NULL) {
			pc->count++;
			break;
		} else {
			pc = pc->child;
		}
	}
	lkPrintValue(ctx,L_STRING,0,0,(void*)" liszt: ");
#endif
}

L_RUNTIME_NESTED
void lkPrintEnd(lkContext * ctx) {
#ifndef NDEBUG
	lkPrintValue(ctx,L_STRING,0,0,(void*)"\n");
#endif
}

L_RUNTIME_NESTED
void lkPrintValue(lkContext * ctx, lType typ, size_t r, size_t c, const void * value) {
#ifndef NDEBUG
	//reserve an element
	uint32_t offset = atomicAdd(&ctx->print_buffer->sizes.n_elements,1);
	PrintElement * elem = &ctx->print_buffer->elements[offset];
	elem->type = typ;
	elem->rows = r;
	elem->cols = c;
	uint64_t id = blockIdx.x * blockDim.x + threadIdx.x;
	
	size_t data_size;


	if(typ == L_STRING) {
		const char * str = (const char *) value;
		data_size = deviceStrlen(str) + 1;
	} else {
		size_t rows = (r == 0) ? 1 : r;
		size_t cols = (c == 0) ? 1 : c;
		data_size = lUtilTypeSize(typ) * rows * cols;
	}
	elem->offset = atomicAdd(&ctx->print_buffer->sizes.n_bytes,data_size);
	deviceMemcpy(ctx->print_buffer->data + elem->offset, (const char * ) value, data_size);
	
	//must go last to ensure the host will not see a valid print count before all the data is in place
	//do we need a fence before this?
	elem->order = (id << 32) + ctx->print_count;
	
	ctx->print_count++;
#endif
}

L_RUNTIME_NESTED
void lkPrintElement(lkContext * ctx, lElementType typ, lkElement e) {
	const char * name;
	switch(typ) {
		case L_VERTEX:
			name = "Vertex ";
			break;
		case L_EDGE:
			name = "Edge ";
			break;
		case L_FACE:
			name = "Face ";
			break;
		case L_CELL:
			name = "Cell ";
			break;
	}
	
	lkPrintValue(ctx,L_STRING,0,0,(void*)name);
	int id = ctx->local_to_global[typ][elementID(e)];
	lkPrintValue(ctx,L_INT,0,0,&id);
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
lkFields * lkGetFields(lkContext * ctx) {
	return ctx->fields;
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
lkScalars * lkGetScalars(lkContext * ctx) {
	return ctx->scalars;
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
bool lkGetActiveElement(lkContext * ctx, lkElement * e) {
	int id = blockIdx.x * blockDim.x + threadIdx.x;
	if(id < ctx->n_elements) {
		if(ctx->colored_elements != NULL)
			e->data = ctx->colored_elements[id];
		else
			e->data = id + ctx->first_element;
#ifndef NDEBUG
		ctx->current_print_context = &ctx->print_context;
		ctx->print_context.id = elementID(*e);
		ctx->print_context.count = 0;
#endif
		return true;
	} else 
		return false;
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkFieldRead(lkField * f, lkElement e, lType element_type, size_t element_length, size_t val_offset, size_t val_length,  void * result) {
	size_t offset,size;
	lUtilValueLocation(elementID(e),element_type,element_length,val_offset,val_length,&offset,&size);
	lUtilValueReduce(result, L_ASSIGN, element_type, val_length,f->data + offset);
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkFieldWrite(lkField * field, lkElement e, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value) {
	uint32_t id = elementID(e);
	uint32_t adjusted_id = (id >= field->first_ghost) ? (id + field->write_ghosts_offset) : id; 
	size_t offset,size;
	lUtilValueLocation(adjusted_id,element_type,element_length,val_offset,val_length,&offset,&size);
	lUtilValueReduce(field->data + offset, reduction, element_type, val_length, value);
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkScalarRead(lkContext * ctx, lkScalar * scalar, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * result){
	lUtilValueReduce(result, L_ASSIGN, element_type, val_length, scalar->data + val_offset * lUtilTypeSize(element_type));
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkScalarWrite(lkContext * ctx, lkScalar * scalar, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value_) {
	int id = blockDim.x * blockIdx.x + threadIdx.x;
	size_t elem_size = lUtilTypeSize(element_type);
	byte * value =  (byte *) value_;
	for(int i = 0; i < val_length; i++) {
		size_t offset_elem = (ctx->n_elements_max * (val_offset + i) + id);
		//printf("(%d) reducing into element %d\n",(int)id,(int)offset_elem);
		size_t offset = elem_size * offset_elem;
		lUtilValueReduce(scalar->write_table + offset, reduction, element_type, 1, value + elem_size * i);
	}
}

L_ALWAYS_INLINE L_RUNTIME_NESTED
size_t lkSetSize(lkContext * ctx, lkSet * set) {
	return set->set.size;
}
L_ALWAYS_INLINE L_RUNTIME_NESTED
void lkSetGetIterator(lkContext * ctx, lkSet * set, lElementType typ, lkIterator * iterator) {

	lNestedSetGetIterator(&set->set,&iterator->it);
	
#ifndef NDEBUG

	iterator->print_context.type = typ;
	iterator->print_context.count = 0;
	iterator->print_context.parent = ctx->current_print_context;
	ctx->current_print_context->child = &iterator->print_context;
	iterator->print_context.child = NULL;
	ctx->current_print_context = ctx->current_print_context->child;

#endif
}

L_ALWAYS_INLINE
bool lkIteratorNext(lkContext * ctx, lkIterator * iterator, lElementType typ, lkElement * e, int * lbl) {
	bool result = lNestedIteratorNext(&iterator->it, e, lbl);
#ifndef NDEBUG
	if(result) {
		iterator->print_context.id = elementID(*e);
		iterator->print_context.count = 0;
	} else {
		ctx->current_print_context = ctx->current_print_context->parent;
		ctx->current_print_context->child = NULL;
		ctx->current_print_context->count++;
	}
#endif
	return result;
}

//now all the topological functions ....
void lVerticesOfMesh(lContext * ctx, lSet * set) {
	set->id = -L_VERTEX - 1;
	set->size = numberOfGlobalElementsOfType(ctx,L_VERTEX);
	set->local_size = numberOfOwnedElementsOfType(ctx,L_VERTEX);
	set->ranges.push_back(std::make_pair(0, set->local_size));
}
void lEdgesOfMesh(lContext * ctx, lSet * set) {
	set->id = -L_EDGE - 1;
	set->size = numberOfGlobalElementsOfType(ctx,L_EDGE);
	set->local_size = numberOfOwnedElementsOfType(ctx,L_EDGE);
    set->ranges.push_back(std::make_pair(0, set->local_size));
}
void lFacesOfMesh(lContext * ctx, lSet * set) {
	set->id = -L_FACE - 1;
	set->size = numberOfGlobalElementsOfType(ctx,L_FACE);
	set->local_size = numberOfOwnedElementsOfType(ctx,L_FACE);
    set->ranges.push_back(std::make_pair(0, set->local_size));
}
void lCellsOfMesh(lContext * ctx, lSet * set) {
	set->id = -L_CELL - 1;
	set->size = numberOfGlobalElementsOfType(ctx,L_CELL) - 1;
	set->local_size = numberOfOwnedElementsOfType(ctx,L_CELL);
    set->ranges.push_back(std::make_pair(1, set->local_size));
}

int lkIDOfVertex(lkContext * ctx, lkElement e) {
	return ctx->local_to_global[L_VERTEX][elementID(e)];
}
int lkIDOfEdge(lkContext * ctx, lkElement e) {
	return ctx->local_to_global[L_EDGE][elementID(e)];
}
int lkIDOfFace(lkContext * ctx, lkElement e) {
	return ctx->local_to_global[L_FACE][elementID(e)];
}
int lkIDOfCell(lkContext * ctx, lkElement e) {
	return ctx->local_to_global[L_CELL][elementID(e)];
}

#define NESTED_TOPOLOGY_GET_MESH(ctx) (ctx->mesh)
#define NESTED_TOPOLOGY_GET_SET(set) (set->set)
#include "nested_topology_functions.inc"

struct lsElement {  lColoringElement c; lDiscoveryElement d;};
struct lsSet {  lColoringElement c; lDiscoverySet d;};
struct lsIterator {  lColoringElement c; lDiscoveryIterator d; };


lDiscoveryElement * lDiscoveryGetElement(lsElement * e) { return &e->d; }
lDiscoverySet * lDiscoveryGetSet(lsSet * e) { return &e->d; }
lDiscoveryIterator * lDiscoveryGetIterator(lsIterator * e) { return &e->d; }
lDiscoveryContext * lDiscoveryGetContext(lsContext * e) { return &e->d; }



#include "vector.h"
#include "matrix.h"
#include "liszt_math.h"

//code generated from the liszt compiler is now included here after runtime-specific
//headers have been included

#include "generated.cpp"

