
#include <sys/time.h>
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <assert.h>

#include "liszt_runtime.h"
#include "boundary_set_reader.h"
#include "mesh.h"
#include "print_context.h"
#include "runtime_util.h"
#include "Array.h"
#include "ParallelLoad/ParallelLoad.h"
#include "nested_topology.h"
#include "ghost_discovery.h"
#include "vector.h"

#define L_KERNEL L_ALWAYS_INLINE
#define L_NESTED L_ALWAYS_INLINE
#define L_UNNESTED
#define L_STENCIL

//define opaque data types and implementation for specific runtime

struct lContext {
	lFields * fields;
	lSets * sets;
	lScalars * scalars;
	shared_ptr<MeshIO::LisztFileReader> mesh_reader; //we hang on to this so we can load the positions on demand
	shared_ptr<BoundarySetReader> boundary_reader;
	
	Mesh mesh;
	PrintContext print_context;


	shared_ptr<NameService> name_service;
	Array<uint32_t> local_to_global[L_ELEMENT_TYPE_SIZE];
	size_t num_global_elements[L_ELEMENT_TYPE_SIZE];
	size_t first_ghost[L_ELEMENT_TYPE_SIZE];
	size_t num_ghosts[L_ELEMENT_TYPE_SIZE];
	Array<FieldCommunicationPattern> field_communication_patterns; //map from field_id -> communication pattern
};


//lField and lkField are the same object in this runtime
struct lField {
	byte * data;
	FieldCommunicator communicator;
	size_t first_ghost;
	size_t write_ghosts_offset;
	size_t n_write_ghosts;
	lPhase phase;
};

struct lkField {
	byte * data;
	FieldCommunicator communicator;
	size_t first_ghost;
	size_t write_ghosts_offset;
	size_t n_write_ghosts;
	lPhase phase;
};

//lScalar and lkScalar are the same object as well
struct lScalar {
	byte * data;
	byte * write_entry;
	lPhase phase;
};
struct lkScalar {
	byte * data;
	byte * write_entry;
	lPhase phase;
};

struct lSet {
	BoundarySetReader::range_list_t ranges;
	size_t size; //total size, not partitioned size
};

struct lkContext {
	lContext * ctx;
	lkElement element;
};

struct lkSet {
	lNestedSet set;
};
struct lkIterator {
	lNestedIterator it;
};

L_ALWAYS_INLINE 
size_t numberOfOwnedElementsOfType(lContext * ctx, lElementType typ) {
	return ctx->first_ghost[(int)typ];
}
L_ALWAYS_INLINE 
size_t numberOfGlobalElementsOfType(lContext * ctx, lElementType typ) {
	return ctx->num_global_elements[(int)typ];
}

L_ALWAYS_INLINE 
size_t numberOfGhostElementsOfType(lContext * ctx, lElementType typ) {
	return ctx->num_ghosts[(int)typ];
}

L_ALWAYS_INLINE 
size_t numberOfLocalElementsOfType(lContext * ctx, lElementType typ) {
	return numberOfOwnedElementsOfType(ctx,typ) 
	     + numberOfGhostElementsOfType(ctx,typ);
}


//Unnested Runtime Calls           

struct lsElement { lDiscoveryElement d; };
struct lsSet { lDiscoverySet d; };
struct lsIterator { lDiscoveryIterator d; };
struct lsContext { lDiscoveryContext d; };

void initializeContext(lContext * ctx, void (*entry_stencil)(lsFunctionTable*,lsContext*), lProgramArguments * arguments, size_t n_fields, size_t n_sets, size_t n_scalars) {
	
	int zero = 0;
	MPI_Init(&zero,NULL);
	
	if(arguments->redirect_output_to_log) {
		int mpi_rank;
		std::ostringstream ss;
		MPI_Comm_rank(MPI_COMM_WORLD,&mpi_rank);
		ss << "liszt_runtime_" << mpi_rank << ".log";
		int file = fileno(fopen( /*(mpi_rank == 0) ?*/ ss.str().c_str() /*: "/dev/null"*/,"a"));
		dup2(file,1);
		dup2(file,2);
	}
	
	//allocate memory to hold global state
	ctx->fields = (lFields*) new lField[n_fields];
	ctx->sets = (lSets*) new lSet[n_sets];
	ctx->scalars = (lScalars*) new lScalar[n_scalars];
	
	
	//set up the file reader which reads a mesh files, note that this will only read objects from a file when requested
	ctx->mesh_reader.reset(new MeshIO::LisztFileReader()); 
	ctx->mesh_reader->init(arguments->mesh_file);
	//every node reads the entire header and boundary set list from liszt mesh file
	const MeshIO::LisztHeader & h = ctx->mesh_reader->header();
	ctx->num_global_elements[L_VERTEX] = h.nV;
	ctx->num_global_elements[L_EDGE] = h.nE;
	ctx->num_global_elements[L_FACE] = h.nF;
	ctx->num_global_elements[L_CELL] = h.nC;
	
	ctx->boundary_reader.reset(new BoundarySetReader);
	ctx->boundary_reader->init(h.nBoundaries,ctx->mesh_reader->boundaries());
	
	

	const LisztHeader * header = &ctx->mesh_reader->header();
	ctx->boundary_reader->init(header->nBoundaries, ctx->mesh_reader->boundaries());
	

	//see individual objects for what each does during initialization in detail
	//name_service answers queries about who has mesh information
	//e.g. who is responsible for knowing the owner of cell x?
	ctx->name_service.reset(new NameService(header));

	//the topology service answers queries about topology
	//e.g. get me all the facet edges for cell x? individual requests are aggregated for performance
	shared_ptr<TopologyService> topology_service(new TopologyService(ctx->name_service, ctx->mesh_reader));

	//owner service determines ownership through partition and then answers questions like
	//what node owns cell x? or what are all the nodes that my local node owns?
	shared_ptr<OwnerService> owner_service( new OwnerService(ctx->name_service,topology_service->partitionedTopology()));

	//local topology stores the topology necessary to run the program and is used to construct the mesh that will be used
	//at runtime
	shared_ptr<LocalTopology> local_topology(new LocalTopology(ctx->name_service->self(),owner_service->locallyOwnedElements()));

	//local memory stores the memory access patterns to field, used to determine field communication.
	shared_ptr<LocalMemory> local_memory(new LocalMemory(n_fields));

	//The stencil is the abstract representation of the program for the purpose of topology discovery and field memory discovery
	//it will populate local_topology and local_memory based on the initial partition
	lsContext lsCtx;
	lDiscoveryRun(entry_stencil,&lsCtx, ctx->name_service,topology_service,owner_service,ctx->boundary_reader,local_topology,local_memory);
	
	//convert the facet edge mesh of local topology into a fast Mesh object.
	local_topology->buildLocalMesh(&ctx->mesh);
	
	//initialize meta-data needed to later initialize fields
	//local_topology will no longer exist when fields are constructed, so we save the relevent information here
	for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++) {
		ctx->first_ghost[i] = local_topology->firstGhost((lElementType)i);
		ctx->num_ghosts[i] = local_topology->numberGhost((lElementType)i);
	}
	//we need to maintain a local ID -> global ID mapping for correct implementation of liszt's ID function
	for(size_t i = 0; i < L_ELEMENT_TYPE_SIZE; i++) {
		local_topology->buildLocalToGlobal((lElementType)i,&ctx->local_to_global[i]);
	}
	
	//construct a map from field id to the information about field reads/writes it performs
	//when a field is initialized it will use these values to set up its comm structures
	
	//Fill in field communication patterns
	ctx->field_communication_patterns.reset(n_fields);
	for(size_t i = 0; i < n_fields; i++) {
		ctx->field_communication_patterns[i].init(local_topology,&local_memory->fields[i]);
	}

	local_topology->print_debug(ctx->name_service);

	//note that most of the parallel load infrastructure goes out of scope when init exits, freeing all memory but the things
	//that are necessary to perform the run
}
void lExec(void (*entry_point)(lContext*),void (*entry_stencil)(lsFunctionTable*,lsContext*), lProgramArguments * arguments, size_t n_fields, size_t n_sets, size_t n_scalars) {
	
	lContext ctx;
	
	initializeContext(&ctx,entry_stencil,arguments,n_fields,n_sets,n_scalars);
	
	//launch the program
	
	entry_point(&ctx);
	
    MPI_Finalize();
	//do cleanup
}


L_ALWAYS_INLINE
lFields * lGetFields(lContext * ctx) {
	return ctx->fields;
}
L_ALWAYS_INLINE
lSets * lGetSets(lContext * ctx) {
	return ctx->sets;
}
L_ALWAYS_INLINE
lScalars * lGetScalars(lContext * ctx) {
	return ctx->scalars;
}
void lFieldInit(lContext * ctx, lField * field, int id, lElementType key_type, lType val_type, size_t val_length) {

	field->communicator.init(&ctx->field_communication_patterns[id]);
	
	size_t n_owned = numberOfOwnedElementsOfType(ctx, key_type);
	size_t n_ghosts = numberOfGhostElementsOfType(ctx,key_type);
	size_t n_read_ghosts = (field->communicator.communication_pattern->reads_ghosts) ? n_ghosts : 0;
	field->n_write_ghosts = (field->communicator.communication_pattern->writes_ghosts) ? n_ghosts : 0;
	
	field->first_ghost = n_owned;
	field->write_ghosts_offset = n_read_ghosts;
	
	size_t n_elems = n_owned + n_read_ghosts + field->n_write_ghosts;
	
	field->data = (byte*) malloc(n_elems * val_length * lUtilTypeSize(val_type));
	field->phase = L_READ_ONLY;
}

size_t fieldNumberOfReadElements(lContext * ctx, lField * field, lElementType key_type) {
	size_t n_owned = numberOfOwnedElementsOfType(ctx, key_type);
	size_t n_ghosts = numberOfGhostElementsOfType(ctx,key_type);
	size_t n_read_ghosts = (field->communicator.communication_pattern->reads_ghosts) ? n_ghosts : 0;
	return n_owned + n_read_ghosts;
}
L_ALWAYS_INLINE
void lFieldBroadcast(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, void * data) {
	size_t n_elems = fieldNumberOfReadElements(ctx,field,key_type);
	for(size_t i = 0; i < n_elems; i++) {
		size_t offset,size;
		lUtilValueLocation(i,val_type,val_length,0,val_length,&offset,&size);
		lUtilValueReduce(field->data + offset, L_ASSIGN, val_type, val_length, data);
	}
}

L_RUNTIME_UNNESTED void lFieldLoadData(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, const char * key) {
	if (!strcmp(key,"position")) {
        assert(key_type == L_VERTEX);
        assert(val_type == L_DOUBLE || val_type == L_FLOAT);
    }

    size_t n_read_elems = fieldNumberOfReadElements(ctx,field,key_type);

    // load field data from our mesh file in parallel
    uint32_t begin, end;

    // query name service to see which range this node is responsible for
    ctx->name_service->localPartition(key_type, &begin, &end);
    // read field data range from file
    const byte *data = ctx->mesh_reader->fieldData(key_type, val_type, val_length, key, begin, end);

    SizedAllToAll<uint32_t> request;
    request.beginCount();

    // Field data now needs to be requested from node responsible for partition and sent 
    // to whoever potentially reads them

    // for everything in this field's read entries, request field data
    for (size_t i = 0; i < n_read_elems; i++) {
        uint32_t global_id = ctx->local_to_global[key_type][i];
        request.count(ctx->name_service->partitionFor(key_type,global_id));
    }

    request.beginData();
    for (size_t  i = 0; i < n_read_elems; i++) {
        uint32_t global_id = ctx->local_to_global[key_type][i];
        request.data(ctx->name_service->partitionFor(key_type,global_id),global_id);
    }
        
    size_t n_part;
    const int *offsets;
    const uint32_t *rq_data;
    size_t total_data = request.endData(&n_part, &offsets, &rq_data);
    
    size_t elem_size = val_length * lUtilTypeSize(val_type);
    SizedAllToAll<uint8_t> response;

    response.beginCount();
    for (size_t i = 0; i < n_part; i++) {
        size_t begin = offsets[i];
        size_t end   = (i + 1 == n_part) ? total_data : offsets[i + 1];
        response.count(i, (end - begin) * elem_size);
    }

    // offsets - array listing the offset in rq_data where each node's request begins/ends
    // rq_data - array of global ids of requested elements
    // begin   - id of global element that is the 0th element of this node's partition

    // respond to the initial request with the data read from the local range we are responsible for
    response.beginData();
    for (size_t i = 0; i < n_part; i++) {
        size_t pend   = (i + 1 == n_part) ? total_data : offsets[i + 1];
        size_t pbegin = offsets[i];
        for (size_t j = pbegin; j < pend; j++) {
            size_t e = rq_data[j] - begin;
            
            uint8_t elem_data[elem_size];
            memcpy(elem_data, data + e * elem_size, elem_size);
            
            for (int k = 0; k < elem_size; k++)
                response.data(i, elem_data[k]);
        }
    }
    std::free((void *) data);


    // Reading request response as a client
    const uint8_t *rdata;
    total_data = response.endData(&n_part,&offsets,&rdata);
  
    Array<int> roffsets(n_part);
    std::copy(offsets,offsets+n_part,roffsets.begin());

    // fill the field with the response data
    //   (Remember, elem_size elems are read for each field entry)
    for (size_t i = 0; i < n_read_elems; i++) {
        uint32_t global_id = ctx->local_to_global[key_type][i];
        PartitionID part   = ctx->name_service->partitionFor(key_type, global_id);

        memcpy(field->data + i * elem_size, rdata + roffsets[part], elem_size);
        roffsets[part] += elem_size;
    }
}

L_RUNTIME_UNNESTED void lFieldSaveData(lContext * ctx, lField * field, lElementType key_type, lType val_type, size_t val_length, const char * key) {
	lFieldEnterPhase(field,val_type, val_length, L_READ_ONLY);
	printf("TODO: save field %s\n",key);
}

void lFieldEnterPhase(lField * field, lType val_type, size_t val_length, lPhase phase) {
	if(phase == L_MODIFY)
		phase = L_WRITE_ONLY; //L_MODIFY and L_WRITE_ONLY have the same behavior since read/writes to non-ghosts always go directly to the element
		
	if(field->phase != phase) {
		if(field->phase != L_READ_ONLY) {
			//performs the communication to complete the reductions
			field->communicator.update(field->data, field->first_ghost, field->write_ghosts_offset, val_type, val_length, field->phase, phase);
		}
		if(phase != L_READ_ONLY && phase != L_WRITE_ONLY && field->n_write_ghosts > 0) {
			lUtilValue v;
			lUtilValueIdentity(phase, val_type, &v);
			//set the identity
			size_t type_size = lUtilTypeSize(val_type);
			size_t element_size = type_size * val_length;
			byte * write_ghosts = field->data + element_size * (field->first_ghost + field->write_ghosts_offset);
			for(size_t i = 0; i < val_length * field->n_write_ghosts; i++)
				memcpy(write_ghosts + i * type_size, &v, type_size);
			
		}
		field->phase = phase;
	}
}

void lScalarInit(lContext * ctx, lScalar * scalar, lType val_type, size_t val_length) {
	size_t entry_size = lUtilTypeSize(val_type) * val_length;
	scalar->data = (byte*) malloc(2 * entry_size);
	scalar->write_entry = scalar->data + entry_size; 
	scalar->phase = L_READ_ONLY;
}


L_ALWAYS_INLINE 
void lScalarRead(lContext * ctx, lScalar * scalar, lType element_type, size_t element_length, size_t offset, size_t val_length, void * result) {
	size_t scalar_size = lUtilTypeSize(element_type);
	memcpy(result,scalar->data + offset * scalar_size, scalar_size * val_length);
}

L_ALWAYS_INLINE 
void lScalarWrite(lContext * ctx, lScalar * scalar, lReduction reduction, lType element_type, size_t element_length, size_t offset, size_t val_length, void * value) {
	size_t scalar_size = lUtilTypeSize(element_type);
	lUtilValueReduce(scalar->data + offset * scalar_size, reduction, element_type, val_length, value);
}

MPI_Datatype lTypeToMPIType(lType type) {
	switch(type) {
		case L_INT: return MPI_INT;
		case L_FLOAT: return MPI_FLOAT;
		case L_DOUBLE: return MPI_DOUBLE;
		case L_BOOL: return MPI_UNSIGNED_CHAR;
		case L_STRING: abort();
	}
}

MPI_Op lPhaseToMPIOp(lPhase phase) {
	switch(phase) {
		case L_READ_ONLY: abort();
		case L_WRITE_ONLY: abort();
		case L_MODIFY: abort();
		case L_REDUCE_PLUS: return MPI_SUM;
		case L_REDUCE_MULTIPLY: return MPI_PROD;
		case L_REDUCE_MIN: return MPI_MIN;
		case L_REDUCE_MAX: return MPI_MAX;
		case L_REDUCE_BOR: return MPI_BOR;
		case L_REDUCE_BAND: return MPI_BAND;
		case L_REDUCE_XOR: return MPI_BXOR;
		case L_REDUCE_AND: return MPI_LAND;
		case L_REDUCE_OR: return MPI_LOR;
	}
}


L_ALWAYS_INLINE
void lScalarEnterPhase(lScalar * scalar, lType val_type, size_t val_length, lPhase phase) {
	if(scalar->phase != phase) {
		
		//perform the reduction required by scalar->phase (or do nothing if we are currently in a read phase)
		if(scalar->phase != L_READ_ONLY) {
			if(scalar->phase == L_WRITE_ONLY) {
				memcpy(scalar->data, scalar->write_entry, lUtilTypeSize(val_type) * val_length);
			} else {
				MPI_Op op = lPhaseToMPIOp(scalar->phase);
				MPI_Datatype type = lTypeToMPIType(val_type);
                byte temp[lUtilTypeSize(val_type) * val_length];
				MPI_Allreduce(scalar->write_entry, temp, val_length, type, op, MPI_COMM_WORLD);
                memcpy(scalar->write_entry, temp, lUtilTypeSize(val_type) * val_length);
				lUtilValueReduce(scalar->data, lUtilPhaseToReduction(scalar->phase), val_type, val_length, scalar->write_entry);
			}
		}
		
		//set identity for upcoming reduction if the new phase is a reduction
		if(phase != L_READ_ONLY && phase != L_WRITE_ONLY) {
			
			for(size_t i = 0; i < val_length; i++) {
				byte * data = (scalar->write_entry + i * lUtilTypeSize(val_type));
				lUtilValueIdentity(phase, val_type, (lUtilValue*) data);
			}
		}
		scalar->phase = phase;
	}
	
}

void lSetInitBoundary(lContext * ctx, lSet * set, lElementType type, const char * boundary_name) {
	int id;
	lDiscoverySetInitRanges(ctx->boundary_reader, type, boundary_name, numberOfOwnedElementsOfType(ctx,type), ctx->local_to_global[(int)type].begin(), &set->ranges, &set->size,NULL,&id); 
}

L_ALWAYS_INLINE 
size_t lSetSize(lContext * ctx, lSet * set) {
	return set->size;
}

L_ALWAYS_INLINE
void lKernelRun(lContext * ctx,  lSet * set, lElementType typ, int id, void (*kernel)(lkContext), lStencilData data) {

#ifndef NDEBUG
	ctx->print_context.push();
#endif

	for(size_t i = 0; i < set->ranges.size(); i++) {
		uint32_t start = set->ranges[i].first;
		uint32_t end = set->ranges[i].second;
		for(size_t j = start; j < end; j++) {

#ifndef NDEBUG
			ctx->print_context.mark(typ, ctx->local_to_global[typ][j]);
#endif
			lkContext kctx;
			kctx.ctx  = ctx;
			kctx.element.data = j;
			kernel(kctx);
		}
	}
	
#ifndef NDEBUG
	ctx->print_context.pop();
#endif

}
/*kernels have the signature and structure: 
void kernel_name(lkContext * ctx) {
	lkElement e;
	if(lkGetActiveElement(ctx,&e)) {
	  <...>
	}
}*/

void lPrintBegin(lContext * ctx) {
#ifndef NDEBUG
	std::cout << ctx->print_context;
#endif
	std::cout << "liszt: ";
}
void lPrintEnd(lContext * ctx) {
	std::cout << "\n";
}

void printValue(lType typ, const void * value) {
	switch(typ) {
		case L_FLOAT:
			std::cout << *(float*)value;
			break;
		case L_INT:
			std::cout << *(int*)value;
			break;
		case L_BOOL:
			std::cout << *(bool*)value;
			break;
		case L_DOUBLE:
			std::cout << *(double*)value;
			break;
		case L_STRING:
			std::cout << (const char *) value;
			break;
	}
}

void lPrintValue(lContext * ctx, lType typ, size_t r, size_t c, const void * value) {
	byte * v = (byte *) value;
	size_t size = lUtilTypeSize(typ);
	if(c != 0) { 
		std::cout << "[";
		for(size_t i = 0; i < r; i++) {
			for (size_t j = 0; j < c; j++) {
				printValue(typ,v);
				v += size;
				if(j < c - 1) std::cout << ",";
			}
			if(i < r - 1) std::cout << "|";
			else std::cout << "]";  
		}
	} else if(r != 0) {
		std::cout << "(";
		for(int i = 0; i < r - 1; i++) {
			printValue(typ,v);
			std::cout << ",";
			v += size;
		}
		printValue(typ,v);
		std::cout << ")";
	} else {
		printValue(typ,v);
	}
}
double lWallTime(lContext * ctx) { //counter, time reported in seconds
	timeval currentTime;
	gettimeofday(&currentTime, 0);
	return currentTime.tv_sec + 1e-6 * currentTime.tv_usec;
}
//Nested Runtime Calls
void lkPrintBegin(lkContext * ctx) {
	lPrintBegin(ctx->ctx);
}
void lkPrintEnd(lkContext * ctx) {
	lPrintEnd(ctx->ctx);
}
void lkPrintValue(lkContext * ctx, lType typ, size_t r, size_t c, const void * value) {
	lPrintValue(ctx->ctx, typ, r, c, value);
}
void lkPrintElement(lkContext * ctx, lElementType typ, lkElement e) {
	switch(typ) {
		case L_VERTEX:
			printf("Vertex %d", lkIDOfVertex(ctx,e));
			break;
		case L_EDGE:
			printf("Edge %d", lkIDOfEdge(ctx,e));
			break;
		case L_FACE:
			printf("Face %d", lkIDOfFace(ctx,e));
			break;
		case L_CELL:
			printf("Cell %d", lkIDOfCell(ctx,e));
			break;
		default:
			abort();
	}
	//if(elementID(e) >= numberOfOwnedElementsOfType(ctx->ctx,typ))
	//	printf("(g)");
}

L_ALWAYS_INLINE 
lkFields * lkGetFields(lkContext * ctx) {
	return reinterpret_cast<lkFields*>(ctx->ctx->fields);
}

L_ALWAYS_INLINE 
lkScalars * lkGetScalars(lkContext * ctx) {
	return reinterpret_cast<lkScalars*>(ctx->ctx->scalars);
}

L_ALWAYS_INLINE 
bool lkGetActiveElement(lkContext * ctx, lkElement * e) {
	*e = ctx->element;
	return true;
}

void lkFieldRead(lkField * f, lkElement e, lType element_type, size_t element_length, size_t val_offset, size_t val_length,  void * result) {
	size_t offset,size;
	lUtilValueLocation(elementID(e),element_type,element_length,val_offset,val_length,&offset,&size);
	memcpy(result,f->data + offset, size);
}

L_ALWAYS_INLINE 
void lkFieldWrite(lkField * field, lkElement e, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value) {
	uint32_t id = elementID(e);
	uint32_t adjusted_id = (id >= field->first_ghost) ? (id + field->write_ghosts_offset) : id; 
	size_t offset,size;
	lUtilValueLocation(adjusted_id,element_type,element_length,val_offset,val_length,&offset,&size);
	lUtilValueReduce(field->data + offset, reduction, element_type, val_length, value);
}

L_ALWAYS_INLINE
void lkScalarRead(lkContext * ctx, lkScalar * scalar, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * result){
	lScalarRead(ctx->ctx,(lScalar*)scalar,element_type, element_length, val_offset, val_length, result);
}

L_ALWAYS_INLINE
void lkScalarWrite(lkContext * ctx, lkScalar * scalar, lReduction reduction, lType element_type, size_t element_length, size_t val_offset, size_t val_length, void * value) {
	size_t scalar_size = lUtilTypeSize(element_type);
	lUtilValueReduce(scalar->write_entry + val_offset * scalar_size, reduction, element_type, val_length, value);
}

L_ALWAYS_INLINE 
size_t lkSetSize(lkContext * ctx, lkSet * set) {
	return set->set.size;
}
L_ALWAYS_INLINE
void lkSetGetIterator(lkContext * ctx, lkSet * set, lElementType typ, lkIterator * iterator) {
	
	lNestedSetGetIterator(&set->set,&iterator->it);
	
#ifndef NDEBUG
	ctx->ctx->print_context.push();
#endif
}

L_ALWAYS_INLINE
bool lkIteratorNext(lkContext * ctx, lkIterator * iterator, lElementType typ, lkElement * e, int * lbl) {
	bool result = lNestedIteratorNext(&iterator->it,e,lbl);
	
		
#ifndef NDEBUG
	if(result) {
		ctx->ctx->print_context.mark(typ,  ctx->ctx->local_to_global[typ][elementID(*e)]);
	} else {
		ctx->ctx->print_context.pop();
	}
#endif
	return result;
}

//now all the topological functions ....
void lVerticesOfMesh(lContext * ctx, lSet * set) {
	set->size = numberOfGlobalElementsOfType(ctx,L_VERTEX);
	set->ranges.push_back(std::make_pair(0, numberOfOwnedElementsOfType(ctx,L_VERTEX)));
}
void lEdgesOfMesh(lContext * ctx, lSet * set) {
	set->size = numberOfGlobalElementsOfType(ctx,L_EDGE);
	set->ranges.push_back(std::make_pair(0, numberOfOwnedElementsOfType(ctx,L_EDGE)));
}
void lFacesOfMesh(lContext * ctx, lSet * set) {
	set->size = numberOfGlobalElementsOfType(ctx,L_FACE);
	set->ranges.push_back(std::make_pair(0, numberOfOwnedElementsOfType(ctx,L_FACE)));
}
void lCellsOfMesh(lContext * ctx, lSet * set) {
	set->size = numberOfGlobalElementsOfType(ctx,L_CELL) - 1;
	set->ranges.push_back(std::make_pair(1, numberOfOwnedElementsOfType(ctx,L_CELL)));
}

int lkIDOfVertex(lkContext * ctx, lkElement e) {
	return ctx->ctx->local_to_global[L_VERTEX][elementID(e)];
}
int lkIDOfEdge(lkContext * ctx, lkElement e) {
	return ctx->ctx->local_to_global[L_EDGE][elementID(e)];
}
int lkIDOfFace(lkContext * ctx, lkElement e) {
	return ctx->ctx->local_to_global[L_FACE][elementID(e)];
}
int lkIDOfCell(lkContext * ctx, lkElement e) {
	return ctx->ctx->local_to_global[L_CELL][elementID(e)];
}

#define NESTED_TOPOLOGY_GET_MESH(ctx) (&ctx->ctx->mesh)
#define NESTED_TOPOLOGY_GET_SET(set) (set->set)
#include "nested_topology_functions.inc"

lDiscoveryElement * lDiscoveryGetElement(lsElement * e) { return &e->d; }
lDiscoverySet * lDiscoveryGetSet(lsSet * e) { return &e->d; }
lDiscoveryIterator * lDiscoveryGetIterator(lsIterator * e) { return &e->d; }
lDiscoveryContext * lDiscoveryGetContext(lsContext * e) { return &e->d; }

#include "vector.h"
#include "matrix.h"
#include "liszt_math.h"

//code generated from the liszt compiler is now included here after runtime-specific
//headers have been included

#include "generated.cpp"

