#include <cstdlib>
#include <cstdio>
#include <vector>
#include <iostream>
#include <cassert>
#include <cmath>

using namespace std;

enum axis {AXIS_X=0, AXIS_Y=1, AXIS_Z=2};
enum genstate {INLET,OUTLET,BOUNDARY,INTERIOR};
struct face;

struct cell {
	cell() : ooo(-1), loo(-1), olo(-1), llo(-1), ool(-1), lol(-1), oll(-1), lll(-1), fx0(-1), fx1(-1), fy0(-1), fy1(-1), fz0(-1), fz1(-1) {};
	int ooo, loo, olo, llo, ool, lol, oll, lll;
	int fx0,fx1;
	int fy0,fy1;
	int fz0,fz1;
	int id;
};
struct face {
	face() : ll(-1), lr(-1), ul(-1), ur(-1) {};
	int ll, lr, ul, ur;
};
struct vertex {
	int x, y, z;
	vertex() : x(-1), y(-1), z(-1) {}
	vertex(int _x, int _y, int _z) : x(_x), y(_y), z(_z) {}
	void print() { cout << "(" << x << "," << y << "," << z << ")"; }
};

struct mesh {
	int cn[3]; //cells along each axis
	int vn[3]; //vertices along each axis
	int nfaces_a[3]; //faces along each axis
	int ncells, nverts, nfaces;
	FILE* fp;
	
	mesh(int _cx, int _cy, int _cz) : 
		ncells(_cx*_cy*_cz), fp(NULL) {
			cn[0] = _cx;
			cn[1] = _cy;
			cn[2] = _cz;
			vn[0] = _cx + 1;
			vn[1] = _cy + 1;
			vn[2] = _cz + 1;
			nverts = (_cx + 1)*(_cy + 1)*(_cz + 1);
			nfaces_a[0] = (cn[0]+1)*cn[1]*cn[2];
			nfaces_a[1] = cn[0]*(cn[1]+1)*cn[2];
			nfaces_a[2] = cn[0]*cn[1]*(cn[2]+1);
			nfaces = nfaces_a[0] + nfaces_a[1] + nfaces_a[2];
	}
	
	int vertex_position_to_id(vertex v) {
		assert (v.x >= 0 && v.x < vn[0]);
		assert (v.y >= 0 && v.y < vn[1]);
		assert (v.z >= 0 && v.z < vn[2]);
		return v.x + vn[0]*v.y + vn[0]*vn[1]*v.z;
	}
	vertex vertex_id_to_vertex(int id) {
		int x = id%(vn[0]);
		int y = (id/vn[0])%vn[1];
		int z = id/(vn[0]*vn[1]);		
		return vertex(x, y, z);
	}
	
	//faces are uniquely identified by their lower-left vertex and their perpendicular axis.
	int face_id_from_ll(vertex ll, axis a) {
		switch (a) {
			case AXIS_X:
				return ll.x + (cn[0]+1)*ll.y + (cn[0]+1)*cn[1]*ll.z;
			case AXIS_Y:
				return ll.y + (cn[1]+1)*ll.x + (cn[1]+1)*cn[0]*ll.z + nfaces_a[0];
			case AXIS_Z:
				return ll.z + (cn[2]+1)*ll.x + (cn[2]+1)*cn[0]*ll.y + nfaces_a[0] + nfaces_a[1];
		}
		return -1;
	}
	
	face face_from_point(float x, float y, float z, axis a) {
		face f;
		if (a == AXIS_X) {
			f.ll = vertex_position_to_id(vertex((int)x, (int)floor(y), (int)floor(z)));
			f.lr = vertex_position_to_id(vertex((int)x, (int)floor(y), (int)ceil(z)));
			f.ul = vertex_position_to_id(vertex((int)x, (int)ceil(y),  (int)floor(z)));
			f.ur = vertex_position_to_id(vertex((int)x, (int)ceil(y),  (int)ceil(z)));
		} else if (a == AXIS_Y) {
			f.ll = vertex_position_to_id(vertex((int)floor(x), (int)y,  (int)floor(z)));
			f.ul = vertex_position_to_id(vertex((int)floor(x), (int)y,  (int)ceil(z)));
			f.lr = vertex_position_to_id(vertex((int)ceil(x),  (int)y,  (int)floor(z)));
			f.ur = vertex_position_to_id(vertex((int)ceil(x),  (int)y,  (int)ceil(z)));			
		} else if (a == AXIS_Z) {
			f.lr = vertex_position_to_id(vertex((int)floor(x), (int)floor(y), (int)z));
			f.ll = vertex_position_to_id(vertex((int)ceil(x),  (int)floor(y), (int)z));
			f.ur = vertex_position_to_id(vertex((int)floor(x), (int)ceil(y),  (int)z));
			f.ul = vertex_position_to_id(vertex((int)ceil(x),  (int)ceil(y),  (int)z));
		}
		return f;
	}
	
	cell cell_from_point(float x, float y, float z) { //given point in cell, find cell.
		if (x < 0 || x > cn[0] ||  y < 0 || y > cn[1] || z < 0 || z > cn[2]) {
			cell c;
			c.id = 0; //outside
			return c;
		} else {
			cell c;
			c.id = 1;
			c.ooo = vertex_position_to_id(vertex((int)floor(x), (int)floor(y), (int)floor(z)));
			c.loo = vertex_position_to_id(vertex((int)ceil(x),  (int)floor(y), (int)floor(z)));
			c.olo = vertex_position_to_id(vertex((int)floor(x), (int)ceil(y),  (int)floor(z)));
			c.llo = vertex_position_to_id(vertex((int)ceil(x),  (int)ceil(y),  (int)floor(z)));
			c.ool = vertex_position_to_id(vertex((int)floor(x), (int)floor(y), (int)ceil(z)));
			c.lol = vertex_position_to_id(vertex((int)ceil(x),  (int)floor(y), (int)ceil(z)));
			c.oll = vertex_position_to_id(vertex((int)floor(x), (int)ceil(y),  (int)ceil(z)));
			c.lll = vertex_position_to_id(vertex((int)ceil(x),  (int)ceil(y),  (int)ceil(z)));
			return c;
		}
	}
	int cell_id_from_cell(cell c) {
		if (c.id == 0)
			return 0;
		vertex ooo = vertex_id_to_vertex(c.ooo);
		return 1 + ooo.x + ooo.y*cn[0] + ooo.z*cn[0]*cn[1];
	}
	
	void print_cell(cell & c) {
		if (c.id == 0) {
			cout << "Outside cell" << endl;
		} else {
		cout << "Cell with vertices: " << endl;
		cout << "  "; vertex_id_to_vertex(c.ooo).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(c.loo).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(c.olo).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(c.llo).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(c.ool).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(c.lol).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(c.oll).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(c.lll).print(); cout << endl;
			
		}
	}
	void print_face(face & f) {
		cout << "Face with vertices: " << endl;
		cout << "  "; vertex_id_to_vertex(f.ll).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(f.lr).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(f.ur).print(); cout << endl;
		cout << "  "; vertex_id_to_vertex(f.ul).print(); cout << endl;

	}
	
	//Filters to generate different types of sets.
	struct face_filter {
		virtual bool operator()(int x,int y,int z,int a, int cbid, int ciid) { return true; }
	};

	struct inlet_face_filter : public face_filter {
		virtual bool operator()(int x,int y,int z, int a, int cbid, int ciid) { return (a == 0 && cbid == 0); }
	};
	struct outlet_face_filter : public face_filter {
		virtual bool operator()(int x,int y,int z,int a, int cbid, int ciid) { return (a == 0 && ciid == 0); }
	};
	struct interior_face_filter : public face_filter {
		virtual bool operator()(int x,int y,int z,int a, int cbid, int ciid) { return (ciid != 0 && cbid != 0); }
	};
	struct far_field_face_filter : public face_filter {
		virtual bool operator()(int x,int y,int z,int a, int cbid, int ciid) { return (a != 0 && (cbid == 0 || ciid == 0)); }
	};
	
	void gen_vertices() {
		int id = 0;
		fprintf(fp, "(0 \"Vertices:\")\n");
		fprintf(fp, "(10 (%d %x %x %d %d))\n", 0, 1, nverts, 1, 3);             
		fprintf(fp, "(10 (%d %x %x %d %d)(\n", 1, 1, nverts, 1, 3);
		for (int z = 0.0; z < vn[2]; z++) {
				for (int y = 0.0; y < vn[1]; y++) {
						for (int x = 0.0; x < vn[0]; x++) {
								vertex v(x, y, z);
								//vertices[id] = v;
								fprintf(fp, "  %e %e %e\n", (float)x, (float)y, (float)z);
								id++;
						}
				}
		}
		fprintf(fp, "))\n\n");
	}
	
	int gen_faces_for_filter(face_filter & filter, int zone_id, int first_index, char* bc_type, bool orient_outside) {
		int total_faces = 0;
		for (int stage = 0; stage < 2; stage++) {
		for (int a = 0; a < 3; a++) {
			//cout << "Generating " << nfaces_a[a] << " faces along " << a << " axis." << endl;			
			for (float i = 0; i < vn[a]; i+=1) { //the plane we are sweeping through
				for (float j = 0.5; j < cn[(a+2)%3]; j+= 1.0) { 
					for (float k = 0.5; k < cn[(a+1)%3]; k+= 1.0) {
						float ijk[3] = {i,j,k};
						float x = ijk[a];
						float y = ijk[(a+2)%3];
						float z = ijk[(a+1)%3];
						face f = face_from_point(x, y, z, (axis) a);
						vertex ll = vertex_id_to_vertex(f.ll);
						//int fid_fv = face_id_from_ll(ll, (axis) a);

						//get inner and outer cell
						float ijk_behind[3] = {i-0.5,j,k};
						float ijk_infront[3] = {i+0.5,j,k};
						cell behind = cell_from_point(ijk_behind[a], ijk_behind[(a+2)%3], ijk_behind[(a+1)%3]);
						cell infront = cell_from_point(ijk_infront[a], ijk_infront[(a+2)%3], ijk_infront[(a+1)%3]);
						int cbid = cell_id_from_cell(behind);
						int ciid = cell_id_from_cell(infront);
						if (filter(x,y,z,a,cbid,ciid)) {	
							if (stage == 0) {
								total_faces++;
							} else {
								fprintf(fp, "   %d %x %x %x %x %x %x\n", 4, f.ll+1, f.ul+1, f.ur+1, f.lr+1, ciid, cbid);
							}
						}
					}
				}
			}	
		}
		if (stage == 0) {
			cout << "  Faces in this set: " << total_faces << endl;
			fprintf(fp, "(13 (%d %x %x %s %d)(\n", zone_id, first_index, first_index+total_faces-1, bc_type, 0);
		} else {
			fprintf(fp, "))\n");
		}
		}
		return total_faces;
	}
	void gen(FILE* _fp) {
		fp = _fp;
		cout << "GridGen creating grid with " << nverts << " vertices, " << nfaces << " faces and " << ncells << " cells arranged as (" << cn[0] << "," << cn[1] << "," << cn[2] << ") cells." << endl;	
	
		fprintf(fp, "(0 \"Liszt Grid Generator\")\n\n");
		fprintf(fp,"(0 \"Dimensions:\")\n");
		fprintf(fp, "(%d %d)\n\n", 2, 3);
		
		gen_vertices();

		/////// FACES
		
		fprintf(fp, "(0 \"Faces:\")\n");
		fprintf(fp, "(13 (%d %x %x %d))\n", 0, 1, nfaces, 0);
		inlet_face_filter iff;
		outlet_face_filter off;
		far_field_face_filter far_fff;
		interior_face_filter interior_ff;
		
		int first_index = 1;
		cout << "Generating x-outlet..." << endl;
		first_index += gen_faces_for_filter(off, 3, first_index, "24", false);
		cout << "Generating x-inlet..." << endl;
		first_index += gen_faces_for_filter(iff, 4, first_index, "a", false);
		cout << "Generating far_field..." << endl;
		first_index += gen_faces_for_filter(far_fff, 5, first_index, "9", false);
		cout << "Generating interior..." << endl;
		first_index += gen_faces_for_filter(interior_ff, 7, first_index, "2", false);
		
		fprintf(fp, "\n");
		fprintf(fp, "(0 \"Cells:\")\n");
		fprintf(fp, "(12 (%d %x %x %d))\n", 0, 1, ncells, 0);
		fprintf(fp, "(12 (%d %x %x %d %d))\n\n", 2, 1, ncells, 1, 4);			

		fprintf(fp, "(0 \"Zones:\")\n");
		fprintf(fp, "(45 (%d %s %s)())\n", 2, "fluid", "fluid");
		fprintf(fp, "(45 (%d %s %s)())\n", 3, "outflow", "outlet");
		fprintf(fp, "(45 (%d %s %s)())\n", 4, "velocity-inlet", "inlet");
		fprintf(fp, "(45 (%d %s %s)())\n", 5, "pressure-far-field", "far_field");
		fprintf(fp, "(45 (%d %s %s)())\n", 7, "interior", "default-interior");
			
		
	}
};

int main(int argc, char* argv[]) {
	if (argc != 5) {
		cout << "Usage:\n  gridgen out.msh cx cy cz" << endl;
		exit(1);
	}
	char* filename = argv[1];
	int cx = atoi(argv[2]);
	int cy = atoi(argv[3]);
	int cz = atoi(argv[4]);
	
	mesh m(cx, cy, cz);
	FILE* fp = fopen(filename, "w");
	m.gen(fp);
	fprintf(fp,"\n");
	fclose(fp);
	
}
