/*
 *  Encoding.h
 *  lisztvis
 *
 *  Created by Zach Devito on 11/20/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */
#ifndef _ENCODING_H
#define _ENCODING_H

#include "interpret.h"
#include <map>
#include "Filter.h"
#import "Glue.h"

static const int MAX_CATEGORY_SIZE = 11;

class Encoding {
public:
	void init(ElemType e, Filter * f) {
		std::cout << "INIT: " << e << std::endl;
		elem_type = e;
		filter = f;
		new (&encodingExp) ExpRef();
		new (&category_map) category_map_type();
		calcStats();
	}
	bool setExp(const std::string & expstr) {
		ExpRef exp = interpret_compile(expstr);
		if(exp->isError()) {
			encodingExp.reset();
			calcStats();
			return false;
		} else {
			encodingExp = exp;
			calcStats();	
			return true;
		}
	}
	bool isActive() const {
		return encodingExp.get() != NULL;
	}
	template<typename MO>
	IObject operator()(const MO & mo) const {
		assert(MO::type::value == elem_type);
		return valueForID(mo.ID());
	}
	template<typename MO>
	IObject linearScale(const MO & mo) const {
		IObject zero(0.);
		IObject v = (*this)(mo);
		return (v - minv) / (maxv - minv + zero);
	}
	template<typename MO>
	IObject zscore(const MO & mo) const {
		IObject v = (*this)(mo);
		return (v - mean) / stddev;
	}
	template<typename MO>
	id_type category(const MO & mo) const {
		assert(isCategorical);
		IObject v = (*this)(mo);
		assert(category_map.count(v));
		category_map_type::const_iterator it = category_map.find(v);
		assert(it != category_map.end());
		return it->second;
	}
	size_t numCategories() const {
		assert(categorical());
		return category_map.size();
	}
	bool categorical() const {
		return isCategorical;
	}
	void createLabels(size_t requested_size, std::vector<std::string> * lbls, bool * isCat) {
		lbls->clear();
		if(categorical()) {
			*isCat = true;
			lbls->insert(lbls->end(),category_map.size(),"");
			for(category_map_type::const_iterator it = category_map.begin(), end = category_map.end();
				it != end;
				it++) {
				std::ostringstream ss;
				ss << " " << it->first;
				(*lbls)[it->second] = ss.str();
			}
		} else {
			*isCat = false;
			double min = minv.doubleValue(0.);
			double max = maxv.doubleValue(1.);
			double interval = (max - min) / requested_size;
			for(size_t i = 0; i < requested_size; i++) {
				std::ostringstream ss;
				ss << " " << min + i * interval << " to " << min + (i+1) * interval;
				lbls->push_back(ss.str());
			}
		}
	}
	void calcStats() {
		category_map.clear();
		
		if(isActive()) {
			size_t nElems = numElems();
			bool first = true;
			isCategorical = true;
			size_t nSelected = 0;
			for(id_type i = 0; i < nElems; i++) {
				IObject v = valueForID(i);
				if(v.isUnknown()) {
					setExp("");
					return; //some field has an error, we are erroring out
				}
				if( filter->shouldDisplay(i) && !v.isUnknown()) {
					if(first && !v.isNaN()) {
						mean = minv = maxv = v;
						stddev = mean * mean;
						first = false;
					} else {
						maxv = IObjectLib::max(maxv,v);
						minv = IObjectLib::min(minv,v);
						mean = mean + v;
						stddev = stddev + v * v;
					}
					nSelected++;
				}
				if(v.getType() == TDOUBLE || v.isVec())
					isCategorical = false;
				size_t cats = category_map.size();
				if(category_map.size() <= MAX_CATEGORY_SIZE) {
					category_map_type::const_iterator it = category_map.find(v);
					if(it == category_map.end()) {
						category_map[v] = cats;
					}
				}
			}
			cout << category_map.size() << endl;
			IObject nElemso = IObject((int)nSelected);
			mean = mean / nElemso;
			stddev = stddev / nElemso - mean * mean;
			if(category_map.size() > MAX_CATEGORY_SIZE)
				isCategorical = false;
		} else {
			isCategorical = true;
			category_map[IObject()] = 0;
		}
		//renumber categories in sorted order
		if(categorical()) {
			size_t i = 0;
			for(category_map_type::iterator it = category_map.begin(), end = category_map.end();
			    it != end;
			    it++) {
				it->second = i;
				i++;
			}
		}
		cout << "is categorical? " << categorical() << endl;
		
	}
	ElemType getElemType() {
		return elem_type;
	}
private:
	typedef std::map<IObject,id_type,IObjectLib> category_map_type;
	
	IObject valueForID(id_type idx) const {
		if(isActive()) {
			return getValueForExpWithID(encodingExp,idx);
		} else {
			return IObject();
		}
	}
	size_t numElems() {
		switch(elem_type) {
			case VERTEX:
				return mesh()->vertices().size();
			case EDGE:
				return mesh()->edges().size();
			case FACE:
				return mesh()->faces().size();
			case CELL:
				return mesh()->cells().size();
			default:
				assert(false);
				return 0;
		}
	}
	IObject mean,stddev,minv,maxv;
	ElemType elem_type;
	bool isCategorical;
	ExpRef encodingExp;
	category_map_type category_map;
	Filter * filter;
};
#endif