package ScalarConvection
import Liszt.Builders._
import Liszt.MetaInteger._
import Liszt.Types._
import Liszt.Exps._
import Liszt.Language._
import Liszt.StdLib._
object ScalarConvection extends LisztApplication {
  val face_set = Var of loadFieldWithConst[FaceB,IntB](0,"face_sets")
  val position = Var of loadFieldWithURL[VertexB,float3]("mesh://position","position")
  val interior_set = Var of loadSet[FaceB]("default-interior")
  val inlet_set = Var of loadSet[FaceB]("inlet")
  val outlet_set = Var of loadSet[FaceB]("outlet")
  val far_field_set = Var of loadSet[FaceB]("far_field")
  val float3_zero : LValueB[float3] = Var of Vec(0.f,0.f,0.f)

  val faceCenter = Var of loadFieldWithConst[FaceB,float3](float3_zero,"faceCenter")
  val faceArea = Var of loadFieldWithConst[FaceB,FloatB](0.f,"faceArea")

  val Phi = Var of loadFieldWithConst[CellB,FloatB](0.f,"Phi")
  val Flux = Var of loadFieldWithConst[CellB,FloatB](0.f,"Flux")
  val globalVelocity = Var of Vec(1.f,0.f,0.f)


  //some geometry fields
  val face_centroid = Var of loadFieldWithConst[FaceB,float3](float3_zero,"face_centroid")
  val face_area = Var of loadFieldWithConst[FaceB,FloatB](0.f,"face_area")
  val face_normal = Var of loadFieldWithConst[FaceB,float3](float3_zero,"face_normal")
  val face_unit_normal = Var of loadFieldWithConst[FaceB,float3](float3_zero,"face_unit_normal")

  val cell_centroid = Var of loadFieldWithConst[CellB,float3](float3_zero,"cell_centroid")
  val cell_volume = Var of loadFieldWithConst[CellB,FloatB](0.f,"cell_volume")
  //some geometry functions
  val calcFaceCenter : (FaceB) => float3
  = lisztfn {
    (f : Arg[FaceB]) =>
    val center = Var of Vec(0.f,0.f,0.f)
    val test = Var of Vec(0.f,0.f,0.f,0.f)
    for(v <- vertices(f)) {
      center += position(v)
    }
    center := center / itof(size(vertices(f)))
    Return(center -> calcFaceCenter)
  }
  val calcCellCenter : (CellB) => float3
  = lisztfn {
    (c : Arg[CellB]) =>
    val center = Var of Vec(0.f,0.f,0.f)
    for(v <- vertices(c)) {
      center += position(v)
    }
    center := center / size(vertices(c))
    Return(center -> calcCellCenter)
  }
  val calcFaceGeom = lisztfn {
    (f : Arg[FaceB]) =>
    val approxCenter = Var of calcFaceCenter(f)
    val normal = Var of Vec(0.f,0.f,0.f)
    for(e <- edgesCCW(f)) {
      val v0 = Var of position(head(e)) - approxCenter
      val v1 = Var of position(tail(e)) - approxCenter
      normal += cross(v1,v0)
    }
    normal := normalize(normal)
    val center = Var of Vec(0.f,0.f,0.f)
    val area = Var of 0.f
    for(e <- edgesCCW(f)) {
      val v0 = Var of position(head(e)) - approxCenter
      val v1 = Var of position(tail(e)) - approxCenter
      val tmp_area = Var of dot(normal,cross(v1,v0))
      area += tmp_area
      center += tmp_area * ( approxCenter + position(head(e)) + position(tail(e)))
    }  
    face_centroid(f) := center / (3.f * area)
    face_area(f) := area / 2.f
    face_normal(f) := normal*face_area(f)
    face_unit_normal(f) := normal
  }

  val calcCellGeom = lisztfn {
    (c : Arg[CellB]) =>
    val approxCenter = Var of calcCellCenter(c)
    val volume = Var of 0.f
    val center = Var of Vec(0.f,0.f,0.f)
    for(f <- faces(c)) {
      val v0 = Var of face_centroid(f) - approxCenter
      for(e <- edgesCCW(towards(f,c))) {
        val v1 = Var of position(head(e)) - approxCenter
        val v2 = Var of position(tail(e)) - approxCenter
        val tetVol = Var of dot(v0,cross(v1,v2))
        volume += tetVol
        center += tetVol * ( approxCenter + face_centroid(f) + position(head(e)) + position(tail(e)))
      }
    }
    cell_centroid(c) := center / (volume * 4.f)
    cell_volume(c) := volume / 6.f
  }
  val phi_sine_function : (FloatB) => FloatB
  = lisztfn {
    (t : Arg[FloatB]) =>
      Return(10.f.C * sin(t*2.f.C*MATH_PI) -> phi_sine_function)
  }
  def main() {
    for(f <- interior_set) { face_set(f) := 1 }
    for(f <- outlet_set) { face_set(f) := 2 }
    for(f <- inlet_set) { face_set(f) := 3 }
    for(f <- far_field_set) { face_set(f) := 4 }
    //initialize geometry fields
    for(f <- faces(mesh)) {
      If(ID(outside(f)) < ID(inside(f))) {
        calcFaceGeom(flip(f))
      } Else {
        calcFaceGeom(f)
      }
    }
    for(f <- faces(mesh)) {
      Print(ID(f),"FaceArea: ",face_area(f),"normal: ",face_unit_normal(f),"face_centroid",face_centroid(f))
    }
    for(c <- cells(mesh)) {
      calcCellGeom(c)
    }
    for(c <- cells(mesh)) {
      Print("c: ",ID(c)," ",cell_volume(c)," ",cell_centroid(c))
    }
    //calculate bounding box
    val maxf : Float = java.lang.Float.MAX_VALUE
    val minf : Float = java.lang.Float.MIN_VALUE
    val ll = Var of Vec(maxf,maxf,maxf)
    val ur = Var of Vec(minf,minf,minf)

    for(v <- vertices(mesh)) {
      ll minWith position(v)
      ur maxWith position(v)
    }
    val mesh_center = Var of .5f.C * (ll + ur)
    for(c <- cells(mesh)) {
      val center = Var of cell_centroid(c)
      val x = Var of normal_pdf(center.x - mesh_center.x)
      val y = Var of normal_pdf(center.y - mesh_center.y)
      val z = Var of normal_pdf(center.z - mesh_center.z)
      Phi(c) := x * y * z
    }
    val deltat = Var of .015f
    val t = Var of 0.f
    for(c <- cells(mesh)) {
      Print("before cell number: ",ID(c)," -> phi value: ",Phi(c))
    }
    While(t < 2.f) {

      for(f <- interior_set) {
        val normal = Var of face_unit_normal(f)
        val vDotN = Var of dot(globalVelocity,normal)
        val area = Var of face_area(f)
        val flux = Var of 0.f
          If(vDotN >= 0.f) {
            flux := area * vDotN * Phi(inside(f))
          } Else {
            flux := area * vDotN * Phi(outside(f))
          }

        Flux(inside(f)) -= flux
        Flux(outside(f)) += flux
      }
      for(f <- outlet_set) {
        val normal = Var of face_unit_normal(f)
        If(ID(outside(f)) =? 0) {
          Flux(inside(f)) -= face_area(f) * dot(normal,globalVelocity) * Phi(inside(f))
        } Else {
          Flux(outside(f)) -= face_area(f) * dot(-normal,globalVelocity) * Phi(outside(f))
        }
      }
      for(f <- inlet_set) {
        val area = Var of face_area(f)
        val vDotN = Var of dot(globalVelocity,face_unit_normal(f))
        Flux(outside(f)) += area * vDotN * phi_sine_function(t)
      }
      for(f <- far_field_set) {
         val normal = Var of face_unit_normal(f)
         If(ID(outside(f)) =? 0) {
           Flux(inside(f)) -= dot(normal,globalVelocity) * face_area(f) * Phi(inside(f))
         } Else {
           Flux(outside(f)) -= dot(-normal,globalVelocity) * face_area(f) * Phi(outside(f))
         }
      }

      //TODO(zach): some more loops for boundary conditions go here
      for(c <- cells(mesh)) {
        Phi(c) += deltat * Flux(c) / cell_volume(c)
        Flux(c) := 0.f
      }
      t += deltat
    }
    for(c <- cells(mesh)) {
      Print("cell number: ",ID(c)," -> phi value: ",Phi(c))
    }
  }
} //Scalar Convection
