#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cctype>
#include <stdint.h>

bool USE_DOUBLES;

enum TokenType { TOK_DBL, TOK_FLT, TOK_INT, TOK_CHAR, TOK_EOF };
struct Token {
    TokenType type;
    float flt;
    double dbl;
    int integer;
    char chr;
};

bool compareDoubles(double A, double B, uint64_t maxUlps) {
    union {
        double dbl;
        uint64_t longlong;
    } pun;
    pun.dbl = A;
    uint64_t aInt = pun.longlong;
    if (aInt < 0)
        aInt = 0x8000000000000000LL - aInt;
    pun.dbl = B;
    uint64_t bInt = pun.longlong;
    if (bInt < 0)
        bInt = 0x8000000000000000LL - bInt;
    uint64_t intDiff = abs(aInt - bInt);
    if (intDiff <= maxUlps)
        return true;
    return false;
}

bool compareFloats(float A, float B, uint32_t maxUlps) {
    union {
        float flt;
        int integer;
    } pun;
    pun.flt = A;
    int aInt = pun.integer;
    if (aInt < 0)
        aInt = 0x80000000 - aInt;
    pun.flt = B;
    int bInt = pun.integer;
    if (bInt < 0)
        bInt = 0x80000000 - bInt;
    uint32_t intDiff = abs(aInt - bInt);
    if (intDiff <= maxUlps)
        return true;
    printf("floats differ by %d units.\n",intDiff);
    return false;
}

bool compareInts(int a, int b, int fuzziness) {
    return (b == a);
}

void promoteTo(Token * tok1, TokenType typ) {
    tok1->type = typ;
    if(typ == TOK_FLT) {
        tok1->flt = (float) tok1->integer;
    } else if(typ == TOK_DBL) {
        tok1->dbl = (double) tok1->integer;
    }
}
void checkPromote(Token * tok1, Token * tok2) {
    if(tok1->type == TOK_INT && (tok2->type == TOK_FLT || tok2->type == TOK_DBL)) {
        promoteTo(tok1,tok2->type);
    }
}
//returns true if tokens are equal
bool compareTokens(Token* tok1, Token* tok2, int fuzziness) {
    checkPromote(tok1,tok2);
    checkPromote(tok2,tok1);
    if (tok1->type != tok2->type)
        return false;
    switch (tok1->type) {
        case TOK_DBL:
        return compareDoubles(tok1->dbl, tok2->dbl, fuzziness);    
        case TOK_FLT:
        return compareFloats(tok1->flt, tok2->flt, fuzziness);
        case TOK_INT:
        return compareInts(tok1->integer, tok2->integer, fuzziness);
        case TOK_CHAR:
        return (tok1->chr == tok2->chr);
        case TOK_EOF:
        return true;
    }
    return true;
}

void printToken(Token* tok, unsigned int indent) {
    for (unsigned int i = 0; i < indent; i++) {
        printf(" ");
    }
    switch (tok->type) {
        case TOK_DBL:
            printf("%.13f\n", tok->dbl); //todo(njoubert): is this how you print doubles?
            break;
        case TOK_FLT:
            printf("%.10f\n", tok->flt);
            break;
        case TOK_INT:
            printf("%d\n", tok->integer);
            break;
        case TOK_CHAR:
            printf("%c (char code %d)\n", tok->chr, (int) tok->chr);
            break;
        case TOK_EOF:
            printf("EOF\n");
            break;
    }
}

//todo(njoubert): still need to handle the -.{digits} case
void parseNumberToken(FILE* fp, int a, Token* tok, bool negative) {
    char buf[61];
    int count = 0;
    int period_count = 0;
    int e_count = 0;
    if (negative) {
        buf[count] = '-';
        count++;
    }
    tok->type = TOK_INT;
    while(count < 60) {
        if(a == '.') {
            period_count++;
            if(period_count > 1)
                break;
        } else if(isdigit(a)) {}
        else if(e_count == 0 && a == 'e') {
        	e_count++;
        	period_count = 1;
        	int temp = getc(fp);
        	if(temp == '-') {
        		buf[count++] = (char) 'e';
        		buf[count++] = (char) '-';
        		a = getc(fp);
        		continue;
        	} else ungetc(temp,fp);

        } else break;
        
        buf[count] = (char) a;
        count++;
        a = getc(fp);
    }
    if(period_count > 0)
        tok->type = (USE_DOUBLES) ? TOK_DBL : TOK_FLT;
    ungetc(a, fp);
    buf[count] = 0;
    if (count >= 60) {
        printf("Buffer Overflow during number parsing!\n");
        exit(1);
    }
    if (tok->type == TOK_FLT) {
        tok->flt = strtod(buf,NULL);
    } else if (tok->type == TOK_DBL) {
        tok->dbl = strtod(buf,NULL);
    } else {
        tok->integer = atoi(buf);
    }
}

void parseToken(FILE* fp, Token* tok) {
    int a = getc(fp);
    if (a == EOF) {
        tok->type = TOK_EOF;
    } else if (isdigit(a) == 0) { //not a number
        int b = getc(fp);
        if (a == '-' && isdigit(b)) {
            parseNumberToken(fp, b, tok, true);
        } else {
            ungetc(b, fp);
            tok->type = TOK_CHAR;
            tok->chr = (char) a;            
        }
    } else {
        parseNumberToken(fp, a, tok, false);
    }
}

int main(int argc, char *argv[]) {
    if (argc != 4 && argc != 5) {
        printf("Usage:  fuzzydiff file1 file2 fuzziness [--dbl]\n\n");
        printf("\tfuzziness - maximum representable floating point numbers allowed between two 'equal' floats.\n");
        printf("\t[--dbl] - Option to use doubles rather than floats\n");
        return 1;
    }
    
    FILE *fp1 = fopen(argv[1], "r");
    FILE *fp2 = fopen(argv[2], "r");
    int fuzziness = atoi(argv[3]);
    USE_DOUBLES = (argc == 5);
    
    Token tok1;
    Token tok2;
    tok1.type = TOK_CHAR;
    tok2.type = TOK_CHAR;
    int linecount = 0;
    while((tok1.type != TOK_EOF) && (tok2.type != TOK_EOF)) {
        parseToken(fp1, &tok1);
        parseToken(fp2, &tok2);
        if (compareTokens(&tok1, &tok2, fuzziness) != true) {
            printf("Line %d, Tokens do not match:\n", linecount);
            printToken(&tok1, 2);
            printToken(&tok2, 2);
            return 1;
        }
        if (tok1.type == TOK_CHAR && tok1.chr == '\n')
            linecount++;
    }
    fclose(fp1);
    fclose(fp2);    
    return 0;
}